from decimal import Decimal

NaN = float("nan")
s_decimal_NaN = Decimal("nan")
s_decimal_0 = Decimal(0)
SECOND = 1
MINUTE = 60
TWELVE_HOURS = MINUTE * 60 * 12
DAY = TWELVE_HOURS * 2

FUNDING_FEE_POLL_INTERVAL = 120
