cdef object c_calculate_bid_ask_ratios_from_base_asset_ratio(double base_asset_amount,
                                                             double quote_asset_amount,
                                                             double price,
                                                             double target_base_asset_ratio,
                                                             double base_asset_range)
