/* Generated by Cython 3.2.4 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "hummingbot/core/cpp/LimitOrder.h",
            "hummingbot/core/cpp/OrderBookEntry.h",
            "hummingbot/core/cpp/PyRef.h"
        ],
        "include_dirs": [
            "hummingbot/core",
            "hummingbot/core/data_type"
        ],
        "language": "c++",
        "name": "hummingbot.strategy.pure_market_making.pure_market_making",
        "sources": [
            "hummingbot/strategy/pure_market_making/pure_market_making.pyx"
        ]
    },
    "module_name": "hummingbot.strategy.pure_market_making.pure_market_making"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CppInitCode */
#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(const U& other) const { return *ptr == other; }
    template<typename U> bool operator !=(const U& other) const { return *ptr != other; }
    template<typename U> bool operator==(const __Pyx_FakeReference<U>& other) const { return *ptr == *other.ptr; }
    template<typename U> bool operator!=(const __Pyx_FakeReference<U>& other) const { return *ptr != *other.ptr; }
  private:
    T *ptr;
};

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#define __PYX_HAVE__hummingbot__strategy__pure_market_making__pure_market_making
#define __PYX_HAVE_API__hummingbot__strategy__pure_market_making__pure_market_making
/* Early includes */
#include <stdint.h>
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <utility>

    #if __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600)
    // move should be defined for these versions of MSVC, but __cplusplus isn't set usefully
    #include <type_traits>

    namespace cython_std {
    template <typename T> typename std::remove_reference<T>::type&& move(T& t) noexcept { return std::move(t); }
    template <typename T> typename std::remove_reference<T>::type&& move(T&& t) noexcept { return std::move(t); }
    }

    #endif
    
#include <unordered_map>
#include <unordered_set>
#include "cpp/PyRef.h"
#include <string.h>
#include <string_view>
#include <string>
#include "../cpp/LimitOrder.h"
#include <set>
#include <vector>
#include "../cpp/OrderBookEntry.h"
#include <stdio.h>

    /* Using NumPy API declarations from "numpy/__init__.cython-30.pxd" */
    
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif (defined(_Complex_I) && !defined(_MSC_VER)) || ((defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) && !defined(__STDC_NO_COMPLEX__) && !defined(_MSC_VER))
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "hummingbot/strategy/pure_market_making/pure_market_making.pyx",
  "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd",
  "<stringsource>",
  "hummingbot/core/clock.pxd",
  "hummingbot/core/event/event_listener.pxd",
  "hummingbot/core/pubsub.pxd",
  "hummingbot/core/time_iterator.pxd",
  "hummingbot/core/data_type/limit_order.pxd",
  "hummingbot/strategy/order_tracker.pxd",
  "hummingbot/core/event/event_reporter.pxd",
  "hummingbot/core/event/event_logger.pxd",
  "hummingbot/core/data_type/order_book_query_result.pxd",
  "cpython/type.pxd",
  "hummingbot/core/data_type/order_book.pxd",
  "hummingbot/core/network_iterator.pxd",
  "hummingbot/connector/connector_base.pxd",
  "hummingbot/connector/exchange_base.pxd",
  "hummingbot/strategy/asset_price_delegate.pxd",
  "hummingbot/strategy/order_book_asset_price_delegate.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* #### Code section: numeric_typedefs ### */

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":744
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
*/
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":745
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
*/
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":746
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * 
*/
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":747
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_uint8      uint8_t
*/
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":749
 * ctypedef npy_int64      int64_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
*/
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":750
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
*/
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":751
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * 
*/
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":752
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_float32    float32_t
*/
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":754
 * ctypedef npy_uint64     uint64_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
*/
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":755
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
*/
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":762
 * ctypedef double complex complex128_t
 * 
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
*/
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":763
 * 
 * ctypedef npy_longlong   longlong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
*/
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":765
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
*/
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":766
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
*/
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":768
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
*/
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":769
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
*/
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":770
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef float complex       cfloat_t
*/
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* #### Code section: complex_type_declarations ### */
/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< long double > __pyx_t_long_double_complex;
  #else
    typedef long double _Complex __pyx_t_long_double_complex;
  #endif
#else
    typedef struct { long double real, imag; } __pyx_t_long_double_complex;
#endif
static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double, long double);

/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_10hummingbot_4core_5clock_Clock;
struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener;
struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub;
struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator;
struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder;
struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker;
struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase;
struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter;
struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger;
struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult;
struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult;
struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook;
struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator;
struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase;
struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase;
struct __pyx_obj_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate;
struct __pyx_obj_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate;
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy;
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start;
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr;
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit;
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr;
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal;
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr;
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr;

/* "hummingbot/core/pubsub.pxd":10
 * from hummingbot.core.event.event_listener cimport EventListener
 * 
 * ctypedef unordered_set[PyRef] EventListenersCollection             # <<<<<<<<<<<<<<
 * ctypedef unordered_set[PyRef].iterator EventListenersIterator
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events
*/
typedef std::unordered_set<PyRef>  __pyx_t_10hummingbot_4core_6pubsub_EventListenersCollection;

/* "hummingbot/core/pubsub.pxd":11
 * 
 * ctypedef unordered_set[PyRef] EventListenersCollection
 * ctypedef unordered_set[PyRef].iterator EventListenersIterator             # <<<<<<<<<<<<<<
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events
 * ctypedef unordered_map[int64_t, EventListenersCollection].iterator EventsIterator
*/
typedef std::unordered_set<PyRef> ::iterator __pyx_t_10hummingbot_4core_6pubsub_EventListenersIterator;

/* "hummingbot/core/pubsub.pxd":12
 * ctypedef unordered_set[PyRef] EventListenersCollection
 * ctypedef unordered_set[PyRef].iterator EventListenersIterator
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events             # <<<<<<<<<<<<<<
 * ctypedef unordered_map[int64_t, EventListenersCollection].iterator EventsIterator
 * ctypedef pair[int64_t, EventListenersCollection] EventsPair
*/
typedef std::unordered_map<int64_t,__pyx_t_10hummingbot_4core_6pubsub_EventListenersCollection>  __pyx_t_10hummingbot_4core_6pubsub_Events;

/* "hummingbot/core/pubsub.pxd":13
 * ctypedef unordered_set[PyRef].iterator EventListenersIterator
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events
 * ctypedef unordered_map[int64_t, EventListenersCollection].iterator EventsIterator             # <<<<<<<<<<<<<<
 * ctypedef pair[int64_t, EventListenersCollection] EventsPair
 * 
*/
typedef std::unordered_map<int64_t,__pyx_t_10hummingbot_4core_6pubsub_EventListenersCollection> ::iterator __pyx_t_10hummingbot_4core_6pubsub_EventsIterator;

/* "hummingbot/core/pubsub.pxd":14
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events
 * ctypedef unordered_map[int64_t, EventListenersCollection].iterator EventsIterator
 * ctypedef pair[int64_t, EventListenersCollection] EventsPair             # <<<<<<<<<<<<<<
 * 
 * 
*/
typedef std::pair<int64_t,__pyx_t_10hummingbot_4core_6pubsub_EventListenersCollection>  __pyx_t_10hummingbot_4core_6pubsub_EventsPair;
struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_buy_with_specific_market;
struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_sell_with_specific_market;

/* "hummingbot/strategy/strategy_base.pxd":54
 *     cdef c_did_complete_sell_order_tracker(self, object order_completed_event)
 * 
 *     cdef str c_buy_with_specific_market(self, object market_trading_pair_tuple, object amount, object order_type = *,             # <<<<<<<<<<<<<<
 *                                         object price = *, double expiration_seconds = *, position_action = *)
 *     cdef str c_sell_with_specific_market(self, object market_trading_pair_tuple, object amount, object order_type = *,
*/
struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_buy_with_specific_market {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  double expiration_seconds;
  PyObject *position_action;
};

/* "hummingbot/strategy/strategy_base.pxd":56
 *     cdef str c_buy_with_specific_market(self, object market_trading_pair_tuple, object amount, object order_type = *,
 *                                         object price = *, double expiration_seconds = *, position_action = *)
 *     cdef str c_sell_with_specific_market(self, object market_trading_pair_tuple, object amount, object order_type = *,             # <<<<<<<<<<<<<<
 *                                          object price = *, double expiration_seconds = *, position_action = *, )
 *     cdef c_cancel_order(self, object market_pair, str order_id)
*/
struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_sell_with_specific_market {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  double expiration_seconds;
  PyObject *position_action;
};
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_buy;
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_sell;
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_quantize_order_amount;

/* "hummingbot/connector/connector_base.pxd":21
 *         public object _balance_asset_limit
 * 
 *     cdef str c_buy(self, str trading_pair, object amount, object order_type=*, object price=*, dict kwargs=*)             # <<<<<<<<<<<<<<
 *     cdef str c_sell(self, str trading_pair, object amount, object order_type=*, object price=*, dict kwargs=*)
 *     cdef c_cancel(self, str trading_pair, str client_order_id)
*/
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_buy {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/connector_base.pxd":22
 * 
 *     cdef str c_buy(self, str trading_pair, object amount, object order_type=*, object price=*, dict kwargs=*)
 *     cdef str c_sell(self, str trading_pair, object amount, object order_type=*, object price=*, dict kwargs=*)             # <<<<<<<<<<<<<<
 *     cdef c_cancel(self, str trading_pair, str client_order_id)
 *     cdef c_stop_tracking_order(self, str order_id)
*/
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_sell {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/connector_base.pxd":31
 *     cdef object c_get_order_size_quantum(self, str trading_pair, object order_size)
 *     cdef object c_quantize_order_price(self, str trading_pair, object price)
 *     cdef object c_quantize_order_amount(self, str trading_pair, object amount, object price=*)             # <<<<<<<<<<<<<<
*/
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_quantize_order_amount {
  int __pyx_n;
  PyObject *price;
};
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_buy;
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_sell;
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_get_fee;

/* "hummingbot/connector/exchange_base.pxd":17
 *         object _mapping_initialization_lock
 * 
 *     cdef str c_buy(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)             # <<<<<<<<<<<<<<
 *     cdef str c_sell(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)
 *     cdef c_cancel(self, str trading_pair, str client_order_id)
*/
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_buy {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/exchange_base.pxd":18
 * 
 *     cdef str c_buy(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)
 *     cdef str c_sell(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)             # <<<<<<<<<<<<<<
 *     cdef c_cancel(self, str trading_pair, str client_order_id)
 *     cdef c_stop_tracking_order(self, str order_id)
*/
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_sell {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/exchange_base.pxd":33
 *     cdef ClientOrderBookQueryResult c_get_price_for_quote_volume(self, str trading_pair, bint is_buy, double volume)
 *     cdef ClientOrderBookQueryResult c_get_price_for_volume(self, str trading_pair, bint is_buy, object volume)
 *     cdef object c_get_fee(             # <<<<<<<<<<<<<<
 *         self,
 *         str base_currency,
*/
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_get_fee {
  int __pyx_n;
  PyObject *is_maker;
};

/* "hummingbot/core/clock.pxd":3
 * # distutils: language=c++
 * 
 * cdef class Clock:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _clock_mode
*/
struct __pyx_obj_10hummingbot_4core_5clock_Clock {
  PyObject_HEAD
  PyObject *_clock_mode;
  double _tick_size;
  double _start_time;
  double _end_time;
  PyObject *_child_iterators;
  PyObject *_current_context;
  double _current_tick;
  int _started;
};


/* "hummingbot/core/event/event_listener.pxd":6
 * 
 * 
 * cdef class EventListener:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object __weakref__
*/
struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener {
  PyObject_HEAD
  struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener *__pyx_vtab;
  PyObject *__weakref__;
  int64_t _current_event_tag;
  struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *_current_event_caller;
};


/* "hummingbot/core/pubsub.pxd":17
 * 
 * 
 * cdef class PubSub:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Events _events
*/
struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub {
  PyObject_HEAD
  struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub *__pyx_vtab;
  __pyx_t_10hummingbot_4core_6pubsub_Events _events;
  PyObject *__weakref__;
};


/* "hummingbot/core/time_iterator.pxd":7
 * 
 * 
 * cdef class TimeIterator(PubSub):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double _current_timestamp
*/
struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator {
  struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub __pyx_base;
  double _current_timestamp;
  struct __pyx_obj_10hummingbot_4core_5clock_Clock *_clock;
};


/* "hummingbot/core/data_type/limit_order.pxd":6
 * 
 * 
 * cdef class LimitOrder:             # <<<<<<<<<<<<<<
 *     cdef:
 *         CPPLimitOrder _cpp_limit_order
*/
struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder {
  PyObject_HEAD
  struct __pyx_vtabstruct_10hummingbot_4core_9data_type_11limit_order_LimitOrder *__pyx_vtab;
  LimitOrder _cpp_limit_order;
};


/* "order_tracker.pxd":7
 * 
 * 
 * cdef class OrderTracker(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict _tracked_limit_orders
*/
struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker {
  struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
  PyObject *_tracked_limit_orders;
  PyObject *_tracked_market_orders;
  PyObject *_order_id_to_market_pair;
  PyObject *_shadow_tracked_limit_orders;
  PyObject *_shadow_order_id_to_market_pair;
  PyObject *_shadow_gc_requests;
  PyObject *_in_flight_cancels;
  PyObject *_in_flight_pending_created;
};


/* "hummingbot/strategy/strategy_base.pxd":8
 * from .order_tracker cimport OrderTracker
 * 
 * cdef class StrategyBase(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         set _sb_markets
*/
struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase {
  struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
  PyObject *_sb_markets;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_create_buy_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_create_sell_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_fill_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_fail_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_cancel_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_expire_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_complete_buy_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_complete_sell_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_complete_funding_payment_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_position_mode_change_success_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_position_mode_change_failure_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_range_position_liquidity_added_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_range_position_liquidity_removed_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_range_position_update_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_range_position_update_failure_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_range_position_fee_collected_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_range_position_closed_listener;
  int _sb_delegate_lock;
  struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *_sb_order_tracker;
};


/* "hummingbot/core/event/event_reporter.pxd":4
 * 
 * 
 * cdef class EventReporter(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str event_source
*/
struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter {
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
  PyObject *event_source;
};


/* "hummingbot/core/event/event_logger.pxd":4
 * 
 * 
 * cdef class EventLogger(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str _event_source
*/
struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger {
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
  PyObject *_event_source;
  PyObject *_logged_events;
  PyObject *_generic_logged_events;
  PyObject *_order_filled_logged_events;
  PyObject *_waiting;
  PyObject *_wait_returns;
};


/* "order_book_query_result.pxd":3
 * # distutils: language=c++
 * 
 * cdef class OrderBookQueryResult:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double query_price
*/
struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult {
  PyObject_HEAD
  double query_price;
  double query_volume;
  double result_price;
  double result_volume;
};


/* "order_book_query_result.pxd":11
 * 
 * 
 * cdef class ClientOrderBookQueryResult:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object query_price
*/
struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult {
  PyObject_HEAD
  PyObject *query_price;
  PyObject *query_volume;
  PyObject *result_price;
  PyObject *result_volume;
};


/* "hummingbot/core/data_type/order_book.pxd":12
 * 
 * 
 * cdef class OrderBook(PubSub):             # <<<<<<<<<<<<<<
 *     cdef set[OrderBookEntry] _bid_book
 *     cdef set[OrderBookEntry] _ask_book
*/
struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook {
  struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub __pyx_base;
  std::set<OrderBookEntry>  _bid_book;
  std::set<OrderBookEntry>  _ask_book;
  int64_t _snapshot_uid;
  int64_t _last_diff_uid;
  double _best_bid;
  double _best_ask;
  double _last_trade_price;
  double _last_applied_trade;
  double _last_trade_price_rest_updated;
  int _dex;
};


/* "hummingbot/core/network_iterator.pxd":6
 * 
 * 
 * cdef class NetworkIterator(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _network_status
*/
struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator {
  struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
  PyObject *_network_status;
  double _last_connected_timestamp;
  double _check_network_interval;
  double _check_network_timeout;
  double _network_error_wait_time;
  PyObject *_check_network_task;
};


/* "hummingbot/connector/connector_base.pxd":5
 * from hummingbot.core.network_iterator cimport NetworkIterator
 * 
 * cdef class ConnectorBase(NetworkIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         EventReporter _event_reporter
*/
struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase {
  struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator __pyx_base;
  struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter *_event_reporter;
  struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger *_event_logger;
  int _trading_required;
  PyObject *_account_available_balances;
  PyObject *_account_balances;
  int _real_time_balance_update;
  PyObject *_in_flight_orders_snapshot;
  double _in_flight_orders_snapshot_timestamp;
  PyObject *_current_trade_fills;
  PyObject *_exchange_order_ids;
  PyObject *_trade_fee_schema;
  PyObject *_trade_volume_metric_collector;
  PyObject *_balance_asset_limit;
};


/* "hummingbot/connector/exchange_base.pxd":10
 * )
 * 
 * cdef class ExchangeBase(ConnectorBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _order_book_tracker
*/
struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase {
  struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase __pyx_base;
  PyObject *_order_book_tracker;
  PyObject *_budget_checker;
  PyObject *_trading_pair_symbol_map;
  PyObject *_mapping_initialization_lock;
};


/* "hummingbot/strategy/asset_price_delegate.pxd":2
 * 
 * cdef class AssetPriceDelegate:             # <<<<<<<<<<<<<<
 *     cdef object c_get_mid_price(self)
*/
struct __pyx_obj_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate {
  PyObject_HEAD
  struct __pyx_vtabstruct_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate *__pyx_vtab;
};


/* "hummingbot/strategy/order_book_asset_price_delegate.pxd":4
 * from .asset_price_delegate cimport AssetPriceDelegate
 * 
 * cdef class OrderBookAssetPriceDelegate(AssetPriceDelegate):             # <<<<<<<<<<<<<<
 *     cdef:
 *         ExchangeBase _market
*/
struct __pyx_obj_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate {
  struct __pyx_obj_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate __pyx_base;
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *_market;
  PyObject *_trading_pair;
};


/* "hummingbot/strategy/pure_market_making/pure_market_making.pxd":8
 * 
 * 
 * cdef class PureMarketMakingStrategy(StrategyBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _market_info
*/
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy {
  struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase __pyx_base;
  PyObject *_market_info;
  PyObject *_bid_spread;
  PyObject *_ask_spread;
  PyObject *_minimum_spread;
  PyObject *_order_amount;
  int _order_levels;
  int _buy_levels;
  int _sell_levels;
  PyObject *_split_order_levels_enabled;
  PyObject *_bid_order_level_spreads;
  PyObject *_ask_order_level_spreads;
  PyObject *_order_level_spread;
  PyObject *_order_level_amount;
  double _order_refresh_time;
  double _max_order_age;
  PyObject *_order_refresh_tolerance_pct;
  double _filled_order_delay;
  int _inventory_skew_enabled;
  PyObject *_inventory_target_base_pct;
  PyObject *_inventory_range_multiplier;
  int _hanging_orders_enabled;
  PyObject *_hanging_orders_tracker;
  int _order_optimization_enabled;
  PyObject *_ask_order_optimization_depth;
  PyObject *_bid_order_optimization_depth;
  int _add_transaction_costs_to_orders;
  PyObject *_asset_price_delegate;
  PyObject *_inventory_cost_price_delegate;
  PyObject *_price_type;
  int _take_if_crossed;
  PyObject *_price_ceiling;
  PyObject *_price_floor;
  int _ping_pong_enabled;
  PyObject *_ping_pong_warning_lines;
  int _hb_app_notification;
  PyObject *_order_override;
  double _cancel_timestamp;
  double _create_timestamp;
  PyObject *_limit_order_type;
  int _all_markets_ready;
  int _filled_buys_balance;
  int _filled_sells_balance;
  double _last_timestamp;
  double _status_report_interval;
  int64_t _logging_options;
  PyObject *_last_own_trade_price;
  int _should_wait_order_cancel_confirmation;
  PyObject *_moving_price_band;
};


/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":696
 *     # ---------------------------------------------------------------
 * 
 *     cdef c_start(self, Clock clock, double timestamp):             # <<<<<<<<<<<<<<
 *         StrategyBase.c_start(self, clock, timestamp)
 *         self._last_timestamp = timestamp
*/
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start {
  PyObject_HEAD
  PyObject *__pyx_v_order_id;
};


/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":707
 *             # make restored order hanging orders
 *             for order_id in restored_order_ids:
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)             # <<<<<<<<<<<<<<
 *                 if order:
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)
*/
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_o;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1177
 *         return True
 * 
 *     cdef c_cancel_active_orders_on_max_age_limit(self):             # <<<<<<<<<<<<<<
 *         """
 *         Cancels active non hanging orders if they are older than max age limit
*/
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit {
  PyObject_HEAD
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self;
};


/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1184
 *             list active_orders = self.active_non_hanging_orders
 * 
 *         if active_orders and any(order_age(o, self._current_timestamp) > self._max_order_age for o in active_orders):             # <<<<<<<<<<<<<<
 *             for order in active_orders:
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
*/
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr {
  PyObject_HEAD
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_o;
};


/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1247
 *                 and len(non_hanging_orders_non_cancelled) == 0)
 * 
 *     cdef c_execute_orders_proposal(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double expiration_seconds = NaN
*/
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal {
  PyObject_HEAD
  PyObject *__pyx_v_ask_order_id;
  PyObject *__pyx_v_bid_order_id;
};


/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1274
 *                 orders_created = True
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))             # <<<<<<<<<<<<<<
 *                     if order:
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
*/
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr {
  PyObject_HEAD
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_o;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1297
 *                 orders_created = True
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))             # <<<<<<<<<<<<<<
 *                     if order:
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
*/
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr {
  PyObject_HEAD
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_o;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};



/* "hummingbot/core/event/event_listener.pxd":6
 * 
 * 
 * cdef class EventListener:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object __weakref__
*/

struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener {
  PyObject *(*c_set_event_info)(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *, int64_t, struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *);
  PyObject *(*c_call)(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *, PyObject *);
};
static struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener *__pyx_vtabptr_10hummingbot_4core_5event_14event_listener_EventListener;


/* "hummingbot/core/pubsub.pxd":17
 * 
 * 
 * cdef class PubSub:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Events _events
*/

struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub {
  PyObject *(*c_log_exception)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t, PyObject *);
  PyObject *(*c_add_listener)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t, struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *);
  PyObject *(*c_remove_listener)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t, struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *);
  PyObject *(*c_remove_dead_listeners)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t);
  PyObject *(*c_get_listeners)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t);
  PyObject *(*c_trigger_event)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t, PyObject *);
};
static struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub *__pyx_vtabptr_10hummingbot_4core_6pubsub_PubSub;


/* "hummingbot/core/time_iterator.pxd":7
 * 
 * 
 * cdef class TimeIterator(PubSub):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double _current_timestamp
*/

struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator {
  struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub __pyx_base;
  PyObject *(*c_start)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, struct __pyx_obj_10hummingbot_4core_5clock_Clock *, double);
  PyObject *(*c_stop)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, struct __pyx_obj_10hummingbot_4core_5clock_Clock *);
  PyObject *(*c_tick)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, double);
};
static struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator *__pyx_vtabptr_10hummingbot_4core_13time_iterator_TimeIterator;


/* "hummingbot/core/data_type/limit_order.pxd":6
 * 
 * 
 * cdef class LimitOrder:             # <<<<<<<<<<<<<<
 *     cdef:
 *         CPPLimitOrder _cpp_limit_order
*/

struct __pyx_vtabstruct_10hummingbot_4core_9data_type_11limit_order_LimitOrder {
  PY_LONG_LONG (*c_age)(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *);
  PY_LONG_LONG (*c_age_til)(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *, PY_LONG_LONG);
};
static struct __pyx_vtabstruct_10hummingbot_4core_9data_type_11limit_order_LimitOrder *__pyx_vtabptr_10hummingbot_4core_9data_type_11limit_order_LimitOrder;


/* "order_tracker.pxd":7
 * 
 * 
 * cdef class OrderTracker(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict _tracked_limit_orders
*/

struct __pyx_vtabstruct_10hummingbot_8strategy_13order_tracker_OrderTracker {
  struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
  PyObject *(*c_get_limit_orders)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *);
  PyObject *(*c_get_market_orders)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *);
  PyObject *(*c_get_shadow_limit_orders)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *);
  int (*c_has_in_flight_cancel)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
  int (*c_check_and_track_cancel)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
  PyObject *(*c_get_market_pair_from_order_id)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
  PyObject *(*c_get_shadow_market_pair_from_order_id)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *(*c_get_limit_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *, PyObject *);
  PyObject *(*c_get_market_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *(*c_get_shadow_limit_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
  PyObject *(*c_start_tracking_limit_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *, PyObject *, int, PyObject *, PyObject *);
  PyObject *(*c_stop_tracking_limit_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *, PyObject *);
  PyObject *(*c_start_tracking_market_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *, PyObject *, int, PyObject *);
  PyObject *(*c_stop_tracking_market_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *, PyObject *);
  PyObject *(*c_check_and_cleanup_shadow_records)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *);
  PyObject *(*c_add_create_order_pending)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
  PyObject *(*c_remove_create_order_pending)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
};
static struct __pyx_vtabstruct_10hummingbot_8strategy_13order_tracker_OrderTracker *__pyx_vtabptr_10hummingbot_8strategy_13order_tracker_OrderTracker;


/* "hummingbot/strategy/strategy_base.pxd":8
 * from .order_tracker cimport OrderTracker
 * 
 * cdef class StrategyBase(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         set _sb_markets
*/

struct __pyx_vtabstruct_10hummingbot_8strategy_13strategy_base_StrategyBase {
  struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
  PyObject *(*c_add_markets)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_remove_markets)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_create_buy_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_create_sell_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_fill_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_fail_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_cancel_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_expire_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_complete_buy_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_complete_sell_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_complete_funding_payment)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_change_position_mode_succeed)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_change_position_mode_fail)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_add_liquidity)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_remove_liquidity)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_fail_lp_update)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_fail_order_tracker)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_cancel_order_tracker)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_expire_order_tracker)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_complete_buy_order_tracker)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_complete_sell_order_tracker)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_buy_with_specific_market)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_buy_with_specific_market *__pyx_optional_args);
  PyObject *(*c_sell_with_specific_market)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_sell_with_specific_market *__pyx_optional_args);
  PyObject *(*c_cancel_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *);
  PyObject *(*c_start_tracking_limit_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *, int, PyObject *, PyObject *);
  PyObject *(*c_stop_tracking_limit_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *);
  PyObject *(*c_start_tracking_market_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *, int, PyObject *);
  PyObject *(*c_stop_tracking_market_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *);
  PyObject *(*c_track_restored_orders)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_sum_flat_fees)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_10hummingbot_8strategy_13strategy_base_StrategyBase *__pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase;


/* "hummingbot/core/event/event_reporter.pxd":4
 * 
 * 
 * cdef class EventReporter(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str event_source
*/

struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_reporter_EventReporter {
  struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_reporter_EventReporter *__pyx_vtabptr_10hummingbot_4core_5event_14event_reporter_EventReporter;


/* "hummingbot/core/event/event_logger.pxd":4
 * 
 * 
 * cdef class EventLogger(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str _event_source
*/

struct __pyx_vtabstruct_10hummingbot_4core_5event_12event_logger_EventLogger {
  struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_4core_5event_12event_logger_EventLogger *__pyx_vtabptr_10hummingbot_4core_5event_12event_logger_EventLogger;


/* "hummingbot/core/data_type/order_book.pxd":12
 * 
 * 
 * cdef class OrderBook(PubSub):             # <<<<<<<<<<<<<<
 *     cdef set[OrderBookEntry] _bid_book
 *     cdef set[OrderBookEntry] _ask_book
*/

struct __pyx_vtabstruct_10hummingbot_4core_9data_type_10order_book_OrderBook {
  struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub __pyx_base;
  PyObject *(*c_apply_diffs)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, std::vector<OrderBookEntry> , std::vector<OrderBookEntry> , int64_t);
  PyObject *(*c_apply_snapshot)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, std::vector<OrderBookEntry> , std::vector<OrderBookEntry> , int64_t);
  PyObject *(*c_apply_trade)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, PyObject *);
  PyObject *(*c_apply_numpy_diffs)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, PyArrayObject *, PyArrayObject *);
  PyObject *(*c_apply_numpy_snapshot)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, PyArrayObject *, PyArrayObject *);
  double (*c_get_price)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_price_for_volume)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_price_for_quote_volume)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_volume_for_price)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_quote_volume_for_price)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_vwap_for_volume)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_quote_volume_for_base_amount)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
};
static struct __pyx_vtabstruct_10hummingbot_4core_9data_type_10order_book_OrderBook *__pyx_vtabptr_10hummingbot_4core_9data_type_10order_book_OrderBook;


/* "hummingbot/core/network_iterator.pxd":6
 * 
 * 
 * cdef class NetworkIterator(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _network_status
*/

struct __pyx_vtabstruct_10hummingbot_4core_16network_iterator_NetworkIterator {
  struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_4core_16network_iterator_NetworkIterator *__pyx_vtabptr_10hummingbot_4core_16network_iterator_NetworkIterator;


/* "hummingbot/connector/connector_base.pxd":5
 * from hummingbot.core.network_iterator cimport NetworkIterator
 * 
 * cdef class ConnectorBase(NetworkIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         EventReporter _event_reporter
*/

struct __pyx_vtabstruct_10hummingbot_9connector_14connector_base_ConnectorBase {
  struct __pyx_vtabstruct_10hummingbot_4core_16network_iterator_NetworkIterator __pyx_base;
  PyObject *(*c_buy)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_buy *__pyx_optional_args);
  PyObject *(*c_sell)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_sell *__pyx_optional_args);
  PyObject *(*c_cancel)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *);
  PyObject *(*c_stop_tracking_order)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *);
  PyObject *(*c_get_balance)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *);
  PyObject *(*c_get_available_balance)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *);
  PyObject *(*c_get_price)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, int);
  PyObject *(*c_get_order_price_quantum)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *);
  PyObject *(*c_get_order_size_quantum)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *);
  PyObject *(*c_quantize_order_price)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *);
  PyObject *(*c_quantize_order_amount)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_quantize_order_amount *__pyx_optional_args);
};
static struct __pyx_vtabstruct_10hummingbot_9connector_14connector_base_ConnectorBase *__pyx_vtabptr_10hummingbot_9connector_14connector_base_ConnectorBase;


/* "hummingbot/connector/exchange_base.pxd":10
 * )
 * 
 * cdef class ExchangeBase(ConnectorBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _order_book_tracker
*/

struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase {
  struct __pyx_vtabstruct_10hummingbot_9connector_14connector_base_ConnectorBase __pyx_base;
  struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *(*c_get_order_book)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_quote_volume_for_base_amount)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_volume_for_price)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_quote_volume_for_price)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_vwap_for_volume)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_price_for_quote_volume)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_price_for_volume)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  PyObject *(*c_get_fee)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_get_fee *__pyx_optional_args);
};
static struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_vtabptr_10hummingbot_9connector_13exchange_base_ExchangeBase;


/* "hummingbot/strategy/asset_price_delegate.pxd":2
 * 
 * cdef class AssetPriceDelegate:             # <<<<<<<<<<<<<<
 *     cdef object c_get_mid_price(self)
*/

struct __pyx_vtabstruct_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate {
  PyObject *(*c_get_mid_price)(struct __pyx_obj_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate *);
};
static struct __pyx_vtabstruct_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate *__pyx_vtabptr_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate;


/* "hummingbot/strategy/order_book_asset_price_delegate.pxd":4
 * from .asset_price_delegate cimport AssetPriceDelegate
 * 
 * cdef class OrderBookAssetPriceDelegate(AssetPriceDelegate):             # <<<<<<<<<<<<<<
 *     cdef:
 *         ExchangeBase _market
*/

struct __pyx_vtabstruct_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate {
  struct __pyx_vtabstruct_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate *__pyx_vtabptr_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate;


/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":38
 * 
 * 
 * cdef class PureMarketMakingStrategy(StrategyBase):             # <<<<<<<<<<<<<<
 *     OPTION_LOG_CREATE_ORDER = 1 << 3
 *     OPTION_LOG_MAKER_ORDER_FILLED = 1 << 4
*/

struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy {
  struct __pyx_vtabstruct_10hummingbot_8strategy_13strategy_base_StrategyBase __pyx_base;
  PyObject *(*c_get_mid_price)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *);
  PyObject *(*c_create_base_proposal)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *);
  PyObject *(*c_get_adjusted_available_balance)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_order_levels_modifiers)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_price_band)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_ping_pong)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_order_price_modifiers)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_order_size_modifiers)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_inventory_skew)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_budget_constraint)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  PyObject *(*c_filter_out_takers)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_order_optimization)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_add_transaction_costs)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  int (*c_is_within_tolerance)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *, PyObject *);
  PyObject *(*c_cancel_active_orders)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  PyObject *(*c_cancel_orders_below_min_spread)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *);
  PyObject *(*c_cancel_active_orders_on_max_age_limit)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *);
  int (*c_to_create_orders)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  PyObject *(*c_execute_orders_proposal)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
  PyObject *(*set_timers)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *);
  PyObject *(*c_apply_moving_price_band)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *);
};
static struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_vtabptr_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetTopmostException.proto (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyImportError_Check.proto */
#define __Pyx_PyExc_ImportError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ImportError)

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* TupleAndListFromArray.proto (used by fastcall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by BytesEquals) */
#include <string.h>

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyObjectCallOneArg.proto (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by RejectKeywords) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RejectKeywords.export */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* PyValueError_Check.proto */
#define __Pyx_PyExc_ValueError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ValueError)

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* IterFinish.proto (used by set_iter) */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* set_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set);
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectDelAttr.proto (used by PyObjectSetAttrStr) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
#define __Pyx_PyObject_DelAttr(o, n) PyObject_SetAttr(o, n, NULL)
#else
#define __Pyx_PyObject_DelAttr(o, n) PyObject_DelAttr(o, n)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   __Pyx_PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_SubtractCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_SubtractCObj(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* pybytes_as_double.proto (used by pyobject_as_double) */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj);
static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length);
static CYTHON_INLINE double __Pyx_PyBytes_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    as_c_string = PyBytes_AS_STRING(obj);
    size = PyBytes_GET_SIZE(obj);
#else
    if (PyBytes_AsStringAndSize(obj, &as_c_string, &size) < 0) {
        return (double)-1;
    }
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}
static CYTHON_INLINE double __Pyx_PyByteArray_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    as_c_string = PyByteArray_AS_STRING(obj);
    size = PyByteArray_GET_SIZE(obj);
#else
    as_c_string = PyByteArray_AsString(obj);
    if (as_c_string == NULL) {
        return (double)-1;
    }
    size = PyByteArray_Size(obj);
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}

/* pyunicode_as_double.proto (used by pyobject_as_double) */
#if !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
static const char* __Pyx__PyUnicode_AsDouble_Copy(const void* data, const int kind, char* buffer, Py_ssize_t start, Py_ssize_t end) {
    int last_was_punctuation;
    Py_ssize_t i;
    last_was_punctuation = 1;
    for (i=start; i <= end; i++) {
        Py_UCS4 chr = PyUnicode_READ(kind, data, i);
        int is_punctuation = (chr == '_') | (chr == '.');
        *buffer = (char)chr;
        buffer += (chr != '_');
        if (unlikely(chr > 127)) goto parse_failure;
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyUnicode_AsDouble_inf_nan(const void* data, int kind, Py_ssize_t start, Py_ssize_t length) {
    int matches = 1;
    Py_UCS4 chr;
    Py_UCS4 sign = PyUnicode_READ(kind, data, start);
    int is_signed = (sign == '-') | (sign == '+');
    start += is_signed;
    length -= is_signed;
    switch (PyUnicode_READ(kind, data, start)) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'a') | (chr == 'A');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'n') | (chr == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'f') | (chr == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+3);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+4);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+5);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+6);
            matches &= (chr == 't') | (chr == 'T');
            chr = PyUnicode_READ(kind, data, start+7);
            matches &= (chr == 'y') | (chr == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static double __Pyx_PyUnicode_AsDouble_WithSpaces(PyObject *obj) {
    double value;
    const char *last;
    char *end;
    Py_ssize_t start, length = PyUnicode_GET_LENGTH(obj);
    const int kind = PyUnicode_KIND(obj);
    const void* data = PyUnicode_DATA(obj);
    start = 0;
    while (Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, start)))
        start++;
    while (start < length - 1 && Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, length - 1)))
        length--;
    length -= start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyUnicode_AsDouble_inf_nan(data, kind, start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    if (length < 40) {
        char number[40];
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((length + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}
#endif
static CYTHON_INLINE double __Pyx_PyUnicode_AsDouble(PyObject *obj) {
#if !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(__Pyx_PyUnicode_READY(obj) == -1))
        return (double)-1;
    if (likely(PyUnicode_IS_ASCII(obj))) {
        const char *s;
        Py_ssize_t length;
        s = PyUnicode_AsUTF8AndSize(obj, &length);
        return __Pyx__PyBytes_AsDouble(obj, s, length);
    }
    return __Pyx_PyUnicode_AsDouble_WithSpaces(obj);
#else
    return __Pyx_SlowPyString_AsDouble(obj);
#endif
}

/* pyobject_as_double.proto */
static double __Pyx__PyObject_AsDouble(PyObject* obj);
#if CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyObject_AsDouble(obj)\
(likely(PyFloat_CheckExact(obj)) ? PyFloat_AS_DOUBLE(obj) :\
 likely(PyLong_CheckExact(obj)) ?\
 PyFloat_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#else
#define __Pyx_PyObject_AsDouble(obj)\
((likely(PyFloat_CheckExact(obj))) ?  __Pyx_PyFloat_AS_DOUBLE(obj) :\
 likely(PyLong_CheckExact(obj)) ?\
 PyLong_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#endif

/* FloatExceptionCheck.proto */
#define __PYX_CHECK_FLOAT_EXCEPTION(value, error_value)\
    ((error_value) == (error_value) ?\
     (value) == (error_value) :\
     (value) != (value))

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* JoinPyUnicode.export */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* PyObjectFormat.proto */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* s, PyObject* f);
#else
#define __Pyx_PyObject_Format(s, f) PyObject_Format(s, f)
#endif

/* PyObjectVectorCallKwBuilder.proto */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto (used by FetchCommonType) */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* CallTypeTraverse.proto (used by CythonFunctionShared) */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* PyObjectVectorCallMethodKwBuilder.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_VectorcallMethod_CallFromBuilder PyObject_VectorcallMethod
#else
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#endif

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyLongCompare.proto */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* py_abs.proto */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *num);
#define __Pyx_PyNumber_Absolute(x)\
    ((likely(PyLong_CheckExact(x))) ?\
         (likely(__Pyx_PyLong_IsNonNeg(x)) ? __Pyx_NewRef(x) : __Pyx_PyLong_AbsNeg(x)) :\
         PyNumber_Absolute(x))
#else
#define __Pyx_PyNumber_Absolute(x)  PyNumber_Absolute(x)
#endif

/* pynumber_float.proto */
static CYTHON_INLINE PyObject* __Pyx__PyNumber_Float(PyObject* obj);
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : __Pyx__PyNumber_Float(x))

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00a2
    return PyList_Extend(L, v);
#elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PyObject_Unicode.proto */
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))

/* UnicodeConcatInPlace.proto */
# if CYTHON_COMPILING_IN_CPYTHON
    #if CYTHON_REFNANNY
        #define __Pyx_PyUnicode_ConcatInPlace(left, right, unsafe_shared) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, unsafe_shared, __pyx_refnanny)
    #else
        #define __Pyx_PyUnicode_ConcatInPlace(left, right, unsafe_shared) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, unsafe_shared)
    #endif
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_DefinitelyUnique)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_OwnStrongReference)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_DefinitelyUnique)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_SharedReference)
    static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right, int unsafe_shared
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    );
#else
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace __Pyx_PyUnicode_Concat
#endif
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace(left, right))

/* RaiseUnboundLocalError.proto */
static void __Pyx_RaiseUnboundLocalError(const char *varname);

/* RaiseClosureNameError.proto */
static void __Pyx_RaiseClosureNameError(const char *varname);

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* IterNextPlain.proto (used by IterNext) */
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next_Plain(PyObject *iterator);
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *__Pyx_GetBuiltinNext_LimitedAPI(void);
#endif

/* IterNext.proto */
#define __Pyx_PyIter_Next(obj) __Pyx_PyIter_Next2(obj, NULL)
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject *, PyObject *);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyRuntimeError_Check.proto */
#define __Pyx_PyExc_RuntimeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_RuntimeError)

/* PyObjectCallNoArg.proto (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);
#endif

/* PyObjectCallMethod0.proto (used by dict_iter) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* UnpackTupleError.proto (used by UnpackTuple2) */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto (used by dict_iter) */
static CYTHON_INLINE int __Pyx_unpack_tuple2(
    PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple);
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* PyRange_Check.proto */
#if CYTHON_COMPILING_IN_PYPY && !defined(PyRange_Check)
  #define PyRange_Check(obj)  __Pyx_TypeCheck((obj), &PyRange_Type)
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* BuildPyUnicode.proto (used by COrdinalToPyUnicode) */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, const char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* COrdinalToPyUnicode.proto (used by CIntToPyUnicode) */
static CYTHON_INLINE int __Pyx_CheckUnicodeValue(int value);
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromOrdinal_Padded(int value, Py_ssize_t width, char padding_char);

/* GCCDiagnostics.proto (used by CIntToPyUnicode) */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* IncludeStdlibH.proto (used by CIntToPyUnicode) */
#include <stdlib.h>

/* CIntToPyUnicode.proto */
#define __Pyx_PyUnicode_From_int(value, width, padding_char, format_char) (\
    ((format_char) == ('c')) ?\
        __Pyx_uchar___Pyx_PyUnicode_From_int(value, width, padding_char) :\
        __Pyx____Pyx_PyUnicode_From_int(value, width, padding_char, format_char)\
    )
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char);
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_AddCObj(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* CIntToPyUnicode.proto */
#define __Pyx_PyUnicode_From_Py_ssize_t(value, width, padding_char, format_char) (\
    ((format_char) == ('c')) ?\
        __Pyx_uchar___Pyx_PyUnicode_From_Py_ssize_t(value, width, padding_char) :\
        __Pyx____Pyx_PyUnicode_From_Py_ssize_t(value, width, padding_char, format_char)\
    )
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char);
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* PyTypeError_Check.proto */
#define __Pyx_PyExc_TypeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_TypeError)

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear);

/* CheckTypeForFreelists.proto */
#if CYTHON_USE_FREELISTS
#if CYTHON_USE_TYPE_SPECS
#define __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, expected_tp, expected_size) ((int) ((t) == (expected_tp)))
#define __PYX_CHECK_TYPE_FOR_FREELIST_FLAGS  Py_TPFLAGS_IS_ABSTRACT
#else
#define __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, expected_tp, expected_size) ((int) ((t)->tp_basicsize == (expected_size)))
#define __PYX_CHECK_TYPE_FOR_FREELIST_FLAGS  (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)
#endif
#define __PYX_CHECK_TYPE_FOR_FREELISTS(t, expected_tp, expected_size)\
    (__PYX_CHECK_FINAL_TYPE_FOR_FREELISTS((t), (expected_tp), (expected_size)) &\
     (int) (!__Pyx_PyType_HasFeature((t), __PYX_CHECK_TYPE_FOR_FREELIST_FLAGS)))
#endif

/* AllocateExtensionType.proto */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_2_4
#define __PYX_HAVE_RT_ImportType_proto_3_2_4
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_2_4 {
   __Pyx_ImportType_CheckSize_Error_3_2_4 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_2_4 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_2_4 = 2
};
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* ValidateBasesTuple.proto (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* MergeVTables.proto */
static int __Pyx_MergeVtables(PyTypeObject *type);

/* DelItemOnTypeDict.proto (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* FunctionImport.proto */
static int __Pyx_ImportFunction_3_2_4(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* ClassMethod.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
#include "descrobject.h"
#endif
CYTHON_UNUSED static PyObject* __Pyx_Method_ClassMethod(PyObject *method);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_long__double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_long__double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_long__double(a, b) ((a)-(b))
    #define __Pyx_c_prod_long__double(a, b) ((a)*(b))
    #define __Pyx_c_quot_long__double(a, b) ((a)/(b))
    #define __Pyx_c_neg_long__double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_long__double(z) ((z)==(long double)0)
    #define __Pyx_c_conj_long__double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_long__double(z)     (::std::abs(z))
        #define __Pyx_c_pow_long__double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_long__double(z) ((z)==0)
    #define __Pyx_c_conj_long__double(z)    (conjl(z))
    #if 1
        #define __Pyx_c_abs_long__double(z)     (cabsl(z))
        #define __Pyx_c_pow_long__double(a, b)  (cpowl(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_sum_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_diff_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_prod_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_neg_long__double(__pyx_t_long_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_long__double(__pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_conj_long__double(__pyx_t_long_double_complex);
    #if 1
        static CYTHON_INLINE long double __Pyx_c_abs_long__double(__pyx_t_long_double_complex);
        static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_pow_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    #endif
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyLong_As_int64_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int64_t(int64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* PyObjectCall2Args.proto (used by PyObjectCallMethod1) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethod1.proto (used by CoroutineBase) */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* ReturnWithStopIteration.proto (used by CoroutineBase) */
static CYTHON_INLINE void __Pyx_ReturnWithStopIteration(PyObject* value, int async, int iternext);

/* CoroutineBase.proto (used by Generator) */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *gi_weakreflist;
#endif
    PyObject *classobj;
    PyObject *yieldfrom;
    __Pyx_pyiter_sendfunc yieldfrom_am_send;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
#if CYTHON_USE_SYS_MONITORING && (CYTHON_PROFILE || CYTHON_TRACE)
    PyMonitoringState __pyx_pymonitoring_state[__Pyx_MonitoringEventTypes_CyGen_count];
    uint64_t __pyx_pymonitoring_version;
#endif
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static __Pyx_PySendResult __Pyx_Coroutine_AmSend(PyObject *self, PyObject *value, PyObject **retval);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static __Pyx_PySendResult __Pyx_Coroutine_Close(PyObject *self, PyObject **retval);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);
static char __Pyx_Coroutine_test_and_set_is_running(__pyx_CoroutineObject *gen);
static void __Pyx_Coroutine_unset_is_running(__pyx_CoroutineObject *gen);
static char __Pyx_Coroutine_get_is_running(__pyx_CoroutineObject *gen);
static PyObject *__Pyx_Coroutine_get_is_running_getter(PyObject *gen, void *closure);
#if __PYX_HAS_PY_AM_SEND == 2
static void __Pyx_SetBackportTypeAmSend(PyTypeObject *type, __Pyx_PyAsyncMethodsStruct *static_amsend_methods, __Pyx_pyiter_sendfunc am_send);
#endif
static PyObject *__Pyx_Coroutine_fail_reduce_ex(PyObject *self, PyObject *arg);

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_mstate_global->__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);
static CYTHON_INLINE PyObject *__Pyx_Generator_GetInlinedResult(PyObject *self);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_8itemsize_itemsize(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_9alignment_alignment(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_6fields_fields(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_5names_names(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_ArrayDescr *__pyx_f_5numpy_5dtype_8subarray_subarray(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_uint64 __pyx_f_5numpy_5dtype_5flags_flags(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_7numiter_numiter(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_4size_size(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_5index_index(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_2nd_nd(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_9broadcast_10dimensions_dimensions(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE void **__pyx_f_5numpy_9broadcast_5iters_iters(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_get_mid_price(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_start(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, struct __pyx_obj_10hummingbot_4core_5clock_Clock *__pyx_v_clock, double __pyx_v_timestamp); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_stop(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, struct __pyx_obj_10hummingbot_4core_5clock_Clock *__pyx_v_clock); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_tick(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, double __pyx_v_timestamp); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_create_base_proposal(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_get_adjusted_available_balance(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_orders); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_order_levels_modifiers(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_price_band(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_moving_price_band(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_ping_pong(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_order_price_modifiers(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_order_size_modifiers(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_inventory_skew(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_budget_constraint(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_filter_out_takers(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_order_optimization(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_add_transaction_costs(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_did_fill_order(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_filled_event); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_did_complete_buy_order(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_completed_event); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_did_complete_sell_order(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_completed_event); /* proto*/
static int __pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_is_within_tolerance(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_current_prices, PyObject *__pyx_v_proposal_prices); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_cancel_active_orders_on_max_age_limit(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_cancel_active_orders(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_cancel_orders_below_min_spread(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto*/
static int __pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_to_create_orders(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_execute_orders_proposal(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_set_timers(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto*/

/* Module declarations from "libc.stdint" */

/* Module declarations from "hummingbot.core.clock" */

/* Module declarations from "libcpp.utility" */

/* Module declarations from "libcpp.unordered_map" */

/* Module declarations from "libcpp.unordered_set" */

/* Module declarations from "hummingbot.core.PyRef" */

/* Module declarations from "hummingbot.core.event.event_listener" */

/* Module declarations from "hummingbot.core.pubsub" */

/* Module declarations from "hummingbot.core.time_iterator" */

/* Module declarations from "libcpp" */

/* Module declarations from "libc.string" */

/* Module declarations from "libcpp.string_view" */

/* Module declarations from "libcpp.string" */

/* Module declarations from "hummingbot.core.data_type.LimitOrder" */

/* Module declarations from "hummingbot.core.data_type.limit_order" */

/* Module declarations from "hummingbot.strategy.order_tracker" */

/* Module declarations from "hummingbot.strategy.strategy_base" */

/* Module declarations from "hummingbot.core.event.event_reporter" */

/* Module declarations from "hummingbot.core.event.event_logger" */

/* Module declarations from "libcpp.set" */

/* Module declarations from "libcpp.vector" */

/* Module declarations from "hummingbot.core.data_type.OrderBookEntry" */

/* Module declarations from "hummingbot.core.data_type.order_book_query_result" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "numpy" */

/* Module declarations from "numpy" */

/* Module declarations from "hummingbot.core.data_type.order_book" */

/* Module declarations from "hummingbot.core.network_iterator" */

/* Module declarations from "hummingbot.connector.connector_base" */

/* Module declarations from "hummingbot.connector.exchange_base" */

/* Module declarations from "hummingbot.strategy.asset_price_delegate" */

/* Module declarations from "hummingbot.strategy.order_book_asset_price_delegate" */

/* Module declarations from "hummingbot.strategy.pure_market_making.inventory_skew_calculator" */
static PyObject *(*__pyx_f_10hummingbot_8strategy_18pure_market_making_25inventory_skew_calculator_c_calculate_bid_ask_ratios_from_base_asset_ratio)(double, double, double, double, double); /*proto*/

/* Module declarations from "hummingbot.strategy.pure_market_making.pure_market_making" */
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "hummingbot.strategy.pure_market_making.pure_market_making"
extern int __pyx_module_is_main_hummingbot__strategy__pure_market_making__pure_market_making;
int __pyx_module_is_main_hummingbot__strategy__pure_market_making__pure_market_making = 0;

/* Implementation of "hummingbot.strategy.pure_market_making.pure_market_making" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_all;
static PyObject *__pyx_builtin_round;
static PyObject *__pyx_builtin_max;
static PyObject *__pyx_builtin_sorted;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_super;
/* #### Code section: string_decls ### */
/* #### Code section: decls ### */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_logger(CYTHON_UNUSED PyTypeObject *__pyx_v_cls); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_2init_params(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_market_info, PyObject *__pyx_v_bid_spread, PyObject *__pyx_v_ask_spread, PyObject *__pyx_v_order_amount, PyObject *__pyx_v_order_levels, PyObject *__pyx_v_order_level_spread, PyObject *__pyx_v_order_level_amount, PyObject *__pyx_v_order_refresh_time, PyObject *__pyx_v_max_order_age, PyObject *__pyx_v_order_refresh_tolerance_pct, PyObject *__pyx_v_filled_order_delay, PyObject *__pyx_v_inventory_skew_enabled, PyObject *__pyx_v_inventory_target_base_pct, PyObject *__pyx_v_inventory_range_multiplier, PyObject *__pyx_v_hanging_orders_enabled, PyObject *__pyx_v_hanging_orders_cancel_pct, PyObject *__pyx_v_order_optimization_enabled, PyObject *__pyx_v_ask_order_optimization_depth, PyObject *__pyx_v_bid_order_optimization_depth, PyObject *__pyx_v_add_transaction_costs_to_orders, PyObject *__pyx_v_asset_price_delegate, PyObject *__pyx_v_inventory_cost_price_delegate, PyObject *__pyx_v_price_type, PyObject *__pyx_v_take_if_crossed, PyObject *__pyx_v_price_ceiling, PyObject *__pyx_v_price_floor, PyObject *__pyx_v_ping_pong_enabled, PyObject *__pyx_v_logging_options, PyObject *__pyx_v_status_report_interval, PyObject *__pyx_v_minimum_spread, PyObject *__pyx_v_hb_app_notification, PyObject *__pyx_v_order_override, PyObject *__pyx_v_split_order_levels_enabled, PyObject *__pyx_v_bid_order_level_spreads, PyObject *__pyx_v_ask_order_level_spreads, PyObject *__pyx_v_should_wait_order_cancel_confirmation, PyObject *__pyx_v_moving_price_band); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_4all_markets_ready(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11market_info___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13max_order_age___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14minimum_spread___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17ping_pong_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28ask_order_optimization_depth___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28bid_order_optimization_depth___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10price_type___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27order_refresh_tolerance_pct___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27order_refresh_tolerance_pct_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_amount___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_amount_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_levels___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_levels_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10buy_levels___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10buy_levels_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11sell_levels___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11sell_levels_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_amount___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_amount_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_spread___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_spread_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22inventory_skew_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22inventory_skew_enabled_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25inventory_target_base_pct___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25inventory_target_base_pct_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26inventory_range_multiplier___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26inventory_range_multiplier_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_enabled_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25hanging_orders_cancel_pct___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25hanging_orders_cancel_pct_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10bid_spread___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10bid_spread_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10ask_spread___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10ask_spread_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26order_optimization_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26order_optimization_enabled_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_refresh_time___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_refresh_time_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18filled_order_delay___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18filled_order_delay_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31add_transaction_costs_to_orders___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31add_transaction_costs_to_orders_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13price_ceiling___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13price_ceiling_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11price_floor___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11price_floor_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10base_asset___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11quote_asset___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12trading_pair___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14order_override___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14order_override_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26split_order_levels_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23bid_order_level_spreads___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23ask_order_level_spreads___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25moving_price_band_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25moving_price_band_enabled_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17price_ceiling_pct___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17price_ceiling_pct_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15price_floor_pct___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15price_floor_pct_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23price_band_refresh_time___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23price_band_refresh_time_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17moving_price_band___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_6get_price(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_8get_mid_price(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17hanging_order_ids___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28market_info_to_active_orders___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13active_orders___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11active_buys___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12active_sells___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25active_non_hanging_orders___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15logging_options___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15logging_options_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, int64_t __pyx_v_logging_options); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_tracker___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20asset_price_delegate___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20asset_price_delegate_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29inventory_cost_price_delegate___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29inventory_cost_price_delegate_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10inventory_skew_stats_data_frame(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12pure_mm_assets_df(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_to_show_current_pct); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14active_orders_df(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_16market_status_data_frame(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_market_trading_pair_tuples); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18format_status(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20execute_orders_proposal(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22cancel_order(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_id); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7c_start_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_24adjusted_available_balance_for_orders_budget_constrain(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda2(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda3(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_39c_cancel_active_orders_on_max_age_limit_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_3genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26notify_hb_app(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_msg); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28get_price_type(CYTHON_UNUSED struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_price_type_str); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_30__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_32__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyTypeObject *__pyx_ptype_10hummingbot_4core_5clock_Clock;
  PyTypeObject *__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener;
  PyTypeObject *__pyx_ptype_10hummingbot_4core_6pubsub_PubSub;
  PyTypeObject *__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator;
  PyTypeObject *__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase;
  PyTypeObject *__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter;
  PyTypeObject *__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger;
  PyTypeObject *__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult;
  PyTypeObject *__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult;
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyTypeObject *__pyx_ptype_5numpy_dtype;
  PyTypeObject *__pyx_ptype_5numpy_flatiter;
  PyTypeObject *__pyx_ptype_5numpy_broadcast;
  PyTypeObject *__pyx_ptype_5numpy_ndarray;
  PyTypeObject *__pyx_ptype_5numpy_generic;
  PyTypeObject *__pyx_ptype_5numpy_number;
  PyTypeObject *__pyx_ptype_5numpy_integer;
  PyTypeObject *__pyx_ptype_5numpy_signedinteger;
  PyTypeObject *__pyx_ptype_5numpy_unsignedinteger;
  PyTypeObject *__pyx_ptype_5numpy_inexact;
  PyTypeObject *__pyx_ptype_5numpy_floating;
  PyTypeObject *__pyx_ptype_5numpy_complexfloating;
  PyTypeObject *__pyx_ptype_5numpy_flexible;
  PyTypeObject *__pyx_ptype_5numpy_character;
  PyTypeObject *__pyx_ptype_5numpy_ufunc;
  PyTypeObject *__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook;
  PyTypeObject *__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator;
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase;
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate;
  PyObject *__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy;
  PyObject *__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start;
  PyObject *__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr;
  PyObject *__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit;
  PyObject *__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr;
  PyObject *__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal;
  PyObject *__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr;
  PyObject *__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  PyObject *__pyx_k_;
  PyObject *__pyx_k__2;
  PyObject *__pyx_k__3;
  PyObject *__pyx_k__4;
  PyObject *__pyx_k__5;
  PyObject *__pyx_k__6;
  PyObject *__pyx_k__7;
  PyObject *__pyx_k__8;
  PyObject *__pyx_k__9;
  PyObject *__pyx_k__10;
  PyObject *__pyx_k__11;
  PyObject *__pyx_k__12;
  PyObject *__pyx_codeobj_tab[24];
  PyObject *__pyx_string_tab[486];
  PyObject *__pyx_number_tab[12];
/* #### Code section: module_state_contents ### */
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* IterNextPlain.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
PyObject *__Pyx_GetBuiltinNext_LimitedAPI_cache;
#endif


#if CYTHON_USE_FREELISTS
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start[8];
int __pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start;
#endif

#if CYTHON_USE_FREELISTS
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr *__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr[8];
int __pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr;
#endif

#if CYTHON_USE_FREELISTS
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit *__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit[8];
int __pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit;
#endif

#if CYTHON_USE_FREELISTS
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr *__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr[8];
int __pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr;
#endif

#if CYTHON_USE_FREELISTS
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal *__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal[8];
int __pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal;
#endif

#if CYTHON_USE_FREELISTS
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr *__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr[8];
int __pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr;
#endif

#if CYTHON_USE_FREELISTS
struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr *__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr[8];
int __pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr;
#endif
/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* Generator.module_state_decls */
PyTypeObject *__pyx_GeneratorType;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_0_1 __pyx_string_tab[0]
#define __pyx_kp_u_0_5 __pyx_string_tab[1]
#define __pyx_kp_u_1 __pyx_string_tab[2]
#define __pyx_kp_u_100 __pyx_string_tab[3]
#define __pyx_kp_u_1_2 __pyx_string_tab[4]
#define __pyx_kp_u_2 __pyx_string_tab[5]
#define __pyx_kp_u_2_2 __pyx_string_tab[6]
#define __pyx_kp_u_4f __pyx_string_tab[7]
#define __pyx_kp_u_Amount_Adj __pyx_string_tab[8]
#define __pyx_kp_u_Amount_Orig __pyx_string_tab[9]
#define __pyx_kp_u_Assets __pyx_string_tab[10]
#define __pyx_kp_u_Available_Balance __pyx_string_tab[11]
#define __pyx_kp_u_Best_Ask __pyx_string_tab[12]
#define __pyx_kp_u_Best_Bid __pyx_string_tab[13]
#define __pyx_kp_u_Canceling_Order __pyx_string_tab[14]
#define __pyx_kp_u_Creating __pyx_string_tab[15]
#define __pyx_kp_u_Current __pyx_string_tab[16]
#define __pyx_kp_u_Current_Value __pyx_string_tab[17]
#define __pyx_kp_u_Dict_str_List_str __pyx_string_tab[18]
#define __pyx_kp_u_H_M_S __pyx_string_tab[19]
#define __pyx_kp_u_Hanging_maker_BUY_order __pyx_string_tab[20]
#define __pyx_kp_u_Hanging_maker_SELL_order __pyx_string_tab[21]
#define __pyx_kp_u_Hanging_maker_buy_order __pyx_string_tab[22]
#define __pyx_kp_u_Hanging_maker_sell_order __pyx_string_tab[23]
#define __pyx_kp_u_ID __pyx_string_tab[24]
#define __pyx_kp_u_Initial_inventory_price_is_not_s __pyx_string_tab[25]
#define __pyx_kp_u_Inventory_Range __pyx_string_tab[26]
#define __pyx_kp_u_List_Decimal __pyx_string_tab[27]
#define __pyx_kp_u_List_MarketTradingPairTuple __pyx_string_tab[28]
#define __pyx_kp_u_Maker_BUY_order __pyx_string_tab[29]
#define __pyx_kp_u_Maker_SELL_order __pyx_string_tab[30]
#define __pyx_kp_u_Maker_buy_order __pyx_string_tab[31]
#define __pyx_kp_u_Maker_buy_order_of __pyx_string_tab[32]
#define __pyx_kp_u_Maker_sell_order __pyx_string_tab[33]
#define __pyx_kp_u_Maker_sell_order_of __pyx_string_tab[34]
#define __pyx_kp_u_Market_connectors_are_not_ready __pyx_string_tab[35]
#define __pyx_kp_u_Markets __pyx_string_tab[36]
#define __pyx_kp_u_Markets_are_not_ready_No_market __pyx_string_tab[37]
#define __pyx_kp_u_No_active_maker_orders __pyx_string_tab[38]
#define __pyx_kp_u_None __pyx_string_tab[39]
#define __pyx_kp_u_Optional_MovingPriceBand __pyx_string_tab[40]
#define __pyx_kp_u_Optional_pd_DataFrame __pyx_string_tab[41]
#define __pyx_kp_u_Order_Adjust __pyx_string_tab[42]
#define __pyx_kp_u_Order_is_below_minimum_spread __pyx_string_tab[43]
#define __pyx_kp_u_Orders __pyx_string_tab[44]
#define __pyx_kp_u_Parameter_price_ceiling_cannot_b __pyx_string_tab[45]
#define __pyx_kp_u_Ping_pong_removed __pyx_string_tab[46]
#define __pyx_kp_u_Ref_Price __pyx_string_tab[47]
#define __pyx_kp_u_Ref_Price_MidPrice __pyx_string_tab[48]
#define __pyx_kp_u_Target __pyx_string_tab[49]
#define __pyx_kp_u_Target_Value __pyx_string_tab[50]
#define __pyx_kp_u_Total_Balance __pyx_string_tab[51]
#define __pyx_kp_u_Unrecognized_price_type_string __pyx_string_tab[52]
#define __pyx_kp_u_WARNINGS __pyx_string_tab[53]
#define __pyx_kp_u_WARNING_Some_markets_are_not_con __pyx_string_tab[54]
#define __pyx_kp_u__13 __pyx_string_tab[55]
#define __pyx_kp_u__14 __pyx_string_tab[56]
#define __pyx_kp_u__15 __pyx_string_tab[57]
#define __pyx_kp_u__16 __pyx_string_tab[58]
#define __pyx_kp_u__17 __pyx_string_tab[59]
#define __pyx_kp_u__18 __pyx_string_tab[60]
#define __pyx_kp_u__19 __pyx_string_tab[61]
#define __pyx_kp_u__20 __pyx_string_tab[62]
#define __pyx_kp_u__21 __pyx_string_tab[63]
#define __pyx_kp_u__22 __pyx_string_tab[64]
#define __pyx_kp_u__23 __pyx_string_tab[65]
#define __pyx_kp_u__24 __pyx_string_tab[66]
#define __pyx_kp_u__25 __pyx_string_tab[67]
#define __pyx_kp_u__26 __pyx_string_tab[68]
#define __pyx_kp_u_ask_orders_at_Size_Price __pyx_string_tab[69]
#define __pyx_kp_u_bid_orders_at_Size_Price __pyx_string_tab[70]
#define __pyx_kp_u_buy_orders __pyx_string_tab[71]
#define __pyx_kp_u_disable __pyx_string_tab[72]
#define __pyx_kp_u_enable __pyx_string_tab[73]
#define __pyx_kp_u_filled __pyx_string_tab[74]
#define __pyx_kp_u_gc __pyx_string_tab[75]
#define __pyx_kp_u_has_been_completely_filled __pyx_string_tab[76]
#define __pyx_kp_u_hummingbot_strategy_pure_market_2 __pyx_string_tab[77]
#define __pyx_kp_u_hummingbot_strategy_pure_market_3 __pyx_string_tab[78]
#define __pyx_kp_u_hummingbot_strategy_pure_market_4 __pyx_string_tab[79]
#define __pyx_kp_u_hummingbot_strategy_pure_market_5 __pyx_string_tab[80]
#define __pyx_kp_u_hummingbot_strategy_pure_market_6 __pyx_string_tab[81]
#define __pyx_kp_u_hummingbot_strategy_pure_market_7 __pyx_string_tab[82]
#define __pyx_kp_u_is_filled __pyx_string_tab[83]
#define __pyx_kp_u_isenabled __pyx_string_tab[84]
#define __pyx_kp_u_numpy__core_multiarray_failed_to __pyx_string_tab[85]
#define __pyx_kp_u_numpy__core_umath_failed_to_impo __pyx_string_tab[86]
#define __pyx_kp_u_pd_DataFrame __pyx_string_tab[87]
#define __pyx_kp_u_self__events_cannot_be_converted __pyx_string_tab[88]
#define __pyx_kp_u_sell_orders __pyx_string_tab[89]
#define __pyx_kp_u_stringsource __pyx_string_tab[90]
#define __pyx_n_u_Age __pyx_string_tab[91]
#define __pyx_n_u_AssetPriceDelegate __pyx_string_tab[92]
#define __pyx_n_u_BUY __pyx_string_tab[93]
#define __pyx_n_u_BestAsk __pyx_string_tab[94]
#define __pyx_n_u_BestBid __pyx_string_tab[95]
#define __pyx_n_u_Buy __pyx_string_tab[96]
#define __pyx_n_u_CONNECTED __pyx_string_tab[97]
#define __pyx_n_u_CreatedPairOfOrders __pyx_string_tab[98]
#define __pyx_n_u_Custom __pyx_string_tab[99]
#define __pyx_n_u_DataFrame __pyx_string_tab[100]
#define __pyx_n_u_Decimal __pyx_string_tab[101]
#define __pyx_n_u_Dict __pyx_string_tab[102]
#define __pyx_n_u_Exchange __pyx_string_tab[103]
#define __pyx_n_u_ExchangeBase __pyx_string_tab[104]
#define __pyx_n_u_HangingOrdersTracker __pyx_string_tab[105]
#define __pyx_n_u_INFO __pyx_string_tab[106]
#define __pyx_n_u_InventoryCost __pyx_string_tab[107]
#define __pyx_n_u_InventoryCostPriceDelegate __pyx_string_tab[108]
#define __pyx_n_u_LIMIT __pyx_string_tab[109]
#define __pyx_n_u_LastOwnTrade __pyx_string_tab[110]
#define __pyx_n_u_LastTrade __pyx_string_tab[111]
#define __pyx_n_u_Level __pyx_string_tab[112]
#define __pyx_n_u_LimitOrder __pyx_string_tab[113]
#define __pyx_n_u_List __pyx_string_tab[114]
#define __pyx_n_u_Market __pyx_string_tab[115]
#define __pyx_n_u_MarketTradingPairTuple __pyx_string_tab[116]
#define __pyx_n_u_MidPrice __pyx_string_tab[117]
#define __pyx_n_u_MovingPriceBand __pyx_string_tab[118]
#define __pyx_n_u_NaN __pyx_string_tab[119]
#define __pyx_n_u_NetworkStatus __pyx_string_tab[120]
#define __pyx_n_u_OPTION_LOG_ALL __pyx_string_tab[121]
#define __pyx_n_u_OPTION_LOG_CREATE_ORDER __pyx_string_tab[122]
#define __pyx_n_u_OPTION_LOG_MAKER_ORDER_FILLED __pyx_string_tab[123]
#define __pyx_n_u_OPTION_LOG_STATUS_REPORT __pyx_string_tab[124]
#define __pyx_n_u_Optional __pyx_string_tab[125]
#define __pyx_n_u_OrderType __pyx_string_tab[126]
#define __pyx_n_u_Price __pyx_string_tab[127]
#define __pyx_n_u_PriceSize __pyx_string_tab[128]
#define __pyx_n_u_PriceType __pyx_string_tab[129]
#define __pyx_n_u_Proposal __pyx_string_tab[130]
#define __pyx_n_u_PureMarketMakingOrderTracker __pyx_string_tab[131]
#define __pyx_n_u_PureMarketMakingStrategy __pyx_string_tab[132]
#define __pyx_n_u_PureMarketMakingStrategy___reduc __pyx_string_tab[133]
#define __pyx_n_u_PureMarketMakingStrategy___setst __pyx_string_tab[134]
#define __pyx_n_u_PureMarketMakingStrategy_active __pyx_string_tab[135]
#define __pyx_n_u_PureMarketMakingStrategy_adjuste __pyx_string_tab[136]
#define __pyx_n_u_PureMarketMakingStrategy_all_mar __pyx_string_tab[137]
#define __pyx_n_u_PureMarketMakingStrategy_c_apply __pyx_string_tab[138]
#define __pyx_n_u_PureMarketMakingStrategy_c_cance __pyx_string_tab[139]
#define __pyx_n_u_PureMarketMakingStrategy_c_execu __pyx_string_tab[140]
#define __pyx_n_u_PureMarketMakingStrategy_c_start __pyx_string_tab[141]
#define __pyx_n_u_PureMarketMakingStrategy_cancel __pyx_string_tab[142]
#define __pyx_n_u_PureMarketMakingStrategy_execute __pyx_string_tab[143]
#define __pyx_n_u_PureMarketMakingStrategy_format __pyx_string_tab[144]
#define __pyx_n_u_PureMarketMakingStrategy_get_mid __pyx_string_tab[145]
#define __pyx_n_u_PureMarketMakingStrategy_get_pri __pyx_string_tab[146]
#define __pyx_n_u_PureMarketMakingStrategy_get_pri_2 __pyx_string_tab[147]
#define __pyx_n_u_PureMarketMakingStrategy_init_pa __pyx_string_tab[148]
#define __pyx_n_u_PureMarketMakingStrategy_invento __pyx_string_tab[149]
#define __pyx_n_u_PureMarketMakingStrategy_logger __pyx_string_tab[150]
#define __pyx_n_u_PureMarketMakingStrategy_market __pyx_string_tab[151]
#define __pyx_n_u_PureMarketMakingStrategy_notify __pyx_string_tab[152]
#define __pyx_n_u_PureMarketMakingStrategy_pure_mm __pyx_string_tab[153]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[154]
#define __pyx_n_u_SELL __pyx_string_tab[155]
#define __pyx_n_u_Sell __pyx_string_tab[156]
#define __pyx_n_u_Spread __pyx_string_tab[157]
#define __pyx_n_u_StrategyBase __pyx_string_tab[158]
#define __pyx_n_u_Timestamp __pyx_string_tab[159]
#define __pyx_n_u_TradeType __pyx_string_tab[160]
#define __pyx_n_u_Type __pyx_string_tab[161]
#define __pyx_n_u_active_markets __pyx_string_tab[162]
#define __pyx_n_u_active_non_hanging_orders __pyx_string_tab[163]
#define __pyx_n_u_active_orders __pyx_string_tab[164]
#define __pyx_n_u_active_orders_df __pyx_string_tab[165]
#define __pyx_n_u_active_orders_df_locals_lambda __pyx_string_tab[166]
#define __pyx_n_u_add_as_hanging_order __pyx_string_tab[167]
#define __pyx_n_u_add_current_pairs_of_proposal_or __pyx_string_tab[168]
#define __pyx_n_u_add_transaction_costs_to_orders __pyx_string_tab[169]
#define __pyx_n_u_adjusted_available_balance_for_o __pyx_string_tab[170]
#define __pyx_n_u_age __pyx_string_tab[171]
#define __pyx_n_u_all __pyx_string_tab[172]
#define __pyx_n_u_all_markets_ready __pyx_string_tab[173]
#define __pyx_n_u_all_non_hanging_orders __pyx_string_tab[174]
#define __pyx_n_u_all_orders __pyx_string_tab[175]
#define __pyx_n_u_amount __pyx_string_tab[176]
#define __pyx_n_u_amount_orig __pyx_string_tab[177]
#define __pyx_n_u_apply __pyx_string_tab[178]
#define __pyx_n_u_ask_order_level_spreads __pyx_string_tab[179]
#define __pyx_n_u_ask_order_optimization_depth __pyx_string_tab[180]
#define __pyx_n_u_ask_price __pyx_string_tab[181]
#define __pyx_n_u_ask_ratio __pyx_string_tab[182]
#define __pyx_n_u_ask_spread __pyx_string_tab[183]
#define __pyx_n_u_asset_price_delegate __pyx_string_tab[184]
#define __pyx_n_u_assets_df __pyx_string_tab[185]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[186]
#define __pyx_n_u_available_base_balance __pyx_string_tab[187]
#define __pyx_n_u_available_quote_balance __pyx_string_tab[188]
#define __pyx_n_u_balance_warning __pyx_string_tab[189]
#define __pyx_n_u_base_asset __pyx_string_tab[190]
#define __pyx_n_u_base_asset_amount __pyx_string_tab[191]
#define __pyx_n_u_base_asset_range __pyx_string_tab[192]
#define __pyx_n_u_base_asset_ratio __pyx_string_tab[193]
#define __pyx_n_u_base_asset_value __pyx_string_tab[194]
#define __pyx_n_u_base_balance __pyx_string_tab[195]
#define __pyx_n_u_base_currency __pyx_string_tab[196]
#define __pyx_n_u_base_ratio __pyx_string_tab[197]
#define __pyx_n_u_base_value __pyx_string_tab[198]
#define __pyx_n_u_best_ask __pyx_string_tab[199]
#define __pyx_n_u_best_bid __pyx_string_tab[200]
#define __pyx_n_u_bid_ask_ratios __pyx_string_tab[201]
#define __pyx_n_u_bid_order_level_spreads __pyx_string_tab[202]
#define __pyx_n_u_bid_order_optimization_depth __pyx_string_tab[203]
#define __pyx_n_u_bid_price __pyx_string_tab[204]
#define __pyx_n_u_bid_ratio __pyx_string_tab[205]
#define __pyx_n_u_bid_spread __pyx_string_tab[206]
#define __pyx_n_u_bool __pyx_string_tab[207]
#define __pyx_n_u_buy __pyx_string_tab[208]
#define __pyx_n_u_buys __pyx_string_tab[209]
#define __pyx_n_u_calculate_total_order_size __pyx_string_tab[210]
#define __pyx_n_u_cancel_order __pyx_string_tab[211]
#define __pyx_n_u_candidate_hanging_orders __pyx_string_tab[212]
#define __pyx_n_u_candidate_hanging_orders_from_pa __pyx_string_tab[213]
#define __pyx_n_u_ceil __pyx_string_tab[214]
#define __pyx_n_u_check_and_update_price_band __pyx_string_tab[215]
#define __pyx_n_u_check_price_ceiling_exceeded __pyx_string_tab[216]
#define __pyx_n_u_check_price_floor_exceeded __pyx_string_tab[217]
#define __pyx_n_u_class_getitem __pyx_string_tab[218]
#define __pyx_n_u_client_order_id __pyx_string_tab[219]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[220]
#define __pyx_n_u_close __pyx_string_tab[221]
#define __pyx_n_u_cls __pyx_string_tab[222]
#define __pyx_n_u_columns __pyx_string_tab[223]
#define __pyx_n_u_concat __pyx_string_tab[224]
#define __pyx_n_u_current_created_pairs_of_orders __pyx_string_tab[225]
#define __pyx_n_u_current_timestamp __pyx_string_tab[226]
#define __pyx_n_u_custom __pyx_string_tab[227]
#define __pyx_n_u_data __pyx_string_tab[228]
#define __pyx_n_u_data_types __pyx_string_tab[229]
#define __pyx_n_u_decimal __pyx_string_tab[230]
#define __pyx_n_u_df __pyx_string_tab[231]
#define __pyx_n_u_df_lines __pyx_string_tab[232]
#define __pyx_n_u_dict __pyx_string_tab[233]
#define __pyx_n_u_display_name __pyx_string_tab[234]
#define __pyx_n_u_enabled __pyx_string_tab[235]
#define __pyx_n_u_enumerate __pyx_string_tab[236]
#define __pyx_n_u_execute_orders_proposal __pyx_string_tab[237]
#define __pyx_n_u_filled_order_delay __pyx_string_tab[238]
#define __pyx_n_u_first_col_length __pyx_string_tab[239]
#define __pyx_n_u_float __pyx_string_tab[240]
#define __pyx_n_u_floor __pyx_string_tab[241]
#define __pyx_n_u_format __pyx_string_tab[242]
#define __pyx_n_u_format_status __pyx_string_tab[243]
#define __pyx_n_u_formatters __pyx_string_tab[244]
#define __pyx_n_u_func __pyx_string_tab[245]
#define __pyx_n_u_genexpr __pyx_string_tab[246]
#define __pyx_n_u_get __pyx_string_tab[247]
#define __pyx_n_u_getLogger __pyx_string_tab[248]
#define __pyx_n_u_get_available_balance __pyx_string_tab[249]
#define __pyx_n_u_get_balance __pyx_string_tab[250]
#define __pyx_n_u_get_limit_orders __pyx_string_tab[251]
#define __pyx_n_u_get_maker_order_type __pyx_string_tab[252]
#define __pyx_n_u_get_mid_price __pyx_string_tab[253]
#define __pyx_n_u_get_price __pyx_string_tab[254]
#define __pyx_n_u_get_price_by_type __pyx_string_tab[255]
#define __pyx_n_u_get_price_for_volume __pyx_string_tab[256]
#define __pyx_n_u_get_price_type __pyx_string_tab[257]
#define __pyx_n_u_getstate __pyx_string_tab[258]
#define __pyx_n_u_hang __pyx_string_tab[259]
#define __pyx_n_u_hanging_order_ids __pyx_string_tab[260]
#define __pyx_n_u_hanging_orders_cancel_pct __pyx_string_tab[261]
#define __pyx_n_u_hanging_orders_cancel_pct_2 __pyx_string_tab[262]
#define __pyx_n_u_hanging_orders_enabled __pyx_string_tab[263]
#define __pyx_n_u_hanging_orders_tracker __pyx_string_tab[264]
#define __pyx_n_u_hb_app_notification __pyx_string_tab[265]
#define __pyx_n_u_header __pyx_string_tab[266]
#define __pyx_n_u_high_water_mark __pyx_string_tab[267]
#define __pyx_n_u_high_water_mark_ratio __pyx_string_tab[268]
#define __pyx_n_u_hummingbot_connector_exchange_ba __pyx_string_tab[269]
#define __pyx_n_u_hummingbot_core_data_type_common __pyx_string_tab[270]
#define __pyx_n_u_hummingbot_core_data_type_limit __pyx_string_tab[271]
#define __pyx_n_u_hummingbot_core_network_iterator __pyx_string_tab[272]
#define __pyx_n_u_hummingbot_core_utils __pyx_string_tab[273]
#define __pyx_n_u_hummingbot_strategy_asset_price __pyx_string_tab[274]
#define __pyx_n_u_hummingbot_strategy_hanging_orde __pyx_string_tab[275]
#define __pyx_n_u_hummingbot_strategy_market_tradi __pyx_string_tab[276]
#define __pyx_n_u_hummingbot_strategy_pure_market __pyx_string_tab[277]
#define __pyx_n_u_hummingbot_strategy_strategy_bas __pyx_string_tab[278]
#define __pyx_n_u_hummingbot_strategy_utils __pyx_string_tab[279]
#define __pyx_n_u_idx __pyx_string_tab[280]
#define __pyx_n_u_in_flight_cancels __pyx_string_tab[281]
#define __pyx_n_u_index __pyx_string_tab[282]
#define __pyx_n_u_info __pyx_string_tab[283]
#define __pyx_n_u_init_params __pyx_string_tab[284]
#define __pyx_n_u_inner __pyx_string_tab[285]
#define __pyx_n_u_int __pyx_string_tab[286]
#define __pyx_n_u_inventory_cost __pyx_string_tab[287]
#define __pyx_n_u_inventory_cost_price_delegate __pyx_string_tab[288]
#define __pyx_n_u_inventory_range_multiplier __pyx_string_tab[289]
#define __pyx_n_u_inventory_skew_calculator __pyx_string_tab[290]
#define __pyx_n_u_inventory_skew_df __pyx_string_tab[291]
#define __pyx_n_u_inventory_skew_enabled __pyx_string_tab[292]
#define __pyx_n_u_inventory_skew_stats_data_frame __pyx_string_tab[293]
#define __pyx_n_u_inventory_target_base_pct __pyx_string_tab[294]
#define __pyx_n_u_is_buy __pyx_string_tab[295]
#define __pyx_n_u_is_coroutine __pyx_string_tab[296]
#define __pyx_n_u_is_hanging_order __pyx_string_tab[297]
#define __pyx_n_u_is_nan __pyx_string_tab[298]
#define __pyx_n_u_is_order_id_in_hanging_orders __pyx_string_tab[299]
#define __pyx_n_u_is_potential_hanging_order __pyx_string_tab[300]
#define __pyx_n_u_items __pyx_string_tab[301]
#define __pyx_n_u_join __pyx_string_tab[302]
#define __pyx_n_u_key __pyx_string_tab[303]
#define __pyx_n_u_lambda __pyx_string_tab[304]
#define __pyx_n_u_last_own_trade_price __pyx_string_tab[305]
#define __pyx_n_u_last_price __pyx_string_tab[306]
#define __pyx_n_u_len __pyx_string_tab[307]
#define __pyx_n_u_level __pyx_string_tab[308]
#define __pyx_n_u_level_for_calculation __pyx_string_tab[309]
#define __pyx_n_u_limit_orders __pyx_string_tab[310]
#define __pyx_n_u_line __pyx_string_tab[311]
#define __pyx_n_u_lines __pyx_string_tab[312]
#define __pyx_n_u_log_with_clock __pyx_string_tab[313]
#define __pyx_n_u_logger __pyx_string_tab[314]
#define __pyx_n_u_logging __pyx_string_tab[315]
#define __pyx_n_u_logging_options __pyx_string_tab[316]
#define __pyx_n_u_low_water_mark __pyx_string_tab[317]
#define __pyx_n_u_low_water_mark_ratio __pyx_string_tab[318]
#define __pyx_n_u_lvl_buy __pyx_string_tab[319]
#define __pyx_n_u_lvl_sell __pyx_string_tab[320]
#define __pyx_n_u_main __pyx_string_tab[321]
#define __pyx_n_u_map_df_to_str __pyx_string_tab[322]
#define __pyx_n_u_market __pyx_string_tab[323]
#define __pyx_n_u_market_books __pyx_string_tab[324]
#define __pyx_n_u_market_info __pyx_string_tab[325]
#define __pyx_n_u_market_info_to_active_orders __pyx_string_tab[326]
#define __pyx_n_u_market_pair_to_active_orders __pyx_string_tab[327]
#define __pyx_n_u_market_status_data_frame __pyx_string_tab[328]
#define __pyx_n_u_market_trading_pair_tuples __pyx_string_tab[329]
#define __pyx_n_u_markets_columns __pyx_string_tab[330]
#define __pyx_n_u_markets_data __pyx_string_tab[331]
#define __pyx_n_u_markets_df __pyx_string_tab[332]
#define __pyx_n_u_math __pyx_string_tab[333]
#define __pyx_n_u_max __pyx_string_tab[334]
#define __pyx_n_u_max_order_age __pyx_string_tab[335]
#define __pyx_n_u_mid_price __pyx_string_tab[336]
#define __pyx_n_u_minimum_spread __pyx_string_tab[337]
#define __pyx_n_u_module __pyx_string_tab[338]
#define __pyx_n_u_moving_price_band __pyx_string_tab[339]
#define __pyx_n_u_moving_price_band_enabled __pyx_string_tab[340]
#define __pyx_n_u_msg __pyx_string_tab[341]
#define __pyx_n_u_name __pyx_string_tab[342]
#define __pyx_n_u_name_2 __pyx_string_tab[343]
#define __pyx_n_u_nan __pyx_string_tab[344]
#define __pyx_n_u_network_status __pyx_string_tab[345]
#define __pyx_n_u_network_warning __pyx_string_tab[346]
#define __pyx_n_u_next __pyx_string_tab[347]
#define __pyx_n_u_no_sells __pyx_string_tab[348]
#define __pyx_n_u_non_hanging __pyx_string_tab[349]
#define __pyx_n_u_normalize __pyx_string_tab[350]
#define __pyx_n_u_notify_hb_app __pyx_string_tab[351]
#define __pyx_n_u_notify_hb_app_with_timestamp __pyx_string_tab[352]
#define __pyx_n_u_np __pyx_string_tab[353]
#define __pyx_n_u_numpy __pyx_string_tab[354]
#define __pyx_n_u_o __pyx_string_tab[355]
#define __pyx_n_u_order __pyx_string_tab[356]
#define __pyx_n_u_order_age __pyx_string_tab[357]
#define __pyx_n_u_order_amount __pyx_string_tab[358]
#define __pyx_n_u_order_id __pyx_string_tab[359]
#define __pyx_n_u_order_level_amount __pyx_string_tab[360]
#define __pyx_n_u_order_level_spread __pyx_string_tab[361]
#define __pyx_n_u_order_levels __pyx_string_tab[362]
#define __pyx_n_u_order_optimization_enabled __pyx_string_tab[363]
#define __pyx_n_u_order_override __pyx_string_tab[364]
#define __pyx_n_u_order_refresh_time __pyx_string_tab[365]
#define __pyx_n_u_order_refresh_tolerance_pct __pyx_string_tab[366]
#define __pyx_n_u_p __pyx_string_tab[367]
#define __pyx_n_u_pandas __pyx_string_tab[368]
#define __pyx_n_u_pd __pyx_string_tab[369]
#define __pyx_n_u_percent __pyx_string_tab[370]
#define __pyx_n_u_ping_pong_enabled __pyx_string_tab[371]
#define __pyx_n_u_pmm_logger __pyx_string_tab[372]
#define __pyx_n_u_pop __pyx_string_tab[373]
#define __pyx_n_u_price __pyx_string_tab[374]
#define __pyx_n_u_price_band_refresh_time __pyx_string_tab[375]
#define __pyx_n_u_price_ceiling __pyx_string_tab[376]
#define __pyx_n_u_price_ceiling_pct __pyx_string_tab[377]
#define __pyx_n_u_price_floor __pyx_string_tab[378]
#define __pyx_n_u_price_floor_pct __pyx_string_tab[379]
#define __pyx_n_u_price_provider __pyx_string_tab[380]
#define __pyx_n_u_price_type __pyx_string_tab[381]
#define __pyx_n_u_price_type_str __pyx_string_tab[382]
#define __pyx_n_u_process_order_fill_event __pyx_string_tab[383]
#define __pyx_n_u_process_tick __pyx_string_tab[384]
#define __pyx_n_u_proposal __pyx_string_tab[385]
#define __pyx_n_u_pure_market_making_order_tracker __pyx_string_tab[386]
#define __pyx_n_u_pure_mm_assets_df __pyx_string_tab[387]
#define __pyx_n_u_pyx_state __pyx_string_tab[388]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[389]
#define __pyx_n_u_qualname __pyx_string_tab[390]
#define __pyx_n_u_quantity __pyx_string_tab[391]
#define __pyx_n_u_quote_asset __pyx_string_tab[392]
#define __pyx_n_u_quote_asset_amount __pyx_string_tab[393]
#define __pyx_n_u_quote_asset_ratio __pyx_string_tab[394]
#define __pyx_n_u_quote_asset_value __pyx_string_tab[395]
#define __pyx_n_u_quote_balance __pyx_string_tab[396]
#define __pyx_n_u_quote_currency __pyx_string_tab[397]
#define __pyx_n_u_quote_ratio __pyx_string_tab[398]
#define __pyx_n_u_ready __pyx_string_tab[399]
#define __pyx_n_u_reduce __pyx_string_tab[400]
#define __pyx_n_u_reduce_cython __pyx_string_tab[401]
#define __pyx_n_u_reduce_ex __pyx_string_tab[402]
#define __pyx_n_u_ref_price __pyx_string_tab[403]
#define __pyx_n_u_regex __pyx_string_tab[404]
#define __pyx_n_u_register_events __pyx_string_tab[405]
#define __pyx_n_u_replace __pyx_string_tab[406]
#define __pyx_n_u_result_price __pyx_string_tab[407]
#define __pyx_n_u_return __pyx_string_tab[408]
#define __pyx_n_u_reverse __pyx_string_tab[409]
#define __pyx_n_u_round __pyx_string_tab[410]
#define __pyx_n_u_s __pyx_string_tab[411]
#define __pyx_n_u_s_decimal_neg_one __pyx_string_tab[412]
#define __pyx_n_u_s_decimal_zero __pyx_string_tab[413]
#define __pyx_n_u_self __pyx_string_tab[414]
#define __pyx_n_u_sell __pyx_string_tab[415]
#define __pyx_n_u_sell_order __pyx_string_tab[416]
#define __pyx_n_u_sells __pyx_string_tab[417]
#define __pyx_n_u_send __pyx_string_tab[418]
#define __pyx_n_u_set_name __pyx_string_tab[419]
#define __pyx_n_u_setdefault __pyx_string_tab[420]
#define __pyx_n_u_setstate __pyx_string_tab[421]
#define __pyx_n_u_setstate_cython __pyx_string_tab[422]
#define __pyx_n_u_should_wait_order_cancel_confirm __pyx_string_tab[423]
#define __pyx_n_u_size __pyx_string_tab[424]
#define __pyx_n_u_sort __pyx_string_tab[425]
#define __pyx_n_u_sorted __pyx_string_tab[426]
#define __pyx_n_u_split __pyx_string_tab[427]
#define __pyx_n_u_split_order_levels_enabled __pyx_string_tab[428]
#define __pyx_n_u_spread __pyx_string_tab[429]
#define __pyx_n_u_status_report_interval __pyx_string_tab[430]
#define __pyx_n_u_str __pyx_string_tab[431]
#define __pyx_n_u_strategy_current_hanging_orders __pyx_string_tab[432]
#define __pyx_n_u_strftime __pyx_string_tab[433]
#define __pyx_n_u_super __pyx_string_tab[434]
#define __pyx_n_u_switch __pyx_string_tab[435]
#define __pyx_n_u_take_if_crossed __pyx_string_tab[436]
#define __pyx_n_u_target_base_amount __pyx_string_tab[437]
#define __pyx_n_u_target_base_amount_in_quote __pyx_string_tab[438]
#define __pyx_n_u_target_base_ratio __pyx_string_tab[439]
#define __pyx_n_u_target_quote_amount __pyx_string_tab[440]
#define __pyx_n_u_test __pyx_string_tab[441]
#define __pyx_n_u_throw __pyx_string_tab[442]
#define __pyx_n_u_to_show_current_pct __pyx_string_tab[443]
#define __pyx_n_u_to_string __pyx_string_tab[444]
#define __pyx_n_u_total_in_quote __pyx_string_tab[445]
#define __pyx_n_u_total_order_size __pyx_string_tab[446]
#define __pyx_n_u_total_order_size_ratio __pyx_string_tab[447]
#define __pyx_n_u_total_value __pyx_string_tab[448]
#define __pyx_n_u_total_value_in_quote __pyx_string_tab[449]
#define __pyx_n_u_trade_type __pyx_string_tab[450]
#define __pyx_n_u_trading_pair __pyx_string_tab[451]
#define __pyx_n_u_typing __pyx_string_tab[452]
#define __pyx_n_u_unit __pyx_string_tab[453]
#define __pyx_n_u_unregister_events __pyx_string_tab[454]
#define __pyx_n_u_update __pyx_string_tab[455]
#define __pyx_n_u_update_strategy_orders_with_equi __pyx_string_tab[456]
#define __pyx_n_u_value __pyx_string_tab[457]
#define __pyx_n_u_values __pyx_string_tab[458]
#define __pyx_n_u_warning __pyx_string_tab[459]
#define __pyx_n_u_warning_lines __pyx_string_tab[460]
#define __pyx_n_u_x __pyx_string_tab[461]
#define __pyx_n_u_zip __pyx_string_tab[462]
#define __pyx_kp_b_PyObject_double_double_double_do __pyx_string_tab[463]
#define __pyx_kp_b_iso88591_0_t_Qa __pyx_string_tab[464]
#define __pyx_kp_b_iso88591_1_3 __pyx_string_tab[465]
#define __pyx_kp_b_iso88591_1_4Ct1_4_Cy_D_c_N_4AYa_N_4AT_5 __pyx_string_tab[466]
#define __pyx_kp_b_iso88591_1_m1_Ja_0_4UUVVZZ_5Qd_J_Nddhhi __pyx_string_tab[467]
#define __pyx_kp_b_iso88591_4q __pyx_string_tab[468]
#define __pyx_kp_b_iso88591_5Ya_l_A_M_q_uAV_q_QfL_av_C1A_q __pyx_string_tab[469]
#define __pyx_kp_b_iso88591_AQ __pyx_string_tab[470]
#define __pyx_kp_b_iso88591_A_2_2_0_11_1_0_31_9_0_31_9_07q2 __pyx_string_tab[471]
#define __pyx_kp_b_iso88591_A_4_bbc_a_4_Ct_55Fa_4_T_6_Ya_d_b __pyx_string_tab[472]
#define __pyx_kp_b_iso88591_A_c_1_q __pyx_string_tab[473]
#define __pyx_kp_b_iso88591_A_s_1F_Zt1 __pyx_string_tab[474]
#define __pyx_kp_b_iso88591_CCdde_q_z_VWW_ggh_4_Cy_d_TT_2U __pyx_string_tab[475]
#define __pyx_kp_b_iso88591_Q __pyx_string_tab[476]
#define __pyx_kp_b_iso88591_Q_2 __pyx_string_tab[477]
#define __pyx_kp_b_iso88591_Q_t_4_a __pyx_string_tab[478]
#define __pyx_kp_b_iso88591_XQ_Q_9A_Cq_9A_Cq_9A_Cq_9A_Cq_9A __pyx_string_tab[479]
#define __pyx_kp_b_iso88591__27 __pyx_string_tab[480]
#define __pyx_kp_b_iso88591_l_A_Ja_A_3aq_e_D_QFWWX_D_88VVWW __pyx_string_tab[481]
#define __pyx_kp_b_iso88591_q __pyx_string_tab[482]
#define __pyx_kp_b_iso88591_q_4t1_1_WAT_WAT_1_4q_4_a_WAQd_7 __pyx_string_tab[483]
#define __pyx_kp_b_iso88591_q_t_1 __pyx_string_tab[484]
#define __pyx_kp_b_iso88591_z __pyx_string_tab[485]
#define __pyx_float_30_0 __pyx_number_tab[0]
#define __pyx_float_60_0 __pyx_number_tab[1]
#define __pyx_float_1800_0 __pyx_number_tab[2]
#define __pyx_int_0 __pyx_number_tab[3]
#define __pyx_int_neg_1 __pyx_number_tab[4]
#define __pyx_int_1 __pyx_number_tab[5]
#define __pyx_int_4 __pyx_number_tab[6]
#define __pyx_int_8 __pyx_number_tab[7]
#define __pyx_int_16 __pyx_number_tab[8]
#define __pyx_int_32 __pyx_number_tab[9]
#define __pyx_int_900 __pyx_number_tab[10]
#define __pyx_int_0x7fffffffffffffff __pyx_number_tab[11]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_5clock_Clock);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_6pubsub_PubSub);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_dtype);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flatiter);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_broadcast);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ndarray);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_generic);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_number);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_integer);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_inexact);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_floating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flexible);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_character);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ufunc);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr);
  Py_CLEAR(clear_module_state->__pyx_k_);
  Py_CLEAR(clear_module_state->__pyx_k__2);
  Py_CLEAR(clear_module_state->__pyx_k__3);
  Py_CLEAR(clear_module_state->__pyx_k__4);
  Py_CLEAR(clear_module_state->__pyx_k__5);
  Py_CLEAR(clear_module_state->__pyx_k__6);
  Py_CLEAR(clear_module_state->__pyx_k__7);
  Py_CLEAR(clear_module_state->__pyx_k__8);
  Py_CLEAR(clear_module_state->__pyx_k__9);
  Py_CLEAR(clear_module_state->__pyx_k__10);
  Py_CLEAR(clear_module_state->__pyx_k__11);
  Py_CLEAR(clear_module_state->__pyx_k__12);
  for (int i=0; i<24; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<486; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<12; ++i) { Py_CLEAR(clear_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* Generator.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_GeneratorType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_5clock_Clock);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_6pubsub_PubSub);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_dtype);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flatiter);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_broadcast);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ndarray);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_generic);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_number);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_integer);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_inexact);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_floating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flexible);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_character);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ufunc);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr);
  Py_VISIT(traverse_module_state->__pyx_k_);
  Py_VISIT(traverse_module_state->__pyx_k__2);
  Py_VISIT(traverse_module_state->__pyx_k__3);
  Py_VISIT(traverse_module_state->__pyx_k__4);
  Py_VISIT(traverse_module_state->__pyx_k__5);
  Py_VISIT(traverse_module_state->__pyx_k__6);
  Py_VISIT(traverse_module_state->__pyx_k__7);
  Py_VISIT(traverse_module_state->__pyx_k__8);
  Py_VISIT(traverse_module_state->__pyx_k__9);
  Py_VISIT(traverse_module_state->__pyx_k__10);
  Py_VISIT(traverse_module_state->__pyx_k__11);
  Py_VISIT(traverse_module_state->__pyx_k__12);
  for (int i=0; i<24; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<486; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<12; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* Generator.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_GeneratorType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":243
 *         cdef int type_num
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp itemsize(self) noexcept nogil:
 *             return PyDataType_ELSIZE(self)
*/

static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_8itemsize_itemsize(PyArray_Descr *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":245
 *         @property
 *         cdef inline npy_intp itemsize(self) noexcept nogil:
 *             return PyDataType_ELSIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyDataType_ELSIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":243
 *         cdef int type_num
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp itemsize(self) noexcept nogil:
 *             return PyDataType_ELSIZE(self)
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":247
 *             return PyDataType_ELSIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp alignment(self) noexcept nogil:
 *             return PyDataType_ALIGNMENT(self)
*/

static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_9alignment_alignment(PyArray_Descr *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":249
 *         @property
 *         cdef inline npy_intp alignment(self) noexcept nogil:
 *             return PyDataType_ALIGNMENT(self)             # <<<<<<<<<<<<<<
 * 
 *         # Use fields/names with care as they may be NULL.  You must check
*/
  __pyx_r = PyDataType_ALIGNMENT(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":247
 *             return PyDataType_ELSIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp alignment(self) noexcept nogil:
 *             return PyDataType_ALIGNMENT(self)
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":253
 *         # Use fields/names with care as they may be NULL.  You must check
 *         # for this using PyDataType_HASFIELDS.
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline object fields(self):
 *             return <object>PyDataType_FIELDS(self)
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_6fields_fields(PyArray_Descr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("fields", 0);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":255
 *         @property
 *         cdef inline object fields(self):
 *             return <object>PyDataType_FIELDS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDataType_FIELDS(__pyx_v_self);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":253
 *         # Use fields/names with care as they may be NULL.  You must check
 *         # for this using PyDataType_HASFIELDS.
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline object fields(self):
 *             return <object>PyDataType_FIELDS(self)
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":257
 *             return <object>PyDataType_FIELDS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline tuple names(self):
 *             return <tuple>PyDataType_NAMES(self)
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_5names_names(PyArray_Descr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("names", 0);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":259
 *         @property
 *         cdef inline tuple names(self):
 *             return <tuple>PyDataType_NAMES(self)             # <<<<<<<<<<<<<<
 * 
 *         # Use PyDataType_HASSUBARRAY to test whether this field is
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDataType_NAMES(__pyx_v_self);
  __Pyx_INCREF(((PyObject*)__pyx_t_1));
  __pyx_r = ((PyObject*)__pyx_t_1);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":257
 *             return <object>PyDataType_FIELDS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline tuple names(self):
 *             return <tuple>PyDataType_NAMES(self)
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":264
 *         # valid (the pointer can be NULL). Most users should access
 *         # this field via the inline helper method PyDataType_SHAPE.
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:
 *             return PyDataType_SUBARRAY(self)
*/

static CYTHON_INLINE PyArray_ArrayDescr *__pyx_f_5numpy_5dtype_8subarray_subarray(PyArray_Descr *__pyx_v_self) {
  PyArray_ArrayDescr *__pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":266
 *         @property
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:
 *             return PyDataType_SUBARRAY(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyDataType_SUBARRAY(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":264
 *         # valid (the pointer can be NULL). Most users should access
 *         # this field via the inline helper method PyDataType_SHAPE.
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:
 *             return PyDataType_SUBARRAY(self)
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":268
 *             return PyDataType_SUBARRAY(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_uint64 flags(self) noexcept nogil:
 *             """The data types flags."""
*/

static CYTHON_INLINE npy_uint64 __pyx_f_5numpy_5dtype_5flags_flags(PyArray_Descr *__pyx_v_self) {
  npy_uint64 __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":271
 *         cdef inline npy_uint64 flags(self) noexcept nogil:
 *             """The data types flags."""
 *             return PyDataType_FLAGS(self)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = PyDataType_FLAGS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":268
 *             return PyDataType_SUBARRAY(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_uint64 flags(self) noexcept nogil:
 *             """The data types flags."""
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":280
 *     ctypedef class numpy.broadcast [object PyArrayMultiIterObject, check_size ignore]:
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int numiter(self) noexcept nogil:
 *             """The number of arrays that need to be broadcast to the same shape."""
*/

static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_7numiter_numiter(PyArrayMultiIterObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":283
 *         cdef inline int numiter(self) noexcept nogil:
 *             """The number of arrays that need to be broadcast to the same shape."""
 *             return PyArray_MultiIter_NUMITER(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_MultiIter_NUMITER(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":280
 *     ctypedef class numpy.broadcast [object PyArrayMultiIterObject, check_size ignore]:
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int numiter(self) noexcept nogil:
 *             """The number of arrays that need to be broadcast to the same shape."""
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":285
 *             return PyArray_MultiIter_NUMITER(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """The total broadcasted size."""
*/

static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_4size_size(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":288
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """The total broadcasted size."""
 *             return PyArray_MultiIter_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_MultiIter_SIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":285
 *             return PyArray_MultiIter_NUMITER(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """The total broadcasted size."""
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":290
 *             return PyArray_MultiIter_SIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp index(self) noexcept nogil:
 *             """The current (1-d) index into the broadcasted result."""
*/

static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_5index_index(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":293
 *         cdef inline npy_intp index(self) noexcept nogil:
 *             """The current (1-d) index into the broadcasted result."""
 *             return PyArray_MultiIter_INDEX(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_MultiIter_INDEX(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":290
 *             return PyArray_MultiIter_SIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp index(self) noexcept nogil:
 *             """The current (1-d) index into the broadcasted result."""
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":295
 *             return PyArray_MultiIter_INDEX(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int nd(self) noexcept nogil:
 *             """The number of dimensions in the broadcasted result."""
*/

static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_2nd_nd(PyArrayMultiIterObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":298
 *         cdef inline int nd(self) noexcept nogil:
 *             """The number of dimensions in the broadcasted result."""
 *             return PyArray_MultiIter_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_MultiIter_NDIM(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":295
 *             return PyArray_MultiIter_INDEX(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int nd(self) noexcept nogil:
 *             """The number of dimensions in the broadcasted result."""
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":300
 *             return PyArray_MultiIter_NDIM(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:
 *             """The shape of the broadcasted result."""
*/

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_9broadcast_10dimensions_dimensions(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":303
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:
 *             """The shape of the broadcasted result."""
 *             return PyArray_MultiIter_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_MultiIter_DIMS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":300
 *             return PyArray_MultiIter_NDIM(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:
 *             """The shape of the broadcasted result."""
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":305
 *             return PyArray_MultiIter_DIMS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline void** iters(self) noexcept nogil:
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
*/

static CYTHON_INLINE void **__pyx_f_5numpy_9broadcast_5iters_iters(PyArrayMultiIterObject *__pyx_v_self) {
  void **__pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":309
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
 *             On return, the iterators are adjusted for broadcasting."""
 *             return PyArray_MultiIter_ITERS(self)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = PyArray_MultiIter_ITERS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":305
 *             return PyArray_MultiIter_DIMS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline void** iters(self) noexcept nogil:
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":323
 *         # Instead, we use properties that map to the corresponding C-API functions.
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline PyObject* base(self) noexcept nogil:
 *             """Returns a borrowed reference to the object owning the data/memory.
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self) {
  PyObject *__pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":327
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 *             return PyArray_BASE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_BASE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":323
 *         # Instead, we use properties that map to the corresponding C-API functions.
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline PyObject* base(self) noexcept nogil:
 *             """Returns a borrowed reference to the object owning the data/memory.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":329
 *             return PyArray_BASE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline dtype descr(self):
 *             """Returns an owned reference to the dtype of the array.
*/

static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self) {
  PyArray_Descr *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyArray_Descr *__pyx_t_1;
  __Pyx_RefNannySetupContext("descr", 0);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":333
 *             """Returns an owned reference to the dtype of the array.
 *             """
 *             return <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyArray_DESCR(__pyx_v_self);
  __Pyx_INCREF((PyObject *)((PyArray_Descr *)__pyx_t_1));
  __pyx_r = ((PyArray_Descr *)__pyx_t_1);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":329
 *             return PyArray_BASE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline dtype descr(self):
 *             """Returns an owned reference to the dtype of the array.
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":335
 *             return <dtype>PyArray_DESCR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int ndim(self) noexcept nogil:
 *             """Returns the number of dimensions in the array.
*/

static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":339
 *             """Returns the number of dimensions in the array.
 *             """
 *             return PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_NDIM(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":335
 *             return <dtype>PyArray_DESCR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int ndim(self) noexcept nogil:
 *             """Returns the number of dimensions in the array.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":341
 *             return PyArray_NDIM(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp *shape(self) noexcept nogil:
 *             """Returns a pointer to the dimensions/shape of the array.
*/

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":347
 *             Can return NULL for 0-dimensional arrays.
 *             """
 *             return PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_DIMS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":341
 *             return PyArray_NDIM(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp *shape(self) noexcept nogil:
 *             """Returns a pointer to the dimensions/shape of the array.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":349
 *             return PyArray_DIMS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp *strides(self) noexcept nogil:
 *             """Returns a pointer to the strides of the array.
*/

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":354
 *             The number of elements matches the number of dimensions of the array (ndim).
 *             """
 *             return PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_STRIDES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":349
 *             return PyArray_DIMS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp *strides(self) noexcept nogil:
 *             """Returns a pointer to the strides of the array.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":356
 *             return PyArray_STRIDES(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """Returns the total size (in number of elements) of the array.
*/

static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":360
 *             """Returns the total size (in number of elements) of the array.
 *             """
 *             return PyArray_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_SIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":356
 *             return PyArray_STRIDES(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """Returns the total size (in number of elements) of the array.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":362
 *             return PyArray_SIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline char* data(self) noexcept nogil:
 *             """The pointer to the data buffer as a char*.
*/

static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self) {
  char *__pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":369
 *             of `PyArray_DATA()` instead, which returns a 'void*'.
 *             """
 *             return PyArray_BYTES(self)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = PyArray_BYTES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":362
 *             return PyArray_SIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline char* data(self) noexcept nogil:
 *             """The pointer to the data buffer as a char*.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":777
 * ctypedef long double complex clongdouble_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":778
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":777
 * ctypedef long double complex clongdouble_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":780
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":781
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":780
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":783
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":784
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":783
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":786
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":787
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":786
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":789
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":790
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":789
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":792
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":793
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
*/
  __pyx_t_1 = PyDataType_HASSUBARRAY(__pyx_v_d);
  if (__pyx_t_1) {

    /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":794
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_5numpy_5dtype_8subarray_subarray(__pyx_v_d)->shape;
    __Pyx_INCREF(((PyObject*)__pyx_t_2));
    __pyx_r = ((PyObject*)__pyx_t_2);
    goto __pyx_L0;

    /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":793
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
*/
  }

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":796
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_empty_tuple);
    __pyx_r = __pyx_mstate_global->__pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":792
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":995
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
*/

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":996
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
*/
  Py_INCREF(__pyx_v_base);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":997
 * cdef inline void set_array_base(ndarray arr, object base) except *:
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
*/
  __pyx_t_1 = PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 997, __pyx_L1_error)

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":995
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("numpy.set_array_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":999
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1000
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
*/
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1001
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
*/
  __pyx_t_1 = (__pyx_v_base == NULL);
  if (__pyx_t_1) {

    /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1002
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1001
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
*/
  }

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1003
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":999
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1007
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
*/

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1008
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1009
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")
*/
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 1009, __pyx_L3_error)

      /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1008
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1010
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
*/
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1010, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1011
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
*/
      __pyx_t_9 = NULL;
      __pyx_t_10 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_mstate_global->__pyx_kp_u_numpy__core_multiarray_failed_to};
        __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ImportError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1011, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1011, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1008
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1007
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1013
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
*/

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1015
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")
*/
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 1015, __pyx_L3_error)

      /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1016
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.umath failed to import")
 * 
*/
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1016, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1017
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
*/
      __pyx_t_9 = NULL;
      __pyx_t_10 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_mstate_global->__pyx_kp_u_numpy__core_umath_failed_to_impo};
        __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ImportError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1017, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1017, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1013
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1019
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
*/

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1020
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1021
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")
*/
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 1021, __pyx_L3_error)

      /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1020
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1022
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.umath failed to import")
 * 
*/
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1022, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1023
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * 
*/
      __pyx_t_9 = NULL;
      __pyx_t_10 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_mstate_global->__pyx_kp_u_numpy__core_umath_failed_to_impo};
        __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ImportError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1023, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1023, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1020
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1019
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1026
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
*/

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1038
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1026
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1041
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
*/

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1053
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1041
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1056
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
*/

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1063
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1056
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1066
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
*/

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1070
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1066
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1073
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
*/

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1077
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../../../private/var/folders/mg/h161hv09185018vpst9gtclm0000gn/T/build-env-qlwg9ih7/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1073
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":44
 *     OPTION_LOG_ALL = 0x7fffffffffffffff
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def logger(cls):
 *         global pmm_logger
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_1logger(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_1logger = {"logger", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_1logger, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_1logger(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("logger (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("logger", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("logger", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_logger(((PyTypeObject*)__pyx_v_cls));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_logger(CYTHON_UNUSED PyTypeObject *__pyx_v_cls) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("logger", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":47
 *     def logger(cls):
 *         global pmm_logger
 *         if pmm_logger is None:             # <<<<<<<<<<<<<<
 *             pmm_logger = logging.getLogger(__name__)
 *         return pmm_logger
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_pmm_logger); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":48
 *         global pmm_logger
 *         if pmm_logger is None:
 *             pmm_logger = logging.getLogger(__name__)             # <<<<<<<<<<<<<<
 *         return pmm_logger
 * 
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_logging); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_getLogger); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pmm_logger, __pyx_t_1) < (0)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":47
 *     def logger(cls):
 *         global pmm_logger
 *         if pmm_logger is None:             # <<<<<<<<<<<<<<
 *             pmm_logger = logging.getLogger(__name__)
 *         return pmm_logger
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":49
 *         if pmm_logger is None:
 *             pmm_logger = logging.getLogger(__name__)
 *         return pmm_logger             # <<<<<<<<<<<<<<
 * 
 *     def init_params(self,
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_pmm_logger); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":44
 *     OPTION_LOG_ALL = 0x7fffffffffffffff
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def logger(cls):
 *         global pmm_logger
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.logger", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":51
 *         return pmm_logger
 * 
 *     def init_params(self,             # <<<<<<<<<<<<<<
 *                     market_info: MarketTradingPairTuple,
 *                     bid_spread: Decimal,
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_3init_params(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_3init_params = {"init_params", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_3init_params, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_3init_params(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_market_info = 0;
  PyObject *__pyx_v_bid_spread = 0;
  PyObject *__pyx_v_ask_spread = 0;
  PyObject *__pyx_v_order_amount = 0;
  PyObject *__pyx_v_order_levels = 0;
  PyObject *__pyx_v_order_level_spread = 0;
  PyObject *__pyx_v_order_level_amount = 0;
  PyObject *__pyx_v_order_refresh_time = 0;
  PyObject *__pyx_v_max_order_age = 0;
  PyObject *__pyx_v_order_refresh_tolerance_pct = 0;
  PyObject *__pyx_v_filled_order_delay = 0;
  PyObject *__pyx_v_inventory_skew_enabled = 0;
  PyObject *__pyx_v_inventory_target_base_pct = 0;
  PyObject *__pyx_v_inventory_range_multiplier = 0;
  PyObject *__pyx_v_hanging_orders_enabled = 0;
  PyObject *__pyx_v_hanging_orders_cancel_pct = 0;
  PyObject *__pyx_v_order_optimization_enabled = 0;
  PyObject *__pyx_v_ask_order_optimization_depth = 0;
  PyObject *__pyx_v_bid_order_optimization_depth = 0;
  PyObject *__pyx_v_add_transaction_costs_to_orders = 0;
  PyObject *__pyx_v_asset_price_delegate = 0;
  PyObject *__pyx_v_inventory_cost_price_delegate = 0;
  PyObject *__pyx_v_price_type = 0;
  PyObject *__pyx_v_take_if_crossed = 0;
  PyObject *__pyx_v_price_ceiling = 0;
  PyObject *__pyx_v_price_floor = 0;
  PyObject *__pyx_v_ping_pong_enabled = 0;
  PyObject *__pyx_v_logging_options = 0;
  PyObject *__pyx_v_status_report_interval = 0;
  PyObject *__pyx_v_minimum_spread = 0;
  PyObject *__pyx_v_hb_app_notification = 0;
  PyObject *__pyx_v_order_override = 0;
  PyObject *__pyx_v_split_order_levels_enabled = 0;
  PyObject *__pyx_v_bid_order_level_spreads = 0;
  PyObject *__pyx_v_ask_order_level_spreads = 0;
  PyObject *__pyx_v_should_wait_order_cancel_confirmation = 0;
  PyObject *__pyx_v_moving_price_band = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[37] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("init_params (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_market_info,&__pyx_mstate_global->__pyx_n_u_bid_spread,&__pyx_mstate_global->__pyx_n_u_ask_spread,&__pyx_mstate_global->__pyx_n_u_order_amount,&__pyx_mstate_global->__pyx_n_u_order_levels,&__pyx_mstate_global->__pyx_n_u_order_level_spread,&__pyx_mstate_global->__pyx_n_u_order_level_amount,&__pyx_mstate_global->__pyx_n_u_order_refresh_time,&__pyx_mstate_global->__pyx_n_u_max_order_age,&__pyx_mstate_global->__pyx_n_u_order_refresh_tolerance_pct,&__pyx_mstate_global->__pyx_n_u_filled_order_delay,&__pyx_mstate_global->__pyx_n_u_inventory_skew_enabled,&__pyx_mstate_global->__pyx_n_u_inventory_target_base_pct,&__pyx_mstate_global->__pyx_n_u_inventory_range_multiplier,&__pyx_mstate_global->__pyx_n_u_hanging_orders_enabled,&__pyx_mstate_global->__pyx_n_u_hanging_orders_cancel_pct,&__pyx_mstate_global->__pyx_n_u_order_optimization_enabled,&__pyx_mstate_global->__pyx_n_u_ask_order_optimization_depth,&__pyx_mstate_global->__pyx_n_u_bid_order_optimization_depth,&__pyx_mstate_global->__pyx_n_u_add_transaction_costs_to_orders,&__pyx_mstate_global->__pyx_n_u_asset_price_delegate,&__pyx_mstate_global->__pyx_n_u_inventory_cost_price_delegate,&__pyx_mstate_global->__pyx_n_u_price_type,&__pyx_mstate_global->__pyx_n_u_take_if_crossed,&__pyx_mstate_global->__pyx_n_u_price_ceiling,&__pyx_mstate_global->__pyx_n_u_price_floor,&__pyx_mstate_global->__pyx_n_u_ping_pong_enabled,&__pyx_mstate_global->__pyx_n_u_logging_options,&__pyx_mstate_global->__pyx_n_u_status_report_interval,&__pyx_mstate_global->__pyx_n_u_minimum_spread,&__pyx_mstate_global->__pyx_n_u_hb_app_notification,&__pyx_mstate_global->__pyx_n_u_order_override,&__pyx_mstate_global->__pyx_n_u_split_order_levels_enabled,&__pyx_mstate_global->__pyx_n_u_bid_order_level_spreads,&__pyx_mstate_global->__pyx_n_u_ask_order_level_spreads,&__pyx_mstate_global->__pyx_n_u_should_wait_order_cancel_confirm,&__pyx_mstate_global->__pyx_n_u_moving_price_band,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 51, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case 37:
        values[36] = __Pyx_ArgRef_FASTCALL(__pyx_args, 36);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[36])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 36:
        values[35] = __Pyx_ArgRef_FASTCALL(__pyx_args, 35);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[35])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 35:
        values[34] = __Pyx_ArgRef_FASTCALL(__pyx_args, 34);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[34])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 34:
        values[33] = __Pyx_ArgRef_FASTCALL(__pyx_args, 33);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[33])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 33:
        values[32] = __Pyx_ArgRef_FASTCALL(__pyx_args, 32);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[32])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 32:
        values[31] = __Pyx_ArgRef_FASTCALL(__pyx_args, 31);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[31])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 31:
        values[30] = __Pyx_ArgRef_FASTCALL(__pyx_args, 30);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[30])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 30:
        values[29] = __Pyx_ArgRef_FASTCALL(__pyx_args, 29);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[29])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 29:
        values[28] = __Pyx_ArgRef_FASTCALL(__pyx_args, 28);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[28])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 28:
        values[27] = __Pyx_ArgRef_FASTCALL(__pyx_args, 27);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[27])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 27:
        values[26] = __Pyx_ArgRef_FASTCALL(__pyx_args, 26);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[26])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 26:
        values[25] = __Pyx_ArgRef_FASTCALL(__pyx_args, 25);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[25])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 25:
        values[24] = __Pyx_ArgRef_FASTCALL(__pyx_args, 24);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[24])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 24:
        values[23] = __Pyx_ArgRef_FASTCALL(__pyx_args, 23);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[23])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 23:
        values[22] = __Pyx_ArgRef_FASTCALL(__pyx_args, 22);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[22])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 22:
        values[21] = __Pyx_ArgRef_FASTCALL(__pyx_args, 21);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[21])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 21:
        values[20] = __Pyx_ArgRef_FASTCALL(__pyx_args, 20);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[20])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 20:
        values[19] = __Pyx_ArgRef_FASTCALL(__pyx_args, 19);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[19])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 19:
        values[18] = __Pyx_ArgRef_FASTCALL(__pyx_args, 18);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[18])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 18:
        values[17] = __Pyx_ArgRef_FASTCALL(__pyx_args, 17);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[17])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 17:
        values[16] = __Pyx_ArgRef_FASTCALL(__pyx_args, 16);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[16])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 16:
        values[15] = __Pyx_ArgRef_FASTCALL(__pyx_args, 15);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[15])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 15:
        values[14] = __Pyx_ArgRef_FASTCALL(__pyx_args, 14);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[14])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 14:
        values[13] = __Pyx_ArgRef_FASTCALL(__pyx_args, 13);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[13])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 13:
        values[12] = __Pyx_ArgRef_FASTCALL(__pyx_args, 12);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[12])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 12:
        values[11] = __Pyx_ArgRef_FASTCALL(__pyx_args, 11);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[11])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 11:
        values[10] = __Pyx_ArgRef_FASTCALL(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_FASTCALL(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_FASTCALL(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "init_params", 0) < (0)) __PYX_ERR(0, 51, __pyx_L3_error)
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[5]) values[5] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k_);
      if (!values[6]) values[6] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__2);
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_float_30_0));
      if (!values[8]) values[8] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_float_1800_0));
      if (!values[9]) values[9] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__3);
      if (!values[10]) values[10] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_float_60_0));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":63
 *                     order_refresh_tolerance_pct: Decimal = s_decimal_neg_one,
 *                     filled_order_delay: float = 60.0,
 *                     inventory_skew_enabled: bool = False,             # <<<<<<<<<<<<<<
 *                     inventory_target_base_pct: Decimal = s_decimal_zero,
 *                     inventory_range_multiplier: Decimal = s_decimal_zero,
*/
      if (!values[11]) values[11] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[12]) values[12] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__4);
      if (!values[13]) values[13] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__5);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":66
 *                     inventory_target_base_pct: Decimal = s_decimal_zero,
 *                     inventory_range_multiplier: Decimal = s_decimal_zero,
 *                     hanging_orders_enabled: bool = False,             # <<<<<<<<<<<<<<
 *                     hanging_orders_cancel_pct: Decimal = Decimal("0.1"),
 *                     order_optimization_enabled: bool = False,
*/
      if (!values[14]) values[14] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[15]) values[15] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__6);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":68
 *                     hanging_orders_enabled: bool = False,
 *                     hanging_orders_cancel_pct: Decimal = Decimal("0.1"),
 *                     order_optimization_enabled: bool = False,             # <<<<<<<<<<<<<<
 *                     ask_order_optimization_depth: Decimal = s_decimal_zero,
 *                     bid_order_optimization_depth: Decimal = s_decimal_zero,
*/
      if (!values[16]) values[16] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[17]) values[17] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__7);
      if (!values[18]) values[18] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__8);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":71
 *                     ask_order_optimization_depth: Decimal = s_decimal_zero,
 *                     bid_order_optimization_depth: Decimal = s_decimal_zero,
 *                     add_transaction_costs_to_orders: bool = False,             # <<<<<<<<<<<<<<
 *                     asset_price_delegate: AssetPriceDelegate = None,
 *                     inventory_cost_price_delegate: InventoryCostPriceDelegate = None,
*/
      if (!values[19]) values[19] = __Pyx_NewRef(((PyObject *)Py_False));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":72
 *                     bid_order_optimization_depth: Decimal = s_decimal_zero,
 *                     add_transaction_costs_to_orders: bool = False,
 *                     asset_price_delegate: AssetPriceDelegate = None,             # <<<<<<<<<<<<<<
 *                     inventory_cost_price_delegate: InventoryCostPriceDelegate = None,
 *                     price_type: str = "mid_price",
*/
      if (!values[20]) values[20] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":73
 *                     add_transaction_costs_to_orders: bool = False,
 *                     asset_price_delegate: AssetPriceDelegate = None,
 *                     inventory_cost_price_delegate: InventoryCostPriceDelegate = None,             # <<<<<<<<<<<<<<
 *                     price_type: str = "mid_price",
 *                     take_if_crossed: bool = False,
*/
      if (!values[21]) values[21] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[22]) values[22] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_n_u_mid_price));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":75
 *                     inventory_cost_price_delegate: InventoryCostPriceDelegate = None,
 *                     price_type: str = "mid_price",
 *                     take_if_crossed: bool = False,             # <<<<<<<<<<<<<<
 *                     price_ceiling: Decimal = s_decimal_neg_one,
 *                     price_floor: Decimal = s_decimal_neg_one,
*/
      if (!values[23]) values[23] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[24]) values[24] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__9);
      if (!values[25]) values[25] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__10);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":78
 *                     price_ceiling: Decimal = s_decimal_neg_one,
 *                     price_floor: Decimal = s_decimal_neg_one,
 *                     ping_pong_enabled: bool = False,             # <<<<<<<<<<<<<<
 *                     logging_options: int = OPTION_LOG_ALL,
 *                     status_report_interval: float = 900,
*/
      if (!values[26]) values[26] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[27]) values[27] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__11);
      if (!values[28]) values[28] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_900));
      if (!values[29]) values[29] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__12);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":82
 *                     status_report_interval: float = 900,
 *                     minimum_spread: Decimal = Decimal(0),
 *                     hb_app_notification: bool = False,             # <<<<<<<<<<<<<<
 *                     order_override: Dict[str, List[str]] = None,
 *                     split_order_levels_enabled: bool = False,
*/
      if (!values[30]) values[30] = __Pyx_NewRef(((PyObject *)Py_False));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":83
 *                     minimum_spread: Decimal = Decimal(0),
 *                     hb_app_notification: bool = False,
 *                     order_override: Dict[str, List[str]] = None,             # <<<<<<<<<<<<<<
 *                     split_order_levels_enabled: bool = False,
 *                     bid_order_level_spreads: List[Decimal] = None,
*/
      if (!values[31]) values[31] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":84
 *                     hb_app_notification: bool = False,
 *                     order_override: Dict[str, List[str]] = None,
 *                     split_order_levels_enabled: bool = False,             # <<<<<<<<<<<<<<
 *                     bid_order_level_spreads: List[Decimal] = None,
 *                     ask_order_level_spreads: List[Decimal] = None,
*/
      if (!values[32]) values[32] = __Pyx_NewRef(((PyObject *)Py_False));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":85
 *                     order_override: Dict[str, List[str]] = None,
 *                     split_order_levels_enabled: bool = False,
 *                     bid_order_level_spreads: List[Decimal] = None,             # <<<<<<<<<<<<<<
 *                     ask_order_level_spreads: List[Decimal] = None,
 *                     should_wait_order_cancel_confirmation: bool = True,
*/
      if (!values[33]) values[33] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":86
 *                     split_order_levels_enabled: bool = False,
 *                     bid_order_level_spreads: List[Decimal] = None,
 *                     ask_order_level_spreads: List[Decimal] = None,             # <<<<<<<<<<<<<<
 *                     should_wait_order_cancel_confirmation: bool = True,
 *                     moving_price_band: Optional[MovingPriceBand] = None
*/
      if (!values[34]) values[34] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":87
 *                     bid_order_level_spreads: List[Decimal] = None,
 *                     ask_order_level_spreads: List[Decimal] = None,
 *                     should_wait_order_cancel_confirmation: bool = True,             # <<<<<<<<<<<<<<
 *                     moving_price_band: Optional[MovingPriceBand] = None
 *                     ):
*/
      if (!values[35]) values[35] = __Pyx_NewRef(((PyObject *)Py_True));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":88
 *                     ask_order_level_spreads: List[Decimal] = None,
 *                     should_wait_order_cancel_confirmation: bool = True,
 *                     moving_price_band: Optional[MovingPriceBand] = None             # <<<<<<<<<<<<<<
 *                     ):
 *         if order_override is None:
*/
      if (!values[36]) values[36] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("init_params", 0, 4, 37, i); __PYX_ERR(0, 51, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case 37:
        values[36] = __Pyx_ArgRef_FASTCALL(__pyx_args, 36);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[36])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 36:
        values[35] = __Pyx_ArgRef_FASTCALL(__pyx_args, 35);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[35])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 35:
        values[34] = __Pyx_ArgRef_FASTCALL(__pyx_args, 34);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[34])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 34:
        values[33] = __Pyx_ArgRef_FASTCALL(__pyx_args, 33);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[33])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 33:
        values[32] = __Pyx_ArgRef_FASTCALL(__pyx_args, 32);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[32])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 32:
        values[31] = __Pyx_ArgRef_FASTCALL(__pyx_args, 31);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[31])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 31:
        values[30] = __Pyx_ArgRef_FASTCALL(__pyx_args, 30);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[30])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 30:
        values[29] = __Pyx_ArgRef_FASTCALL(__pyx_args, 29);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[29])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 29:
        values[28] = __Pyx_ArgRef_FASTCALL(__pyx_args, 28);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[28])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 28:
        values[27] = __Pyx_ArgRef_FASTCALL(__pyx_args, 27);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[27])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 27:
        values[26] = __Pyx_ArgRef_FASTCALL(__pyx_args, 26);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[26])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 26:
        values[25] = __Pyx_ArgRef_FASTCALL(__pyx_args, 25);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[25])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 25:
        values[24] = __Pyx_ArgRef_FASTCALL(__pyx_args, 24);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[24])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 24:
        values[23] = __Pyx_ArgRef_FASTCALL(__pyx_args, 23);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[23])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 23:
        values[22] = __Pyx_ArgRef_FASTCALL(__pyx_args, 22);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[22])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 22:
        values[21] = __Pyx_ArgRef_FASTCALL(__pyx_args, 21);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[21])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 21:
        values[20] = __Pyx_ArgRef_FASTCALL(__pyx_args, 20);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[20])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 20:
        values[19] = __Pyx_ArgRef_FASTCALL(__pyx_args, 19);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[19])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 19:
        values[18] = __Pyx_ArgRef_FASTCALL(__pyx_args, 18);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[18])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 18:
        values[17] = __Pyx_ArgRef_FASTCALL(__pyx_args, 17);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[17])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 17:
        values[16] = __Pyx_ArgRef_FASTCALL(__pyx_args, 16);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[16])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 16:
        values[15] = __Pyx_ArgRef_FASTCALL(__pyx_args, 15);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[15])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 15:
        values[14] = __Pyx_ArgRef_FASTCALL(__pyx_args, 14);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[14])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 14:
        values[13] = __Pyx_ArgRef_FASTCALL(__pyx_args, 13);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[13])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 13:
        values[12] = __Pyx_ArgRef_FASTCALL(__pyx_args, 12);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[12])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 12:
        values[11] = __Pyx_ArgRef_FASTCALL(__pyx_args, 11);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[11])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 11:
        values[10] = __Pyx_ArgRef_FASTCALL(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_FASTCALL(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_FASTCALL(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 51, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 51, __pyx_L3_error)
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 51, __pyx_L3_error)
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 51, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 51, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[5]) values[5] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k_);
      if (!values[6]) values[6] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__2);
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_float_30_0));
      if (!values[8]) values[8] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_float_1800_0));
      if (!values[9]) values[9] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__3);
      if (!values[10]) values[10] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_float_60_0));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":63
 *                     order_refresh_tolerance_pct: Decimal = s_decimal_neg_one,
 *                     filled_order_delay: float = 60.0,
 *                     inventory_skew_enabled: bool = False,             # <<<<<<<<<<<<<<
 *                     inventory_target_base_pct: Decimal = s_decimal_zero,
 *                     inventory_range_multiplier: Decimal = s_decimal_zero,
*/
      if (!values[11]) values[11] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[12]) values[12] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__4);
      if (!values[13]) values[13] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__5);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":66
 *                     inventory_target_base_pct: Decimal = s_decimal_zero,
 *                     inventory_range_multiplier: Decimal = s_decimal_zero,
 *                     hanging_orders_enabled: bool = False,             # <<<<<<<<<<<<<<
 *                     hanging_orders_cancel_pct: Decimal = Decimal("0.1"),
 *                     order_optimization_enabled: bool = False,
*/
      if (!values[14]) values[14] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[15]) values[15] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__6);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":68
 *                     hanging_orders_enabled: bool = False,
 *                     hanging_orders_cancel_pct: Decimal = Decimal("0.1"),
 *                     order_optimization_enabled: bool = False,             # <<<<<<<<<<<<<<
 *                     ask_order_optimization_depth: Decimal = s_decimal_zero,
 *                     bid_order_optimization_depth: Decimal = s_decimal_zero,
*/
      if (!values[16]) values[16] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[17]) values[17] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__7);
      if (!values[18]) values[18] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__8);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":71
 *                     ask_order_optimization_depth: Decimal = s_decimal_zero,
 *                     bid_order_optimization_depth: Decimal = s_decimal_zero,
 *                     add_transaction_costs_to_orders: bool = False,             # <<<<<<<<<<<<<<
 *                     asset_price_delegate: AssetPriceDelegate = None,
 *                     inventory_cost_price_delegate: InventoryCostPriceDelegate = None,
*/
      if (!values[19]) values[19] = __Pyx_NewRef(((PyObject *)Py_False));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":72
 *                     bid_order_optimization_depth: Decimal = s_decimal_zero,
 *                     add_transaction_costs_to_orders: bool = False,
 *                     asset_price_delegate: AssetPriceDelegate = None,             # <<<<<<<<<<<<<<
 *                     inventory_cost_price_delegate: InventoryCostPriceDelegate = None,
 *                     price_type: str = "mid_price",
*/
      if (!values[20]) values[20] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":73
 *                     add_transaction_costs_to_orders: bool = False,
 *                     asset_price_delegate: AssetPriceDelegate = None,
 *                     inventory_cost_price_delegate: InventoryCostPriceDelegate = None,             # <<<<<<<<<<<<<<
 *                     price_type: str = "mid_price",
 *                     take_if_crossed: bool = False,
*/
      if (!values[21]) values[21] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[22]) values[22] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_n_u_mid_price));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":75
 *                     inventory_cost_price_delegate: InventoryCostPriceDelegate = None,
 *                     price_type: str = "mid_price",
 *                     take_if_crossed: bool = False,             # <<<<<<<<<<<<<<
 *                     price_ceiling: Decimal = s_decimal_neg_one,
 *                     price_floor: Decimal = s_decimal_neg_one,
*/
      if (!values[23]) values[23] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[24]) values[24] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__9);
      if (!values[25]) values[25] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__10);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":78
 *                     price_ceiling: Decimal = s_decimal_neg_one,
 *                     price_floor: Decimal = s_decimal_neg_one,
 *                     ping_pong_enabled: bool = False,             # <<<<<<<<<<<<<<
 *                     logging_options: int = OPTION_LOG_ALL,
 *                     status_report_interval: float = 900,
*/
      if (!values[26]) values[26] = __Pyx_NewRef(((PyObject *)Py_False));
      if (!values[27]) values[27] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__11);
      if (!values[28]) values[28] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_900));
      if (!values[29]) values[29] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__12);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":82
 *                     status_report_interval: float = 900,
 *                     minimum_spread: Decimal = Decimal(0),
 *                     hb_app_notification: bool = False,             # <<<<<<<<<<<<<<
 *                     order_override: Dict[str, List[str]] = None,
 *                     split_order_levels_enabled: bool = False,
*/
      if (!values[30]) values[30] = __Pyx_NewRef(((PyObject *)Py_False));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":83
 *                     minimum_spread: Decimal = Decimal(0),
 *                     hb_app_notification: bool = False,
 *                     order_override: Dict[str, List[str]] = None,             # <<<<<<<<<<<<<<
 *                     split_order_levels_enabled: bool = False,
 *                     bid_order_level_spreads: List[Decimal] = None,
*/
      if (!values[31]) values[31] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":84
 *                     hb_app_notification: bool = False,
 *                     order_override: Dict[str, List[str]] = None,
 *                     split_order_levels_enabled: bool = False,             # <<<<<<<<<<<<<<
 *                     bid_order_level_spreads: List[Decimal] = None,
 *                     ask_order_level_spreads: List[Decimal] = None,
*/
      if (!values[32]) values[32] = __Pyx_NewRef(((PyObject *)Py_False));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":85
 *                     order_override: Dict[str, List[str]] = None,
 *                     split_order_levels_enabled: bool = False,
 *                     bid_order_level_spreads: List[Decimal] = None,             # <<<<<<<<<<<<<<
 *                     ask_order_level_spreads: List[Decimal] = None,
 *                     should_wait_order_cancel_confirmation: bool = True,
*/
      if (!values[33]) values[33] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":86
 *                     split_order_levels_enabled: bool = False,
 *                     bid_order_level_spreads: List[Decimal] = None,
 *                     ask_order_level_spreads: List[Decimal] = None,             # <<<<<<<<<<<<<<
 *                     should_wait_order_cancel_confirmation: bool = True,
 *                     moving_price_band: Optional[MovingPriceBand] = None
*/
      if (!values[34]) values[34] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":87
 *                     bid_order_level_spreads: List[Decimal] = None,
 *                     ask_order_level_spreads: List[Decimal] = None,
 *                     should_wait_order_cancel_confirmation: bool = True,             # <<<<<<<<<<<<<<
 *                     moving_price_band: Optional[MovingPriceBand] = None
 *                     ):
*/
      if (!values[35]) values[35] = __Pyx_NewRef(((PyObject *)Py_True));

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":88
 *                     ask_order_level_spreads: List[Decimal] = None,
 *                     should_wait_order_cancel_confirmation: bool = True,
 *                     moving_price_band: Optional[MovingPriceBand] = None             # <<<<<<<<<<<<<<
 *                     ):
 *         if order_override is None:
*/
      if (!values[36]) values[36] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_market_info = values[0];
    __pyx_v_bid_spread = values[1];
    __pyx_v_ask_spread = values[2];
    __pyx_v_order_amount = values[3];
    __pyx_v_order_levels = values[4];
    __pyx_v_order_level_spread = values[5];
    __pyx_v_order_level_amount = values[6];
    __pyx_v_order_refresh_time = values[7];
    __pyx_v_max_order_age = values[8];
    __pyx_v_order_refresh_tolerance_pct = values[9];
    __pyx_v_filled_order_delay = values[10];
    __pyx_v_inventory_skew_enabled = values[11];
    __pyx_v_inventory_target_base_pct = values[12];
    __pyx_v_inventory_range_multiplier = values[13];
    __pyx_v_hanging_orders_enabled = values[14];
    __pyx_v_hanging_orders_cancel_pct = values[15];
    __pyx_v_order_optimization_enabled = values[16];
    __pyx_v_ask_order_optimization_depth = values[17];
    __pyx_v_bid_order_optimization_depth = values[18];
    __pyx_v_add_transaction_costs_to_orders = values[19];
    __pyx_v_asset_price_delegate = values[20];
    __pyx_v_inventory_cost_price_delegate = values[21];
    __pyx_v_price_type = values[22];
    __pyx_v_take_if_crossed = values[23];
    __pyx_v_price_ceiling = values[24];
    __pyx_v_price_floor = values[25];
    __pyx_v_ping_pong_enabled = values[26];
    __pyx_v_logging_options = values[27];
    __pyx_v_status_report_interval = values[28];
    __pyx_v_minimum_spread = values[29];
    __pyx_v_hb_app_notification = values[30];
    __pyx_v_order_override = values[31];
    __pyx_v_split_order_levels_enabled = values[32];
    __pyx_v_bid_order_level_spreads = values[33];
    __pyx_v_ask_order_level_spreads = values[34];
    __pyx_v_should_wait_order_cancel_confirmation = values[35];
    __pyx_v_moving_price_band = values[36];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("init_params", 0, 4, 37, __pyx_nargs); __PYX_ERR(0, 51, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.init_params", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_2init_params(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), __pyx_v_market_info, __pyx_v_bid_spread, __pyx_v_ask_spread, __pyx_v_order_amount, __pyx_v_order_levels, __pyx_v_order_level_spread, __pyx_v_order_level_amount, __pyx_v_order_refresh_time, __pyx_v_max_order_age, __pyx_v_order_refresh_tolerance_pct, __pyx_v_filled_order_delay, __pyx_v_inventory_skew_enabled, __pyx_v_inventory_target_base_pct, __pyx_v_inventory_range_multiplier, __pyx_v_hanging_orders_enabled, __pyx_v_hanging_orders_cancel_pct, __pyx_v_order_optimization_enabled, __pyx_v_ask_order_optimization_depth, __pyx_v_bid_order_optimization_depth, __pyx_v_add_transaction_costs_to_orders, __pyx_v_asset_price_delegate, __pyx_v_inventory_cost_price_delegate, __pyx_v_price_type, __pyx_v_take_if_crossed, __pyx_v_price_ceiling, __pyx_v_price_floor, __pyx_v_ping_pong_enabled, __pyx_v_logging_options, __pyx_v_status_report_interval, __pyx_v_minimum_spread, __pyx_v_hb_app_notification, __pyx_v_order_override, __pyx_v_split_order_levels_enabled, __pyx_v_bid_order_level_spreads, __pyx_v_ask_order_level_spreads, __pyx_v_should_wait_order_cancel_confirmation, __pyx_v_moving_price_band);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":51
 *         return pmm_logger
 * 
 *     def init_params(self,             # <<<<<<<<<<<<<<
 *                     market_info: MarketTradingPairTuple,
 *                     bid_spread: Decimal,
*/

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_2init_params(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_market_info, PyObject *__pyx_v_bid_spread, PyObject *__pyx_v_ask_spread, PyObject *__pyx_v_order_amount, PyObject *__pyx_v_order_levels, PyObject *__pyx_v_order_level_spread, PyObject *__pyx_v_order_level_amount, PyObject *__pyx_v_order_refresh_time, PyObject *__pyx_v_max_order_age, PyObject *__pyx_v_order_refresh_tolerance_pct, PyObject *__pyx_v_filled_order_delay, PyObject *__pyx_v_inventory_skew_enabled, PyObject *__pyx_v_inventory_target_base_pct, PyObject *__pyx_v_inventory_range_multiplier, PyObject *__pyx_v_hanging_orders_enabled, PyObject *__pyx_v_hanging_orders_cancel_pct, PyObject *__pyx_v_order_optimization_enabled, PyObject *__pyx_v_ask_order_optimization_depth, PyObject *__pyx_v_bid_order_optimization_depth, PyObject *__pyx_v_add_transaction_costs_to_orders, PyObject *__pyx_v_asset_price_delegate, PyObject *__pyx_v_inventory_cost_price_delegate, PyObject *__pyx_v_price_type, PyObject *__pyx_v_take_if_crossed, PyObject *__pyx_v_price_ceiling, PyObject *__pyx_v_price_floor, PyObject *__pyx_v_ping_pong_enabled, PyObject *__pyx_v_logging_options, PyObject *__pyx_v_status_report_interval, PyObject *__pyx_v_minimum_spread, PyObject *__pyx_v_hb_app_notification, PyObject *__pyx_v_order_override, PyObject *__pyx_v_split_order_levels_enabled, PyObject *__pyx_v_bid_order_level_spreads, PyObject *__pyx_v_ask_order_level_spreads, PyObject *__pyx_v_should_wait_order_cancel_confirmation, PyObject *__pyx_v_moving_price_band) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  int64_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init_params", 0);
  __Pyx_INCREF(__pyx_v_order_override);
  __Pyx_INCREF(__pyx_v_moving_price_band);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":90
 *                     moving_price_band: Optional[MovingPriceBand] = None
 *                     ):
 *         if order_override is None:             # <<<<<<<<<<<<<<
 *             order_override = {}
 *         if moving_price_band is None:
*/
  __pyx_t_1 = (__pyx_v_order_override == Py_None);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":91
 *                     ):
 *         if order_override is None:
 *             order_override = {}             # <<<<<<<<<<<<<<
 *         if moving_price_band is None:
 *             moving_price_band = MovingPriceBand()
*/
    __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_order_override, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":90
 *                     moving_price_band: Optional[MovingPriceBand] = None
 *                     ):
 *         if order_override is None:             # <<<<<<<<<<<<<<
 *             order_override = {}
 *         if moving_price_band is None:
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":92
 *         if order_override is None:
 *             order_override = {}
 *         if moving_price_band is None:             # <<<<<<<<<<<<<<
 *             moving_price_band = MovingPriceBand()
 *         if price_ceiling != s_decimal_neg_one and price_ceiling < price_floor:
*/
  __pyx_t_1 = (__pyx_v_moving_price_band == Py_None);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":93
 *             order_override = {}
 *         if moving_price_band is None:
 *             moving_price_band = MovingPriceBand()             # <<<<<<<<<<<<<<
 *         if price_ceiling != s_decimal_neg_one and price_ceiling < price_floor:
 *             raise ValueError("Parameter price_ceiling cannot be lower than price_floor.")
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_MovingPriceBand); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF_SET(__pyx_v_moving_price_band, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":92
 *         if order_override is None:
 *             order_override = {}
 *         if moving_price_band is None:             # <<<<<<<<<<<<<<
 *             moving_price_band = MovingPriceBand()
 *         if price_ceiling != s_decimal_neg_one and price_ceiling < price_floor:
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":94
 *         if moving_price_band is None:
 *             moving_price_band = MovingPriceBand()
 *         if price_ceiling != s_decimal_neg_one and price_ceiling < price_floor:             # <<<<<<<<<<<<<<
 *             raise ValueError("Parameter price_ceiling cannot be lower than price_floor.")
 *         self._sb_order_tracker = PureMarketMakingOrderTracker()
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_s_decimal_neg_one); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_price_ceiling, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_1 = __pyx_t_6;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_price_ceiling, __pyx_v_price_floor, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 94, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_6;
  __pyx_L6_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":95
 *             moving_price_band = MovingPriceBand()
 *         if price_ceiling != s_decimal_neg_one and price_ceiling < price_floor:
 *             raise ValueError("Parameter price_ceiling cannot be lower than price_floor.")             # <<<<<<<<<<<<<<
 *         self._sb_order_tracker = PureMarketMakingOrderTracker()
 *         self._market_info = market_info
*/
    __pyx_t_2 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Parameter_price_ceiling_cannot_b};
      __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 95, __pyx_L1_error)

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":94
 *         if moving_price_band is None:
 *             moving_price_band = MovingPriceBand()
 *         if price_ceiling != s_decimal_neg_one and price_ceiling < price_floor:             # <<<<<<<<<<<<<<
 *             raise ValueError("Parameter price_ceiling cannot be lower than price_floor.")
 *         self._sb_order_tracker = PureMarketMakingOrderTracker()
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":96
 *         if price_ceiling != s_decimal_neg_one and price_ceiling < price_floor:
 *             raise ValueError("Parameter price_ceiling cannot be lower than price_floor.")
 *         self._sb_order_tracker = PureMarketMakingOrderTracker()             # <<<<<<<<<<<<<<
 *         self._market_info = market_info
 *         self._bid_spread = bid_spread
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PureMarketMakingOrderTracker); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker))))) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->__pyx_base._sb_order_tracker);
  __Pyx_DECREF((PyObject *)__pyx_v_self->__pyx_base._sb_order_tracker);
  __pyx_v_self->__pyx_base._sb_order_tracker = ((struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":97
 *             raise ValueError("Parameter price_ceiling cannot be lower than price_floor.")
 *         self._sb_order_tracker = PureMarketMakingOrderTracker()
 *         self._market_info = market_info             # <<<<<<<<<<<<<<
 *         self._bid_spread = bid_spread
 *         self._ask_spread = ask_spread
*/
  __Pyx_INCREF(__pyx_v_market_info);
  __Pyx_GIVEREF(__pyx_v_market_info);
  __Pyx_GOTREF(__pyx_v_self->_market_info);
  __Pyx_DECREF(__pyx_v_self->_market_info);
  __pyx_v_self->_market_info = __pyx_v_market_info;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":98
 *         self._sb_order_tracker = PureMarketMakingOrderTracker()
 *         self._market_info = market_info
 *         self._bid_spread = bid_spread             # <<<<<<<<<<<<<<
 *         self._ask_spread = ask_spread
 *         self._minimum_spread = minimum_spread
*/
  __Pyx_INCREF(__pyx_v_bid_spread);
  __Pyx_GIVEREF(__pyx_v_bid_spread);
  __Pyx_GOTREF(__pyx_v_self->_bid_spread);
  __Pyx_DECREF(__pyx_v_self->_bid_spread);
  __pyx_v_self->_bid_spread = __pyx_v_bid_spread;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":99
 *         self._market_info = market_info
 *         self._bid_spread = bid_spread
 *         self._ask_spread = ask_spread             # <<<<<<<<<<<<<<
 *         self._minimum_spread = minimum_spread
 *         self._order_amount = order_amount
*/
  __Pyx_INCREF(__pyx_v_ask_spread);
  __Pyx_GIVEREF(__pyx_v_ask_spread);
  __Pyx_GOTREF(__pyx_v_self->_ask_spread);
  __Pyx_DECREF(__pyx_v_self->_ask_spread);
  __pyx_v_self->_ask_spread = __pyx_v_ask_spread;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":100
 *         self._bid_spread = bid_spread
 *         self._ask_spread = ask_spread
 *         self._minimum_spread = minimum_spread             # <<<<<<<<<<<<<<
 *         self._order_amount = order_amount
 *         self._order_levels = order_levels
*/
  __Pyx_INCREF(__pyx_v_minimum_spread);
  __Pyx_GIVEREF(__pyx_v_minimum_spread);
  __Pyx_GOTREF(__pyx_v_self->_minimum_spread);
  __Pyx_DECREF(__pyx_v_self->_minimum_spread);
  __pyx_v_self->_minimum_spread = __pyx_v_minimum_spread;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":101
 *         self._ask_spread = ask_spread
 *         self._minimum_spread = minimum_spread
 *         self._order_amount = order_amount             # <<<<<<<<<<<<<<
 *         self._order_levels = order_levels
 *         self._buy_levels = order_levels
*/
  __Pyx_INCREF(__pyx_v_order_amount);
  __Pyx_GIVEREF(__pyx_v_order_amount);
  __Pyx_GOTREF(__pyx_v_self->_order_amount);
  __Pyx_DECREF(__pyx_v_self->_order_amount);
  __pyx_v_self->_order_amount = __pyx_v_order_amount;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":102
 *         self._minimum_spread = minimum_spread
 *         self._order_amount = order_amount
 *         self._order_levels = order_levels             # <<<<<<<<<<<<<<
 *         self._buy_levels = order_levels
 *         self._sell_levels = order_levels
*/
  __pyx_t_7 = __Pyx_PyLong_As_int(__pyx_v_order_levels); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 102, __pyx_L1_error)
  __pyx_v_self->_order_levels = __pyx_t_7;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":103
 *         self._order_amount = order_amount
 *         self._order_levels = order_levels
 *         self._buy_levels = order_levels             # <<<<<<<<<<<<<<
 *         self._sell_levels = order_levels
 *         self._order_level_spread = order_level_spread
*/
  __pyx_t_7 = __Pyx_PyLong_As_int(__pyx_v_order_levels); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 103, __pyx_L1_error)
  __pyx_v_self->_buy_levels = __pyx_t_7;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":104
 *         self._order_levels = order_levels
 *         self._buy_levels = order_levels
 *         self._sell_levels = order_levels             # <<<<<<<<<<<<<<
 *         self._order_level_spread = order_level_spread
 *         self._order_level_amount = order_level_amount
*/
  __pyx_t_7 = __Pyx_PyLong_As_int(__pyx_v_order_levels); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 104, __pyx_L1_error)
  __pyx_v_self->_sell_levels = __pyx_t_7;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":105
 *         self._buy_levels = order_levels
 *         self._sell_levels = order_levels
 *         self._order_level_spread = order_level_spread             # <<<<<<<<<<<<<<
 *         self._order_level_amount = order_level_amount
 *         self._order_refresh_time = order_refresh_time
*/
  __Pyx_INCREF(__pyx_v_order_level_spread);
  __Pyx_GIVEREF(__pyx_v_order_level_spread);
  __Pyx_GOTREF(__pyx_v_self->_order_level_spread);
  __Pyx_DECREF(__pyx_v_self->_order_level_spread);
  __pyx_v_self->_order_level_spread = __pyx_v_order_level_spread;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":106
 *         self._sell_levels = order_levels
 *         self._order_level_spread = order_level_spread
 *         self._order_level_amount = order_level_amount             # <<<<<<<<<<<<<<
 *         self._order_refresh_time = order_refresh_time
 *         self._max_order_age = max_order_age
*/
  __Pyx_INCREF(__pyx_v_order_level_amount);
  __Pyx_GIVEREF(__pyx_v_order_level_amount);
  __Pyx_GOTREF(__pyx_v_self->_order_level_amount);
  __Pyx_DECREF(__pyx_v_self->_order_level_amount);
  __pyx_v_self->_order_level_amount = __pyx_v_order_level_amount;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":107
 *         self._order_level_spread = order_level_spread
 *         self._order_level_amount = order_level_amount
 *         self._order_refresh_time = order_refresh_time             # <<<<<<<<<<<<<<
 *         self._max_order_age = max_order_age
 *         self._order_refresh_tolerance_pct = order_refresh_tolerance_pct
*/
  __pyx_t_8 = __Pyx_PyFloat_AsDouble(__pyx_v_order_refresh_time); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 107, __pyx_L1_error)
  __pyx_v_self->_order_refresh_time = __pyx_t_8;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":108
 *         self._order_level_amount = order_level_amount
 *         self._order_refresh_time = order_refresh_time
 *         self._max_order_age = max_order_age             # <<<<<<<<<<<<<<
 *         self._order_refresh_tolerance_pct = order_refresh_tolerance_pct
 *         self._filled_order_delay = filled_order_delay
*/
  __pyx_t_8 = __Pyx_PyFloat_AsDouble(__pyx_v_max_order_age); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 108, __pyx_L1_error)
  __pyx_v_self->_max_order_age = __pyx_t_8;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":109
 *         self._order_refresh_time = order_refresh_time
 *         self._max_order_age = max_order_age
 *         self._order_refresh_tolerance_pct = order_refresh_tolerance_pct             # <<<<<<<<<<<<<<
 *         self._filled_order_delay = filled_order_delay
 *         self._inventory_skew_enabled = inventory_skew_enabled
*/
  __Pyx_INCREF(__pyx_v_order_refresh_tolerance_pct);
  __Pyx_GIVEREF(__pyx_v_order_refresh_tolerance_pct);
  __Pyx_GOTREF(__pyx_v_self->_order_refresh_tolerance_pct);
  __Pyx_DECREF(__pyx_v_self->_order_refresh_tolerance_pct);
  __pyx_v_self->_order_refresh_tolerance_pct = __pyx_v_order_refresh_tolerance_pct;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":110
 *         self._max_order_age = max_order_age
 *         self._order_refresh_tolerance_pct = order_refresh_tolerance_pct
 *         self._filled_order_delay = filled_order_delay             # <<<<<<<<<<<<<<
 *         self._inventory_skew_enabled = inventory_skew_enabled
 *         self._inventory_target_base_pct = inventory_target_base_pct
*/
  __pyx_t_8 = __Pyx_PyFloat_AsDouble(__pyx_v_filled_order_delay); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)
  __pyx_v_self->_filled_order_delay = __pyx_t_8;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":111
 *         self._order_refresh_tolerance_pct = order_refresh_tolerance_pct
 *         self._filled_order_delay = filled_order_delay
 *         self._inventory_skew_enabled = inventory_skew_enabled             # <<<<<<<<<<<<<<
 *         self._inventory_target_base_pct = inventory_target_base_pct
 *         self._inventory_range_multiplier = inventory_range_multiplier
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_inventory_skew_enabled); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 111, __pyx_L1_error)
  __pyx_v_self->_inventory_skew_enabled = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":112
 *         self._filled_order_delay = filled_order_delay
 *         self._inventory_skew_enabled = inventory_skew_enabled
 *         self._inventory_target_base_pct = inventory_target_base_pct             # <<<<<<<<<<<<<<
 *         self._inventory_range_multiplier = inventory_range_multiplier
 *         self._hanging_orders_enabled = hanging_orders_enabled
*/
  __Pyx_INCREF(__pyx_v_inventory_target_base_pct);
  __Pyx_GIVEREF(__pyx_v_inventory_target_base_pct);
  __Pyx_GOTREF(__pyx_v_self->_inventory_target_base_pct);
  __Pyx_DECREF(__pyx_v_self->_inventory_target_base_pct);
  __pyx_v_self->_inventory_target_base_pct = __pyx_v_inventory_target_base_pct;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":113
 *         self._inventory_skew_enabled = inventory_skew_enabled
 *         self._inventory_target_base_pct = inventory_target_base_pct
 *         self._inventory_range_multiplier = inventory_range_multiplier             # <<<<<<<<<<<<<<
 *         self._hanging_orders_enabled = hanging_orders_enabled
 *         self._hanging_orders_tracker = HangingOrdersTracker(self, hanging_orders_cancel_pct)
*/
  __Pyx_INCREF(__pyx_v_inventory_range_multiplier);
  __Pyx_GIVEREF(__pyx_v_inventory_range_multiplier);
  __Pyx_GOTREF(__pyx_v_self->_inventory_range_multiplier);
  __Pyx_DECREF(__pyx_v_self->_inventory_range_multiplier);
  __pyx_v_self->_inventory_range_multiplier = __pyx_v_inventory_range_multiplier;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":114
 *         self._inventory_target_base_pct = inventory_target_base_pct
 *         self._inventory_range_multiplier = inventory_range_multiplier
 *         self._hanging_orders_enabled = hanging_orders_enabled             # <<<<<<<<<<<<<<
 *         self._hanging_orders_tracker = HangingOrdersTracker(self, hanging_orders_cancel_pct)
 *         self._order_optimization_enabled = order_optimization_enabled
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_hanging_orders_enabled); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 114, __pyx_L1_error)
  __pyx_v_self->_hanging_orders_enabled = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":115
 *         self._inventory_range_multiplier = inventory_range_multiplier
 *         self._hanging_orders_enabled = hanging_orders_enabled
 *         self._hanging_orders_tracker = HangingOrdersTracker(self, hanging_orders_cancel_pct)             # <<<<<<<<<<<<<<
 *         self._order_optimization_enabled = order_optimization_enabled
 *         self._ask_order_optimization_depth = ask_order_optimization_depth
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_HangingOrdersTracker); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_v_hanging_orders_cancel_pct};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_hanging_orders_tracker);
  __Pyx_DECREF(__pyx_v_self->_hanging_orders_tracker);
  __pyx_v_self->_hanging_orders_tracker = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":116
 *         self._hanging_orders_enabled = hanging_orders_enabled
 *         self._hanging_orders_tracker = HangingOrdersTracker(self, hanging_orders_cancel_pct)
 *         self._order_optimization_enabled = order_optimization_enabled             # <<<<<<<<<<<<<<
 *         self._ask_order_optimization_depth = ask_order_optimization_depth
 *         self._bid_order_optimization_depth = bid_order_optimization_depth
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_order_optimization_enabled); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 116, __pyx_L1_error)
  __pyx_v_self->_order_optimization_enabled = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":117
 *         self._hanging_orders_tracker = HangingOrdersTracker(self, hanging_orders_cancel_pct)
 *         self._order_optimization_enabled = order_optimization_enabled
 *         self._ask_order_optimization_depth = ask_order_optimization_depth             # <<<<<<<<<<<<<<
 *         self._bid_order_optimization_depth = bid_order_optimization_depth
 *         self._add_transaction_costs_to_orders = add_transaction_costs_to_orders
*/
  __Pyx_INCREF(__pyx_v_ask_order_optimization_depth);
  __Pyx_GIVEREF(__pyx_v_ask_order_optimization_depth);
  __Pyx_GOTREF(__pyx_v_self->_ask_order_optimization_depth);
  __Pyx_DECREF(__pyx_v_self->_ask_order_optimization_depth);
  __pyx_v_self->_ask_order_optimization_depth = __pyx_v_ask_order_optimization_depth;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":118
 *         self._order_optimization_enabled = order_optimization_enabled
 *         self._ask_order_optimization_depth = ask_order_optimization_depth
 *         self._bid_order_optimization_depth = bid_order_optimization_depth             # <<<<<<<<<<<<<<
 *         self._add_transaction_costs_to_orders = add_transaction_costs_to_orders
 *         self._asset_price_delegate = asset_price_delegate
*/
  __Pyx_INCREF(__pyx_v_bid_order_optimization_depth);
  __Pyx_GIVEREF(__pyx_v_bid_order_optimization_depth);
  __Pyx_GOTREF(__pyx_v_self->_bid_order_optimization_depth);
  __Pyx_DECREF(__pyx_v_self->_bid_order_optimization_depth);
  __pyx_v_self->_bid_order_optimization_depth = __pyx_v_bid_order_optimization_depth;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":119
 *         self._ask_order_optimization_depth = ask_order_optimization_depth
 *         self._bid_order_optimization_depth = bid_order_optimization_depth
 *         self._add_transaction_costs_to_orders = add_transaction_costs_to_orders             # <<<<<<<<<<<<<<
 *         self._asset_price_delegate = asset_price_delegate
 *         self._inventory_cost_price_delegate = inventory_cost_price_delegate
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_add_transaction_costs_to_orders); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 119, __pyx_L1_error)
  __pyx_v_self->_add_transaction_costs_to_orders = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":120
 *         self._bid_order_optimization_depth = bid_order_optimization_depth
 *         self._add_transaction_costs_to_orders = add_transaction_costs_to_orders
 *         self._asset_price_delegate = asset_price_delegate             # <<<<<<<<<<<<<<
 *         self._inventory_cost_price_delegate = inventory_cost_price_delegate
 *         self._price_type = self.get_price_type(price_type)
*/
  __Pyx_INCREF(__pyx_v_asset_price_delegate);
  __Pyx_GIVEREF(__pyx_v_asset_price_delegate);
  __Pyx_GOTREF(__pyx_v_self->_asset_price_delegate);
  __Pyx_DECREF(__pyx_v_self->_asset_price_delegate);
  __pyx_v_self->_asset_price_delegate = __pyx_v_asset_price_delegate;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":121
 *         self._add_transaction_costs_to_orders = add_transaction_costs_to_orders
 *         self._asset_price_delegate = asset_price_delegate
 *         self._inventory_cost_price_delegate = inventory_cost_price_delegate             # <<<<<<<<<<<<<<
 *         self._price_type = self.get_price_type(price_type)
 *         self._take_if_crossed = take_if_crossed
*/
  __Pyx_INCREF(__pyx_v_inventory_cost_price_delegate);
  __Pyx_GIVEREF(__pyx_v_inventory_cost_price_delegate);
  __Pyx_GOTREF(__pyx_v_self->_inventory_cost_price_delegate);
  __Pyx_DECREF(__pyx_v_self->_inventory_cost_price_delegate);
  __pyx_v_self->_inventory_cost_price_delegate = __pyx_v_inventory_cost_price_delegate;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":122
 *         self._asset_price_delegate = asset_price_delegate
 *         self._inventory_cost_price_delegate = inventory_cost_price_delegate
 *         self._price_type = self.get_price_type(price_type)             # <<<<<<<<<<<<<<
 *         self._take_if_crossed = take_if_crossed
 *         self._price_ceiling = price_ceiling
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_price_type};
    __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price_type, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_price_type);
  __Pyx_DECREF(__pyx_v_self->_price_type);
  __pyx_v_self->_price_type = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":123
 *         self._inventory_cost_price_delegate = inventory_cost_price_delegate
 *         self._price_type = self.get_price_type(price_type)
 *         self._take_if_crossed = take_if_crossed             # <<<<<<<<<<<<<<
 *         self._price_ceiling = price_ceiling
 *         self._price_floor = price_floor
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_take_if_crossed); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 123, __pyx_L1_error)
  __pyx_v_self->_take_if_crossed = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":124
 *         self._price_type = self.get_price_type(price_type)
 *         self._take_if_crossed = take_if_crossed
 *         self._price_ceiling = price_ceiling             # <<<<<<<<<<<<<<
 *         self._price_floor = price_floor
 *         self._ping_pong_enabled = ping_pong_enabled
*/
  __Pyx_INCREF(__pyx_v_price_ceiling);
  __Pyx_GIVEREF(__pyx_v_price_ceiling);
  __Pyx_GOTREF(__pyx_v_self->_price_ceiling);
  __Pyx_DECREF(__pyx_v_self->_price_ceiling);
  __pyx_v_self->_price_ceiling = __pyx_v_price_ceiling;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":125
 *         self._take_if_crossed = take_if_crossed
 *         self._price_ceiling = price_ceiling
 *         self._price_floor = price_floor             # <<<<<<<<<<<<<<
 *         self._ping_pong_enabled = ping_pong_enabled
 *         self._ping_pong_warning_lines = []
*/
  __Pyx_INCREF(__pyx_v_price_floor);
  __Pyx_GIVEREF(__pyx_v_price_floor);
  __Pyx_GOTREF(__pyx_v_self->_price_floor);
  __Pyx_DECREF(__pyx_v_self->_price_floor);
  __pyx_v_self->_price_floor = __pyx_v_price_floor;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":126
 *         self._price_ceiling = price_ceiling
 *         self._price_floor = price_floor
 *         self._ping_pong_enabled = ping_pong_enabled             # <<<<<<<<<<<<<<
 *         self._ping_pong_warning_lines = []
 *         self._hb_app_notification = hb_app_notification
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_ping_pong_enabled); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 126, __pyx_L1_error)
  __pyx_v_self->_ping_pong_enabled = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":127
 *         self._price_floor = price_floor
 *         self._ping_pong_enabled = ping_pong_enabled
 *         self._ping_pong_warning_lines = []             # <<<<<<<<<<<<<<
 *         self._hb_app_notification = hb_app_notification
 *         self._order_override = order_override
*/
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_ping_pong_warning_lines);
  __Pyx_DECREF(__pyx_v_self->_ping_pong_warning_lines);
  __pyx_v_self->_ping_pong_warning_lines = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":128
 *         self._ping_pong_enabled = ping_pong_enabled
 *         self._ping_pong_warning_lines = []
 *         self._hb_app_notification = hb_app_notification             # <<<<<<<<<<<<<<
 *         self._order_override = order_override
 *         self._split_order_levels_enabled=split_order_levels_enabled
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_hb_app_notification); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 128, __pyx_L1_error)
  __pyx_v_self->_hb_app_notification = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":129
 *         self._ping_pong_warning_lines = []
 *         self._hb_app_notification = hb_app_notification
 *         self._order_override = order_override             # <<<<<<<<<<<<<<
 *         self._split_order_levels_enabled=split_order_levels_enabled
 *         self._bid_order_level_spreads=bid_order_level_spreads
*/
  __Pyx_INCREF(__pyx_v_order_override);
  __Pyx_GIVEREF(__pyx_v_order_override);
  __Pyx_GOTREF(__pyx_v_self->_order_override);
  __Pyx_DECREF(__pyx_v_self->_order_override);
  __pyx_v_self->_order_override = __pyx_v_order_override;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":130
 *         self._hb_app_notification = hb_app_notification
 *         self._order_override = order_override
 *         self._split_order_levels_enabled=split_order_levels_enabled             # <<<<<<<<<<<<<<
 *         self._bid_order_level_spreads=bid_order_level_spreads
 *         self._ask_order_level_spreads=ask_order_level_spreads
*/
  __Pyx_INCREF(__pyx_v_split_order_levels_enabled);
  __Pyx_GIVEREF(__pyx_v_split_order_levels_enabled);
  __Pyx_GOTREF(__pyx_v_self->_split_order_levels_enabled);
  __Pyx_DECREF(__pyx_v_self->_split_order_levels_enabled);
  __pyx_v_self->_split_order_levels_enabled = __pyx_v_split_order_levels_enabled;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":131
 *         self._order_override = order_override
 *         self._split_order_levels_enabled=split_order_levels_enabled
 *         self._bid_order_level_spreads=bid_order_level_spreads             # <<<<<<<<<<<<<<
 *         self._ask_order_level_spreads=ask_order_level_spreads
 *         self._cancel_timestamp = 0
*/
  __Pyx_INCREF(__pyx_v_bid_order_level_spreads);
  __Pyx_GIVEREF(__pyx_v_bid_order_level_spreads);
  __Pyx_GOTREF(__pyx_v_self->_bid_order_level_spreads);
  __Pyx_DECREF(__pyx_v_self->_bid_order_level_spreads);
  __pyx_v_self->_bid_order_level_spreads = __pyx_v_bid_order_level_spreads;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":132
 *         self._split_order_levels_enabled=split_order_levels_enabled
 *         self._bid_order_level_spreads=bid_order_level_spreads
 *         self._ask_order_level_spreads=ask_order_level_spreads             # <<<<<<<<<<<<<<
 *         self._cancel_timestamp = 0
 *         self._create_timestamp = 0
*/
  __Pyx_INCREF(__pyx_v_ask_order_level_spreads);
  __Pyx_GIVEREF(__pyx_v_ask_order_level_spreads);
  __Pyx_GOTREF(__pyx_v_self->_ask_order_level_spreads);
  __Pyx_DECREF(__pyx_v_self->_ask_order_level_spreads);
  __pyx_v_self->_ask_order_level_spreads = __pyx_v_ask_order_level_spreads;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":133
 *         self._bid_order_level_spreads=bid_order_level_spreads
 *         self._ask_order_level_spreads=ask_order_level_spreads
 *         self._cancel_timestamp = 0             # <<<<<<<<<<<<<<
 *         self._create_timestamp = 0
 *         self._limit_order_type = self._market_info.market.get_maker_order_type()
*/
  __pyx_v_self->_cancel_timestamp = 0.0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":134
 *         self._ask_order_level_spreads=ask_order_level_spreads
 *         self._cancel_timestamp = 0
 *         self._create_timestamp = 0             # <<<<<<<<<<<<<<
 *         self._limit_order_type = self._market_info.market.get_maker_order_type()
 *         if take_if_crossed:
*/
  __pyx_v_self->_create_timestamp = 0.0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":135
 *         self._cancel_timestamp = 0
 *         self._create_timestamp = 0
 *         self._limit_order_type = self._market_info.market.get_maker_order_type()             # <<<<<<<<<<<<<<
 *         if take_if_crossed:
 *             self._limit_order_type = OrderType.LIMIT
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_maker_order_type, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_limit_order_type);
  __Pyx_DECREF(__pyx_v_self->_limit_order_type);
  __pyx_v_self->_limit_order_type = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":136
 *         self._create_timestamp = 0
 *         self._limit_order_type = self._market_info.market.get_maker_order_type()
 *         if take_if_crossed:             # <<<<<<<<<<<<<<
 *             self._limit_order_type = OrderType.LIMIT
 *         self._all_markets_ready = False
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_take_if_crossed); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 136, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":137
 *         self._limit_order_type = self._market_info.market.get_maker_order_type()
 *         if take_if_crossed:
 *             self._limit_order_type = OrderType.LIMIT             # <<<<<<<<<<<<<<
 *         self._all_markets_ready = False
 *         self._filled_buys_balance = 0
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_OrderType); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_LIMIT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_limit_order_type);
    __Pyx_DECREF(__pyx_v_self->_limit_order_type);
    __pyx_v_self->_limit_order_type = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":136
 *         self._create_timestamp = 0
 *         self._limit_order_type = self._market_info.market.get_maker_order_type()
 *         if take_if_crossed:             # <<<<<<<<<<<<<<
 *             self._limit_order_type = OrderType.LIMIT
 *         self._all_markets_ready = False
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":138
 *         if take_if_crossed:
 *             self._limit_order_type = OrderType.LIMIT
 *         self._all_markets_ready = False             # <<<<<<<<<<<<<<
 *         self._filled_buys_balance = 0
 *         self._filled_sells_balance = 0
*/
  __pyx_v_self->_all_markets_ready = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":139
 *             self._limit_order_type = OrderType.LIMIT
 *         self._all_markets_ready = False
 *         self._filled_buys_balance = 0             # <<<<<<<<<<<<<<
 *         self._filled_sells_balance = 0
 *         self._logging_options = logging_options
*/
  __pyx_v_self->_filled_buys_balance = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":140
 *         self._all_markets_ready = False
 *         self._filled_buys_balance = 0
 *         self._filled_sells_balance = 0             # <<<<<<<<<<<<<<
 *         self._logging_options = logging_options
 *         self._last_timestamp = 0
*/
  __pyx_v_self->_filled_sells_balance = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":141
 *         self._filled_buys_balance = 0
 *         self._filled_sells_balance = 0
 *         self._logging_options = logging_options             # <<<<<<<<<<<<<<
 *         self._last_timestamp = 0
 *         self._status_report_interval = status_report_interval
*/
  __pyx_t_9 = __Pyx_PyLong_As_int64_t(__pyx_v_logging_options); if (unlikely((__pyx_t_9 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 141, __pyx_L1_error)
  __pyx_v_self->_logging_options = __pyx_t_9;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":142
 *         self._filled_sells_balance = 0
 *         self._logging_options = logging_options
 *         self._last_timestamp = 0             # <<<<<<<<<<<<<<
 *         self._status_report_interval = status_report_interval
 *         self._last_own_trade_price = Decimal('nan')
*/
  __pyx_v_self->_last_timestamp = 0.0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":143
 *         self._logging_options = logging_options
 *         self._last_timestamp = 0
 *         self._status_report_interval = status_report_interval             # <<<<<<<<<<<<<<
 *         self._last_own_trade_price = Decimal('nan')
 *         self._should_wait_order_cancel_confirmation = should_wait_order_cancel_confirmation
*/
  __pyx_t_8 = __Pyx_PyFloat_AsDouble(__pyx_v_status_report_interval); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 143, __pyx_L1_error)
  __pyx_v_self->_status_report_interval = __pyx_t_8;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":144
 *         self._last_timestamp = 0
 *         self._status_report_interval = status_report_interval
 *         self._last_own_trade_price = Decimal('nan')             # <<<<<<<<<<<<<<
 *         self._should_wait_order_cancel_confirmation = should_wait_order_cancel_confirmation
 *         self._moving_price_band = moving_price_band
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_nan};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_last_own_trade_price);
  __Pyx_DECREF(__pyx_v_self->_last_own_trade_price);
  __pyx_v_self->_last_own_trade_price = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":145
 *         self._status_report_interval = status_report_interval
 *         self._last_own_trade_price = Decimal('nan')
 *         self._should_wait_order_cancel_confirmation = should_wait_order_cancel_confirmation             # <<<<<<<<<<<<<<
 *         self._moving_price_band = moving_price_band
 *         self.c_add_markets([market_info.market])
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_should_wait_order_cancel_confirmation); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 145, __pyx_L1_error)
  __pyx_v_self->_should_wait_order_cancel_confirmation = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":146
 *         self._last_own_trade_price = Decimal('nan')
 *         self._should_wait_order_cancel_confirmation = should_wait_order_cancel_confirmation
 *         self._moving_price_band = moving_price_band             # <<<<<<<<<<<<<<
 *         self.c_add_markets([market_info.market])
 * 
*/
  __Pyx_INCREF(__pyx_v_moving_price_band);
  __Pyx_GIVEREF(__pyx_v_moving_price_band);
  __Pyx_GOTREF(__pyx_v_self->_moving_price_band);
  __Pyx_DECREF(__pyx_v_self->_moving_price_band);
  __pyx_v_self->_moving_price_band = __pyx_v_moving_price_band;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":147
 *         self._should_wait_order_cancel_confirmation = should_wait_order_cancel_confirmation
 *         self._moving_price_band = moving_price_band
 *         self.c_add_markets([market_info.market])             # <<<<<<<<<<<<<<
 * 
 *     def all_markets_ready(self):
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_market_info, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_3) != (0)) __PYX_ERR(0, 147, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_3 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_add_markets(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":51
 *         return pmm_logger
 * 
 *     def init_params(self,             # <<<<<<<<<<<<<<
 *                     market_info: MarketTradingPairTuple,
 *                     bid_spread: Decimal,
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.init_params", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_order_override);
  __Pyx_XDECREF(__pyx_v_moving_price_band);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":149
 *         self.c_add_markets([market_info.market])
 * 
 *     def all_markets_ready(self):             # <<<<<<<<<<<<<<
 *         return all([market.ready for market in self._sb_markets])
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_5all_markets_ready(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_5all_markets_ready = {"all_markets_ready", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_5all_markets_ready, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_5all_markets_ready(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("all_markets_ready (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("all_markets_ready", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("all_markets_ready", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_4all_markets_ready(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_4all_markets_ready(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_7genexpr__pyx_v_market = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("all_markets_ready", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":150
 * 
 *     def all_markets_ready(self):
 *         return all([market.ready for market in self._sb_markets])             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 150, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = 0;
    __pyx_t_8 = __Pyx_set_iterator(__pyx_v_self->__pyx_base._sb_markets, 1, (&__pyx_t_6), (&__pyx_t_7)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 150, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_4);
    __pyx_t_4 = __pyx_t_8;
    __pyx_t_8 = 0;
    while (1) {
      __pyx_t_9 = __Pyx_set_iter_next(__pyx_t_4, __pyx_t_6, &__pyx_t_5, &__pyx_t_8, __pyx_t_7);
      if (unlikely(__pyx_t_9 == 0)) break;
      if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 150, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_market, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_7genexpr__pyx_v_market, __pyx_mstate_global->__pyx_n_u_ready); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 150, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 150, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_market); __pyx_7genexpr__pyx_v_market = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_market); __pyx_7genexpr__pyx_v_market = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_t_10 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_all, __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":149
 *         self.c_add_markets([market_info.market])
 * 
 *     def all_markets_ready(self):             # <<<<<<<<<<<<<<
 *         return all([market.ready for market in self._sb_markets])
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.all_markets_ready", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_market);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":152
 *         return all([market.ready for market in self._sb_markets])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def market_info(self) -> MarketTradingPairTuple:
 *         return self._market_info
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11market_info_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11market_info_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11market_info___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11market_info___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":154
 *     @property
 *     def market_info(self) -> MarketTradingPairTuple:
 *         return self._market_info             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_market_info);
  __pyx_r = __pyx_v_self->_market_info;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":152
 *         return all([market.ready for market in self._sb_markets])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def market_info(self) -> MarketTradingPairTuple:
 *         return self._market_info
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":156
 *         return self._market_info
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_order_age(self) -> float:
 *         return self._max_order_age
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13max_order_age_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13max_order_age_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13max_order_age___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13max_order_age___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":158
 *     @property
 *     def max_order_age(self) -> float:
 *         return self._max_order_age             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_max_order_age); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":156
 *         return self._market_info
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_order_age(self) -> float:
 *         return self._max_order_age
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.max_order_age.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":160
 *         return self._max_order_age
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def minimum_spread(self) -> Decimal:
 *         return self._minimum_spread
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14minimum_spread_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14minimum_spread_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14minimum_spread___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14minimum_spread___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":162
 *     @property
 *     def minimum_spread(self) -> Decimal:
 *         return self._minimum_spread             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_minimum_spread);
  __pyx_r = __pyx_v_self->_minimum_spread;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":160
 *         return self._max_order_age
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def minimum_spread(self) -> Decimal:
 *         return self._minimum_spread
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":164
 *         return self._minimum_spread
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ping_pong_enabled(self) -> bool:
 *         return self._ping_pong_enabled
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17ping_pong_enabled_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17ping_pong_enabled_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17ping_pong_enabled___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17ping_pong_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":166
 *     @property
 *     def ping_pong_enabled(self) -> bool:
 *         return self._ping_pong_enabled             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_ping_pong_enabled); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":164
 *         return self._minimum_spread
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ping_pong_enabled(self) -> bool:
 *         return self._ping_pong_enabled
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.ping_pong_enabled.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":168
 *         return self._ping_pong_enabled
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ask_order_optimization_depth(self) -> Decimal:
 *         return self._ask_order_optimization_depth
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28ask_order_optimization_depth_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28ask_order_optimization_depth_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28ask_order_optimization_depth___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28ask_order_optimization_depth___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":170
 *     @property
 *     def ask_order_optimization_depth(self) -> Decimal:
 *         return self._ask_order_optimization_depth             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_ask_order_optimization_depth);
  __pyx_r = __pyx_v_self->_ask_order_optimization_depth;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":168
 *         return self._ping_pong_enabled
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ask_order_optimization_depth(self) -> Decimal:
 *         return self._ask_order_optimization_depth
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":172
 *         return self._ask_order_optimization_depth
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def bid_order_optimization_depth(self) -> Decimal:
 *         return self._bid_order_optimization_depth
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28bid_order_optimization_depth_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28bid_order_optimization_depth_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28bid_order_optimization_depth___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28bid_order_optimization_depth___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":174
 *     @property
 *     def bid_order_optimization_depth(self) -> Decimal:
 *         return self._bid_order_optimization_depth             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_bid_order_optimization_depth);
  __pyx_r = __pyx_v_self->_bid_order_optimization_depth;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":172
 *         return self._ask_order_optimization_depth
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def bid_order_optimization_depth(self) -> Decimal:
 *         return self._bid_order_optimization_depth
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":176
 *         return self._bid_order_optimization_depth
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def price_type(self) -> PriceType:
 *         return self._price_type
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10price_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10price_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10price_type___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10price_type___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":178
 *     @property
 *     def price_type(self) -> PriceType:
 *         return self._price_type             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_price_type);
  __pyx_r = __pyx_v_self->_price_type;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":176
 *         return self._bid_order_optimization_depth
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def price_type(self) -> PriceType:
 *         return self._price_type
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":180
 *         return self._price_type
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_refresh_tolerance_pct(self) -> Decimal:
 *         return self._order_refresh_tolerance_pct
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27order_refresh_tolerance_pct_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27order_refresh_tolerance_pct_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27order_refresh_tolerance_pct___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27order_refresh_tolerance_pct___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":182
 *     @property
 *     def order_refresh_tolerance_pct(self) -> Decimal:
 *         return self._order_refresh_tolerance_pct             # <<<<<<<<<<<<<<
 * 
 *     @order_refresh_tolerance_pct.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_order_refresh_tolerance_pct);
  __pyx_r = __pyx_v_self->_order_refresh_tolerance_pct;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":180
 *         return self._price_type
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_refresh_tolerance_pct(self) -> Decimal:
 *         return self._order_refresh_tolerance_pct
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":184
 *         return self._order_refresh_tolerance_pct
 * 
 *     @order_refresh_tolerance_pct.setter             # <<<<<<<<<<<<<<
 *     def order_refresh_tolerance_pct(self, value: Decimal):
 *         self._order_refresh_tolerance_pct = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27order_refresh_tolerance_pct_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27order_refresh_tolerance_pct_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27order_refresh_tolerance_pct_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27order_refresh_tolerance_pct_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":186
 *     @order_refresh_tolerance_pct.setter
 *     def order_refresh_tolerance_pct(self, value: Decimal):
 *         self._order_refresh_tolerance_pct = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_order_refresh_tolerance_pct);
  __Pyx_DECREF(__pyx_v_self->_order_refresh_tolerance_pct);
  __pyx_v_self->_order_refresh_tolerance_pct = __pyx_v_value;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":184
 *         return self._order_refresh_tolerance_pct
 * 
 *     @order_refresh_tolerance_pct.setter             # <<<<<<<<<<<<<<
 *     def order_refresh_tolerance_pct(self, value: Decimal):
 *         self._order_refresh_tolerance_pct = value
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":188
 *         self._order_refresh_tolerance_pct = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_amount(self) -> Decimal:
 *         return self._order_amount
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_amount_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_amount_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_amount___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_amount___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":190
 *     @property
 *     def order_amount(self) -> Decimal:
 *         return self._order_amount             # <<<<<<<<<<<<<<
 * 
 *     @order_amount.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_order_amount);
  __pyx_r = __pyx_v_self->_order_amount;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":188
 *         self._order_refresh_tolerance_pct = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_amount(self) -> Decimal:
 *         return self._order_amount
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":192
 *         return self._order_amount
 * 
 *     @order_amount.setter             # <<<<<<<<<<<<<<
 *     def order_amount(self, value: Decimal):
 *         self._order_amount = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_amount_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_amount_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_amount_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_amount_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":194
 *     @order_amount.setter
 *     def order_amount(self, value: Decimal):
 *         self._order_amount = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_order_amount);
  __Pyx_DECREF(__pyx_v_self->_order_amount);
  __pyx_v_self->_order_amount = __pyx_v_value;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":192
 *         return self._order_amount
 * 
 *     @order_amount.setter             # <<<<<<<<<<<<<<
 *     def order_amount(self, value: Decimal):
 *         self._order_amount = value
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":196
 *         self._order_amount = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_levels(self) -> int:
 *         return self._order_levels
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_levels_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_levels_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_levels___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_levels___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":198
 *     @property
 *     def order_levels(self) -> int:
 *         return self._order_levels             # <<<<<<<<<<<<<<
 * 
 *     @order_levels.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->_order_levels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":196
 *         self._order_amount = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_levels(self) -> int:
 *         return self._order_levels
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.order_levels.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":200
 *         return self._order_levels
 * 
 *     @order_levels.setter             # <<<<<<<<<<<<<<
 *     def order_levels(self, value: int):
 *         self._order_levels = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_levels_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_levels_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_levels_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_levels_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":202
 *     @order_levels.setter
 *     def order_levels(self, value: int):
 *         self._order_levels = value             # <<<<<<<<<<<<<<
 *         self._buy_levels = value
 *         self._sell_levels = value
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 202, __pyx_L1_error)
  __pyx_v_self->_order_levels = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":203
 *     def order_levels(self, value: int):
 *         self._order_levels = value
 *         self._buy_levels = value             # <<<<<<<<<<<<<<
 *         self._sell_levels = value
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 203, __pyx_L1_error)
  __pyx_v_self->_buy_levels = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":204
 *         self._order_levels = value
 *         self._buy_levels = value
 *         self._sell_levels = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 204, __pyx_L1_error)
  __pyx_v_self->_sell_levels = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":200
 *         return self._order_levels
 * 
 *     @order_levels.setter             # <<<<<<<<<<<<<<
 *     def order_levels(self, value: int):
 *         self._order_levels = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.order_levels.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":206
 *         self._sell_levels = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def buy_levels(self) -> int:
 *         return self._buy_levels
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10buy_levels_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10buy_levels_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10buy_levels___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10buy_levels___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":208
 *     @property
 *     def buy_levels(self) -> int:
 *         return self._buy_levels             # <<<<<<<<<<<<<<
 * 
 *     @buy_levels.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->_buy_levels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":206
 *         self._sell_levels = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def buy_levels(self) -> int:
 *         return self._buy_levels
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.buy_levels.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":210
 *         return self._buy_levels
 * 
 *     @buy_levels.setter             # <<<<<<<<<<<<<<
 *     def buy_levels(self, value: int):
 *         self._buy_levels = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10buy_levels_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10buy_levels_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10buy_levels_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10buy_levels_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":212
 *     @buy_levels.setter
 *     def buy_levels(self, value: int):
 *         self._buy_levels = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 212, __pyx_L1_error)
  __pyx_v_self->_buy_levels = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":210
 *         return self._buy_levels
 * 
 *     @buy_levels.setter             # <<<<<<<<<<<<<<
 *     def buy_levels(self, value: int):
 *         self._buy_levels = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.buy_levels.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":214
 *         self._buy_levels = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sell_levels(self) -> int:
 *         return self._sell_levels
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11sell_levels_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11sell_levels_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11sell_levels___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11sell_levels___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":216
 *     @property
 *     def sell_levels(self) -> int:
 *         return self._sell_levels             # <<<<<<<<<<<<<<
 * 
 *     @sell_levels.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->_sell_levels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":214
 *         self._buy_levels = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sell_levels(self) -> int:
 *         return self._sell_levels
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.sell_levels.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":218
 *         return self._sell_levels
 * 
 *     @sell_levels.setter             # <<<<<<<<<<<<<<
 *     def sell_levels(self, value: int):
 *         self._sell_levels = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11sell_levels_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11sell_levels_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11sell_levels_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11sell_levels_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":220
 *     @sell_levels.setter
 *     def sell_levels(self, value: int):
 *         self._sell_levels = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_v_self->_sell_levels = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":218
 *         return self._sell_levels
 * 
 *     @sell_levels.setter             # <<<<<<<<<<<<<<
 *     def sell_levels(self, value: int):
 *         self._sell_levels = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.sell_levels.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":222
 *         self._sell_levels = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_level_amount(self) -> Decimal:
 *         return self._order_level_amount
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_amount_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_amount_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_amount___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_amount___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":224
 *     @property
 *     def order_level_amount(self) -> Decimal:
 *         return self._order_level_amount             # <<<<<<<<<<<<<<
 * 
 *     @order_level_amount.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_order_level_amount);
  __pyx_r = __pyx_v_self->_order_level_amount;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":222
 *         self._sell_levels = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_level_amount(self) -> Decimal:
 *         return self._order_level_amount
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":226
 *         return self._order_level_amount
 * 
 *     @order_level_amount.setter             # <<<<<<<<<<<<<<
 *     def order_level_amount(self, value: Decimal):
 *         self._order_level_amount = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_amount_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_amount_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_amount_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_amount_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":228
 *     @order_level_amount.setter
 *     def order_level_amount(self, value: Decimal):
 *         self._order_level_amount = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_order_level_amount);
  __Pyx_DECREF(__pyx_v_self->_order_level_amount);
  __pyx_v_self->_order_level_amount = __pyx_v_value;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":226
 *         return self._order_level_amount
 * 
 *     @order_level_amount.setter             # <<<<<<<<<<<<<<
 *     def order_level_amount(self, value: Decimal):
 *         self._order_level_amount = value
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":230
 *         self._order_level_amount = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_level_spread(self) -> Decimal:
 *         return self._order_level_spread
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_spread_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_spread_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_spread___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_spread___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":232
 *     @property
 *     def order_level_spread(self) -> Decimal:
 *         return self._order_level_spread             # <<<<<<<<<<<<<<
 * 
 *     @order_level_spread.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_order_level_spread);
  __pyx_r = __pyx_v_self->_order_level_spread;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":230
 *         self._order_level_amount = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_level_spread(self) -> Decimal:
 *         return self._order_level_spread
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":234
 *         return self._order_level_spread
 * 
 *     @order_level_spread.setter             # <<<<<<<<<<<<<<
 *     def order_level_spread(self, value: Decimal):
 *         self._order_level_spread = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_spread_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_spread_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_spread_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_spread_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":236
 *     @order_level_spread.setter
 *     def order_level_spread(self, value: Decimal):
 *         self._order_level_spread = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_order_level_spread);
  __Pyx_DECREF(__pyx_v_self->_order_level_spread);
  __pyx_v_self->_order_level_spread = __pyx_v_value;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":234
 *         return self._order_level_spread
 * 
 *     @order_level_spread.setter             # <<<<<<<<<<<<<<
 *     def order_level_spread(self, value: Decimal):
 *         self._order_level_spread = value
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":238
 *         self._order_level_spread = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inventory_skew_enabled(self) -> bool:
 *         return self._inventory_skew_enabled
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22inventory_skew_enabled_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22inventory_skew_enabled_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22inventory_skew_enabled___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22inventory_skew_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":240
 *     @property
 *     def inventory_skew_enabled(self) -> bool:
 *         return self._inventory_skew_enabled             # <<<<<<<<<<<<<<
 * 
 *     @inventory_skew_enabled.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_inventory_skew_enabled); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":238
 *         self._order_level_spread = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inventory_skew_enabled(self) -> bool:
 *         return self._inventory_skew_enabled
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.inventory_skew_enabled.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":242
 *         return self._inventory_skew_enabled
 * 
 *     @inventory_skew_enabled.setter             # <<<<<<<<<<<<<<
 *     def inventory_skew_enabled(self, value: bool):
 *         self._inventory_skew_enabled = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22inventory_skew_enabled_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22inventory_skew_enabled_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22inventory_skew_enabled_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22inventory_skew_enabled_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":244
 *     @inventory_skew_enabled.setter
 *     def inventory_skew_enabled(self, value: bool):
 *         self._inventory_skew_enabled = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 244, __pyx_L1_error)
  __pyx_v_self->_inventory_skew_enabled = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":242
 *         return self._inventory_skew_enabled
 * 
 *     @inventory_skew_enabled.setter             # <<<<<<<<<<<<<<
 *     def inventory_skew_enabled(self, value: bool):
 *         self._inventory_skew_enabled = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.inventory_skew_enabled.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":246
 *         self._inventory_skew_enabled = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inventory_target_base_pct(self) -> Decimal:
 *         return self._inventory_target_base_pct
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25inventory_target_base_pct_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25inventory_target_base_pct_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25inventory_target_base_pct___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25inventory_target_base_pct___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":248
 *     @property
 *     def inventory_target_base_pct(self) -> Decimal:
 *         return self._inventory_target_base_pct             # <<<<<<<<<<<<<<
 * 
 *     @inventory_target_base_pct.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_inventory_target_base_pct);
  __pyx_r = __pyx_v_self->_inventory_target_base_pct;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":246
 *         self._inventory_skew_enabled = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inventory_target_base_pct(self) -> Decimal:
 *         return self._inventory_target_base_pct
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":250
 *         return self._inventory_target_base_pct
 * 
 *     @inventory_target_base_pct.setter             # <<<<<<<<<<<<<<
 *     def inventory_target_base_pct(self, value: Decimal):
 *         self._inventory_target_base_pct = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25inventory_target_base_pct_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25inventory_target_base_pct_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25inventory_target_base_pct_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25inventory_target_base_pct_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":252
 *     @inventory_target_base_pct.setter
 *     def inventory_target_base_pct(self, value: Decimal):
 *         self._inventory_target_base_pct = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_inventory_target_base_pct);
  __Pyx_DECREF(__pyx_v_self->_inventory_target_base_pct);
  __pyx_v_self->_inventory_target_base_pct = __pyx_v_value;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":250
 *         return self._inventory_target_base_pct
 * 
 *     @inventory_target_base_pct.setter             # <<<<<<<<<<<<<<
 *     def inventory_target_base_pct(self, value: Decimal):
 *         self._inventory_target_base_pct = value
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":254
 *         self._inventory_target_base_pct = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inventory_range_multiplier(self) -> Decimal:
 *         return self._inventory_range_multiplier
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26inventory_range_multiplier_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26inventory_range_multiplier_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26inventory_range_multiplier___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26inventory_range_multiplier___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":256
 *     @property
 *     def inventory_range_multiplier(self) -> Decimal:
 *         return self._inventory_range_multiplier             # <<<<<<<<<<<<<<
 * 
 *     @inventory_range_multiplier.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_inventory_range_multiplier);
  __pyx_r = __pyx_v_self->_inventory_range_multiplier;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":254
 *         self._inventory_target_base_pct = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inventory_range_multiplier(self) -> Decimal:
 *         return self._inventory_range_multiplier
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":258
 *         return self._inventory_range_multiplier
 * 
 *     @inventory_range_multiplier.setter             # <<<<<<<<<<<<<<
 *     def inventory_range_multiplier(self, value: Decimal):
 *         self._inventory_range_multiplier = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26inventory_range_multiplier_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26inventory_range_multiplier_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26inventory_range_multiplier_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26inventory_range_multiplier_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":260
 *     @inventory_range_multiplier.setter
 *     def inventory_range_multiplier(self, value: Decimal):
 *         self._inventory_range_multiplier = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_inventory_range_multiplier);
  __Pyx_DECREF(__pyx_v_self->_inventory_range_multiplier);
  __pyx_v_self->_inventory_range_multiplier = __pyx_v_value;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":258
 *         return self._inventory_range_multiplier
 * 
 *     @inventory_range_multiplier.setter             # <<<<<<<<<<<<<<
 *     def inventory_range_multiplier(self, value: Decimal):
 *         self._inventory_range_multiplier = value
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":262
 *         self._inventory_range_multiplier = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def hanging_orders_enabled(self) -> bool:
 *         return self._hanging_orders_enabled
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_enabled_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_enabled_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_enabled___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":264
 *     @property
 *     def hanging_orders_enabled(self) -> bool:
 *         return self._hanging_orders_enabled             # <<<<<<<<<<<<<<
 * 
 *     @hanging_orders_enabled.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_hanging_orders_enabled); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":262
 *         self._inventory_range_multiplier = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def hanging_orders_enabled(self) -> bool:
 *         return self._hanging_orders_enabled
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.hanging_orders_enabled.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":266
 *         return self._hanging_orders_enabled
 * 
 *     @hanging_orders_enabled.setter             # <<<<<<<<<<<<<<
 *     def hanging_orders_enabled(self, value: bool):
 *         self._hanging_orders_enabled = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_enabled_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_enabled_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_enabled_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_enabled_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":268
 *     @hanging_orders_enabled.setter
 *     def hanging_orders_enabled(self, value: bool):
 *         self._hanging_orders_enabled = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 268, __pyx_L1_error)
  __pyx_v_self->_hanging_orders_enabled = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":266
 *         return self._hanging_orders_enabled
 * 
 *     @hanging_orders_enabled.setter             # <<<<<<<<<<<<<<
 *     def hanging_orders_enabled(self, value: bool):
 *         self._hanging_orders_enabled = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.hanging_orders_enabled.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":270
 *         self._hanging_orders_enabled = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def hanging_orders_cancel_pct(self) -> Decimal:
 *         return self._hanging_orders_tracker._hanging_orders_cancel_pct
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25hanging_orders_cancel_pct_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25hanging_orders_cancel_pct_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25hanging_orders_cancel_pct___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25hanging_orders_cancel_pct___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":272
 *     @property
 *     def hanging_orders_cancel_pct(self) -> Decimal:
 *         return self._hanging_orders_tracker._hanging_orders_cancel_pct             # <<<<<<<<<<<<<<
 * 
 *     @hanging_orders_cancel_pct.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_mstate_global->__pyx_n_u_hanging_orders_cancel_pct_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":270
 *         self._hanging_orders_enabled = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def hanging_orders_cancel_pct(self) -> Decimal:
 *         return self._hanging_orders_tracker._hanging_orders_cancel_pct
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.hanging_orders_cancel_pct.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":274
 *         return self._hanging_orders_tracker._hanging_orders_cancel_pct
 * 
 *     @hanging_orders_cancel_pct.setter             # <<<<<<<<<<<<<<
 *     def hanging_orders_cancel_pct(self, value: Decimal):
 *         self._hanging_orders_tracker._hanging_orders_cancel_pct = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25hanging_orders_cancel_pct_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25hanging_orders_cancel_pct_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25hanging_orders_cancel_pct_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25hanging_orders_cancel_pct_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":276
 *     @hanging_orders_cancel_pct.setter
 *     def hanging_orders_cancel_pct(self, value: Decimal):
 *         self._hanging_orders_tracker._hanging_orders_cancel_pct = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_mstate_global->__pyx_n_u_hanging_orders_cancel_pct_2, __pyx_v_value) < (0)) __PYX_ERR(0, 276, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":274
 *         return self._hanging_orders_tracker._hanging_orders_cancel_pct
 * 
 *     @hanging_orders_cancel_pct.setter             # <<<<<<<<<<<<<<
 *     def hanging_orders_cancel_pct(self, value: Decimal):
 *         self._hanging_orders_tracker._hanging_orders_cancel_pct = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.hanging_orders_cancel_pct.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":278
 *         self._hanging_orders_tracker._hanging_orders_cancel_pct = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def bid_spread(self) -> Decimal:
 *         return self._bid_spread
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10bid_spread_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10bid_spread_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10bid_spread___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10bid_spread___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":280
 *     @property
 *     def bid_spread(self) -> Decimal:
 *         return self._bid_spread             # <<<<<<<<<<<<<<
 * 
 *     @bid_spread.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_bid_spread);
  __pyx_r = __pyx_v_self->_bid_spread;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":278
 *         self._hanging_orders_tracker._hanging_orders_cancel_pct = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def bid_spread(self) -> Decimal:
 *         return self._bid_spread
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":282
 *         return self._bid_spread
 * 
 *     @bid_spread.setter             # <<<<<<<<<<<<<<
 *     def bid_spread(self, value: Decimal):
 *         self._bid_spread = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10bid_spread_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10bid_spread_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10bid_spread_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10bid_spread_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":284
 *     @bid_spread.setter
 *     def bid_spread(self, value: Decimal):
 *         self._bid_spread = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_bid_spread);
  __Pyx_DECREF(__pyx_v_self->_bid_spread);
  __pyx_v_self->_bid_spread = __pyx_v_value;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":282
 *         return self._bid_spread
 * 
 *     @bid_spread.setter             # <<<<<<<<<<<<<<
 *     def bid_spread(self, value: Decimal):
 *         self._bid_spread = value
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":286
 *         self._bid_spread = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ask_spread(self) -> Decimal:
 *         return self._ask_spread
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10ask_spread_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10ask_spread_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10ask_spread___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10ask_spread___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":288
 *     @property
 *     def ask_spread(self) -> Decimal:
 *         return self._ask_spread             # <<<<<<<<<<<<<<
 * 
 *     @ask_spread.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_ask_spread);
  __pyx_r = __pyx_v_self->_ask_spread;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":286
 *         self._bid_spread = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ask_spread(self) -> Decimal:
 *         return self._ask_spread
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":290
 *         return self._ask_spread
 * 
 *     @ask_spread.setter             # <<<<<<<<<<<<<<
 *     def ask_spread(self, value: Decimal):
 *         self._ask_spread = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10ask_spread_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10ask_spread_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10ask_spread_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10ask_spread_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":292
 *     @ask_spread.setter
 *     def ask_spread(self, value: Decimal):
 *         self._ask_spread = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_ask_spread);
  __Pyx_DECREF(__pyx_v_self->_ask_spread);
  __pyx_v_self->_ask_spread = __pyx_v_value;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":290
 *         return self._ask_spread
 * 
 *     @ask_spread.setter             # <<<<<<<<<<<<<<
 *     def ask_spread(self, value: Decimal):
 *         self._ask_spread = value
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":294
 *         self._ask_spread = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_optimization_enabled(self) -> bool:
 *         return self._order_optimization_enabled
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26order_optimization_enabled_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26order_optimization_enabled_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26order_optimization_enabled___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26order_optimization_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":296
 *     @property
 *     def order_optimization_enabled(self) -> bool:
 *         return self._order_optimization_enabled             # <<<<<<<<<<<<<<
 * 
 *     @order_optimization_enabled.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_order_optimization_enabled); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":294
 *         self._ask_spread = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_optimization_enabled(self) -> bool:
 *         return self._order_optimization_enabled
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.order_optimization_enabled.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":298
 *         return self._order_optimization_enabled
 * 
 *     @order_optimization_enabled.setter             # <<<<<<<<<<<<<<
 *     def order_optimization_enabled(self, value: bool):
 *         self._order_optimization_enabled = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26order_optimization_enabled_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26order_optimization_enabled_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26order_optimization_enabled_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26order_optimization_enabled_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":300
 *     @order_optimization_enabled.setter
 *     def order_optimization_enabled(self, value: bool):
 *         self._order_optimization_enabled = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 300, __pyx_L1_error)
  __pyx_v_self->_order_optimization_enabled = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":298
 *         return self._order_optimization_enabled
 * 
 *     @order_optimization_enabled.setter             # <<<<<<<<<<<<<<
 *     def order_optimization_enabled(self, value: bool):
 *         self._order_optimization_enabled = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.order_optimization_enabled.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":302
 *         self._order_optimization_enabled = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_refresh_time(self) -> float:
 *         return self._order_refresh_time
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_refresh_time_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_refresh_time_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_refresh_time___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_refresh_time___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":304
 *     @property
 *     def order_refresh_time(self) -> float:
 *         return self._order_refresh_time             # <<<<<<<<<<<<<<
 * 
 *     @order_refresh_time.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_order_refresh_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":302
 *         self._order_optimization_enabled = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_refresh_time(self) -> float:
 *         return self._order_refresh_time
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.order_refresh_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":306
 *         return self._order_refresh_time
 * 
 *     @order_refresh_time.setter             # <<<<<<<<<<<<<<
 *     def order_refresh_time(self, value: float):
 *         self._order_refresh_time = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_refresh_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_refresh_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_refresh_time_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_refresh_time_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":308
 *     @order_refresh_time.setter
 *     def order_refresh_time(self, value: float):
 *         self._order_refresh_time = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 308, __pyx_L1_error)
  __pyx_v_self->_order_refresh_time = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":306
 *         return self._order_refresh_time
 * 
 *     @order_refresh_time.setter             # <<<<<<<<<<<<<<
 *     def order_refresh_time(self, value: float):
 *         self._order_refresh_time = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.order_refresh_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":310
 *         self._order_refresh_time = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def filled_order_delay(self) -> float:
 *         return self._filled_order_delay
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18filled_order_delay_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18filled_order_delay_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18filled_order_delay___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18filled_order_delay___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":312
 *     @property
 *     def filled_order_delay(self) -> float:
 *         return self._filled_order_delay             # <<<<<<<<<<<<<<
 * 
 *     @filled_order_delay.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_filled_order_delay); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":310
 *         self._order_refresh_time = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def filled_order_delay(self) -> float:
 *         return self._filled_order_delay
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.filled_order_delay.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":314
 *         return self._filled_order_delay
 * 
 *     @filled_order_delay.setter             # <<<<<<<<<<<<<<
 *     def filled_order_delay(self, value: float):
 *         self._filled_order_delay = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18filled_order_delay_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18filled_order_delay_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18filled_order_delay_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18filled_order_delay_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":316
 *     @filled_order_delay.setter
 *     def filled_order_delay(self, value: float):
 *         self._filled_order_delay = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 316, __pyx_L1_error)
  __pyx_v_self->_filled_order_delay = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":314
 *         return self._filled_order_delay
 * 
 *     @filled_order_delay.setter             # <<<<<<<<<<<<<<
 *     def filled_order_delay(self, value: float):
 *         self._filled_order_delay = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.filled_order_delay.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":318
 *         self._filled_order_delay = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def add_transaction_costs_to_orders(self) -> bool:
 *         return self._add_transaction_costs_to_orders
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31add_transaction_costs_to_orders_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31add_transaction_costs_to_orders_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31add_transaction_costs_to_orders___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31add_transaction_costs_to_orders___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":320
 *     @property
 *     def add_transaction_costs_to_orders(self) -> bool:
 *         return self._add_transaction_costs_to_orders             # <<<<<<<<<<<<<<
 * 
 *     @add_transaction_costs_to_orders.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_add_transaction_costs_to_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":318
 *         self._filled_order_delay = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def add_transaction_costs_to_orders(self) -> bool:
 *         return self._add_transaction_costs_to_orders
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.add_transaction_costs_to_orders.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":322
 *         return self._add_transaction_costs_to_orders
 * 
 *     @add_transaction_costs_to_orders.setter             # <<<<<<<<<<<<<<
 *     def add_transaction_costs_to_orders(self, value: bool):
 *         self._add_transaction_costs_to_orders = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31add_transaction_costs_to_orders_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31add_transaction_costs_to_orders_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31add_transaction_costs_to_orders_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31add_transaction_costs_to_orders_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":324
 *     @add_transaction_costs_to_orders.setter
 *     def add_transaction_costs_to_orders(self, value: bool):
 *         self._add_transaction_costs_to_orders = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 324, __pyx_L1_error)
  __pyx_v_self->_add_transaction_costs_to_orders = __pyx_t_1;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":322
 *         return self._add_transaction_costs_to_orders
 * 
 *     @add_transaction_costs_to_orders.setter             # <<<<<<<<<<<<<<
 *     def add_transaction_costs_to_orders(self, value: bool):
 *         self._add_transaction_costs_to_orders = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.add_transaction_costs_to_orders.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":326
 *         self._add_transaction_costs_to_orders = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def price_ceiling(self) -> Decimal:
 *         return self._price_ceiling
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13price_ceiling_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13price_ceiling_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13price_ceiling___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13price_ceiling___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":328
 *     @property
 *     def price_ceiling(self) -> Decimal:
 *         return self._price_ceiling             # <<<<<<<<<<<<<<
 * 
 *     @price_ceiling.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_price_ceiling);
  __pyx_r = __pyx_v_self->_price_ceiling;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":326
 *         self._add_transaction_costs_to_orders = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def price_ceiling(self) -> Decimal:
 *         return self._price_ceiling
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":330
 *         return self._price_ceiling
 * 
 *     @price_ceiling.setter             # <<<<<<<<<<<<<<
 *     def price_ceiling(self, value: Decimal):
 *         self._price_ceiling = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13price_ceiling_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13price_ceiling_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13price_ceiling_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13price_ceiling_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":332
 *     @price_ceiling.setter
 *     def price_ceiling(self, value: Decimal):
 *         self._price_ceiling = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_price_ceiling);
  __Pyx_DECREF(__pyx_v_self->_price_ceiling);
  __pyx_v_self->_price_ceiling = __pyx_v_value;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":330
 *         return self._price_ceiling
 * 
 *     @price_ceiling.setter             # <<<<<<<<<<<<<<
 *     def price_ceiling(self, value: Decimal):
 *         self._price_ceiling = value
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":334
 *         self._price_ceiling = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def price_floor(self) -> Decimal:
 *         return self._price_floor
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11price_floor_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11price_floor_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11price_floor___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11price_floor___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":336
 *     @property
 *     def price_floor(self) -> Decimal:
 *         return self._price_floor             # <<<<<<<<<<<<<<
 * 
 *     @price_floor.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_price_floor);
  __pyx_r = __pyx_v_self->_price_floor;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":334
 *         self._price_ceiling = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def price_floor(self) -> Decimal:
 *         return self._price_floor
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":338
 *         return self._price_floor
 * 
 *     @price_floor.setter             # <<<<<<<<<<<<<<
 *     def price_floor(self, value: Decimal):
 *         self._price_floor = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11price_floor_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11price_floor_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11price_floor_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11price_floor_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":340
 *     @price_floor.setter
 *     def price_floor(self, value: Decimal):
 *         self._price_floor = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_price_floor);
  __Pyx_DECREF(__pyx_v_self->_price_floor);
  __pyx_v_self->_price_floor = __pyx_v_value;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":338
 *         return self._price_floor
 * 
 *     @price_floor.setter             # <<<<<<<<<<<<<<
 *     def price_floor(self, value: Decimal):
 *         self._price_floor = value
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":342
 *         self._price_floor = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def base_asset(self):
 *         return self._market_info.base_asset
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10base_asset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10base_asset_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10base_asset___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10base_asset___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":344
 *     @property
 *     def base_asset(self):
 *         return self._market_info.base_asset             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":342
 *         self._price_floor = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def base_asset(self):
 *         return self._market_info.base_asset
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.base_asset.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":346
 *         return self._market_info.base_asset
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def quote_asset(self):
 *         return self._market_info.quote_asset
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11quote_asset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11quote_asset_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11quote_asset___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11quote_asset___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":348
 *     @property
 *     def quote_asset(self):
 *         return self._market_info.quote_asset             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_quote_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":346
 *         return self._market_info.base_asset
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def quote_asset(self):
 *         return self._market_info.quote_asset
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.quote_asset.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":350
 *         return self._market_info.quote_asset
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def trading_pair(self):
 *         return self._market_info.trading_pair
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12trading_pair_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12trading_pair_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12trading_pair___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12trading_pair___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":352
 *     @property
 *     def trading_pair(self):
 *         return self._market_info.trading_pair             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":350
 *         return self._market_info.quote_asset
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def trading_pair(self):
 *         return self._market_info.trading_pair
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.trading_pair.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":354
 *         return self._market_info.trading_pair
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_override(self):
 *         return self._order_override
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14order_override_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14order_override_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14order_override___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14order_override___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":356
 *     @property
 *     def order_override(self):
 *         return self._order_override             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_order_override);
  __pyx_r = __pyx_v_self->_order_override;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":354
 *         return self._market_info.trading_pair
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_override(self):
 *         return self._order_override
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":370
 *         return self._ask_order_level_spreads
 * 
 *     @order_override.setter             # <<<<<<<<<<<<<<
 *     def order_override(self, value: Dict[str, List[str]]):
 *         self._order_override = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14order_override_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14order_override_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14order_override_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14order_override_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":372
 *     @order_override.setter
 *     def order_override(self, value: Dict[str, List[str]]):
 *         self._order_override = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_order_override);
  __Pyx_DECREF(__pyx_v_self->_order_override);
  __pyx_v_self->_order_override = __pyx_v_value;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":370
 *         return self._ask_order_level_spreads
 * 
 *     @order_override.setter             # <<<<<<<<<<<<<<
 *     def order_override(self, value: Dict[str, List[str]]):
 *         self._order_override = value
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":358
 *         return self._order_override
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def split_order_levels_enabled(self):
 *         return self._split_order_levels_enabled
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26split_order_levels_enabled_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26split_order_levels_enabled_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26split_order_levels_enabled___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26split_order_levels_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":360
 *     @property
 *     def split_order_levels_enabled(self):
 *         return self._split_order_levels_enabled             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_split_order_levels_enabled);
  __pyx_r = __pyx_v_self->_split_order_levels_enabled;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":358
 *         return self._order_override
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def split_order_levels_enabled(self):
 *         return self._split_order_levels_enabled
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":362
 *         return self._split_order_levels_enabled
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def bid_order_level_spreads(self):
 *         return self._bid_order_level_spreads
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23bid_order_level_spreads_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23bid_order_level_spreads_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23bid_order_level_spreads___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23bid_order_level_spreads___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":364
 *     @property
 *     def bid_order_level_spreads(self):
 *         return self._bid_order_level_spreads             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_bid_order_level_spreads);
  __pyx_r = __pyx_v_self->_bid_order_level_spreads;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":362
 *         return self._split_order_levels_enabled
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def bid_order_level_spreads(self):
 *         return self._bid_order_level_spreads
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":366
 *         return self._bid_order_level_spreads
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ask_order_level_spreads(self):
 *         return self._ask_order_level_spreads
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23ask_order_level_spreads_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23ask_order_level_spreads_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23ask_order_level_spreads___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23ask_order_level_spreads___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":368
 *     @property
 *     def ask_order_level_spreads(self):
 *         return self._ask_order_level_spreads             # <<<<<<<<<<<<<<
 * 
 *     @order_override.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_ask_order_level_spreads);
  __pyx_r = __pyx_v_self->_ask_order_level_spreads;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":366
 *         return self._bid_order_level_spreads
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ask_order_level_spreads(self):
 *         return self._ask_order_level_spreads
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":374
 *         self._order_override = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def moving_price_band_enabled(self) -> bool:
 *         return self._moving_price_band.enabled
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25moving_price_band_enabled_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25moving_price_band_enabled_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25moving_price_band_enabled___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25moving_price_band_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":376
 *     @property
 *     def moving_price_band_enabled(self) -> bool:
 *         return self._moving_price_band.enabled             # <<<<<<<<<<<<<<
 * 
 *     @moving_price_band_enabled.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_moving_price_band, __pyx_mstate_global->__pyx_n_u_enabled); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":374
 *         self._order_override = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def moving_price_band_enabled(self) -> bool:
 *         return self._moving_price_band.enabled
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.moving_price_band_enabled.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":378
 *         return self._moving_price_band.enabled
 * 
 *     @moving_price_band_enabled.setter             # <<<<<<<<<<<<<<
 *     def moving_price_band_enabled(self, value: bool):
 *         self._moving_price_band.switch(value)
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25moving_price_band_enabled_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25moving_price_band_enabled_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25moving_price_band_enabled_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25moving_price_band_enabled_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":380
 *     @moving_price_band_enabled.setter
 *     def moving_price_band_enabled(self, value: bool):
 *         self._moving_price_band.switch(value)             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_2 = __pyx_v_self->_moving_price_band;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_switch, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":378
 *         return self._moving_price_band.enabled
 * 
 *     @moving_price_band_enabled.setter             # <<<<<<<<<<<<<<
 *     def moving_price_band_enabled(self, value: bool):
 *         self._moving_price_band.switch(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.moving_price_band_enabled.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":382
 *         self._moving_price_band.switch(value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def price_ceiling_pct(self) -> Decimal:
 *         return self._moving_price_band.price_ceiling_pct
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17price_ceiling_pct_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17price_ceiling_pct_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17price_ceiling_pct___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17price_ceiling_pct___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":384
 *     @property
 *     def price_ceiling_pct(self) -> Decimal:
 *         return self._moving_price_band.price_ceiling_pct             # <<<<<<<<<<<<<<
 * 
 *     @price_ceiling_pct.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_moving_price_band, __pyx_mstate_global->__pyx_n_u_price_ceiling_pct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":382
 *         self._moving_price_band.switch(value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def price_ceiling_pct(self) -> Decimal:
 *         return self._moving_price_band.price_ceiling_pct
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.price_ceiling_pct.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":386
 *         return self._moving_price_band.price_ceiling_pct
 * 
 *     @price_ceiling_pct.setter             # <<<<<<<<<<<<<<
 *     def price_ceiling_pct(self, value: Decimal):
 *         self._moving_price_band.price_ceiling_pct = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17price_ceiling_pct_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17price_ceiling_pct_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17price_ceiling_pct_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17price_ceiling_pct_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":388
 *     @price_ceiling_pct.setter
 *     def price_ceiling_pct(self, value: Decimal):
 *         self._moving_price_band.price_ceiling_pct = value             # <<<<<<<<<<<<<<
 *         self._moving_price_band.update(self._current_timestamp, self.get_price())
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_moving_price_band, __pyx_mstate_global->__pyx_n_u_price_ceiling_pct, __pyx_v_value) < (0)) __PYX_ERR(0, 388, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":389
 *     def price_ceiling_pct(self, value: Decimal):
 *         self._moving_price_band.price_ceiling_pct = value
 *         self._moving_price_band.update(self._current_timestamp, self.get_price())             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_2 = __pyx_v_self->_moving_price_band;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_update, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":386
 *         return self._moving_price_band.price_ceiling_pct
 * 
 *     @price_ceiling_pct.setter             # <<<<<<<<<<<<<<
 *     def price_ceiling_pct(self, value: Decimal):
 *         self._moving_price_band.price_ceiling_pct = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.price_ceiling_pct.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":391
 *         self._moving_price_band.update(self._current_timestamp, self.get_price())
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def price_floor_pct(self) -> Decimal:
 *         return self._moving_price_band.price_floor_pct
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15price_floor_pct_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15price_floor_pct_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15price_floor_pct___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15price_floor_pct___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":393
 *     @property
 *     def price_floor_pct(self) -> Decimal:
 *         return self._moving_price_band.price_floor_pct             # <<<<<<<<<<<<<<
 * 
 *     @price_floor_pct.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_moving_price_band, __pyx_mstate_global->__pyx_n_u_price_floor_pct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":391
 *         self._moving_price_band.update(self._current_timestamp, self.get_price())
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def price_floor_pct(self) -> Decimal:
 *         return self._moving_price_band.price_floor_pct
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.price_floor_pct.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":395
 *         return self._moving_price_band.price_floor_pct
 * 
 *     @price_floor_pct.setter             # <<<<<<<<<<<<<<
 *     def price_floor_pct(self, value: Decimal):
 *         self._moving_price_band.price_floor_pct = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15price_floor_pct_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15price_floor_pct_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15price_floor_pct_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15price_floor_pct_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":397
 *     @price_floor_pct.setter
 *     def price_floor_pct(self, value: Decimal):
 *         self._moving_price_band.price_floor_pct = value             # <<<<<<<<<<<<<<
 *         self._moving_price_band.update(self._current_timestamp, self.get_price())
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_moving_price_band, __pyx_mstate_global->__pyx_n_u_price_floor_pct, __pyx_v_value) < (0)) __PYX_ERR(0, 397, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":398
 *     def price_floor_pct(self, value: Decimal):
 *         self._moving_price_band.price_floor_pct = value
 *         self._moving_price_band.update(self._current_timestamp, self.get_price())             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_2 = __pyx_v_self->_moving_price_band;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_update, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":395
 *         return self._moving_price_band.price_floor_pct
 * 
 *     @price_floor_pct.setter             # <<<<<<<<<<<<<<
 *     def price_floor_pct(self, value: Decimal):
 *         self._moving_price_band.price_floor_pct = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.price_floor_pct.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":400
 *         self._moving_price_band.update(self._current_timestamp, self.get_price())
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def price_band_refresh_time(self) -> float:
 *         return self._moving_price_band.price_band_refresh_time
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23price_band_refresh_time_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23price_band_refresh_time_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23price_band_refresh_time___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23price_band_refresh_time___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":402
 *     @property
 *     def price_band_refresh_time(self) -> float:
 *         return self._moving_price_band.price_band_refresh_time             # <<<<<<<<<<<<<<
 * 
 *     @price_band_refresh_time.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_moving_price_band, __pyx_mstate_global->__pyx_n_u_price_band_refresh_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":400
 *         self._moving_price_band.update(self._current_timestamp, self.get_price())
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def price_band_refresh_time(self) -> float:
 *         return self._moving_price_band.price_band_refresh_time
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.price_band_refresh_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":404
 *         return self._moving_price_band.price_band_refresh_time
 * 
 *     @price_band_refresh_time.setter             # <<<<<<<<<<<<<<
 *     def price_band_refresh_time(self, value: Decimal):
 *         self._moving_price_band.price_band_refresh_time = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23price_band_refresh_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23price_band_refresh_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23price_band_refresh_time_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23price_band_refresh_time_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":406
 *     @price_band_refresh_time.setter
 *     def price_band_refresh_time(self, value: Decimal):
 *         self._moving_price_band.price_band_refresh_time = value             # <<<<<<<<<<<<<<
 *         self._moving_price_band.update(self._current_timestamp, self.get_price())
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_moving_price_band, __pyx_mstate_global->__pyx_n_u_price_band_refresh_time, __pyx_v_value) < (0)) __PYX_ERR(0, 406, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":407
 *     def price_band_refresh_time(self, value: Decimal):
 *         self._moving_price_band.price_band_refresh_time = value
 *         self._moving_price_band.update(self._current_timestamp, self.get_price())             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_2 = __pyx_v_self->_moving_price_band;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_update, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":404
 *         return self._moving_price_band.price_band_refresh_time
 * 
 *     @price_band_refresh_time.setter             # <<<<<<<<<<<<<<
 *     def price_band_refresh_time(self, value: Decimal):
 *         self._moving_price_band.price_band_refresh_time = value
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.price_band_refresh_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":409
 *         self._moving_price_band.update(self._current_timestamp, self.get_price())
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def moving_price_band(self) -> MovingPriceBand:
 *         return self._moving_price_band
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17moving_price_band_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17moving_price_band_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17moving_price_band___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17moving_price_band___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":411
 *     @property
 *     def moving_price_band(self) -> MovingPriceBand:
 *         return self._moving_price_band             # <<<<<<<<<<<<<<
 * 
 *     def get_price(self) -> Decimal:
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_moving_price_band);
  __pyx_r = __pyx_v_self->_moving_price_band;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":409
 *         self._moving_price_band.update(self._current_timestamp, self.get_price())
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def moving_price_band(self) -> MovingPriceBand:
 *         return self._moving_price_band
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":413
 *         return self._moving_price_band
 * 
 *     def get_price(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         price_provider = self._asset_price_delegate or self._market_info
 *         if self._price_type is PriceType.LastOwnTrade:
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7get_price(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7get_price = {"get_price", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7get_price, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7get_price(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_price (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("get_price", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("get_price", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_6get_price(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_6get_price(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_price_provider = NULL;
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_price", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":414
 * 
 *     def get_price(self) -> Decimal:
 *         price_provider = self._asset_price_delegate or self._market_info             # <<<<<<<<<<<<<<
 *         if self._price_type is PriceType.LastOwnTrade:
 *             price = self._last_own_trade_price
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->_asset_price_delegate); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 414, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_self->_asset_price_delegate);
    __pyx_t_1 = __pyx_v_self->_asset_price_delegate;
    goto __pyx_L3_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_v_self->_market_info);
  __pyx_t_1 = __pyx_v_self->_market_info;
  __pyx_L3_bool_binop_done:;
  __pyx_v_price_provider = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":415
 *     def get_price(self) -> Decimal:
 *         price_provider = self._asset_price_delegate or self._market_info
 *         if self._price_type is PriceType.LastOwnTrade:             # <<<<<<<<<<<<<<
 *             price = self._last_own_trade_price
 *         elif self._price_type is PriceType.InventoryCost:
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_PriceType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_LastOwnTrade); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_v_self->_price_type == __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":416
 *         price_provider = self._asset_price_delegate or self._market_info
 *         if self._price_type is PriceType.LastOwnTrade:
 *             price = self._last_own_trade_price             # <<<<<<<<<<<<<<
 *         elif self._price_type is PriceType.InventoryCost:
 *             price = price_provider.get_price_by_type(PriceType.MidPrice)
*/
    __pyx_t_3 = __pyx_v_self->_last_own_trade_price;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_price = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":415
 *     def get_price(self) -> Decimal:
 *         price_provider = self._asset_price_delegate or self._market_info
 *         if self._price_type is PriceType.LastOwnTrade:             # <<<<<<<<<<<<<<
 *             price = self._last_own_trade_price
 *         elif self._price_type is PriceType.InventoryCost:
*/
    goto __pyx_L5;
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":417
 *         if self._price_type is PriceType.LastOwnTrade:
 *             price = self._last_own_trade_price
 *         elif self._price_type is PriceType.InventoryCost:             # <<<<<<<<<<<<<<
 *             price = price_provider.get_price_by_type(PriceType.MidPrice)
 *         else:
*/
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PriceType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_InventoryCost); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = (__pyx_v_self->_price_type == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":418
 *             price = self._last_own_trade_price
 *         elif self._price_type is PriceType.InventoryCost:
 *             price = price_provider.get_price_by_type(PriceType.MidPrice)             # <<<<<<<<<<<<<<
 *         else:
 *             price = price_provider.get_price_by_type(self._price_type)
*/
    __pyx_t_3 = __pyx_v_price_provider;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PriceType); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_MidPrice); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price_by_type, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_v_price = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":417
 *         if self._price_type is PriceType.LastOwnTrade:
 *             price = self._last_own_trade_price
 *         elif self._price_type is PriceType.InventoryCost:             # <<<<<<<<<<<<<<
 *             price = price_provider.get_price_by_type(PriceType.MidPrice)
 *         else:
*/
    goto __pyx_L5;
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":420
 *             price = price_provider.get_price_by_type(PriceType.MidPrice)
 *         else:
 *             price = price_provider.get_price_by_type(self._price_type)             # <<<<<<<<<<<<<<
 * 
 *         if price.is_nan():
*/
  /*else*/ {
    __pyx_t_5 = __pyx_v_price_provider;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_self->_price_type};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price_by_type, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_v_price = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L5:;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":422
 *             price = price_provider.get_price_by_type(self._price_type)
 * 
 *         if price.is_nan():             # <<<<<<<<<<<<<<
 *             price = price_provider.get_price_by_type(PriceType.MidPrice)
 * 
*/
  __pyx_t_5 = __pyx_v_price;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_nan, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":423
 * 
 *         if price.is_nan():
 *             price = price_provider.get_price_by_type(PriceType.MidPrice)             # <<<<<<<<<<<<<<
 * 
 *         return price
*/
    __pyx_t_5 = __pyx_v_price_provider;
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PriceType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_MidPrice); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price_by_type, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF_SET(__pyx_v_price, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":422
 *             price = price_provider.get_price_by_type(self._price_type)
 * 
 *         if price.is_nan():             # <<<<<<<<<<<<<<
 *             price = price_provider.get_price_by_type(PriceType.MidPrice)
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":425
 *             price = price_provider.get_price_by_type(PriceType.MidPrice)
 * 
 *         return price             # <<<<<<<<<<<<<<
 * 
 *     def get_mid_price(self) -> Decimal:
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_price);
  __pyx_r = __pyx_v_price;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":413
 *         return self._moving_price_band
 * 
 *     def get_price(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         price_provider = self._asset_price_delegate or self._market_info
 *         if self._price_type is PriceType.LastOwnTrade:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.get_price", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_price_provider);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":427
 *         return price
 * 
 *     def get_mid_price(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_mid_price()
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_9get_mid_price(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_9get_mid_price = {"get_mid_price", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_9get_mid_price, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_9get_mid_price(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_mid_price (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("get_mid_price", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("get_mid_price", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_8get_mid_price(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_8get_mid_price(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_mid_price", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":428
 * 
 *     def get_mid_price(self) -> Decimal:
 *         return self.c_get_mid_price()             # <<<<<<<<<<<<<<
 * 
 *     cdef object c_get_mid_price(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_mid_price(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":427
 *         return price
 * 
 *     def get_mid_price(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_mid_price()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.get_mid_price", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":430
 *         return self.c_get_mid_price()
 * 
 *     cdef object c_get_mid_price(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             AssetPriceDelegate delegate = self._asset_price_delegate
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_get_mid_price(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate *__pyx_v_delegate = 0;
  PyObject *__pyx_v_mid_price = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_get_mid_price", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":432
 *     cdef object c_get_mid_price(self):
 *         cdef:
 *             AssetPriceDelegate delegate = self._asset_price_delegate             # <<<<<<<<<<<<<<
 *             object mid_price
 *         if self._asset_price_delegate is not None:
*/
  __pyx_t_1 = __pyx_v_self->_asset_price_delegate;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate))))) __PYX_ERR(0, 432, __pyx_L1_error)
  __pyx_v_delegate = ((struct __pyx_obj_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":434
 *             AssetPriceDelegate delegate = self._asset_price_delegate
 *             object mid_price
 *         if self._asset_price_delegate is not None:             # <<<<<<<<<<<<<<
 *             mid_price = delegate.c_get_mid_price()
 *         else:
*/
  __pyx_t_2 = (__pyx_v_self->_asset_price_delegate != Py_None);
  if (__pyx_t_2) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":435
 *             object mid_price
 *         if self._asset_price_delegate is not None:
 *             mid_price = delegate.c_get_mid_price()             # <<<<<<<<<<<<<<
 *         else:
 *             mid_price = self._market_info.get_mid_price()
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate *)__pyx_v_delegate->__pyx_vtab)->c_get_mid_price(__pyx_v_delegate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_mid_price = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":434
 *             AssetPriceDelegate delegate = self._asset_price_delegate
 *             object mid_price
 *         if self._asset_price_delegate is not None:             # <<<<<<<<<<<<<<
 *             mid_price = delegate.c_get_mid_price()
 *         else:
*/
    goto __pyx_L3;
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":437
 *             mid_price = delegate.c_get_mid_price()
 *         else:
 *             mid_price = self._market_info.get_mid_price()             # <<<<<<<<<<<<<<
 *         return mid_price
 * 
*/
  /*else*/ {
    __pyx_t_3 = __pyx_v_self->_market_info;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_mid_price, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_v_mid_price = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":438
 *         else:
 *             mid_price = self._market_info.get_mid_price()
 *         return mid_price             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_mid_price);
  __pyx_r = __pyx_v_mid_price;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":430
 *         return self.c_get_mid_price()
 * 
 *     cdef object c_get_mid_price(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             AssetPriceDelegate delegate = self._asset_price_delegate
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_get_mid_price", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_delegate);
  __Pyx_XDECREF(__pyx_v_mid_price);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":440
 *         return mid_price
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def hanging_order_ids(self) -> List[str]:
 *         return [o.order_id for o in self._hanging_orders_tracker.strategy_current_hanging_orders]
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17hanging_order_ids_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17hanging_order_ids_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17hanging_order_ids___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17hanging_order_ids___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_8genexpr1__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":442
 *     @property
 *     def hanging_order_ids(self) -> List[str]:
 *         return [o.order_id for o in self._hanging_orders_tracker.strategy_current_hanging_orders]             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_mstate_global->__pyx_n_u_strategy_current_hanging_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 442, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 442, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 442, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 442, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_4, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_4;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 442, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4));
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4);
          #endif
          ++__pyx_t_4;
        }
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 442, __pyx_L5_error)
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 442, __pyx_L5_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_o, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr1__pyx_v_o, __pyx_mstate_global->__pyx_n_u_order_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 442, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 442, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_o); __pyx_8genexpr1__pyx_v_o = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_o); __pyx_8genexpr1__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":440
 *         return mid_price
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def hanging_order_ids(self) -> List[str]:
 *         return [o.order_id for o in self._hanging_orders_tracker.strategy_current_hanging_orders]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.hanging_order_ids.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":444
 *         return [o.order_id for o in self._hanging_orders_tracker.strategy_current_hanging_orders]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def market_info_to_active_orders(self) -> Dict[MarketTradingPairTuple, List[LimitOrder]]:
 *         return self._sb_order_tracker.market_pair_to_active_orders
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28market_info_to_active_orders_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28market_info_to_active_orders_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28market_info_to_active_orders___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28market_info_to_active_orders___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":446
 *     @property
 *     def market_info_to_active_orders(self) -> Dict[MarketTradingPairTuple, List[LimitOrder]]:
 *         return self._sb_order_tracker.market_pair_to_active_orders             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base._sb_order_tracker), __pyx_mstate_global->__pyx_n_u_market_pair_to_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":444
 *         return [o.order_id for o in self._hanging_orders_tracker.strategy_current_hanging_orders]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def market_info_to_active_orders(self) -> Dict[MarketTradingPairTuple, List[LimitOrder]]:
 *         return self._sb_order_tracker.market_pair_to_active_orders
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.market_info_to_active_orders.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":448
 *         return self._sb_order_tracker.market_pair_to_active_orders
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_orders(self) -> List[LimitOrder]:
 *         if self._market_info not in self.market_info_to_active_orders:
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13active_orders_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13active_orders_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13active_orders___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13active_orders___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":450
 *     @property
 *     def active_orders(self) -> List[LimitOrder]:
 *         if self._market_info not in self.market_info_to_active_orders:             # <<<<<<<<<<<<<<
 *             return []
 *         return self.market_info_to_active_orders[self._market_info]
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_market_info_to_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_self->_market_info, __pyx_t_1, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":451
 *     def active_orders(self) -> List[LimitOrder]:
 *         if self._market_info not in self.market_info_to_active_orders:
 *             return []             # <<<<<<<<<<<<<<
 *         return self.market_info_to_active_orders[self._market_info]
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":450
 *     @property
 *     def active_orders(self) -> List[LimitOrder]:
 *         if self._market_info not in self.market_info_to_active_orders:             # <<<<<<<<<<<<<<
 *             return []
 *         return self.market_info_to_active_orders[self._market_info]
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":452
 *         if self._market_info not in self.market_info_to_active_orders:
 *             return []
 *         return self.market_info_to_active_orders[self._market_info]             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_market_info_to_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_self->_market_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":448
 *         return self._sb_order_tracker.market_pair_to_active_orders
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_orders(self) -> List[LimitOrder]:
 *         if self._market_info not in self.market_info_to_active_orders:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.active_orders.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":454
 *         return self.market_info_to_active_orders[self._market_info]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_buys(self) -> List[LimitOrder]:
 *         return [o for o in self.active_orders if o.is_buy]
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11active_buys_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11active_buys_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11active_buys___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11active_buys___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_8genexpr2__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":456
 *     @property
 *     def active_buys(self) -> List[LimitOrder]:
 *         return [o for o in self.active_orders if o.is_buy]             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 456, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 456, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 456, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 456, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_4, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_4;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 456, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4));
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4);
          #endif
          ++__pyx_t_4;
        }
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 456, __pyx_L5_error)
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 456, __pyx_L5_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_o, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr2__pyx_v_o, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 456, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 456, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr2__pyx_v_o))) __PYX_ERR(0, 456, __pyx_L5_error)
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_o); __pyx_8genexpr2__pyx_v_o = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_o); __pyx_8genexpr2__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":454
 *         return self.market_info_to_active_orders[self._market_info]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_buys(self) -> List[LimitOrder]:
 *         return [o for o in self.active_orders if o.is_buy]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.active_buys.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":458
 *         return [o for o in self.active_orders if o.is_buy]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_sells(self) -> List[LimitOrder]:
 *         return [o for o in self.active_orders if not o.is_buy]
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12active_sells_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12active_sells_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12active_sells___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12active_sells___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_8genexpr3__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":460
 *     @property
 *     def active_sells(self) -> List[LimitOrder]:
 *         return [o for o in self.active_orders if not o.is_buy]             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 460, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 460, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 460, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 460, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 460, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_4, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_4;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 460, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4));
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4);
          #endif
          ++__pyx_t_4;
        }
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 460, __pyx_L5_error)
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 460, __pyx_L5_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_o, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr3__pyx_v_o, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 460, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 460, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_7 = (!__pyx_t_6);
      if (__pyx_t_7) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr3__pyx_v_o))) __PYX_ERR(0, 460, __pyx_L5_error)
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_o); __pyx_8genexpr3__pyx_v_o = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_o); __pyx_8genexpr3__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":458
 *         return [o for o in self.active_orders if o.is_buy]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_sells(self) -> List[LimitOrder]:
 *         return [o for o in self.active_orders if not o.is_buy]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.active_sells.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":462
 *         return [o for o in self.active_orders if not o.is_buy]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_non_hanging_orders(self) -> List[LimitOrder]:
 *         orders = [o for o in self.active_orders if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)]
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25active_non_hanging_orders_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25active_non_hanging_orders_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25active_non_hanging_orders___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25active_non_hanging_orders___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_orders = NULL;
  PyObject *__pyx_8genexpr4__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":464
 *     @property
 *     def active_non_hanging_orders(self) -> List[LimitOrder]:
 *         orders = [o for o in self.active_orders if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)]             # <<<<<<<<<<<<<<
 *         return orders
 * 
*/
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 464, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 464, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 464, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 464, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_4, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_4;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 464, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4));
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4);
          #endif
          ++__pyx_t_4;
        }
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 464, __pyx_L5_error)
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 464, __pyx_L5_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_8genexpr4__pyx_v_o, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_6 = __pyx_v_self->_hanging_orders_tracker;
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr4__pyx_v_o, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 464, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_7};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_order_id_in_hanging_orders, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 464, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 464, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_10 = (!__pyx_t_9);
      if (__pyx_t_10) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr4__pyx_v_o))) __PYX_ERR(0, 464, __pyx_L5_error)
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_o); __pyx_8genexpr4__pyx_v_o = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_o); __pyx_8genexpr4__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_v_orders = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":465
 *     def active_non_hanging_orders(self) -> List[LimitOrder]:
 *         orders = [o for o in self.active_orders if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)]
 *         return orders             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_orders);
  __pyx_r = __pyx_v_orders;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":462
 *         return [o for o in self.active_orders if not o.is_buy]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_non_hanging_orders(self) -> List[LimitOrder]:
 *         orders = [o for o in self.active_orders if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.active_non_hanging_orders.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_orders);
  __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":467
 *         return orders
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def logging_options(self) -> int:
 *         return self._logging_options
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15logging_options_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15logging_options_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15logging_options___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15logging_options___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":469
 *     @property
 *     def logging_options(self) -> int:
 *         return self._logging_options             # <<<<<<<<<<<<<<
 * 
 *     @logging_options.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->_logging_options); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":467
 *         return orders
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def logging_options(self) -> int:
 *         return self._logging_options
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.logging_options.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":471
 *         return self._logging_options
 * 
 *     @logging_options.setter             # <<<<<<<<<<<<<<
 *     def logging_options(self, int64_t logging_options):
 *         self._logging_options = logging_options
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15logging_options_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_logging_options); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15logging_options_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_logging_options) {
  int64_t __pyx_v_logging_options;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_logging_options); {
    __pyx_v_logging_options = __Pyx_PyLong_As_int64_t(__pyx_arg_logging_options); if (unlikely((__pyx_v_logging_options == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 472, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.logging_options.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15logging_options_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((int64_t)__pyx_v_logging_options));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15logging_options_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, int64_t __pyx_v_logging_options) {
  int __pyx_r;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":473
 *     @logging_options.setter
 *     def logging_options(self, int64_t logging_options):
 *         self._logging_options = logging_options             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_v_self->_logging_options = __pyx_v_logging_options;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":471
 *         return self._logging_options
 * 
 *     @logging_options.setter             # <<<<<<<<<<<<<<
 *     def logging_options(self, int64_t logging_options):
 *         self._logging_options = logging_options
*/

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":475
 *         self._logging_options = logging_options
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def hanging_orders_tracker(self):
 *         return self._hanging_orders_tracker
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_tracker_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_tracker_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_tracker___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_tracker___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":477
 *     @property
 *     def hanging_orders_tracker(self):
 *         return self._hanging_orders_tracker             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_hanging_orders_tracker);
  __pyx_r = __pyx_v_self->_hanging_orders_tracker;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":475
 *         self._logging_options = logging_options
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def hanging_orders_tracker(self):
 *         return self._hanging_orders_tracker
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":479
 *         return self._hanging_orders_tracker
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def asset_price_delegate(self) -> AssetPriceDelegate:
 *         return self._asset_price_delegate
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20asset_price_delegate_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20asset_price_delegate_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20asset_price_delegate___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20asset_price_delegate___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":481
 *     @property
 *     def asset_price_delegate(self) -> AssetPriceDelegate:
 *         return self._asset_price_delegate             # <<<<<<<<<<<<<<
 * 
 *     @asset_price_delegate.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_asset_price_delegate);
  __pyx_r = __pyx_v_self->_asset_price_delegate;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":479
 *         return self._hanging_orders_tracker
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def asset_price_delegate(self) -> AssetPriceDelegate:
 *         return self._asset_price_delegate
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":483
 *         return self._asset_price_delegate
 * 
 *     @asset_price_delegate.setter             # <<<<<<<<<<<<<<
 *     def asset_price_delegate(self, value):
 *         self._asset_price_delegate = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20asset_price_delegate_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20asset_price_delegate_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20asset_price_delegate_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20asset_price_delegate_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":485
 *     @asset_price_delegate.setter
 *     def asset_price_delegate(self, value):
 *         self._asset_price_delegate = value             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_asset_price_delegate);
  __Pyx_DECREF(__pyx_v_self->_asset_price_delegate);
  __pyx_v_self->_asset_price_delegate = __pyx_v_value;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":483
 *         return self._asset_price_delegate
 * 
 *     @asset_price_delegate.setter             # <<<<<<<<<<<<<<
 *     def asset_price_delegate(self, value):
 *         self._asset_price_delegate = value
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":487
 *         self._asset_price_delegate = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inventory_cost_price_delegate(self) -> AssetPriceDelegate:
 *         return self._inventory_cost_price_delegate
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29inventory_cost_price_delegate_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29inventory_cost_price_delegate_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29inventory_cost_price_delegate___get__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29inventory_cost_price_delegate___get__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":489
 *     @property
 *     def inventory_cost_price_delegate(self) -> AssetPriceDelegate:
 *         return self._inventory_cost_price_delegate             # <<<<<<<<<<<<<<
 * 
 *     @inventory_cost_price_delegate.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_inventory_cost_price_delegate);
  __pyx_r = __pyx_v_self->_inventory_cost_price_delegate;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":487
 *         self._asset_price_delegate = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inventory_cost_price_delegate(self) -> AssetPriceDelegate:
 *         return self._inventory_cost_price_delegate
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":491
 *         return self._inventory_cost_price_delegate
 * 
 *     @inventory_cost_price_delegate.setter             # <<<<<<<<<<<<<<
 *     def inventory_cost_price_delegate(self, value):
 *         self._inventory_cost_price_delegate = value
*/

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29inventory_cost_price_delegate_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29inventory_cost_price_delegate_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29inventory_cost_price_delegate_2__set__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29inventory_cost_price_delegate_2__set__(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":493
 *     @inventory_cost_price_delegate.setter
 *     def inventory_cost_price_delegate(self, value):
 *         self._inventory_cost_price_delegate = value             # <<<<<<<<<<<<<<
 * 
 *     def inventory_skew_stats_data_frame(self) -> Optional[pd.DataFrame]:
*/
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_inventory_cost_price_delegate);
  __Pyx_DECREF(__pyx_v_self->_inventory_cost_price_delegate);
  __pyx_v_self->_inventory_cost_price_delegate = __pyx_v_value;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":491
 *         return self._inventory_cost_price_delegate
 * 
 *     @inventory_cost_price_delegate.setter             # <<<<<<<<<<<<<<
 *     def inventory_cost_price_delegate(self, value):
 *         self._inventory_cost_price_delegate = value
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":495
 *         self._inventory_cost_price_delegate = value
 * 
 *     def inventory_skew_stats_data_frame(self) -> Optional[pd.DataFrame]:             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11inventory_skew_stats_data_frame(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11inventory_skew_stats_data_frame = {"inventory_skew_stats_data_frame", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11inventory_skew_stats_data_frame, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11inventory_skew_stats_data_frame(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("inventory_skew_stats_data_frame (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("inventory_skew_stats_data_frame", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("inventory_skew_stats_data_frame", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10inventory_skew_stats_data_frame(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10inventory_skew_stats_data_frame(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  CYTHON_UNUSED struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_v_base_asset_amount = NULL;
  PyObject *__pyx_v_quote_asset_amount = NULL;
  PyObject *__pyx_v_total_order_size = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_base_asset_value = NULL;
  PyObject *__pyx_v_quote_asset_value = NULL;
  PyObject *__pyx_v_total_value = NULL;
  PyObject *__pyx_v_total_value_in_quote = NULL;
  PyObject *__pyx_v_base_asset_ratio = NULL;
  PyObject *__pyx_v_quote_asset_ratio = NULL;
  PyObject *__pyx_v_target_base_ratio = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_inventory_range_multiplier = NULL;
  PyObject *__pyx_v_target_base_amount = NULL;
  PyObject *__pyx_v_target_base_amount_in_quote = NULL;
  PyObject *__pyx_v_target_quote_amount = NULL;
  PyObject *__pyx_v_base_asset_range = NULL;
  PyObject *__pyx_v_high_water_mark = NULL;
  PyObject *__pyx_v_low_water_mark = NULL;
  PyObject *__pyx_v_low_water_mark_ratio = NULL;
  PyObject *__pyx_v_high_water_mark_ratio = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_total_order_size_ratio = NULL;
  PyObject *__pyx_v_bid_ask_ratios = NULL;
  PyObject *__pyx_v_inventory_skew_df = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  double __pyx_t_7;
  double __pyx_t_8;
  double __pyx_t_9;
  double __pyx_t_10;
  double __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13[3];
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("inventory_skew_stats_data_frame", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":497
 *     def inventory_skew_stats_data_frame(self) -> Optional[pd.DataFrame]:
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 * 
 *         price = self.get_price()
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 497, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":499
 *             ExchangeBase market = self._market_info.market
 * 
 *         price = self.get_price()             # <<<<<<<<<<<<<<
 *         base_asset_amount, quote_asset_amount = self.c_get_adjusted_available_balance(self.active_orders)
 *         total_order_size = calculate_total_order_size(self._order_amount, self._order_level_amount, self._order_levels)
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":500
 * 
 *         price = self.get_price()
 *         base_asset_amount, quote_asset_amount = self.c_get_adjusted_available_balance(self.active_orders)             # <<<<<<<<<<<<<<
 *         total_order_size = calculate_total_order_size(self._order_amount, self._order_level_amount, self._order_levels)
 * 
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 500, __pyx_L1_error)
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_adjusted_available_balance(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PyTuple_GET_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 500, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_1 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 500, __pyx_L1_error)
  }
  __pyx_v_base_asset_amount = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_quote_asset_amount = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":501
 *         price = self.get_price()
 *         base_asset_amount, quote_asset_amount = self.c_get_adjusted_available_balance(self.active_orders)
 *         total_order_size = calculate_total_order_size(self._order_amount, self._order_level_amount, self._order_levels)             # <<<<<<<<<<<<<<
 * 
 *         base_asset_value = base_asset_amount * price
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_calculate_total_order_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyLong_From_int(__pyx_v_self->_order_levels); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
    __pyx_t_3 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_v_self->_order_amount, __pyx_v_self->_order_level_amount, __pyx_t_5};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_3, (4-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_total_order_size = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":503
 *         total_order_size = calculate_total_order_size(self._order_amount, self._order_level_amount, self._order_levels)
 * 
 *         base_asset_value = base_asset_amount * price             # <<<<<<<<<<<<<<
 *         quote_asset_value = quote_asset_amount / price if price > s_decimal_zero else s_decimal_zero
 *         total_value = base_asset_amount + quote_asset_value
*/
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_base_asset_amount, __pyx_v_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_base_asset_value = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":504
 * 
 *         base_asset_value = base_asset_amount * price
 *         quote_asset_value = quote_asset_amount / price if price > s_decimal_zero else s_decimal_zero             # <<<<<<<<<<<<<<
 *         total_value = base_asset_amount + quote_asset_value
 *         total_value_in_quote = (base_asset_amount * price) + quote_asset_amount
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_price, __pyx_t_1, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_6) {
    __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_v_quote_asset_amount, __pyx_v_price); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __pyx_t_5;
    __pyx_t_5 = 0;
  } else {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __pyx_t_5;
    __pyx_t_5 = 0;
  }
  __pyx_v_quote_asset_value = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":505
 *         base_asset_value = base_asset_amount * price
 *         quote_asset_value = quote_asset_amount / price if price > s_decimal_zero else s_decimal_zero
 *         total_value = base_asset_amount + quote_asset_value             # <<<<<<<<<<<<<<
 *         total_value_in_quote = (base_asset_amount * price) + quote_asset_amount
 * 
*/
  __pyx_t_2 = PyNumber_Add(__pyx_v_base_asset_amount, __pyx_v_quote_asset_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_total_value = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":506
 *         quote_asset_value = quote_asset_amount / price if price > s_decimal_zero else s_decimal_zero
 *         total_value = base_asset_amount + quote_asset_value
 *         total_value_in_quote = (base_asset_amount * price) + quote_asset_amount             # <<<<<<<<<<<<<<
 * 
 *         base_asset_ratio = (base_asset_amount / total_value
*/
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_base_asset_amount, __pyx_v_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyNumber_Add(__pyx_t_2, __pyx_v_quote_asset_amount); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_total_value_in_quote = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":509
 * 
 *         base_asset_ratio = (base_asset_amount / total_value
 *                             if total_value > s_decimal_zero             # <<<<<<<<<<<<<<
 *                             else s_decimal_zero)
 *         quote_asset_ratio = Decimal("1") - base_asset_ratio if total_value > 0 else 0
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_total_value, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":508
 *         total_value_in_quote = (base_asset_amount * price) + quote_asset_amount
 * 
 *         base_asset_ratio = (base_asset_amount / total_value             # <<<<<<<<<<<<<<
 *                             if total_value > s_decimal_zero
 *                             else s_decimal_zero)
*/
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_v_base_asset_amount, __pyx_v_total_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_t_1;
    __pyx_t_1 = 0;
  } else {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":510
 *         base_asset_ratio = (base_asset_amount / total_value
 *                             if total_value > s_decimal_zero
 *                             else s_decimal_zero)             # <<<<<<<<<<<<<<
 *         quote_asset_ratio = Decimal("1") - base_asset_ratio if total_value > 0 else 0
 *         target_base_ratio = self._inventory_target_base_pct
*/
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_v_base_asset_ratio = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":511
 *                             if total_value > s_decimal_zero
 *                             else s_decimal_zero)
 *         quote_asset_ratio = Decimal("1") - base_asset_ratio if total_value > 0 else 0             # <<<<<<<<<<<<<<
 *         target_base_ratio = self._inventory_target_base_pct
 *         inventory_range_multiplier = self._inventory_range_multiplier
*/
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_total_value, __pyx_mstate_global->__pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_1};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_4 = PyNumber_Subtract(__pyx_t_1, __pyx_v_base_asset_ratio); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_0);
    __pyx_t_5 = __pyx_mstate_global->__pyx_int_0;
  }
  __pyx_v_quote_asset_ratio = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":512
 *                             else s_decimal_zero)
 *         quote_asset_ratio = Decimal("1") - base_asset_ratio if total_value > 0 else 0
 *         target_base_ratio = self._inventory_target_base_pct             # <<<<<<<<<<<<<<
 *         inventory_range_multiplier = self._inventory_range_multiplier
 *         target_base_amount = (total_value * target_base_ratio
*/
  __pyx_t_5 = __pyx_v_self->_inventory_target_base_pct;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_v_target_base_ratio = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":513
 *         quote_asset_ratio = Decimal("1") - base_asset_ratio if total_value > 0 else 0
 *         target_base_ratio = self._inventory_target_base_pct
 *         inventory_range_multiplier = self._inventory_range_multiplier             # <<<<<<<<<<<<<<
 *         target_base_amount = (total_value * target_base_ratio
 *                               if price > s_decimal_zero
*/
  __pyx_t_5 = __pyx_v_self->_inventory_range_multiplier;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_v_inventory_range_multiplier = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":515
 *         inventory_range_multiplier = self._inventory_range_multiplier
 *         target_base_amount = (total_value * target_base_ratio
 *                               if price > s_decimal_zero             # <<<<<<<<<<<<<<
 *                               else s_decimal_zero)
 *         target_base_amount_in_quote = target_base_ratio * total_value_in_quote
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_price, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":514
 *         target_base_ratio = self._inventory_target_base_pct
 *         inventory_range_multiplier = self._inventory_range_multiplier
 *         target_base_amount = (total_value * target_base_ratio             # <<<<<<<<<<<<<<
 *                               if price > s_decimal_zero
 *                               else s_decimal_zero)
*/
    __pyx_t_1 = PyNumber_Multiply(__pyx_v_total_value, __pyx_v_target_base_ratio); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_t_1;
    __pyx_t_1 = 0;
  } else {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":516
 *         target_base_amount = (total_value * target_base_ratio
 *                               if price > s_decimal_zero
 *                               else s_decimal_zero)             # <<<<<<<<<<<<<<
 *         target_base_amount_in_quote = target_base_ratio * total_value_in_quote
 *         target_quote_amount = (1 - target_base_ratio) * total_value_in_quote
*/
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_v_target_base_amount = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":517
 *                               if price > s_decimal_zero
 *                               else s_decimal_zero)
 *         target_base_amount_in_quote = target_base_ratio * total_value_in_quote             # <<<<<<<<<<<<<<
 *         target_quote_amount = (1 - target_base_ratio) * total_value_in_quote
 * 
*/
  __pyx_t_5 = PyNumber_Multiply(__pyx_v_target_base_ratio, __pyx_v_total_value_in_quote); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_target_base_amount_in_quote = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":518
 *                               else s_decimal_zero)
 *         target_base_amount_in_quote = target_base_ratio * total_value_in_quote
 *         target_quote_amount = (1 - target_base_ratio) * total_value_in_quote             # <<<<<<<<<<<<<<
 * 
 *         base_asset_range = total_order_size * self._inventory_range_multiplier
*/
  __pyx_t_5 = __Pyx_PyLong_SubtractCObj(__pyx_mstate_global->__pyx_int_1, __pyx_v_target_base_ratio, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = PyNumber_Multiply(__pyx_t_5, __pyx_v_total_value_in_quote); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_target_quote_amount = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":520
 *         target_quote_amount = (1 - target_base_ratio) * total_value_in_quote
 * 
 *         base_asset_range = total_order_size * self._inventory_range_multiplier             # <<<<<<<<<<<<<<
 *         base_asset_range = min(base_asset_range, total_value * Decimal("0.5"))
 *         high_water_mark = target_base_amount + base_asset_range
*/
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_total_order_size, __pyx_v_self->_inventory_range_multiplier); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_base_asset_range = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":521
 * 
 *         base_asset_range = total_order_size * self._inventory_range_multiplier
 *         base_asset_range = min(base_asset_range, total_value * Decimal("0.5"))             # <<<<<<<<<<<<<<
 *         high_water_mark = target_base_amount + base_asset_range
 *         low_water_mark = max(target_base_amount - base_asset_range, s_decimal_zero)
*/
  __pyx_t_5 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_5);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_3 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_0_5};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_4 = PyNumber_Multiply(__pyx_v_total_value, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_INCREF(__pyx_v_base_asset_range);
  __pyx_t_1 = __pyx_v_base_asset_range;
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 521, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_5 = __pyx_t_4;
  } else {
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = __pyx_t_1;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_t_5;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF_SET(__pyx_v_base_asset_range, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":522
 *         base_asset_range = total_order_size * self._inventory_range_multiplier
 *         base_asset_range = min(base_asset_range, total_value * Decimal("0.5"))
 *         high_water_mark = target_base_amount + base_asset_range             # <<<<<<<<<<<<<<
 *         low_water_mark = max(target_base_amount - base_asset_range, s_decimal_zero)
 *         low_water_mark_ratio = (low_water_mark / total_value
*/
  __pyx_t_4 = PyNumber_Add(__pyx_v_target_base_amount, __pyx_v_base_asset_range); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_high_water_mark = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":523
 *         base_asset_range = min(base_asset_range, total_value * Decimal("0.5"))
 *         high_water_mark = target_base_amount + base_asset_range
 *         low_water_mark = max(target_base_amount - base_asset_range, s_decimal_zero)             # <<<<<<<<<<<<<<
 *         low_water_mark_ratio = (low_water_mark / total_value
 *                                 if total_value > s_decimal_zero
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyNumber_Subtract(__pyx_v_target_base_amount, __pyx_v_base_asset_range); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_5, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 523, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = __pyx_t_4;
  } else {
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_1 = __pyx_t_5;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_low_water_mark = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":525
 *         low_water_mark = max(target_base_amount - base_asset_range, s_decimal_zero)
 *         low_water_mark_ratio = (low_water_mark / total_value
 *                                 if total_value > s_decimal_zero             # <<<<<<<<<<<<<<
 *                                 else s_decimal_zero)
 *         high_water_mark_ratio = (high_water_mark / total_value
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_total_value, __pyx_t_1, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_6) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":524
 *         high_water_mark = target_base_amount + base_asset_range
 *         low_water_mark = max(target_base_amount - base_asset_range, s_decimal_zero)
 *         low_water_mark_ratio = (low_water_mark / total_value             # <<<<<<<<<<<<<<
 *                                 if total_value > s_decimal_zero
 *                                 else s_decimal_zero)
*/
    __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_v_low_water_mark, __pyx_v_total_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __pyx_t_5;
    __pyx_t_5 = 0;
  } else {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":526
 *         low_water_mark_ratio = (low_water_mark / total_value
 *                                 if total_value > s_decimal_zero
 *                                 else s_decimal_zero)             # <<<<<<<<<<<<<<
 *         high_water_mark_ratio = (high_water_mark / total_value
 *                                  if total_value > s_decimal_zero
*/
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __pyx_t_5;
    __pyx_t_5 = 0;
  }
  __pyx_v_low_water_mark_ratio = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":528
 *                                 else s_decimal_zero)
 *         high_water_mark_ratio = (high_water_mark / total_value
 *                                  if total_value > s_decimal_zero             # <<<<<<<<<<<<<<
 *                                  else s_decimal_zero)
 *         high_water_mark_ratio = min(1.0, high_water_mark_ratio)
*/
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_total_value, __pyx_t_5, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":527
 *                                 if total_value > s_decimal_zero
 *                                 else s_decimal_zero)
 *         high_water_mark_ratio = (high_water_mark / total_value             # <<<<<<<<<<<<<<
 *                                  if total_value > s_decimal_zero
 *                                  else s_decimal_zero)
*/
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_v_high_water_mark, __pyx_v_total_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __pyx_t_1;
    __pyx_t_1 = 0;
  } else {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":529
 *         high_water_mark_ratio = (high_water_mark / total_value
 *                                  if total_value > s_decimal_zero
 *                                  else s_decimal_zero)             # <<<<<<<<<<<<<<
 *         high_water_mark_ratio = min(1.0, high_water_mark_ratio)
 *         total_order_size_ratio = (self._order_amount * Decimal("2") / total_value
*/
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_v_high_water_mark_ratio = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":530
 *                                  if total_value > s_decimal_zero
 *                                  else s_decimal_zero)
 *         high_water_mark_ratio = min(1.0, high_water_mark_ratio)             # <<<<<<<<<<<<<<
 *         total_order_size_ratio = (self._order_amount * Decimal("2") / total_value
 *                                   if total_value > s_decimal_zero
*/
  __Pyx_INCREF(__pyx_v_high_water_mark_ratio);
  __pyx_t_4 = __pyx_v_high_water_mark_ratio;
  __pyx_t_7 = 1.0;
  __pyx_t_5 = PyFloat_FromDouble(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_5, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = __pyx_t_4;
  } else {
    __pyx_t_2 = PyFloat_FromDouble(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 530, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_high_water_mark_ratio, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":532
 *         high_water_mark_ratio = min(1.0, high_water_mark_ratio)
 *         total_order_size_ratio = (self._order_amount * Decimal("2") / total_value
 *                                   if total_value > s_decimal_zero             # <<<<<<<<<<<<<<
 *                                   else s_decimal_zero)
 *         bid_ask_ratios = c_calculate_bid_ask_ratios_from_base_asset_ratio(
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_total_value, __pyx_t_1, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":531
 *                                  else s_decimal_zero)
 *         high_water_mark_ratio = min(1.0, high_water_mark_ratio)
 *         total_order_size_ratio = (self._order_amount * Decimal("2") / total_value             # <<<<<<<<<<<<<<
 *                                   if total_value > s_decimal_zero
 *                                   else s_decimal_zero)
*/
    __pyx_t_1 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_2};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_5 = PyNumber_Multiply(__pyx_v_self->_order_amount, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_5, __pyx_v_total_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_4 = __pyx_t_2;
    __pyx_t_2 = 0;
  } else {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":533
 *         total_order_size_ratio = (self._order_amount * Decimal("2") / total_value
 *                                   if total_value > s_decimal_zero
 *                                   else s_decimal_zero)             # <<<<<<<<<<<<<<
 *         bid_ask_ratios = c_calculate_bid_ask_ratios_from_base_asset_ratio(
 *             float(base_asset_amount),
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_v_total_order_size_ratio = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":535
 *                                   else s_decimal_zero)
 *         bid_ask_ratios = c_calculate_bid_ask_ratios_from_base_asset_ratio(
 *             float(base_asset_amount),             # <<<<<<<<<<<<<<
 *             float(quote_asset_amount),
 *             float(price),
*/
  __pyx_t_7 = __Pyx_PyObject_AsDouble(__pyx_v_base_asset_amount); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_7, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 535, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":536
 *         bid_ask_ratios = c_calculate_bid_ask_ratios_from_base_asset_ratio(
 *             float(base_asset_amount),
 *             float(quote_asset_amount),             # <<<<<<<<<<<<<<
 *             float(price),
 *             float(target_base_ratio),
*/
  __pyx_t_8 = __Pyx_PyObject_AsDouble(__pyx_v_quote_asset_amount); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_8, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 536, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":537
 *             float(base_asset_amount),
 *             float(quote_asset_amount),
 *             float(price),             # <<<<<<<<<<<<<<
 *             float(target_base_ratio),
 *             float(base_asset_range)
*/
  __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_v_price); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_9, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 537, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":538
 *             float(quote_asset_amount),
 *             float(price),
 *             float(target_base_ratio),             # <<<<<<<<<<<<<<
 *             float(base_asset_range)
 *         )
*/
  __pyx_t_10 = __Pyx_PyObject_AsDouble(__pyx_v_target_base_ratio); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_10, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 538, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":539
 *             float(price),
 *             float(target_base_ratio),
 *             float(base_asset_range)             # <<<<<<<<<<<<<<
 *         )
 *         inventory_skew_df = pd.DataFrame(data=[
*/
  __pyx_t_11 = __Pyx_PyObject_AsDouble(__pyx_v_base_asset_range); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_11, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 539, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":534
 *                                   if total_value > s_decimal_zero
 *                                   else s_decimal_zero)
 *         bid_ask_ratios = c_calculate_bid_ask_ratios_from_base_asset_ratio(             # <<<<<<<<<<<<<<
 *             float(base_asset_amount),
 *             float(quote_asset_amount),
*/
  __pyx_t_4 = __pyx_f_10hummingbot_8strategy_18pure_market_making_25inventory_skew_calculator_c_calculate_bid_ask_ratios_from_base_asset_ratio(__pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_bid_ask_ratios = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":541
 *             float(base_asset_range)
 *         )
 *         inventory_skew_df = pd.DataFrame(data=[             # <<<<<<<<<<<<<<
 *             [f"Target Value ({self.quote_asset})", f"{target_base_amount_in_quote:.4f}",
 *              f"{target_quote_amount:.4f}"],
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pd); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_DataFrame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":542
 *         )
 *         inventory_skew_df = pd.DataFrame(data=[
 *             [f"Target Value ({self.quote_asset})", f"{target_base_amount_in_quote:.4f}",             # <<<<<<<<<<<<<<
 *              f"{target_quote_amount:.4f}"],
 *             ["Current %", f"{base_asset_ratio:.1%}", f"{quote_asset_ratio:.1%}"],
*/
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_quote_asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_12 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_13[0] = __pyx_mstate_global->__pyx_kp_u_Target_Value;
  __pyx_t_13[1] = __pyx_t_12;
  __pyx_t_13[2] = __pyx_mstate_global->__pyx_kp_u__13;
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_13, 3, 14 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12) + 1, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12));
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyObject_Format(__pyx_v_target_base_amount_in_quote, __pyx_mstate_global->__pyx_kp_u_4f); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":543
 *         inventory_skew_df = pd.DataFrame(data=[
 *             [f"Target Value ({self.quote_asset})", f"{target_base_amount_in_quote:.4f}",
 *              f"{target_quote_amount:.4f}"],             # <<<<<<<<<<<<<<
 *             ["Current %", f"{base_asset_ratio:.1%}", f"{quote_asset_ratio:.1%}"],
 *             ["Target %", f"{target_base_ratio:.1%}", f"{1 - target_base_ratio:.1%}"],
*/
  __pyx_t_14 = __Pyx_PyObject_Format(__pyx_v_target_quote_amount, __pyx_mstate_global->__pyx_kp_u_4f); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":542
 *         )
 *         inventory_skew_df = pd.DataFrame(data=[
 *             [f"Target Value ({self.quote_asset})", f"{target_base_amount_in_quote:.4f}",             # <<<<<<<<<<<<<<
 *              f"{target_quote_amount:.4f}"],
 *             ["Current %", f"{base_asset_ratio:.1%}", f"{quote_asset_ratio:.1%}"],
*/
  __pyx_t_15 = PyList_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_15, 0, __pyx_t_5) != (0)) __PYX_ERR(0, 542, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_15, 1, __pyx_t_12) != (0)) __PYX_ERR(0, 542, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_14);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_15, 2, __pyx_t_14) != (0)) __PYX_ERR(0, 542, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_12 = 0;
  __pyx_t_14 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":544
 *             [f"Target Value ({self.quote_asset})", f"{target_base_amount_in_quote:.4f}",
 *              f"{target_quote_amount:.4f}"],
 *             ["Current %", f"{base_asset_ratio:.1%}", f"{quote_asset_ratio:.1%}"],             # <<<<<<<<<<<<<<
 *             ["Target %", f"{target_base_ratio:.1%}", f"{1 - target_base_ratio:.1%}"],
 *             ["Inventory Range", f"{low_water_mark_ratio:.1%} - {high_water_mark_ratio:.1%}",
*/
  __pyx_t_14 = __Pyx_PyObject_Format(__pyx_v_base_asset_ratio, __pyx_mstate_global->__pyx_kp_u_1_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_12 = __Pyx_PyObject_Format(__pyx_v_quote_asset_ratio, __pyx_mstate_global->__pyx_kp_u_1_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_5 = PyList_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Current);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Current);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 0, __pyx_mstate_global->__pyx_kp_u_Current) != (0)) __PYX_ERR(0, 544, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_14);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_14) != (0)) __PYX_ERR(0, 544, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 2, __pyx_t_12) != (0)) __PYX_ERR(0, 544, __pyx_L1_error);
  __pyx_t_14 = 0;
  __pyx_t_12 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":545
 *              f"{target_quote_amount:.4f}"],
 *             ["Current %", f"{base_asset_ratio:.1%}", f"{quote_asset_ratio:.1%}"],
 *             ["Target %", f"{target_base_ratio:.1%}", f"{1 - target_base_ratio:.1%}"],             # <<<<<<<<<<<<<<
 *             ["Inventory Range", f"{low_water_mark_ratio:.1%} - {high_water_mark_ratio:.1%}",
 *              f"{1 - high_water_mark_ratio:.1%} - {1 - low_water_mark_ratio:.1%}"],
*/
  __pyx_t_12 = __Pyx_PyObject_Format(__pyx_v_target_base_ratio, __pyx_mstate_global->__pyx_kp_u_1_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_14 = __Pyx_PyLong_SubtractCObj(__pyx_mstate_global->__pyx_int_1, __pyx_v_target_base_ratio, 1, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_16 = __Pyx_PyObject_Format(__pyx_t_14, __pyx_mstate_global->__pyx_kp_u_1_2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = PyList_New(3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Target);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Target);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 0, __pyx_mstate_global->__pyx_kp_u_Target) != (0)) __PYX_ERR(0, 545, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 1, __pyx_t_12) != (0)) __PYX_ERR(0, 545, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 2, __pyx_t_16) != (0)) __PYX_ERR(0, 545, __pyx_L1_error);
  __pyx_t_12 = 0;
  __pyx_t_16 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":546
 *             ["Current %", f"{base_asset_ratio:.1%}", f"{quote_asset_ratio:.1%}"],
 *             ["Target %", f"{target_base_ratio:.1%}", f"{1 - target_base_ratio:.1%}"],
 *             ["Inventory Range", f"{low_water_mark_ratio:.1%} - {high_water_mark_ratio:.1%}",             # <<<<<<<<<<<<<<
 *              f"{1 - high_water_mark_ratio:.1%} - {1 - low_water_mark_ratio:.1%}"],
 *             ["Order Adjust %", f"{bid_ask_ratios.bid_ratio:.1%}", f"{bid_ask_ratios.ask_ratio:.1%}"]
*/
  __pyx_t_16 = __Pyx_PyObject_Format(__pyx_v_low_water_mark_ratio, __pyx_mstate_global->__pyx_kp_u_1_2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_12 = __Pyx_PyObject_Format(__pyx_v_high_water_mark_ratio, __pyx_mstate_global->__pyx_kp_u_1_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13[0] = __pyx_t_16;
  __pyx_t_13[1] = __pyx_mstate_global->__pyx_kp_u__14;
  __pyx_t_13[2] = __pyx_t_12;
  __pyx_t_17 = __Pyx_PyUnicode_Join(__pyx_t_13, 3, __Pyx_PyUnicode_GET_LENGTH(__pyx_t_16) + 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_16) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12));
  if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":547
 *             ["Target %", f"{target_base_ratio:.1%}", f"{1 - target_base_ratio:.1%}"],
 *             ["Inventory Range", f"{low_water_mark_ratio:.1%} - {high_water_mark_ratio:.1%}",
 *              f"{1 - high_water_mark_ratio:.1%} - {1 - low_water_mark_ratio:.1%}"],             # <<<<<<<<<<<<<<
 *             ["Order Adjust %", f"{bid_ask_ratios.bid_ratio:.1%}", f"{bid_ask_ratios.ask_ratio:.1%}"]
 *         ])
*/
  __pyx_t_12 = __Pyx_PyLong_SubtractCObj(__pyx_mstate_global->__pyx_int_1, __pyx_v_high_water_mark_ratio, 1, 0, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_16 = __Pyx_PyObject_Format(__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_1_2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyLong_SubtractCObj(__pyx_mstate_global->__pyx_int_1, __pyx_v_low_water_mark_ratio, 1, 0, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_18 = __Pyx_PyObject_Format(__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_1_2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_13[0] = __pyx_t_16;
  __pyx_t_13[1] = __pyx_mstate_global->__pyx_kp_u__14;
  __pyx_t_13[2] = __pyx_t_18;
  __pyx_t_12 = __Pyx_PyUnicode_Join(__pyx_t_13, 3, __Pyx_PyUnicode_GET_LENGTH(__pyx_t_16) + 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_18), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_16) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_18));
  if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":546
 *             ["Current %", f"{base_asset_ratio:.1%}", f"{quote_asset_ratio:.1%}"],
 *             ["Target %", f"{target_base_ratio:.1%}", f"{1 - target_base_ratio:.1%}"],
 *             ["Inventory Range", f"{low_water_mark_ratio:.1%} - {high_water_mark_ratio:.1%}",             # <<<<<<<<<<<<<<
 *              f"{1 - high_water_mark_ratio:.1%} - {1 - low_water_mark_ratio:.1%}"],
 *             ["Order Adjust %", f"{bid_ask_ratios.bid_ratio:.1%}", f"{bid_ask_ratios.ask_ratio:.1%}"]
*/
  __pyx_t_18 = PyList_New(3); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Inventory_Range);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Inventory_Range);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 0, __pyx_mstate_global->__pyx_kp_u_Inventory_Range) != (0)) __PYX_ERR(0, 546, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_17);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 1, __pyx_t_17) != (0)) __PYX_ERR(0, 546, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 2, __pyx_t_12) != (0)) __PYX_ERR(0, 546, __pyx_L1_error);
  __pyx_t_17 = 0;
  __pyx_t_12 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":548
 *             ["Inventory Range", f"{low_water_mark_ratio:.1%} - {high_water_mark_ratio:.1%}",
 *              f"{1 - high_water_mark_ratio:.1%} - {1 - low_water_mark_ratio:.1%}"],
 *             ["Order Adjust %", f"{bid_ask_ratios.bid_ratio:.1%}", f"{bid_ask_ratios.ask_ratio:.1%}"]             # <<<<<<<<<<<<<<
 *         ])
 *         return inventory_skew_df
*/
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_bid_ask_ratios, __pyx_mstate_global->__pyx_n_u_bid_ratio); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_17 = __Pyx_PyObject_Format(__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_1_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_bid_ask_ratios, __pyx_mstate_global->__pyx_n_u_ask_ratio); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_16 = __Pyx_PyObject_Format(__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_1_2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = PyList_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Order_Adjust);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Order_Adjust);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_12, 0, __pyx_mstate_global->__pyx_kp_u_Order_Adjust) != (0)) __PYX_ERR(0, 548, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_17);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_12, 1, __pyx_t_17) != (0)) __PYX_ERR(0, 548, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_12, 2, __pyx_t_16) != (0)) __PYX_ERR(0, 548, __pyx_L1_error);
  __pyx_t_17 = 0;
  __pyx_t_16 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":541
 *             float(base_asset_range)
 *         )
 *         inventory_skew_df = pd.DataFrame(data=[             # <<<<<<<<<<<<<<
 *             [f"Target Value ({self.quote_asset})", f"{target_base_amount_in_quote:.4f}",
 *              f"{target_quote_amount:.4f}"],
*/
  __pyx_t_16 = PyList_New(5); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_15);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_16, 0, __pyx_t_15) != (0)) __PYX_ERR(0, 541, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_16, 1, __pyx_t_5) != (0)) __PYX_ERR(0, 541, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_14);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_16, 2, __pyx_t_14) != (0)) __PYX_ERR(0, 541, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_18);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_16, 3, __pyx_t_18) != (0)) __PYX_ERR(0, 541, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_16, 4, __pyx_t_12) != (0)) __PYX_ERR(0, 541, __pyx_L1_error);
  __pyx_t_15 = 0;
  __pyx_t_5 = 0;
  __pyx_t_14 = 0;
  __pyx_t_18 = 0;
  __pyx_t_12 = 0;
  __pyx_t_3 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
    __pyx_t_3 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, NULL};
    __pyx_t_12 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_data, __pyx_t_16, __pyx_t_12, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 541, __pyx_L1_error)
    __pyx_t_4 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_12);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_v_inventory_skew_df = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":550
 *             ["Order Adjust %", f"{bid_ask_ratios.bid_ratio:.1%}", f"{bid_ask_ratios.ask_ratio:.1%}"]
 *         ])
 *         return inventory_skew_df             # <<<<<<<<<<<<<<
 * 
 *     def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_inventory_skew_df);
  __pyx_r = __pyx_v_inventory_skew_df;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":495
 *         self._inventory_cost_price_delegate = value
 * 
 *     def inventory_skew_stats_data_frame(self) -> Optional[pd.DataFrame]:             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.inventory_skew_stats_data_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_base_asset_amount);
  __Pyx_XDECREF(__pyx_v_quote_asset_amount);
  __Pyx_XDECREF(__pyx_v_total_order_size);
  __Pyx_XDECREF(__pyx_v_base_asset_value);
  __Pyx_XDECREF(__pyx_v_quote_asset_value);
  __Pyx_XDECREF(__pyx_v_total_value);
  __Pyx_XDECREF(__pyx_v_total_value_in_quote);
  __Pyx_XDECREF(__pyx_v_base_asset_ratio);
  __Pyx_XDECREF(__pyx_v_quote_asset_ratio);
  __Pyx_XDECREF(__pyx_v_target_base_ratio);
  __Pyx_XDECREF(__pyx_v_inventory_range_multiplier);
  __Pyx_XDECREF(__pyx_v_target_base_amount);
  __Pyx_XDECREF(__pyx_v_target_base_amount_in_quote);
  __Pyx_XDECREF(__pyx_v_target_quote_amount);
  __Pyx_XDECREF(__pyx_v_base_asset_range);
  __Pyx_XDECREF(__pyx_v_high_water_mark);
  __Pyx_XDECREF(__pyx_v_low_water_mark);
  __Pyx_XDECREF(__pyx_v_low_water_mark_ratio);
  __Pyx_XDECREF(__pyx_v_high_water_mark_ratio);
  __Pyx_XDECREF(__pyx_v_total_order_size_ratio);
  __Pyx_XDECREF(__pyx_v_bid_ask_ratios);
  __Pyx_XDECREF(__pyx_v_inventory_skew_df);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":552
 *         return inventory_skew_df
 * 
 *     def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self._market_info.get_mid_price()
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13pure_mm_assets_df(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13pure_mm_assets_df = {"pure_mm_assets_df", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13pure_mm_assets_df, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13pure_mm_assets_df(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_to_show_current_pct = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pure_mm_assets_df (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_to_show_current_pct,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 552, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 552, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "pure_mm_assets_df", 0) < (0)) __PYX_ERR(0, 552, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("pure_mm_assets_df", 1, 1, 1, i); __PYX_ERR(0, 552, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 552, __pyx_L3_error)
    }
    __pyx_v_to_show_current_pct = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pure_mm_assets_df", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 552, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.pure_mm_assets_df", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12pure_mm_assets_df(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), __pyx_v_to_show_current_pct);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12pure_mm_assets_df(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_to_show_current_pct) {
  PyObject *__pyx_v_market = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_trading_pair = NULL;
  PyObject *__pyx_v_base_asset = NULL;
  PyObject *__pyx_v_quote_asset = NULL;
  PyObject *__pyx_v_price = NULL;
  double __pyx_v_base_balance;
  double __pyx_v_quote_balance;
  double __pyx_v_available_base_balance;
  double __pyx_v_available_quote_balance;
  double __pyx_v_base_value;
  double __pyx_v_total_in_quote;
  double __pyx_v_base_ratio;
  double __pyx_v_quote_ratio;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_df = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  size_t __pyx_t_8;
  double __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11[3];
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pure_mm_assets_df", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":553
 * 
 *     def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:
 *         market, trading_pair, base_asset, quote_asset = self._market_info             # <<<<<<<<<<<<<<
 *         price = self._market_info.get_mid_price()
 *         base_balance = float(market.get_balance(base_asset))
*/
  __pyx_t_1 = __pyx_v_self->_market_info;
  __Pyx_INCREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 4)) {
      if (size > 4) __Pyx_RaiseTooManyValuesError(4);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 553, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3);
      __Pyx_INCREF(__pyx_t_5);
    } else {
      __pyx_t_2 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 553, __pyx_L1_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 553, __pyx_L1_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyList_GetItemRefFast(sequence, 2, __Pyx_ReferenceSharing_SharedReference);
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 553, __pyx_L1_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyList_GetItemRefFast(sequence, 3, __Pyx_ReferenceSharing_SharedReference);
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 553, __pyx_L1_error)
      __Pyx_XGOTREF(__pyx_t_5);
    }
    #else
    {
      Py_ssize_t i;
      PyObject** temps[4] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
      for (i=0; i < 4; i++) {
        PyObject* item = __Pyx_PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 553, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[4] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
    __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
    for (index=0; index < 4; index++) {
      PyObject* item = __pyx_t_7(__pyx_t_6); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 4) < (0)) __PYX_ERR(0, 553, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 553, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_market = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_trading_pair = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_base_asset = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_quote_asset = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":554
 *     def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self._market_info.get_mid_price()             # <<<<<<<<<<<<<<
 *         base_balance = float(market.get_balance(base_asset))
 *         quote_balance = float(market.get_balance(quote_asset))
*/
  __pyx_t_5 = __pyx_v_self->_market_info;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_8 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_mid_price, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":555
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self._market_info.get_mid_price()
 *         base_balance = float(market.get_balance(base_asset))             # <<<<<<<<<<<<<<
 *         quote_balance = float(market.get_balance(quote_asset))
 *         available_base_balance = float(market.get_available_balance(base_asset))
*/
  __pyx_t_5 = __pyx_v_market;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_8 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_base_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_balance, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_t_1); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_9, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_base_balance = __pyx_t_9;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":556
 *         price = self._market_info.get_mid_price()
 *         base_balance = float(market.get_balance(base_asset))
 *         quote_balance = float(market.get_balance(quote_asset))             # <<<<<<<<<<<<<<
 *         available_base_balance = float(market.get_available_balance(base_asset))
 *         available_quote_balance = float(market.get_available_balance(quote_asset))
*/
  __pyx_t_5 = __pyx_v_market;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_8 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_quote_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_balance, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_t_1); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_9, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_quote_balance = __pyx_t_9;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":557
 *         base_balance = float(market.get_balance(base_asset))
 *         quote_balance = float(market.get_balance(quote_asset))
 *         available_base_balance = float(market.get_available_balance(base_asset))             # <<<<<<<<<<<<<<
 *         available_quote_balance = float(market.get_available_balance(quote_asset))
 *         base_value = base_balance * float(price)
*/
  __pyx_t_5 = __pyx_v_market;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_8 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_base_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_available_balance, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 557, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_t_1); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_9, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_available_base_balance = __pyx_t_9;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":558
 *         quote_balance = float(market.get_balance(quote_asset))
 *         available_base_balance = float(market.get_available_balance(base_asset))
 *         available_quote_balance = float(market.get_available_balance(quote_asset))             # <<<<<<<<<<<<<<
 *         base_value = base_balance * float(price)
 *         total_in_quote = base_value + quote_balance
*/
  __pyx_t_5 = __pyx_v_market;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_8 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_quote_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_available_balance, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_t_1); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_9, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_available_quote_balance = __pyx_t_9;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":559
 *         available_base_balance = float(market.get_available_balance(base_asset))
 *         available_quote_balance = float(market.get_available_balance(quote_asset))
 *         base_value = base_balance * float(price)             # <<<<<<<<<<<<<<
 *         total_in_quote = base_value + quote_balance
 *         base_ratio = base_value / total_in_quote if total_in_quote > 0 else 0
*/
  __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_v_price); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_9, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 559, __pyx_L1_error)
  __pyx_v_base_value = (__pyx_v_base_balance * __pyx_t_9);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":560
 *         available_quote_balance = float(market.get_available_balance(quote_asset))
 *         base_value = base_balance * float(price)
 *         total_in_quote = base_value + quote_balance             # <<<<<<<<<<<<<<
 *         base_ratio = base_value / total_in_quote if total_in_quote > 0 else 0
 *         quote_ratio = quote_balance / total_in_quote if total_in_quote > 0 else 0
*/
  __pyx_v_total_in_quote = (__pyx_v_base_value + __pyx_v_quote_balance);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":561
 *         base_value = base_balance * float(price)
 *         total_in_quote = base_value + quote_balance
 *         base_ratio = base_value / total_in_quote if total_in_quote > 0 else 0             # <<<<<<<<<<<<<<
 *         quote_ratio = quote_balance / total_in_quote if total_in_quote > 0 else 0
 *         data=[
*/
  __pyx_t_10 = (__pyx_v_total_in_quote > 0.0);
  if (__pyx_t_10) {
    if (unlikely(__pyx_v_total_in_quote == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 561, __pyx_L1_error)
    }
    __pyx_t_9 = (__pyx_v_base_value / __pyx_v_total_in_quote);
  } else {
    __pyx_t_9 = 0.0;
  }
  __pyx_v_base_ratio = __pyx_t_9;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":562
 *         total_in_quote = base_value + quote_balance
 *         base_ratio = base_value / total_in_quote if total_in_quote > 0 else 0
 *         quote_ratio = quote_balance / total_in_quote if total_in_quote > 0 else 0             # <<<<<<<<<<<<<<
 *         data=[
 *             ["", base_asset, quote_asset],
*/
  __pyx_t_10 = (__pyx_v_total_in_quote > 0.0);
  if (__pyx_t_10) {
    if (unlikely(__pyx_v_total_in_quote == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 562, __pyx_L1_error)
    }
    __pyx_t_9 = (__pyx_v_quote_balance / __pyx_v_total_in_quote);
  } else {
    __pyx_t_9 = 0.0;
  }
  __pyx_v_quote_ratio = __pyx_t_9;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":564
 *         quote_ratio = quote_balance / total_in_quote if total_in_quote > 0 else 0
 *         data=[
 *             ["", base_asset, quote_asset],             # <<<<<<<<<<<<<<
 *             ["Total Balance", round(base_balance, 4), round(quote_balance, 4)],
 *             ["Available Balance", round(available_base_balance, 4), round(available_quote_balance, 4)],
*/
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__15);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u__15);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_mstate_global->__pyx_kp_u__15) != (0)) __PYX_ERR(0, 564, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_base_asset);
  __Pyx_GIVEREF(__pyx_v_base_asset);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_v_base_asset) != (0)) __PYX_ERR(0, 564, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_quote_asset);
  __Pyx_GIVEREF(__pyx_v_quote_asset);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_v_quote_asset) != (0)) __PYX_ERR(0, 564, __pyx_L1_error);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":565
 *         data=[
 *             ["", base_asset, quote_asset],
 *             ["Total Balance", round(base_balance, 4), round(quote_balance, 4)],             # <<<<<<<<<<<<<<
 *             ["Available Balance", round(available_base_balance, 4), round(available_quote_balance, 4)],
 *             [f"Current Value ({quote_asset})", round(base_value, 4), round(quote_balance, 4)]
*/
  __pyx_t_4 = NULL;
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_base_balance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_3, __pyx_mstate_global->__pyx_int_4};
    __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_round, __pyx_callargs+__pyx_t_8, (3-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  }
  __pyx_t_4 = NULL;
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_quote_balance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_2, __pyx_mstate_global->__pyx_int_4};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_round, __pyx_callargs+__pyx_t_8, (3-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Total_Balance);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Total_Balance);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_mstate_global->__pyx_kp_u_Total_Balance) != (0)) __PYX_ERR(0, 565, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_5) != (0)) __PYX_ERR(0, 565, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 2, __pyx_t_3) != (0)) __PYX_ERR(0, 565, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":566
 *             ["", base_asset, quote_asset],
 *             ["Total Balance", round(base_balance, 4), round(quote_balance, 4)],
 *             ["Available Balance", round(available_base_balance, 4), round(available_quote_balance, 4)],             # <<<<<<<<<<<<<<
 *             [f"Current Value ({quote_asset})", round(base_value, 4), round(quote_balance, 4)]
 *         ]
*/
  __pyx_t_5 = NULL;
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_available_base_balance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_4, __pyx_mstate_global->__pyx_int_4};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_round, __pyx_callargs+__pyx_t_8, (3-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_5 = NULL;
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_available_quote_balance); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_6, __pyx_mstate_global->__pyx_int_4};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_round, __pyx_callargs+__pyx_t_8, (3-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_6 = PyList_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Available_Balance);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Available_Balance);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 0, __pyx_mstate_global->__pyx_kp_u_Available_Balance) != (0)) __PYX_ERR(0, 566, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 1, __pyx_t_3) != (0)) __PYX_ERR(0, 566, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 2, __pyx_t_4) != (0)) __PYX_ERR(0, 566, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":567
 *             ["Total Balance", round(base_balance, 4), round(quote_balance, 4)],
 *             ["Available Balance", round(available_base_balance, 4), round(available_quote_balance, 4)],
 *             [f"Current Value ({quote_asset})", round(base_value, 4), round(quote_balance, 4)]             # <<<<<<<<<<<<<<
 *         ]
 *         if to_show_current_pct:
*/
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_quote_asset, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11[0] = __pyx_mstate_global->__pyx_kp_u_Current_Value;
  __pyx_t_11[1] = __pyx_t_4;
  __pyx_t_11[2] = __pyx_mstate_global->__pyx_kp_u__13;
  __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_11, 3, 15 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 1, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4));
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_12 = PyFloat_FromDouble(__pyx_v_base_value); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_8 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_12, __pyx_mstate_global->__pyx_int_4};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_round, __pyx_callargs+__pyx_t_8, (3-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_5 = NULL;
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_quote_balance); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_8 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_13, __pyx_mstate_global->__pyx_int_4};
    __pyx_t_12 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_round, __pyx_callargs+__pyx_t_8, (3-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
  }
  __pyx_t_13 = PyList_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_13, 0, __pyx_t_3) != (0)) __PYX_ERR(0, 567, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_13, 1, __pyx_t_4) != (0)) __PYX_ERR(0, 567, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_13, 2, __pyx_t_12) != (0)) __PYX_ERR(0, 567, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_12 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":563
 *         base_ratio = base_value / total_in_quote if total_in_quote > 0 else 0
 *         quote_ratio = quote_balance / total_in_quote if total_in_quote > 0 else 0
 *         data=[             # <<<<<<<<<<<<<<
 *             ["", base_asset, quote_asset],
 *             ["Total Balance", round(base_balance, 4), round(quote_balance, 4)],
*/
  __pyx_t_12 = PyList_New(4); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_12, 0, __pyx_t_1) != (0)) __PYX_ERR(0, 563, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_12, 1, __pyx_t_2) != (0)) __PYX_ERR(0, 563, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_12, 2, __pyx_t_6) != (0)) __PYX_ERR(0, 563, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_13);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_12, 3, __pyx_t_13) != (0)) __PYX_ERR(0, 563, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_13 = 0;
  __pyx_v_data = ((PyObject*)__pyx_t_12);
  __pyx_t_12 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":569
 *             [f"Current Value ({quote_asset})", round(base_value, 4), round(quote_balance, 4)]
 *         ]
 *         if to_show_current_pct:             # <<<<<<<<<<<<<<
 *             data.append(["Current %", f"{base_ratio:.1%}", f"{quote_ratio:.1%}"])
 *         df = pd.DataFrame(data=data)
*/
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_to_show_current_pct); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 569, __pyx_L1_error)
  if (__pyx_t_10) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":570
 *         ]
 *         if to_show_current_pct:
 *             data.append(["Current %", f"{base_ratio:.1%}", f"{quote_ratio:.1%}"])             # <<<<<<<<<<<<<<
 *         df = pd.DataFrame(data=data)
 *         return df
*/
    __pyx_t_12 = PyFloat_FromDouble(__pyx_v_base_ratio); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_13 = __Pyx_PyObject_Format(__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_1_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = PyFloat_FromDouble(__pyx_v_quote_ratio); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_6 = __Pyx_PyObject_Format(__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_1_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = PyList_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Current);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Current);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_12, 0, __pyx_mstate_global->__pyx_kp_u_Current) != (0)) __PYX_ERR(0, 570, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_13);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_12, 1, __pyx_t_13) != (0)) __PYX_ERR(0, 570, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_12, 2, __pyx_t_6) != (0)) __PYX_ERR(0, 570, __pyx_L1_error);
    __pyx_t_13 = 0;
    __pyx_t_6 = 0;
    __pyx_t_14 = __Pyx_PyList_Append(__pyx_v_data, __pyx_t_12); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 570, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":569
 *             [f"Current Value ({quote_asset})", round(base_value, 4), round(quote_balance, 4)]
 *         ]
 *         if to_show_current_pct:             # <<<<<<<<<<<<<<
 *             data.append(["Current %", f"{base_ratio:.1%}", f"{quote_ratio:.1%}"])
 *         df = pd.DataFrame(data=data)
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":571
 *         if to_show_current_pct:
 *             data.append(["Current %", f"{base_ratio:.1%}", f"{quote_ratio:.1%}"])
 *         df = pd.DataFrame(data=data)             # <<<<<<<<<<<<<<
 *         return df
 * 
*/
  __pyx_t_6 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_pd); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_DataFrame); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_8 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_6);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_8 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_6, NULL};
    __pyx_t_13 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_data, __pyx_v_data, __pyx_t_13, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 571, __pyx_L1_error)
    __pyx_t_12 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_13);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
  }
  __pyx_v_df = __pyx_t_12;
  __pyx_t_12 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":572
 *             data.append(["Current %", f"{base_ratio:.1%}", f"{quote_ratio:.1%}"])
 *         df = pd.DataFrame(data=data)
 *         return df             # <<<<<<<<<<<<<<
 * 
 *     def active_orders_df(self) -> pd.DataFrame:
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_df);
  __pyx_r = __pyx_v_df;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":552
 *         return inventory_skew_df
 * 
 *     def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self._market_info.get_mid_price()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.pure_mm_assets_df", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_trading_pair);
  __Pyx_XDECREF(__pyx_v_base_asset);
  __Pyx_XDECREF(__pyx_v_quote_asset);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_df);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":574
 *         return df
 * 
 *     def active_orders_df(self) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self.get_price()
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15active_orders_df(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15active_orders_df = {"active_orders_df", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15active_orders_df, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15active_orders_df(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("active_orders_df (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("active_orders_df", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("active_orders_df", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14active_orders_df(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":580
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])
 *         active_orders.sort(key=lambda x: x.price, reverse=True)             # <<<<<<<<<<<<<<
 *         columns = ["Level", "Type", "Price", "Spread", "Amount (Orig)", "Amount (Adj)", "Age"]
 *         data = []
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_16active_orders_df_lambda(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_16active_orders_df_lambda = {"lambda", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_16active_orders_df_lambda, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_16active_orders_df_lambda(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_x = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_x,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 580, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 580, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "lambda", 0) < (0)) __PYX_ERR(0, 580, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("lambda", 1, 1, 1, i); __PYX_ERR(0, 580, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 580, __pyx_L3_error)
    }
    __pyx_v_x = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 580, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.active_orders_df.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda(__pyx_self, __pyx_v_x);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.active_orders_df.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":574
 *         return df
 * 
 *     def active_orders_df(self) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self.get_price()
*/

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14active_orders_df(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  CYTHON_UNUSED PyObject *__pyx_v_market = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_trading_pair = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_base_asset = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_quote_asset = NULL;
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_v_active_orders = NULL;
  PyObject *__pyx_v_no_sells = NULL;
  PyObject *__pyx_v_columns = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_lvl_buy = NULL;
  PyObject *__pyx_v_lvl_sell = NULL;
  Py_ssize_t __pyx_v_idx;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_v_is_hanging_order = NULL;
  PyObject *__pyx_v_amount_orig = NULL;
  PyObject *__pyx_v_level = NULL;
  PyObject *__pyx_v_level_for_calculation = NULL;
  PyObject *__pyx_v_spread = NULL;
  PyObject *__pyx_v_age = NULL;
  PyObject *__pyx_8genexpr5__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  size_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("active_orders_df", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":575
 * 
 *     def active_orders_df(self) -> pd.DataFrame:
 *         market, trading_pair, base_asset, quote_asset = self._market_info             # <<<<<<<<<<<<<<
 *         price = self.get_price()
 *         active_orders = self.active_orders
*/
  __pyx_t_1 = __pyx_v_self->_market_info;
  __Pyx_INCREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 4)) {
      if (size > 4) __Pyx_RaiseTooManyValuesError(4);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 575, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3);
      __Pyx_INCREF(__pyx_t_5);
    } else {
      __pyx_t_2 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyList_GetItemRefFast(sequence, 2, __Pyx_ReferenceSharing_SharedReference);
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyList_GetItemRefFast(sequence, 3, __Pyx_ReferenceSharing_SharedReference);
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_XGOTREF(__pyx_t_5);
    }
    #else
    {
      Py_ssize_t i;
      PyObject** temps[4] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
      for (i=0; i < 4; i++) {
        PyObject* item = __Pyx_PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 575, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[4] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
    __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
    for (index=0; index < 4; index++) {
      PyObject* item = __pyx_t_7(__pyx_t_6); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 4) < (0)) __PYX_ERR(0, 575, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 575, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_market = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_trading_pair = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_base_asset = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_quote_asset = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":576
 *     def active_orders_df(self) -> pd.DataFrame:
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self.get_price()             # <<<<<<<<<<<<<<
 *         active_orders = self.active_orders
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and
*/
  __pyx_t_5 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_8 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":577
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self.get_price()
 *         active_orders = self.active_orders             # <<<<<<<<<<<<<<
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_active_orders = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":578
 *         price = self.get_price()
 *         active_orders = self.active_orders
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and             # <<<<<<<<<<<<<<
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])
 *         active_orders.sort(key=lambda x: x.price, reverse=True)
*/
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 578, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_v_active_orders)) || PyTuple_CheckExact(__pyx_v_active_orders)) {
      __pyx_t_5 = __pyx_v_active_orders; __Pyx_INCREF(__pyx_t_5);
      __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_active_orders); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 578, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 578, __pyx_L7_error)
    }
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 578, __pyx_L7_error)
            #endif
            if (__pyx_t_9 >= __pyx_temp) break;
          }
          __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_5, __pyx_t_9, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_9;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 578, __pyx_L7_error)
            #endif
            if (__pyx_t_9 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_9));
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_9);
          #endif
          ++__pyx_t_9;
        }
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 578, __pyx_L7_error)
      } else {
        __pyx_t_4 = __pyx_t_10(__pyx_t_5);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 578, __pyx_L7_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_8genexpr5__pyx_v_o, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr5__pyx_v_o, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 578, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 578, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_13 = (!__pyx_t_12);
      if (__pyx_t_13) {
      } else {
        __pyx_t_11 = __pyx_t_13;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr5__pyx_v_o, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 578, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 578, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_13) {
      } else {
        __pyx_t_11 = __pyx_t_13;
        goto __pyx_L11_bool_binop_done;
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":579
 *         active_orders = self.active_orders
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])             # <<<<<<<<<<<<<<
 *         active_orders.sort(key=lambda x: x.price, reverse=True)
 *         columns = ["Level", "Type", "Price", "Spread", "Amount (Orig)", "Amount (Adj)", "Age"]
*/
      __pyx_t_3 = __pyx_v_self->_hanging_orders_tracker;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr5__pyx_v_o, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_2};
        __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_order_id_in_hanging_orders, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 579, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 579, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_12 = (!__pyx_t_13);
      __pyx_t_11 = __pyx_t_12;
      __pyx_L11_bool_binop_done:;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":578
 *         price = self.get_price()
 *         active_orders = self.active_orders
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and             # <<<<<<<<<<<<<<
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])
 *         active_orders.sort(key=lambda x: x.price, reverse=True)
*/
      if (__pyx_t_11) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr5__pyx_v_o))) __PYX_ERR(0, 578, __pyx_L7_error)
      }
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_o); __pyx_8genexpr5__pyx_v_o = 0;
    goto __pyx_L15_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_o); __pyx_8genexpr5__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L15_exit_scope:;
  } /* exit inner scope */
  __pyx_t_9 = __Pyx_PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyLong_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_no_sells = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":580
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])
 *         active_orders.sort(key=lambda x: x.price, reverse=True)             # <<<<<<<<<<<<<<
 *         columns = ["Level", "Type", "Price", "Spread", "Amount (Orig)", "Amount (Adj)", "Age"]
 *         data = []
*/
  __pyx_t_5 = __pyx_v_active_orders;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_16active_orders_df_lambda, 0, __pyx_mstate_global->__pyx_n_u_active_orders_df_locals_lambda, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = 0;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 2 : 0)] = {__pyx_t_5, NULL};
    __pyx_t_2 = __Pyx_MakeVectorcallBuilderKwds(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_key, __pyx_t_4, __pyx_t_2, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 580, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_reverse, Py_True, __pyx_t_2, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 580, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_sort, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":581
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])
 *         active_orders.sort(key=lambda x: x.price, reverse=True)
 *         columns = ["Level", "Type", "Price", "Spread", "Amount (Orig)", "Amount (Adj)", "Age"]             # <<<<<<<<<<<<<<
 *         data = []
 *         lvl_buy, lvl_sell = 0, 0
*/
  __pyx_t_1 = PyList_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_Level);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_Level);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_mstate_global->__pyx_n_u_Level) != (0)) __PYX_ERR(0, 581, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_Type);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_Type);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_n_u_Type) != (0)) __PYX_ERR(0, 581, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_Price);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_Price);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_mstate_global->__pyx_n_u_Price) != (0)) __PYX_ERR(0, 581, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_Spread);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_Spread);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 3, __pyx_mstate_global->__pyx_n_u_Spread) != (0)) __PYX_ERR(0, 581, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Amount_Orig);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Amount_Orig);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 4, __pyx_mstate_global->__pyx_kp_u_Amount_Orig) != (0)) __PYX_ERR(0, 581, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Amount_Adj);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Amount_Adj);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 5, __pyx_mstate_global->__pyx_kp_u_Amount_Adj) != (0)) __PYX_ERR(0, 581, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_Age);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_Age);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 6, __pyx_mstate_global->__pyx_n_u_Age) != (0)) __PYX_ERR(0, 581, __pyx_L1_error);
  __pyx_v_columns = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":582
 *         active_orders.sort(key=lambda x: x.price, reverse=True)
 *         columns = ["Level", "Type", "Price", "Spread", "Amount (Orig)", "Amount (Adj)", "Age"]
 *         data = []             # <<<<<<<<<<<<<<
 *         lvl_buy, lvl_sell = 0, 0
 *         for idx in range(0, len(active_orders)):
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":583
 *         columns = ["Level", "Type", "Price", "Spread", "Amount (Orig)", "Amount (Adj)", "Age"]
 *         data = []
 *         lvl_buy, lvl_sell = 0, 0             # <<<<<<<<<<<<<<
 *         for idx in range(0, len(active_orders)):
 *             order = active_orders[idx]
*/
  __pyx_t_1 = __pyx_mstate_global->__pyx_int_0;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_mstate_global->__pyx_int_0;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_lvl_buy = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_lvl_sell = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":584
 *         data = []
 *         lvl_buy, lvl_sell = 0, 0
 *         for idx in range(0, len(active_orders)):             # <<<<<<<<<<<<<<
 *             order = active_orders[idx]
 *             is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)
*/
  __pyx_t_9 = PyObject_Length(__pyx_v_active_orders); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 584, __pyx_L1_error)
  __pyx_t_14 = __pyx_t_9;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
    __pyx_v_idx = __pyx_t_15;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":585
 *         lvl_buy, lvl_sell = 0, 0
 *         for idx in range(0, len(active_orders)):
 *             order = active_orders[idx]             # <<<<<<<<<<<<<<
 *             is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)
 *             amount_orig = ""
*/
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_active_orders, __pyx_v_idx, Py_ssize_t, 1, PyLong_FromSsize_t, 0, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 585, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":586
 *         for idx in range(0, len(active_orders)):
 *             order = active_orders[idx]
 *             is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)             # <<<<<<<<<<<<<<
 *             amount_orig = ""
 *             if not is_hanging_order:
*/
    __pyx_t_1 = __pyx_v_self->_hanging_orders_tracker;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_order_id_in_hanging_orders, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_is_hanging_order, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":587
 *             order = active_orders[idx]
 *             is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)
 *             amount_orig = ""             # <<<<<<<<<<<<<<
 *             if not is_hanging_order:
 *                 if order.is_buy:
*/
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__15);
    __Pyx_XDECREF_SET(__pyx_v_amount_orig, __pyx_mstate_global->__pyx_kp_u__15);

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":588
 *             is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)
 *             amount_orig = ""
 *             if not is_hanging_order:             # <<<<<<<<<<<<<<
 *                 if order.is_buy:
 *                     level = lvl_buy + 1
*/
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_is_hanging_order); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 588, __pyx_L1_error)
    __pyx_t_12 = (!__pyx_t_11);
    if (__pyx_t_12) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":589
 *             amount_orig = ""
 *             if not is_hanging_order:
 *                 if order.is_buy:             # <<<<<<<<<<<<<<
 *                     level = lvl_buy + 1
 *                     lvl_buy += 1
*/
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 589, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 589, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_12) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":590
 *             if not is_hanging_order:
 *                 if order.is_buy:
 *                     level = lvl_buy + 1             # <<<<<<<<<<<<<<
 *                     lvl_buy += 1
 *                 else:
*/
        __pyx_t_2 = __Pyx_PyLong_AddObjC(__pyx_v_lvl_buy, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 590, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_level, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":591
 *                 if order.is_buy:
 *                     level = lvl_buy + 1
 *                     lvl_buy += 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     level = no_sells - lvl_sell
*/
        __pyx_t_2 = __Pyx_PyLong_AddObjC(__pyx_v_lvl_buy, __pyx_mstate_global->__pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_lvl_buy, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":589
 *             amount_orig = ""
 *             if not is_hanging_order:
 *                 if order.is_buy:             # <<<<<<<<<<<<<<
 *                     level = lvl_buy + 1
 *                     lvl_buy += 1
*/
        goto __pyx_L19;
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":593
 *                     lvl_buy += 1
 *                 else:
 *                     level = no_sells - lvl_sell             # <<<<<<<<<<<<<<
 *                     lvl_sell += 1
 *                 amount_orig = self._order_amount + ((level - 1) * self._order_level_amount)
*/
      /*else*/ {
        __pyx_t_2 = PyNumber_Subtract(__pyx_v_no_sells, __pyx_v_lvl_sell); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 593, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_level, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":594
 *                 else:
 *                     level = no_sells - lvl_sell
 *                     lvl_sell += 1             # <<<<<<<<<<<<<<
 *                 amount_orig = self._order_amount + ((level - 1) * self._order_level_amount)
 *             else:
*/
        __pyx_t_2 = __Pyx_PyLong_AddObjC(__pyx_v_lvl_sell, __pyx_mstate_global->__pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 594, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_lvl_sell, __pyx_t_2);
        __pyx_t_2 = 0;
      }
      __pyx_L19:;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":595
 *                     level = no_sells - lvl_sell
 *                     lvl_sell += 1
 *                 amount_orig = self._order_amount + ((level - 1) * self._order_level_amount)             # <<<<<<<<<<<<<<
 *             else:
 *                 level_for_calculation = lvl_buy if order.is_buy else lvl_sell
*/
      __pyx_t_2 = __Pyx_PyLong_SubtractObjC(__pyx_v_level, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 595, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PyNumber_Multiply(__pyx_t_2, __pyx_v_self->_order_level_amount); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 595, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Add(__pyx_v_self->_order_amount, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 595, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_amount_orig, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":588
 *             is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)
 *             amount_orig = ""
 *             if not is_hanging_order:             # <<<<<<<<<<<<<<
 *                 if order.is_buy:
 *                     level = lvl_buy + 1
*/
      goto __pyx_L18;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":597
 *                 amount_orig = self._order_amount + ((level - 1) * self._order_level_amount)
 *             else:
 *                 level_for_calculation = lvl_buy if order.is_buy else lvl_sell             # <<<<<<<<<<<<<<
 *                 amount_orig = self._order_amount + ((level_for_calculation - 1) * self._order_level_amount)
 *                 level = "hang"
*/
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 597, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 597, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_12) {
        __Pyx_INCREF(__pyx_v_lvl_buy);
        __pyx_t_2 = __pyx_v_lvl_buy;
      } else {
        __Pyx_INCREF(__pyx_v_lvl_sell);
        __pyx_t_2 = __pyx_v_lvl_sell;
      }
      __Pyx_XDECREF_SET(__pyx_v_level_for_calculation, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":598
 *             else:
 *                 level_for_calculation = lvl_buy if order.is_buy else lvl_sell
 *                 amount_orig = self._order_amount + ((level_for_calculation - 1) * self._order_level_amount)             # <<<<<<<<<<<<<<
 *                 level = "hang"
 *             spread = 0 if price == 0 else abs(order.price - price)/price
*/
      __pyx_t_2 = __Pyx_PyLong_SubtractObjC(__pyx_v_level_for_calculation, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PyNumber_Multiply(__pyx_t_2, __pyx_v_self->_order_level_amount); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Add(__pyx_v_self->_order_amount, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_amount_orig, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":599
 *                 level_for_calculation = lvl_buy if order.is_buy else lvl_sell
 *                 amount_orig = self._order_amount + ((level_for_calculation - 1) * self._order_level_amount)
 *                 level = "hang"             # <<<<<<<<<<<<<<
 *             spread = 0 if price == 0 else abs(order.price - price)/price
 *             age = pd.Timestamp(order_age(order, self._current_timestamp), unit='s').strftime('%H:%M:%S')
*/
      __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_hang);
      __Pyx_XDECREF_SET(__pyx_v_level, __pyx_mstate_global->__pyx_n_u_hang);
    }
    __pyx_L18:;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":600
 *                 amount_orig = self._order_amount + ((level_for_calculation - 1) * self._order_level_amount)
 *                 level = "hang"
 *             spread = 0 if price == 0 else abs(order.price - price)/price             # <<<<<<<<<<<<<<
 *             age = pd.Timestamp(order_age(order, self._current_timestamp), unit='s').strftime('%H:%M:%S')
 *             data.append([
*/
    __pyx_t_12 = (__Pyx_PyLong_BoolEqObjC(__pyx_v_price, __pyx_mstate_global->__pyx_int_0, 0, 0)); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 600, __pyx_L1_error)
    if (__pyx_t_12) {
      __Pyx_INCREF(__pyx_mstate_global->__pyx_int_0);
      __pyx_t_2 = __pyx_mstate_global->__pyx_int_0;
    } else {
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 600, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyNumber_Subtract(__pyx_t_4, __pyx_v_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 600, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyNumber_Absolute(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 600, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_4, __pyx_v_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 600, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = __pyx_t_1;
      __pyx_t_1 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_spread, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":601
 *                 level = "hang"
 *             spread = 0 if price == 0 else abs(order.price - price)/price
 *             age = pd.Timestamp(order_age(order, self._current_timestamp), unit='s').strftime('%H:%M:%S')             # <<<<<<<<<<<<<<
 *             data.append([
 *                 level,
*/
    __pyx_t_5 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_pd); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_Timestamp); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_16 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_order_age); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_18 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_8 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_17))) {
      __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_17);
      assert(__pyx_t_16);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_17);
      __Pyx_INCREF(__pyx_t_16);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_17, __pyx__function);
      __pyx_t_8 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_16, __pyx_v_order, __pyx_t_18};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_8, (3-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 601, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_t_8 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      assert(__pyx_t_5);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
      __pyx_t_8 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_5, __pyx_t_3};
      __pyx_t_17 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 601, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_unit, __pyx_mstate_global->__pyx_n_u_s, __pyx_t_17, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 601, __pyx_L1_error)
      __pyx_t_4 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_17);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 601, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __pyx_t_1 = __pyx_t_4;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_8 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_H_M_S};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_strftime, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 601, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_age, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":604
 *             data.append([
 *                 level,
 *                 "buy" if order.is_buy else "sell",             # <<<<<<<<<<<<<<
 *                 float(order.price),
 *                 f"{spread:.2%}",
*/
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 604, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 604, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_12) {
      __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_buy);
      __pyx_t_2 = __pyx_mstate_global->__pyx_n_u_buy;
    } else {
      __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_sell);
      __pyx_t_2 = __pyx_mstate_global->__pyx_n_u_sell;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":605
 *                 level,
 *                 "buy" if order.is_buy else "sell",
 *                 float(order.price),             # <<<<<<<<<<<<<<
 *                 f"{spread:.2%}",
 *                 amount_orig,
*/
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":606
 *                 "buy" if order.is_buy else "sell",
 *                 float(order.price),
 *                 f"{spread:.2%}",             # <<<<<<<<<<<<<<
 *                 amount_orig,
 *                 float(order.quantity),
*/
    __pyx_t_4 = __Pyx_PyObject_Format(__pyx_v_spread, __pyx_mstate_global->__pyx_kp_u_2_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 606, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":608
 *                 f"{spread:.2%}",
 *                 amount_orig,
 *                 float(order.quantity),             # <<<<<<<<<<<<<<
 *                 age
 *             ])
*/
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_quantity); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 608, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_17 = __Pyx_PyNumber_Float(__pyx_t_6); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 608, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":602
 *             spread = 0 if price == 0 else abs(order.price - price)/price
 *             age = pd.Timestamp(order_age(order, self._current_timestamp), unit='s').strftime('%H:%M:%S')
 *             data.append([             # <<<<<<<<<<<<<<
 *                 level,
 *                 "buy" if order.is_buy else "sell",
*/
    __pyx_t_6 = PyList_New(7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_level);
    __Pyx_GIVEREF(__pyx_v_level);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 0, __pyx_v_level) != (0)) __PYX_ERR(0, 602, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 1, __pyx_t_2) != (0)) __PYX_ERR(0, 602, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 2, __pyx_t_1) != (0)) __PYX_ERR(0, 602, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 3, __pyx_t_4) != (0)) __PYX_ERR(0, 602, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_amount_orig);
    __Pyx_GIVEREF(__pyx_v_amount_orig);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 4, __pyx_v_amount_orig) != (0)) __PYX_ERR(0, 602, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_17);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 5, __pyx_t_17) != (0)) __PYX_ERR(0, 602, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_age);
    __Pyx_GIVEREF(__pyx_v_age);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 6, __pyx_v_age) != (0)) __PYX_ERR(0, 602, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_t_17 = 0;
    __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_data, __pyx_t_6); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 602, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":612
 *             ])
 * 
 *         return pd.DataFrame(data=data, columns=columns)             # <<<<<<<<<<<<<<
 * 
 *     def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_17 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pd); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DataFrame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_8 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_1);
    assert(__pyx_t_17);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_17);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
    __pyx_t_8 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 2 : 0)] = {__pyx_t_17, NULL};
    __pyx_t_4 = __Pyx_MakeVectorcallBuilderKwds(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_data, __pyx_v_data, __pyx_t_4, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 612, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_columns, __pyx_v_columns, __pyx_t_4, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 612, __pyx_L1_error)
    __pyx_t_6 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
  }
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":574
 *         return df
 * 
 *     def active_orders_df(self) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self.get_price()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.active_orders_df", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_trading_pair);
  __Pyx_XDECREF(__pyx_v_base_asset);
  __Pyx_XDECREF(__pyx_v_quote_asset);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_active_orders);
  __Pyx_XDECREF(__pyx_v_no_sells);
  __Pyx_XDECREF(__pyx_v_columns);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_lvl_buy);
  __Pyx_XDECREF(__pyx_v_lvl_sell);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XDECREF(__pyx_v_is_hanging_order);
  __Pyx_XDECREF(__pyx_v_amount_orig);
  __Pyx_XDECREF(__pyx_v_level);
  __Pyx_XDECREF(__pyx_v_level_for_calculation);
  __Pyx_XDECREF(__pyx_v_spread);
  __Pyx_XDECREF(__pyx_v_age);
  __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":614
 *         return pd.DataFrame(data=data, columns=columns)
 * 
 *     def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         markets_data = []
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"Ref Price ({self._price_type.name})"]
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17market_status_data_frame(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17market_status_data_frame = {"market_status_data_frame", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17market_status_data_frame, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17market_status_data_frame(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_market_trading_pair_tuples = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("market_status_data_frame (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_market_trading_pair_tuples,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 614, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 614, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "market_status_data_frame", 0) < (0)) __PYX_ERR(0, 614, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("market_status_data_frame", 1, 1, 1, i); __PYX_ERR(0, 614, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 614, __pyx_L3_error)
    }
    __pyx_v_market_trading_pair_tuples = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("market_status_data_frame", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 614, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.market_status_data_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_16market_status_data_frame(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), __pyx_v_market_trading_pair_tuples);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_16market_status_data_frame(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_market_trading_pair_tuples) {
  PyObject *__pyx_v_markets_data = NULL;
  PyObject *__pyx_v_markets_columns = NULL;
  PyObject *__pyx_v_market_books = NULL;
  PyObject *__pyx_v_market = NULL;
  PyObject *__pyx_v_trading_pair = NULL;
  PyObject *__pyx_v_bid_price = NULL;
  PyObject *__pyx_v_ask_price = NULL;
  PyObject *__pyx_v_ref_price = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3[3];
  int __pyx_t_4;
  int __pyx_t_5;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  double __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("market_status_data_frame", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":615
 * 
 *     def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:
 *         markets_data = []             # <<<<<<<<<<<<<<
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"Ref Price ({self._price_type.name})"]
 *         if self._price_type is PriceType.LastOwnTrade and self._last_own_trade_price.is_nan():
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_markets_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":616
 *     def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:
 *         markets_data = []
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"Ref Price ({self._price_type.name})"]             # <<<<<<<<<<<<<<
 *         if self._price_type is PriceType.LastOwnTrade and self._last_own_trade_price.is_nan():
 *             markets_columns[-1] = "Ref Price (MidPrice)"
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_price_type, __pyx_mstate_global->__pyx_n_u_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3[0] = __pyx_mstate_global->__pyx_kp_u_Ref_Price;
  __pyx_t_3[1] = __pyx_t_2;
  __pyx_t_3[2] = __pyx_mstate_global->__pyx_kp_u__13;
  __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, 11 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2) + 1, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2));
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_Exchange);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_Exchange);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_mstate_global->__pyx_n_u_Exchange) != (0)) __PYX_ERR(0, 616, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_Market);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_Market);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_mstate_global->__pyx_n_u_Market) != (0)) __PYX_ERR(0, 616, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Best_Bid);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Best_Bid);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 2, __pyx_mstate_global->__pyx_kp_u_Best_Bid) != (0)) __PYX_ERR(0, 616, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Best_Ask);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Best_Ask);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 3, __pyx_mstate_global->__pyx_kp_u_Best_Ask) != (0)) __PYX_ERR(0, 616, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 4, __pyx_t_1) != (0)) __PYX_ERR(0, 616, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_markets_columns = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":617
 *         markets_data = []
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"Ref Price ({self._price_type.name})"]
 *         if self._price_type is PriceType.LastOwnTrade and self._last_own_trade_price.is_nan():             # <<<<<<<<<<<<<<
 *             markets_columns[-1] = "Ref Price (MidPrice)"
 *         market_books = [(self._market_info.market, self._market_info.trading_pair)]
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PriceType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_LastOwnTrade); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_v_self->_price_type == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __pyx_v_self->_last_own_trade_price;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_nan, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":618
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"Ref Price ({self._price_type.name})"]
 *         if self._price_type is PriceType.LastOwnTrade and self._last_own_trade_price.is_nan():
 *             markets_columns[-1] = "Ref Price (MidPrice)"             # <<<<<<<<<<<<<<
 *         market_books = [(self._market_info.market, self._market_info.trading_pair)]
 *         if type(self._asset_price_delegate) is OrderBookAssetPriceDelegate:
*/
    if (unlikely((__Pyx_SetItemInt(__pyx_v_markets_columns, -1L, __pyx_mstate_global->__pyx_kp_u_Ref_Price_MidPrice, long, 1, __Pyx_PyLong_From_long, 1, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference) < 0))) __PYX_ERR(0, 618, __pyx_L1_error)

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":617
 *         markets_data = []
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"Ref Price ({self._price_type.name})"]
 *         if self._price_type is PriceType.LastOwnTrade and self._last_own_trade_price.is_nan():             # <<<<<<<<<<<<<<
 *             markets_columns[-1] = "Ref Price (MidPrice)"
 *         market_books = [(self._market_info.market, self._market_info.trading_pair)]
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":619
 *         if self._price_type is PriceType.LastOwnTrade and self._last_own_trade_price.is_nan():
 *             markets_columns[-1] = "Ref Price (MidPrice)"
 *         market_books = [(self._market_info.market, self._market_info.trading_pair)]             # <<<<<<<<<<<<<<
 *         if type(self._asset_price_delegate) is OrderBookAssetPriceDelegate:
 *             market_books.append((self._asset_price_delegate.market, self._asset_price_delegate.trading_pair))
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1) != (0)) __PYX_ERR(0, 619, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2) != (0)) __PYX_ERR(0, 619, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_7) != (0)) __PYX_ERR(0, 619, __pyx_L1_error);
  __pyx_t_7 = 0;
  __pyx_v_market_books = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":620
 *             markets_columns[-1] = "Ref Price (MidPrice)"
 *         market_books = [(self._market_info.market, self._market_info.trading_pair)]
 *         if type(self._asset_price_delegate) is OrderBookAssetPriceDelegate:             # <<<<<<<<<<<<<<
 *             market_books.append((self._asset_price_delegate.market, self._asset_price_delegate.trading_pair))
 *         for market, trading_pair in market_books:
*/
  __pyx_t_4 = (((PyObject *)Py_TYPE(__pyx_v_self->_asset_price_delegate)) == ((PyObject *)__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate));
  if (__pyx_t_4) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":621
 *         market_books = [(self._market_info.market, self._market_info.trading_pair)]
 *         if type(self._asset_price_delegate) is OrderBookAssetPriceDelegate:
 *             market_books.append((self._asset_price_delegate.market, self._asset_price_delegate.trading_pair))             # <<<<<<<<<<<<<<
 *         for market, trading_pair in market_books:
 *             bid_price = market.get_price(trading_pair, False)
*/
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_asset_price_delegate, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_asset_price_delegate, __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2) != (0)) __PYX_ERR(0, 621, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_7) != (0)) __PYX_ERR(0, 621, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_7 = 0;
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_market_books, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":620
 *             markets_columns[-1] = "Ref Price (MidPrice)"
 *         market_books = [(self._market_info.market, self._market_info.trading_pair)]
 *         if type(self._asset_price_delegate) is OrderBookAssetPriceDelegate:             # <<<<<<<<<<<<<<
 *             market_books.append((self._asset_price_delegate.market, self._asset_price_delegate.trading_pair))
 *         for market, trading_pair in market_books:
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":622
 *         if type(self._asset_price_delegate) is OrderBookAssetPriceDelegate:
 *             market_books.append((self._asset_price_delegate.market, self._asset_price_delegate.trading_pair))
 *         for market, trading_pair in market_books:             # <<<<<<<<<<<<<<
 *             bid_price = market.get_price(trading_pair, False)
 *             ask_price = market.get_price(trading_pair, True)
*/
  __pyx_t_1 = __pyx_v_market_books; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_9 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 622, __pyx_L1_error)
      #endif
      if (__pyx_t_9 >= __pyx_temp) break;
    }
    __pyx_t_7 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_9, __Pyx_ReferenceSharing_OwnStrongReference);
    ++__pyx_t_9;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
      PyObject* sequence = __pyx_t_7;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 622, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1);
        __Pyx_INCREF(__pyx_t_10);
      } else {
        __pyx_t_2 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 622, __pyx_L1_error)
        __Pyx_XGOTREF(__pyx_t_2);
        __pyx_t_10 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
        if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 622, __pyx_L1_error)
        __Pyx_XGOTREF(__pyx_t_10);
      }
      #else
      __pyx_t_2 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_11 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_12 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_11);
      index = 0; __pyx_t_2 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_2)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_11), 2) < (0)) __PYX_ERR(0, 622, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 622, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_market, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_trading_pair, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":623
 *             market_books.append((self._asset_price_delegate.market, self._asset_price_delegate.trading_pair))
 *         for market, trading_pair in market_books:
 *             bid_price = market.get_price(trading_pair, False)             # <<<<<<<<<<<<<<
 *             ask_price = market.get_price(trading_pair, True)
 *             ref_price = float("nan")
*/
    __pyx_t_10 = __pyx_v_market;
    __Pyx_INCREF(__pyx_t_10);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_10, __pyx_v_trading_pair, Py_False};
      __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 623, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_bid_price, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":624
 *         for market, trading_pair in market_books:
 *             bid_price = market.get_price(trading_pair, False)
 *             ask_price = market.get_price(trading_pair, True)             # <<<<<<<<<<<<<<
 *             ref_price = float("nan")
 *             if market == self._market_info.market and self._inventory_cost_price_delegate is not None:
*/
    __pyx_t_10 = __pyx_v_market;
    __Pyx_INCREF(__pyx_t_10);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_10, __pyx_v_trading_pair, Py_True};
      __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 624, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_ask_price, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":625
 *             bid_price = market.get_price(trading_pair, False)
 *             ask_price = market.get_price(trading_pair, True)
 *             ref_price = float("nan")             # <<<<<<<<<<<<<<
 *             if market == self._market_info.market and self._inventory_cost_price_delegate is not None:
 *                 # We're using inventory_cost, show it's price
*/
    __pyx_t_13 = __Pyx_PyUnicode_AsDouble(__pyx_mstate_global->__pyx_n_u_nan); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_13, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 625, __pyx_L1_error)
    __pyx_t_7 = PyFloat_FromDouble(__pyx_t_13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 625, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_ref_price, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":626
 *             ask_price = market.get_price(trading_pair, True)
 *             ref_price = float("nan")
 *             if market == self._market_info.market and self._inventory_cost_price_delegate is not None:             # <<<<<<<<<<<<<<
 *                 # We're using inventory_cost, show it's price
 *                 ref_price = self._inventory_cost_price_delegate.get_price()
*/
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = PyObject_RichCompare(__pyx_v_market, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 626, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_inventory_cost_price_delegate != Py_None);
    __pyx_t_4 = __pyx_t_5;
    __pyx_L12_bool_binop_done:;
    if (__pyx_t_4) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":628
 *             if market == self._market_info.market and self._inventory_cost_price_delegate is not None:
 *                 # We're using inventory_cost, show it's price
 *                 ref_price = self._inventory_cost_price_delegate.get_price()             # <<<<<<<<<<<<<<
 *                 if ref_price is None:
 *                     ref_price = self.get_price()
*/
      __pyx_t_7 = __pyx_v_self->_inventory_cost_price_delegate;
      __Pyx_INCREF(__pyx_t_7);
      __pyx_t_6 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
        __pyx_t_10 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 628, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
      }
      __Pyx_DECREF_SET(__pyx_v_ref_price, __pyx_t_10);
      __pyx_t_10 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":629
 *                 # We're using inventory_cost, show it's price
 *                 ref_price = self._inventory_cost_price_delegate.get_price()
 *                 if ref_price is None:             # <<<<<<<<<<<<<<
 *                     ref_price = self.get_price()
 *             elif market == self._market_info.market and self._asset_price_delegate is None:
*/
      __pyx_t_4 = (__pyx_v_ref_price == Py_None);
      if (__pyx_t_4) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":630
 *                 ref_price = self._inventory_cost_price_delegate.get_price()
 *                 if ref_price is None:
 *                     ref_price = self.get_price()             # <<<<<<<<<<<<<<
 *             elif market == self._market_info.market and self._asset_price_delegate is None:
 *                 ref_price = self.get_price()
*/
        __pyx_t_7 = ((PyObject *)__pyx_v_self);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_t_6 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
          __pyx_t_10 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 630, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
        }
        __Pyx_DECREF_SET(__pyx_v_ref_price, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":629
 *                 # We're using inventory_cost, show it's price
 *                 ref_price = self._inventory_cost_price_delegate.get_price()
 *                 if ref_price is None:             # <<<<<<<<<<<<<<
 *                     ref_price = self.get_price()
 *             elif market == self._market_info.market and self._asset_price_delegate is None:
*/
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":626
 *             ask_price = market.get_price(trading_pair, True)
 *             ref_price = float("nan")
 *             if market == self._market_info.market and self._inventory_cost_price_delegate is not None:             # <<<<<<<<<<<<<<
 *                 # We're using inventory_cost, show it's price
 *                 ref_price = self._inventory_cost_price_delegate.get_price()
*/
      goto __pyx_L11;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":631
 *                 if ref_price is None:
 *                     ref_price = self.get_price()
 *             elif market == self._market_info.market and self._asset_price_delegate is None:             # <<<<<<<<<<<<<<
 *                 ref_price = self.get_price()
 *             elif (
*/
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 631, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = PyObject_RichCompare(__pyx_v_market, __pyx_t_10, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 631, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 631, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_asset_price_delegate == Py_None);
    __pyx_t_4 = __pyx_t_5;
    __pyx_L15_bool_binop_done:;
    if (__pyx_t_4) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":632
 *                     ref_price = self.get_price()
 *             elif market == self._market_info.market and self._asset_price_delegate is None:
 *                 ref_price = self.get_price()             # <<<<<<<<<<<<<<
 *             elif (
 *                 self._asset_price_delegate is not None
*/
      __pyx_t_10 = ((PyObject *)__pyx_v_self);
      __Pyx_INCREF(__pyx_t_10);
      __pyx_t_6 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
        __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 632, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_DECREF_SET(__pyx_v_ref_price, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":631
 *                 if ref_price is None:
 *                     ref_price = self.get_price()
 *             elif market == self._market_info.market and self._asset_price_delegate is None:             # <<<<<<<<<<<<<<
 *                 ref_price = self.get_price()
 *             elif (
*/
      goto __pyx_L11;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":634
 *                 ref_price = self.get_price()
 *             elif (
 *                 self._asset_price_delegate is not None             # <<<<<<<<<<<<<<
 *                 and market == self._asset_price_delegate.market
 *                 and self._price_type is not PriceType.LastOwnTrade
*/
    __pyx_t_5 = (__pyx_v_self->_asset_price_delegate != Py_None);
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L17_bool_binop_done;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":635
 *             elif (
 *                 self._asset_price_delegate is not None
 *                 and market == self._asset_price_delegate.market             # <<<<<<<<<<<<<<
 *                 and self._price_type is not PriceType.LastOwnTrade
 *             ):
*/
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_asset_price_delegate, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = PyObject_RichCompare(__pyx_v_market, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L17_bool_binop_done;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":636
 *                 self._asset_price_delegate is not None
 *                 and market == self._asset_price_delegate.market
 *                 and self._price_type is not PriceType.LastOwnTrade             # <<<<<<<<<<<<<<
 *             ):
 *                 ref_price = self._asset_price_delegate.get_price_by_type(self._price_type)
*/
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_PriceType); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_LastOwnTrade); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_5 = (__pyx_v_self->_price_type != __pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_4 = __pyx_t_5;
    __pyx_L17_bool_binop_done:;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":633
 *             elif market == self._market_info.market and self._asset_price_delegate is None:
 *                 ref_price = self.get_price()
 *             elif (             # <<<<<<<<<<<<<<
 *                 self._asset_price_delegate is not None
 *                 and market == self._asset_price_delegate.market
*/
    if (__pyx_t_4) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":638
 *                 and self._price_type is not PriceType.LastOwnTrade
 *             ):
 *                 ref_price = self._asset_price_delegate.get_price_by_type(self._price_type)             # <<<<<<<<<<<<<<
 *             markets_data.append([
 *                 market.display_name,
*/
      __pyx_t_10 = __pyx_v_self->_asset_price_delegate;
      __Pyx_INCREF(__pyx_t_10);
      __pyx_t_6 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_v_self->_price_type};
        __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price_by_type, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 638, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_DECREF_SET(__pyx_v_ref_price, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":633
 *             elif market == self._market_info.market and self._asset_price_delegate is None:
 *                 ref_price = self.get_price()
 *             elif (             # <<<<<<<<<<<<<<
 *                 self._asset_price_delegate is not None
 *                 and market == self._asset_price_delegate.market
*/
    }
    __pyx_L11:;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":640
 *                 ref_price = self._asset_price_delegate.get_price_by_type(self._price_type)
 *             markets_data.append([
 *                 market.display_name,             # <<<<<<<<<<<<<<
 *                 trading_pair,
 *                 float(bid_price),
*/
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_market, __pyx_mstate_global->__pyx_n_u_display_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":642
 *                 market.display_name,
 *                 trading_pair,
 *                 float(bid_price),             # <<<<<<<<<<<<<<
 *                 float(ask_price),
 *                 float(ref_price)
*/
    __pyx_t_10 = __Pyx_PyNumber_Float(__pyx_v_bid_price); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 642, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":643
 *                 trading_pair,
 *                 float(bid_price),
 *                 float(ask_price),             # <<<<<<<<<<<<<<
 *                 float(ref_price)
 *             ])
*/
    __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_v_ask_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":644
 *                 float(bid_price),
 *                 float(ask_price),
 *                 float(ref_price)             # <<<<<<<<<<<<<<
 *             ])
 *         return pd.DataFrame(data=markets_data, columns=markets_columns).replace(np.nan, '', regex=True)
*/
    __pyx_t_11 = __Pyx_PyNumber_Float(__pyx_v_ref_price); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":639
 *             ):
 *                 ref_price = self._asset_price_delegate.get_price_by_type(self._price_type)
 *             markets_data.append([             # <<<<<<<<<<<<<<
 *                 market.display_name,
 *                 trading_pair,
*/
    __pyx_t_14 = PyList_New(5); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 0, __pyx_t_7) != (0)) __PYX_ERR(0, 639, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_trading_pair);
    __Pyx_GIVEREF(__pyx_v_trading_pair);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 1, __pyx_v_trading_pair) != (0)) __PYX_ERR(0, 639, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_10);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 2, __pyx_t_10) != (0)) __PYX_ERR(0, 639, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 3, __pyx_t_2) != (0)) __PYX_ERR(0, 639, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_11);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 4, __pyx_t_11) != (0)) __PYX_ERR(0, 639, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_10 = 0;
    __pyx_t_2 = 0;
    __pyx_t_11 = 0;
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_markets_data, __pyx_t_14); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":622
 *         if type(self._asset_price_delegate) is OrderBookAssetPriceDelegate:
 *             market_books.append((self._asset_price_delegate.market, self._asset_price_delegate.trading_pair))
 *         for market, trading_pair in market_books:             # <<<<<<<<<<<<<<
 *             bid_price = market.get_price(trading_pair, False)
 *             ask_price = market.get_price(trading_pair, True)
*/
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":646
 *                 float(ref_price)
 *             ])
 *         return pd.DataFrame(data=markets_data, columns=markets_columns).replace(np.nan, '', regex=True)             # <<<<<<<<<<<<<<
 * 
 *     def format_status(self) -> str:
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_pd); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_DataFrame); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 2 : 0)] = {__pyx_t_2, NULL};
    __pyx_t_10 = __Pyx_MakeVectorcallBuilderKwds(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_data, __pyx_v_markets_data, __pyx_t_10, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 646, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_columns, __pyx_v_markets_columns, __pyx_t_10, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 646, __pyx_L1_error)
    __pyx_t_11 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
  }
  __pyx_t_14 = __pyx_t_11;
  __Pyx_INCREF(__pyx_t_14);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_nan); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_14, __pyx_t_10, __pyx_mstate_global->__pyx_kp_u__15};
    __pyx_t_7 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_regex, Py_True, __pyx_t_7, __pyx_callargs+3, 0) < (0)) __PYX_ERR(0, 646, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_replace, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":614
 *         return pd.DataFrame(data=data, columns=columns)
 * 
 *     def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         markets_data = []
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"Ref Price ({self._price_type.name})"]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.market_status_data_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_markets_data);
  __Pyx_XDECREF(__pyx_v_markets_columns);
  __Pyx_XDECREF(__pyx_v_market_books);
  __Pyx_XDECREF(__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_trading_pair);
  __Pyx_XDECREF(__pyx_v_bid_price);
  __Pyx_XDECREF(__pyx_v_ask_price);
  __Pyx_XDECREF(__pyx_v_ref_price);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":648
 *         return pd.DataFrame(data=markets_data, columns=markets_columns).replace(np.nan, '', regex=True)
 * 
 *     def format_status(self) -> str:             # <<<<<<<<<<<<<<
 *         if not self._all_markets_ready:
 *             return "Market connectors are not ready."
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_19format_status(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_19format_status = {"format_status", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_19format_status, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_19format_status(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("format_status (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("format_status", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("format_status", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18format_status(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18format_status(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_lines = 0;
  PyObject *__pyx_v_warning_lines = 0;
  PyObject *__pyx_v_markets_df = NULL;
  PyObject *__pyx_v_assets_df = NULL;
  PyObject *__pyx_v_inventory_skew_df = NULL;
  PyObject *__pyx_v_first_col_length = NULL;
  PyObject *__pyx_v_df_lines = NULL;
  PyObject *__pyx_v_df = NULL;
  PyObject *__pyx_8genexpr6__pyx_v_line = NULL;
  PyObject *__pyx_8genexpr7__pyx_v_line = NULL;
  PyObject *__pyx_8genexpr8__pyx_v_line = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("format_status", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":649
 * 
 *     def format_status(self) -> str:
 *         if not self._all_markets_ready:             # <<<<<<<<<<<<<<
 *             return "Market connectors are not ready."
 *         cdef:
*/
  __pyx_t_1 = (!__pyx_v_self->_all_markets_ready);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":650
 *     def format_status(self) -> str:
 *         if not self._all_markets_ready:
 *             return "Market connectors are not ready."             # <<<<<<<<<<<<<<
 *         cdef:
 *             list lines = []
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Market_connectors_are_not_ready);
    __pyx_r = __pyx_mstate_global->__pyx_kp_u_Market_connectors_are_not_ready;
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":649
 * 
 *     def format_status(self) -> str:
 *         if not self._all_markets_ready:             # <<<<<<<<<<<<<<
 *             return "Market connectors are not ready."
 *         cdef:
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":652
 *             return "Market connectors are not ready."
 *         cdef:
 *             list lines = []             # <<<<<<<<<<<<<<
 *             list warning_lines = []
 *         warning_lines.extend(self._ping_pong_warning_lines)
*/
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_lines = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":653
 *         cdef:
 *             list lines = []
 *             list warning_lines = []             # <<<<<<<<<<<<<<
 *         warning_lines.extend(self._ping_pong_warning_lines)
 *         warning_lines.extend(self.network_warning([self._market_info]))
*/
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_warning_lines = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":654
 *             list lines = []
 *             list warning_lines = []
 *         warning_lines.extend(self._ping_pong_warning_lines)             # <<<<<<<<<<<<<<
 *         warning_lines.extend(self.network_warning([self._market_info]))
 * 
*/
  __pyx_t_2 = __pyx_v_self->_ping_pong_warning_lines;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyList_Extend(__pyx_v_warning_lines, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":655
 *             list warning_lines = []
 *         warning_lines.extend(self._ping_pong_warning_lines)
 *         warning_lines.extend(self.network_warning([self._market_info]))             # <<<<<<<<<<<<<<
 * 
 *         markets_df = map_df_to_str(self.market_status_data_frame([self._market_info]))
*/
  __pyx_t_4 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_self->_market_info);
  __Pyx_GIVEREF(__pyx_v_self->_market_info);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_self->_market_info) != (0)) __PYX_ERR(0, 655, __pyx_L1_error);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_network_warning, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 655, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_3 = __Pyx_PyList_Extend(__pyx_v_warning_lines, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":657
 *         warning_lines.extend(self.network_warning([self._market_info]))
 * 
 *         markets_df = map_df_to_str(self.market_status_data_frame([self._market_info]))             # <<<<<<<<<<<<<<
 *         lines.extend(["", "  Markets:"] + ["    " + line for line in markets_df.to_string(index=False).split("\n")])
 * 
*/
  __pyx_t_5 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_map_df_to_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_8);
  __pyx_t_9 = PyList_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_v_self->_market_info);
  __Pyx_GIVEREF(__pyx_v_self->_market_info);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 0, __pyx_v_self->_market_info) != (0)) __PYX_ERR(0, 657, __pyx_L1_error);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_9};
    __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_market_status_data_frame, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
  }
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_5);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_7};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_markets_df = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":658
 * 
 *         markets_df = map_df_to_str(self.market_status_data_frame([self._market_info]))
 *         lines.extend(["", "  Markets:"] + ["    " + line for line in markets_df.to_string(index=False).split("\n")])             # <<<<<<<<<<<<<<
 * 
 *         assets_df = map_df_to_str(self.pure_mm_assets_df(not self._inventory_skew_enabled))
*/
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__15);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u__15);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_mstate_global->__pyx_kp_u__15) != (0)) __PYX_ERR(0, 658, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Markets);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Markets);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_mstate_global->__pyx_kp_u_Markets) != (0)) __PYX_ERR(0, 658, __pyx_L1_error);
  { /* enter inner scope */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 658, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = __pyx_v_markets_df;
    __Pyx_INCREF(__pyx_t_8);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_8, NULL};
      __pyx_t_10 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 658, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_index, Py_False, __pyx_t_10, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 658, __pyx_L6_error)
      __pyx_t_9 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_to_string, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_10);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 658, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_9);
    }
    __pyx_t_5 = __pyx_t_9;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_kp_u__16};
      __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_split, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 658, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_7);
    }
    if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
      __pyx_t_9 = __pyx_t_7; __Pyx_INCREF(__pyx_t_9);
      __pyx_t_11 = 0;
      __pyx_t_12 = NULL;
    } else {
      __pyx_t_11 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 658, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_12 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_9); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 658, __pyx_L6_error)
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    for (;;) {
      if (likely(!__pyx_t_12)) {
        if (likely(PyList_CheckExact(__pyx_t_9))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_9);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 658, __pyx_L6_error)
            #endif
            if (__pyx_t_11 >= __pyx_temp) break;
          }
          __pyx_t_7 = __Pyx_PyList_GetItemRefFast(__pyx_t_9, __pyx_t_11, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_11;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_9);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 658, __pyx_L6_error)
            #endif
            if (__pyx_t_11 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_11));
          #else
          __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_11);
          #endif
          ++__pyx_t_11;
        }
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 658, __pyx_L6_error)
      } else {
        __pyx_t_7 = __pyx_t_12(__pyx_t_9);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 658, __pyx_L6_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_line, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = PyNumber_Add(__pyx_mstate_global->__pyx_kp_u__17, __pyx_8genexpr6__pyx_v_line); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 658, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 658, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_line); __pyx_8genexpr6__pyx_v_line = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_line); __pyx_8genexpr6__pyx_v_line = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_t_9 = PyNumber_Add(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyList_Extend(__pyx_v_lines, __pyx_t_9); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":660
 *         lines.extend(["", "  Markets:"] + ["    " + line for line in markets_df.to_string(index=False).split("\n")])
 * 
 *         assets_df = map_df_to_str(self.pure_mm_assets_df(not self._inventory_skew_enabled))             # <<<<<<<<<<<<<<
 *         # append inventory skew stats.
 *         if self._inventory_skew_enabled:
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_map_df_to_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyBool_FromLong((!__pyx_v_self->_inventory_skew_enabled)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_10};
    __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_pure_mm_assets_df, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 660, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
  }
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_7};
    __pyx_t_9 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 660, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
  }
  __pyx_v_assets_df = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":662
 *         assets_df = map_df_to_str(self.pure_mm_assets_df(not self._inventory_skew_enabled))
 *         # append inventory skew stats.
 *         if self._inventory_skew_enabled:             # <<<<<<<<<<<<<<
 *             inventory_skew_df = map_df_to_str(self.inventory_skew_stats_data_frame())
 *             assets_df = pd.concat(
*/
  if (__pyx_v_self->_inventory_skew_enabled) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":663
 *         # append inventory skew stats.
 *         if self._inventory_skew_enabled:
 *             inventory_skew_df = map_df_to_str(self.inventory_skew_stats_data_frame())             # <<<<<<<<<<<<<<
 *             assets_df = pd.concat(
 *                 [assets_df, inventory_skew_df], join="inner")
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_map_df_to_str); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 663, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_10);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_inventory_skew_stats_data_frame, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 663, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
      __pyx_t_9 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 663, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
    }
    __pyx_v_inventory_skew_df = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":664
 *         if self._inventory_skew_enabled:
 *             inventory_skew_df = map_df_to_str(self.inventory_skew_stats_data_frame())
 *             assets_df = pd.concat(             # <<<<<<<<<<<<<<
 *                 [assets_df, inventory_skew_df], join="inner")
 * 
*/
    __pyx_t_7 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pd); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_concat); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":665
 *             inventory_skew_df = map_df_to_str(self.inventory_skew_stats_data_frame())
 *             assets_df = pd.concat(
 *                 [assets_df, inventory_skew_df], join="inner")             # <<<<<<<<<<<<<<
 * 
 *         first_col_length = max(*assets_df[0].apply(len))
*/
    __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_assets_df);
    __Pyx_GIVEREF(__pyx_v_assets_df);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_assets_df) != (0)) __PYX_ERR(0, 665, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_inventory_skew_df);
    __Pyx_GIVEREF(__pyx_v_inventory_skew_df);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_v_inventory_skew_df) != (0)) __PYX_ERR(0, 665, __pyx_L1_error);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
      assert(__pyx_t_7);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_7, __pyx_t_4};
      __pyx_t_10 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_join, __pyx_mstate_global->__pyx_n_u_inner, __pyx_t_10, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 664, __pyx_L1_error)
      __pyx_t_9 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_10);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_DECREF_SET(__pyx_v_assets_df, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":662
 *         assets_df = map_df_to_str(self.pure_mm_assets_df(not self._inventory_skew_enabled))
 *         # append inventory skew stats.
 *         if self._inventory_skew_enabled:             # <<<<<<<<<<<<<<
 *             inventory_skew_df = map_df_to_str(self.inventory_skew_stats_data_frame())
 *             assets_df = pd.concat(
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":667
 *                 [assets_df, inventory_skew_df], join="inner")
 * 
 *         first_col_length = max(*assets_df[0].apply(len))             # <<<<<<<<<<<<<<
 *         df_lines = assets_df.to_string(index=False, header=False,
 *                                        formatters={0: ("{:<" + str(first_col_length) + "}").format}).split("\n")
*/
  __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_assets_df, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_2 = __pyx_t_10;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_GetBuiltinName(__pyx_mstate_global->__pyx_n_u_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_9 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_apply, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
  }
  __pyx_t_10 = __Pyx_PySequence_Tuple(__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_max, __pyx_t_10, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_first_col_length = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":668
 * 
 *         first_col_length = max(*assets_df[0].apply(len))
 *         df_lines = assets_df.to_string(index=False, header=False,             # <<<<<<<<<<<<<<
 *                                        formatters={0: ("{:<" + str(first_col_length) + "}").format}).split("\n")
 *         lines.extend(["", "  Assets:"] + ["    " + line for line in df_lines])
*/
  __pyx_t_2 = __pyx_v_assets_df;
  __Pyx_INCREF(__pyx_t_2);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":669
 *         first_col_length = max(*assets_df[0].apply(len))
 *         df_lines = assets_df.to_string(index=False, header=False,
 *                                        formatters={0: ("{:<" + str(first_col_length) + "}").format}).split("\n")             # <<<<<<<<<<<<<<
 *         lines.extend(["", "  Assets:"] + ["    " + line for line in df_lines])
 * 
*/
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Unicode(__pyx_v_first_col_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u__18, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_8, __pyx_mstate_global->__pyx_kp_u__19); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_format); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_mstate_global->__pyx_int_0, __pyx_t_8) < (0)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 3 : 0)] = {__pyx_t_2, NULL};
    __pyx_t_8 = __Pyx_MakeVectorcallBuilderKwds(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_index, Py_False, __pyx_t_8, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 668, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_header, Py_False, __pyx_t_8, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 668, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_formatters, __pyx_t_7, __pyx_t_8, __pyx_callargs+1, 2) < (0)) __PYX_ERR(0, 668, __pyx_L1_error)
    __pyx_t_4 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_to_string, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_10 = __pyx_t_4;
  __Pyx_INCREF(__pyx_t_10);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_mstate_global->__pyx_kp_u__16};
    __pyx_t_9 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_split, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
  }
  __pyx_v_df_lines = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":670
 *         df_lines = assets_df.to_string(index=False, header=False,
 *                                        formatters={0: ("{:<" + str(first_col_length) + "}").format}).split("\n")
 *         lines.extend(["", "  Assets:"] + ["    " + line for line in df_lines])             # <<<<<<<<<<<<<<
 * 
 *         # See if there're any open orders.
*/
  __pyx_t_9 = PyList_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__15);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u__15);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 0, __pyx_mstate_global->__pyx_kp_u__15) != (0)) __PYX_ERR(0, 670, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Assets);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Assets);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 1, __pyx_mstate_global->__pyx_kp_u_Assets) != (0)) __PYX_ERR(0, 670, __pyx_L1_error);
  { /* enter inner scope */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 670, __pyx_L14_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(__pyx_v_df_lines)) || PyTuple_CheckExact(__pyx_v_df_lines)) {
      __pyx_t_10 = __pyx_v_df_lines; __Pyx_INCREF(__pyx_t_10);
      __pyx_t_11 = 0;
      __pyx_t_12 = NULL;
    } else {
      __pyx_t_11 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_v_df_lines); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 670, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_12 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 670, __pyx_L14_error)
    }
    for (;;) {
      if (likely(!__pyx_t_12)) {
        if (likely(PyList_CheckExact(__pyx_t_10))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_10);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 670, __pyx_L14_error)
            #endif
            if (__pyx_t_11 >= __pyx_temp) break;
          }
          __pyx_t_8 = __Pyx_PyList_GetItemRefFast(__pyx_t_10, __pyx_t_11, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_11;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_10);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 670, __pyx_L14_error)
            #endif
            if (__pyx_t_11 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_11));
          #else
          __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_10, __pyx_t_11);
          #endif
          ++__pyx_t_11;
        }
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 670, __pyx_L14_error)
      } else {
        __pyx_t_8 = __pyx_t_12(__pyx_t_10);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 670, __pyx_L14_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_8genexpr7__pyx_v_line, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_8 = PyNumber_Add(__pyx_mstate_global->__pyx_kp_u__17, __pyx_8genexpr7__pyx_v_line); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 670, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 670, __pyx_L14_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_line); __pyx_8genexpr7__pyx_v_line = 0;
    goto __pyx_L18_exit_scope;
    __pyx_L14_error:;
    __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_line); __pyx_8genexpr7__pyx_v_line = 0;
    goto __pyx_L1_error;
    __pyx_L18_exit_scope:;
  } /* exit inner scope */
  __pyx_t_10 = PyNumber_Add(__pyx_t_9, __pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyList_Extend(__pyx_v_lines, __pyx_t_10); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":673
 * 
 *         # See if there're any open orders.
 *         if len(self.active_orders) > 0:             # <<<<<<<<<<<<<<
 *             df = map_df_to_str(self.active_orders_df())
 *             lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])
*/
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_orders); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = PyObject_Length(__pyx_t_10); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_1 = (__pyx_t_11 > 0);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":674
 *         # See if there're any open orders.
 *         if len(self.active_orders) > 0:
 *             df = map_df_to_str(self.active_orders_df())             # <<<<<<<<<<<<<<
 *             lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])
 *         else:
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_map_df_to_str); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_7 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
      __pyx_t_8 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_active_orders_df, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
    }
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_9, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_8};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_9, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_v_df = __pyx_t_10;
    __pyx_t_10 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":675
 *         if len(self.active_orders) > 0:
 *             df = map_df_to_str(self.active_orders_df())
 *             lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])             # <<<<<<<<<<<<<<
 *         else:
 *             lines.extend(["", "  No active maker orders."])
*/
    __pyx_t_10 = PyList_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__15);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u__15);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 0, __pyx_mstate_global->__pyx_kp_u__15) != (0)) __PYX_ERR(0, 675, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Orders);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_Orders);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 1, __pyx_mstate_global->__pyx_kp_u_Orders) != (0)) __PYX_ERR(0, 675, __pyx_L1_error);
    { /* enter inner scope */
      __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 675, __pyx_L22_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_2 = __pyx_v_df;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_6 = 0;
      {
        PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, NULL};
        __pyx_t_5 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 675, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_index, Py_False, __pyx_t_5, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 675, __pyx_L22_error)
        __pyx_t_7 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_to_string, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_5);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 675, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_7);
      }
      __pyx_t_4 = __pyx_t_7;
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_6 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u__16};
        __pyx_t_8 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_split, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 675, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_8);
      }
      if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
        __pyx_t_7 = __pyx_t_8; __Pyx_INCREF(__pyx_t_7);
        __pyx_t_11 = 0;
        __pyx_t_12 = NULL;
      } else {
        __pyx_t_11 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 675, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_12 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 675, __pyx_L22_error)
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      for (;;) {
        if (likely(!__pyx_t_12)) {
          if (likely(PyList_CheckExact(__pyx_t_7))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_7);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 675, __pyx_L22_error)
              #endif
              if (__pyx_t_11 >= __pyx_temp) break;
            }
            __pyx_t_8 = __Pyx_PyList_GetItemRefFast(__pyx_t_7, __pyx_t_11, __Pyx_ReferenceSharing_OwnStrongReference);
            ++__pyx_t_11;
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_7);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 675, __pyx_L22_error)
              #endif
              if (__pyx_t_11 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_8 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_11));
            #else
            __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_11);
            #endif
            ++__pyx_t_11;
          }
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 675, __pyx_L22_error)
        } else {
          __pyx_t_8 = __pyx_t_12(__pyx_t_7);
          if (unlikely(!__pyx_t_8)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 675, __pyx_L22_error)
              PyErr_Clear();
            }
            break;
          }
        }
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_XDECREF_SET(__pyx_8genexpr8__pyx_v_line, __pyx_t_8);
        __pyx_t_8 = 0;
        __pyx_t_8 = PyNumber_Add(__pyx_mstate_global->__pyx_kp_u__17, __pyx_8genexpr8__pyx_v_line); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 675, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_9, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 675, __pyx_L22_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_line); __pyx_8genexpr8__pyx_v_line = 0;
      goto __pyx_L26_exit_scope;
      __pyx_L22_error:;
      __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_line); __pyx_8genexpr8__pyx_v_line = 0;
      goto __pyx_L1_error;
      __pyx_L26_exit_scope:;
    } /* exit inner scope */
    __pyx_t_7 = PyNumber_Add(__pyx_t_10, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_3 = __Pyx_PyList_Extend(__pyx_v_lines, __pyx_t_7); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":673
 * 
 *         # See if there're any open orders.
 *         if len(self.active_orders) > 0:             # <<<<<<<<<<<<<<
 *             df = map_df_to_str(self.active_orders_df())
 *             lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])
*/
    goto __pyx_L19;
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":677
 *             lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])
 *         else:
 *             lines.extend(["", "  No active maker orders."])             # <<<<<<<<<<<<<<
 * 
 *         warning_lines.extend(self.balance_warning([self._market_info]))
*/
  /*else*/ {
    __pyx_t_3 = __Pyx_ListComp_Append(__pyx_v_lines, __pyx_mstate_global->__pyx_kp_u__15); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 677, __pyx_L1_error)
    __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_lines, __pyx_mstate_global->__pyx_kp_u_No_active_maker_orders); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 677, __pyx_L1_error)
    (void)((__pyx_t_3 | __pyx_t_13));
  }
  __pyx_L19:;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":679
 *             lines.extend(["", "  No active maker orders."])
 * 
 *         warning_lines.extend(self.balance_warning([self._market_info]))             # <<<<<<<<<<<<<<
 * 
 *         if len(warning_lines) > 0:
*/
  __pyx_t_9 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_9);
  __pyx_t_10 = PyList_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_v_self->_market_info);
  __Pyx_GIVEREF(__pyx_v_self->_market_info);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 0, __pyx_v_self->_market_info) != (0)) __PYX_ERR(0, 679, __pyx_L1_error);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_10};
    __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_balance_warning, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
  }
  __pyx_t_13 = __Pyx_PyList_Extend(__pyx_v_warning_lines, __pyx_t_7); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":681
 *         warning_lines.extend(self.balance_warning([self._market_info]))
 * 
 *         if len(warning_lines) > 0:             # <<<<<<<<<<<<<<
 *             lines.extend(["", "*** WARNINGS ***"] + warning_lines)
 * 
*/
  __pyx_t_11 = __Pyx_PyList_GET_SIZE(__pyx_v_warning_lines); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 681, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_11 > 0);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":682
 * 
 *         if len(warning_lines) > 0:
 *             lines.extend(["", "*** WARNINGS ***"] + warning_lines)             # <<<<<<<<<<<<<<
 * 
 *         return "\n".join(lines)
*/
    __pyx_t_7 = PyList_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 682, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__15);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u__15);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 0, __pyx_mstate_global->__pyx_kp_u__15) != (0)) __PYX_ERR(0, 682, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_WARNINGS);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_kp_u_WARNINGS);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 1, __pyx_mstate_global->__pyx_kp_u_WARNINGS) != (0)) __PYX_ERR(0, 682, __pyx_L1_error);
    __pyx_t_10 = PyNumber_Add(__pyx_t_7, __pyx_v_warning_lines); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 682, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_13 = __Pyx_PyList_Extend(__pyx_v_lines, __pyx_t_10); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 682, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":681
 *         warning_lines.extend(self.balance_warning([self._market_info]))
 * 
 *         if len(warning_lines) > 0:             # <<<<<<<<<<<<<<
 *             lines.extend(["", "*** WARNINGS ***"] + warning_lines)
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":684
 *             lines.extend(["", "*** WARNINGS ***"] + warning_lines)
 * 
 *         return "\n".join(lines)             # <<<<<<<<<<<<<<
 * 
 *     # The following exposed Python functions are meant for unit tests
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_10 = PyUnicode_Join(__pyx_mstate_global->__pyx_kp_u__16, __pyx_v_lines); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":648
 *         return pd.DataFrame(data=markets_data, columns=markets_columns).replace(np.nan, '', regex=True)
 * 
 *     def format_status(self) -> str:             # <<<<<<<<<<<<<<
 *         if not self._all_markets_ready:
 *             return "Market connectors are not ready."
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.format_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lines);
  __Pyx_XDECREF(__pyx_v_warning_lines);
  __Pyx_XDECREF(__pyx_v_markets_df);
  __Pyx_XDECREF(__pyx_v_assets_df);
  __Pyx_XDECREF(__pyx_v_inventory_skew_df);
  __Pyx_XDECREF(__pyx_v_first_col_length);
  __Pyx_XDECREF(__pyx_v_df_lines);
  __Pyx_XDECREF(__pyx_v_df);
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_line);
  __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_line);
  __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_line);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":688
 *     # The following exposed Python functions are meant for unit tests
 *     # ---------------------------------------------------------------
 *     def execute_orders_proposal(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_execute_orders_proposal(proposal)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_21execute_orders_proposal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_21execute_orders_proposal = {"execute_orders_proposal", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_21execute_orders_proposal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_21execute_orders_proposal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_proposal = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("execute_orders_proposal (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_proposal,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 688, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 688, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "execute_orders_proposal", 0) < (0)) __PYX_ERR(0, 688, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("execute_orders_proposal", 1, 1, 1, i); __PYX_ERR(0, 688, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 688, __pyx_L3_error)
    }
    __pyx_v_proposal = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("execute_orders_proposal", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 688, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.execute_orders_proposal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20execute_orders_proposal(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), __pyx_v_proposal);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20execute_orders_proposal(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute_orders_proposal", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":689
 *     # ---------------------------------------------------------------
 *     def execute_orders_proposal(self, proposal: Proposal):
 *         return self.c_execute_orders_proposal(proposal)             # <<<<<<<<<<<<<<
 * 
 *     def cancel_order(self, order_id: str):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_execute_orders_proposal(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":688
 *     # The following exposed Python functions are meant for unit tests
 *     # ---------------------------------------------------------------
 *     def execute_orders_proposal(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_execute_orders_proposal(proposal)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.execute_orders_proposal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":691
 *         return self.c_execute_orders_proposal(proposal)
 * 
 *     def cancel_order(self, order_id: str):             # <<<<<<<<<<<<<<
 *         return self.c_cancel_order(self._market_info, order_id)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23cancel_order(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23cancel_order = {"cancel_order", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23cancel_order, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23cancel_order(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_order_id = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel_order (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_order_id,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 691, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 691, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "cancel_order", 0) < (0)) __PYX_ERR(0, 691, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("cancel_order", 1, 1, 1, i); __PYX_ERR(0, 691, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 691, __pyx_L3_error)
    }
    __pyx_v_order_id = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cancel_order", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 691, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.cancel_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22cancel_order(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), __pyx_v_order_id);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22cancel_order(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_id) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel_order", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":692
 * 
 *     def cancel_order(self, order_id: str):
 *         return self.c_cancel_order(self._market_info, order_id)             # <<<<<<<<<<<<<<
 * 
 *     # ---------------------------------------------------------------
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->_market_info;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_order_id;
  __Pyx_INCREF(__pyx_t_2);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(0, 692, __pyx_L1_error)
  __pyx_t_3 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_cancel_order(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), __pyx_t_1, ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":691
 *         return self.c_execute_orders_proposal(proposal)
 * 
 *     def cancel_order(self, order_id: str):             # <<<<<<<<<<<<<<
 *         return self.c_cancel_order(self._market_info, order_id)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.cancel_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7c_start_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":707
 *             # make restored order hanging orders
 *             for order_id in restored_order_ids:
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)             # <<<<<<<<<<<<<<
 *                 if order:
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)
*/

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7c_start_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 707, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7c_start_2generator, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1]), (PyObject *) __pyx_cur_scope, __pyx_mstate_global->__pyx_n_u_genexpr, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_c_start, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market); if (unlikely(!gen)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_start.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7c_start_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(__pyx_sent_value != Py_None)) {
    if (unlikely(__pyx_sent_value)) PyErr_SetString(PyExc_TypeError, "can't send non-None value to a just-started generator");
    __PYX_ERR(0, 707, __pyx_L1_error)
  }
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 707, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 707, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 707, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_2, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_2;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 707, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2));
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2);
        #endif
        ++__pyx_t_2;
      }
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 707, __pyx_L1_error)
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 707, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_o);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_o, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_o, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_order_id)) { __Pyx_RaiseClosureNameError("order_id"); __PYX_ERR(0, 707, __pyx_L1_error) }
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_order_id, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_6) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_o);
      __pyx_r = __pyx_cur_scope->__pyx_v_o;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 707, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__Pyx_PyErr_Occurred()) {
    __Pyx_Generator_Replace_StopIteration(0);
    __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  }
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":696
 *     # ---------------------------------------------------------------
 * 
 *     cdef c_start(self, Clock clock, double timestamp):             # <<<<<<<<<<<<<<
 *         StrategyBase.c_start(self, clock, timestamp)
 *         self._last_timestamp = timestamp
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_start(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, struct __pyx_obj_10hummingbot_4core_5clock_Clock *__pyx_v_clock, double __pyx_v_timestamp) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *__pyx_cur_scope;
  PyObject *__pyx_v_restored_order_ids = NULL;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7c_start_2generator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_start", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 696, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":697
 * 
 *     cdef c_start(self, Clock clock, double timestamp):
 *         StrategyBase.c_start(self, clock, timestamp)             # <<<<<<<<<<<<<<
 *         self._last_timestamp = timestamp
 * 
*/
  __pyx_t_1 = __pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase->__pyx_base.c_start(((struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *)__pyx_v_self), __pyx_v_clock, __pyx_v_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":698
 *     cdef c_start(self, Clock clock, double timestamp):
 *         StrategyBase.c_start(self, clock, timestamp)
 *         self._last_timestamp = timestamp             # <<<<<<<<<<<<<<
 * 
 *         self._hanging_orders_tracker.register_events(self.active_markets)
*/
  __pyx_v_self->_last_timestamp = __pyx_v_timestamp;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":700
 *         self._last_timestamp = timestamp
 * 
 *         self._hanging_orders_tracker.register_events(self.active_markets)             # <<<<<<<<<<<<<<
 * 
 *         if self._hanging_orders_enabled:
*/
  __pyx_t_2 = __pyx_v_self->_hanging_orders_tracker;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_markets); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_register_events, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":702
 *         self._hanging_orders_tracker.register_events(self.active_markets)
 * 
 *         if self._hanging_orders_enabled:             # <<<<<<<<<<<<<<
 *             # start tracking any restored limit order
 *             restored_order_ids = self.c_track_restored_orders(self.market_info)
*/
  if (__pyx_v_self->_hanging_orders_enabled) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":704
 *         if self._hanging_orders_enabled:
 *             # start tracking any restored limit order
 *             restored_order_ids = self.c_track_restored_orders(self.market_info)             # <<<<<<<<<<<<<<
 *             # make restored order hanging orders
 *             for order_id in restored_order_ids:
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_market_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_track_restored_orders(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_restored_order_ids = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":706
 *             restored_order_ids = self.c_track_restored_orders(self.market_info)
 *             # make restored order hanging orders
 *             for order_id in restored_order_ids:             # <<<<<<<<<<<<<<
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)
 *                 if order:
*/
    if (likely(PyList_CheckExact(__pyx_v_restored_order_ids)) || PyTuple_CheckExact(__pyx_v_restored_order_ids)) {
      __pyx_t_3 = __pyx_v_restored_order_ids; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_restored_order_ids); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 706, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 706, __pyx_L1_error)
            #endif
            if (__pyx_t_5 >= __pyx_temp) break;
          }
          __pyx_t_1 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_5, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_5;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 706, __pyx_L1_error)
            #endif
            if (__pyx_t_5 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_5));
          #else
          __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_5);
          #endif
          ++__pyx_t_5;
        }
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 706, __pyx_L1_error)
      } else {
        __pyx_t_1 = __pyx_t_6(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 706, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_order_id);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_order_id, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      __pyx_t_1 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":707
 *             # make restored order hanging orders
 *             for order_id in restored_order_ids:
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)             # <<<<<<<<<<<<<<
 *                 if order:
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_market_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_limit_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7c_start_genexpr(((PyObject*)__pyx_cur_scope), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyIter_Next(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":708
 *             for order_id in restored_order_ids:
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)
 *                 if order:             # <<<<<<<<<<<<<<
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)
 * 
*/
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_order); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 708, __pyx_L1_error)
      if (__pyx_t_7) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":709
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)
 *                 if order:
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_stop(self, Clock clock):
*/
        __pyx_t_2 = __pyx_v_self->_hanging_orders_tracker;
        __Pyx_INCREF(__pyx_t_2);
        __pyx_t_4 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_order};
          __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add_as_hanging_order, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":708
 *             for order_id in restored_order_ids:
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)
 *                 if order:             # <<<<<<<<<<<<<<
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)
 * 
*/
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":706
 *             restored_order_ids = self.c_track_restored_orders(self.market_info)
 *             # make restored order hanging orders
 *             for order_id in restored_order_ids:             # <<<<<<<<<<<<<<
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)
 *                 if order:
*/
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":702
 *         self._hanging_orders_tracker.register_events(self.active_markets)
 * 
 *         if self._hanging_orders_enabled:             # <<<<<<<<<<<<<<
 *             # start tracking any restored limit order
 *             restored_order_ids = self.c_track_restored_orders(self.market_info)
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":696
 *     # ---------------------------------------------------------------
 * 
 *     cdef c_start(self, Clock clock, double timestamp):             # <<<<<<<<<<<<<<
 *         StrategyBase.c_start(self, clock, timestamp)
 *         self._last_timestamp = timestamp
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_restored_order_ids);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XDECREF(__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7c_start_2generator);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":711
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)
 * 
 *     cdef c_stop(self, Clock clock):             # <<<<<<<<<<<<<<
 *         self._hanging_orders_tracker.unregister_events(self.active_markets)
 *         StrategyBase.c_stop(self, clock)
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_stop(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, struct __pyx_obj_10hummingbot_4core_5clock_Clock *__pyx_v_clock) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_stop", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":712
 * 
 *     cdef c_stop(self, Clock clock):
 *         self._hanging_orders_tracker.unregister_events(self.active_markets)             # <<<<<<<<<<<<<<
 *         StrategyBase.c_stop(self, clock)
 * 
*/
  __pyx_t_2 = __pyx_v_self->_hanging_orders_tracker;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_markets); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_unregister_events, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":713
 *     cdef c_stop(self, Clock clock):
 *         self._hanging_orders_tracker.unregister_events(self.active_markets)
 *         StrategyBase.c_stop(self, clock)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_tick(self, double timestamp):
*/
  __pyx_t_1 = __pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase->__pyx_base.c_stop(((struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *)__pyx_v_self), __pyx_v_clock); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":711
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)
 * 
 *     cdef c_stop(self, Clock clock):             # <<<<<<<<<<<<<<
 *         self._hanging_orders_tracker.unregister_events(self.active_markets)
 *         StrategyBase.c_stop(self, clock)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_stop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":715
 *         StrategyBase.c_stop(self, clock)
 * 
 *     cdef c_tick(self, double timestamp):             # <<<<<<<<<<<<<<
 *         StrategyBase.c_tick(self, timestamp)
 * 
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_tick(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, double __pyx_v_timestamp) {
  int64_t __pyx_v_current_tick;
  int64_t __pyx_v_last_tick;
  int __pyx_v_should_report_warnings;
  PyObject *__pyx_v_proposal = 0;
  PyObject *__pyx_9genexpr10__pyx_v_market = NULL;
  PyObject *__pyx_9genexpr11__pyx_v_market = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  size_t __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_tick", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":716
 * 
 *     cdef c_tick(self, double timestamp):
 *         StrategyBase.c_tick(self, timestamp)             # <<<<<<<<<<<<<<
 * 
 *         cdef:
*/
  __pyx_t_1 = __pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase->__pyx_base.c_tick(((struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *)__pyx_v_self), __pyx_v_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":719
 * 
 *         cdef:
 *             int64_t current_tick = <int64_t>(timestamp // self._status_report_interval)             # <<<<<<<<<<<<<<
 *             int64_t last_tick = <int64_t>(self._last_timestamp // self._status_report_interval)
 *             bint should_report_warnings = ((current_tick > last_tick) and
*/
  if (unlikely(__pyx_v_self->_status_report_interval == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 719, __pyx_L1_error)
  }
  __pyx_v_current_tick = ((int64_t)floor(__pyx_v_timestamp / __pyx_v_self->_status_report_interval));

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":720
 *         cdef:
 *             int64_t current_tick = <int64_t>(timestamp // self._status_report_interval)
 *             int64_t last_tick = <int64_t>(self._last_timestamp // self._status_report_interval)             # <<<<<<<<<<<<<<
 *             bint should_report_warnings = ((current_tick > last_tick) and
 *                                            (self._logging_options & self.OPTION_LOG_STATUS_REPORT))
*/
  if (unlikely(__pyx_v_self->_status_report_interval == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 720, __pyx_L1_error)
  }
  __pyx_v_last_tick = ((int64_t)floor(__pyx_v_self->_last_timestamp / __pyx_v_self->_status_report_interval));

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":721
 *             int64_t current_tick = <int64_t>(timestamp // self._status_report_interval)
 *             int64_t last_tick = <int64_t>(self._last_timestamp // self._status_report_interval)
 *             bint should_report_warnings = ((current_tick > last_tick) and             # <<<<<<<<<<<<<<
 *                                            (self._logging_options & self.OPTION_LOG_STATUS_REPORT))
 *             cdef object proposal
*/
  __pyx_t_3 = (__pyx_v_current_tick > __pyx_v_last_tick);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L3_bool_binop_done;
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":722
 *             int64_t last_tick = <int64_t>(self._last_timestamp // self._status_report_interval)
 *             bint should_report_warnings = ((current_tick > last_tick) and
 *                                            (self._logging_options & self.OPTION_LOG_STATUS_REPORT))             # <<<<<<<<<<<<<<
 *             cdef object proposal
 *         try:
*/
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->_logging_options); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_OPTION_LOG_STATUS_REPORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyNumber_And(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __pyx_t_3;
  __pyx_L3_bool_binop_done:;
  __pyx_v_should_report_warnings = __pyx_t_2;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":724
 *                                            (self._logging_options & self.OPTION_LOG_STATUS_REPORT))
 *             cdef object proposal
 *         try:             # <<<<<<<<<<<<<<
 *             if not self._all_markets_ready:
 *                 self._all_markets_ready = all([market.ready for market in self._sb_markets])
*/
  /*try:*/ {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":725
 *             cdef object proposal
 *         try:
 *             if not self._all_markets_ready:             # <<<<<<<<<<<<<<
 *                 self._all_markets_ready = all([market.ready for market in self._sb_markets])
 *                 if self._asset_price_delegate is not None and self._all_markets_ready:
*/
    __pyx_t_2 = (!__pyx_v_self->_all_markets_ready);
    if (__pyx_t_2) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":726
 *         try:
 *             if not self._all_markets_ready:
 *                 self._all_markets_ready = all([market.ready for market in self._sb_markets])             # <<<<<<<<<<<<<<
 *                 if self._asset_price_delegate is not None and self._all_markets_ready:
 *                     self._all_markets_ready = self._asset_price_delegate.ready
*/
      __pyx_t_4 = NULL;
      { /* enter inner scope */
        __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = 0;
        __pyx_t_10 = __Pyx_set_iterator(__pyx_v_self->__pyx_base._sb_markets, 1, (&__pyx_t_8), (&__pyx_t_9)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 726, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_6);
        __pyx_t_6 = __pyx_t_10;
        __pyx_t_10 = 0;
        while (1) {
          __pyx_t_11 = __Pyx_set_iter_next(__pyx_t_6, __pyx_t_8, &__pyx_t_7, &__pyx_t_10, __pyx_t_9);
          if (unlikely(__pyx_t_11 == 0)) break;
          if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 726, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_9genexpr10__pyx_v_market, __pyx_t_10);
          __pyx_t_10 = 0;
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr10__pyx_v_market, __pyx_mstate_global->__pyx_n_u_ready); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 726, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 726, __pyx_L11_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_market); __pyx_9genexpr10__pyx_v_market = 0;
        goto __pyx_L14_exit_scope;
        __pyx_L11_error:;
        __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_market); __pyx_9genexpr10__pyx_v_market = 0;
        goto __pyx_L6_error;
        __pyx_L14_exit_scope:;
      } /* exit inner scope */
      __pyx_t_12 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_1};
        __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_all, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 726, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
      }
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_self->_all_markets_ready = __pyx_t_2;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":727
 *             if not self._all_markets_ready:
 *                 self._all_markets_ready = all([market.ready for market in self._sb_markets])
 *                 if self._asset_price_delegate is not None and self._all_markets_ready:             # <<<<<<<<<<<<<<
 *                     self._all_markets_ready = self._asset_price_delegate.ready
 *                 if not self._all_markets_ready:
*/
      __pyx_t_3 = (__pyx_v_self->_asset_price_delegate != Py_None);
      if (__pyx_t_3) {
      } else {
        __pyx_t_2 = __pyx_t_3;
        goto __pyx_L16_bool_binop_done;
      }
      __pyx_t_2 = __pyx_v_self->_all_markets_ready;
      __pyx_L16_bool_binop_done:;
      if (__pyx_t_2) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":728
 *                 self._all_markets_ready = all([market.ready for market in self._sb_markets])
 *                 if self._asset_price_delegate is not None and self._all_markets_ready:
 *                     self._all_markets_ready = self._asset_price_delegate.ready             # <<<<<<<<<<<<<<
 *                 if not self._all_markets_ready:
 *                     # Markets not ready yet. Don't do anything.
*/
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_asset_price_delegate, __pyx_mstate_global->__pyx_n_u_ready); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 728, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 728, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_self->_all_markets_ready = __pyx_t_2;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":727
 *             if not self._all_markets_ready:
 *                 self._all_markets_ready = all([market.ready for market in self._sb_markets])
 *                 if self._asset_price_delegate is not None and self._all_markets_ready:             # <<<<<<<<<<<<<<
 *                     self._all_markets_ready = self._asset_price_delegate.ready
 *                 if not self._all_markets_ready:
*/
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":729
 *                 if self._asset_price_delegate is not None and self._all_markets_ready:
 *                     self._all_markets_ready = self._asset_price_delegate.ready
 *                 if not self._all_markets_ready:             # <<<<<<<<<<<<<<
 *                     # Markets not ready yet. Don't do anything.
 *                     if should_report_warnings:
*/
      __pyx_t_2 = (!__pyx_v_self->_all_markets_ready);
      if (__pyx_t_2) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":731
 *                 if not self._all_markets_ready:
 *                     # Markets not ready yet. Don't do anything.
 *                     if should_report_warnings:             # <<<<<<<<<<<<<<
 *                         self.logger().warning(f"Markets are not ready. No market making trades are permitted.")
 *                     return
*/
        if (__pyx_v_should_report_warnings) {

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":732
 *                     # Markets not ready yet. Don't do anything.
 *                     if should_report_warnings:
 *                         self.logger().warning(f"Markets are not ready. No market making trades are permitted.")             # <<<<<<<<<<<<<<
 *                     return
 * 
*/
          __pyx_t_6 = ((PyObject *)__pyx_v_self);
          __Pyx_INCREF(__pyx_t_6);
          __pyx_t_12 = 0;
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
            __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_logger, __pyx_callargs+__pyx_t_12, (1-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 732, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
          }
          __pyx_t_1 = __pyx_t_4;
          __Pyx_INCREF(__pyx_t_1);
          __pyx_t_12 = 0;
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_Markets_are_not_ready_No_market};
            __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warning, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 732, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":731
 *                 if not self._all_markets_ready:
 *                     # Markets not ready yet. Don't do anything.
 *                     if should_report_warnings:             # <<<<<<<<<<<<<<
 *                         self.logger().warning(f"Markets are not ready. No market making trades are permitted.")
 *                     return
*/
        }

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":733
 *                     if should_report_warnings:
 *                         self.logger().warning(f"Markets are not ready. No market making trades are permitted.")
 *                     return             # <<<<<<<<<<<<<<
 * 
 *             if should_report_warnings:
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L5_return;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":729
 *                 if self._asset_price_delegate is not None and self._all_markets_ready:
 *                     self._all_markets_ready = self._asset_price_delegate.ready
 *                 if not self._all_markets_ready:             # <<<<<<<<<<<<<<
 *                     # Markets not ready yet. Don't do anything.
 *                     if should_report_warnings:
*/
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":725
 *             cdef object proposal
 *         try:
 *             if not self._all_markets_ready:             # <<<<<<<<<<<<<<
 *                 self._all_markets_ready = all([market.ready for market in self._sb_markets])
 *                 if self._asset_price_delegate is not None and self._all_markets_ready:
*/
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":735
 *                     return
 * 
 *             if should_report_warnings:             # <<<<<<<<<<<<<<
 *                 if not all([market.network_status is NetworkStatus.CONNECTED for market in self._sb_markets]):
 *                     self.logger().warning(f"WARNING: Some markets are not connected or are down at the moment. Market "
*/
    if (__pyx_v_should_report_warnings) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":736
 * 
 *             if should_report_warnings:
 *                 if not all([market.network_status is NetworkStatus.CONNECTED for market in self._sb_markets]):             # <<<<<<<<<<<<<<
 *                     self.logger().warning(f"WARNING: Some markets are not connected or are down at the moment. Market "
 *                                           f"making may be dangerous when markets or networks are unstable.")
*/
      __pyx_t_4 = NULL;
      { /* enter inner scope */
        __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 736, __pyx_L24_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = 0;
        __pyx_t_10 = __Pyx_set_iterator(__pyx_v_self->__pyx_base._sb_markets, 1, (&__pyx_t_7), (&__pyx_t_9)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 736, __pyx_L24_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_6);
        __pyx_t_6 = __pyx_t_10;
        __pyx_t_10 = 0;
        while (1) {
          __pyx_t_11 = __Pyx_set_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_8, &__pyx_t_10, __pyx_t_9);
          if (unlikely(__pyx_t_11 == 0)) break;
          if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 736, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_9genexpr11__pyx_v_market, __pyx_t_10);
          __pyx_t_10 = 0;
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr11__pyx_v_market, __pyx_mstate_global->__pyx_n_u_network_status); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 736, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_NetworkStatus); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 736, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_CONNECTED); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 736, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_2 = (__pyx_t_10 == __pyx_t_14);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_14 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 736, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_14);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_14))) __PYX_ERR(0, 736, __pyx_L24_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_market); __pyx_9genexpr11__pyx_v_market = 0;
        goto __pyx_L27_exit_scope;
        __pyx_L24_error:;
        __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_market); __pyx_9genexpr11__pyx_v_market = 0;
        goto __pyx_L6_error;
        __pyx_L27_exit_scope:;
      } /* exit inner scope */
      __pyx_t_12 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_1};
        __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_all, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 736, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
      }
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 736, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_3 = (!__pyx_t_2);
      if (__pyx_t_3) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":737
 *             if should_report_warnings:
 *                 if not all([market.network_status is NetworkStatus.CONNECTED for market in self._sb_markets]):
 *                     self.logger().warning(f"WARNING: Some markets are not connected or are down at the moment. Market "             # <<<<<<<<<<<<<<
 *                                           f"making may be dangerous when markets or networks are unstable.")
 * 
*/
        __pyx_t_6 = ((PyObject *)__pyx_v_self);
        __Pyx_INCREF(__pyx_t_6);
        __pyx_t_12 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
          __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_logger, __pyx_callargs+__pyx_t_12, (1-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 737, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
        }
        __pyx_t_1 = __pyx_t_4;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_12 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_WARNING_Some_markets_are_not_con};
          __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warning, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 737, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":736
 * 
 *             if should_report_warnings:
 *                 if not all([market.network_status is NetworkStatus.CONNECTED for market in self._sb_markets]):             # <<<<<<<<<<<<<<
 *                     self.logger().warning(f"WARNING: Some markets are not connected or are down at the moment. Market "
 *                                           f"making may be dangerous when markets or networks are unstable.")
*/
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":735
 *                     return
 * 
 *             if should_report_warnings:             # <<<<<<<<<<<<<<
 *                 if not all([market.network_status is NetworkStatus.CONNECTED for market in self._sb_markets]):
 *                     self.logger().warning(f"WARNING: Some markets are not connected or are down at the moment. Market "
*/
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":740
 *                                           f"making may be dangerous when markets or networks are unstable.")
 * 
 *             proposal = None             # <<<<<<<<<<<<<<
 *             if self._create_timestamp <= self._current_timestamp:
 *                 # 1. Create base order proposals
*/
    __Pyx_INCREF(Py_None);
    __pyx_v_proposal = Py_None;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":741
 * 
 *             proposal = None
 *             if self._create_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *                 # 1. Create base order proposals
 *                 proposal = self.c_create_base_proposal()
*/
    __pyx_t_3 = (__pyx_v_self->_create_timestamp <= __pyx_v_self->__pyx_base.__pyx_base._current_timestamp);
    if (__pyx_t_3) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":743
 *             if self._create_timestamp <= self._current_timestamp:
 *                 # 1. Create base order proposals
 *                 proposal = self.c_create_base_proposal()             # <<<<<<<<<<<<<<
 *                 # 2. Apply functions that limit numbers of buys and sells proposal
 *                 self.c_apply_order_levels_modifiers(proposal)
*/
      __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_create_base_proposal(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 743, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_proposal, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":745
 *                 proposal = self.c_create_base_proposal()
 *                 # 2. Apply functions that limit numbers of buys and sells proposal
 *                 self.c_apply_order_levels_modifiers(proposal)             # <<<<<<<<<<<<<<
 *                 # 3. Apply functions that modify orders price
 *                 self.c_apply_order_price_modifiers(proposal)
*/
      __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_order_levels_modifiers(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 745, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":747
 *                 self.c_apply_order_levels_modifiers(proposal)
 *                 # 3. Apply functions that modify orders price
 *                 self.c_apply_order_price_modifiers(proposal)             # <<<<<<<<<<<<<<
 *                 # 4. Apply functions that modify orders size
 *                 self.c_apply_order_size_modifiers(proposal)
*/
      __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_order_price_modifiers(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 747, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":749
 *                 self.c_apply_order_price_modifiers(proposal)
 *                 # 4. Apply functions that modify orders size
 *                 self.c_apply_order_size_modifiers(proposal)             # <<<<<<<<<<<<<<
 *                 # 5. Apply budget constraint, i.e. can't buy/sell more than what you have.
 *                 self.c_apply_budget_constraint(proposal)
*/
      __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_order_size_modifiers(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 749, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":751
 *                 self.c_apply_order_size_modifiers(proposal)
 *                 # 5. Apply budget constraint, i.e. can't buy/sell more than what you have.
 *                 self.c_apply_budget_constraint(proposal)             # <<<<<<<<<<<<<<
 * 
 *                 if not self._take_if_crossed:
*/
      __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_budget_constraint(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 751, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":753
 *                 self.c_apply_budget_constraint(proposal)
 * 
 *                 if not self._take_if_crossed:             # <<<<<<<<<<<<<<
 *                     self.c_filter_out_takers(proposal)
 * 
*/
      __pyx_t_3 = (!__pyx_v_self->_take_if_crossed);
      if (__pyx_t_3) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":754
 * 
 *                 if not self._take_if_crossed:
 *                     self.c_filter_out_takers(proposal)             # <<<<<<<<<<<<<<
 * 
 *             self._hanging_orders_tracker.process_tick()
*/
        __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_filter_out_takers(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 754, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":753
 *                 self.c_apply_budget_constraint(proposal)
 * 
 *                 if not self._take_if_crossed:             # <<<<<<<<<<<<<<
 *                     self.c_filter_out_takers(proposal)
 * 
*/
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":741
 * 
 *             proposal = None
 *             if self._create_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *                 # 1. Create base order proposals
 *                 proposal = self.c_create_base_proposal()
*/
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":756
 *                     self.c_filter_out_takers(proposal)
 * 
 *             self._hanging_orders_tracker.process_tick()             # <<<<<<<<<<<<<<
 * 
 *             self.c_cancel_active_orders_on_max_age_limit()
*/
    __pyx_t_4 = __pyx_v_self->_hanging_orders_tracker;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_12 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_process_tick, __pyx_callargs+__pyx_t_12, (1-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 756, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":758
 *             self._hanging_orders_tracker.process_tick()
 * 
 *             self.c_cancel_active_orders_on_max_age_limit()             # <<<<<<<<<<<<<<
 *             self.c_cancel_active_orders(proposal)
 *             self.c_cancel_orders_below_min_spread()
*/
    __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_cancel_active_orders_on_max_age_limit(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 758, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":759
 * 
 *             self.c_cancel_active_orders_on_max_age_limit()
 *             self.c_cancel_active_orders(proposal)             # <<<<<<<<<<<<<<
 *             self.c_cancel_orders_below_min_spread()
 *             if self.c_to_create_orders(proposal):
*/
    __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_cancel_active_orders(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 759, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":760
 *             self.c_cancel_active_orders_on_max_age_limit()
 *             self.c_cancel_active_orders(proposal)
 *             self.c_cancel_orders_below_min_spread()             # <<<<<<<<<<<<<<
 *             if self.c_to_create_orders(proposal):
 *                 self.c_execute_orders_proposal(proposal)
*/
    __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_cancel_orders_below_min_spread(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 760, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":761
 *             self.c_cancel_active_orders(proposal)
 *             self.c_cancel_orders_below_min_spread()
 *             if self.c_to_create_orders(proposal):             # <<<<<<<<<<<<<<
 *                 self.c_execute_orders_proposal(proposal)
 *         finally:
*/
    __pyx_t_3 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_to_create_orders(__pyx_v_self, __pyx_v_proposal); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 761, __pyx_L6_error)
    if (__pyx_t_3) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":762
 *             self.c_cancel_orders_below_min_spread()
 *             if self.c_to_create_orders(proposal):
 *                 self.c_execute_orders_proposal(proposal)             # <<<<<<<<<<<<<<
 *         finally:
 *             self._last_timestamp = timestamp
*/
      __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_execute_orders_proposal(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 762, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":761
 *             self.c_cancel_active_orders(proposal)
 *             self.c_cancel_orders_below_min_spread()
 *             if self.c_to_create_orders(proposal):             # <<<<<<<<<<<<<<
 *                 self.c_execute_orders_proposal(proposal)
 *         finally:
*/
    }
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":764
 *                 self.c_execute_orders_proposal(proposal)
 *         finally:
 *             self._last_timestamp = timestamp             # <<<<<<<<<<<<<<
 * 
 *     cdef object c_create_base_proposal(self):
*/
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_v_self->_last_timestamp = __pyx_v_timestamp;
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
      if ( unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_21);
      __pyx_t_9 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
      {
        __pyx_v_self->_last_timestamp = __pyx_v_timestamp;
      }
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_XGIVEREF(__pyx_t_20);
      __Pyx_XGIVEREF(__pyx_t_21);
      __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_XGIVEREF(__pyx_t_18);
      __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
      __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
      __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_15;
      goto __pyx_L1_error;
    }
    __pyx_L5_return: {
      __pyx_t_21 = __pyx_r;
      __pyx_r = 0;
      __pyx_v_self->_last_timestamp = __pyx_v_timestamp;
      __pyx_r = __pyx_t_21;
      __pyx_t_21 = 0;
      goto __pyx_L0;
    }
    __pyx_L7:;
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":715
 *         StrategyBase.c_stop(self, clock)
 * 
 *     cdef c_tick(self, double timestamp):             # <<<<<<<<<<<<<<
 *         StrategyBase.c_tick(self, timestamp)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_tick", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_proposal);
  __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_market);
  __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_market);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":766
 *             self._last_timestamp = timestamp
 * 
 *     cdef object c_create_base_proposal(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_create_base_proposal(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_buys = 0;
  PyObject *__pyx_v_sells = 0;
  PyObject *__pyx_v_buy_reference_price = NULL;
  PyObject *__pyx_v_sell_reference_price = NULL;
  PyObject *__pyx_v_inventory_cost_price = NULL;
  double __pyx_v_base_balance;
  PyObject *__pyx_v_order_override = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_v_size = NULL;
  PyObject *__pyx_v_level = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  double __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  PyObject *(*__pyx_t_18)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_create_base_proposal", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":768
 *     cdef object c_create_base_proposal(self):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             list buys = []
 *             list sells = []
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 768, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":769
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             list buys = []             # <<<<<<<<<<<<<<
 *             list sells = []
 * 
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":770
 *             ExchangeBase market = self._market_info.market
 *             list buys = []
 *             list sells = []             # <<<<<<<<<<<<<<
 * 
 *         buy_reference_price = sell_reference_price = self.get_price()
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sells = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":772
 *             list sells = []
 * 
 *         buy_reference_price = sell_reference_price = self.get_price()             # <<<<<<<<<<<<<<
 * 
 *         if self._inventory_cost_price_delegate is not None:
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 772, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_buy_reference_price = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_sell_reference_price = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":774
 *         buy_reference_price = sell_reference_price = self.get_price()
 * 
 *         if self._inventory_cost_price_delegate is not None:             # <<<<<<<<<<<<<<
 *             inventory_cost_price = self._inventory_cost_price_delegate.get_price()
 *             if inventory_cost_price is not None:
*/
  __pyx_t_4 = (__pyx_v_self->_inventory_cost_price_delegate != Py_None);
  if (__pyx_t_4) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":775
 * 
 *         if self._inventory_cost_price_delegate is not None:
 *             inventory_cost_price = self._inventory_cost_price_delegate.get_price()             # <<<<<<<<<<<<<<
 *             if inventory_cost_price is not None:
 *                 # Only limit sell price. Buy are always allowed.
*/
    __pyx_t_2 = __pyx_v_self->_inventory_cost_price_delegate;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 775, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_v_inventory_cost_price = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":776
 *         if self._inventory_cost_price_delegate is not None:
 *             inventory_cost_price = self._inventory_cost_price_delegate.get_price()
 *             if inventory_cost_price is not None:             # <<<<<<<<<<<<<<
 *                 # Only limit sell price. Buy are always allowed.
 *                 sell_reference_price = max(inventory_cost_price, sell_reference_price)
*/
    __pyx_t_4 = (__pyx_v_inventory_cost_price != Py_None);
    if (__pyx_t_4) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":778
 *             if inventory_cost_price is not None:
 *                 # Only limit sell price. Buy are always allowed.
 *                 sell_reference_price = max(inventory_cost_price, sell_reference_price)             # <<<<<<<<<<<<<<
 *             else:
 *                 base_balance = float(market.get_balance(self._market_info.base_asset))
*/
      __Pyx_INCREF(__pyx_v_sell_reference_price);
      __pyx_t_1 = __pyx_v_sell_reference_price;
      __Pyx_INCREF(__pyx_v_inventory_cost_price);
      __pyx_t_2 = __pyx_v_inventory_cost_price;
      __pyx_t_6 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 778, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 778, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (__pyx_t_4) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1;
      } else {
        __Pyx_INCREF(__pyx_t_2);
        __pyx_t_5 = __pyx_t_2;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __pyx_t_5;
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_sell_reference_price, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":776
 *         if self._inventory_cost_price_delegate is not None:
 *             inventory_cost_price = self._inventory_cost_price_delegate.get_price()
 *             if inventory_cost_price is not None:             # <<<<<<<<<<<<<<
 *                 # Only limit sell price. Buy are always allowed.
 *                 sell_reference_price = max(inventory_cost_price, sell_reference_price)
*/
      goto __pyx_L4;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":780
 *                 sell_reference_price = max(inventory_cost_price, sell_reference_price)
 *             else:
 *                 base_balance = float(market.get_balance(self._market_info.base_asset))             # <<<<<<<<<<<<<<
 *                 if base_balance > 0:
 *                     raise RuntimeError("Initial inventory price is not set while inventory_cost feature is active.")
*/
    /*else*/ {
      __pyx_t_5 = ((PyObject *)__pyx_v_market);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_base_asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_2};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_balance, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __pyx_t_7 = __Pyx_PyObject_AsDouble(__pyx_t_1); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_7, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 780, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_base_balance = __pyx_t_7;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":781
 *             else:
 *                 base_balance = float(market.get_balance(self._market_info.base_asset))
 *                 if base_balance > 0:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError("Initial inventory price is not set while inventory_cost feature is active.")
 * 
*/
      __pyx_t_4 = (__pyx_v_base_balance > 0.0);
      if (unlikely(__pyx_t_4)) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":782
 *                 base_balance = float(market.get_balance(self._market_info.base_asset))
 *                 if base_balance > 0:
 *                     raise RuntimeError("Initial inventory price is not set while inventory_cost feature is active.")             # <<<<<<<<<<<<<<
 * 
 *         # First to check if a customized order override is configured, otherwise the proposal will be created according
*/
        __pyx_t_2 = NULL;
        __pyx_t_3 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Initial_inventory_price_is_not_s};
          __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_RuntimeError)), __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 782, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 782, __pyx_L1_error)

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":781
 *             else:
 *                 base_balance = float(market.get_balance(self._market_info.base_asset))
 *                 if base_balance > 0:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError("Initial inventory price is not set while inventory_cost feature is active.")
 * 
*/
      }
    }
    __pyx_L4:;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":774
 *         buy_reference_price = sell_reference_price = self.get_price()
 * 
 *         if self._inventory_cost_price_delegate is not None:             # <<<<<<<<<<<<<<
 *             inventory_cost_price = self._inventory_cost_price_delegate.get_price()
 *             if inventory_cost_price is not None:
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":786
 *         # First to check if a customized order override is configured, otherwise the proposal will be created according
 *         # to order spread, amount, and levels setting.
 *         order_override = self._order_override             # <<<<<<<<<<<<<<
 *         if order_override is not None and len(order_override) > 0:
 *             for key, value in order_override.items():
*/
  __pyx_t_1 = __pyx_v_self->_order_override;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_order_override = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":787
 *         # to order spread, amount, and levels setting.
 *         order_override = self._order_override
 *         if order_override is not None and len(order_override) > 0:             # <<<<<<<<<<<<<<
 *             for key, value in order_override.items():
 *                 if str(value[0]) in ["buy", "sell"]:
*/
  __pyx_t_8 = (__pyx_v_order_override != Py_None);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_9 = PyObject_Length(__pyx_v_order_override); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 787, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_9 > 0);
  __pyx_t_4 = __pyx_t_8;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_4) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":788
 *         order_override = self._order_override
 *         if order_override is not None and len(order_override) > 0:
 *             for key, value in order_override.items():             # <<<<<<<<<<<<<<
 *                 if str(value[0]) in ["buy", "sell"]:
 *                     if str(value[0]) == "buy" and not buy_reference_price.is_nan():
*/
    __pyx_t_9 = 0;
    if (unlikely(__pyx_v_order_override == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 788, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_dict_iterator(__pyx_v_order_override, 0, __pyx_mstate_global->__pyx_n_u_items, (&__pyx_t_10), (&__pyx_t_11)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
    while (1) {
      __pyx_t_12 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_10, &__pyx_t_9, &__pyx_t_2, &__pyx_t_5, NULL, __pyx_t_11);
      if (unlikely(__pyx_t_12 == 0)) break;
      if (unlikely(__pyx_t_12 == -1)) __PYX_ERR(0, 788, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":789
 *         if order_override is not None and len(order_override) > 0:
 *             for key, value in order_override.items():
 *                 if str(value[0]) in ["buy", "sell"]:             # <<<<<<<<<<<<<<
 *                     if str(value[0]) == "buy" and not buy_reference_price.is_nan():
 *                         price = buy_reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))
*/
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_value, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 789, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_Unicode(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 789, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_buy, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 789, __pyx_L1_error)
      if (!__pyx_t_8) {
      } else {
        __pyx_t_4 = __pyx_t_8;
        goto __pyx_L12_bool_binop_done;
      }
      __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_sell, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 789, __pyx_L1_error)
      __pyx_t_4 = __pyx_t_8;
      __pyx_L12_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = __pyx_t_4;
      if (__pyx_t_8) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":790
 *             for key, value in order_override.items():
 *                 if str(value[0]) in ["buy", "sell"]:
 *                     if str(value[0]) == "buy" and not buy_reference_price.is_nan():             # <<<<<<<<<<<<<<
 *                         price = buy_reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))
 *                         price = market.c_quantize_order_price(self.trading_pair, price)
*/
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_value, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = __Pyx_PyObject_Unicode(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 790, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_buy, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 790, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (__pyx_t_4) {
        } else {
          __pyx_t_8 = __pyx_t_4;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_2 = __pyx_v_buy_reference_price;
        __Pyx_INCREF(__pyx_t_2);
        __pyx_t_3 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
          __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_nan, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 790, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
        }
        __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 790, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_13 = (!__pyx_t_4);
        __pyx_t_8 = __pyx_t_13;
        __pyx_L15_bool_binop_done:;
        if (__pyx_t_8) {

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":791
 *                 if str(value[0]) in ["buy", "sell"]:
 *                     if str(value[0]) == "buy" and not buy_reference_price.is_nan():
 *                         price = buy_reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))             # <<<<<<<<<<<<<<
 *                         price = market.c_quantize_order_price(self.trading_pair, price)
 *                         size = Decimal(str(value[2]))
*/
          __pyx_t_2 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 791, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_3 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
            assert(__pyx_t_2);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
            __pyx_t_3 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_1};
            __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 791, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
          }
          __pyx_t_2 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 791, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_value, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 791, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_16 = __Pyx_PyObject_Unicode(__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 791, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_3 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_14))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_14);
            assert(__pyx_t_2);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_14);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_14, __pyx__function);
            __pyx_t_3 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_16};
            __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_14, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 791, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
          }
          __pyx_t_16 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 791, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_2);
            assert(__pyx_t_16);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_16);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
            __pyx_t_3 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_mstate_global->__pyx_kp_u_100};
            __pyx_t_14 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 791, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
          }
          __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_6, __pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 791, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_14 = PyNumber_Subtract(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 791, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Multiply(__pyx_v_buy_reference_price, __pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 791, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF_SET(__pyx_v_price, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":792
 *                     if str(value[0]) == "buy" and not buy_reference_price.is_nan():
 *                         price = buy_reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))
 *                         price = market.c_quantize_order_price(self.trading_pair, price)             # <<<<<<<<<<<<<<
 *                         size = Decimal(str(value[2]))
 *                         size = market.c_quantize_order_amount(self.trading_pair, size)
*/
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 792, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(0, 792, __pyx_L1_error)
          __pyx_t_14 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_2), __pyx_v_price); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 792, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_price, __pyx_t_14);
          __pyx_t_14 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":793
 *                         price = buy_reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))
 *                         price = market.c_quantize_order_price(self.trading_pair, price)
 *                         size = Decimal(str(value[2]))             # <<<<<<<<<<<<<<
 *                         size = market.c_quantize_order_amount(self.trading_pair, size)
 *                         if size > 0 and price > 0:
*/
          __pyx_t_2 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 793, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_value, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 793, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_16 = __Pyx_PyObject_Unicode(__pyx_t_6); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 793, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_3 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_2);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_3 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_16};
            __pyx_t_14 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 793, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
          }
          __Pyx_XDECREF_SET(__pyx_v_size, __pyx_t_14);
          __pyx_t_14 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":794
 *                         price = market.c_quantize_order_price(self.trading_pair, price)
 *                         size = Decimal(str(value[2]))
 *                         size = market.c_quantize_order_amount(self.trading_pair, size)             # <<<<<<<<<<<<<<
 *                         if size > 0 and price > 0:
 *                             buys.append(PriceSize(price, size))
*/
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 794, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          if (!(likely(PyUnicode_CheckExact(__pyx_t_14))||((__pyx_t_14) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_14))) __PYX_ERR(0, 794, __pyx_L1_error)
          __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_14), __pyx_v_size, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 794, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_DECREF_SET(__pyx_v_size, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":795
 *                         size = Decimal(str(value[2]))
 *                         size = market.c_quantize_order_amount(self.trading_pair, size)
 *                         if size > 0 and price > 0:             # <<<<<<<<<<<<<<
 *                             buys.append(PriceSize(price, size))
 *                     elif str(value[0]) == "sell" and not sell_reference_price.is_nan():
*/
          __pyx_t_5 = PyObject_RichCompare(__pyx_v_size, __pyx_mstate_global->__pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 795, __pyx_L1_error)
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 795, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (__pyx_t_13) {
          } else {
            __pyx_t_8 = __pyx_t_13;
            goto __pyx_L18_bool_binop_done;
          }
          __pyx_t_5 = PyObject_RichCompare(__pyx_v_price, __pyx_mstate_global->__pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 795, __pyx_L1_error)
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 795, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_8 = __pyx_t_13;
          __pyx_L18_bool_binop_done:;
          if (__pyx_t_8) {

            /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":796
 *                         size = market.c_quantize_order_amount(self.trading_pair, size)
 *                         if size > 0 and price > 0:
 *                             buys.append(PriceSize(price, size))             # <<<<<<<<<<<<<<
 *                     elif str(value[0]) == "sell" and not sell_reference_price.is_nan():
 *                         price = sell_reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))
*/
            __pyx_t_14 = NULL;
            __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_PriceSize); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 796, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_3 = 1;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_16))) {
              __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_16);
              assert(__pyx_t_14);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_16, __pyx__function);
              __pyx_t_3 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[3] = {__pyx_t_14, __pyx_v_price, __pyx_v_size};
              __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_16, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 796, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
            }
            __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_buys, __pyx_t_5); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 796, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

            /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":795
 *                         size = Decimal(str(value[2]))
 *                         size = market.c_quantize_order_amount(self.trading_pair, size)
 *                         if size > 0 and price > 0:             # <<<<<<<<<<<<<<
 *                             buys.append(PriceSize(price, size))
 *                     elif str(value[0]) == "sell" and not sell_reference_price.is_nan():
*/
          }

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":790
 *             for key, value in order_override.items():
 *                 if str(value[0]) in ["buy", "sell"]:
 *                     if str(value[0]) == "buy" and not buy_reference_price.is_nan():             # <<<<<<<<<<<<<<
 *                         price = buy_reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))
 *                         price = market.c_quantize_order_price(self.trading_pair, price)
*/
          goto __pyx_L14;
        }

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":797
 *                         if size > 0 and price > 0:
 *                             buys.append(PriceSize(price, size))
 *                     elif str(value[0]) == "sell" and not sell_reference_price.is_nan():             # <<<<<<<<<<<<<<
 *                         price = sell_reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))
 *                         price = market.c_quantize_order_price(self.trading_pair, price)
*/
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_value, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 797, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_16 = __Pyx_PyObject_Unicode(__pyx_t_5); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 797, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_sell, Py_EQ)); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 797, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        if (__pyx_t_13) {
        } else {
          __pyx_t_8 = __pyx_t_13;
          goto __pyx_L20_bool_binop_done;
        }
        __pyx_t_5 = __pyx_v_sell_reference_price;
        __Pyx_INCREF(__pyx_t_5);
        __pyx_t_3 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
          __pyx_t_16 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_nan, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 797, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
        }
        __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_16); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 797, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __pyx_t_4 = (!__pyx_t_13);
        __pyx_t_8 = __pyx_t_4;
        __pyx_L20_bool_binop_done:;
        if (__pyx_t_8) {

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":798
 *                             buys.append(PriceSize(price, size))
 *                     elif str(value[0]) == "sell" and not sell_reference_price.is_nan():
 *                         price = sell_reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))             # <<<<<<<<<<<<<<
 *                         price = market.c_quantize_order_price(self.trading_pair, price)
 *                         size = Decimal(str(value[2]))
*/
          __pyx_t_5 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 798, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_3 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_14))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_14);
            assert(__pyx_t_5);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_14);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_14, __pyx__function);
            __pyx_t_3 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_1};
            __pyx_t_16 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_14, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 798, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
          }
          __pyx_t_5 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_value, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 798, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_15 = __Pyx_PyObject_Unicode(__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 798, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_3 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            assert(__pyx_t_5);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
            __pyx_t_3 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_15};
            __pyx_t_14 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 798, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
          }
          __pyx_t_15 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 798, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_15);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_3 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_mstate_global->__pyx_kp_u_100};
            __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
          }
          __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_14, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 798, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Add(__pyx_t_16, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = PyNumber_Multiply(__pyx_v_sell_reference_price, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 798, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF_SET(__pyx_v_price, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":799
 *                     elif str(value[0]) == "sell" and not sell_reference_price.is_nan():
 *                         price = sell_reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))
 *                         price = market.c_quantize_order_price(self.trading_pair, price)             # <<<<<<<<<<<<<<
 *                         size = Decimal(str(value[2]))
 *                         size = market.c_quantize_order_amount(self.trading_pair, size)
*/
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 799, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_5))) __PYX_ERR(0, 799, __pyx_L1_error)
          __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_5), __pyx_v_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 799, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF_SET(__pyx_v_price, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":800
 *                         price = sell_reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))
 *                         price = market.c_quantize_order_price(self.trading_pair, price)
 *                         size = Decimal(str(value[2]))             # <<<<<<<<<<<<<<
 *                         size = market.c_quantize_order_amount(self.trading_pair, size)
 *                         if size > 0 and price > 0:
*/
          __pyx_t_5 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 800, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_14 = __Pyx_GetItemInt(__pyx_v_value, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 800, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_15 = __Pyx_PyObject_Unicode(__pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 800, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_3 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_16))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_16);
            assert(__pyx_t_5);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_16);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_16, __pyx__function);
            __pyx_t_3 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_15};
            __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_16, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 800, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
          }
          __Pyx_XDECREF_SET(__pyx_v_size, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":801
 *                         price = market.c_quantize_order_price(self.trading_pair, price)
 *                         size = Decimal(str(value[2]))
 *                         size = market.c_quantize_order_amount(self.trading_pair, size)             # <<<<<<<<<<<<<<
 *                         if size > 0 and price > 0:
 *                             sells.append(PriceSize(price, size))
*/
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 801, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(0, 801, __pyx_L1_error)
          __pyx_t_16 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_2), __pyx_v_size, NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 801, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_size, __pyx_t_16);
          __pyx_t_16 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":802
 *                         size = Decimal(str(value[2]))
 *                         size = market.c_quantize_order_amount(self.trading_pair, size)
 *                         if size > 0 and price > 0:             # <<<<<<<<<<<<<<
 *                             sells.append(PriceSize(price, size))
 *         else:
*/
          __pyx_t_16 = PyObject_RichCompare(__pyx_v_size, __pyx_mstate_global->__pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_16); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 802, __pyx_L1_error)
          __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_16); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 802, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          if (__pyx_t_4) {
          } else {
            __pyx_t_8 = __pyx_t_4;
            goto __pyx_L23_bool_binop_done;
          }
          __pyx_t_16 = PyObject_RichCompare(__pyx_v_price, __pyx_mstate_global->__pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_16); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 802, __pyx_L1_error)
          __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_16); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 802, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_8 = __pyx_t_4;
          __pyx_L23_bool_binop_done:;
          if (__pyx_t_8) {

            /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":803
 *                         size = market.c_quantize_order_amount(self.trading_pair, size)
 *                         if size > 0 and price > 0:
 *                             sells.append(PriceSize(price, size))             # <<<<<<<<<<<<<<
 *         else:
 *             if not buy_reference_price.is_nan():
*/
            __pyx_t_2 = NULL;
            __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_PriceSize); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 803, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_3 = 1;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_15))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_15);
              assert(__pyx_t_2);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_15);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_15, __pyx__function);
              __pyx_t_3 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_price, __pyx_v_size};
              __pyx_t_16 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_15, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 803, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_16);
            }
            __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_sells, __pyx_t_16); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 803, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

            /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":802
 *                         size = Decimal(str(value[2]))
 *                         size = market.c_quantize_order_amount(self.trading_pair, size)
 *                         if size > 0 and price > 0:             # <<<<<<<<<<<<<<
 *                             sells.append(PriceSize(price, size))
 *         else:
*/
          }

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":797
 *                         if size > 0 and price > 0:
 *                             buys.append(PriceSize(price, size))
 *                     elif str(value[0]) == "sell" and not sell_reference_price.is_nan():             # <<<<<<<<<<<<<<
 *                         price = sell_reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))
 *                         price = market.c_quantize_order_price(self.trading_pair, price)
*/
        }
        __pyx_L14:;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":789
 *         if order_override is not None and len(order_override) > 0:
 *             for key, value in order_override.items():
 *                 if str(value[0]) in ["buy", "sell"]:             # <<<<<<<<<<<<<<
 *                     if str(value[0]) == "buy" and not buy_reference_price.is_nan():
 *                         price = buy_reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))
*/
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":787
 *         # to order spread, amount, and levels setting.
 *         order_override = self._order_override
 *         if order_override is not None and len(order_override) > 0:             # <<<<<<<<<<<<<<
 *             for key, value in order_override.items():
 *                 if str(value[0]) in ["buy", "sell"]:
*/
    goto __pyx_L6;
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":805
 *                             sells.append(PriceSize(price, size))
 *         else:
 *             if not buy_reference_price.is_nan():             # <<<<<<<<<<<<<<
 *                 for level in range(0, self._buy_levels):
 *                     price = buy_reference_price * (Decimal("1") - self._bid_spread - (level * self._order_level_spread))
*/
  /*else*/ {
    __pyx_t_16 = __pyx_v_buy_reference_price;
    __Pyx_INCREF(__pyx_t_16);
    __pyx_t_3 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_16, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_nan, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 805, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 805, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = (!__pyx_t_8);
    if (__pyx_t_4) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":806
 *         else:
 *             if not buy_reference_price.is_nan():
 *                 for level in range(0, self._buy_levels):             # <<<<<<<<<<<<<<
 *                     price = buy_reference_price * (Decimal("1") - self._bid_spread - (level * self._order_level_spread))
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
*/
      __pyx_t_16 = NULL;
      __pyx_t_15 = __Pyx_PyLong_From_int(__pyx_v_self->_buy_levels); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 806, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_3 = 1;
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_16, __pyx_mstate_global->__pyx_int_0, __pyx_t_15};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(&PyRange_Type), __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 806, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __pyx_t_15 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 806, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_18 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_15); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 806, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      for (;;) {
        {
          __pyx_t_1 = __pyx_t_18(__pyx_t_15);
          if (unlikely(!__pyx_t_1)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 806, __pyx_L1_error)
              PyErr_Clear();
            }
            break;
          }
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_level, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":807
 *             if not buy_reference_price.is_nan():
 *                 for level in range(0, self._buy_levels):
 *                     price = buy_reference_price * (Decimal("1") - self._bid_spread - (level * self._order_level_spread))             # <<<<<<<<<<<<<<
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
 *                     size = self._order_amount + (self._order_level_amount * level)
*/
        __pyx_t_16 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 807, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_2);
          assert(__pyx_t_16);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_16);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
          __pyx_t_3 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_mstate_global->__pyx_kp_u_1};
          __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 807, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __pyx_t_2 = PyNumber_Subtract(__pyx_t_1, __pyx_v_self->_bid_spread); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 807, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyNumber_Multiply(__pyx_v_level, __pyx_v_self->_order_level_spread); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 807, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_16 = PyNumber_Subtract(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 807, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyNumber_Multiply(__pyx_v_buy_reference_price, __pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 807, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF_SET(__pyx_v_price, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":808
 *                 for level in range(0, self._buy_levels):
 *                     price = buy_reference_price * (Decimal("1") - self._bid_spread - (level * self._order_level_spread))
 *                     price = market.c_quantize_order_price(self.trading_pair, price)             # <<<<<<<<<<<<<<
 *                     size = self._order_amount + (self._order_level_amount * level)
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
*/
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 808, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 808, __pyx_L1_error)
        __pyx_t_16 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_1), __pyx_v_price); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 808, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_price, __pyx_t_16);
        __pyx_t_16 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":809
 *                     price = buy_reference_price * (Decimal("1") - self._bid_spread - (level * self._order_level_spread))
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
 *                     size = self._order_amount + (self._order_level_amount * level)             # <<<<<<<<<<<<<<
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
 *                     if size > 0:
*/
        __pyx_t_16 = PyNumber_Multiply(__pyx_v_self->_order_level_amount, __pyx_v_level); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 809, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_1 = PyNumber_Add(__pyx_v_self->_order_amount, __pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 809, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF_SET(__pyx_v_size, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":810
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
 *                     size = self._order_amount + (self._order_level_amount * level)
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)             # <<<<<<<<<<<<<<
 *                     if size > 0:
 *                         buys.append(PriceSize(price, size))
*/
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 810, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 810, __pyx_L1_error)
        __pyx_t_16 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_1), __pyx_v_size, NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 810, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_size, __pyx_t_16);
        __pyx_t_16 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":811
 *                     size = self._order_amount + (self._order_level_amount * level)
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
 *                     if size > 0:             # <<<<<<<<<<<<<<
 *                         buys.append(PriceSize(price, size))
 *             if not sell_reference_price.is_nan():
*/
        __pyx_t_16 = PyObject_RichCompare(__pyx_v_size, __pyx_mstate_global->__pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_16); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 811, __pyx_L1_error)
        __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_16); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 811, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        if (__pyx_t_4) {

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":812
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
 *                     if size > 0:
 *                         buys.append(PriceSize(price, size))             # <<<<<<<<<<<<<<
 *             if not sell_reference_price.is_nan():
 *                 for level in range(0, self._sell_levels):
*/
          __pyx_t_1 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PriceSize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 812, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            assert(__pyx_t_1);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
            __pyx_t_3 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_price, __pyx_v_size};
            __pyx_t_16 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 812, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
          }
          __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_buys, __pyx_t_16); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 812, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":811
 *                     size = self._order_amount + (self._order_level_amount * level)
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
 *                     if size > 0:             # <<<<<<<<<<<<<<
 *                         buys.append(PriceSize(price, size))
 *             if not sell_reference_price.is_nan():
*/
        }

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":806
 *         else:
 *             if not buy_reference_price.is_nan():
 *                 for level in range(0, self._buy_levels):             # <<<<<<<<<<<<<<
 *                     price = buy_reference_price * (Decimal("1") - self._bid_spread - (level * self._order_level_spread))
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
*/
      }
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":805
 *                             sells.append(PriceSize(price, size))
 *         else:
 *             if not buy_reference_price.is_nan():             # <<<<<<<<<<<<<<
 *                 for level in range(0, self._buy_levels):
 *                     price = buy_reference_price * (Decimal("1") - self._bid_spread - (level * self._order_level_spread))
*/
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":813
 *                     if size > 0:
 *                         buys.append(PriceSize(price, size))
 *             if not sell_reference_price.is_nan():             # <<<<<<<<<<<<<<
 *                 for level in range(0, self._sell_levels):
 *                     price = sell_reference_price * (Decimal("1") + self._ask_spread + (level * self._order_level_spread))
*/
    __pyx_t_16 = __pyx_v_sell_reference_price;
    __Pyx_INCREF(__pyx_t_16);
    __pyx_t_3 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_16, NULL};
      __pyx_t_15 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_nan, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 813, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
    }
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_15); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_8 = (!__pyx_t_4);
    if (__pyx_t_8) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":814
 *                         buys.append(PriceSize(price, size))
 *             if not sell_reference_price.is_nan():
 *                 for level in range(0, self._sell_levels):             # <<<<<<<<<<<<<<
 *                     price = sell_reference_price * (Decimal("1") + self._ask_spread + (level * self._order_level_spread))
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
*/
      __pyx_t_16 = NULL;
      __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_v_self->_sell_levels); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 814, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = 1;
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_16, __pyx_mstate_global->__pyx_int_0, __pyx_t_2};
        __pyx_t_15 = __Pyx_PyObject_FastCall((PyObject*)(&PyRange_Type), __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 814, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
      }
      __pyx_t_2 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 814, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_18 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 814, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      for (;;) {
        {
          __pyx_t_15 = __pyx_t_18(__pyx_t_2);
          if (unlikely(!__pyx_t_15)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 814, __pyx_L1_error)
              PyErr_Clear();
            }
            break;
          }
        }
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_XDECREF_SET(__pyx_v_level, __pyx_t_15);
        __pyx_t_15 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":815
 *             if not sell_reference_price.is_nan():
 *                 for level in range(0, self._sell_levels):
 *                     price = sell_reference_price * (Decimal("1") + self._ask_spread + (level * self._order_level_spread))             # <<<<<<<<<<<<<<
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
 *                     size = self._order_amount + (self._order_level_amount * level)
*/
        __pyx_t_16 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 815, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_1);
          assert(__pyx_t_16);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_16);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
          __pyx_t_3 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_mstate_global->__pyx_kp_u_1};
          __pyx_t_15 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 815, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
        }
        __pyx_t_1 = PyNumber_Add(__pyx_t_15, __pyx_v_self->_ask_spread); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 815, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_15 = PyNumber_Multiply(__pyx_v_level, __pyx_v_self->_order_level_spread); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 815, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __pyx_t_16 = PyNumber_Add(__pyx_t_1, __pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 815, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_15 = PyNumber_Multiply(__pyx_v_sell_reference_price, __pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 815, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF_SET(__pyx_v_price, __pyx_t_15);
        __pyx_t_15 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":816
 *                 for level in range(0, self._sell_levels):
 *                     price = sell_reference_price * (Decimal("1") + self._ask_spread + (level * self._order_level_spread))
 *                     price = market.c_quantize_order_price(self.trading_pair, price)             # <<<<<<<<<<<<<<
 *                     size = self._order_amount + (self._order_level_amount * level)
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
*/
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 816, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_15))||((__pyx_t_15) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_15))) __PYX_ERR(0, 816, __pyx_L1_error)
        __pyx_t_16 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_15), __pyx_v_price); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 816, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF_SET(__pyx_v_price, __pyx_t_16);
        __pyx_t_16 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":817
 *                     price = sell_reference_price * (Decimal("1") + self._ask_spread + (level * self._order_level_spread))
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
 *                     size = self._order_amount + (self._order_level_amount * level)             # <<<<<<<<<<<<<<
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
 *                     if size > 0:
*/
        __pyx_t_16 = PyNumber_Multiply(__pyx_v_self->_order_level_amount, __pyx_v_level); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 817, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_15 = PyNumber_Add(__pyx_v_self->_order_amount, __pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 817, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF_SET(__pyx_v_size, __pyx_t_15);
        __pyx_t_15 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":818
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
 *                     size = self._order_amount + (self._order_level_amount * level)
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)             # <<<<<<<<<<<<<<
 *                     if size > 0:
 *                         sells.append(PriceSize(price, size))
*/
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 818, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_15))||((__pyx_t_15) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_15))) __PYX_ERR(0, 818, __pyx_L1_error)
        __pyx_t_16 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_15), __pyx_v_size, NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 818, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF_SET(__pyx_v_size, __pyx_t_16);
        __pyx_t_16 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":819
 *                     size = self._order_amount + (self._order_level_amount * level)
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
 *                     if size > 0:             # <<<<<<<<<<<<<<
 *                         sells.append(PriceSize(price, size))
 * 
*/
        __pyx_t_16 = PyObject_RichCompare(__pyx_v_size, __pyx_mstate_global->__pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_16); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 819, __pyx_L1_error)
        __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_16); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 819, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        if (__pyx_t_8) {

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":820
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
 *                     if size > 0:
 *                         sells.append(PriceSize(price, size))             # <<<<<<<<<<<<<<
 * 
 *         return Proposal(buys, sells)
*/
          __pyx_t_15 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_PriceSize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 820, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_1);
            assert(__pyx_t_15);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
            __pyx_t_3 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_15, __pyx_v_price, __pyx_v_size};
            __pyx_t_16 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 820, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
          }
          __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_sells, __pyx_t_16); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 820, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":819
 *                     size = self._order_amount + (self._order_level_amount * level)
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
 *                     if size > 0:             # <<<<<<<<<<<<<<
 *                         sells.append(PriceSize(price, size))
 * 
*/
        }

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":814
 *                         buys.append(PriceSize(price, size))
 *             if not sell_reference_price.is_nan():
 *                 for level in range(0, self._sell_levels):             # <<<<<<<<<<<<<<
 *                     price = sell_reference_price * (Decimal("1") + self._ask_spread + (level * self._order_level_spread))
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
*/
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":813
 *                     if size > 0:
 *                         buys.append(PriceSize(price, size))
 *             if not sell_reference_price.is_nan():             # <<<<<<<<<<<<<<
 *                 for level in range(0, self._sell_levels):
 *                     price = sell_reference_price * (Decimal("1") + self._ask_spread + (level * self._order_level_spread))
*/
    }
  }
  __pyx_L6:;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":822
 *                         sells.append(PriceSize(price, size))
 * 
 *         return Proposal(buys, sells)             # <<<<<<<<<<<<<<
 * 
 *     cdef tuple c_get_adjusted_available_balance(self, list orders):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_16 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_Proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_1);
    assert(__pyx_t_16);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_16);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
    __pyx_t_3 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_16, __pyx_v_buys, __pyx_v_sells};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 822, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":766
 *             self._last_timestamp = timestamp
 * 
 *     cdef object c_create_base_proposal(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_create_base_proposal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_buys);
  __Pyx_XDECREF(__pyx_v_sells);
  __Pyx_XDECREF(__pyx_v_buy_reference_price);
  __Pyx_XDECREF(__pyx_v_sell_reference_price);
  __Pyx_XDECREF(__pyx_v_inventory_cost_price);
  __Pyx_XDECREF(__pyx_v_order_override);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XDECREF(__pyx_v_level);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":824
 *         return Proposal(buys, sells)
 * 
 *     cdef tuple c_get_adjusted_available_balance(self, list orders):             # <<<<<<<<<<<<<<
 *         """
 *         Calculates the available balance, plus the amount attributed to orders.
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_get_adjusted_available_balance(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_orders) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_base_balance = 0;
  PyObject *__pyx_v_quote_balance = 0;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_get_adjusted_available_balance", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":830
 *         """
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             object base_balance = market.c_get_available_balance(self.base_asset)
 *             object quote_balance = market.c_get_available_balance(self.quote_asset)
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 830, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":831
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             object base_balance = market.c_get_available_balance(self.base_asset)             # <<<<<<<<<<<<<<
 *             object quote_balance = market.c_get_available_balance(self.quote_asset)
 * 
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 831, __pyx_L1_error)
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_available_balance(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_base_balance = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":832
 *             ExchangeBase market = self._market_info.market
 *             object base_balance = market.c_get_available_balance(self.base_asset)
 *             object quote_balance = market.c_get_available_balance(self.quote_asset)             # <<<<<<<<<<<<<<
 * 
 *         for order in orders:
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_quote_asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(0, 832, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_available_balance(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_quote_balance = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":834
 *             object quote_balance = market.c_get_available_balance(self.quote_asset)
 * 
 *         for order in orders:             # <<<<<<<<<<<<<<
 *             if order.is_buy:
 *                 quote_balance += order.quantity * order.price
*/
  if (unlikely(__pyx_v_orders == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 834, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_orders; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 834, __pyx_L1_error)
      #endif
      if (__pyx_t_3 >= __pyx_temp) break;
    }
    __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
    ++__pyx_t_3;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 834, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":835
 * 
 *         for order in orders:
 *             if order.is_buy:             # <<<<<<<<<<<<<<
 *                 quote_balance += order.quantity * order.price
 *             else:
*/
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 835, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 835, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_4) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":836
 *         for order in orders:
 *             if order.is_buy:
 *                 quote_balance += order.quantity * order.price             # <<<<<<<<<<<<<<
 *             else:
 *                 base_balance += order.quantity
*/
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_quantity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyNumber_Multiply(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_quote_balance, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF_SET(__pyx_v_quote_balance, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":835
 * 
 *         for order in orders:
 *             if order.is_buy:             # <<<<<<<<<<<<<<
 *                 quote_balance += order.quantity * order.price
 *             else:
*/
      goto __pyx_L5;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":838
 *                 quote_balance += order.quantity * order.price
 *             else:
 *                 base_balance += order.quantity             # <<<<<<<<<<<<<<
 * 
 *         return base_balance, quote_balance
*/
    /*else*/ {
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_quantity); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 838, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_base_balance, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 838, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_base_balance, __pyx_t_6);
      __pyx_t_6 = 0;
    }
    __pyx_L5:;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":834
 *             object quote_balance = market.c_get_available_balance(self.quote_asset)
 * 
 *         for order in orders:             # <<<<<<<<<<<<<<
 *             if order.is_buy:
 *                 quote_balance += order.quantity * order.price
*/
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":840
 *                 base_balance += order.quantity
 * 
 *         return base_balance, quote_balance             # <<<<<<<<<<<<<<
 * 
 *     cdef c_apply_order_levels_modifiers(self, proposal):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_base_balance);
  __Pyx_GIVEREF(__pyx_v_base_balance);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_base_balance) != (0)) __PYX_ERR(0, 840, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_quote_balance);
  __Pyx_GIVEREF(__pyx_v_quote_balance);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_quote_balance) != (0)) __PYX_ERR(0, 840, __pyx_L1_error);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":824
 *         return Proposal(buys, sells)
 * 
 *     cdef tuple c_get_adjusted_available_balance(self, list orders):             # <<<<<<<<<<<<<<
 *         """
 *         Calculates the available balance, plus the amount attributed to orders.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_get_adjusted_available_balance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_base_balance);
  __Pyx_XDECREF(__pyx_v_quote_balance);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":842
 *         return base_balance, quote_balance
 * 
 *     cdef c_apply_order_levels_modifiers(self, proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_price_band(proposal)
 *         if self.moving_price_band_enabled:
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_order_levels_modifiers(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_order_levels_modifiers", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":843
 * 
 *     cdef c_apply_order_levels_modifiers(self, proposal):
 *         self.c_apply_price_band(proposal)             # <<<<<<<<<<<<<<
 *         if self.moving_price_band_enabled:
 *             self.c_apply_moving_price_band(proposal)
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_price_band(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":844
 *     cdef c_apply_order_levels_modifiers(self, proposal):
 *         self.c_apply_price_band(proposal)
 *         if self.moving_price_band_enabled:             # <<<<<<<<<<<<<<
 *             self.c_apply_moving_price_band(proposal)
 *         if self._ping_pong_enabled:
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_moving_price_band_enabled); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":845
 *         self.c_apply_price_band(proposal)
 *         if self.moving_price_band_enabled:
 *             self.c_apply_moving_price_band(proposal)             # <<<<<<<<<<<<<<
 *         if self._ping_pong_enabled:
 *             self.c_apply_ping_pong(proposal)
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_moving_price_band(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 845, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":844
 *     cdef c_apply_order_levels_modifiers(self, proposal):
 *         self.c_apply_price_band(proposal)
 *         if self.moving_price_band_enabled:             # <<<<<<<<<<<<<<
 *             self.c_apply_moving_price_band(proposal)
 *         if self._ping_pong_enabled:
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":846
 *         if self.moving_price_band_enabled:
 *             self.c_apply_moving_price_band(proposal)
 *         if self._ping_pong_enabled:             # <<<<<<<<<<<<<<
 *             self.c_apply_ping_pong(proposal)
 * 
*/
  if (__pyx_v_self->_ping_pong_enabled) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":847
 *             self.c_apply_moving_price_band(proposal)
 *         if self._ping_pong_enabled:
 *             self.c_apply_ping_pong(proposal)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_apply_price_band(self, proposal):
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_ping_pong(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":846
 *         if self.moving_price_band_enabled:
 *             self.c_apply_moving_price_band(proposal)
 *         if self._ping_pong_enabled:             # <<<<<<<<<<<<<<
 *             self.c_apply_ping_pong(proposal)
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":842
 *         return base_balance, quote_balance
 * 
 *     cdef c_apply_order_levels_modifiers(self, proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_price_band(proposal)
 *         if self.moving_price_band_enabled:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_order_levels_modifiers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":849
 *             self.c_apply_ping_pong(proposal)
 * 
 *     cdef c_apply_price_band(self, proposal):             # <<<<<<<<<<<<<<
 *         if self._price_ceiling > 0 and self.get_price() >= self._price_ceiling:
 *             proposal.buys = []
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_price_band(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_price_band", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":850
 * 
 *     cdef c_apply_price_band(self, proposal):
 *         if self._price_ceiling > 0 and self.get_price() >= self._price_ceiling:             # <<<<<<<<<<<<<<
 *             proposal.buys = []
 *         if self._price_floor > 0 and self.get_price() <= self._price_floor:
*/
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_self->_price_ceiling, __pyx_mstate_global->__pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 850, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 850, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_v_self->_price_ceiling, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":851
 *     cdef c_apply_price_band(self, proposal):
 *         if self._price_ceiling > 0 and self.get_price() >= self._price_ceiling:
 *             proposal.buys = []             # <<<<<<<<<<<<<<
 *         if self._price_floor > 0 and self.get_price() <= self._price_floor:
 *             proposal.sells = []
*/
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys, __pyx_t_4) < (0)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":850
 * 
 *     cdef c_apply_price_band(self, proposal):
 *         if self._price_ceiling > 0 and self.get_price() >= self._price_ceiling:             # <<<<<<<<<<<<<<
 *             proposal.buys = []
 *         if self._price_floor > 0 and self.get_price() <= self._price_floor:
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":852
 *         if self._price_ceiling > 0 and self.get_price() >= self._price_ceiling:
 *             proposal.buys = []
 *         if self._price_floor > 0 and self.get_price() <= self._price_floor:             # <<<<<<<<<<<<<<
 *             proposal.sells = []
 * 
*/
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_self->_price_floor, __pyx_mstate_global->__pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 852, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_v_self->_price_floor, Py_LE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":853
 *             proposal.buys = []
 *         if self._price_floor > 0 and self.get_price() <= self._price_floor:
 *             proposal.sells = []             # <<<<<<<<<<<<<<
 * 
 *     cdef c_apply_moving_price_band(self, proposal):
*/
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 853, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells, __pyx_t_2) < (0)) __PYX_ERR(0, 853, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":852
 *         if self._price_ceiling > 0 and self.get_price() >= self._price_ceiling:
 *             proposal.buys = []
 *         if self._price_floor > 0 and self.get_price() <= self._price_floor:             # <<<<<<<<<<<<<<
 *             proposal.sells = []
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":849
 *             self.c_apply_ping_pong(proposal)
 * 
 *     cdef c_apply_price_band(self, proposal):             # <<<<<<<<<<<<<<
 *         if self._price_ceiling > 0 and self.get_price() >= self._price_ceiling:
 *             proposal.buys = []
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_price_band", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":855
 *             proposal.sells = []
 * 
 *     cdef c_apply_moving_price_band(self, proposal):             # <<<<<<<<<<<<<<
 *         price = self.get_price()
 *         self._moving_price_band.check_and_update_price_band(
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_moving_price_band(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_moving_price_band", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":856
 * 
 *     cdef c_apply_moving_price_band(self, proposal):
 *         price = self.get_price()             # <<<<<<<<<<<<<<
 *         self._moving_price_band.check_and_update_price_band(
 *             self.current_timestamp, price)
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 856, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":857
 *     cdef c_apply_moving_price_band(self, proposal):
 *         price = self.get_price()
 *         self._moving_price_band.check_and_update_price_band(             # <<<<<<<<<<<<<<
 *             self.current_timestamp, price)
 *         if self._moving_price_band.check_price_ceiling_exceeded(price):
*/
  __pyx_t_2 = __pyx_v_self->_moving_price_band;
  __Pyx_INCREF(__pyx_t_2);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":858
 *         price = self.get_price()
 *         self._moving_price_band.check_and_update_price_band(
 *             self.current_timestamp, price)             # <<<<<<<<<<<<<<
 *         if self._moving_price_band.check_price_ceiling_exceeded(price):
 *             proposal.buys = []
*/
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_current_timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_4, __pyx_v_price};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_and_update_price_band, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 857, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":859
 *         self._moving_price_band.check_and_update_price_band(
 *             self.current_timestamp, price)
 *         if self._moving_price_band.check_price_ceiling_exceeded(price):             # <<<<<<<<<<<<<<
 *             proposal.buys = []
 *         if self._moving_price_band.check_price_floor_exceeded(price):
*/
  __pyx_t_4 = __pyx_v_self->_moving_price_band;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_price};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_price_ceiling_exceeded, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":860
 *             self.current_timestamp, price)
 *         if self._moving_price_band.check_price_ceiling_exceeded(price):
 *             proposal.buys = []             # <<<<<<<<<<<<<<
 *         if self._moving_price_band.check_price_floor_exceeded(price):
 *             proposal.sells = []
*/
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys, __pyx_t_1) < (0)) __PYX_ERR(0, 860, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":859
 *         self._moving_price_band.check_and_update_price_band(
 *             self.current_timestamp, price)
 *         if self._moving_price_band.check_price_ceiling_exceeded(price):             # <<<<<<<<<<<<<<
 *             proposal.buys = []
 *         if self._moving_price_band.check_price_floor_exceeded(price):
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":861
 *         if self._moving_price_band.check_price_ceiling_exceeded(price):
 *             proposal.buys = []
 *         if self._moving_price_band.check_price_floor_exceeded(price):             # <<<<<<<<<<<<<<
 *             proposal.sells = []
 * 
*/
  __pyx_t_4 = __pyx_v_self->_moving_price_band;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_price};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_check_price_floor_exceeded, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":862
 *             proposal.buys = []
 *         if self._moving_price_band.check_price_floor_exceeded(price):
 *             proposal.sells = []             # <<<<<<<<<<<<<<
 * 
 *     cdef c_apply_ping_pong(self, object proposal):
*/
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells, __pyx_t_1) < (0)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":861
 *         if self._moving_price_band.check_price_ceiling_exceeded(price):
 *             proposal.buys = []
 *         if self._moving_price_band.check_price_floor_exceeded(price):             # <<<<<<<<<<<<<<
 *             proposal.sells = []
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":855
 *             proposal.sells = []
 * 
 *     cdef c_apply_moving_price_band(self, proposal):             # <<<<<<<<<<<<<<
 *         price = self.get_price()
 *         self._moving_price_band.check_and_update_price_band(
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_moving_price_band", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":864
 *             proposal.sells = []
 * 
 *     cdef c_apply_ping_pong(self, object proposal):             # <<<<<<<<<<<<<<
 *         self._ping_pong_warning_lines = []
 *         if self._filled_buys_balance == self._filled_sells_balance:
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_ping_pong(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4[3];
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_ping_pong", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":865
 * 
 *     cdef c_apply_ping_pong(self, object proposal):
 *         self._ping_pong_warning_lines = []             # <<<<<<<<<<<<<<
 *         if self._filled_buys_balance == self._filled_sells_balance:
 *             self._filled_buys_balance = self._filled_sells_balance = 0
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_ping_pong_warning_lines);
  __Pyx_DECREF(__pyx_v_self->_ping_pong_warning_lines);
  __pyx_v_self->_ping_pong_warning_lines = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":866
 *     cdef c_apply_ping_pong(self, object proposal):
 *         self._ping_pong_warning_lines = []
 *         if self._filled_buys_balance == self._filled_sells_balance:             # <<<<<<<<<<<<<<
 *             self._filled_buys_balance = self._filled_sells_balance = 0
 *         if self._filled_buys_balance > 0:
*/
  __pyx_t_2 = (__pyx_v_self->_filled_buys_balance == __pyx_v_self->_filled_sells_balance);
  if (__pyx_t_2) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":867
 *         self._ping_pong_warning_lines = []
 *         if self._filled_buys_balance == self._filled_sells_balance:
 *             self._filled_buys_balance = self._filled_sells_balance = 0             # <<<<<<<<<<<<<<
 *         if self._filled_buys_balance > 0:
 *             proposal.buys = proposal.buys[self._filled_buys_balance:]
*/
    __pyx_v_self->_filled_buys_balance = 0;
    __pyx_v_self->_filled_sells_balance = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":866
 *     cdef c_apply_ping_pong(self, object proposal):
 *         self._ping_pong_warning_lines = []
 *         if self._filled_buys_balance == self._filled_sells_balance:             # <<<<<<<<<<<<<<
 *             self._filled_buys_balance = self._filled_sells_balance = 0
 *         if self._filled_buys_balance > 0:
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":868
 *         if self._filled_buys_balance == self._filled_sells_balance:
 *             self._filled_buys_balance = self._filled_sells_balance = 0
 *         if self._filled_buys_balance > 0:             # <<<<<<<<<<<<<<
 *             proposal.buys = proposal.buys[self._filled_buys_balance:]
 *             self._ping_pong_warning_lines.extend(
*/
  __pyx_t_2 = (__pyx_v_self->_filled_buys_balance > 0);
  if (__pyx_t_2) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":869
 *             self._filled_buys_balance = self._filled_sells_balance = 0
 *         if self._filled_buys_balance > 0:
 *             proposal.buys = proposal.buys[self._filled_buys_balance:]             # <<<<<<<<<<<<<<
 *             self._ping_pong_warning_lines.extend(
 *                 [f"  Ping-pong removed {self._filled_buys_balance} buy orders."]
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_1, __pyx_v_self->_filled_buys_balance, 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys, __pyx_t_3) < (0)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":870
 *         if self._filled_buys_balance > 0:
 *             proposal.buys = proposal.buys[self._filled_buys_balance:]
 *             self._ping_pong_warning_lines.extend(             # <<<<<<<<<<<<<<
 *                 [f"  Ping-pong removed {self._filled_buys_balance} buy orders."]
 *             )
*/
    if (unlikely(__pyx_v_self->_ping_pong_warning_lines == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
      __PYX_ERR(0, 870, __pyx_L1_error)
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":871
 *             proposal.buys = proposal.buys[self._filled_buys_balance:]
 *             self._ping_pong_warning_lines.extend(
 *                 [f"  Ping-pong removed {self._filled_buys_balance} buy orders."]             # <<<<<<<<<<<<<<
 *             )
 *         if self._filled_sells_balance > 0:
*/
    __pyx_t_3 = __Pyx_PyUnicode_From_int(__pyx_v_self->_filled_buys_balance, 0, ' ', 'd'); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 871, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4[0] = __pyx_mstate_global->__pyx_kp_u_Ping_pong_removed;
    __pyx_t_4[1] = __pyx_t_3;
    __pyx_t_4[2] = __pyx_mstate_global->__pyx_kp_u_buy_orders;
    __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_4, 3, 20 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 12, 127);
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 871, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":870
 *         if self._filled_buys_balance > 0:
 *             proposal.buys = proposal.buys[self._filled_buys_balance:]
 *             self._ping_pong_warning_lines.extend(             # <<<<<<<<<<<<<<
 *                 [f"  Ping-pong removed {self._filled_buys_balance} buy orders."]
 *             )
*/
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_self->_ping_pong_warning_lines, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 870, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":868
 *         if self._filled_buys_balance == self._filled_sells_balance:
 *             self._filled_buys_balance = self._filled_sells_balance = 0
 *         if self._filled_buys_balance > 0:             # <<<<<<<<<<<<<<
 *             proposal.buys = proposal.buys[self._filled_buys_balance:]
 *             self._ping_pong_warning_lines.extend(
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":873
 *                 [f"  Ping-pong removed {self._filled_buys_balance} buy orders."]
 *             )
 *         if self._filled_sells_balance > 0:             # <<<<<<<<<<<<<<
 *             proposal.sells = proposal.sells[self._filled_sells_balance:]
 *             self._ping_pong_warning_lines.extend(
*/
  __pyx_t_2 = (__pyx_v_self->_filled_sells_balance > 0);
  if (__pyx_t_2) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":874
 *             )
 *         if self._filled_sells_balance > 0:
 *             proposal.sells = proposal.sells[self._filled_sells_balance:]             # <<<<<<<<<<<<<<
 *             self._ping_pong_warning_lines.extend(
 *                 [f"  Ping-pong removed {self._filled_sells_balance} sell orders."]
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_1, __pyx_v_self->_filled_sells_balance, 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells, __pyx_t_3) < (0)) __PYX_ERR(0, 874, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":875
 *         if self._filled_sells_balance > 0:
 *             proposal.sells = proposal.sells[self._filled_sells_balance:]
 *             self._ping_pong_warning_lines.extend(             # <<<<<<<<<<<<<<
 *                 [f"  Ping-pong removed {self._filled_sells_balance} sell orders."]
 *             )
*/
    if (unlikely(__pyx_v_self->_ping_pong_warning_lines == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
      __PYX_ERR(0, 875, __pyx_L1_error)
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":876
 *             proposal.sells = proposal.sells[self._filled_sells_balance:]
 *             self._ping_pong_warning_lines.extend(
 *                 [f"  Ping-pong removed {self._filled_sells_balance} sell orders."]             # <<<<<<<<<<<<<<
 *             )
 * 
*/
    __pyx_t_3 = __Pyx_PyUnicode_From_int(__pyx_v_self->_filled_sells_balance, 0, ' ', 'd'); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4[0] = __pyx_mstate_global->__pyx_kp_u_Ping_pong_removed;
    __pyx_t_4[1] = __pyx_t_3;
    __pyx_t_4[2] = __pyx_mstate_global->__pyx_kp_u_sell_orders;
    __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_4, 3, 20 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3) + 13, 127);
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":875
 *         if self._filled_sells_balance > 0:
 *             proposal.sells = proposal.sells[self._filled_sells_balance:]
 *             self._ping_pong_warning_lines.extend(             # <<<<<<<<<<<<<<
 *                 [f"  Ping-pong removed {self._filled_sells_balance} sell orders."]
 *             )
*/
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_self->_ping_pong_warning_lines, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 875, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":873
 *                 [f"  Ping-pong removed {self._filled_buys_balance} buy orders."]
 *             )
 *         if self._filled_sells_balance > 0:             # <<<<<<<<<<<<<<
 *             proposal.sells = proposal.sells[self._filled_sells_balance:]
 *             self._ping_pong_warning_lines.extend(
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":864
 *             proposal.sells = []
 * 
 *     cdef c_apply_ping_pong(self, object proposal):             # <<<<<<<<<<<<<<
 *         self._ping_pong_warning_lines = []
 *         if self._filled_buys_balance == self._filled_sells_balance:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_ping_pong", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":879
 *             )
 * 
 *     cdef c_apply_order_price_modifiers(self, object proposal):             # <<<<<<<<<<<<<<
 *         if self._order_optimization_enabled:
 *             self.c_apply_order_optimization(proposal)
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_order_price_modifiers(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_order_price_modifiers", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":880
 * 
 *     cdef c_apply_order_price_modifiers(self, object proposal):
 *         if self._order_optimization_enabled:             # <<<<<<<<<<<<<<
 *             self.c_apply_order_optimization(proposal)
 * 
*/
  if (__pyx_v_self->_order_optimization_enabled) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":881
 *     cdef c_apply_order_price_modifiers(self, object proposal):
 *         if self._order_optimization_enabled:
 *             self.c_apply_order_optimization(proposal)             # <<<<<<<<<<<<<<
 * 
 *         if self._add_transaction_costs_to_orders:
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_order_optimization(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 881, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":880
 * 
 *     cdef c_apply_order_price_modifiers(self, object proposal):
 *         if self._order_optimization_enabled:             # <<<<<<<<<<<<<<
 *             self.c_apply_order_optimization(proposal)
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":883
 *             self.c_apply_order_optimization(proposal)
 * 
 *         if self._add_transaction_costs_to_orders:             # <<<<<<<<<<<<<<
 *             self.c_apply_add_transaction_costs(proposal)
 * 
*/
  if (__pyx_v_self->_add_transaction_costs_to_orders) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":884
 * 
 *         if self._add_transaction_costs_to_orders:
 *             self.c_apply_add_transaction_costs(proposal)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_apply_order_size_modifiers(self, object proposal):
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_add_transaction_costs(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":883
 *             self.c_apply_order_optimization(proposal)
 * 
 *         if self._add_transaction_costs_to_orders:             # <<<<<<<<<<<<<<
 *             self.c_apply_add_transaction_costs(proposal)
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":879
 *             )
 * 
 *     cdef c_apply_order_price_modifiers(self, object proposal):             # <<<<<<<<<<<<<<
 *         if self._order_optimization_enabled:
 *             self.c_apply_order_optimization(proposal)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_order_price_modifiers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":886
 *             self.c_apply_add_transaction_costs(proposal)
 * 
 *     cdef c_apply_order_size_modifiers(self, object proposal):             # <<<<<<<<<<<<<<
 *         if self._inventory_skew_enabled:
 *             self.c_apply_inventory_skew(proposal)
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_order_size_modifiers(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_order_size_modifiers", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":887
 * 
 *     cdef c_apply_order_size_modifiers(self, object proposal):
 *         if self._inventory_skew_enabled:             # <<<<<<<<<<<<<<
 *             self.c_apply_inventory_skew(proposal)
 * 
*/
  if (__pyx_v_self->_inventory_skew_enabled) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":888
 *     cdef c_apply_order_size_modifiers(self, object proposal):
 *         if self._inventory_skew_enabled:
 *             self.c_apply_inventory_skew(proposal)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_apply_inventory_skew(self, object proposal):
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_inventory_skew(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":887
 * 
 *     cdef c_apply_order_size_modifiers(self, object proposal):
 *         if self._inventory_skew_enabled:             # <<<<<<<<<<<<<<
 *             self.c_apply_inventory_skew(proposal)
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":886
 *             self.c_apply_add_transaction_costs(proposal)
 * 
 *     cdef c_apply_order_size_modifiers(self, object proposal):             # <<<<<<<<<<<<<<
 *         if self._inventory_skew_enabled:
 *             self.c_apply_inventory_skew(proposal)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_order_size_modifiers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":890
 *             self.c_apply_inventory_skew(proposal)
 * 
 *     cdef c_apply_inventory_skew(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_inventory_skew(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_bid_adj_ratio = 0;
  PyObject *__pyx_v_ask_adj_ratio = 0;
  PyObject *__pyx_v_size = 0;
  PyObject *__pyx_v_base_balance = NULL;
  PyObject *__pyx_v_quote_balance = NULL;
  PyObject *__pyx_v_total_order_size = NULL;
  PyObject *__pyx_v_bid_ask_ratios = NULL;
  PyObject *__pyx_v_buy = NULL;
  PyObject *__pyx_v_sell = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  double __pyx_t_6;
  double __pyx_t_7;
  double __pyx_t_8;
  double __pyx_t_9;
  double __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_quantize_order_amount __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_inventory_skew", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":892
 *     cdef c_apply_inventory_skew(self, object proposal):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             object bid_adj_ratio
 *             object ask_adj_ratio
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 892, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":897
 *             object size
 * 
 *         base_balance, quote_balance = self.c_get_adjusted_available_balance(self.active_orders)             # <<<<<<<<<<<<<<
 * 
 *         total_order_size = calculate_total_order_size(self._order_amount, self._order_level_amount, self._order_levels)
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 897, __pyx_L1_error)
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_adjusted_available_balance(__pyx_v_self, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PyTuple_GET_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 897, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_1 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 897, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 897, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 897, __pyx_L1_error)
  }
  __pyx_v_base_balance = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_quote_balance = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":899
 *         base_balance, quote_balance = self.c_get_adjusted_available_balance(self.active_orders)
 * 
 *         total_order_size = calculate_total_order_size(self._order_amount, self._order_level_amount, self._order_levels)             # <<<<<<<<<<<<<<
 *         bid_ask_ratios = c_calculate_bid_ask_ratios_from_base_asset_ratio(
 *             float(base_balance),
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_calculate_total_order_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyLong_From_int(__pyx_v_self->_order_levels); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_self->_order_amount, __pyx_v_self->_order_level_amount, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_5, (4-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 899, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_total_order_size = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":901
 *         total_order_size = calculate_total_order_size(self._order_amount, self._order_level_amount, self._order_levels)
 *         bid_ask_ratios = c_calculate_bid_ask_ratios_from_base_asset_ratio(
 *             float(base_balance),             # <<<<<<<<<<<<<<
 *             float(quote_balance),
 *             float(self.get_price()),
*/
  __pyx_t_6 = __Pyx_PyObject_AsDouble(__pyx_v_base_balance); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_6, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 901, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":902
 *         bid_ask_ratios = c_calculate_bid_ask_ratios_from_base_asset_ratio(
 *             float(base_balance),
 *             float(quote_balance),             # <<<<<<<<<<<<<<
 *             float(self.get_price()),
 *             float(self._inventory_target_base_pct),
*/
  __pyx_t_7 = __Pyx_PyObject_AsDouble(__pyx_v_quote_balance); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_7, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 902, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":903
 *             float(base_balance),
 *             float(quote_balance),
 *             float(self.get_price()),             # <<<<<<<<<<<<<<
 *             float(self._inventory_target_base_pct),
 *             float(total_order_size * self._inventory_range_multiplier)
*/
  __pyx_t_1 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 903, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_8 = __Pyx_PyObject_AsDouble(__pyx_t_2); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_8, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":904
 *             float(quote_balance),
 *             float(self.get_price()),
 *             float(self._inventory_target_base_pct),             # <<<<<<<<<<<<<<
 *             float(total_order_size * self._inventory_range_multiplier)
 *         )
*/
  __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_v_self->_inventory_target_base_pct); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_9, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 904, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":905
 *             float(self.get_price()),
 *             float(self._inventory_target_base_pct),
 *             float(total_order_size * self._inventory_range_multiplier)             # <<<<<<<<<<<<<<
 *         )
 *         bid_adj_ratio = Decimal(bid_ask_ratios.bid_ratio)
*/
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_total_order_size, __pyx_v_self->_inventory_range_multiplier); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = __Pyx_PyObject_AsDouble(__pyx_t_2); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_10, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":900
 * 
 *         total_order_size = calculate_total_order_size(self._order_amount, self._order_level_amount, self._order_levels)
 *         bid_ask_ratios = c_calculate_bid_ask_ratios_from_base_asset_ratio(             # <<<<<<<<<<<<<<
 *             float(base_balance),
 *             float(quote_balance),
*/
  __pyx_t_2 = __pyx_f_10hummingbot_8strategy_18pure_market_making_25inventory_skew_calculator_c_calculate_bid_ask_ratios_from_base_asset_ratio(__pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 900, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_bid_ask_ratios = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":907
 *             float(total_order_size * self._inventory_range_multiplier)
 *         )
 *         bid_adj_ratio = Decimal(bid_ask_ratios.bid_ratio)             # <<<<<<<<<<<<<<
 *         ask_adj_ratio = Decimal(bid_ask_ratios.ask_ratio)
 * 
*/
  __pyx_t_1 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_bid_ask_ratios, __pyx_mstate_global->__pyx_n_u_bid_ratio); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_1);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_t_3};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_bid_adj_ratio = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":908
 *         )
 *         bid_adj_ratio = Decimal(bid_ask_ratios.bid_ratio)
 *         ask_adj_ratio = Decimal(bid_ask_ratios.ask_ratio)             # <<<<<<<<<<<<<<
 * 
 *         for buy in proposal.buys:
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_bid_ask_ratios, __pyx_mstate_global->__pyx_n_u_ask_ratio); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_1};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_ask_adj_ratio = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":910
 *         ask_adj_ratio = Decimal(bid_ask_ratios.ask_ratio)
 * 
 *         for buy in proposal.buys:             # <<<<<<<<<<<<<<
 *             size = buy.size * bid_adj_ratio
 *             size = market.c_quantize_order_amount(self.trading_pair, size)
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3);
    __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_11 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_12 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 910, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 910, __pyx_L1_error)
          #endif
          if (__pyx_t_11 >= __pyx_temp) break;
        }
        __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_11, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_11;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 910, __pyx_L1_error)
          #endif
          if (__pyx_t_11 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_11));
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_11);
        #endif
        ++__pyx_t_11;
      }
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 910, __pyx_L1_error)
    } else {
      __pyx_t_2 = __pyx_t_12(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 910, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_buy, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":911
 * 
 *         for buy in proposal.buys:
 *             size = buy.size * bid_adj_ratio             # <<<<<<<<<<<<<<
 *             size = market.c_quantize_order_amount(self.trading_pair, size)
 *             buy.size = size
*/
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyNumber_Multiply(__pyx_t_2, __pyx_v_bid_adj_ratio); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_size, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":912
 *         for buy in proposal.buys:
 *             size = buy.size * bid_adj_ratio
 *             size = market.c_quantize_order_amount(self.trading_pair, size)             # <<<<<<<<<<<<<<
 *             buy.size = size
 * 
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 912, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 912, __pyx_L1_error)
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_1), __pyx_v_size, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 912, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_size, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":913
 *             size = buy.size * bid_adj_ratio
 *             size = market.c_quantize_order_amount(self.trading_pair, size)
 *             buy.size = size             # <<<<<<<<<<<<<<
 * 
 *         for sell in proposal.sells:
*/
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_size, __pyx_v_size) < (0)) __PYX_ERR(0, 913, __pyx_L1_error)

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":910
 *         ask_adj_ratio = Decimal(bid_ask_ratios.ask_ratio)
 * 
 *         for buy in proposal.buys:             # <<<<<<<<<<<<<<
 *             size = buy.size * bid_adj_ratio
 *             size = market.c_quantize_order_amount(self.trading_pair, size)
*/
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":915
 *             buy.size = size
 * 
 *         for sell in proposal.sells:             # <<<<<<<<<<<<<<
 *             size = sell.size * ask_adj_ratio
 *             size = market.c_quantize_order_amount(self.trading_pair, size, sell.price)
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_11 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 915, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 915, __pyx_L1_error)
          #endif
          if (__pyx_t_11 >= __pyx_temp) break;
        }
        __pyx_t_3 = __Pyx_PyList_GetItemRefFast(__pyx_t_2, __pyx_t_11, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_11;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 915, __pyx_L1_error)
          #endif
          if (__pyx_t_11 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_11));
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_11);
        #endif
        ++__pyx_t_11;
      }
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 915, __pyx_L1_error)
    } else {
      __pyx_t_3 = __pyx_t_12(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 915, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_sell, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":916
 * 
 *         for sell in proposal.sells:
 *             size = sell.size * ask_adj_ratio             # <<<<<<<<<<<<<<
 *             size = market.c_quantize_order_amount(self.trading_pair, size, sell.price)
 *             sell.size = size
*/
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyNumber_Multiply(__pyx_t_3, __pyx_v_ask_adj_ratio); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_size, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":917
 *         for sell in proposal.sells:
 *             size = sell.size * ask_adj_ratio
 *             size = market.c_quantize_order_amount(self.trading_pair, size, sell.price)             # <<<<<<<<<<<<<<
 *             sell.size = size
 * 
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 917, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13.__pyx_n = 1;
    __pyx_t_13.price = __pyx_t_3;
    __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_1), __pyx_v_size, &__pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_size, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":918
 *             size = sell.size * ask_adj_ratio
 *             size = market.c_quantize_order_amount(self.trading_pair, size, sell.price)
 *             sell.size = size             # <<<<<<<<<<<<<<
 * 
 *     def adjusted_available_balance_for_orders_budget_constrain(self):
*/
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_size, __pyx_v_size) < (0)) __PYX_ERR(0, 918, __pyx_L1_error)

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":915
 *             buy.size = size
 * 
 *         for sell in proposal.sells:             # <<<<<<<<<<<<<<
 *             size = sell.size * ask_adj_ratio
 *             size = market.c_quantize_order_amount(self.trading_pair, size, sell.price)
*/
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":890
 *             self.c_apply_inventory_skew(proposal)
 * 
 *     cdef c_apply_inventory_skew(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_inventory_skew", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_bid_adj_ratio);
  __Pyx_XDECREF(__pyx_v_ask_adj_ratio);
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XDECREF(__pyx_v_base_balance);
  __Pyx_XDECREF(__pyx_v_quote_balance);
  __Pyx_XDECREF(__pyx_v_total_order_size);
  __Pyx_XDECREF(__pyx_v_bid_ask_ratios);
  __Pyx_XDECREF(__pyx_v_buy);
  __Pyx_XDECREF(__pyx_v_sell);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":920
 *             sell.size = size
 * 
 *     def adjusted_available_balance_for_orders_budget_constrain(self):             # <<<<<<<<<<<<<<
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
 *         non_hanging = []
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25adjusted_available_balance_for_orders_budget_constrain(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25adjusted_available_balance_for_orders_budget_constrain = {"adjusted_available_balance_for_orders_budget_constrain", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25adjusted_available_balance_for_orders_budget_constrain, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25adjusted_available_balance_for_orders_budget_constrain(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("adjusted_available_balance_for_orders_budget_constrain (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("adjusted_available_balance_for_orders_budget_constrain", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("adjusted_available_balance_for_orders_budget_constrain", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_24adjusted_available_balance_for_orders_budget_constrain(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_24adjusted_available_balance_for_orders_budget_constrain(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_candidate_hanging_orders = NULL;
  PyObject *__pyx_v_non_hanging = NULL;
  PyObject *__pyx_v_all_orders = NULL;
  PyObject *__pyx_v_all_non_hanging_orders = NULL;
  PyObject *__pyx_9genexpr12__pyx_v_order = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("adjusted_available_balance_for_orders_budget_constrain", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":921
 * 
 *     def adjusted_available_balance_for_orders_budget_constrain(self):
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()             # <<<<<<<<<<<<<<
 *         non_hanging = []
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_hanging_orders_tracker); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_candidate_hanging_orders_from_pa, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 921, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_candidate_hanging_orders = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":922
 *     def adjusted_available_balance_for_orders_budget_constrain(self):
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
 *         non_hanging = []             # <<<<<<<<<<<<<<
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_non_hanging = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":923
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
 *         non_hanging = []
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():             # <<<<<<<<<<<<<<
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_market_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)__pyx_v_self->__pyx_base._sb_order_tracker);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_limit_orders, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_t_3, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 923, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_5) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":924
 *         non_hanging = []
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()             # <<<<<<<<<<<<<<
 *             non_hanging = [order for order in all_orders
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]
*/
    __pyx_t_6 = ((PyObject *)__pyx_v_self->__pyx_base._sb_order_tracker);
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_limit_orders, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 924, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_market_info); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_1 = __pyx_t_7;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_values, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 924, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_v_all_orders = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":925
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders             # <<<<<<<<<<<<<<
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))
*/
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 925, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (likely(PyList_CheckExact(__pyx_v_all_orders)) || PyTuple_CheckExact(__pyx_v_all_orders)) {
        __pyx_t_7 = __pyx_v_all_orders; __Pyx_INCREF(__pyx_t_7);
        __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_v_all_orders); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 925, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_9 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 925, __pyx_L6_error)
      }
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_7))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_7);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 925, __pyx_L6_error)
              #endif
              if (__pyx_t_8 >= __pyx_temp) break;
            }
            __pyx_t_1 = __Pyx_PyList_GetItemRefFast(__pyx_t_7, __pyx_t_8, __Pyx_ReferenceSharing_OwnStrongReference);
            ++__pyx_t_8;
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_7);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 925, __pyx_L6_error)
              #endif
              if (__pyx_t_8 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_8));
            #else
            __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_8);
            #endif
            ++__pyx_t_8;
          }
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 925, __pyx_L6_error)
        } else {
          __pyx_t_1 = __pyx_t_9(__pyx_t_7);
          if (unlikely(!__pyx_t_1)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 925, __pyx_L6_error)
              PyErr_Clear();
            }
            break;
          }
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_order, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":926
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]             # <<<<<<<<<<<<<<
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))
 *         return self.c_get_adjusted_available_balance(all_non_hanging_orders)
*/
        __pyx_t_6 = __pyx_v_self->_hanging_orders_tracker;
        __Pyx_INCREF(__pyx_t_6);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr12__pyx_v_order, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 926, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_2};
          __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_order_id_in_hanging_orders, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 926, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 926, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_10 = (!__pyx_t_5);
        if (__pyx_t_10) {

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":925
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders             # <<<<<<<<<<<<<<
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))
*/
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_9genexpr12__pyx_v_order))) __PYX_ERR(0, 925, __pyx_L6_error)

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":926
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]             # <<<<<<<<<<<<<<
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))
 *         return self.c_get_adjusted_available_balance(all_non_hanging_orders)
*/
        }

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":925
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders             # <<<<<<<<<<<<<<
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))
*/
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_order); __pyx_9genexpr12__pyx_v_order = 0;
      goto __pyx_L11_exit_scope;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_order); __pyx_9genexpr12__pyx_v_order = 0;
      goto __pyx_L1_error;
      __pyx_L11_exit_scope:;
    } /* exit inner scope */
    __Pyx_DECREF_SET(__pyx_v_non_hanging, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":923
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
 *         non_hanging = []
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():             # <<<<<<<<<<<<<<
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":927
 *             non_hanging = [order for order in all_orders
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))             # <<<<<<<<<<<<<<
 *         return self.c_get_adjusted_available_balance(all_non_hanging_orders)
 * 
*/
  __pyx_t_3 = PySet_New(__pyx_v_non_hanging); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 927, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PySet_New(__pyx_v_candidate_hanging_orders); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 927, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = PyNumber_Subtract(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 927, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 927, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_all_non_hanging_orders = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":928
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))
 *         return self.c_get_adjusted_available_balance(all_non_hanging_orders)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_apply_budget_constraint(self, object proposal):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_adjusted_available_balance(__pyx_v_self, __pyx_v_all_non_hanging_orders); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":920
 *             sell.size = size
 * 
 *     def adjusted_available_balance_for_orders_budget_constrain(self):             # <<<<<<<<<<<<<<
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
 *         non_hanging = []
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.adjusted_available_balance_for_orders_budget_constrain", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_candidate_hanging_orders);
  __Pyx_XDECREF(__pyx_v_non_hanging);
  __Pyx_XDECREF(__pyx_v_all_orders);
  __Pyx_XDECREF(__pyx_v_all_non_hanging_orders);
  __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_order);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":930
 *         return self.c_get_adjusted_available_balance(all_non_hanging_orders)
 * 
 *     cdef c_apply_budget_constraint(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_budget_constraint(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_quote_size = 0;
  PyObject *__pyx_v_base_size = 0;
  PyObject *__pyx_v_adjusted_amount = 0;
  PyObject *__pyx_v_base_balance = NULL;
  PyObject *__pyx_v_quote_balance = NULL;
  PyObject *__pyx_v_buy = NULL;
  PyObject *__pyx_v_buy_fee = NULL;
  PyObject *__pyx_v_sell = NULL;
  PyObject *__pyx_9genexpr13__pyx_v_o = NULL;
  PyObject *__pyx_9genexpr14__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_budget_constraint", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":932
 *     cdef c_apply_budget_constraint(self, object proposal):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             object quote_size
 *             object base_size
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 932, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":937
 *             object adjusted_amount
 * 
 *         base_balance, quote_balance = self.adjusted_available_balance_for_orders_budget_constrain()             # <<<<<<<<<<<<<<
 * 
 *         for buy in proposal.buys:
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_adjusted_available_balance_for_o, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 937, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1);
      __Pyx_INCREF(__pyx_t_4);
    } else {
      __pyx_t_2 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 937, __pyx_L1_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 937, __pyx_L1_error)
      __Pyx_XGOTREF(__pyx_t_4);
    }
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_5);
    index = 0; __pyx_t_2 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < (0)) __PYX_ERR(0, 937, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 937, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_base_balance = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_quote_balance = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":939
 *         base_balance, quote_balance = self.adjusted_available_balance_for_orders_budget_constrain()
 * 
 *         for buy in proposal.buys:             # <<<<<<<<<<<<<<
 *             buy_fee = market.c_get_fee(self.base_asset, self.quote_asset, OrderType.LIMIT, TradeType.BUY,
 *                                        buy.size, buy.price)
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_4 = __pyx_t_1; __Pyx_INCREF(__pyx_t_4);
    __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 939, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 939, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 939, __pyx_L1_error)
          #endif
          if (__pyx_t_7 >= __pyx_temp) break;
        }
        __pyx_t_1 = __Pyx_PyList_GetItemRefFast(__pyx_t_4, __pyx_t_7, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_7;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 939, __pyx_L1_error)
          #endif
          if (__pyx_t_7 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_7));
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_7);
        #endif
        ++__pyx_t_7;
      }
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 939, __pyx_L1_error)
    } else {
      __pyx_t_1 = __pyx_t_8(__pyx_t_4);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 939, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_buy, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":940
 * 
 *         for buy in proposal.buys:
 *             buy_fee = market.c_get_fee(self.base_asset, self.quote_asset, OrderType.LIMIT, TradeType.BUY,             # <<<<<<<<<<<<<<
 *                                        buy.size, buy.price)
 *             quote_size = buy.size * buy.price * (Decimal(1) + buy_fee.percent)
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 940, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_quote_asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(0, 940, __pyx_L1_error)
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_OrderType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_LIMIT); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_TradeType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_BUY); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":941
 *         for buy in proposal.buys:
 *             buy_fee = market.c_get_fee(self.base_asset, self.quote_asset, OrderType.LIMIT, TradeType.BUY,
 *                                        buy.size, buy.price)             # <<<<<<<<<<<<<<
 *             quote_size = buy.size * buy.price * (Decimal(1) + buy_fee.percent)
 * 
*/
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":940
 * 
 *         for buy in proposal.buys:
 *             buy_fee = market.c_get_fee(self.base_asset, self.quote_asset, OrderType.LIMIT, TradeType.BUY,             # <<<<<<<<<<<<<<
 *                                        buy.size, buy.price)
 *             quote_size = buy.size * buy.price * (Decimal(1) + buy_fee.percent)
*/
    __pyx_t_12 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_fee(__pyx_v_market, ((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2), __pyx_t_9, __pyx_t_10, __pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_buy_fee, __pyx_t_12);
    __pyx_t_12 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":942
 *             buy_fee = market.c_get_fee(self.base_asset, self.quote_asset, OrderType.LIMIT, TradeType.BUY,
 *                                        buy.size, buy.price)
 *             quote_size = buy.size * buy.price * (Decimal(1) + buy_fee.percent)             # <<<<<<<<<<<<<<
 * 
 *             # Adjust buy order size to use remaining balance if less than the order amount
*/
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_5 = PyNumber_Multiply(__pyx_t_12, __pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
      assert(__pyx_t_12);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_12);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_10, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_int_1};
      __pyx_t_11 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_10, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 942, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
    }
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy_fee, __pyx_mstate_global->__pyx_n_u_percent); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = PyNumber_Add(__pyx_t_11, __pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyNumber_Multiply(__pyx_t_5, __pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF_SET(__pyx_v_quote_size, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":945
 * 
 *             # Adjust buy order size to use remaining balance if less than the order amount
 *             if quote_balance < quote_size:             # <<<<<<<<<<<<<<
 *                 adjusted_amount = quote_balance / (buy.price * (Decimal("1") + buy_fee.percent))
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, adjusted_amount)
*/
    __pyx_t_10 = PyObject_RichCompare(__pyx_v_quote_balance, __pyx_v_quote_size, Py_LT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 945, __pyx_L1_error)
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (__pyx_t_13) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":946
 *             # Adjust buy order size to use remaining balance if less than the order amount
 *             if quote_balance < quote_size:
 *                 adjusted_amount = quote_balance / (buy.price * (Decimal("1") + buy_fee.percent))             # <<<<<<<<<<<<<<
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, adjusted_amount)
 *                 buy.size = adjusted_amount
*/
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 946, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_5 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 946, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_3 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_11);
        assert(__pyx_t_5);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
        __pyx_t_3 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_1};
        __pyx_t_12 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 946, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
      }
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy_fee, __pyx_mstate_global->__pyx_n_u_percent); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 946, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_5 = PyNumber_Add(__pyx_t_12, __pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 946, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = PyNumber_Multiply(__pyx_t_10, __pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 946, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_v_quote_balance, __pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 946, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF_SET(__pyx_v_adjusted_amount, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":947
 *             if quote_balance < quote_size:
 *                 adjusted_amount = quote_balance / (buy.price * (Decimal("1") + buy_fee.percent))
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, adjusted_amount)             # <<<<<<<<<<<<<<
 *                 buy.size = adjusted_amount
 *                 quote_balance = s_decimal_zero
*/
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 947, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_5))) __PYX_ERR(0, 947, __pyx_L1_error)
      __pyx_t_11 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_5), __pyx_v_adjusted_amount, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 947, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_adjusted_amount, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":948
 *                 adjusted_amount = quote_balance / (buy.price * (Decimal("1") + buy_fee.percent))
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, adjusted_amount)
 *                 buy.size = adjusted_amount             # <<<<<<<<<<<<<<
 *                 quote_balance = s_decimal_zero
 *             elif quote_balance == s_decimal_zero:
*/
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_size, __pyx_v_adjusted_amount) < (0)) __PYX_ERR(0, 948, __pyx_L1_error)

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":949
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, adjusted_amount)
 *                 buy.size = adjusted_amount
 *                 quote_balance = s_decimal_zero             # <<<<<<<<<<<<<<
 *             elif quote_balance == s_decimal_zero:
 *                 buy.size = s_decimal_zero
*/
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 949, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF_SET(__pyx_v_quote_balance, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":945
 * 
 *             # Adjust buy order size to use remaining balance if less than the order amount
 *             if quote_balance < quote_size:             # <<<<<<<<<<<<<<
 *                 adjusted_amount = quote_balance / (buy.price * (Decimal("1") + buy_fee.percent))
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, adjusted_amount)
*/
      goto __pyx_L7;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":950
 *                 buy.size = adjusted_amount
 *                 quote_balance = s_decimal_zero
 *             elif quote_balance == s_decimal_zero:             # <<<<<<<<<<<<<<
 *                 buy.size = s_decimal_zero
 *             else:
*/
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 950, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_quote_balance, __pyx_t_11, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 950, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 950, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_13) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":951
 *                 quote_balance = s_decimal_zero
 *             elif quote_balance == s_decimal_zero:
 *                 buy.size = s_decimal_zero             # <<<<<<<<<<<<<<
 *             else:
 *                 quote_balance -= quote_size
*/
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 951, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_size, __pyx_t_5) < (0)) __PYX_ERR(0, 951, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":950
 *                 buy.size = adjusted_amount
 *                 quote_balance = s_decimal_zero
 *             elif quote_balance == s_decimal_zero:             # <<<<<<<<<<<<<<
 *                 buy.size = s_decimal_zero
 *             else:
*/
      goto __pyx_L7;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":953
 *                 buy.size = s_decimal_zero
 *             else:
 *                 quote_balance -= quote_size             # <<<<<<<<<<<<<<
 * 
 *         proposal.buys = [o for o in proposal.buys if o.size > 0]
*/
    /*else*/ {
      __pyx_t_5 = PyNumber_InPlaceSubtract(__pyx_v_quote_balance, __pyx_v_quote_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 953, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_quote_balance, __pyx_t_5);
      __pyx_t_5 = 0;
    }
    __pyx_L7:;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":939
 *         base_balance, quote_balance = self.adjusted_available_balance_for_orders_budget_constrain()
 * 
 *         for buy in proposal.buys:             # <<<<<<<<<<<<<<
 *             buy_fee = market.c_get_fee(self.base_asset, self.quote_asset, OrderType.LIMIT, TradeType.BUY,
 *                                        buy.size, buy.price)
*/
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":955
 *                 quote_balance -= quote_size
 * 
 *         proposal.buys = [o for o in proposal.buys if o.size > 0]             # <<<<<<<<<<<<<<
 * 
 *         for sell in proposal.sells:
*/
  { /* enter inner scope */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 955, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 955, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_11 = __pyx_t_5; __Pyx_INCREF(__pyx_t_11);
      __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 955, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_8 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 955, __pyx_L11_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_11))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_11);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 955, __pyx_L11_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          __pyx_t_5 = __Pyx_PyList_GetItemRefFast(__pyx_t_11, __pyx_t_7, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_7;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_11);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 955, __pyx_L11_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_7));
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_11, __pyx_t_7);
          #endif
          ++__pyx_t_7;
        }
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 955, __pyx_L11_error)
      } else {
        __pyx_t_5 = __pyx_t_8(__pyx_t_11);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 955, __pyx_L11_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_9genexpr13__pyx_v_o, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr13__pyx_v_o, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 955, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PyObject_RichCompare(__pyx_t_5, __pyx_mstate_global->__pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 955, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 955, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (__pyx_t_13) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_9genexpr13__pyx_v_o))) __PYX_ERR(0, 955, __pyx_L11_error)
      }
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_o); __pyx_9genexpr13__pyx_v_o = 0;
    goto __pyx_L16_exit_scope;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_o); __pyx_9genexpr13__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L16_exit_scope:;
  } /* exit inner scope */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys, __pyx_t_4) < (0)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":957
 *         proposal.buys = [o for o in proposal.buys if o.size > 0]
 * 
 *         for sell in proposal.sells:             # <<<<<<<<<<<<<<
 *             base_size = sell.size
 * 
*/
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_11 = __pyx_t_4; __Pyx_INCREF(__pyx_t_11);
    __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_8 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 957, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_11))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_11);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 957, __pyx_L1_error)
          #endif
          if (__pyx_t_7 >= __pyx_temp) break;
        }
        __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_11, __pyx_t_7, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_7;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_11);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 957, __pyx_L1_error)
          #endif
          if (__pyx_t_7 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_7));
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_11, __pyx_t_7);
        #endif
        ++__pyx_t_7;
      }
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 957, __pyx_L1_error)
    } else {
      __pyx_t_4 = __pyx_t_8(__pyx_t_11);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 957, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_sell, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":958
 * 
 *         for sell in proposal.sells:
 *             base_size = sell.size             # <<<<<<<<<<<<<<
 * 
 *             # Adjust sell order size to use remaining balance if less than the order amount
*/
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 958, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_base_size, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":961
 * 
 *             # Adjust sell order size to use remaining balance if less than the order amount
 *             if base_balance < base_size:             # <<<<<<<<<<<<<<
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, base_balance)
 *                 sell.size = adjusted_amount
*/
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_base_balance, __pyx_v_base_size, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 961, __pyx_L1_error)
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 961, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_13) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":962
 *             # Adjust sell order size to use remaining balance if less than the order amount
 *             if base_balance < base_size:
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, base_balance)             # <<<<<<<<<<<<<<
 *                 sell.size = adjusted_amount
 *                 base_balance = s_decimal_zero
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 962, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_4))) __PYX_ERR(0, 962, __pyx_L1_error)
      __pyx_t_10 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_4), __pyx_v_base_balance, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 962, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_adjusted_amount, __pyx_t_10);
      __pyx_t_10 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":963
 *             if base_balance < base_size:
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, base_balance)
 *                 sell.size = adjusted_amount             # <<<<<<<<<<<<<<
 *                 base_balance = s_decimal_zero
 *             elif base_balance == s_decimal_zero:
*/
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_size, __pyx_v_adjusted_amount) < (0)) __PYX_ERR(0, 963, __pyx_L1_error)

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":964
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, base_balance)
 *                 sell.size = adjusted_amount
 *                 base_balance = s_decimal_zero             # <<<<<<<<<<<<<<
 *             elif base_balance == s_decimal_zero:
 *                 sell.size = s_decimal_zero
*/
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 964, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF_SET(__pyx_v_base_balance, __pyx_t_10);
      __pyx_t_10 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":961
 * 
 *             # Adjust sell order size to use remaining balance if less than the order amount
 *             if base_balance < base_size:             # <<<<<<<<<<<<<<
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, base_balance)
 *                 sell.size = adjusted_amount
*/
      goto __pyx_L19;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":965
 *                 sell.size = adjusted_amount
 *                 base_balance = s_decimal_zero
 *             elif base_balance == s_decimal_zero:             # <<<<<<<<<<<<<<
 *                 sell.size = s_decimal_zero
 *             else:
*/
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_base_balance, __pyx_t_10, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_13) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":966
 *                 base_balance = s_decimal_zero
 *             elif base_balance == s_decimal_zero:
 *                 sell.size = s_decimal_zero             # <<<<<<<<<<<<<<
 *             else:
 *                 base_balance -= base_size
*/
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 966, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_size, __pyx_t_4) < (0)) __PYX_ERR(0, 966, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":965
 *                 sell.size = adjusted_amount
 *                 base_balance = s_decimal_zero
 *             elif base_balance == s_decimal_zero:             # <<<<<<<<<<<<<<
 *                 sell.size = s_decimal_zero
 *             else:
*/
      goto __pyx_L19;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":968
 *                 sell.size = s_decimal_zero
 *             else:
 *                 base_balance -= base_size             # <<<<<<<<<<<<<<
 * 
 *         proposal.sells = [o for o in proposal.sells if o.size > 0]
*/
    /*else*/ {
      __pyx_t_4 = PyNumber_InPlaceSubtract(__pyx_v_base_balance, __pyx_v_base_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 968, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_base_balance, __pyx_t_4);
      __pyx_t_4 = 0;
    }
    __pyx_L19:;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":957
 *         proposal.buys = [o for o in proposal.buys if o.size > 0]
 * 
 *         for sell in proposal.sells:             # <<<<<<<<<<<<<<
 *             base_size = sell.size
 * 
*/
  }
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":970
 *                 base_balance -= base_size
 * 
 *         proposal.sells = [o for o in proposal.sells if o.size > 0]             # <<<<<<<<<<<<<<
 * 
 *     cdef c_filter_out_takers(self, object proposal):
*/
  { /* enter inner scope */
    __pyx_t_11 = PyList_New(0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 970, __pyx_L23_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 970, __pyx_L23_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_10 = __pyx_t_4; __Pyx_INCREF(__pyx_t_10);
      __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 970, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_8 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 970, __pyx_L23_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_10))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_10);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 970, __pyx_L23_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_10, __pyx_t_7, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_7;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_10);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 970, __pyx_L23_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_7));
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_10, __pyx_t_7);
          #endif
          ++__pyx_t_7;
        }
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 970, __pyx_L23_error)
      } else {
        __pyx_t_4 = __pyx_t_8(__pyx_t_10);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 970, __pyx_L23_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_9genexpr14__pyx_v_o, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_o, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 970, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_mstate_global->__pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 970, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 970, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_13) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_11, (PyObject*)__pyx_9genexpr14__pyx_v_o))) __PYX_ERR(0, 970, __pyx_L23_error)
      }
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_o); __pyx_9genexpr14__pyx_v_o = 0;
    goto __pyx_L28_exit_scope;
    __pyx_L23_error:;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_o); __pyx_9genexpr14__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L28_exit_scope:;
  } /* exit inner scope */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells, __pyx_t_11) < (0)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":930
 *         return self.c_get_adjusted_available_balance(all_non_hanging_orders)
 * 
 *     cdef c_apply_budget_constraint(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_budget_constraint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_quote_size);
  __Pyx_XDECREF(__pyx_v_base_size);
  __Pyx_XDECREF(__pyx_v_adjusted_amount);
  __Pyx_XDECREF(__pyx_v_base_balance);
  __Pyx_XDECREF(__pyx_v_quote_balance);
  __Pyx_XDECREF(__pyx_v_buy);
  __Pyx_XDECREF(__pyx_v_buy_fee);
  __Pyx_XDECREF(__pyx_v_sell);
  __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_o);
  __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":972
 *         proposal.sells = [o for o in proposal.sells if o.size > 0]
 * 
 *     cdef c_filter_out_takers(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_filter_out_takers(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  CYTHON_UNUSED PyObject *__pyx_v_new_buys = 0;
  CYTHON_UNUSED PyObject *__pyx_v_new_sells = 0;
  PyObject *__pyx_v_top_ask = NULL;
  PyObject *__pyx_v_top_bid = NULL;
  PyObject *__pyx_9genexpr15__pyx_v_buy = NULL;
  PyObject *__pyx_9genexpr16__pyx_v_sell = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_filter_out_takers", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":974
 *     cdef c_filter_out_takers(self, object proposal):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             list new_buys = []
 *             list new_sells = []
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 974, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":975
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             list new_buys = []             # <<<<<<<<<<<<<<
 *             list new_sells = []
 *         top_ask = market.c_get_price(self.trading_pair, True)
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_new_buys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":976
 *             ExchangeBase market = self._market_info.market
 *             list new_buys = []
 *             list new_sells = []             # <<<<<<<<<<<<<<
 *         top_ask = market.c_get_price(self.trading_pair, True)
 *         if not top_ask.is_nan():
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_new_sells = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":977
 *             list new_buys = []
 *             list new_sells = []
 *         top_ask = market.c_get_price(self.trading_pair, True)             # <<<<<<<<<<<<<<
 *         if not top_ask.is_nan():
 *             proposal.buys = [buy for buy in proposal.buys if buy.price < top_ask]
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 977, __pyx_L1_error)
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_1), 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_top_ask = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":978
 *             list new_sells = []
 *         top_ask = market.c_get_price(self.trading_pair, True)
 *         if not top_ask.is_nan():             # <<<<<<<<<<<<<<
 *             proposal.buys = [buy for buy in proposal.buys if buy.price < top_ask]
 *         top_bid = market.c_get_price(self.trading_pair, False)
*/
  __pyx_t_1 = __pyx_v_top_ask;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_nan, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 978, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (!__pyx_t_4);
  if (__pyx_t_5) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":979
 *         top_ask = market.c_get_price(self.trading_pair, True)
 *         if not top_ask.is_nan():
 *             proposal.buys = [buy for buy in proposal.buys if buy.price < top_ask]             # <<<<<<<<<<<<<<
 *         top_bid = market.c_get_price(self.trading_pair, False)
 *         if not top_bid.is_nan():
*/
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 979, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 979, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
        __pyx_t_6 = __pyx_t_1; __Pyx_INCREF(__pyx_t_6);
        __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 979, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_8 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 979, __pyx_L6_error)
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_6))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 979, __pyx_L6_error)
              #endif
              if (__pyx_t_7 >= __pyx_temp) break;
            }
            __pyx_t_1 = __Pyx_PyList_GetItemRefFast(__pyx_t_6, __pyx_t_7, __Pyx_ReferenceSharing_OwnStrongReference);
            ++__pyx_t_7;
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 979, __pyx_L6_error)
              #endif
              if (__pyx_t_7 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7));
            #else
            __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_7);
            #endif
            ++__pyx_t_7;
          }
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 979, __pyx_L6_error)
        } else {
          __pyx_t_1 = __pyx_t_8(__pyx_t_6);
          if (unlikely(!__pyx_t_1)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 979, __pyx_L6_error)
              PyErr_Clear();
            }
            break;
          }
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_9genexpr15__pyx_v_buy, __pyx_t_1);
        __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr15__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 979, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = PyObject_RichCompare(__pyx_t_1, __pyx_v_top_ask, Py_LT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 979, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 979, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (__pyx_t_5) {
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_9genexpr15__pyx_v_buy))) __PYX_ERR(0, 979, __pyx_L6_error)
        }
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_buy); __pyx_9genexpr15__pyx_v_buy = 0;
      goto __pyx_L11_exit_scope;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_buy); __pyx_9genexpr15__pyx_v_buy = 0;
      goto __pyx_L1_error;
      __pyx_L11_exit_scope:;
    } /* exit inner scope */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys, __pyx_t_2) < (0)) __PYX_ERR(0, 979, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":978
 *             list new_sells = []
 *         top_ask = market.c_get_price(self.trading_pair, True)
 *         if not top_ask.is_nan():             # <<<<<<<<<<<<<<
 *             proposal.buys = [buy for buy in proposal.buys if buy.price < top_ask]
 *         top_bid = market.c_get_price(self.trading_pair, False)
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":980
 *         if not top_ask.is_nan():
 *             proposal.buys = [buy for buy in proposal.buys if buy.price < top_ask]
 *         top_bid = market.c_get_price(self.trading_pair, False)             # <<<<<<<<<<<<<<
 *         if not top_bid.is_nan():
 *             proposal.sells = [sell for sell in proposal.sells if sell.price > top_bid]
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(0, 980, __pyx_L1_error)
  __pyx_t_6 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_2), 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_top_bid = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":981
 *             proposal.buys = [buy for buy in proposal.buys if buy.price < top_ask]
 *         top_bid = market.c_get_price(self.trading_pair, False)
 *         if not top_bid.is_nan():             # <<<<<<<<<<<<<<
 *             proposal.sells = [sell for sell in proposal.sells if sell.price > top_bid]
 * 
*/
  __pyx_t_2 = __pyx_v_top_bid;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_6 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_nan, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 981, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_4 = (!__pyx_t_5);
  if (__pyx_t_4) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":982
 *         top_bid = market.c_get_price(self.trading_pair, False)
 *         if not top_bid.is_nan():
 *             proposal.sells = [sell for sell in proposal.sells if sell.price > top_bid]             # <<<<<<<<<<<<<<
 * 
 *     # Compare the market price with the top bid and top ask price
*/
    { /* enter inner scope */
      __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 982, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 982, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
        __pyx_t_9 = __pyx_t_2; __Pyx_INCREF(__pyx_t_9);
        __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 982, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_8 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 982, __pyx_L15_error)
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_9))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_9);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 982, __pyx_L15_error)
              #endif
              if (__pyx_t_7 >= __pyx_temp) break;
            }
            __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_9, __pyx_t_7, __Pyx_ReferenceSharing_OwnStrongReference);
            ++__pyx_t_7;
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_9);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 982, __pyx_L15_error)
              #endif
              if (__pyx_t_7 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_7));
            #else
            __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_7);
            #endif
            ++__pyx_t_7;
          }
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 982, __pyx_L15_error)
        } else {
          __pyx_t_2 = __pyx_t_8(__pyx_t_9);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 982, __pyx_L15_error)
              PyErr_Clear();
            }
            break;
          }
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_9genexpr16__pyx_v_sell, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr16__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 982, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_v_top_bid, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 982, __pyx_L15_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 982, __pyx_L15_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_4) {
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_9genexpr16__pyx_v_sell))) __PYX_ERR(0, 982, __pyx_L15_error)
        }
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_sell); __pyx_9genexpr16__pyx_v_sell = 0;
      goto __pyx_L20_exit_scope;
      __pyx_L15_error:;
      __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_sell); __pyx_9genexpr16__pyx_v_sell = 0;
      goto __pyx_L1_error;
      __pyx_L20_exit_scope:;
    } /* exit inner scope */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells, __pyx_t_6) < (0)) __PYX_ERR(0, 982, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":981
 *             proposal.buys = [buy for buy in proposal.buys if buy.price < top_ask]
 *         top_bid = market.c_get_price(self.trading_pair, False)
 *         if not top_bid.is_nan():             # <<<<<<<<<<<<<<
 *             proposal.sells = [sell for sell in proposal.sells if sell.price > top_bid]
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":972
 *         proposal.sells = [o for o in proposal.sells if o.size > 0]
 * 
 *     cdef c_filter_out_takers(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_filter_out_takers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_new_buys);
  __Pyx_XDECREF(__pyx_v_new_sells);
  __Pyx_XDECREF(__pyx_v_top_ask);
  __Pyx_XDECREF(__pyx_v_top_bid);
  __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_buy);
  __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_sell);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1010
 *             # If the price_above_bid is lower than the price suggested by the top pricing proposal,
 *             # lower the price and from there apply the order_level_spread to each order in the next levels
 *             proposal.buys = sorted(proposal.buys, key = lambda p: p.price, reverse = True)             # <<<<<<<<<<<<<<
 *             lower_buy_price = min(proposal.buys[0].price, price_above_bid)
 *             for i, proposed in enumerate(proposal.buys):
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26c_apply_order_optimization_lambda2(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26c_apply_order_optimization_lambda2 = {"lambda2", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26c_apply_order_optimization_lambda2, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26c_apply_order_optimization_lambda2(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_p = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda2 (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_p,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1010, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1010, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "lambda2", 0) < (0)) __PYX_ERR(0, 1010, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("lambda2", 1, 1, 1, i); __PYX_ERR(0, 1010, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1010, __pyx_L3_error)
    }
    __pyx_v_p = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda2", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1010, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_order_optimization.lambda2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda2(__pyx_self, __pyx_v_p);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda2(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda2", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_order_optimization.lambda2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1033
 *             # If the price_below_ask is higher than the price suggested by the pricing proposal,
 *             # increase your price and from there apply the order_level_spread to each order in the next levels
 *             proposal.sells = sorted(proposal.sells, key = lambda p: p.price)             # <<<<<<<<<<<<<<
 *             higher_sell_price = max(proposal.sells[0].price, price_below_ask)
 *             for i, proposed in enumerate(proposal.sells):
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26c_apply_order_optimization_1lambda3(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26c_apply_order_optimization_1lambda3 = {"lambda3", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26c_apply_order_optimization_1lambda3, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26c_apply_order_optimization_1lambda3(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_p = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda3 (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_p,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1033, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1033, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "lambda3", 0) < (0)) __PYX_ERR(0, 1033, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("lambda3", 1, 1, 1, i); __PYX_ERR(0, 1033, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1033, __pyx_L3_error)
    }
    __pyx_v_p = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda3", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1033, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_order_optimization.lambda3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda3(__pyx_self, __pyx_v_p);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda3(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda3", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_order_optimization.lambda3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":985
 * 
 *     # Compare the market price with the top bid and top ask price
 *     cdef c_apply_order_optimization(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_order_optimization(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_own_buy_size = 0;
  PyObject *__pyx_v_own_sell_size = 0;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_v_top_bid_price = NULL;
  PyObject *__pyx_v_price_quantum = NULL;
  PyObject *__pyx_v_price_above_bid = NULL;
  PyObject *__pyx_v_lower_buy_price = NULL;
  PyObject *__pyx_v_i = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_proposed = NULL;
  PyObject *__pyx_v_top_ask_price = NULL;
  PyObject *__pyx_v_price_below_ask = NULL;
  PyObject *__pyx_v_higher_sell_price = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_order_optimization", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":987
 *     cdef c_apply_order_optimization(self, object proposal):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             object own_buy_size = s_decimal_zero
 *             object own_sell_size = s_decimal_zero
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 987, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":988
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             object own_buy_size = s_decimal_zero             # <<<<<<<<<<<<<<
 *             object own_sell_size = s_decimal_zero
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_own_buy_size = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":989
 *             ExchangeBase market = self._market_info.market
 *             object own_buy_size = s_decimal_zero
 *             object own_sell_size = s_decimal_zero             # <<<<<<<<<<<<<<
 * 
 *         for order in self.active_orders:
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_own_sell_size = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":991
 *             object own_sell_size = s_decimal_zero
 * 
 *         for order in self.active_orders:             # <<<<<<<<<<<<<<
 *             if order.is_buy:
 *                 own_buy_size = order.quantity
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 991, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 991, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 991, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        __pyx_t_1 = __Pyx_PyList_GetItemRefFast(__pyx_t_2, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_3;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 991, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3));
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3);
        #endif
        ++__pyx_t_3;
      }
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 991, __pyx_L1_error)
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 991, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":992
 * 
 *         for order in self.active_orders:
 *             if order.is_buy:             # <<<<<<<<<<<<<<
 *                 own_buy_size = order.quantity
 *             else:
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":993
 *         for order in self.active_orders:
 *             if order.is_buy:
 *                 own_buy_size = order.quantity             # <<<<<<<<<<<<<<
 *             else:
 *                 own_sell_size = order.quantity
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_quantity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_own_buy_size, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":992
 * 
 *         for order in self.active_orders:
 *             if order.is_buy:             # <<<<<<<<<<<<<<
 *                 own_buy_size = order.quantity
 *             else:
*/
      goto __pyx_L5;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":995
 *                 own_buy_size = order.quantity
 *             else:
 *                 own_sell_size = order.quantity             # <<<<<<<<<<<<<<
 * 
 *         if len(proposal.buys) > 0:
*/
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_quantity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 995, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_own_sell_size, __pyx_t_1);
      __pyx_t_1 = 0;
    }
    __pyx_L5:;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":991
 *             object own_sell_size = s_decimal_zero
 * 
 *         for order in self.active_orders:             # <<<<<<<<<<<<<<
 *             if order.is_buy:
 *                 own_buy_size = order.quantity
*/
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":997
 *                 own_sell_size = order.quantity
 * 
 *         if len(proposal.buys) > 0:             # <<<<<<<<<<<<<<
 *             # Get the top bid price in the market using order_optimization_depth and your buy order volume
 *             top_bid_price = self._market_info.get_price_for_volume(
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_3 > 0);
  if (__pyx_t_5) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":999
 *         if len(proposal.buys) > 0:
 *             # Get the top bid price in the market using order_optimization_depth and your buy order volume
 *             top_bid_price = self._market_info.get_price_for_volume(             # <<<<<<<<<<<<<<
 *                 False, self._bid_order_optimization_depth + own_buy_size).result_price
 *             price_quantum = market.c_get_order_price_quantum(
*/
    __pyx_t_1 = __pyx_v_self->_market_info;
    __Pyx_INCREF(__pyx_t_1);

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1000
 *             # Get the top bid price in the market using order_optimization_depth and your buy order volume
 *             top_bid_price = self._market_info.get_price_for_volume(
 *                 False, self._bid_order_optimization_depth + own_buy_size).result_price             # <<<<<<<<<<<<<<
 *             price_quantum = market.c_get_order_price_quantum(
 *                 self.trading_pair,
*/
    __pyx_t_6 = PyNumber_Add(__pyx_v_self->_bid_order_optimization_depth, __pyx_v_own_buy_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1000, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = 0;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_1, Py_False, __pyx_t_6};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price_for_volume, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_result_price); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1000, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_top_bid_price = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1002
 *                 False, self._bid_order_optimization_depth + own_buy_size).result_price
 *             price_quantum = market.c_get_order_price_quantum(
 *                 self.trading_pair,             # <<<<<<<<<<<<<<
 *                 top_bid_price
 *             )
*/
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1002, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_6))) __PYX_ERR(0, 1002, __pyx_L1_error)

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1001
 *             top_bid_price = self._market_info.get_price_for_volume(
 *                 False, self._bid_order_optimization_depth + own_buy_size).result_price
 *             price_quantum = market.c_get_order_price_quantum(             # <<<<<<<<<<<<<<
 *                 self.trading_pair,
 *                 top_bid_price
*/
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_order_price_quantum(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_6), __pyx_v_top_bid_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_price_quantum = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1006
 *             )
 *             # Get the price above the top bid
 *             price_above_bid = (ceil(top_bid_price / price_quantum) + 1) * price_quantum             # <<<<<<<<<<<<<<
 * 
 *             # If the price_above_bid is lower than the price suggested by the top pricing proposal,
*/
    __pyx_t_6 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_ceil); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyNumber_Divide(__pyx_v_top_bid_price, __pyx_v_price_quantum); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      assert(__pyx_t_6);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_8};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1006, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_1 = __Pyx_PyLong_AddObjC(__pyx_t_2, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_1, __pyx_v_price_quantum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_price_above_bid = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1010
 *             # If the price_above_bid is lower than the price suggested by the top pricing proposal,
 *             # lower the price and from there apply the order_level_spread to each order in the next levels
 *             proposal.buys = sorted(proposal.buys, key = lambda p: p.price, reverse = True)             # <<<<<<<<<<<<<<
 *             lower_buy_price = min(proposal.buys[0].price, price_above_bid)
 *             for i, proposed in enumerate(proposal.buys):
*/
    __pyx_t_1 = NULL;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26c_apply_order_optimization_lambda2, 0, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_c_apply, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 2 : 0)] = {__pyx_t_1, __pyx_t_8};
      __pyx_t_9 = __Pyx_MakeVectorcallBuilderKwds(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_key, __pyx_t_6, __pyx_t_9, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 1010, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_reverse, Py_True, __pyx_t_9, __pyx_callargs+2, 1) < (0)) __PYX_ERR(0, 1010, __pyx_L1_error)
      __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_builtin_sorted, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_9);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys, __pyx_t_2) < (0)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1011
 *             # lower the price and from there apply the order_level_spread to each order in the next levels
 *             proposal.buys = sorted(proposal.buys, key = lambda p: p.price, reverse = True)
 *             lower_buy_price = min(proposal.buys[0].price, price_above_bid)             # <<<<<<<<<<<<<<
 *             for i, proposed in enumerate(proposal.buys):
 *                 if self._split_order_levels_enabled:
*/
    __Pyx_INCREF(__pyx_v_price_above_bid);
    __pyx_t_2 = __pyx_v_price_above_bid;
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_9, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_2, __pyx_t_9, Py_LT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_5) {
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_6 = __pyx_t_2;
    } else {
      __Pyx_INCREF(__pyx_t_9);
      __pyx_t_6 = __pyx_t_9;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __pyx_t_6;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_lower_buy_price = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1012
 *             proposal.buys = sorted(proposal.buys, key = lambda p: p.price, reverse = True)
 *             lower_buy_price = min(proposal.buys[0].price, price_above_bid)
 *             for i, proposed in enumerate(proposal.buys):             # <<<<<<<<<<<<<<
 *                 if self._split_order_levels_enabled:
 *                     proposal.buys[i].price = (market.c_quantize_order_price(self.trading_pair, lower_buy_price)
*/
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_0);
    __pyx_t_2 = __pyx_mstate_global->__pyx_int_0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_9 = __pyx_t_6; __Pyx_INCREF(__pyx_t_9);
      __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1012, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_4 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1012, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_9))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_9);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1012, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          __pyx_t_6 = __Pyx_PyList_GetItemRefFast(__pyx_t_9, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_3;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_9);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1012, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_3));
          #else
          __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_3);
          #endif
          ++__pyx_t_3;
        }
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1012, __pyx_L1_error)
      } else {
        __pyx_t_6 = __pyx_t_4(__pyx_t_9);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1012, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_proposed, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
      __pyx_t_6 = __Pyx_PyLong_AddObjC(__pyx_t_2, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1012, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2);
      __pyx_t_2 = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1013
 *             lower_buy_price = min(proposal.buys[0].price, price_above_bid)
 *             for i, proposed in enumerate(proposal.buys):
 *                 if self._split_order_levels_enabled:             # <<<<<<<<<<<<<<
 *                     proposal.buys[i].price = (market.c_quantize_order_price(self.trading_pair, lower_buy_price)
 *                                               * (1 - self._bid_order_level_spreads[i] / Decimal("100"))
*/
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_self->_split_order_levels_enabled); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1013, __pyx_L1_error)
      if (__pyx_t_5) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1014
 *             for i, proposed in enumerate(proposal.buys):
 *                 if self._split_order_levels_enabled:
 *                     proposal.buys[i].price = (market.c_quantize_order_price(self.trading_pair, lower_buy_price)             # <<<<<<<<<<<<<<
 *                                               * (1 - self._bid_order_level_spreads[i] / Decimal("100"))
 *                                               / (1-self._bid_order_level_spreads[0] / Decimal("100")))
*/
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_6))) __PYX_ERR(0, 1014, __pyx_L1_error)
        __pyx_t_8 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_6), __pyx_v_lower_buy_price); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1015
 *                 if self._split_order_levels_enabled:
 *                     proposal.buys[i].price = (market.c_quantize_order_price(self.trading_pair, lower_buy_price)
 *                                               * (1 - self._bid_order_level_spreads[i] / Decimal("100"))             # <<<<<<<<<<<<<<
 *                                               / (1-self._bid_order_level_spreads[0] / Decimal("100")))
 *                     continue
*/
        __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_self->_bid_order_level_spreads, __pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1015, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_10 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1015, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_7 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
          assert(__pyx_t_10);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
          __pyx_t_7 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_mstate_global->__pyx_kp_u_100};
          __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __pyx_t_11 = __Pyx_PyNumber_Divide(__pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1015, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyLong_SubtractCObj(__pyx_mstate_global->__pyx_int_1, __pyx_t_11, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = PyNumber_Multiply(__pyx_t_8, __pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1015, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1016
 *                     proposal.buys[i].price = (market.c_quantize_order_price(self.trading_pair, lower_buy_price)
 *                                               * (1 - self._bid_order_level_spreads[i] / Decimal("100"))
 *                                               / (1-self._bid_order_level_spreads[0] / Decimal("100")))             # <<<<<<<<<<<<<<
 *                     continue
 *                 proposal.buys[i].price = market.c_quantize_order_price(self.trading_pair, lower_buy_price) * (1 - self.order_level_spread * i)
*/
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->_bid_order_level_spreads, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_7 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
          assert(__pyx_t_6);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_10, __pyx__function);
          __pyx_t_7 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_mstate_global->__pyx_kp_u_100};
          __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_10, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1016, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
        }
        __pyx_t_10 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyLong_SubtractCObj(__pyx_mstate_global->__pyx_int_1, __pyx_t_10, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_PyNumber_Divide(__pyx_t_11, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1014
 *             for i, proposed in enumerate(proposal.buys):
 *                 if self._split_order_levels_enabled:
 *                     proposal.buys[i].price = (market.c_quantize_order_price(self.trading_pair, lower_buy_price)             # <<<<<<<<<<<<<<
 *                                               * (1 - self._bid_order_level_spreads[i] / Decimal("100"))
 *                                               / (1-self._bid_order_level_spreads[0] / Decimal("100")))
*/
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_t_8, __pyx_v_i); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (__Pyx_PyObject_SetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_price, __pyx_t_10) < (0)) __PYX_ERR(0, 1014, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1017
 *                                               * (1 - self._bid_order_level_spreads[i] / Decimal("100"))
 *                                               / (1-self._bid_order_level_spreads[0] / Decimal("100")))
 *                     continue             # <<<<<<<<<<<<<<
 *                 proposal.buys[i].price = market.c_quantize_order_price(self.trading_pair, lower_buy_price) * (1 - self.order_level_spread * i)
 * 
*/
        goto __pyx_L8_continue;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1013
 *             lower_buy_price = min(proposal.buys[0].price, price_above_bid)
 *             for i, proposed in enumerate(proposal.buys):
 *                 if self._split_order_levels_enabled:             # <<<<<<<<<<<<<<
 *                     proposal.buys[i].price = (market.c_quantize_order_price(self.trading_pair, lower_buy_price)
 *                                               * (1 - self._bid_order_level_spreads[i] / Decimal("100"))
*/
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1018
 *                                               / (1-self._bid_order_level_spreads[0] / Decimal("100")))
 *                     continue
 *                 proposal.buys[i].price = market.c_quantize_order_price(self.trading_pair, lower_buy_price) * (1 - self.order_level_spread * i)             # <<<<<<<<<<<<<<
 * 
 *         if len(proposal.sells) > 0:
*/
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_11))||((__pyx_t_11) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_11))) __PYX_ERR(0, 1018, __pyx_L1_error)
      __pyx_t_10 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_11), __pyx_v_lower_buy_price); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_order_level_spread); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_8 = PyNumber_Multiply(__pyx_t_11, __pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = __Pyx_PyLong_SubtractCObj(__pyx_mstate_global->__pyx_int_1, __pyx_t_8, 1, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyNumber_Multiply(__pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_t_11, __pyx_v_i); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_price, __pyx_t_8) < (0)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1012
 *             proposal.buys = sorted(proposal.buys, key = lambda p: p.price, reverse = True)
 *             lower_buy_price = min(proposal.buys[0].price, price_above_bid)
 *             for i, proposed in enumerate(proposal.buys):             # <<<<<<<<<<<<<<
 *                 if self._split_order_levels_enabled:
 *                     proposal.buys[i].price = (market.c_quantize_order_price(self.trading_pair, lower_buy_price)
*/
      __pyx_L8_continue:;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":997
 *                 own_sell_size = order.quantity
 * 
 *         if len(proposal.buys) > 0:             # <<<<<<<<<<<<<<
 *             # Get the top bid price in the market using order_optimization_depth and your buy order volume
 *             top_bid_price = self._market_info.get_price_for_volume(
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1020
 *                 proposal.buys[i].price = market.c_quantize_order_price(self.trading_pair, lower_buy_price) * (1 - self.order_level_spread * i)
 * 
 *         if len(proposal.sells) > 0:             # <<<<<<<<<<<<<<
 *             # Get the top ask price in the market using order_optimization_depth and your sell order volume
 *             top_ask_price = self._market_info.get_price_for_volume(
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_3 > 0);
  if (__pyx_t_5) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1022
 *         if len(proposal.sells) > 0:
 *             # Get the top ask price in the market using order_optimization_depth and your sell order volume
 *             top_ask_price = self._market_info.get_price_for_volume(             # <<<<<<<<<<<<<<
 *                 True, self._ask_order_optimization_depth + own_sell_size).result_price
 *             price_quantum = market.c_get_order_price_quantum(
*/
    __pyx_t_9 = __pyx_v_self->_market_info;
    __Pyx_INCREF(__pyx_t_9);

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1023
 *             # Get the top ask price in the market using order_optimization_depth and your sell order volume
 *             top_ask_price = self._market_info.get_price_for_volume(
 *                 True, self._ask_order_optimization_depth + own_sell_size).result_price             # <<<<<<<<<<<<<<
 *             price_quantum = market.c_get_order_price_quantum(
 *                 self.trading_pair,
*/
    __pyx_t_10 = PyNumber_Add(__pyx_v_self->_ask_order_optimization_depth, __pyx_v_own_sell_size); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = 0;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_9, Py_True, __pyx_t_10};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price_for_volume, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1022, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_result_price); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_top_ask_price = __pyx_t_10;
    __pyx_t_10 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1025
 *                 True, self._ask_order_optimization_depth + own_sell_size).result_price
 *             price_quantum = market.c_get_order_price_quantum(
 *                 self.trading_pair,             # <<<<<<<<<<<<<<
 *                 top_ask_price
 *             )
*/
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1025, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_10))) __PYX_ERR(0, 1025, __pyx_L1_error)

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1024
 *             top_ask_price = self._market_info.get_price_for_volume(
 *                 True, self._ask_order_optimization_depth + own_sell_size).result_price
 *             price_quantum = market.c_get_order_price_quantum(             # <<<<<<<<<<<<<<
 *                 self.trading_pair,
 *                 top_ask_price
*/
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_order_price_quantum(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_10), __pyx_v_top_ask_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF_SET(__pyx_v_price_quantum, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1029
 *             )
 *             # Get the price below the top ask
 *             price_below_ask = (floor(top_ask_price / price_quantum) - 1) * price_quantum             # <<<<<<<<<<<<<<
 * 
 *             # If the price_below_ask is higher than the price suggested by the pricing proposal,
*/
    __pyx_t_10 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_floor); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_8 = __Pyx_PyNumber_Divide(__pyx_v_top_ask_price, __pyx_v_price_quantum); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      assert(__pyx_t_10);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_9, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_9, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1029, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_9 = __Pyx_PyLong_SubtractObjC(__pyx_t_2, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_9, __pyx_v_price_quantum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_v_price_below_ask = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1033
 *             # If the price_below_ask is higher than the price suggested by the pricing proposal,
 *             # increase your price and from there apply the order_level_spread to each order in the next levels
 *             proposal.sells = sorted(proposal.sells, key = lambda p: p.price)             # <<<<<<<<<<<<<<
 *             higher_sell_price = max(proposal.sells[0].price, price_below_ask)
 *             for i, proposed in enumerate(proposal.sells):
*/
    __pyx_t_9 = NULL;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26c_apply_order_optimization_1lambda3, 0, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_c_apply, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_9, __pyx_t_8};
      __pyx_t_11 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1033, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_key, __pyx_t_10, __pyx_t_11, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 1033, __pyx_L1_error)
      __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_builtin_sorted, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_11);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1033, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells, __pyx_t_2) < (0)) __PYX_ERR(0, 1033, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1034
 *             # increase your price and from there apply the order_level_spread to each order in the next levels
 *             proposal.sells = sorted(proposal.sells, key = lambda p: p.price)
 *             higher_sell_price = max(proposal.sells[0].price, price_below_ask)             # <<<<<<<<<<<<<<
 *             for i, proposed in enumerate(proposal.sells):
 *                 if self._split_order_levels_enabled:
*/
    __Pyx_INCREF(__pyx_v_price_below_ask);
    __pyx_t_2 = __pyx_v_price_below_ask;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_11, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_2, __pyx_t_11, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1034, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1034, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_5) {
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_10 = __pyx_t_2;
    } else {
      __Pyx_INCREF(__pyx_t_11);
      __pyx_t_10 = __pyx_t_11;
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __pyx_t_10;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_v_higher_sell_price = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1035
 *             proposal.sells = sorted(proposal.sells, key = lambda p: p.price)
 *             higher_sell_price = max(proposal.sells[0].price, price_below_ask)
 *             for i, proposed in enumerate(proposal.sells):             # <<<<<<<<<<<<<<
 *                 if self._split_order_levels_enabled:
 *                     proposal.sells[i].price = (market.c_quantize_order_price(self.trading_pair, higher_sell_price)
*/
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_0);
    __pyx_t_2 = __pyx_mstate_global->__pyx_int_0;
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1035, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (likely(PyList_CheckExact(__pyx_t_10)) || PyTuple_CheckExact(__pyx_t_10)) {
      __pyx_t_11 = __pyx_t_10; __Pyx_INCREF(__pyx_t_11);
      __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1035, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_4 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1035, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_11))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_11);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1035, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          __pyx_t_10 = __Pyx_PyList_GetItemRefFast(__pyx_t_11, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_3;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_11);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1035, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_3));
          #else
          __pyx_t_10 = __Pyx_PySequence_ITEM(__pyx_t_11, __pyx_t_3);
          #endif
          ++__pyx_t_3;
        }
        if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1035, __pyx_L1_error)
      } else {
        __pyx_t_10 = __pyx_t_4(__pyx_t_11);
        if (unlikely(!__pyx_t_10)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1035, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_XDECREF_SET(__pyx_v_proposed, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
      __pyx_t_10 = __Pyx_PyLong_AddObjC(__pyx_t_2, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1035, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_2);
      __pyx_t_2 = __pyx_t_10;
      __pyx_t_10 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1036
 *             higher_sell_price = max(proposal.sells[0].price, price_below_ask)
 *             for i, proposed in enumerate(proposal.sells):
 *                 if self._split_order_levels_enabled:             # <<<<<<<<<<<<<<
 *                     proposal.sells[i].price = (market.c_quantize_order_price(self.trading_pair, higher_sell_price)
 *                                                * (1 + self._ask_order_level_spreads[i] / Decimal("100"))
*/
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_self->_split_order_levels_enabled); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1036, __pyx_L1_error)
      if (__pyx_t_5) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1037
 *             for i, proposed in enumerate(proposal.sells):
 *                 if self._split_order_levels_enabled:
 *                     proposal.sells[i].price = (market.c_quantize_order_price(self.trading_pair, higher_sell_price)             # <<<<<<<<<<<<<<
 *                                                * (1 + self._ask_order_level_spreads[i] / Decimal("100"))
 *                                                / (1 + self._ask_order_level_spreads[0] / Decimal("100")))
*/
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1037, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_10))) __PYX_ERR(0, 1037, __pyx_L1_error)
        __pyx_t_8 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_10), __pyx_v_higher_sell_price); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1037, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1038
 *                 if self._split_order_levels_enabled:
 *                     proposal.sells[i].price = (market.c_quantize_order_price(self.trading_pair, higher_sell_price)
 *                                                * (1 + self._ask_order_level_spreads[i] / Decimal("100"))             # <<<<<<<<<<<<<<
 *                                                / (1 + self._ask_order_level_spreads[0] / Decimal("100")))
 *                     continue
*/
        __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_v_self->_ask_order_level_spreads, __pyx_v_i); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1038, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_1 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1038, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
          assert(__pyx_t_1);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
          __pyx_t_7 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_100};
          __pyx_t_9 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1038, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
        }
        __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_t_10, __pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1038, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = __Pyx_PyLong_AddCObj(__pyx_mstate_global->__pyx_int_1, __pyx_t_6, 1, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1038, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = PyNumber_Multiply(__pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1038, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1039
 *                     proposal.sells[i].price = (market.c_quantize_order_price(self.trading_pair, higher_sell_price)
 *                                                * (1 + self._ask_order_level_spreads[i] / Decimal("100"))
 *                                                / (1 + self._ask_order_level_spreads[0] / Decimal("100")))             # <<<<<<<<<<<<<<
 *                     continue
 *                 proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, higher_sell_price) * (1 + self.order_level_spread * i)
*/
        __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_self->_ask_order_level_spreads, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
          assert(__pyx_t_10);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
          __pyx_t_7 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_mstate_global->__pyx_kp_u_100};
          __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1039, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
        }
        __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_9, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyLong_AddCObj(__pyx_mstate_global->__pyx_int_1, __pyx_t_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_6, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1037
 *             for i, proposed in enumerate(proposal.sells):
 *                 if self._split_order_levels_enabled:
 *                     proposal.sells[i].price = (market.c_quantize_order_price(self.trading_pair, higher_sell_price)             # <<<<<<<<<<<<<<
 *                                                * (1 + self._ask_order_level_spreads[i] / Decimal("100"))
 *                                                / (1 + self._ask_order_level_spreads[0] / Decimal("100")))
*/
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1037, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_8, __pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1037, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (__Pyx_PyObject_SetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_price, __pyx_t_1) < (0)) __PYX_ERR(0, 1037, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1040
 *                                                * (1 + self._ask_order_level_spreads[i] / Decimal("100"))
 *                                                / (1 + self._ask_order_level_spreads[0] / Decimal("100")))
 *                     continue             # <<<<<<<<<<<<<<
 *                 proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, higher_sell_price) * (1 + self.order_level_spread * i)
 * 
*/
        goto __pyx_L13_continue;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1036
 *             higher_sell_price = max(proposal.sells[0].price, price_below_ask)
 *             for i, proposed in enumerate(proposal.sells):
 *                 if self._split_order_levels_enabled:             # <<<<<<<<<<<<<<
 *                     proposal.sells[i].price = (market.c_quantize_order_price(self.trading_pair, higher_sell_price)
 *                                                * (1 + self._ask_order_level_spreads[i] / Decimal("100"))
*/
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1041
 *                                                / (1 + self._ask_order_level_spreads[0] / Decimal("100")))
 *                     continue
 *                 proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, higher_sell_price) * (1 + self.order_level_spread * i)             # <<<<<<<<<<<<<<
 * 
 *     cdef object c_apply_add_transaction_costs(self, object proposal):
*/
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_6))) __PYX_ERR(0, 1041, __pyx_L1_error)
      __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_6), __pyx_v_higher_sell_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_order_level_spread); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = PyNumber_Multiply(__pyx_t_6, __pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyLong_AddCObj(__pyx_mstate_global->__pyx_int_1, __pyx_t_8, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyNumber_Multiply(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_6, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_price, __pyx_t_8) < (0)) __PYX_ERR(0, 1041, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1035
 *             proposal.sells = sorted(proposal.sells, key = lambda p: p.price)
 *             higher_sell_price = max(proposal.sells[0].price, price_below_ask)
 *             for i, proposed in enumerate(proposal.sells):             # <<<<<<<<<<<<<<
 *                 if self._split_order_levels_enabled:
 *                     proposal.sells[i].price = (market.c_quantize_order_price(self.trading_pair, higher_sell_price)
*/
      __pyx_L13_continue:;
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1020
 *                 proposal.buys[i].price = market.c_quantize_order_price(self.trading_pair, lower_buy_price) * (1 - self.order_level_spread * i)
 * 
 *         if len(proposal.sells) > 0:             # <<<<<<<<<<<<<<
 *             # Get the top ask price in the market using order_optimization_depth and your sell order volume
 *             top_ask_price = self._market_info.get_price_for_volume(
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":985
 * 
 *     # Compare the market price with the top bid and top ask price
 *     cdef c_apply_order_optimization(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_order_optimization", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_own_buy_size);
  __Pyx_XDECREF(__pyx_v_own_sell_size);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XDECREF(__pyx_v_top_bid_price);
  __Pyx_XDECREF(__pyx_v_price_quantum);
  __Pyx_XDECREF(__pyx_v_price_above_bid);
  __Pyx_XDECREF(__pyx_v_lower_buy_price);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_proposed);
  __Pyx_XDECREF(__pyx_v_top_ask_price);
  __Pyx_XDECREF(__pyx_v_price_below_ask);
  __Pyx_XDECREF(__pyx_v_higher_sell_price);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1043
 *                 proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, higher_sell_price) * (1 + self.order_level_spread * i)
 * 
 *     cdef object c_apply_add_transaction_costs(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_add_transaction_costs(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_buy = NULL;
  PyObject *__pyx_v_fee = NULL;
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_v_sell = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  size_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_add_transaction_costs", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1045
 *     cdef object c_apply_add_transaction_costs(self, object proposal):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *         for buy in proposal.buys:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_mstate_global->__pyx_n_u_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 1045, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1046
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *         for buy in proposal.buys:             # <<<<<<<<<<<<<<
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1046, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1046, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1046, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        __pyx_t_1 = __Pyx_PyList_GetItemRefFast(__pyx_t_2, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_3;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1046, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3));
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3);
        #endif
        ++__pyx_t_3;
      }
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1046, __pyx_L1_error)
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1046, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_buy, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1047
 *             ExchangeBase market = self._market_info.market
 *         for buy in proposal.buys:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,             # <<<<<<<<<<<<<<
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)
 *             price = buy.price * (Decimal(1) - fee.percent)
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1047, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 1047, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_quote_asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1047, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_5))) __PYX_ERR(0, 1047, __pyx_L1_error)

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1048
 *         for buy in proposal.buys:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)             # <<<<<<<<<<<<<<
 *             price = buy.price * (Decimal(1) - fee.percent)
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)
*/
    __pyx_t_6 = __pyx_v_self->_limit_order_type;
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_TradeType); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1048, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_BUY); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1048, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1048, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1048, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1047
 *             ExchangeBase market = self._market_info.market
 *         for buy in proposal.buys:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,             # <<<<<<<<<<<<<<
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)
 *             price = buy.price * (Decimal(1) - fee.percent)
*/
    __pyx_t_10 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_fee(__pyx_v_market, ((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_5), __pyx_t_6, __pyx_t_8, __pyx_t_7, __pyx_t_9, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1047, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF_SET(__pyx_v_fee, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1049
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)
 *             price = buy.price * (Decimal(1) - fee.percent)             # <<<<<<<<<<<<<<
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)
 *         for sell in proposal.sells:
*/
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_11 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      assert(__pyx_t_7);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
      __pyx_t_11 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_int_1};
      __pyx_t_9 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
    }
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_fee, __pyx_mstate_global->__pyx_n_u_percent); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = PyNumber_Subtract(__pyx_t_9, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyNumber_Multiply(__pyx_t_10, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_price, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1050
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)
 *             price = buy.price * (Decimal(1) - fee.percent)
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)             # <<<<<<<<<<<<<<
 *         for sell in proposal.sells:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
*/
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1050, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_8))) __PYX_ERR(0, 1050, __pyx_L1_error)
    __pyx_t_7 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_8), __pyx_v_price); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1050, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_price, __pyx_t_7) < (0)) __PYX_ERR(0, 1050, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1046
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *         for buy in proposal.buys:             # <<<<<<<<<<<<<<
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)
*/
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1051
 *             price = buy.price * (Decimal(1) - fee.percent)
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)
 *         for sell in proposal.sells:             # <<<<<<<<<<<<<<
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_7 = __pyx_t_2; __Pyx_INCREF(__pyx_t_7);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1051, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_7);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1051, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_7, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_3;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_7);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1051, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_3));
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_3);
        #endif
        ++__pyx_t_3;
      }
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1051, __pyx_L1_error)
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_7);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1051, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_sell, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1052
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)
 *         for sell in proposal.sells:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,             # <<<<<<<<<<<<<<
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)
 *             price = sell.price * (Decimal(1) + fee.percent)
*/
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_base_asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(0, 1052, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_quote_asset); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_8))) __PYX_ERR(0, 1052, __pyx_L1_error)

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1053
 *         for sell in proposal.sells:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)             # <<<<<<<<<<<<<<
 *             price = sell.price * (Decimal(1) + fee.percent)
 *             sell.price = market.c_quantize_order_price(self.trading_pair, price)
*/
    __pyx_t_10 = __pyx_v_self->_limit_order_type;
    __Pyx_INCREF(__pyx_t_10);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_TradeType); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_SELL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1052
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)
 *         for sell in proposal.sells:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,             # <<<<<<<<<<<<<<
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)
 *             price = sell.price * (Decimal(1) + fee.percent)
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_fee(__pyx_v_market, ((PyObject*)__pyx_t_2), ((PyObject*)__pyx_t_8), __pyx_t_10, __pyx_t_6, __pyx_t_9, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_fee, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1054
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)
 *             price = sell.price * (Decimal(1) + fee.percent)             # <<<<<<<<<<<<<<
 *             sell.price = market.c_quantize_order_price(self.trading_pair, price)
 * 
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
      assert(__pyx_t_9);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
      __pyx_t_11 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_mstate_global->__pyx_int_1};
      __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_fee, __pyx_mstate_global->__pyx_n_u_percent); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_9 = PyNumber_Add(__pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyNumber_Multiply(__pyx_t_1, __pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF_SET(__pyx_v_price, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1055
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)
 *             price = sell.price * (Decimal(1) + fee.percent)
 *             sell.price = market.c_quantize_order_price(self.trading_pair, price)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_did_fill_order(self, object order_filled_event):
*/
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_6))) __PYX_ERR(0, 1055, __pyx_L1_error)
    __pyx_t_9 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_6), __pyx_v_price); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_price, __pyx_t_9) < (0)) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1051
 *             price = buy.price * (Decimal(1) - fee.percent)
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)
 *         for sell in proposal.sells:             # <<<<<<<<<<<<<<
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)
*/
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1043
 *                 proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, higher_sell_price) * (1 + self.order_level_spread * i)
 * 
 *     cdef object c_apply_add_transaction_costs(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_apply_add_transaction_costs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_buy);
  __Pyx_XDECREF(__pyx_v_fee);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_sell);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1057
 *             sell.price = market.c_quantize_order_price(self.trading_pair, price)
 * 
 *     cdef c_did_fill_order(self, object order_filled_event):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str order_id = order_filled_event.order_id
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_did_fill_order(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_filled_event) {
  PyObject *__pyx_v_order_id = 0;
  PyObject *__pyx_v_market_info = 0;
  CYTHON_UNUSED PyObject *__pyx_v_order_fill_record = 0;
  struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *__pyx_v_limit_order_record = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9[7];
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_did_fill_order", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1059
 *     cdef c_did_fill_order(self, object order_filled_event):
 *         cdef:
 *             str order_id = order_filled_event.order_id             # <<<<<<<<<<<<<<
 *             object market_info = self._sb_order_tracker.c_get_shadow_market_pair_from_order_id(order_id)
 *             tuple order_fill_record
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_filled_event, __pyx_mstate_global->__pyx_n_u_order_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 1059, __pyx_L1_error)
  __pyx_v_order_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1060
 *         cdef:
 *             str order_id = order_filled_event.order_id
 *             object market_info = self._sb_order_tracker.c_get_shadow_market_pair_from_order_id(order_id)             # <<<<<<<<<<<<<<
 *             tuple order_fill_record
 * 
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_13order_tracker_OrderTracker *)__pyx_v_self->__pyx_base._sb_order_tracker->__pyx_base.__pyx_base.__pyx_vtab)->c_get_shadow_market_pair_from_order_id(__pyx_v_self->__pyx_base._sb_order_tracker, __pyx_v_order_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_market_info = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1063
 *             tuple order_fill_record
 * 
 *         if market_info is not None:             # <<<<<<<<<<<<<<
 *             limit_order_record = self._sb_order_tracker.c_get_shadow_limit_order(order_id)
 *             order_fill_record = (limit_order_record, order_filled_event)
*/
  __pyx_t_2 = (__pyx_v_market_info != Py_None);
  if (__pyx_t_2) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1064
 * 
 *         if market_info is not None:
 *             limit_order_record = self._sb_order_tracker.c_get_shadow_limit_order(order_id)             # <<<<<<<<<<<<<<
 *             order_fill_record = (limit_order_record, order_filled_event)
 * 
*/
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_10hummingbot_8strategy_13order_tracker_OrderTracker *)__pyx_v_self->__pyx_base._sb_order_tracker->__pyx_base.__pyx_base.__pyx_vtab)->c_get_shadow_limit_order(__pyx_v_self->__pyx_base._sb_order_tracker, __pyx_v_order_id)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_limit_order_record = ((struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1065
 *         if market_info is not None:
 *             limit_order_record = self._sb_order_tracker.c_get_shadow_limit_order(order_id)
 *             order_fill_record = (limit_order_record, order_filled_event)             # <<<<<<<<<<<<<<
 * 
 *             if order_filled_event.trade_type is TradeType.BUY:
*/
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1065, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF((PyObject *)__pyx_v_limit_order_record);
    __Pyx_GIVEREF((PyObject *)__pyx_v_limit_order_record);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_limit_order_record)) != (0)) __PYX_ERR(0, 1065, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_order_filled_event);
    __Pyx_GIVEREF(__pyx_v_order_filled_event);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_order_filled_event) != (0)) __PYX_ERR(0, 1065, __pyx_L1_error);
    __pyx_v_order_fill_record = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1067
 *             order_fill_record = (limit_order_record, order_filled_event)
 * 
 *             if order_filled_event.trade_type is TradeType.BUY:             # <<<<<<<<<<<<<<
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
 *                     self.log_with_clock(
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_filled_event, __pyx_mstate_global->__pyx_n_u_trade_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_TradeType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_BUY); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = (__pyx_t_1 == __pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1068
 * 
 *             if order_filled_event.trade_type is TradeType.BUY:
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:             # <<<<<<<<<<<<<<
 *                     self.log_with_clock(
 *                         logging.INFO,
*/
      __pyx_t_4 = __Pyx_PyLong_From_int64_t(__pyx_v_self->_logging_options); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1068, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_OPTION_LOG_MAKER_ORDER_FILLED); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1068, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyNumber_And(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1068, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1068, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_2) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1069
 *             if order_filled_event.trade_type is TradeType.BUY:
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
 *                     self.log_with_clock(             # <<<<<<<<<<<<<<
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker buy order of "
*/
        __pyx_t_1 = ((PyObject *)__pyx_v_self);
        __Pyx_INCREF(__pyx_t_1);

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1070
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
 *                     self.log_with_clock(
 *                         logging.INFO,             # <<<<<<<<<<<<<<
 *                         f"({market_info.trading_pair}) Maker buy order of "
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."
*/
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_logging); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_INFO); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1071
 *                     self.log_with_clock(
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker buy order of "             # <<<<<<<<<<<<<<
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."
 *                     )
*/
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_market_info, __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1071, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1071, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1072
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker buy order of "
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."             # <<<<<<<<<<<<<<
 *                     )
 *             else:
*/
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_filled_event, __pyx_mstate_global->__pyx_n_u_amount); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1072, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1072, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_market_info, __pyx_mstate_global->__pyx_n_u_base_asset); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1072, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1072, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_9[0] = __pyx_mstate_global->__pyx_kp_u__20;
        __pyx_t_9[1] = __pyx_t_6;
        __pyx_t_9[2] = __pyx_mstate_global->__pyx_kp_u_Maker_buy_order_of;
        __pyx_t_9[3] = __pyx_t_7;
        __pyx_t_9[4] = __pyx_mstate_global->__pyx_kp_u__21;
        __pyx_t_9[5] = __pyx_t_8;
        __pyx_t_9[6] = __pyx_mstate_global->__pyx_kp_u_filled;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1071
 *                     self.log_with_clock(
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker buy order of "             # <<<<<<<<<<<<<<
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."
 *                     )
*/
        __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_9, 7, 1 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6) + 21 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8) + 8, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8));
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1071, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_10 = 0;
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_t_5, __pyx_t_4};
          __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_log_with_clock, __pyx_callargs+__pyx_t_10, (3-__pyx_t_10) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1069, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1068
 * 
 *             if order_filled_event.trade_type is TradeType.BUY:
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:             # <<<<<<<<<<<<<<
 *                     self.log_with_clock(
 *                         logging.INFO,
*/
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1067
 *             order_fill_record = (limit_order_record, order_filled_event)
 * 
 *             if order_filled_event.trade_type is TradeType.BUY:             # <<<<<<<<<<<<<<
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
 *                     self.log_with_clock(
*/
      goto __pyx_L4;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1075
 *                     )
 *             else:
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:             # <<<<<<<<<<<<<<
 *                     self.log_with_clock(
 *                         logging.INFO,
*/
    /*else*/ {
      __pyx_t_3 = __Pyx_PyLong_From_int64_t(__pyx_v_self->_logging_options); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_OPTION_LOG_MAKER_ORDER_FILLED); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyNumber_And(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_2) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1076
 *             else:
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
 *                     self.log_with_clock(             # <<<<<<<<<<<<<<
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker sell order of "
*/
        __pyx_t_4 = ((PyObject *)__pyx_v_self);
        __Pyx_INCREF(__pyx_t_4);

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1077
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
 *                     self.log_with_clock(
 *                         logging.INFO,             # <<<<<<<<<<<<<<
 *                         f"({market_info.trading_pair}) Maker sell order of "
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."
*/
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_logging); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1077, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_INFO); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1077, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1078
 *                     self.log_with_clock(
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker sell order of "             # <<<<<<<<<<<<<<
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."
 *                     )
*/
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_market_info, __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1078, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_3, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1078, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1079
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker sell order of "
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."             # <<<<<<<<<<<<<<
 *                     )
 * 
*/
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_filled_event, __pyx_mstate_global->__pyx_n_u_amount); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1079, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_3, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1079, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_market_info, __pyx_mstate_global->__pyx_n_u_base_asset); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1079, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_t_3, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1079, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_9[0] = __pyx_mstate_global->__pyx_kp_u__20;
        __pyx_t_9[1] = __pyx_t_8;
        __pyx_t_9[2] = __pyx_mstate_global->__pyx_kp_u_Maker_sell_order_of;
        __pyx_t_9[3] = __pyx_t_7;
        __pyx_t_9[4] = __pyx_mstate_global->__pyx_kp_u__21;
        __pyx_t_9[5] = __pyx_t_6;
        __pyx_t_9[6] = __pyx_mstate_global->__pyx_kp_u_filled;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1078
 *                     self.log_with_clock(
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker sell order of "             # <<<<<<<<<<<<<<
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."
 *                     )
*/
        __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_9, 7, 1 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8) + 22 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6) + 8, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6));
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1078, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_10 = 0;
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_1, __pyx_t_3};
          __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_log_with_clock, __pyx_callargs+__pyx_t_10, (3-__pyx_t_10) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1076, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1075
 *                     )
 *             else:
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:             # <<<<<<<<<<<<<<
 *                     self.log_with_clock(
 *                         logging.INFO,
*/
      }
    }
    __pyx_L4:;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1082
 *                     )
 * 
 *             if self._inventory_cost_price_delegate is not None:             # <<<<<<<<<<<<<<
 *                 self._inventory_cost_price_delegate.process_order_fill_event(order_filled_event)
 * 
*/
    __pyx_t_2 = (__pyx_v_self->_inventory_cost_price_delegate != Py_None);
    if (__pyx_t_2) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1083
 * 
 *             if self._inventory_cost_price_delegate is not None:
 *                 self._inventory_cost_price_delegate.process_order_fill_event(order_filled_event)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_did_complete_buy_order(self, object order_completed_event):
*/
      __pyx_t_3 = __pyx_v_self->_inventory_cost_price_delegate;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_10 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_order_filled_event};
        __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_process_order_fill_event, __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1083, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1082
 *                     )
 * 
 *             if self._inventory_cost_price_delegate is not None:             # <<<<<<<<<<<<<<
 *                 self._inventory_cost_price_delegate.process_order_fill_event(order_filled_event)
 * 
*/
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1063
 *             tuple order_fill_record
 * 
 *         if market_info is not None:             # <<<<<<<<<<<<<<
 *             limit_order_record = self._sb_order_tracker.c_get_shadow_limit_order(order_id)
 *             order_fill_record = (limit_order_record, order_filled_event)
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1057
 *             sell.price = market.c_quantize_order_price(self.trading_pair, price)
 * 
 *     cdef c_did_fill_order(self, object order_filled_event):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str order_id = order_filled_event.order_id
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_did_fill_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_order_id);
  __Pyx_XDECREF(__pyx_v_market_info);
  __Pyx_XDECREF(__pyx_v_order_fill_record);
  __Pyx_XDECREF((PyObject *)__pyx_v_limit_order_record);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1085
 *                 self._inventory_cost_price_delegate.process_order_fill_event(order_filled_event)
 * 
 *     cdef c_did_complete_buy_order(self, object order_completed_event):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str order_id = order_completed_event.order_id
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_did_complete_buy_order(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_completed_event) {
  PyObject *__pyx_v_order_id = 0;
  PyObject *__pyx_v_limit_order_record = 0;
  CYTHON_UNUSED PyObject *__pyx_v_active_sell_ids = NULL;
  PyObject *__pyx_9genexpr17__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15[13];
  size_t __pyx_t_16;
  PyObject *__pyx_t_17[9];
  double __pyx_t_18;
  double __pyx_t_19;
  double __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_did_complete_buy_order", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1087
 *     cdef c_did_complete_buy_order(self, object order_completed_event):
 *         cdef:
 *             str order_id = order_completed_event.order_id             # <<<<<<<<<<<<<<
 *             limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)
 *         if limit_order_record is None:
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_completed_event, __pyx_mstate_global->__pyx_n_u_order_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 1087, __pyx_L1_error)
  __pyx_v_order_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1088
 *         cdef:
 *             str order_id = order_completed_event.order_id
 *             limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)             # <<<<<<<<<<<<<<
 *         if limit_order_record is None:
 *             return
*/
  __pyx_t_1 = __pyx_v_self->_market_info;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_10hummingbot_8strategy_13order_tracker_OrderTracker *)__pyx_v_self->__pyx_base._sb_order_tracker->__pyx_base.__pyx_base.__pyx_vtab)->c_get_limit_order(__pyx_v_self->__pyx_base._sb_order_tracker, __pyx_t_1, __pyx_v_order_id)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_limit_order_record = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1089
 *             str order_id = order_completed_event.order_id
 *             limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)
 *         if limit_order_record is None:             # <<<<<<<<<<<<<<
 *             return
 *         active_sell_ids = [x.client_order_id for x in self.active_orders if not x.is_buy]
*/
  __pyx_t_3 = (__pyx_v_limit_order_record == Py_None);
  if (__pyx_t_3) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1090
 *             limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)
 *         if limit_order_record is None:
 *             return             # <<<<<<<<<<<<<<
 *         active_sell_ids = [x.client_order_id for x in self.active_orders if not x.is_buy]
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1089
 *             str order_id = order_completed_event.order_id
 *             limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)
 *         if limit_order_record is None:             # <<<<<<<<<<<<<<
 *             return
 *         active_sell_ids = [x.client_order_id for x in self.active_orders if not x.is_buy]
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1091
 *         if limit_order_record is None:
 *             return
 *         active_sell_ids = [x.client_order_id for x in self.active_orders if not x.is_buy]             # <<<<<<<<<<<<<<
 * 
 *         if self._hanging_orders_enabled:
*/
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1091, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_4 = __pyx_t_1; __Pyx_INCREF(__pyx_t_4);
      __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1091, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1091, __pyx_L6_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1091, __pyx_L6_error)
            #endif
            if (__pyx_t_5 >= __pyx_temp) break;
          }
          __pyx_t_1 = __Pyx_PyList_GetItemRefFast(__pyx_t_4, __pyx_t_5, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_5;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1091, __pyx_L6_error)
            #endif
            if (__pyx_t_5 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5));
          #else
          __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_5);
          #endif
          ++__pyx_t_5;
        }
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L6_error)
      } else {
        __pyx_t_1 = __pyx_t_6(__pyx_t_4);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1091, __pyx_L6_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_9genexpr17__pyx_v_x, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr17__pyx_v_x, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 1091, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_7 = (!__pyx_t_3);
      if (__pyx_t_7) {
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr17__pyx_v_x, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 1091, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_x); __pyx_9genexpr17__pyx_v_x = 0;
    goto __pyx_L11_exit_scope;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_x); __pyx_9genexpr17__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L11_exit_scope:;
  } /* exit inner scope */
  __pyx_v_active_sell_ids = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1093
 *         active_sell_ids = [x.client_order_id for x in self.active_orders if not x.is_buy]
 * 
 *         if self._hanging_orders_enabled:             # <<<<<<<<<<<<<<
 *             # If the filled order is a hanging order, do nothing
 *             if order_id in self.hanging_order_ids:
*/
  if (__pyx_v_self->_hanging_orders_enabled) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1095
 *         if self._hanging_orders_enabled:
 *             # If the filled order is a hanging order, do nothing
 *             if order_id in self.hanging_order_ids:             # <<<<<<<<<<<<<<
 *                 self.log_with_clock(
 *                     logging.INFO,
*/
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_hanging_order_ids); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_v_order_id, __pyx_t_2, Py_EQ)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1095, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_7) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1096
 *             # If the filled order is a hanging order, do nothing
 *             if order_id in self.hanging_order_ids:
 *                 self.log_with_clock(             # <<<<<<<<<<<<<<
 *                     logging.INFO,
 *                     f"({self.trading_pair}) Hanging maker buy order {order_id} "
*/
      __pyx_t_4 = ((PyObject *)__pyx_v_self);
      __Pyx_INCREF(__pyx_t_4);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1097
 *             if order_id in self.hanging_order_ids:
 *                 self.log_with_clock(
 *                     logging.INFO,             # <<<<<<<<<<<<<<
 *                     f"({self.trading_pair}) Hanging maker buy order {order_id} "
 *                     f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
*/
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_logging); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1097, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_INFO); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1097, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1098
 *                 self.log_with_clock(
 *                     logging.INFO,
 *                     f"({self.trading_pair}) Hanging maker buy order {order_id} "             # <<<<<<<<<<<<<<
 *                     f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1098, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1098, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_v_order_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1098, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1099
 *                     logging.INFO,
 *                     f"({self.trading_pair}) Hanging maker buy order {order_id} "
 *                     f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
 *                 )
*/
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_quantity); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1099, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyObject_FormatSimple(__pyx_t_10, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1099, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_base_currency); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1099, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_12 = __Pyx_PyObject_FormatSimple(__pyx_t_10, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1099, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1100
 *                     f"({self.trading_pair}) Hanging maker buy order {order_id} "
 *                     f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."             # <<<<<<<<<<<<<<
 *                 )
 *                 self.notify_hb_app_with_timestamp(
*/
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_13 = __Pyx_PyObject_FormatSimple(__pyx_t_10, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_quote_currency); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_14 = __Pyx_PyObject_FormatSimple(__pyx_t_10, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_15[0] = __pyx_mstate_global->__pyx_kp_u__20;
      __pyx_t_15[1] = __pyx_t_9;
      __pyx_t_15[2] = __pyx_mstate_global->__pyx_kp_u_Hanging_maker_buy_order;
      __pyx_t_15[3] = __pyx_t_1;
      __pyx_t_15[4] = __pyx_mstate_global->__pyx_kp_u__22;
      __pyx_t_15[5] = __pyx_t_11;
      __pyx_t_15[6] = __pyx_mstate_global->__pyx_kp_u__21;
      __pyx_t_15[7] = __pyx_t_12;
      __pyx_t_15[8] = __pyx_mstate_global->__pyx_kp_u__23;
      __pyx_t_15[9] = __pyx_t_13;
      __pyx_t_15[10] = __pyx_mstate_global->__pyx_kp_u__21;
      __pyx_t_15[11] = __pyx_t_14;
      __pyx_t_15[12] = __pyx_mstate_global->__pyx_kp_u_has_been_completely_filled;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1098
 *                 self.log_with_clock(
 *                     logging.INFO,
 *                     f"({self.trading_pair}) Hanging maker buy order {order_id} "             # <<<<<<<<<<<<<<
 *                     f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
*/
      __pyx_t_10 = __Pyx_PyUnicode_Join(__pyx_t_15, 13, 1 * 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9) + 26 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12) + 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_14) + 29, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14));
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1098, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_16 = 0;
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_8, __pyx_t_10};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_log_with_clock, __pyx_callargs+__pyx_t_16, (3-__pyx_t_16) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1096, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1102
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
 *                 )
 *                 self.notify_hb_app_with_timestamp(             # <<<<<<<<<<<<<<
 *                     f"Hanging maker BUY order {limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
*/
      __pyx_t_10 = ((PyObject *)__pyx_v_self);
      __Pyx_INCREF(__pyx_t_10);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1103
 *                 )
 *                 self.notify_hb_app_with_timestamp(
 *                     f"Hanging maker BUY order {limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
 *                 )
*/
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_quantity); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_base_currency); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_14 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1104
 *                 self.notify_hb_app_with_timestamp(
 *                     f"Hanging maker BUY order {limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."             # <<<<<<<<<<<<<<
 *                 )
 *                 return
*/
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_13 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_quote_currency); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_12 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_17[0] = __pyx_mstate_global->__pyx_kp_u_Hanging_maker_BUY_order;
      __pyx_t_17[1] = __pyx_t_4;
      __pyx_t_17[2] = __pyx_mstate_global->__pyx_kp_u__21;
      __pyx_t_17[3] = __pyx_t_14;
      __pyx_t_17[4] = __pyx_mstate_global->__pyx_kp_u__23;
      __pyx_t_17[5] = __pyx_t_13;
      __pyx_t_17[6] = __pyx_mstate_global->__pyx_kp_u__21;
      __pyx_t_17[7] = __pyx_t_12;
      __pyx_t_17[8] = __pyx_mstate_global->__pyx_kp_u_is_filled;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1103
 *                 )
 *                 self.notify_hb_app_with_timestamp(
 *                     f"Hanging maker BUY order {limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
 *                 )
*/
      __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_17, 9, 24 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 1 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_14) + 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12) + 11, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12));
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_16 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_notify_hb_app_with_timestamp, __pyx_callargs+__pyx_t_16, (2-__pyx_t_16) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1102, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1106
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
 *                 )
 *                 return             # <<<<<<<<<<<<<<
 * 
 *         # delay order creation by filled_order_dalay (in seconds)
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1095
 *         if self._hanging_orders_enabled:
 *             # If the filled order is a hanging order, do nothing
 *             if order_id in self.hanging_order_ids:             # <<<<<<<<<<<<<<
 *                 self.log_with_clock(
 *                     logging.INFO,
*/
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1093
 *         active_sell_ids = [x.client_order_id for x in self.active_orders if not x.is_buy]
 * 
 *         if self._hanging_orders_enabled:             # <<<<<<<<<<<<<<
 *             # If the filled order is a hanging order, do nothing
 *             if order_id in self.hanging_order_ids:
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1109
 * 
 *         # delay order creation by filled_order_dalay (in seconds)
 *         self._create_timestamp = self._current_timestamp + self._filled_order_delay             # <<<<<<<<<<<<<<
 *         self._cancel_timestamp = min(self._cancel_timestamp, self._create_timestamp)
 * 
*/
  __pyx_v_self->_create_timestamp = (__pyx_v_self->__pyx_base.__pyx_base._current_timestamp + __pyx_v_self->_filled_order_delay);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1110
 *         # delay order creation by filled_order_dalay (in seconds)
 *         self._create_timestamp = self._current_timestamp + self._filled_order_delay
 *         self._cancel_timestamp = min(self._cancel_timestamp, self._create_timestamp)             # <<<<<<<<<<<<<<
 * 
 *         self._filled_buys_balance += 1
*/
  __pyx_t_18 = __pyx_v_self->_create_timestamp;
  __pyx_t_19 = __pyx_v_self->_cancel_timestamp;
  __pyx_t_7 = (__pyx_t_18 < __pyx_t_19);
  if (__pyx_t_7) {
    __pyx_t_20 = __pyx_t_18;
  } else {
    __pyx_t_20 = __pyx_t_19;
  }
  __pyx_v_self->_cancel_timestamp = __pyx_t_20;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1112
 *         self._cancel_timestamp = min(self._cancel_timestamp, self._create_timestamp)
 * 
 *         self._filled_buys_balance += 1             # <<<<<<<<<<<<<<
 *         self._last_own_trade_price = limit_order_record.price
 * 
*/
  __pyx_v_self->_filled_buys_balance = (__pyx_v_self->_filled_buys_balance + 1);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1113
 * 
 *         self._filled_buys_balance += 1
 *         self._last_own_trade_price = limit_order_record.price             # <<<<<<<<<<<<<<
 * 
 *         self.log_with_clock(
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_last_own_trade_price);
  __Pyx_DECREF(__pyx_v_self->_last_own_trade_price);
  __pyx_v_self->_last_own_trade_price = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1115
 *         self._last_own_trade_price = limit_order_record.price
 * 
 *         self.log_with_clock(             # <<<<<<<<<<<<<<
 *             logging.INFO,
 *             f"({self.trading_pair}) Maker buy order {order_id} "
*/
  __pyx_t_8 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_8);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1116
 * 
 *         self.log_with_clock(
 *             logging.INFO,             # <<<<<<<<<<<<<<
 *             f"({self.trading_pair}) Maker buy order {order_id} "
 *             f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
*/
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_logging); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_INFO); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1117
 *         self.log_with_clock(
 *             logging.INFO,
 *             f"({self.trading_pair}) Maker buy order {order_id} "             # <<<<<<<<<<<<<<
 *             f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *             f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
*/
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_13 = __Pyx_PyObject_FormatSimple(__pyx_t_10, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyUnicode_Unicode(__pyx_v_order_id); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1118
 *             logging.INFO,
 *             f"({self.trading_pair}) Maker buy order {order_id} "
 *             f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *             f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
 *         )
*/
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_quantity); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_14, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_base_currency); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_11 = __Pyx_PyObject_FormatSimple(__pyx_t_14, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1119
 *             f"({self.trading_pair}) Maker buy order {order_id} "
 *             f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *             f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."             # <<<<<<<<<<<<<<
 *         )
 *         self.notify_hb_app_with_timestamp(
*/
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_14, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_quote_currency); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_t_14, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_15[0] = __pyx_mstate_global->__pyx_kp_u__20;
  __pyx_t_15[1] = __pyx_t_13;
  __pyx_t_15[2] = __pyx_mstate_global->__pyx_kp_u_Maker_buy_order;
  __pyx_t_15[3] = __pyx_t_10;
  __pyx_t_15[4] = __pyx_mstate_global->__pyx_kp_u__22;
  __pyx_t_15[5] = __pyx_t_4;
  __pyx_t_15[6] = __pyx_mstate_global->__pyx_kp_u__21;
  __pyx_t_15[7] = __pyx_t_11;
  __pyx_t_15[8] = __pyx_mstate_global->__pyx_kp_u__23;
  __pyx_t_15[9] = __pyx_t_1;
  __pyx_t_15[10] = __pyx_mstate_global->__pyx_kp_u__21;
  __pyx_t_15[11] = __pyx_t_9;
  __pyx_t_15[12] = __pyx_mstate_global->__pyx_kp_u_has_been_completely_filled;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1117
 *         self.log_with_clock(
 *             logging.INFO,
 *             f"({self.trading_pair}) Maker buy order {order_id} "             # <<<<<<<<<<<<<<
 *             f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *             f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
*/
  __pyx_t_14 = __Pyx_PyUnicode_Join(__pyx_t_15, 13, 1 * 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13) + 18 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11) + 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9) + 29, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9));
  if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_16 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_t_12, __pyx_t_14};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_log_with_clock, __pyx_callargs+__pyx_t_16, (3-__pyx_t_16) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1121
 *             f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
 *         )
 *         self.notify_hb_app_with_timestamp(             # <<<<<<<<<<<<<<
 *             f"Maker BUY order {limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *             f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
*/
  __pyx_t_14 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_14);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1122
 *         )
 *         self.notify_hb_app_with_timestamp(
 *             f"Maker BUY order {limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *             f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
 *         )
*/
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_quantity); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_12, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_base_currency); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_t_12, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1123
 *         self.notify_hb_app_with_timestamp(
 *             f"Maker BUY order {limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *             f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."             # <<<<<<<<<<<<<<
 *         )
 * 
*/
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_12, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order_record, __pyx_mstate_global->__pyx_n_u_quote_currency); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_11 = __Pyx_PyObject_FormatSimple(__pyx_t_12, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_17[0] = __pyx_mstate_global->__pyx_kp_u_Maker_BUY_order;
  __pyx_t_17[1] = __pyx_t_8;
  __pyx_t_17[2] = __pyx_mstate_global->__pyx_kp_u__21;
  __pyx_t_17[3] = __pyx_t_9;
  __pyx_t_17[4] = __pyx_mstate_global->__pyx_kp_u__23;
  __pyx_t_17[5] = __pyx_t_1;
  __pyx_t_17[6] = __pyx_mstate_global->__pyx_kp_u__21;
  __pyx_t_17[7] = __pyx_t_11;
  __pyx_t_17[8] = __pyx_mstate_global->__pyx_kp_u_is_filled;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1122
 *         )
 *         self.notify_hb_app_with_timestamp(
 *             f"Maker BUY order {limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *             f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
 *         )
*/
  __pyx_t_12 = __Pyx_PyUnicode_Join(__pyx_t_17, 9, 16 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8) + 1 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9) + 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11) + 11, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11));
  if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_16 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_14, __pyx_t_12};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_notify_hb_app_with_timestamp, __pyx_callargs+__pyx_t_16, (2-__pyx_t_16) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1085
 *                 self._inventory_cost_price_delegate.process_order_fill_event(order_filled_event)
 * 
 *     cdef c_did_complete_buy_order(self, object order_completed_event):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str order_id = order_completed_event.order_id
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_did_complete_buy_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_order_id);
  __Pyx_XDECREF(__pyx_v_limit_order_record);
  __Pyx_XDECREF(__pyx_v_active_sell_ids);
  __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1126
 *         )
 * 
 *     cdef c_did_complete_sell_order(self, object order_completed_event):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str order_id = order_completed_event.order_id
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_did_complete_sell_order(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_completed_event) {
  PyObject *__pyx_v_order_id = 0;
  struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *__pyx_v_limit_order_record = 0;
  CYTHON_UNUSED PyObject *__pyx_v_active_buy_ids = NULL;
  PyObject *__pyx_9genexpr18__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14[13];
  size_t __pyx_t_15;
  PyObject *__pyx_t_16[9];
  double __pyx_t_17;
  double __pyx_t_18;
  double __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_did_complete_sell_order", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1128
 *     cdef c_did_complete_sell_order(self, object order_completed_event):
 *         cdef:
 *             str order_id = order_completed_event.order_id             # <<<<<<<<<<<<<<
 *             LimitOrder limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)
 *         if limit_order_record is None:
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_completed_event, __pyx_mstate_global->__pyx_n_u_order_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 1128, __pyx_L1_error)
  __pyx_v_order_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1129
 *         cdef:
 *             str order_id = order_completed_event.order_id
 *             LimitOrder limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)             # <<<<<<<<<<<<<<
 *         if limit_order_record is None:
 *             return
*/
  __pyx_t_1 = __pyx_v_self->_market_info;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_10hummingbot_8strategy_13order_tracker_OrderTracker *)__pyx_v_self->__pyx_base._sb_order_tracker->__pyx_base.__pyx_base.__pyx_vtab)->c_get_limit_order(__pyx_v_self->__pyx_base._sb_order_tracker, __pyx_t_1, __pyx_v_order_id)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_limit_order_record = ((struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1130
 *             str order_id = order_completed_event.order_id
 *             LimitOrder limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)
 *         if limit_order_record is None:             # <<<<<<<<<<<<<<
 *             return
 *         active_buy_ids = [x.client_order_id for x in self.active_orders if x.is_buy]
*/
  __pyx_t_3 = (((PyObject *)__pyx_v_limit_order_record) == Py_None);
  if (__pyx_t_3) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1131
 *             LimitOrder limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)
 *         if limit_order_record is None:
 *             return             # <<<<<<<<<<<<<<
 *         active_buy_ids = [x.client_order_id for x in self.active_orders if x.is_buy]
 *         if self._hanging_orders_enabled:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1130
 *             str order_id = order_completed_event.order_id
 *             LimitOrder limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)
 *         if limit_order_record is None:             # <<<<<<<<<<<<<<
 *             return
 *         active_buy_ids = [x.client_order_id for x in self.active_orders if x.is_buy]
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1132
 *         if limit_order_record is None:
 *             return
 *         active_buy_ids = [x.client_order_id for x in self.active_orders if x.is_buy]             # <<<<<<<<<<<<<<
 *         if self._hanging_orders_enabled:
 *             # If the filled order is a hanging order, do nothing
*/
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1132, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1132, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_4 = __pyx_t_1; __Pyx_INCREF(__pyx_t_4);
      __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1132, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1132, __pyx_L6_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1132, __pyx_L6_error)
            #endif
            if (__pyx_t_5 >= __pyx_temp) break;
          }
          __pyx_t_1 = __Pyx_PyList_GetItemRefFast(__pyx_t_4, __pyx_t_5, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_5;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1132, __pyx_L6_error)
            #endif
            if (__pyx_t_5 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5));
          #else
          __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_5);
          #endif
          ++__pyx_t_5;
        }
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1132, __pyx_L6_error)
      } else {
        __pyx_t_1 = __pyx_t_6(__pyx_t_4);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1132, __pyx_L6_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_9genexpr18__pyx_v_x, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr18__pyx_v_x, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1132, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 1132, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr18__pyx_v_x, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1132, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 1132, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_x); __pyx_9genexpr18__pyx_v_x = 0;
    goto __pyx_L11_exit_scope;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_x); __pyx_9genexpr18__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L11_exit_scope:;
  } /* exit inner scope */
  __pyx_v_active_buy_ids = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1133
 *             return
 *         active_buy_ids = [x.client_order_id for x in self.active_orders if x.is_buy]
 *         if self._hanging_orders_enabled:             # <<<<<<<<<<<<<<
 *             # If the filled order is a hanging order, do nothing
 *             if order_id in self.hanging_order_ids:
*/
  if (__pyx_v_self->_hanging_orders_enabled) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1135
 *         if self._hanging_orders_enabled:
 *             # If the filled order is a hanging order, do nothing
 *             if order_id in self.hanging_order_ids:             # <<<<<<<<<<<<<<
 *                 self.log_with_clock(
 *                     logging.INFO,
*/
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_hanging_order_ids); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_order_id, __pyx_t_2, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 1135, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_3) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1136
 *             # If the filled order is a hanging order, do nothing
 *             if order_id in self.hanging_order_ids:
 *                 self.log_with_clock(             # <<<<<<<<<<<<<<
 *                     logging.INFO,
 *                     f"({self.trading_pair}) Hanging maker sell order {order_id} "
*/
      __pyx_t_4 = ((PyObject *)__pyx_v_self);
      __Pyx_INCREF(__pyx_t_4);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1137
 *             if order_id in self.hanging_order_ids:
 *                 self.log_with_clock(
 *                     logging.INFO,             # <<<<<<<<<<<<<<
 *                     f"({self.trading_pair}) Hanging maker sell order {order_id} "
 *                     f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
*/
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_logging); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_INFO); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1138
 *                 self.log_with_clock(
 *                     logging.INFO,
 *                     f"({self.trading_pair}) Hanging maker sell order {order_id} "             # <<<<<<<<<<<<<<
 *                     f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_v_order_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1139
 *                     logging.INFO,
 *                     f"({self.trading_pair}) Hanging maker sell order {order_id} "
 *                     f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
 *                 )
*/
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_quantity); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_FormatSimple(__pyx_t_9, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_base_currency); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_11 = __Pyx_PyObject_FormatSimple(__pyx_t_9, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1140
 *                     f"({self.trading_pair}) Hanging maker sell order {order_id} "
 *                     f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."             # <<<<<<<<<<<<<<
 *                 )
 *                 self.notify_hb_app_with_timestamp(
*/
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_12 = __Pyx_PyObject_FormatSimple(__pyx_t_9, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_quote_currency); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_13 = __Pyx_PyObject_FormatSimple(__pyx_t_9, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_14[0] = __pyx_mstate_global->__pyx_kp_u__20;
      __pyx_t_14[1] = __pyx_t_8;
      __pyx_t_14[2] = __pyx_mstate_global->__pyx_kp_u_Hanging_maker_sell_order;
      __pyx_t_14[3] = __pyx_t_1;
      __pyx_t_14[4] = __pyx_mstate_global->__pyx_kp_u__22;
      __pyx_t_14[5] = __pyx_t_10;
      __pyx_t_14[6] = __pyx_mstate_global->__pyx_kp_u__21;
      __pyx_t_14[7] = __pyx_t_11;
      __pyx_t_14[8] = __pyx_mstate_global->__pyx_kp_u__23;
      __pyx_t_14[9] = __pyx_t_12;
      __pyx_t_14[10] = __pyx_mstate_global->__pyx_kp_u__21;
      __pyx_t_14[11] = __pyx_t_13;
      __pyx_t_14[12] = __pyx_mstate_global->__pyx_kp_u_has_been_completely_filled;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1138
 *                 self.log_with_clock(
 *                     logging.INFO,
 *                     f"({self.trading_pair}) Hanging maker sell order {order_id} "             # <<<<<<<<<<<<<<
 *                     f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
*/
      __pyx_t_9 = __Pyx_PyUnicode_Join(__pyx_t_14, 13, 1 * 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8) + 27 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11) + 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13) + 29, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13));
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_15 = 0;
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_7, __pyx_t_9};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_log_with_clock, __pyx_callargs+__pyx_t_15, (3-__pyx_t_15) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1136, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1142
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
 *                 )
 *                 self.notify_hb_app_with_timestamp(             # <<<<<<<<<<<<<<
 *                     f"Hanging maker SELL order {limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
*/
      __pyx_t_9 = ((PyObject *)__pyx_v_self);
      __Pyx_INCREF(__pyx_t_9);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1143
 *                 )
 *                 self.notify_hb_app_with_timestamp(
 *                     f"Hanging maker SELL order {limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
 *                 )
*/
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_quantity); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_base_currency); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_13 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1144
 *                 self.notify_hb_app_with_timestamp(
 *                     f"Hanging maker SELL order {limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."             # <<<<<<<<<<<<<<
 *                 )
 *                 return
*/
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_12 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_quote_currency); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_11 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_16[0] = __pyx_mstate_global->__pyx_kp_u_Hanging_maker_SELL_order;
      __pyx_t_16[1] = __pyx_t_4;
      __pyx_t_16[2] = __pyx_mstate_global->__pyx_kp_u__21;
      __pyx_t_16[3] = __pyx_t_13;
      __pyx_t_16[4] = __pyx_mstate_global->__pyx_kp_u__23;
      __pyx_t_16[5] = __pyx_t_12;
      __pyx_t_16[6] = __pyx_mstate_global->__pyx_kp_u__21;
      __pyx_t_16[7] = __pyx_t_11;
      __pyx_t_16[8] = __pyx_mstate_global->__pyx_kp_u_is_filled;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1143
 *                 )
 *                 self.notify_hb_app_with_timestamp(
 *                     f"Hanging maker SELL order {limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
 *                 )
*/
      __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_16, 9, 25 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 1 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13) + 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11) + 11, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11));
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_15 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_7};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_notify_hb_app_with_timestamp, __pyx_callargs+__pyx_t_15, (2-__pyx_t_15) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1142, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1146
 *                     f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
 *                 )
 *                 return             # <<<<<<<<<<<<<<
 * 
 *         # delay order creation by filled_order_dalay (in seconds)
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1135
 *         if self._hanging_orders_enabled:
 *             # If the filled order is a hanging order, do nothing
 *             if order_id in self.hanging_order_ids:             # <<<<<<<<<<<<<<
 *                 self.log_with_clock(
 *                     logging.INFO,
*/
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1133
 *             return
 *         active_buy_ids = [x.client_order_id for x in self.active_orders if x.is_buy]
 *         if self._hanging_orders_enabled:             # <<<<<<<<<<<<<<
 *             # If the filled order is a hanging order, do nothing
 *             if order_id in self.hanging_order_ids:
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1149
 * 
 *         # delay order creation by filled_order_dalay (in seconds)
 *         self._create_timestamp = self._current_timestamp + self._filled_order_delay             # <<<<<<<<<<<<<<
 *         self._cancel_timestamp = min(self._cancel_timestamp, self._create_timestamp)
 * 
*/
  __pyx_v_self->_create_timestamp = (__pyx_v_self->__pyx_base.__pyx_base._current_timestamp + __pyx_v_self->_filled_order_delay);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1150
 *         # delay order creation by filled_order_dalay (in seconds)
 *         self._create_timestamp = self._current_timestamp + self._filled_order_delay
 *         self._cancel_timestamp = min(self._cancel_timestamp, self._create_timestamp)             # <<<<<<<<<<<<<<
 * 
 *         self._filled_sells_balance += 1
*/
  __pyx_t_17 = __pyx_v_self->_create_timestamp;
  __pyx_t_18 = __pyx_v_self->_cancel_timestamp;
  __pyx_t_3 = (__pyx_t_17 < __pyx_t_18);
  if (__pyx_t_3) {
    __pyx_t_19 = __pyx_t_17;
  } else {
    __pyx_t_19 = __pyx_t_18;
  }
  __pyx_v_self->_cancel_timestamp = __pyx_t_19;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1152
 *         self._cancel_timestamp = min(self._cancel_timestamp, self._create_timestamp)
 * 
 *         self._filled_sells_balance += 1             # <<<<<<<<<<<<<<
 *         self._last_own_trade_price = limit_order_record.price
 * 
*/
  __pyx_v_self->_filled_sells_balance = (__pyx_v_self->_filled_sells_balance + 1);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1153
 * 
 *         self._filled_sells_balance += 1
 *         self._last_own_trade_price = limit_order_record.price             # <<<<<<<<<<<<<<
 * 
 *         self.log_with_clock(
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_last_own_trade_price);
  __Pyx_DECREF(__pyx_v_self->_last_own_trade_price);
  __pyx_v_self->_last_own_trade_price = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1155
 *         self._last_own_trade_price = limit_order_record.price
 * 
 *         self.log_with_clock(             # <<<<<<<<<<<<<<
 *             logging.INFO,
 *             f"({self.trading_pair}) Maker sell order {order_id} "
*/
  __pyx_t_7 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_7);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1156
 * 
 *         self.log_with_clock(
 *             logging.INFO,             # <<<<<<<<<<<<<<
 *             f"({self.trading_pair}) Maker sell order {order_id} "
 *             f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
*/
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_logging); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_INFO); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1157
 *         self.log_with_clock(
 *             logging.INFO,
 *             f"({self.trading_pair}) Maker sell order {order_id} "             # <<<<<<<<<<<<<<
 *             f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *             f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
*/
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_12 = __Pyx_PyObject_FormatSimple(__pyx_t_9, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyUnicode_Unicode(__pyx_v_order_id); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1158
 *             logging.INFO,
 *             f"({self.trading_pair}) Maker sell order {order_id} "
 *             f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *             f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
 *         )
*/
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_quantity); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_13, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_base_currency); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_10 = __Pyx_PyObject_FormatSimple(__pyx_t_13, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1159
 *             f"({self.trading_pair}) Maker sell order {order_id} "
 *             f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *             f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."             # <<<<<<<<<<<<<<
 *         )
 *         self.notify_hb_app_with_timestamp(
*/
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_13, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_quote_currency); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_13, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_14[0] = __pyx_mstate_global->__pyx_kp_u__20;
  __pyx_t_14[1] = __pyx_t_12;
  __pyx_t_14[2] = __pyx_mstate_global->__pyx_kp_u_Maker_sell_order;
  __pyx_t_14[3] = __pyx_t_9;
  __pyx_t_14[4] = __pyx_mstate_global->__pyx_kp_u__22;
  __pyx_t_14[5] = __pyx_t_4;
  __pyx_t_14[6] = __pyx_mstate_global->__pyx_kp_u__21;
  __pyx_t_14[7] = __pyx_t_10;
  __pyx_t_14[8] = __pyx_mstate_global->__pyx_kp_u__23;
  __pyx_t_14[9] = __pyx_t_1;
  __pyx_t_14[10] = __pyx_mstate_global->__pyx_kp_u__21;
  __pyx_t_14[11] = __pyx_t_8;
  __pyx_t_14[12] = __pyx_mstate_global->__pyx_kp_u_has_been_completely_filled;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1157
 *         self.log_with_clock(
 *             logging.INFO,
 *             f"({self.trading_pair}) Maker sell order {order_id} "             # <<<<<<<<<<<<<<
 *             f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *             f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
*/
  __pyx_t_13 = __Pyx_PyUnicode_Join(__pyx_t_14, 13, 1 * 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12) + 19 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10) + 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8) + 29, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8));
  if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_15 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_11, __pyx_t_13};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_log_with_clock, __pyx_callargs+__pyx_t_15, (3-__pyx_t_15) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1161
 *             f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
 *         )
 *         self.notify_hb_app_with_timestamp(             # <<<<<<<<<<<<<<
 *             f"Maker SELL order {limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *             f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
*/
  __pyx_t_13 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_13);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1162
 *         )
 *         self.notify_hb_app_with_timestamp(
 *             f"Maker SELL order {limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *             f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
 *         )
*/
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_quantity); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_11, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_base_currency); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_11, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1163
 *         self.notify_hb_app_with_timestamp(
 *             f"Maker SELL order {limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *             f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."             # <<<<<<<<<<<<<<
 *         )
 * 
*/
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_11, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_mstate_global->__pyx_n_u_quote_currency); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_10 = __Pyx_PyObject_FormatSimple(__pyx_t_11, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_16[0] = __pyx_mstate_global->__pyx_kp_u_Maker_SELL_order;
  __pyx_t_16[1] = __pyx_t_7;
  __pyx_t_16[2] = __pyx_mstate_global->__pyx_kp_u__21;
  __pyx_t_16[3] = __pyx_t_8;
  __pyx_t_16[4] = __pyx_mstate_global->__pyx_kp_u__23;
  __pyx_t_16[5] = __pyx_t_1;
  __pyx_t_16[6] = __pyx_mstate_global->__pyx_kp_u__21;
  __pyx_t_16[7] = __pyx_t_10;
  __pyx_t_16[8] = __pyx_mstate_global->__pyx_kp_u_is_filled;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1162
 *         )
 *         self.notify_hb_app_with_timestamp(
 *             f"Maker SELL order {limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *             f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
 *         )
*/
  __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_16, 9, 17 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7) + 1 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8) + 3 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10) + 11, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10));
  if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_15 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_t_11};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_notify_hb_app_with_timestamp, __pyx_callargs+__pyx_t_15, (2-__pyx_t_15) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1126
 *         )
 * 
 *     cdef c_did_complete_sell_order(self, object order_completed_event):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str order_id = order_completed_event.order_id
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_did_complete_sell_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_order_id);
  __Pyx_XDECREF((PyObject *)__pyx_v_limit_order_record);
  __Pyx_XDECREF(__pyx_v_active_buy_ids);
  __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1166
 *         )
 * 
 *     cdef bint c_is_within_tolerance(self, list current_prices, list proposal_prices):             # <<<<<<<<<<<<<<
 *         if len(current_prices) != len(proposal_prices):
 *             return False
*/

static int __pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_is_within_tolerance(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_current_prices, PyObject *__pyx_v_proposal_prices) {
  PyObject *__pyx_v_current = NULL;
  PyObject *__pyx_v_proposal = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_is_within_tolerance", 0);
  __Pyx_INCREF(__pyx_v_current_prices);
  __Pyx_INCREF(__pyx_v_proposal_prices);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1167
 * 
 *     cdef bint c_is_within_tolerance(self, list current_prices, list proposal_prices):
 *         if len(current_prices) != len(proposal_prices):             # <<<<<<<<<<<<<<
 *             return False
 *         current_prices = sorted(current_prices)
*/
  if (unlikely(__pyx_v_current_prices == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1167, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_current_prices); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1167, __pyx_L1_error)
  if (unlikely(__pyx_v_proposal_prices == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1167, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_v_proposal_prices); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1167, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_1 != __pyx_t_2);
  if (__pyx_t_3) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1168
 *     cdef bint c_is_within_tolerance(self, list current_prices, list proposal_prices):
 *         if len(current_prices) != len(proposal_prices):
 *             return False             # <<<<<<<<<<<<<<
 *         current_prices = sorted(current_prices)
 *         proposal_prices = sorted(proposal_prices)
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1167
 * 
 *     cdef bint c_is_within_tolerance(self, list current_prices, list proposal_prices):
 *         if len(current_prices) != len(proposal_prices):             # <<<<<<<<<<<<<<
 *             return False
 *         current_prices = sorted(current_prices)
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1169
 *         if len(current_prices) != len(proposal_prices):
 *             return False
 *         current_prices = sorted(current_prices)             # <<<<<<<<<<<<<<
 *         proposal_prices = sorted(proposal_prices)
 *         for current, proposal in zip(current_prices, proposal_prices):
*/
  __pyx_t_4 = PySequence_List(__pyx_v_current_prices); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely((PyList_Sort(__pyx_t_4) < 0))) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_current_prices, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1170
 *             return False
 *         current_prices = sorted(current_prices)
 *         proposal_prices = sorted(proposal_prices)             # <<<<<<<<<<<<<<
 *         for current, proposal in zip(current_prices, proposal_prices):
 *             # if spread diff is more than the tolerance or order quantities are different, return false.
*/
  __pyx_t_4 = PySequence_List(__pyx_v_proposal_prices); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely((PyList_Sort(__pyx_t_4) < 0))) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_proposal_prices, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1171
 *         current_prices = sorted(current_prices)
 *         proposal_prices = sorted(proposal_prices)
 *         for current, proposal in zip(current_prices, proposal_prices):             # <<<<<<<<<<<<<<
 *             # if spread diff is more than the tolerance or order quantities are different, return false.
 *             if abs(proposal - current)/current > self._order_refresh_tolerance_pct:
*/
  __pyx_t_5 = NULL;
  __pyx_t_6 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_current_prices, __pyx_v_proposal_prices};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_zip, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_5 = __pyx_t_4; __Pyx_INCREF(__pyx_t_5);
    __pyx_t_2 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1171, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1171, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_5, __pyx_t_2, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_2;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1171, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_2));
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_2);
        #endif
        ++__pyx_t_2;
      }
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1171, __pyx_L1_error)
    } else {
      __pyx_t_4 = __pyx_t_7(__pyx_t_5);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1171, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_4);
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1171, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0);
        __Pyx_INCREF(__pyx_t_8);
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1);
        __Pyx_INCREF(__pyx_t_9);
      } else {
        __pyx_t_8 = __Pyx_PyList_GetItemRefFast(sequence, 0, __Pyx_ReferenceSharing_SharedReference);
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1171, __pyx_L1_error)
        __Pyx_XGOTREF(__pyx_t_8);
        __pyx_t_9 = __Pyx_PyList_GetItemRefFast(sequence, 1, __Pyx_ReferenceSharing_SharedReference);
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1171, __pyx_L1_error)
        __Pyx_XGOTREF(__pyx_t_9);
      }
      #else
      __pyx_t_8 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1171, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1171, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_10 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1171, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_11 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_10);
      index = 0; __pyx_t_8 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_8)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 1; __pyx_t_9 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_9)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 2) < (0)) __PYX_ERR(0, 1171, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1171, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_current, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_proposal, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1173
 *         for current, proposal in zip(current_prices, proposal_prices):
 *             # if spread diff is more than the tolerance or order quantities are different, return false.
 *             if abs(proposal - current)/current > self._order_refresh_tolerance_pct:             # <<<<<<<<<<<<<<
 *                 return False
 *         return True
*/
    __pyx_t_4 = PyNumber_Subtract(__pyx_v_proposal, __pyx_v_current); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = __Pyx_PyNumber_Absolute(__pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyNumber_Divide(__pyx_t_9, __pyx_v_current); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyObject_RichCompare(__pyx_t_4, __pyx_v_self->_order_refresh_tolerance_pct, Py_GT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1173, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 1173, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (__pyx_t_3) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1174
 *             # if spread diff is more than the tolerance or order quantities are different, return false.
 *             if abs(proposal - current)/current > self._order_refresh_tolerance_pct:
 *                 return False             # <<<<<<<<<<<<<<
 *         return True
 * 
*/
      __pyx_r = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1173
 *         for current, proposal in zip(current_prices, proposal_prices):
 *             # if spread diff is more than the tolerance or order quantities are different, return false.
 *             if abs(proposal - current)/current > self._order_refresh_tolerance_pct:             # <<<<<<<<<<<<<<
 *                 return False
 *         return True
*/
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1171
 *         current_prices = sorted(current_prices)
 *         proposal_prices = sorted(proposal_prices)
 *         for current, proposal in zip(current_prices, proposal_prices):             # <<<<<<<<<<<<<<
 *             # if spread diff is more than the tolerance or order quantities are different, return false.
 *             if abs(proposal - current)/current > self._order_refresh_tolerance_pct:
*/
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1175
 *             if abs(proposal - current)/current > self._order_refresh_tolerance_pct:
 *                 return False
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cdef c_cancel_active_orders_on_max_age_limit(self):
*/
  __pyx_r = 1;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1166
 *         )
 * 
 *     cdef bint c_is_within_tolerance(self, list current_prices, list proposal_prices):             # <<<<<<<<<<<<<<
 *         if len(current_prices) != len(proposal_prices):
 *             return False
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_is_within_tolerance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_current);
  __Pyx_XDECREF(__pyx_v_proposal);
  __Pyx_XDECREF(__pyx_v_current_prices);
  __Pyx_XDECREF(__pyx_v_proposal_prices);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_39c_cancel_active_orders_on_max_age_limit_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1184
 *             list active_orders = self.active_non_hanging_orders
 * 
 *         if active_orders and any(order_age(o, self._current_timestamp) > self._max_order_age for o in active_orders):             # <<<<<<<<<<<<<<
 *             for order in active_orders:
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
*/

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_39c_cancel_active_orders_on_max_age_limit_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1184, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_39c_cancel_active_orders_on_max_age_limit_2generator1, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4]), (PyObject *) __pyx_cur_scope, __pyx_mstate_global->__pyx_n_u_genexpr, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_c_cance, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market); if (unlikely(!gen)) __PYX_ERR(0, 1184, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_cancel_active_orders_on_max_age_limit.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_39c_cancel_active_orders_on_max_age_limit_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr *__pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1184, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 1184, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_genexpr_arg_0 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1184, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1184, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    __pyx_t_3 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_2, __Pyx_ReferenceSharing_OwnStrongReference);
    ++__pyx_t_2;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_o);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_o, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_order_age); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 1184, __pyx_L1_error) }
    __pyx_t_6 = PyFloat_FromDouble(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_o, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 1184, __pyx_L1_error) }
    __pyx_t_5 = PyFloat_FromDouble(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->_max_order_age); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_3, __pyx_t_5, Py_GT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1184, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1184, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_8) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  if (__Pyx_PyErr_Occurred()) {
    __Pyx_Generator_Replace_StopIteration(0);
    __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  }
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1177
 *         return True
 * 
 *     cdef c_cancel_active_orders_on_max_age_limit(self):             # <<<<<<<<<<<<<<
 *         """
 *         Cancels active non hanging orders if they are older than max age limit
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_cancel_active_orders_on_max_age_limit(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit *__pyx_cur_scope;
  PyObject *__pyx_v_active_orders = 0;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_39c_cancel_active_orders_on_max_age_limit_2generator1 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_cancel_active_orders_on_max_age_limit", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1177, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1182
 *         """
 *         cdef:
 *             list active_orders = self.active_non_hanging_orders             # <<<<<<<<<<<<<<
 * 
 *         if active_orders and any(order_age(o, self._current_timestamp) > self._max_order_age for o in active_orders):
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_non_hanging_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 1182, __pyx_L1_error)
  __pyx_v_active_orders = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1184
 *             list active_orders = self.active_non_hanging_orders
 * 
 *         if active_orders and any(order_age(o, self._current_timestamp) > self._max_order_age for o in active_orders):             # <<<<<<<<<<<<<<
 *             for order in active_orders:
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
*/
  if (__pyx_v_active_orders == Py_None) __pyx_t_3 = 0;
  else
  {
    Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_v_active_orders);
    if (unlikely(((!CYTHON_ASSUME_SAFE_SIZE) && __pyx_temp < 0))) __PYX_ERR(0, 1184, __pyx_L1_error)
    __pyx_t_3 = (__pyx_temp != 0);
  }

  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_39c_cancel_active_orders_on_max_age_limit_genexpr(((PyObject*)__pyx_cur_scope), __pyx_v_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_Generator_GetInlinedResult(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1185
 * 
 *         if active_orders and any(order_age(o, self._current_timestamp) > self._max_order_age for o in active_orders):
 *             for order in active_orders:             # <<<<<<<<<<<<<<
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
 * 
*/
    if (unlikely(__pyx_v_active_orders == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1185, __pyx_L1_error)
    }
    __pyx_t_4 = __pyx_v_active_orders; __Pyx_INCREF(__pyx_t_4);
    __pyx_t_5 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1185, __pyx_L1_error)
        #endif
        if (__pyx_t_5 >= __pyx_temp) break;
      }
      __pyx_t_1 = __Pyx_PyList_GetItemRefFast(__pyx_t_4, __pyx_t_5, __Pyx_ReferenceSharing_OwnStrongReference);
      ++__pyx_t_5;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1186
 *         if active_orders and any(order_age(o, self._current_timestamp) > self._max_order_age for o in active_orders):
 *             for order in active_orders:
 *                 self.c_cancel_order(self._market_info, order.client_order_id)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_cancel_active_orders(self, object proposal):
*/
      __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->_market_info;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_6))) __PYX_ERR(0, 1186, __pyx_L1_error)
      __pyx_t_7 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_cancel_order(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_cur_scope->__pyx_v_self), __pyx_t_1, ((PyObject*)__pyx_t_6)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1185
 * 
 *         if active_orders and any(order_age(o, self._current_timestamp) > self._max_order_age for o in active_orders):
 *             for order in active_orders:             # <<<<<<<<<<<<<<
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
 * 
*/
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1184
 *             list active_orders = self.active_non_hanging_orders
 * 
 *         if active_orders and any(order_age(o, self._current_timestamp) > self._max_order_age for o in active_orders):             # <<<<<<<<<<<<<<
 *             for order in active_orders:
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1177
 *         return True
 * 
 *     cdef c_cancel_active_orders_on_max_age_limit(self):             # <<<<<<<<<<<<<<
 *         """
 *         Cancels active non hanging orders if they are older than max age limit
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_cancel_active_orders_on_max_age_limit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_active_orders);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XDECREF(__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_39c_cancel_active_orders_on_max_age_limit_2generator1);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1188
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
 * 
 *     cdef c_cancel_active_orders(self, object proposal):             # <<<<<<<<<<<<<<
 *         """
 *         Cancels active non hanging orders, checks if the order prices are within tolerance threshold
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_cancel_active_orders(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_v_active_orders = 0;
  PyObject *__pyx_v_active_buy_prices = 0;
  CYTHON_UNUSED PyObject *__pyx_v_active_sells = 0;
  int __pyx_v_to_defer_canceling;
  PyObject *__pyx_v_active_sell_prices = NULL;
  PyObject *__pyx_v_proposal_buys = NULL;
  PyObject *__pyx_v_proposal_sells = NULL;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_9genexpr20__pyx_v_o = NULL;
  PyObject *__pyx_9genexpr21__pyx_v_o = NULL;
  PyObject *__pyx_9genexpr22__pyx_v_buy = NULL;
  PyObject *__pyx_9genexpr23__pyx_v_sell = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  size_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_cancel_active_orders", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1192
 *         Cancels active non hanging orders, checks if the order prices are within tolerance threshold
 *         """
 *         if self._cancel_timestamp > self._current_timestamp:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
  __pyx_t_1 = (__pyx_v_self->_cancel_timestamp > __pyx_v_self->__pyx_base.__pyx_base._current_timestamp);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1193
 *         """
 *         if self._cancel_timestamp > self._current_timestamp:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         cdef:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1192
 *         Cancels active non hanging orders, checks if the order prices are within tolerance threshold
 *         """
 *         if self._cancel_timestamp > self._current_timestamp:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1196
 * 
 *         cdef:
 *             list active_orders = self.active_non_hanging_orders             # <<<<<<<<<<<<<<
 *             list active_buy_prices = []
 *             list active_sells = []
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_non_hanging_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 1196, __pyx_L1_error)
  __pyx_v_active_orders = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1197
 *         cdef:
 *             list active_orders = self.active_non_hanging_orders
 *             list active_buy_prices = []             # <<<<<<<<<<<<<<
 *             list active_sells = []
 *             bint to_defer_canceling = False
*/
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_active_buy_prices = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1198
 *             list active_orders = self.active_non_hanging_orders
 *             list active_buy_prices = []
 *             list active_sells = []             # <<<<<<<<<<<<<<
 *             bint to_defer_canceling = False
 *         if len(active_orders) == 0:
*/
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_active_sells = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1199
 *             list active_buy_prices = []
 *             list active_sells = []
 *             bint to_defer_canceling = False             # <<<<<<<<<<<<<<
 *         if len(active_orders) == 0:
 *             return
*/
  __pyx_v_to_defer_canceling = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1200
 *             list active_sells = []
 *             bint to_defer_canceling = False
 *         if len(active_orders) == 0:             # <<<<<<<<<<<<<<
 *             return
 *         if proposal is not None and \
*/
  if (unlikely(__pyx_v_active_orders == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1200, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyList_GET_SIZE(__pyx_v_active_orders); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1200, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_3 == 0);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1201
 *             bint to_defer_canceling = False
 *         if len(active_orders) == 0:
 *             return             # <<<<<<<<<<<<<<
 *         if proposal is not None and \
 *                 self._order_refresh_tolerance_pct >= 0:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1200
 *             list active_sells = []
 *             bint to_defer_canceling = False
 *         if len(active_orders) == 0:             # <<<<<<<<<<<<<<
 *             return
 *         if proposal is not None and \
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1202
 *         if len(active_orders) == 0:
 *             return
 *         if proposal is not None and \             # <<<<<<<<<<<<<<
 *                 self._order_refresh_tolerance_pct >= 0:
 * 
*/
  __pyx_t_4 = (__pyx_v_proposal != Py_None);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1203
 *             return
 *         if proposal is not None and \
 *                 self._order_refresh_tolerance_pct >= 0:             # <<<<<<<<<<<<<<
 * 
 *             active_buy_prices = [Decimal(str(o.price)) for o in active_orders if o.is_buy]
*/
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_self->_order_refresh_tolerance_pct, __pyx_mstate_global->__pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1203, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1203, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L6_bool_binop_done:;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1202
 *         if len(active_orders) == 0:
 *             return
 *         if proposal is not None and \             # <<<<<<<<<<<<<<
 *                 self._order_refresh_tolerance_pct >= 0:
 * 
*/
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1205
 *                 self._order_refresh_tolerance_pct >= 0:
 * 
 *             active_buy_prices = [Decimal(str(o.price)) for o in active_orders if o.is_buy]             # <<<<<<<<<<<<<<
 *             active_sell_prices = [Decimal(str(o.price)) for o in active_orders if not o.is_buy]
 *             proposal_buys = [buy.price for buy in proposal.buys]
*/
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1205, __pyx_L10_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__pyx_v_active_orders == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 1205, __pyx_L10_error)
      }
      __pyx_t_5 = __pyx_v_active_orders; __Pyx_INCREF(__pyx_t_5);
      __pyx_t_3 = 0;
      for (;;) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1205, __pyx_L10_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        __pyx_t_6 = __Pyx_PyList_GetItemRefFast(__pyx_t_5, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_3;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1205, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_XDECREF_SET(__pyx_9genexpr20__pyx_v_o, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr20__pyx_v_o, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1205, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1205, __pyx_L10_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_1) {
          __pyx_t_7 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1205, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr20__pyx_v_o, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1205, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_10 = __Pyx_PyObject_Unicode(__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1205, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_11 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
            assert(__pyx_t_7);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
            __pyx_t_11 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
            __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1205, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_6);
          }
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 1205, __pyx_L10_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_o); __pyx_9genexpr20__pyx_v_o = 0;
      goto __pyx_L15_exit_scope;
      __pyx_L10_error:;
      __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_o); __pyx_9genexpr20__pyx_v_o = 0;
      goto __pyx_L1_error;
      __pyx_L15_exit_scope:;
    } /* exit inner scope */
    __Pyx_DECREF_SET(__pyx_v_active_buy_prices, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1206
 * 
 *             active_buy_prices = [Decimal(str(o.price)) for o in active_orders if o.is_buy]
 *             active_sell_prices = [Decimal(str(o.price)) for o in active_orders if not o.is_buy]             # <<<<<<<<<<<<<<
 *             proposal_buys = [buy.price for buy in proposal.buys]
 *             proposal_sells = [sell.price for sell in proposal.sells]
*/
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1206, __pyx_L18_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__pyx_v_active_orders == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 1206, __pyx_L18_error)
      }
      __pyx_t_5 = __pyx_v_active_orders; __Pyx_INCREF(__pyx_t_5);
      __pyx_t_3 = 0;
      for (;;) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1206, __pyx_L18_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        __pyx_t_6 = __Pyx_PyList_GetItemRefFast(__pyx_t_5, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_3;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1206, __pyx_L18_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_XDECREF_SET(__pyx_9genexpr21__pyx_v_o, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr21__pyx_v_o, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1206, __pyx_L18_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1206, __pyx_L18_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_4 = (!__pyx_t_1);
        if (__pyx_t_4) {
          __pyx_t_8 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1206, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr21__pyx_v_o, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1206, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_9 = __Pyx_PyObject_Unicode(__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1206, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_11 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_10);
            assert(__pyx_t_8);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_10, __pyx__function);
            __pyx_t_11 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_9};
            __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_10, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1206, __pyx_L18_error)
            __Pyx_GOTREF(__pyx_t_6);
          }
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 1206, __pyx_L18_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_o); __pyx_9genexpr21__pyx_v_o = 0;
      goto __pyx_L23_exit_scope;
      __pyx_L18_error:;
      __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_o); __pyx_9genexpr21__pyx_v_o = 0;
      goto __pyx_L1_error;
      __pyx_L23_exit_scope:;
    } /* exit inner scope */
    __pyx_v_active_sell_prices = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1207
 *             active_buy_prices = [Decimal(str(o.price)) for o in active_orders if o.is_buy]
 *             active_sell_prices = [Decimal(str(o.price)) for o in active_orders if not o.is_buy]
 *             proposal_buys = [buy.price for buy in proposal.buys]             # <<<<<<<<<<<<<<
 *             proposal_sells = [sell.price for sell in proposal.sells]
 * 
*/
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1207, __pyx_L26_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1207, __pyx_L26_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
        __pyx_t_6 = __pyx_t_5; __Pyx_INCREF(__pyx_t_6);
        __pyx_t_3 = 0;
        __pyx_t_12 = NULL;
      } else {
        __pyx_t_3 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1207, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_12 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1207, __pyx_L26_error)
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      for (;;) {
        if (likely(!__pyx_t_12)) {
          if (likely(PyList_CheckExact(__pyx_t_6))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1207, __pyx_L26_error)
              #endif
              if (__pyx_t_3 >= __pyx_temp) break;
            }
            __pyx_t_5 = __Pyx_PyList_GetItemRefFast(__pyx_t_6, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
            ++__pyx_t_3;
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1207, __pyx_L26_error)
              #endif
              if (__pyx_t_3 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_3));
            #else
            __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_3);
            #endif
            ++__pyx_t_3;
          }
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1207, __pyx_L26_error)
        } else {
          __pyx_t_5 = __pyx_t_12(__pyx_t_6);
          if (unlikely(!__pyx_t_5)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1207, __pyx_L26_error)
              PyErr_Clear();
            }
            break;
          }
        }
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_XDECREF_SET(__pyx_9genexpr22__pyx_v_buy, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr22__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1207, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 1207, __pyx_L26_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_9genexpr22__pyx_v_buy); __pyx_9genexpr22__pyx_v_buy = 0;
      goto __pyx_L30_exit_scope;
      __pyx_L26_error:;
      __Pyx_XDECREF(__pyx_9genexpr22__pyx_v_buy); __pyx_9genexpr22__pyx_v_buy = 0;
      goto __pyx_L1_error;
      __pyx_L30_exit_scope:;
    } /* exit inner scope */
    __pyx_v_proposal_buys = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1208
 *             active_sell_prices = [Decimal(str(o.price)) for o in active_orders if not o.is_buy]
 *             proposal_buys = [buy.price for buy in proposal.buys]
 *             proposal_sells = [sell.price for sell in proposal.sells]             # <<<<<<<<<<<<<<
 * 
 *             if self.c_is_within_tolerance(active_buy_prices, proposal_buys) and \
*/
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1208, __pyx_L33_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1208, __pyx_L33_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
        __pyx_t_5 = __pyx_t_6; __Pyx_INCREF(__pyx_t_5);
        __pyx_t_3 = 0;
        __pyx_t_12 = NULL;
      } else {
        __pyx_t_3 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1208, __pyx_L33_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_12 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1208, __pyx_L33_error)
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      for (;;) {
        if (likely(!__pyx_t_12)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1208, __pyx_L33_error)
              #endif
              if (__pyx_t_3 >= __pyx_temp) break;
            }
            __pyx_t_6 = __Pyx_PyList_GetItemRefFast(__pyx_t_5, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
            ++__pyx_t_3;
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
              #if !CYTHON_ASSUME_SAFE_SIZE
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1208, __pyx_L33_error)
              #endif
              if (__pyx_t_3 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_3));
            #else
            __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_3);
            #endif
            ++__pyx_t_3;
          }
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1208, __pyx_L33_error)
        } else {
          __pyx_t_6 = __pyx_t_12(__pyx_t_5);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1208, __pyx_L33_error)
              PyErr_Clear();
            }
            break;
          }
        }
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_XDECREF_SET(__pyx_9genexpr23__pyx_v_sell, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr23__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1208, __pyx_L33_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 1208, __pyx_L33_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_sell); __pyx_9genexpr23__pyx_v_sell = 0;
      goto __pyx_L37_exit_scope;
      __pyx_L33_error:;
      __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_sell); __pyx_9genexpr23__pyx_v_sell = 0;
      goto __pyx_L1_error;
      __pyx_L37_exit_scope:;
    } /* exit inner scope */
    __pyx_v_proposal_sells = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1210
 *             proposal_sells = [sell.price for sell in proposal.sells]
 * 
 *             if self.c_is_within_tolerance(active_buy_prices, proposal_buys) and \             # <<<<<<<<<<<<<<
 *                     self.c_is_within_tolerance(active_sell_prices, proposal_sells):
 *                 to_defer_canceling = True
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_is_within_tolerance(__pyx_v_self, __pyx_v_active_buy_prices, __pyx_v_proposal_buys); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1210, __pyx_L1_error)
    if (__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L39_bool_binop_done;
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1211
 * 
 *             if self.c_is_within_tolerance(active_buy_prices, proposal_buys) and \
 *                     self.c_is_within_tolerance(active_sell_prices, proposal_sells):             # <<<<<<<<<<<<<<
 *                 to_defer_canceling = True
 * 
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_is_within_tolerance(__pyx_v_self, __pyx_v_active_sell_prices, __pyx_v_proposal_sells); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1211, __pyx_L1_error)
    __pyx_t_4 = __pyx_t_1;
    __pyx_L39_bool_binop_done:;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1210
 *             proposal_sells = [sell.price for sell in proposal.sells]
 * 
 *             if self.c_is_within_tolerance(active_buy_prices, proposal_buys) and \             # <<<<<<<<<<<<<<
 *                     self.c_is_within_tolerance(active_sell_prices, proposal_sells):
 *                 to_defer_canceling = True
*/
    if (__pyx_t_4) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1212
 *             if self.c_is_within_tolerance(active_buy_prices, proposal_buys) and \
 *                     self.c_is_within_tolerance(active_sell_prices, proposal_sells):
 *                 to_defer_canceling = True             # <<<<<<<<<<<<<<
 * 
 *         if not to_defer_canceling:
*/
      __pyx_v_to_defer_canceling = 1;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1210
 *             proposal_sells = [sell.price for sell in proposal.sells]
 * 
 *             if self.c_is_within_tolerance(active_buy_prices, proposal_buys) and \             # <<<<<<<<<<<<<<
 *                     self.c_is_within_tolerance(active_sell_prices, proposal_sells):
 *                 to_defer_canceling = True
*/
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1202
 *         if len(active_orders) == 0:
 *             return
 *         if proposal is not None and \             # <<<<<<<<<<<<<<
 *                 self._order_refresh_tolerance_pct >= 0:
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1214
 *                 to_defer_canceling = True
 * 
 *         if not to_defer_canceling:             # <<<<<<<<<<<<<<
 *             self._hanging_orders_tracker.update_strategy_orders_with_equivalent_orders()
 *             for order in self.active_non_hanging_orders:
*/
  __pyx_t_4 = (!__pyx_v_to_defer_canceling);
  if (__pyx_t_4) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1215
 * 
 *         if not to_defer_canceling:
 *             self._hanging_orders_tracker.update_strategy_orders_with_equivalent_orders()             # <<<<<<<<<<<<<<
 *             for order in self.active_non_hanging_orders:
 *                 # If is about to be added to hanging_orders then don't cancel
*/
    __pyx_t_5 = __pyx_v_self->_hanging_orders_tracker;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_11 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_update_strategy_orders_with_equi, __pyx_callargs+__pyx_t_11, (1-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1216
 *         if not to_defer_canceling:
 *             self._hanging_orders_tracker.update_strategy_orders_with_equivalent_orders()
 *             for order in self.active_non_hanging_orders:             # <<<<<<<<<<<<<<
 *                 # If is about to be added to hanging_orders then don't cancel
 *                 if not self._hanging_orders_tracker.is_potential_hanging_order(order):
*/
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_non_hanging_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_5 = __pyx_t_2; __Pyx_INCREF(__pyx_t_5);
      __pyx_t_3 = 0;
      __pyx_t_12 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1216, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_12 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1216, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_12)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1216, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_5, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_3;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1216, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_3));
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_3);
          #endif
          ++__pyx_t_3;
        }
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1216, __pyx_L1_error)
      } else {
        __pyx_t_2 = __pyx_t_12(__pyx_t_5);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1216, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1218
 *             for order in self.active_non_hanging_orders:
 *                 # If is about to be added to hanging_orders then don't cancel
 *                 if not self._hanging_orders_tracker.is_potential_hanging_order(order):             # <<<<<<<<<<<<<<
 *                     self.c_cancel_order(self._market_info, order.client_order_id)
 *         # else:
*/
      __pyx_t_6 = __pyx_v_self->_hanging_orders_tracker;
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_11 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_order};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_potential_hanging_order, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1218, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_1 = (!__pyx_t_4);
      if (__pyx_t_1) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1219
 *                 # If is about to be added to hanging_orders then don't cancel
 *                 if not self._hanging_orders_tracker.is_potential_hanging_order(order):
 *                     self.c_cancel_order(self._market_info, order.client_order_id)             # <<<<<<<<<<<<<<
 *         # else:
 *         #     self.set_timers()
*/
        __pyx_t_2 = __pyx_v_self->_market_info;
        __Pyx_INCREF(__pyx_t_2);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1219, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_6))) __PYX_ERR(0, 1219, __pyx_L1_error)
        __pyx_t_10 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_cancel_order(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), __pyx_t_2, ((PyObject*)__pyx_t_6)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1219, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1218
 *             for order in self.active_non_hanging_orders:
 *                 # If is about to be added to hanging_orders then don't cancel
 *                 if not self._hanging_orders_tracker.is_potential_hanging_order(order):             # <<<<<<<<<<<<<<
 *                     self.c_cancel_order(self._market_info, order.client_order_id)
 *         # else:
*/
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1216
 *         if not to_defer_canceling:
 *             self._hanging_orders_tracker.update_strategy_orders_with_equivalent_orders()
 *             for order in self.active_non_hanging_orders:             # <<<<<<<<<<<<<<
 *                 # If is about to be added to hanging_orders then don't cancel
 *                 if not self._hanging_orders_tracker.is_potential_hanging_order(order):
*/
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1214
 *                 to_defer_canceling = True
 * 
 *         if not to_defer_canceling:             # <<<<<<<<<<<<<<
 *             self._hanging_orders_tracker.update_strategy_orders_with_equivalent_orders()
 *             for order in self.active_non_hanging_orders:
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1188
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
 * 
 *     cdef c_cancel_active_orders(self, object proposal):             # <<<<<<<<<<<<<<
 *         """
 *         Cancels active non hanging orders, checks if the order prices are within tolerance threshold
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_cancel_active_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_active_orders);
  __Pyx_XDECREF(__pyx_v_active_buy_prices);
  __Pyx_XDECREF(__pyx_v_active_sells);
  __Pyx_XDECREF(__pyx_v_active_sell_prices);
  __Pyx_XDECREF(__pyx_v_proposal_buys);
  __Pyx_XDECREF(__pyx_v_proposal_sells);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_o);
  __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_o);
  __Pyx_XDECREF(__pyx_9genexpr22__pyx_v_buy);
  __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_sell);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1224
 * 
 *     # Cancel Non-Hanging, Active Orders if Spreads are below minimum_spread
 *     cdef c_cancel_orders_below_min_spread(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list active_orders = self.market_info_to_active_orders.get(self._market_info, [])
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_cancel_orders_below_min_spread(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_active_orders = 0;
  PyObject *__pyx_v_price = 0;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_v_negation = NULL;
  PyObject *__pyx_9genexpr24__pyx_v_order = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12[6];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_cancel_orders_below_min_spread", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1226
 *     cdef c_cancel_orders_below_min_spread(self):
 *         cdef:
 *             list active_orders = self.market_info_to_active_orders.get(self._market_info, [])             # <<<<<<<<<<<<<<
 *             object price = self.get_price()
 *         active_orders = [order for order in active_orders
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_market_info_to_active_orders); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_self->_market_info, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 1226, __pyx_L1_error)
  __pyx_v_active_orders = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1227
 *         cdef:
 *             list active_orders = self.market_info_to_active_orders.get(self._market_info, [])
 *             object price = self.get_price()             # <<<<<<<<<<<<<<
 *         active_orders = [order for order in active_orders
 *                          if order.client_order_id not in self.hanging_order_ids]
*/
  __pyx_t_3 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_price, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1228
 *             list active_orders = self.market_info_to_active_orders.get(self._market_info, [])
 *             object price = self.get_price()
 *         active_orders = [order for order in active_orders             # <<<<<<<<<<<<<<
 *                          if order.client_order_id not in self.hanging_order_ids]
 *         for order in active_orders:
*/
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1228, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_active_orders == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1228, __pyx_L5_error)
    }
    __pyx_t_3 = __pyx_v_active_orders; __Pyx_INCREF(__pyx_t_3);
    __pyx_t_6 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1228, __pyx_L5_error)
        #endif
        if (__pyx_t_6 >= __pyx_temp) break;
      }
      __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_6, __Pyx_ReferenceSharing_OwnStrongReference);
      ++__pyx_t_6;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1228, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_9genexpr24__pyx_v_order, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1229
 *             object price = self.get_price()
 *         active_orders = [order for order in active_orders
 *                          if order.client_order_id not in self.hanging_order_ids]             # <<<<<<<<<<<<<<
 *         for order in active_orders:
 *             negation = -1 if order.is_buy else 1
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr24__pyx_v_order, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1229, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_hanging_order_ids); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1229, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_t_4, __pyx_t_2, Py_NE)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1229, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_7) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1228
 *             list active_orders = self.market_info_to_active_orders.get(self._market_info, [])
 *             object price = self.get_price()
 *         active_orders = [order for order in active_orders             # <<<<<<<<<<<<<<
 *                          if order.client_order_id not in self.hanging_order_ids]
 *         for order in active_orders:
*/
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_9genexpr24__pyx_v_order))) __PYX_ERR(0, 1228, __pyx_L5_error)

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1229
 *             object price = self.get_price()
 *         active_orders = [order for order in active_orders
 *                          if order.client_order_id not in self.hanging_order_ids]             # <<<<<<<<<<<<<<
 *         for order in active_orders:
 *             negation = -1 if order.is_buy else 1
*/
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1228
 *             list active_orders = self.market_info_to_active_orders.get(self._market_info, [])
 *             object price = self.get_price()
 *         active_orders = [order for order in active_orders             # <<<<<<<<<<<<<<
 *                          if order.client_order_id not in self.hanging_order_ids]
 *         for order in active_orders:
*/
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_order); __pyx_9genexpr24__pyx_v_order = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_order); __pyx_9genexpr24__pyx_v_order = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __Pyx_DECREF_SET(__pyx_v_active_orders, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1230
 *         active_orders = [order for order in active_orders
 *                          if order.client_order_id not in self.hanging_order_ids]
 *         for order in active_orders:             # <<<<<<<<<<<<<<
 *             negation = -1 if order.is_buy else 1
 *             if (negation * (order.price - price) / price) < self._minimum_spread:
*/
  __pyx_t_1 = __pyx_v_active_orders; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1230, __pyx_L1_error)
      #endif
      if (__pyx_t_6 >= __pyx_temp) break;
    }
    __pyx_t_3 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_6, __Pyx_ReferenceSharing_OwnStrongReference);
    ++__pyx_t_6;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1231
 *                          if order.client_order_id not in self.hanging_order_ids]
 *         for order in active_orders:
 *             negation = -1 if order.is_buy else 1             # <<<<<<<<<<<<<<
 *             if (negation * (order.price - price) / price) < self._minimum_spread:
 *                 self.logger().info(f"Order is below minimum spread ({self._minimum_spread})."
*/
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1231, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_7) {
      __Pyx_INCREF(__pyx_mstate_global->__pyx_int_neg_1);
      __pyx_t_3 = __pyx_mstate_global->__pyx_int_neg_1;
    } else {
      __Pyx_INCREF(__pyx_mstate_global->__pyx_int_1);
      __pyx_t_3 = __pyx_mstate_global->__pyx_int_1;
    }
    __Pyx_XDECREF_SET(__pyx_v_negation, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1232
 *         for order in active_orders:
 *             negation = -1 if order.is_buy else 1
 *             if (negation * (order.price - price) / price) < self._minimum_spread:             # <<<<<<<<<<<<<<
 *                 self.logger().info(f"Order is below minimum spread ({self._minimum_spread})."
 *                                    f" Canceling Order: ({'Buy' if order.is_buy else 'Sell'}) "
*/
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Multiply(__pyx_v_negation, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_v_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_v_self->_minimum_spread, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1232, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1232, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_7) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1233
 *             negation = -1 if order.is_buy else 1
 *             if (negation * (order.price - price) / price) < self._minimum_spread:
 *                 self.logger().info(f"Order is below minimum spread ({self._minimum_spread})."             # <<<<<<<<<<<<<<
 *                                    f" Canceling Order: ({'Buy' if order.is_buy else 'Sell'}) "
 *                                    f"ID - {order.client_order_id}")
*/
      __pyx_t_8 = ((PyObject *)__pyx_v_self);
      __Pyx_INCREF(__pyx_t_8);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
        __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_logger, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1233, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __pyx_t_2 = __pyx_t_4;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_v_self->_minimum_spread, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1233, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1234
 *             if (negation * (order.price - price) / price) < self._minimum_spread:
 *                 self.logger().info(f"Order is below minimum spread ({self._minimum_spread})."
 *                                    f" Canceling Order: ({'Buy' if order.is_buy else 'Sell'}) "             # <<<<<<<<<<<<<<
 *                                    f"ID - {order.client_order_id}")
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
*/
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_is_buy); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1234, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1234, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (__pyx_t_7) {
        __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_Buy);
        __pyx_t_9 = __pyx_mstate_global->__pyx_n_u_Buy;
      } else {
        __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_Sell);
        __pyx_t_9 = __pyx_mstate_global->__pyx_n_u_Sell;
      }
      __pyx_t_10 = __Pyx_PyUnicode_Unicode(__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1234, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1235
 *                 self.logger().info(f"Order is below minimum spread ({self._minimum_spread})."
 *                                    f" Canceling Order: ({'Buy' if order.is_buy else 'Sell'}) "
 *                                    f"ID - {order.client_order_id}")             # <<<<<<<<<<<<<<
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
 * 
*/
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_11 = __Pyx_PyObject_FormatSimple(__pyx_t_9, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_12[0] = __pyx_mstate_global->__pyx_kp_u_Order_is_below_minimum_spread;
      __pyx_t_12[1] = __pyx_t_8;
      __pyx_t_12[2] = __pyx_mstate_global->__pyx_kp_u_Canceling_Order;
      __pyx_t_12[3] = __pyx_t_10;
      __pyx_t_12[4] = __pyx_mstate_global->__pyx_kp_u_ID;
      __pyx_t_12[5] = __pyx_t_11;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1233
 *             negation = -1 if order.is_buy else 1
 *             if (negation * (order.price - price) / price) < self._minimum_spread:
 *                 self.logger().info(f"Order is below minimum spread ({self._minimum_spread})."             # <<<<<<<<<<<<<<
 *                                    f" Canceling Order: ({'Buy' if order.is_buy else 'Sell'}) "
 *                                    f"ID - {order.client_order_id}")
*/
      __pyx_t_9 = __Pyx_PyUnicode_Join(__pyx_t_12, 6, 31 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8) + 21 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10) + 7 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11));
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1233, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_9};
        __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_info, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1233, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1236
 *                                    f" Canceling Order: ({'Buy' if order.is_buy else 'Sell'}) "
 *                                    f"ID - {order.client_order_id}")
 *                 self.c_cancel_order(self._market_info, order.client_order_id)             # <<<<<<<<<<<<<<
 * 
 *     cdef bint c_to_create_orders(self, object proposal):
*/
      __pyx_t_3 = __pyx_v_self->_market_info;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1236, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_4))) __PYX_ERR(0, 1236, __pyx_L1_error)
      __pyx_t_9 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_cancel_order(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), __pyx_t_3, ((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1236, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1232
 *         for order in active_orders:
 *             negation = -1 if order.is_buy else 1
 *             if (negation * (order.price - price) / price) < self._minimum_spread:             # <<<<<<<<<<<<<<
 *                 self.logger().info(f"Order is below minimum spread ({self._minimum_spread})."
 *                                    f" Canceling Order: ({'Buy' if order.is_buy else 'Sell'}) "
*/
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1230
 *         active_orders = [order for order in active_orders
 *                          if order.client_order_id not in self.hanging_order_ids]
 *         for order in active_orders:             # <<<<<<<<<<<<<<
 *             negation = -1 if order.is_buy else 1
 *             if (negation * (order.price - price) / price) < self._minimum_spread:
*/
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1224
 * 
 *     # Cancel Non-Hanging, Active Orders if Spreads are below minimum_spread
 *     cdef c_cancel_orders_below_min_spread(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list active_orders = self.market_info_to_active_orders.get(self._market_info, [])
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_cancel_orders_below_min_spread", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_active_orders);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XDECREF(__pyx_v_negation);
  __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_order);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1238
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
 * 
 *     cdef bint c_to_create_orders(self, object proposal):             # <<<<<<<<<<<<<<
 *         non_hanging_orders_non_cancelled = [o for o in self.active_non_hanging_orders if not
 *                                             self._hanging_orders_tracker.is_potential_hanging_order(o)]
*/

static int __pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_to_create_orders(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_v_non_hanging_orders_non_cancelled = NULL;
  PyObject *__pyx_9genexpr25__pyx_v_o = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_to_create_orders", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1239
 * 
 *     cdef bint c_to_create_orders(self, object proposal):
 *         non_hanging_orders_non_cancelled = [o for o in self.active_non_hanging_orders if not             # <<<<<<<<<<<<<<
 *                                             self._hanging_orders_tracker.is_potential_hanging_order(o)]
 *         return (self._create_timestamp < self._current_timestamp
*/
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1239, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_non_hanging_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1239, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1239, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1239, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1239, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_3, __pyx_t_4, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_4;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1239, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4));
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4);
          #endif
          ++__pyx_t_4;
        }
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1239, __pyx_L5_error)
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1239, __pyx_L5_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_9genexpr25__pyx_v_o, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1240
 *     cdef bint c_to_create_orders(self, object proposal):
 *         non_hanging_orders_non_cancelled = [o for o in self.active_non_hanging_orders if not
 *                                             self._hanging_orders_tracker.is_potential_hanging_order(o)]             # <<<<<<<<<<<<<<
 *         return (self._create_timestamp < self._current_timestamp
 *                 and (not self._should_wait_order_cancel_confirmation or
*/
      __pyx_t_6 = __pyx_v_self->_hanging_orders_tracker;
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_7 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_9genexpr25__pyx_v_o};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_is_potential_hanging_order, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1240, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1240, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1239
 * 
 *     cdef bint c_to_create_orders(self, object proposal):
 *         non_hanging_orders_non_cancelled = [o for o in self.active_non_hanging_orders if not             # <<<<<<<<<<<<<<
 *                                             self._hanging_orders_tracker.is_potential_hanging_order(o)]
 *         return (self._create_timestamp < self._current_timestamp
*/
      __pyx_t_9 = (!__pyx_t_8);
      if (__pyx_t_9) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_9genexpr25__pyx_v_o))) __PYX_ERR(0, 1239, __pyx_L5_error)
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_o); __pyx_9genexpr25__pyx_v_o = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_o); __pyx_9genexpr25__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_v_non_hanging_orders_non_cancelled = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1241
 *         non_hanging_orders_non_cancelled = [o for o in self.active_non_hanging_orders if not
 *                                             self._hanging_orders_tracker.is_potential_hanging_order(o)]
 *         return (self._create_timestamp < self._current_timestamp             # <<<<<<<<<<<<<<
 *                 and (not self._should_wait_order_cancel_confirmation or
 *                      len(self._sb_order_tracker.in_flight_cancels) == 0)
*/
  __pyx_t_8 = (__pyx_v_self->_create_timestamp < __pyx_v_self->__pyx_base.__pyx_base._current_timestamp);
  if (__pyx_t_8) {
  } else {
    __pyx_t_9 = __pyx_t_8;
    goto __pyx_L11_bool_binop_done;
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1242
 *                                             self._hanging_orders_tracker.is_potential_hanging_order(o)]
 *         return (self._create_timestamp < self._current_timestamp
 *                 and (not self._should_wait_order_cancel_confirmation or             # <<<<<<<<<<<<<<
 *                      len(self._sb_order_tracker.in_flight_cancels) == 0)
 *                 and proposal is not None
*/
  __pyx_t_8 = (!__pyx_v_self->_should_wait_order_cancel_confirmation);
  if (!__pyx_t_8) {
  } else {
    goto __pyx_L13_next_and;
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1243
 *         return (self._create_timestamp < self._current_timestamp
 *                 and (not self._should_wait_order_cancel_confirmation or
 *                      len(self._sb_order_tracker.in_flight_cancels) == 0)             # <<<<<<<<<<<<<<
 *                 and proposal is not None
 *                 and len(non_hanging_orders_non_cancelled) == 0)
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base._sb_order_tracker), __pyx_mstate_global->__pyx_n_u_in_flight_cancels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1243, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = (__pyx_t_4 == 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_9 = __pyx_t_8;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_L13_next_and:;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1244
 *                 and (not self._should_wait_order_cancel_confirmation or
 *                      len(self._sb_order_tracker.in_flight_cancels) == 0)
 *                 and proposal is not None             # <<<<<<<<<<<<<<
 *                 and len(non_hanging_orders_non_cancelled) == 0)
 * 
*/
  __pyx_t_8 = (__pyx_v_proposal != Py_None);
  if (__pyx_t_8) {
  } else {
    __pyx_t_9 = __pyx_t_8;
    goto __pyx_L11_bool_binop_done;
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1245
 *                      len(self._sb_order_tracker.in_flight_cancels) == 0)
 *                 and proposal is not None
 *                 and len(non_hanging_orders_non_cancelled) == 0)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_execute_orders_proposal(self, object proposal):
*/
  __pyx_t_4 = __Pyx_PyList_GET_SIZE(__pyx_v_non_hanging_orders_non_cancelled); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1245, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_4 == 0);
  __pyx_t_9 = __pyx_t_8;
  __pyx_L11_bool_binop_done:;
  __pyx_r = __pyx_t_9;
  goto __pyx_L0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1238
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
 * 
 *     cdef bint c_to_create_orders(self, object proposal):             # <<<<<<<<<<<<<<
 *         non_hanging_orders_non_cancelled = [o for o in self.active_non_hanging_orders if not
 *                                             self._hanging_orders_tracker.is_potential_hanging_order(o)]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_to_create_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_non_hanging_orders_non_cancelled);
  __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_o);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1274
 *                 orders_created = True
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))             # <<<<<<<<<<<<<<
 *                     if order:
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
*/

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1274, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_2generator2, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5]), (PyObject *) __pyx_cur_scope, __pyx_mstate_global->__pyx_n_u_genexpr, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_c_execu, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market); if (unlikely(!gen)) __PYX_ERR(0, 1274, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_execute_orders_proposal.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr *__pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(__pyx_sent_value != Py_None)) {
    if (unlikely(__pyx_sent_value)) PyErr_SetString(PyExc_TypeError, "can't send non-None value to a just-started generator");
    __PYX_ERR(0, 1274, __pyx_L1_error)
  }
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 1274, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1274, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1274, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_2, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_2;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1274, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2));
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2);
        #endif
        ++__pyx_t_2;
      }
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1274, __pyx_L1_error)
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1274, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_o);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_o, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_o, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_bid_order_id)) { __Pyx_RaiseClosureNameError("bid_order_id"); __PYX_ERR(0, 1274, __pyx_L1_error) }
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_bid_order_id, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1274, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_o);
      __pyx_r = __pyx_cur_scope->__pyx_v_o;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1274, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  if (__Pyx_PyErr_Occurred()) {
    __Pyx_Generator_Replace_StopIteration(0);
    __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  }
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_5generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1297
 *                 orders_created = True
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))             # <<<<<<<<<<<<<<
 *                     if order:
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
*/

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_3genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1297, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_5generator3, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6]), (PyObject *) __pyx_cur_scope, __pyx_mstate_global->__pyx_n_u_genexpr, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_c_execu, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market); if (unlikely(!gen)) __PYX_ERR(0, 1297, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_execute_orders_proposal.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_5generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr *__pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(__pyx_sent_value != Py_None)) {
    if (unlikely(__pyx_sent_value)) PyErr_SetString(PyExc_TypeError, "can't send non-None value to a just-started generator");
    __PYX_ERR(0, 1297, __pyx_L1_error)
  }
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 1297, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1297, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1297, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_2, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_2;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1297, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2));
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2);
        #endif
        ++__pyx_t_2;
      }
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1297, __pyx_L1_error)
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1297, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_o);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_o, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_o, __pyx_mstate_global->__pyx_n_u_client_order_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ask_order_id)) { __Pyx_RaiseClosureNameError("ask_order_id"); __PYX_ERR(0, 1297, __pyx_L1_error) }
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_ask_order_id, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1297, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_o);
      __pyx_r = __pyx_cur_scope->__pyx_v_o;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1297, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  if (__Pyx_PyErr_Occurred()) {
    __Pyx_Generator_Replace_StopIteration(0);
    __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  }
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1247
 *                 and len(non_hanging_orders_non_cancelled) == 0)
 * 
 *     cdef c_execute_orders_proposal(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double expiration_seconds = NaN
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_execute_orders_proposal(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal *__pyx_cur_scope;
  double __pyx_v_expiration_seconds;
  int __pyx_v_orders_created;
  Py_ssize_t __pyx_v_number_of_pairs;
  PyObject *__pyx_v_price_quote_str = NULL;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_buy = NULL;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_v_sell = NULL;
  PyObject *__pyx_9genexpr26__pyx_v_buy = NULL;
  PyObject *__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_2generator2 = 0;
  PyObject *__pyx_9genexpr28__pyx_v_sell = NULL;
  PyObject *__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_5generator3 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15[7];
  size_t __pyx_t_16;
  PyObject *__pyx_t_17[6];
  struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_buy_with_specific_market __pyx_t_18;
  struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_sell_with_specific_market __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_execute_orders_proposal", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal, __pyx_mstate_global->__pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1247, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1249
 *     cdef c_execute_orders_proposal(self, object proposal):
 *         cdef:
 *             double expiration_seconds = NaN             # <<<<<<<<<<<<<<
 *             str bid_order_id, ask_order_id
 *             bint orders_created = False
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_NaN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_expiration_seconds = __pyx_t_2;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1251
 *             double expiration_seconds = NaN
 *             str bid_order_id, ask_order_id
 *             bint orders_created = False             # <<<<<<<<<<<<<<
 *         # Number of pair of orders to track for hanging orders
 *         number_of_pairs = min((len(proposal.buys), len(proposal.sells))) if self._hanging_orders_enabled else 0
*/
  __pyx_v_orders_created = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1253
 *             bint orders_created = False
 *         # Number of pair of orders to track for hanging orders
 *         number_of_pairs = min((len(proposal.buys), len(proposal.sells))) if self._hanging_orders_enabled else 0             # <<<<<<<<<<<<<<
 * 
 *         if len(proposal.buys) > 0:
*/
  if (__pyx_v_self->_hanging_orders_enabled) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1253, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1253, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = (__pyx_t_4 < __pyx_t_5);
    if (__pyx_t_7) {
      __pyx_t_6 = __pyx_t_4;
    } else {
      __pyx_t_6 = __pyx_t_5;
    }
    __pyx_t_3 = __pyx_t_6;
  } else {
    __pyx_t_3 = 0;
  }
  __pyx_v_number_of_pairs = __pyx_t_3;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1255
 *         number_of_pairs = min((len(proposal.buys), len(proposal.sells))) if self._hanging_orders_enabled else 0
 * 
 *         if len(proposal.buys) > 0:             # <<<<<<<<<<<<<<
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (__pyx_t_3 > 0);
  if (__pyx_t_7) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1256
 * 
 *         if len(proposal.buys) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:             # <<<<<<<<<<<<<<
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
*/
    __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->_logging_options); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_OPTION_LOG_CREATE_ORDER); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyNumber_And(__pyx_t_1, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1256, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (__pyx_t_7) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1257
 *         if len(proposal.buys) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "             # <<<<<<<<<<<<<<
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 *                                    for buy in proposal.buys]
*/
      { /* enter inner scope */
        __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1257, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_9);

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1259
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 *                                    for buy in proposal.buys]             # <<<<<<<<<<<<<<
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.buys)} bid orders "
*/
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1259, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
          __pyx_t_1 = __pyx_t_8; __Pyx_INCREF(__pyx_t_1);
          __pyx_t_3 = 0;
          __pyx_t_10 = NULL;
        } else {
          __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1259, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_10 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1259, __pyx_L7_error)
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        for (;;) {
          if (likely(!__pyx_t_10)) {
            if (likely(PyList_CheckExact(__pyx_t_1))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
                #if !CYTHON_ASSUME_SAFE_SIZE
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1259, __pyx_L7_error)
                #endif
                if (__pyx_t_3 >= __pyx_temp) break;
              }
              __pyx_t_8 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
              ++__pyx_t_3;
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
                #if !CYTHON_ASSUME_SAFE_SIZE
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1259, __pyx_L7_error)
                #endif
                if (__pyx_t_3 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_8 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3));
              #else
              __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3);
              #endif
              ++__pyx_t_3;
            }
            if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1259, __pyx_L7_error)
          } else {
            __pyx_t_8 = __pyx_t_10(__pyx_t_1);
            if (unlikely(!__pyx_t_8)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1259, __pyx_L7_error)
                PyErr_Clear();
              }
              break;
            }
          }
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_XDECREF_SET(__pyx_9genexpr26__pyx_v_buy, __pyx_t_8);
          __pyx_t_8 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1257
 *         if len(proposal.buys) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "             # <<<<<<<<<<<<<<
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 *                                    for buy in proposal.buys]
*/
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr26__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1257, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_normalize); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1257, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = __Pyx_PyObject_CallNoArg(__pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1257, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1257, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_base_asset); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1257, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_12 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1257, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1258
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
 *                                    f"{buy.price.normalize()} {self.quote_asset}"             # <<<<<<<<<<<<<<
 *                                    for buy in proposal.buys]
 *                 self.logger().info(
*/
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr26__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1258, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_normalize); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1258, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = __Pyx_PyObject_CallNoArg(__pyx_t_13); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1258, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_13 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1258, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_quote_asset); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1258, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_14 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1258, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_15[0] = __pyx_t_11;
          __pyx_t_15[1] = __pyx_mstate_global->__pyx_kp_u__21;
          __pyx_t_15[2] = __pyx_t_12;
          __pyx_t_15[3] = __pyx_mstate_global->__pyx_kp_u__24;
          __pyx_t_15[4] = __pyx_t_13;
          __pyx_t_15[5] = __pyx_mstate_global->__pyx_kp_u__21;
          __pyx_t_15[6] = __pyx_t_14;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1257
 *         if len(proposal.buys) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "             # <<<<<<<<<<<<<<
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 *                                    for buy in proposal.buys]
*/
          __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_15, 7, __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11) + 1 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_14), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14));
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1257, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_9, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 1257, __pyx_L7_error)
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1259
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 *                                    for buy in proposal.buys]             # <<<<<<<<<<<<<<
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.buys)} bid orders "
*/
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_9genexpr26__pyx_v_buy); __pyx_9genexpr26__pyx_v_buy = 0;
        goto __pyx_L11_exit_scope;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_9genexpr26__pyx_v_buy); __pyx_9genexpr26__pyx_v_buy = 0;
        goto __pyx_L1_error;
        __pyx_L11_exit_scope:;
      } /* exit inner scope */
      __pyx_v_price_quote_str = ((PyObject*)__pyx_t_9);
      __pyx_t_9 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1260
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 *                                    for buy in proposal.buys]
 *                 self.logger().info(             # <<<<<<<<<<<<<<
 *                     f"({self.trading_pair}) Creating {len(proposal.buys)} bid orders "
 *                     f"at (Size, Price): {price_quote_str}"
*/
      __pyx_t_14 = ((PyObject *)__pyx_v_self);
      __Pyx_INCREF(__pyx_t_14);
      __pyx_t_16 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_14, NULL};
        __pyx_t_8 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_logger, __pyx_callargs+__pyx_t_16, (1-__pyx_t_16) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1260, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
      }
      __pyx_t_1 = __pyx_t_8;
      __Pyx_INCREF(__pyx_t_1);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1261
 *                                    for buy in proposal.buys]
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.buys)} bid orders "             # <<<<<<<<<<<<<<
 *                     f"at (Size, Price): {price_quote_str}"
 *                 )
*/
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_13 = __Pyx_PyObject_FormatSimple(__pyx_t_14, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_3 = PyObject_Length(__pyx_t_14); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1261, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_14 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_3, 0, ' ', 'd'); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1262
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.buys)} bid orders "
 *                     f"at (Size, Price): {price_quote_str}"             # <<<<<<<<<<<<<<
 *                 )
 *             for idx, buy in enumerate(proposal.buys):
*/
      __pyx_t_12 = __Pyx_PyObject_FormatSimple(__pyx_v_price_quote_str, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1262, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_17[0] = __pyx_mstate_global->__pyx_kp_u__20;
      __pyx_t_17[1] = __pyx_t_13;
      __pyx_t_17[2] = __pyx_mstate_global->__pyx_kp_u_Creating;
      __pyx_t_17[3] = __pyx_t_14;
      __pyx_t_17[4] = __pyx_mstate_global->__pyx_kp_u_bid_orders_at_Size_Price;
      __pyx_t_17[5] = __pyx_t_12;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1261
 *                                    for buy in proposal.buys]
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.buys)} bid orders "             # <<<<<<<<<<<<<<
 *                     f"at (Size, Price): {price_quote_str}"
 *                 )
*/
      __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_17, 6, 1 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13) + 11 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_14) + 30 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12));
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_16 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_t_11};
        __pyx_t_9 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_info, __pyx_callargs+__pyx_t_16, (2-__pyx_t_16) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1260, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1256
 * 
 *         if len(proposal.buys) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:             # <<<<<<<<<<<<<<
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
*/
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1264
 *                     f"at (Size, Price): {price_quote_str}"
 *                 )
 *             for idx, buy in enumerate(proposal.buys):             # <<<<<<<<<<<<<<
 *                 bid_order_id = self.c_buy_with_specific_market(
 *                     self._market_info,
*/
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_0);
    __pyx_t_9 = __pyx_mstate_global->__pyx_int_0;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_buys); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
      __pyx_t_11 = __pyx_t_8; __Pyx_INCREF(__pyx_t_11);
      __pyx_t_3 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_10 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1264, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_11))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_11);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1264, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          __pyx_t_8 = __Pyx_PyList_GetItemRefFast(__pyx_t_11, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_3;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_11);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1264, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_3));
          #else
          __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_11, __pyx_t_3);
          #endif
          ++__pyx_t_3;
        }
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1264, __pyx_L1_error)
      } else {
        __pyx_t_8 = __pyx_t_10(__pyx_t_11);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1264, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_buy, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_9);
      __pyx_t_8 = __Pyx_PyLong_AddObjC(__pyx_t_9, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_9);
      __pyx_t_9 = __pyx_t_8;
      __pyx_t_8 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1266
 *             for idx, buy in enumerate(proposal.buys):
 *                 bid_order_id = self.c_buy_with_specific_market(
 *                     self._market_info,             # <<<<<<<<<<<<<<
 *                     buy.size,
 *                     order_type=self._limit_order_type,
*/
      __pyx_t_8 = __pyx_v_self->_market_info;
      __Pyx_INCREF(__pyx_t_8);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1267
 *                 bid_order_id = self.c_buy_with_specific_market(
 *                     self._market_info,
 *                     buy.size,             # <<<<<<<<<<<<<<
 *                     order_type=self._limit_order_type,
 *                     price=buy.price,
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1268
 *                     self._market_info,
 *                     buy.size,
 *                     order_type=self._limit_order_type,             # <<<<<<<<<<<<<<
 *                     price=buy.price,
 *                     expiration_seconds=expiration_seconds
*/
      __pyx_t_12 = __pyx_v_self->_limit_order_type;
      __Pyx_INCREF(__pyx_t_12);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1269
 *                     buy.size,
 *                     order_type=self._limit_order_type,
 *                     price=buy.price,             # <<<<<<<<<<<<<<
 *                     expiration_seconds=expiration_seconds
 *                 )
*/
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1269, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1265
 *                 )
 *             for idx, buy in enumerate(proposal.buys):
 *                 bid_order_id = self.c_buy_with_specific_market(             # <<<<<<<<<<<<<<
 *                     self._market_info,
 *                     buy.size,
*/
      __pyx_t_18.__pyx_n = 3;
      __pyx_t_18.order_type = __pyx_t_12;
      __pyx_t_18.price = __pyx_t_14;
      __pyx_t_18.expiration_seconds = __pyx_v_expiration_seconds;
      __pyx_t_13 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_buy_with_specific_market(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), __pyx_t_8, __pyx_t_1, &__pyx_t_18); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1265, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_bid_order_id);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_bid_order_id, ((PyObject*)__pyx_t_13));
      __Pyx_GIVEREF(__pyx_t_13);
      __pyx_t_13 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1272
 *                     expiration_seconds=expiration_seconds
 *                 )
 *                 orders_created = True             # <<<<<<<<<<<<<<
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))
*/
      __pyx_v_orders_created = 1;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1273
 *                 )
 *                 orders_created = True
 *                 if idx < number_of_pairs:             # <<<<<<<<<<<<<<
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))
 *                     if order:
*/
      __pyx_t_13 = PyLong_FromSsize_t(__pyx_v_number_of_pairs); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_14 = PyObject_RichCompare(__pyx_v_idx, __pyx_t_13, Py_LT); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (__pyx_t_7) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1274
 *                 orders_created = True
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))             # <<<<<<<<<<<<<<
 *                     if order:
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
*/
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_orders); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1274, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_13 = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_genexpr(((PyObject*)__pyx_cur_scope), __pyx_t_14); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1274, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_14 = __Pyx_PyIter_Next(__pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1274, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_14);
        __pyx_t_14 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1275
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))
 *                     if order:             # <<<<<<<<<<<<<<
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
 *                             CreatedPairOfOrders(order, None))
*/
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_order); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1275, __pyx_L1_error)
        if (__pyx_t_7) {

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1276
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))
 *                     if order:
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(             # <<<<<<<<<<<<<<
 *                             CreatedPairOfOrders(order, None))
 *         if len(proposal.sells) > 0:
*/
          __pyx_t_13 = __pyx_v_self->_hanging_orders_tracker;
          __Pyx_INCREF(__pyx_t_13);

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1277
 *                     if order:
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
 *                             CreatedPairOfOrders(order, None))             # <<<<<<<<<<<<<<
 *         if len(proposal.sells) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
*/
          __pyx_t_1 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_CreatedPairOfOrders); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1277, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_16 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
            assert(__pyx_t_1);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
            __pyx_t_16 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_order, Py_None};
            __pyx_t_12 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_16, (3-__pyx_t_16) | (__pyx_t_16*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1277, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
          }
          __pyx_t_16 = 0;
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_t_12};
            __pyx_t_14 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add_current_pairs_of_proposal_or, __pyx_callargs+__pyx_t_16, (2-__pyx_t_16) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1276, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
          }
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1275
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))
 *                     if order:             # <<<<<<<<<<<<<<
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
 *                             CreatedPairOfOrders(order, None))
*/
        }

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1273
 *                 )
 *                 orders_created = True
 *                 if idx < number_of_pairs:             # <<<<<<<<<<<<<<
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))
 *                     if order:
*/
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1264
 *                     f"at (Size, Price): {price_quote_str}"
 *                 )
 *             for idx, buy in enumerate(proposal.buys):             # <<<<<<<<<<<<<<
 *                 bid_order_id = self.c_buy_with_specific_market(
 *                     self._market_info,
*/
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1255
 *         number_of_pairs = min((len(proposal.buys), len(proposal.sells))) if self._hanging_orders_enabled else 0
 * 
 *         if len(proposal.buys) > 0:             # <<<<<<<<<<<<<<
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1278
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
 *                             CreatedPairOfOrders(order, None))
 *         if len(proposal.sells) > 0:             # <<<<<<<<<<<<<<
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
*/
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_3 = PyObject_Length(__pyx_t_9); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_7 = (__pyx_t_3 > 0);
  if (__pyx_t_7) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1279
 *                             CreatedPairOfOrders(order, None))
 *         if len(proposal.sells) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:             # <<<<<<<<<<<<<<
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
*/
    __pyx_t_9 = __Pyx_PyLong_From_int64_t(__pyx_v_self->_logging_options); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_OPTION_LOG_CREATE_ORDER); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_14 = PyNumber_And(__pyx_t_9, __pyx_t_11); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (__pyx_t_7) {

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1280
 *         if len(proposal.sells) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "             # <<<<<<<<<<<<<<
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 *                                    for sell in proposal.sells]
*/
      { /* enter inner scope */
        __pyx_t_14 = PyList_New(0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1280, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_14);

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1282
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 *                                    for sell in proposal.sells]             # <<<<<<<<<<<<<<
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.sells)} ask "
*/
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1282, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (likely(PyList_CheckExact(__pyx_t_11)) || PyTuple_CheckExact(__pyx_t_11)) {
          __pyx_t_9 = __pyx_t_11; __Pyx_INCREF(__pyx_t_9);
          __pyx_t_3 = 0;
          __pyx_t_10 = NULL;
        } else {
          __pyx_t_3 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_t_11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1282, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_10 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1282, __pyx_L21_error)
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        for (;;) {
          if (likely(!__pyx_t_10)) {
            if (likely(PyList_CheckExact(__pyx_t_9))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_9);
                #if !CYTHON_ASSUME_SAFE_SIZE
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1282, __pyx_L21_error)
                #endif
                if (__pyx_t_3 >= __pyx_temp) break;
              }
              __pyx_t_11 = __Pyx_PyList_GetItemRefFast(__pyx_t_9, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
              ++__pyx_t_3;
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_9);
                #if !CYTHON_ASSUME_SAFE_SIZE
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1282, __pyx_L21_error)
                #endif
                if (__pyx_t_3 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_11 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_3));
              #else
              __pyx_t_11 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_3);
              #endif
              ++__pyx_t_3;
            }
            if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1282, __pyx_L21_error)
          } else {
            __pyx_t_11 = __pyx_t_10(__pyx_t_9);
            if (unlikely(!__pyx_t_11)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1282, __pyx_L21_error)
                PyErr_Clear();
              }
              break;
            }
          }
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_XDECREF_SET(__pyx_9genexpr28__pyx_v_sell, __pyx_t_11);
          __pyx_t_11 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1280
 *         if len(proposal.sells) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "             # <<<<<<<<<<<<<<
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 *                                    for sell in proposal.sells]
*/
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr28__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1280, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_normalize); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1280, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyObject_CallNoArg(__pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1280, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_12 = __Pyx_PyObject_FormatSimple(__pyx_t_11, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1280, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_base_asset); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1280, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_13 = __Pyx_PyObject_FormatSimple(__pyx_t_11, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1280, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1281
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
 *                                    f"{sell.price.normalize()} {self.quote_asset}"             # <<<<<<<<<<<<<<
 *                                    for sell in proposal.sells]
 *                 self.logger().info(
*/
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr28__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1281, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_normalize); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1281, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyObject_CallNoArg(__pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1281, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_11, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1281, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_quote_asset); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1281, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_11, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1281, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_15[0] = __pyx_t_12;
          __pyx_t_15[1] = __pyx_mstate_global->__pyx_kp_u__21;
          __pyx_t_15[2] = __pyx_t_13;
          __pyx_t_15[3] = __pyx_mstate_global->__pyx_kp_u__24;
          __pyx_t_15[4] = __pyx_t_8;
          __pyx_t_15[5] = __pyx_mstate_global->__pyx_kp_u__21;
          __pyx_t_15[6] = __pyx_t_1;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1280
 *         if len(proposal.sells) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "             # <<<<<<<<<<<<<<
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 *                                    for sell in proposal.sells]
*/
          __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_15, 7, __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12) + 1 * 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13) + 2 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8) + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1));
          if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1280, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_14, (PyObject*)__pyx_t_11))) __PYX_ERR(0, 1280, __pyx_L21_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1282
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 *                                    for sell in proposal.sells]             # <<<<<<<<<<<<<<
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.sells)} ask "
*/
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_9genexpr28__pyx_v_sell); __pyx_9genexpr28__pyx_v_sell = 0;
        goto __pyx_L25_exit_scope;
        __pyx_L21_error:;
        __Pyx_XDECREF(__pyx_9genexpr28__pyx_v_sell); __pyx_9genexpr28__pyx_v_sell = 0;
        goto __pyx_L1_error;
        __pyx_L25_exit_scope:;
      } /* exit inner scope */
      __Pyx_XDECREF_SET(__pyx_v_price_quote_str, ((PyObject*)__pyx_t_14));
      __pyx_t_14 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1283
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 *                                    for sell in proposal.sells]
 *                 self.logger().info(             # <<<<<<<<<<<<<<
 *                     f"({self.trading_pair}) Creating {len(proposal.sells)} ask "
 *                     f"orders at (Size, Price): {price_quote_str}"
*/
      __pyx_t_1 = ((PyObject *)__pyx_v_self);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_16 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
        __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_logger, __pyx_callargs+__pyx_t_16, (1-__pyx_t_16) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1283, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
      }
      __pyx_t_9 = __pyx_t_11;
      __Pyx_INCREF(__pyx_t_9);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1284
 *                                    for sell in proposal.sells]
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.sells)} ask "             # <<<<<<<<<<<<<<
 *                     f"orders at (Size, Price): {price_quote_str}"
 *                 )
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1284, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_3, 0, ' ', 'd'); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1285
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.sells)} ask "
 *                     f"orders at (Size, Price): {price_quote_str}"             # <<<<<<<<<<<<<<
 *                 )
 *             for idx, sell in enumerate(proposal.sells):
*/
      __pyx_t_13 = __Pyx_PyObject_FormatSimple(__pyx_v_price_quote_str, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_17[0] = __pyx_mstate_global->__pyx_kp_u__20;
      __pyx_t_17[1] = __pyx_t_8;
      __pyx_t_17[2] = __pyx_mstate_global->__pyx_kp_u_Creating;
      __pyx_t_17[3] = __pyx_t_1;
      __pyx_t_17[4] = __pyx_mstate_global->__pyx_kp_u_ask_orders_at_Size_Price;
      __pyx_t_17[5] = __pyx_t_13;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1284
 *                                    for sell in proposal.sells]
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.sells)} ask "             # <<<<<<<<<<<<<<
 *                     f"orders at (Size, Price): {price_quote_str}"
 *                 )
*/
      __pyx_t_12 = __Pyx_PyUnicode_Join(__pyx_t_17, 6, 1 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8) + 11 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1) + 30 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13), 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13));
      if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_16 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_12};
        __pyx_t_14 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_info, __pyx_callargs+__pyx_t_16, (2-__pyx_t_16) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1283, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
      }
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1279
 *                             CreatedPairOfOrders(order, None))
 *         if len(proposal.sells) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:             # <<<<<<<<<<<<<<
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
*/
    }

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1287
 *                     f"orders at (Size, Price): {price_quote_str}"
 *                 )
 *             for idx, sell in enumerate(proposal.sells):             # <<<<<<<<<<<<<<
 *                 ask_order_id = self.c_sell_with_specific_market(
 *                     self._market_info,
*/
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_0);
    __pyx_t_14 = __pyx_mstate_global->__pyx_int_0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_mstate_global->__pyx_n_u_sells); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (likely(PyList_CheckExact(__pyx_t_11)) || PyTuple_CheckExact(__pyx_t_11)) {
      __pyx_t_12 = __pyx_t_11; __Pyx_INCREF(__pyx_t_12);
      __pyx_t_3 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_12 = PyObject_GetIter(__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_10 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1287, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_12))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_12);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1287, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          __pyx_t_11 = __Pyx_PyList_GetItemRefFast(__pyx_t_12, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
          ++__pyx_t_3;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_12);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1287, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_11 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_12, __pyx_t_3));
          #else
          __pyx_t_11 = __Pyx_PySequence_ITEM(__pyx_t_12, __pyx_t_3);
          #endif
          ++__pyx_t_3;
        }
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1287, __pyx_L1_error)
      } else {
        __pyx_t_11 = __pyx_t_10(__pyx_t_12);
        if (unlikely(!__pyx_t_11)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1287, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_XDECREF_SET(__pyx_v_sell, __pyx_t_11);
      __pyx_t_11 = 0;
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_14);
      __pyx_t_11 = __Pyx_PyLong_AddObjC(__pyx_t_14, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_14);
      __pyx_t_14 = __pyx_t_11;
      __pyx_t_11 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1289
 *             for idx, sell in enumerate(proposal.sells):
 *                 ask_order_id = self.c_sell_with_specific_market(
 *                     self._market_info,             # <<<<<<<<<<<<<<
 *                     sell.size,
 *                     order_type=self._limit_order_type,
*/
      __pyx_t_11 = __pyx_v_self->_market_info;
      __Pyx_INCREF(__pyx_t_11);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1290
 *                 ask_order_id = self.c_sell_with_specific_market(
 *                     self._market_info,
 *                     sell.size,             # <<<<<<<<<<<<<<
 *                     order_type=self._limit_order_type,
 *                     price=sell.price,
*/
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_size); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1290, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1291
 *                     self._market_info,
 *                     sell.size,
 *                     order_type=self._limit_order_type,             # <<<<<<<<<<<<<<
 *                     price=sell.price,
 *                     expiration_seconds=expiration_seconds
*/
      __pyx_t_13 = __pyx_v_self->_limit_order_type;
      __Pyx_INCREF(__pyx_t_13);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1292
 *                     sell.size,
 *                     order_type=self._limit_order_type,
 *                     price=sell.price,             # <<<<<<<<<<<<<<
 *                     expiration_seconds=expiration_seconds
 *                 )
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_mstate_global->__pyx_n_u_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1288
 *                 )
 *             for idx, sell in enumerate(proposal.sells):
 *                 ask_order_id = self.c_sell_with_specific_market(             # <<<<<<<<<<<<<<
 *                     self._market_info,
 *                     sell.size,
*/
      __pyx_t_19.__pyx_n = 3;
      __pyx_t_19.order_type = __pyx_t_13;
      __pyx_t_19.price = __pyx_t_1;
      __pyx_t_19.expiration_seconds = __pyx_v_expiration_seconds;
      __pyx_t_8 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_sell_with_specific_market(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), __pyx_t_11, __pyx_t_9, &__pyx_t_19); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_ask_order_id);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_ask_order_id, ((PyObject*)__pyx_t_8));
      __Pyx_GIVEREF(__pyx_t_8);
      __pyx_t_8 = 0;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1295
 *                     expiration_seconds=expiration_seconds
 *                 )
 *                 orders_created = True             # <<<<<<<<<<<<<<
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))
*/
      __pyx_v_orders_created = 1;

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1296
 *                 )
 *                 orders_created = True
 *                 if idx < number_of_pairs:             # <<<<<<<<<<<<<<
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))
 *                     if order:
*/
      __pyx_t_8 = PyLong_FromSsize_t(__pyx_v_number_of_pairs); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1296, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_idx, __pyx_t_8, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1296, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1296, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_7) {

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1297
 *                 orders_created = True
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))             # <<<<<<<<<<<<<<
 *                     if order:
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
*/
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1297, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_3genexpr(((PyObject*)__pyx_cur_scope), __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1297, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyIter_Next(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1297, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1298
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))
 *                     if order:             # <<<<<<<<<<<<<<
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
 *         if orders_created:
*/
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_order); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1298, __pyx_L1_error)
        if (__pyx_t_7) {

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1299
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))
 *                     if order:
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order             # <<<<<<<<<<<<<<
 *         if orders_created:
 *             self.set_timers()
*/
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_mstate_global->__pyx_n_u_current_created_pairs_of_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1299, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_idx); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1299, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_sell_order, __pyx_v_order) < (0)) __PYX_ERR(0, 1299, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1298
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))
 *                     if order:             # <<<<<<<<<<<<<<
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
 *         if orders_created:
*/
        }

        /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1296
 *                 )
 *                 orders_created = True
 *                 if idx < number_of_pairs:             # <<<<<<<<<<<<<<
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))
 *                     if order:
*/
      }

      /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1287
 *                     f"orders at (Size, Price): {price_quote_str}"
 *                 )
 *             for idx, sell in enumerate(proposal.sells):             # <<<<<<<<<<<<<<
 *                 ask_order_id = self.c_sell_with_specific_market(
 *                     self._market_info,
*/
    }
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1278
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
 *                             CreatedPairOfOrders(order, None))
 *         if len(proposal.sells) > 0:             # <<<<<<<<<<<<<<
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1300
 *                     if order:
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
 *         if orders_created:             # <<<<<<<<<<<<<<
 *             self.set_timers()
 * 
*/
  if (__pyx_v_orders_created) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1301
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
 *         if orders_created:
 *             self.set_timers()             # <<<<<<<<<<<<<<
 * 
 *     cdef set_timers(self):
*/
    __pyx_t_14 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->set_timers(__pyx_v_self); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1300
 *                     if order:
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
 *         if orders_created:             # <<<<<<<<<<<<<<
 *             self.set_timers()
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1247
 *                 and len(non_hanging_orders_non_cancelled) == 0)
 * 
 *     cdef c_execute_orders_proposal(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double expiration_seconds = NaN
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.c_execute_orders_proposal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_price_quote_str);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_buy);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XDECREF(__pyx_v_sell);
  __Pyx_XDECREF(__pyx_9genexpr26__pyx_v_buy);
  __Pyx_XDECREF(__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_2generator2);
  __Pyx_XDECREF(__pyx_9genexpr28__pyx_v_sell);
  __Pyx_XDECREF(__pyx_gb_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25c_execute_orders_proposal_5generator3);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1303
 *             self.set_timers()
 * 
 *     cdef set_timers(self):             # <<<<<<<<<<<<<<
 *         cdef double next_cycle = self._current_timestamp + self._order_refresh_time
 *         if self._create_timestamp <= self._current_timestamp:
*/

static PyObject *__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_set_timers(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  double __pyx_v_next_cycle;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  double __pyx_t_2;
  double __pyx_t_3;
  double __pyx_t_4;
  __Pyx_RefNannySetupContext("set_timers", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1304
 * 
 *     cdef set_timers(self):
 *         cdef double next_cycle = self._current_timestamp + self._order_refresh_time             # <<<<<<<<<<<<<<
 *         if self._create_timestamp <= self._current_timestamp:
 *             self._create_timestamp = next_cycle
*/
  __pyx_v_next_cycle = (__pyx_v_self->__pyx_base.__pyx_base._current_timestamp + __pyx_v_self->_order_refresh_time);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1305
 *     cdef set_timers(self):
 *         cdef double next_cycle = self._current_timestamp + self._order_refresh_time
 *         if self._create_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *             self._create_timestamp = next_cycle
 *         if self._cancel_timestamp <= self._current_timestamp:
*/
  __pyx_t_1 = (__pyx_v_self->_create_timestamp <= __pyx_v_self->__pyx_base.__pyx_base._current_timestamp);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1306
 *         cdef double next_cycle = self._current_timestamp + self._order_refresh_time
 *         if self._create_timestamp <= self._current_timestamp:
 *             self._create_timestamp = next_cycle             # <<<<<<<<<<<<<<
 *         if self._cancel_timestamp <= self._current_timestamp:
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)
*/
    __pyx_v_self->_create_timestamp = __pyx_v_next_cycle;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1305
 *     cdef set_timers(self):
 *         cdef double next_cycle = self._current_timestamp + self._order_refresh_time
 *         if self._create_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *             self._create_timestamp = next_cycle
 *         if self._cancel_timestamp <= self._current_timestamp:
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1307
 *         if self._create_timestamp <= self._current_timestamp:
 *             self._create_timestamp = next_cycle
 *         if self._cancel_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)
 * 
*/
  __pyx_t_1 = (__pyx_v_self->_cancel_timestamp <= __pyx_v_self->__pyx_base.__pyx_base._current_timestamp);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1308
 *             self._create_timestamp = next_cycle
 *         if self._cancel_timestamp <= self._current_timestamp:
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)             # <<<<<<<<<<<<<<
 * 
 *     def notify_hb_app(self, msg: str):
*/
    __pyx_t_2 = __pyx_v_next_cycle;
    __pyx_t_3 = __pyx_v_self->_create_timestamp;
    __pyx_t_1 = (__pyx_t_2 < __pyx_t_3);
    if (__pyx_t_1) {
      __pyx_t_4 = __pyx_t_2;
    } else {
      __pyx_t_4 = __pyx_t_3;
    }
    __pyx_v_self->_cancel_timestamp = __pyx_t_4;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1307
 *         if self._create_timestamp <= self._current_timestamp:
 *             self._create_timestamp = next_cycle
 *         if self._cancel_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1303
 *             self.set_timers()
 * 
 *     cdef set_timers(self):             # <<<<<<<<<<<<<<
 *         cdef double next_cycle = self._current_timestamp + self._order_refresh_time
 *         if self._create_timestamp <= self._current_timestamp:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1310
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)
 * 
 *     def notify_hb_app(self, msg: str):             # <<<<<<<<<<<<<<
 *         if self._hb_app_notification:
 *             super().notify_hb_app(msg)
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27notify_hb_app(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27notify_hb_app = {"notify_hb_app", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27notify_hb_app, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27notify_hb_app(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_msg = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("notify_hb_app (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_msg,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1310, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1310, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "notify_hb_app", 0) < (0)) __PYX_ERR(0, 1310, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("notify_hb_app", 1, 1, 1, i); __PYX_ERR(0, 1310, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1310, __pyx_L3_error)
    }
    __pyx_v_msg = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("notify_hb_app", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1310, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.notify_hb_app", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26notify_hb_app(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), __pyx_v_msg);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26notify_hb_app(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_msg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("notify_hb_app", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1311
 * 
 *     def notify_hb_app(self, msg: str):
 *         if self._hb_app_notification:             # <<<<<<<<<<<<<<
 *             super().notify_hb_app(msg)
 * 
*/
  if (__pyx_v_self->_hb_app_notification) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1312
 *     def notify_hb_app(self, msg: str):
 *         if self._hb_app_notification:
 *             super().notify_hb_app(msg)             # <<<<<<<<<<<<<<
 * 
 *     def get_price_type(self, price_type_str: str) -> PriceType:
*/
    __pyx_t_4 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy), ((PyObject *)__pyx_v_self)};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_t_2 = __pyx_t_3;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_msg};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_notify_hb_app, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1311
 * 
 *     def notify_hb_app(self, msg: str):
 *         if self._hb_app_notification:             # <<<<<<<<<<<<<<
 *             super().notify_hb_app(msg)
 * 
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1310
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)
 * 
 *     def notify_hb_app(self, msg: str):             # <<<<<<<<<<<<<<
 *         if self._hb_app_notification:
 *             super().notify_hb_app(msg)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.notify_hb_app", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1314
 *             super().notify_hb_app(msg)
 * 
 *     def get_price_type(self, price_type_str: str) -> PriceType:             # <<<<<<<<<<<<<<
 *         if price_type_str == "mid_price":
 *             return PriceType.MidPrice
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29get_price_type(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29get_price_type = {"get_price_type", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29get_price_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29get_price_type(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_price_type_str = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_price_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_price_type_str,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1314, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1314, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_price_type", 0) < (0)) __PYX_ERR(0, 1314, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_price_type", 1, 1, 1, i); __PYX_ERR(0, 1314, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1314, __pyx_L3_error)
    }
    __pyx_v_price_type_str = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_price_type", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1314, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.get_price_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28get_price_type(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), __pyx_v_price_type_str);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28get_price_type(CYTHON_UNUSED struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_price_type_str) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5[3];
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_price_type", 0);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1315
 * 
 *     def get_price_type(self, price_type_str: str) -> PriceType:
 *         if price_type_str == "mid_price":             # <<<<<<<<<<<<<<
 *             return PriceType.MidPrice
 *         elif price_type_str == "best_bid":
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_price_type_str, __pyx_mstate_global->__pyx_n_u_mid_price, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1315, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1316
 *     def get_price_type(self, price_type_str: str) -> PriceType:
 *         if price_type_str == "mid_price":
 *             return PriceType.MidPrice             # <<<<<<<<<<<<<<
 *         elif price_type_str == "best_bid":
 *             return PriceType.BestBid
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PriceType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_MidPrice); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1315
 * 
 *     def get_price_type(self, price_type_str: str) -> PriceType:
 *         if price_type_str == "mid_price":             # <<<<<<<<<<<<<<
 *             return PriceType.MidPrice
 *         elif price_type_str == "best_bid":
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1317
 *         if price_type_str == "mid_price":
 *             return PriceType.MidPrice
 *         elif price_type_str == "best_bid":             # <<<<<<<<<<<<<<
 *             return PriceType.BestBid
 *         elif price_type_str == "best_ask":
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_price_type_str, __pyx_mstate_global->__pyx_n_u_best_bid, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1317, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1318
 *             return PriceType.MidPrice
 *         elif price_type_str == "best_bid":
 *             return PriceType.BestBid             # <<<<<<<<<<<<<<
 *         elif price_type_str == "best_ask":
 *             return PriceType.BestAsk
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PriceType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_BestBid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1317
 *         if price_type_str == "mid_price":
 *             return PriceType.MidPrice
 *         elif price_type_str == "best_bid":             # <<<<<<<<<<<<<<
 *             return PriceType.BestBid
 *         elif price_type_str == "best_ask":
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1319
 *         elif price_type_str == "best_bid":
 *             return PriceType.BestBid
 *         elif price_type_str == "best_ask":             # <<<<<<<<<<<<<<
 *             return PriceType.BestAsk
 *         elif price_type_str == "last_price":
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_price_type_str, __pyx_mstate_global->__pyx_n_u_best_ask, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1319, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1320
 *             return PriceType.BestBid
 *         elif price_type_str == "best_ask":
 *             return PriceType.BestAsk             # <<<<<<<<<<<<<<
 *         elif price_type_str == "last_price":
 *             return PriceType.LastTrade
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PriceType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_BestAsk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1319
 *         elif price_type_str == "best_bid":
 *             return PriceType.BestBid
 *         elif price_type_str == "best_ask":             # <<<<<<<<<<<<<<
 *             return PriceType.BestAsk
 *         elif price_type_str == "last_price":
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1321
 *         elif price_type_str == "best_ask":
 *             return PriceType.BestAsk
 *         elif price_type_str == "last_price":             # <<<<<<<<<<<<<<
 *             return PriceType.LastTrade
 *         elif price_type_str == 'last_own_trade_price':
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_price_type_str, __pyx_mstate_global->__pyx_n_u_last_price, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1321, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1322
 *             return PriceType.BestAsk
 *         elif price_type_str == "last_price":
 *             return PriceType.LastTrade             # <<<<<<<<<<<<<<
 *         elif price_type_str == 'last_own_trade_price':
 *             return PriceType.LastOwnTrade
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PriceType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_LastTrade); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1321
 *         elif price_type_str == "best_ask":
 *             return PriceType.BestAsk
 *         elif price_type_str == "last_price":             # <<<<<<<<<<<<<<
 *             return PriceType.LastTrade
 *         elif price_type_str == 'last_own_trade_price':
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1323
 *         elif price_type_str == "last_price":
 *             return PriceType.LastTrade
 *         elif price_type_str == 'last_own_trade_price':             # <<<<<<<<<<<<<<
 *             return PriceType.LastOwnTrade
 *         elif price_type_str == 'inventory_cost':
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_price_type_str, __pyx_mstate_global->__pyx_n_u_last_own_trade_price, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1323, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1324
 *             return PriceType.LastTrade
 *         elif price_type_str == 'last_own_trade_price':
 *             return PriceType.LastOwnTrade             # <<<<<<<<<<<<<<
 *         elif price_type_str == 'inventory_cost':
 *             return PriceType.InventoryCost
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PriceType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_LastOwnTrade); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1323
 *         elif price_type_str == "last_price":
 *             return PriceType.LastTrade
 *         elif price_type_str == 'last_own_trade_price':             # <<<<<<<<<<<<<<
 *             return PriceType.LastOwnTrade
 *         elif price_type_str == 'inventory_cost':
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1325
 *         elif price_type_str == 'last_own_trade_price':
 *             return PriceType.LastOwnTrade
 *         elif price_type_str == 'inventory_cost':             # <<<<<<<<<<<<<<
 *             return PriceType.InventoryCost
 *         elif price_type_str == "custom":
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_price_type_str, __pyx_mstate_global->__pyx_n_u_inventory_cost, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1325, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1326
 *             return PriceType.LastOwnTrade
 *         elif price_type_str == 'inventory_cost':
 *             return PriceType.InventoryCost             # <<<<<<<<<<<<<<
 *         elif price_type_str == "custom":
 *             return PriceType.Custom
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PriceType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_InventoryCost); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1325
 *         elif price_type_str == 'last_own_trade_price':
 *             return PriceType.LastOwnTrade
 *         elif price_type_str == 'inventory_cost':             # <<<<<<<<<<<<<<
 *             return PriceType.InventoryCost
 *         elif price_type_str == "custom":
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1327
 *         elif price_type_str == 'inventory_cost':
 *             return PriceType.InventoryCost
 *         elif price_type_str == "custom":             # <<<<<<<<<<<<<<
 *             return PriceType.Custom
 *         else:
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_price_type_str, __pyx_mstate_global->__pyx_n_u_custom, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1327, __pyx_L1_error)
  if (likely(__pyx_t_1)) {

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1328
 *             return PriceType.InventoryCost
 *         elif price_type_str == "custom":
 *             return PriceType.Custom             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError(f"Unrecognized price type string {price_type_str}.")
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PriceType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_Custom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1327
 *         elif price_type_str == 'inventory_cost':
 *             return PriceType.InventoryCost
 *         elif price_type_str == "custom":             # <<<<<<<<<<<<<<
 *             return PriceType.Custom
 *         else:
*/
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1330
 *             return PriceType.Custom
 *         else:
 *             raise ValueError(f"Unrecognized price type string {price_type_str}.")             # <<<<<<<<<<<<<<
*/
  /*else*/ {
    __pyx_t_2 = NULL;
    __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_price_type_str, __pyx_mstate_global->__pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5[0] = __pyx_mstate_global->__pyx_kp_u_Unrecognized_price_type_string;
    __pyx_t_5[1] = __pyx_t_4;
    __pyx_t_5[2] = __pyx_mstate_global->__pyx_kp_u__25;
    __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_5, 3, 31 + __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4) + 1, 127 | __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4));
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1330, __pyx_L1_error)
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1314
 *             super().notify_hb_app(msg)
 * 
 *     def get_price_type(self, price_type_str: str) -> PriceType:             # <<<<<<<<<<<<<<
 *         if price_type_str == "mid_price":
 *             return PriceType.MidPrice
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.get_price_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_30__reduce_cython__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_30__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self__events_cannot_be_converted, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_33__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_33__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_33__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_33__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_32__setstate_cython__(((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_32__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self__events_cannot_be_converted, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */
static struct __pyx_vtabstruct_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy;

static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub*)__pyx_vtabptr_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy;
  p->_market_info = Py_None; Py_INCREF(Py_None);
  p->_bid_spread = Py_None; Py_INCREF(Py_None);
  p->_ask_spread = Py_None; Py_INCREF(Py_None);
  p->_minimum_spread = Py_None; Py_INCREF(Py_None);
  p->_order_amount = Py_None; Py_INCREF(Py_None);
  p->_split_order_levels_enabled = Py_None; Py_INCREF(Py_None);
  p->_bid_order_level_spreads = Py_None; Py_INCREF(Py_None);
  p->_ask_order_level_spreads = Py_None; Py_INCREF(Py_None);
  p->_order_level_spread = Py_None; Py_INCREF(Py_None);
  p->_order_level_amount = Py_None; Py_INCREF(Py_None);
  p->_order_refresh_tolerance_pct = Py_None; Py_INCREF(Py_None);
  p->_inventory_target_base_pct = Py_None; Py_INCREF(Py_None);
  p->_inventory_range_multiplier = Py_None; Py_INCREF(Py_None);
  p->_hanging_orders_tracker = Py_None; Py_INCREF(Py_None);
  p->_ask_order_optimization_depth = Py_None; Py_INCREF(Py_None);
  p->_bid_order_optimization_depth = Py_None; Py_INCREF(Py_None);
  p->_asset_price_delegate = Py_None; Py_INCREF(Py_None);
  p->_inventory_cost_price_delegate = Py_None; Py_INCREF(Py_None);
  p->_price_type = Py_None; Py_INCREF(Py_None);
  p->_price_ceiling = Py_None; Py_INCREF(Py_None);
  p->_price_floor = Py_None; Py_INCREF(Py_None);
  p->_ping_pong_warning_lines = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_order_override = Py_None; Py_INCREF(Py_None);
  p->_limit_order_type = Py_None; Py_INCREF(Py_None);
  p->_last_own_trade_price = Py_None; Py_INCREF(Py_None);
  p->_moving_price_band = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_market_info);
  Py_CLEAR(p->_bid_spread);
  Py_CLEAR(p->_ask_spread);
  Py_CLEAR(p->_minimum_spread);
  Py_CLEAR(p->_order_amount);
  Py_CLEAR(p->_split_order_levels_enabled);
  Py_CLEAR(p->_bid_order_level_spreads);
  Py_CLEAR(p->_ask_order_level_spreads);
  Py_CLEAR(p->_order_level_spread);
  Py_CLEAR(p->_order_level_amount);
  Py_CLEAR(p->_order_refresh_tolerance_pct);
  Py_CLEAR(p->_inventory_target_base_pct);
  Py_CLEAR(p->_inventory_range_multiplier);
  Py_CLEAR(p->_hanging_orders_tracker);
  Py_CLEAR(p->_ask_order_optimization_depth);
  Py_CLEAR(p->_bid_order_optimization_depth);
  Py_CLEAR(p->_asset_price_delegate);
  Py_CLEAR(p->_inventory_cost_price_delegate);
  Py_CLEAR(p->_price_type);
  Py_CLEAR(p->_price_ceiling);
  Py_CLEAR(p->_price_floor);
  Py_CLEAR(p->_ping_pong_warning_lines);
  Py_CLEAR(p->_order_override);
  Py_CLEAR(p->_limit_order_type);
  Py_CLEAR(p->_last_own_trade_price);
  Py_CLEAR(p->_moving_price_band);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy);
}

static int __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)o;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  if (p->_market_info) {
    e = (*v)(p->_market_info, a); if (e) return e;
  }
  if (p->_bid_spread) {
    e = (*v)(p->_bid_spread, a); if (e) return e;
  }
  if (p->_ask_spread) {
    e = (*v)(p->_ask_spread, a); if (e) return e;
  }
  if (p->_minimum_spread) {
    e = (*v)(p->_minimum_spread, a); if (e) return e;
  }
  if (p->_order_amount) {
    e = (*v)(p->_order_amount, a); if (e) return e;
  }
  if (p->_split_order_levels_enabled) {
    e = (*v)(p->_split_order_levels_enabled, a); if (e) return e;
  }
  if (p->_bid_order_level_spreads) {
    e = (*v)(p->_bid_order_level_spreads, a); if (e) return e;
  }
  if (p->_ask_order_level_spreads) {
    e = (*v)(p->_ask_order_level_spreads, a); if (e) return e;
  }
  if (p->_order_level_spread) {
    e = (*v)(p->_order_level_spread, a); if (e) return e;
  }
  if (p->_order_level_amount) {
    e = (*v)(p->_order_level_amount, a); if (e) return e;
  }
  if (p->_order_refresh_tolerance_pct) {
    e = (*v)(p->_order_refresh_tolerance_pct, a); if (e) return e;
  }
  if (p->_inventory_target_base_pct) {
    e = (*v)(p->_inventory_target_base_pct, a); if (e) return e;
  }
  if (p->_inventory_range_multiplier) {
    e = (*v)(p->_inventory_range_multiplier, a); if (e) return e;
  }
  if (p->_hanging_orders_tracker) {
    e = (*v)(p->_hanging_orders_tracker, a); if (e) return e;
  }
  if (p->_ask_order_optimization_depth) {
    e = (*v)(p->_ask_order_optimization_depth, a); if (e) return e;
  }
  if (p->_bid_order_optimization_depth) {
    e = (*v)(p->_bid_order_optimization_depth, a); if (e) return e;
  }
  if (p->_asset_price_delegate) {
    e = (*v)(p->_asset_price_delegate, a); if (e) return e;
  }
  if (p->_inventory_cost_price_delegate) {
    e = (*v)(p->_inventory_cost_price_delegate, a); if (e) return e;
  }
  if (p->_price_type) {
    e = (*v)(p->_price_type, a); if (e) return e;
  }
  if (p->_price_ceiling) {
    e = (*v)(p->_price_ceiling, a); if (e) return e;
  }
  if (p->_price_floor) {
    e = (*v)(p->_price_floor, a); if (e) return e;
  }
  if (p->_ping_pong_warning_lines) {
    e = (*v)(p->_ping_pong_warning_lines, a); if (e) return e;
  }
  if (p->_order_override) {
    e = (*v)(p->_order_override, a); if (e) return e;
  }
  if (p->_limit_order_type) {
    e = (*v)(p->_limit_order_type, a); if (e) return e;
  }
  if (p->_last_own_trade_price) {
    e = (*v)(p->_last_own_trade_price, a); if (e) return e;
  }
  if (p->_moving_price_band) {
    e = (*v)(p->_moving_price_band, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)o;
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy); }
  tmp = ((PyObject*)p->_market_info);
  p->_market_info = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_bid_spread);
  p->_bid_spread = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_ask_spread);
  p->_ask_spread = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_minimum_spread);
  p->_minimum_spread = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_order_amount);
  p->_order_amount = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_split_order_levels_enabled);
  p->_split_order_levels_enabled = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_bid_order_level_spreads);
  p->_bid_order_level_spreads = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_ask_order_level_spreads);
  p->_ask_order_level_spreads = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_order_level_spread);
  p->_order_level_spread = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_order_level_amount);
  p->_order_level_amount = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_order_refresh_tolerance_pct);
  p->_order_refresh_tolerance_pct = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_inventory_target_base_pct);
  p->_inventory_target_base_pct = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_inventory_range_multiplier);
  p->_inventory_range_multiplier = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_hanging_orders_tracker);
  p->_hanging_orders_tracker = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_ask_order_optimization_depth);
  p->_ask_order_optimization_depth = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_bid_order_optimization_depth);
  p->_bid_order_optimization_depth = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_asset_price_delegate);
  p->_asset_price_delegate = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_inventory_cost_price_delegate);
  p->_inventory_cost_price_delegate = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_price_type);
  p->_price_type = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_price_ceiling);
  p->_price_ceiling = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_price_floor);
  p->_price_floor = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_ping_pong_warning_lines);
  p->_ping_pong_warning_lines = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_order_override);
  p->_order_override = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_limit_order_type);
  p->_limit_order_type = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_last_own_trade_price);
  p->_last_own_trade_price = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_moving_price_band);
  p->_moving_price_band = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_market_info(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11market_info_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_max_order_age(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13max_order_age_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_minimum_spread(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14minimum_spread_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_ping_pong_enabled(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17ping_pong_enabled_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_ask_order_optimization_depth(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28ask_order_optimization_depth_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_bid_order_optimization_depth(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28bid_order_optimization_depth_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10price_type_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_refresh_tolerance_pct(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27order_refresh_tolerance_pct_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_refresh_tolerance_pct(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27order_refresh_tolerance_pct_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_amount(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_amount_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_amount(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_amount_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_levels(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_levels_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_levels(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12order_levels_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_buy_levels(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10buy_levels_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_buy_levels(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10buy_levels_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_sell_levels(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11sell_levels_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_sell_levels(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11sell_levels_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_level_amount(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_amount_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_level_amount(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_amount_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_level_spread(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_spread_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_level_spread(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_level_spread_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_skew_enabled(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22inventory_skew_enabled_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_skew_enabled(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22inventory_skew_enabled_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_target_base_pct(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25inventory_target_base_pct_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_target_base_pct(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25inventory_target_base_pct_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_range_multiplier(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26inventory_range_multiplier_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_range_multiplier(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26inventory_range_multiplier_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_hanging_orders_enabled(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_enabled_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_hanging_orders_enabled(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_enabled_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_hanging_orders_cancel_pct(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25hanging_orders_cancel_pct_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_hanging_orders_cancel_pct(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25hanging_orders_cancel_pct_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_bid_spread(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10bid_spread_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_bid_spread(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10bid_spread_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_ask_spread(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10ask_spread_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_ask_spread(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10ask_spread_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_optimization_enabled(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26order_optimization_enabled_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_optimization_enabled(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26order_optimization_enabled_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_refresh_time(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_refresh_time_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_refresh_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18order_refresh_time_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_filled_order_delay(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18filled_order_delay_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_filled_order_delay(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_18filled_order_delay_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_add_transaction_costs_to_orders(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31add_transaction_costs_to_orders_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_add_transaction_costs_to_orders(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31add_transaction_costs_to_orders_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_ceiling(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13price_ceiling_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_ceiling(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13price_ceiling_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_floor(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11price_floor_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_floor(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11price_floor_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_base_asset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_10base_asset_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_quote_asset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11quote_asset_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_trading_pair(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12trading_pair_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_override(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14order_override_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_override(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_14order_override_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_split_order_levels_enabled(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_26split_order_levels_enabled_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_bid_order_level_spreads(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23bid_order_level_spreads_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_ask_order_level_spreads(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23ask_order_level_spreads_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_moving_price_band_enabled(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25moving_price_band_enabled_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_moving_price_band_enabled(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25moving_price_band_enabled_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_ceiling_pct(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17price_ceiling_pct_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_ceiling_pct(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17price_ceiling_pct_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_floor_pct(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15price_floor_pct_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_floor_pct(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15price_floor_pct_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_band_refresh_time(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23price_band_refresh_time_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_band_refresh_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23price_band_refresh_time_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_moving_price_band(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17moving_price_band_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_hanging_order_ids(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17hanging_order_ids_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_market_info_to_active_orders(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_28market_info_to_active_orders_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_active_orders(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13active_orders_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_active_buys(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11active_buys_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_active_sells(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_12active_sells_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_active_non_hanging_orders(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25active_non_hanging_orders_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_logging_options(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15logging_options_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_logging_options(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15logging_options_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_hanging_orders_tracker(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_22hanging_orders_tracker_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_asset_price_delegate(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20asset_price_delegate_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_asset_price_delegate(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_20asset_price_delegate_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_cost_price_delegate(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29inventory_cost_price_delegate_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_cost_price_delegate(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29inventory_cost_price_delegate_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy[] = {
  {"logger", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_1logger, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"init_params", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_3init_params, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"all_markets_ready", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_5all_markets_ready, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_price", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7get_price, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_mid_price", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_9get_mid_price, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"inventory_skew_stats_data_frame", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11inventory_skew_stats_data_frame, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"pure_mm_assets_df", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13pure_mm_assets_df, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"active_orders_df", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15active_orders_df, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"market_status_data_frame", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17market_status_data_frame, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"format_status", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_19format_status, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"execute_orders_proposal", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_21execute_orders_proposal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"cancel_order", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23cancel_order, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"adjusted_available_balance_for_orders_budget_constrain", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25adjusted_available_balance_for_orders_budget_constrain, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"notify_hb_app", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27notify_hb_app, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_price_type", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29get_price_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_33__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy[] = {
  {"market_info", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_market_info, 0, 0, 0},
  {"max_order_age", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_max_order_age, 0, 0, 0},
  {"minimum_spread", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_minimum_spread, 0, 0, 0},
  {"ping_pong_enabled", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_ping_pong_enabled, 0, 0, 0},
  {"ask_order_optimization_depth", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_ask_order_optimization_depth, 0, 0, 0},
  {"bid_order_optimization_depth", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_bid_order_optimization_depth, 0, 0, 0},
  {"price_type", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_type, 0, 0, 0},
  {"order_refresh_tolerance_pct", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_refresh_tolerance_pct, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_refresh_tolerance_pct, 0, 0},
  {"order_amount", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_amount, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_amount, 0, 0},
  {"order_levels", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_levels, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_levels, 0, 0},
  {"buy_levels", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_buy_levels, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_buy_levels, 0, 0},
  {"sell_levels", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_sell_levels, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_sell_levels, 0, 0},
  {"order_level_amount", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_level_amount, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_level_amount, 0, 0},
  {"order_level_spread", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_level_spread, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_level_spread, 0, 0},
  {"inventory_skew_enabled", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_skew_enabled, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_skew_enabled, 0, 0},
  {"inventory_target_base_pct", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_target_base_pct, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_target_base_pct, 0, 0},
  {"inventory_range_multiplier", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_range_multiplier, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_range_multiplier, 0, 0},
  {"hanging_orders_enabled", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_hanging_orders_enabled, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_hanging_orders_enabled, 0, 0},
  {"hanging_orders_cancel_pct", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_hanging_orders_cancel_pct, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_hanging_orders_cancel_pct, 0, 0},
  {"bid_spread", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_bid_spread, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_bid_spread, 0, 0},
  {"ask_spread", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_ask_spread, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_ask_spread, 0, 0},
  {"order_optimization_enabled", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_optimization_enabled, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_optimization_enabled, 0, 0},
  {"order_refresh_time", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_refresh_time, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_refresh_time, 0, 0},
  {"filled_order_delay", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_filled_order_delay, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_filled_order_delay, 0, 0},
  {"add_transaction_costs_to_orders", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_add_transaction_costs_to_orders, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_add_transaction_costs_to_orders, 0, 0},
  {"price_ceiling", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_ceiling, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_ceiling, 0, 0},
  {"price_floor", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_floor, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_floor, 0, 0},
  {"base_asset", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_base_asset, 0, 0, 0},
  {"quote_asset", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_quote_asset, 0, 0, 0},
  {"trading_pair", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_trading_pair, 0, 0, 0},
  {"order_override", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_override, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_order_override, 0, 0},
  {"split_order_levels_enabled", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_split_order_levels_enabled, 0, 0, 0},
  {"bid_order_level_spreads", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_bid_order_level_spreads, 0, 0, 0},
  {"ask_order_level_spreads", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_ask_order_level_spreads, 0, 0, 0},
  {"moving_price_band_enabled", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_moving_price_band_enabled, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_moving_price_band_enabled, 0, 0},
  {"price_ceiling_pct", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_ceiling_pct, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_ceiling_pct, 0, 0},
  {"price_floor_pct", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_floor_pct, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_floor_pct, 0, 0},
  {"price_band_refresh_time", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_band_refresh_time, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_price_band_refresh_time, 0, 0},
  {"moving_price_band", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_moving_price_band, 0, 0, 0},
  {"hanging_order_ids", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_hanging_order_ids, 0, 0, 0},
  {"market_info_to_active_orders", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_market_info_to_active_orders, 0, 0, 0},
  {"active_orders", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_active_orders, 0, 0, 0},
  {"active_buys", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_active_buys, 0, 0, 0},
  {"active_sells", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_active_sells, 0, 0, 0},
  {"active_non_hanging_orders", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_active_non_hanging_orders, 0, 0, 0},
  {"logging_options", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_logging_options, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_logging_options, 0, 0},
  {"hanging_orders_tracker", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_hanging_orders_tracker, 0, 0, 0},
  {"asset_price_delegate", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_asset_price_delegate, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_asset_price_delegate, 0, 0},
  {"inventory_cost_price_delegate", __pyx_getprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_cost_price_delegate, __pyx_setprop_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_inventory_cost_price_delegate, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy},
  {Py_tp_clear, (void *)__pyx_tp_clear_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy},
  {Py_tp_methods, (void *)__pyx_methods_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy},
  {Py_tp_getset, (void *)__pyx_getsets_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy_spec = {
  "hummingbot.strategy.pure_market_making.pure_market_making.PureMarketMakingStrategy",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.pure_market_making.pure_market_making.""PureMarketMakingStrategy", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, /*tp_traverse*/
  __pyx_tp_clear_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start[--__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_order_id);
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start))))
  {
    __pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start[__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start++] = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->__pyx_v_order_id) {
    e = (*v)(p->__pyx_v_order_id, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *)o;
  tmp = ((PyObject*)p->__pyx_v_order_id);
  p->__pyx_v_order_id = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start},
  {Py_tp_clear, (void *)__pyx_tp_clear_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start_spec = {
  "hummingbot.strategy.pure_market_making.pure_market_making.__pyx_scope_struct____pyx_base____pyx_base__c_start",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.pure_market_making.pure_market_making.""__pyx_scope_struct____pyx_base____pyx_base__c_start", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start, /*tp_traverse*/
  __pyx_tp_clear_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr[--__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_o);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr))))
  {
    __pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr[__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_o) {
    e = (*v)(p->__pyx_v_o, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr_spec = {
  "hummingbot.strategy.pure_market_making.pure_market_making.__pyx_scope_struct_1_genexpr",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.pure_market_making.pure_market_making.""__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit[--__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit))))
  {
    __pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit[__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit++] = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit},
  {Py_tp_clear, (void *)__pyx_tp_clear_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit_spec = {
  "hummingbot.strategy.pure_market_making.pure_market_making.__pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.pure_market_making.pure_market_making.""__pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit, /*tp_traverse*/
  __pyx_tp_clear_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr[--__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_o);
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr))))
  {
    __pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr[__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr++] = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_o) {
    e = (*v)(p->__pyx_v_o, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr_spec = {
  "hummingbot.strategy.pure_market_making.pure_market_making.__pyx_scope_struct_3_genexpr",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.pure_market_making.pure_market_making.""__pyx_scope_struct_3_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal[--__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  Py_CLEAR(p->__pyx_v_ask_order_id);
  Py_CLEAR(p->__pyx_v_bid_order_id);
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal))))
  {
    __pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal[__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal++] = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal_spec = {
  "hummingbot.strategy.pure_market_making.pure_market_making.__pyx_scope_struct_4_c_execute_orders_proposal",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.pure_market_making.pure_market_making.""__pyx_scope_struct_4_c_execute_orders_proposal", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr[--__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_o);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr))))
  {
    __pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr[__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr++] = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_o) {
    e = (*v)(p->__pyx_v_o, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr_spec = {
  "hummingbot.strategy.pure_market_making.pure_market_making.__pyx_scope_struct_5_genexpr",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.pure_market_making.pure_market_making.""__pyx_scope_struct_5_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr > 0) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(t, __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr))))
  {
    o = (PyObject*)__pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr[--__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr];
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(Py_TYPE(o));
    #endif
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr));
    #if CYTHON_COMPILING_IN_LIMITED_API
    (void) PyObject_Init(o, t);
    #else
    (void) PyObject_INIT(o, t);
    #endif
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = __Pyx_AllocateExtensionType(t, 1);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_o);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr < 8) & __PYX_CHECK_FINAL_TYPE_FOR_FREELISTS(Py_TYPE(o), __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr, sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr))))
  {
    __pyx_mstate_global->__pyx_freelist_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr[__pyx_mstate_global->__pyx_freecount_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr++] = ((struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr *)o);
  } else
  #endif
  {
    PyTypeObject *tp = Py_TYPE(o);
    #if CYTHON_USE_TYPE_SLOTS
    (*tp->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
    #if CYTHON_USE_TYPE_SPECS
    Py_DECREF(tp);
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_o) {
    e = (*v)(p->__pyx_v_o, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr_spec = {
  "hummingbot.strategy.pure_market_making.pure_market_making.__pyx_scope_struct_6_genexpr",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.pure_market_making.pure_market_making.""__pyx_scope_struct_6_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_t_1 = PyImport_ImportModule("hummingbot.strategy.strategy_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.strategy.strategy_base", "StrategyBase",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase),
  #else
  sizeof(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase = (struct __pyx_vtabstruct_10hummingbot_8strategy_13strategy_base_StrategyBase*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase); if (unlikely(!__pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy = &__pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.__pyx_base = *__pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.__pyx_base.__pyx_base.c_start = (PyObject *(*)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, struct __pyx_obj_10hummingbot_4core_5clock_Clock *, double))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_start;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.__pyx_base.__pyx_base.c_stop = (PyObject *(*)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, struct __pyx_obj_10hummingbot_4core_5clock_Clock *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_stop;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.__pyx_base.__pyx_base.c_tick = (PyObject *(*)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, double))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_tick;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.__pyx_base.c_did_fill_order = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_did_fill_order;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.__pyx_base.c_did_complete_buy_order = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_did_complete_buy_order;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.__pyx_base.c_did_complete_sell_order = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_did_complete_sell_order;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_get_mid_price = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_get_mid_price;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_create_base_proposal = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_create_base_proposal;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_get_adjusted_available_balance = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_get_adjusted_available_balance;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_apply_order_levels_modifiers = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_order_levels_modifiers;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_apply_price_band = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_price_band;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_apply_ping_pong = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_ping_pong;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_apply_order_price_modifiers = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_order_price_modifiers;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_apply_order_size_modifiers = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_order_size_modifiers;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_apply_inventory_skew = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_inventory_skew;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_apply_budget_constraint = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_budget_constraint;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_filter_out_takers = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_filter_out_takers;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_apply_order_optimization = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_order_optimization;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_apply_add_transaction_costs = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_add_transaction_costs;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_is_within_tolerance = (int (*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_is_within_tolerance;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_cancel_active_orders = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_cancel_active_orders;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_cancel_orders_below_min_spread = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_cancel_orders_below_min_spread;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_cancel_active_orders_on_max_age_limit = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_cancel_active_orders_on_max_age_limit;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_to_create_orders = (int (*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_to_create_orders;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_execute_orders_proposal = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_execute_orders_proposal;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.set_timers = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_set_timers;
  __pyx_vtable_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy.c_apply_moving_price_band = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_c_apply_moving_price_band;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy)) __PYX_ERR(0, 38, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy_spec, __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy) < (0)) __PYX_ERR(0, 38, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy = &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy->tp_base = __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy) < (0)) __PYX_ERR(0, 38, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy->tp_dictoffset && __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_vtabptr_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy) < (0)) __PYX_ERR(0, 38, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy) < (0)) __PYX_ERR(0, 38, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy, (PyObject *) __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy) < (0)) __PYX_ERR(0, 38, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy->tp_weaklistoffset = offsetof(struct __pyx_obj_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_base.__pyx_base.__pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy) < (0)) __PYX_ERR(0, 38, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start)) __PYX_ERR(0, 696, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start_spec, __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start) < (0)) __PYX_ERR(0, 696, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start = &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start) < (0)) __PYX_ERR(0, 696, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start->tp_dictoffset && __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr)) __PYX_ERR(0, 707, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr_spec, __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr) < (0)) __PYX_ERR(0, 707, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr = &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr) < (0)) __PYX_ERR(0, 707, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr->tp_dictoffset && __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_1_genexpr->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit)) __PYX_ERR(0, 1177, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit_spec, __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit) < (0)) __PYX_ERR(0, 1177, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit = &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit) < (0)) __PYX_ERR(0, 1177, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit->tp_dictoffset && __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_2_c_cancel_active_orders_on_max_age_limit->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr)) __PYX_ERR(0, 1184, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr_spec, __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr) < (0)) __PYX_ERR(0, 1184, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr = &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr) < (0)) __PYX_ERR(0, 1184, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr->tp_dictoffset && __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_3_genexpr->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal)) __PYX_ERR(0, 1247, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal_spec, __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal) < (0)) __PYX_ERR(0, 1247, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal = &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal) < (0)) __PYX_ERR(0, 1247, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal->tp_dictoffset && __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_4_c_execute_orders_proposal->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr)) __PYX_ERR(0, 1274, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr_spec, __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr) < (0)) __PYX_ERR(0, 1274, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr = &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr) < (0)) __PYX_ERR(0, 1274, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr->tp_dictoffset && __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_5_genexpr->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr)) __PYX_ERR(0, 1297, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr_spec, __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr) < (0)) __PYX_ERR(0, 1297, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr = &__pyx_type_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr) < (0)) __PYX_ERR(0, 1297, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr->tp_dictoffset && __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making___pyx_scope_struct_6_genexpr->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.clock"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_4core_5clock_Clock = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.core.clock", "Clock",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_4core_5clock_Clock), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_5clock_Clock),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_4core_5clock_Clock), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_5clock_Clock),
  #else
  sizeof(struct __pyx_obj_10hummingbot_4core_5clock_Clock), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_5clock_Clock),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_4core_5clock_Clock) __PYX_ERR(3, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.event.event_listener"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.core.event.event_listener", "EventListener",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener),
  #else
  sizeof(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener) __PYX_ERR(4, 6, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_5event_14event_listener_EventListener = (struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_5event_14event_listener_EventListener)) __PYX_ERR(4, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.pubsub"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_4core_6pubsub_PubSub = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.core.pubsub", "PubSub",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub),
  #else
  sizeof(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_4core_6pubsub_PubSub) __PYX_ERR(5, 17, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_6pubsub_PubSub = (struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_4core_6pubsub_PubSub); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_6pubsub_PubSub)) __PYX_ERR(5, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.time_iterator"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.core.time_iterator", "TimeIterator",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator),
  #else
  sizeof(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator) __PYX_ERR(6, 7, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_13time_iterator_TimeIterator = (struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_13time_iterator_TimeIterator)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.data_type.limit_order"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.core.data_type.limit_order", "LimitOrder",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder),
  #else
  sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder) __PYX_ERR(7, 6, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_9data_type_11limit_order_LimitOrder = (struct __pyx_vtabstruct_10hummingbot_4core_9data_type_11limit_order_LimitOrder*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_9data_type_11limit_order_LimitOrder)) __PYX_ERR(7, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.strategy.order_tracker"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.strategy.order_tracker", "OrderTracker",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker),
  #else
  sizeof(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker) __PYX_ERR(8, 7, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_8strategy_13order_tracker_OrderTracker = (struct __pyx_vtabstruct_10hummingbot_8strategy_13order_tracker_OrderTracker*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker); if (unlikely(!__pyx_vtabptr_10hummingbot_8strategy_13order_tracker_OrderTracker)) __PYX_ERR(8, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.event.event_reporter"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.core.event.event_reporter", "EventReporter",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter),
  #else
  sizeof(struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter) __PYX_ERR(9, 4, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_5event_14event_reporter_EventReporter = (struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_reporter_EventReporter*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_5event_14event_reporter_EventReporter)) __PYX_ERR(9, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.event.event_logger"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.core.event.event_logger", "EventLogger",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger),
  #else
  sizeof(struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger) __PYX_ERR(10, 4, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_5event_12event_logger_EventLogger = (struct __pyx_vtabstruct_10hummingbot_4core_5event_12event_logger_EventLogger*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_5event_12event_logger_EventLogger)) __PYX_ERR(10, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.data_type.order_book_query_result"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.core.data_type.order_book_query_result", "OrderBookQueryResult",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult),
  #else
  sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult) __PYX_ERR(11, 3, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.core.data_type.order_book_query_result", "ClientOrderBookQueryResult",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult),
  #else
  sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult) __PYX_ERR(11, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_2_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_4type_type) __PYX_ERR(12, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_5numpy_dtype = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "dtype",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArray_Descr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArray_Descr),
  #else
  sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArray_Descr),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 229, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_flatiter = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "flatiter",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayIterObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayIterObject),
  #else
  sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayIterObject),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 274, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_broadcast = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "broadcast",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayMultiIterObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayMultiIterObject),
  #else
  sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayMultiIterObject),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 278, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_ndarray = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "ndarray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayObject),
  #else
  sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayObject),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 317, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_generic = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "generic",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_generic) __PYX_ERR(1, 826, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_number = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "number",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_number) __PYX_ERR(1, 828, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_integer = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "integer",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_integer) __PYX_ERR(1, 830, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_signedinteger = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "signedinteger",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_signedinteger) __PYX_ERR(1, 832, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "unsignedinteger",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(1, 834, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_inexact = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "inexact",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_inexact) __PYX_ERR(1, 836, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_floating = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "floating",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_floating) __PYX_ERR(1, 838, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_complexfloating = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "complexfloating",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_complexfloating) __PYX_ERR(1, 840, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_flexible = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "flexible",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_flexible) __PYX_ERR(1, 842, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_character = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "character",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_character) __PYX_ERR(1, 844, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_ufunc = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "ufunc",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyUFuncObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyUFuncObject),
  #else
  sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyUFuncObject),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 908, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.data_type.order_book"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.core.data_type.order_book", "OrderBook",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook),
  #else
  sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook) __PYX_ERR(13, 12, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_9data_type_10order_book_OrderBook = (struct __pyx_vtabstruct_10hummingbot_4core_9data_type_10order_book_OrderBook*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_9data_type_10order_book_OrderBook)) __PYX_ERR(13, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.network_iterator"); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.core.network_iterator", "NetworkIterator",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator),
  #else
  sizeof(struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator) __PYX_ERR(14, 6, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_16network_iterator_NetworkIterator = (struct __pyx_vtabstruct_10hummingbot_4core_16network_iterator_NetworkIterator*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_16network_iterator_NetworkIterator)) __PYX_ERR(14, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.connector.connector_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.connector.connector_base", "ConnectorBase",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase),
  #else
  sizeof(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase) __PYX_ERR(15, 5, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_9connector_14connector_base_ConnectorBase = (struct __pyx_vtabstruct_10hummingbot_9connector_14connector_base_ConnectorBase*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase); if (unlikely(!__pyx_vtabptr_10hummingbot_9connector_14connector_base_ConnectorBase)) __PYX_ERR(15, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.connector.exchange_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.connector.exchange_base", "ExchangeBase",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase),
  #else
  sizeof(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase) __PYX_ERR(16, 10, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_9connector_13exchange_base_ExchangeBase = (struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase); if (unlikely(!__pyx_vtabptr_10hummingbot_9connector_13exchange_base_ExchangeBase)) __PYX_ERR(16, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.strategy.asset_price_delegate"); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.strategy.asset_price_delegate", "AssetPriceDelegate",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate),
  #else
  sizeof(struct __pyx_obj_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate) __PYX_ERR(17, 2, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate = (struct __pyx_vtabstruct_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate); if (unlikely(!__pyx_vtabptr_10hummingbot_8strategy_20asset_price_delegate_AssetPriceDelegate)) __PYX_ERR(17, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.strategy.order_book_asset_price_delegate"); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate = __Pyx_ImportType_3_2_4(__pyx_t_1, "hummingbot.strategy.order_book_asset_price_delegate", "OrderBookAssetPriceDelegate",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate),
  #else
  sizeof(struct __pyx_obj_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate) __PYX_ERR(18, 4, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate = (struct __pyx_vtabstruct_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate); if (unlikely(!__pyx_vtabptr_10hummingbot_8strategy_31order_book_asset_price_delegate_OrderBookAssetPriceDelegate)) __PYX_ERR(18, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  {
    __pyx_t_1 = PyImport_ImportModule("hummingbot.strategy.pure_market_making.inventory_skew_calculator"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    const char * __pyx_import_signature = __Pyx_PyBytes_AsString(__pyx_mstate_global->__pyx_kp_b_PyObject_double_double_double_do);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(!__pyx_import_signature)) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    const char * __pyx_import_name = __pyx_import_signature + 51;
    void (**const __pyx_import_pointers[])(void) = {(void (**)(void))&__pyx_f_10hummingbot_8strategy_18pure_market_making_25inventory_skew_calculator_c_calculate_bid_ask_ratios_from_base_asset_ratio, (void (**)(void)) NULL};
    void (**const *__pyx_import_pointer)(void) = __pyx_import_pointers;
    const char *__pyx_import_current_signature = __pyx_import_signature;
    while (*__pyx_import_pointer) {
      if (__Pyx_ImportFunction_3_2_4(__pyx_t_1, __pyx_import_name, *__pyx_import_pointer, __pyx_import_current_signature) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
      ++__pyx_import_pointer;
      __pyx_import_name = strchr(__pyx_import_name, '\0') + 1;
      __pyx_import_signature = strchr(__pyx_import_signature, '\0') + 1;
      if (*__pyx_import_signature != '\0') __pyx_import_current_signature = __pyx_import_signature;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_pure_market_making(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_pure_market_making},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "pure_market_making",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_pure_market_making(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_pure_market_making(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_pure_market_making(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'pure_market_making' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "pure_market_making" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit_pure_market_making", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_hummingbot__strategy__pure_market_making__pure_market_making) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "hummingbot.strategy.pure_market_making.pure_market_making")) {
      if (unlikely((PyDict_SetItemString(modules, "hummingbot.strategy.pure_market_making.pure_market_making", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_function_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1
 * import logging             # <<<<<<<<<<<<<<
 * from decimal import Decimal
 * from math import ceil, floor
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_logging, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_logging, __pyx_t_2) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":2
 * import logging
 * from decimal import Decimal             # <<<<<<<<<<<<<<
 * from math import ceil, floor
 * from typing import Dict, List, Optional
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_Decimal};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_decimal, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_Decimal};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 2, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":3
 * import logging
 * from decimal import Decimal
 * from math import ceil, floor             # <<<<<<<<<<<<<<
 * from typing import Dict, List, Optional
 * 
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ceil,__pyx_mstate_global->__pyx_n_u_floor};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_math, __pyx_imported_names, 2, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ceil,__pyx_mstate_global->__pyx_n_u_floor};
    for (__pyx_t_3=0; __pyx_t_3 < 2; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 3, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":4
 * from decimal import Decimal
 * from math import ceil, floor
 * from typing import Dict, List, Optional             # <<<<<<<<<<<<<<
 * 
 * import numpy as np
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_Dict,__pyx_mstate_global->__pyx_n_u_List,__pyx_mstate_global->__pyx_n_u_Optional};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_typing, __pyx_imported_names, 3, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_Dict,__pyx_mstate_global->__pyx_n_u_List,__pyx_mstate_global->__pyx_n_u_Optional};
    for (__pyx_t_3=0; __pyx_t_3 < 3; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 4, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":6
 * from typing import Dict, List, Optional
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * import pandas as pd
 * 
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_numpy, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_np, __pyx_t_2) < (0)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":7
 * 
 * import numpy as np
 * import pandas as pd             # <<<<<<<<<<<<<<
 * 
 * from hummingbot.connector.exchange_base import ExchangeBase
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pandas, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pd, __pyx_t_2) < (0)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":9
 * import pandas as pd
 * 
 * from hummingbot.connector.exchange_base import ExchangeBase             # <<<<<<<<<<<<<<
 * from hummingbot.connector.exchange_base cimport ExchangeBase
 * from hummingbot.core.clock cimport Clock
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ExchangeBase};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_hummingbot_connector_exchange_ba, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":12
 * from hummingbot.connector.exchange_base cimport ExchangeBase
 * from hummingbot.core.clock cimport Clock
 * from hummingbot.core.data_type.common import OrderType, PriceType, TradeType             # <<<<<<<<<<<<<<
 * from hummingbot.core.data_type.limit_order cimport LimitOrder
 * from hummingbot.core.data_type.limit_order import LimitOrder
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_OrderType,__pyx_mstate_global->__pyx_n_u_PriceType,__pyx_mstate_global->__pyx_n_u_TradeType};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_hummingbot_core_data_type_common, __pyx_imported_names, 3, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_OrderType,__pyx_mstate_global->__pyx_n_u_PriceType,__pyx_mstate_global->__pyx_n_u_TradeType};
    for (__pyx_t_3=0; __pyx_t_3 < 3; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 12, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 12, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":14
 * from hummingbot.core.data_type.common import OrderType, PriceType, TradeType
 * from hummingbot.core.data_type.limit_order cimport LimitOrder
 * from hummingbot.core.data_type.limit_order import LimitOrder             # <<<<<<<<<<<<<<
 * from hummingbot.core.network_iterator import NetworkStatus
 * from hummingbot.core.utils import map_df_to_str
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_LimitOrder};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_hummingbot_core_data_type_limit, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":15
 * from hummingbot.core.data_type.limit_order cimport LimitOrder
 * from hummingbot.core.data_type.limit_order import LimitOrder
 * from hummingbot.core.network_iterator import NetworkStatus             # <<<<<<<<<<<<<<
 * from hummingbot.core.utils import map_df_to_str
 * from hummingbot.strategy.asset_price_delegate cimport AssetPriceDelegate
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_NetworkStatus};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_hummingbot_core_network_iterator, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_NetworkStatus};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 15, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":16
 * from hummingbot.core.data_type.limit_order import LimitOrder
 * from hummingbot.core.network_iterator import NetworkStatus
 * from hummingbot.core.utils import map_df_to_str             # <<<<<<<<<<<<<<
 * from hummingbot.strategy.asset_price_delegate cimport AssetPriceDelegate
 * from hummingbot.strategy.asset_price_delegate import AssetPriceDelegate
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_map_df_to_str};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_hummingbot_core_utils, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_map_df_to_str};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 16, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 16, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":18
 * from hummingbot.core.utils import map_df_to_str
 * from hummingbot.strategy.asset_price_delegate cimport AssetPriceDelegate
 * from hummingbot.strategy.asset_price_delegate import AssetPriceDelegate             # <<<<<<<<<<<<<<
 * from hummingbot.strategy.hanging_orders_tracker import CreatedPairOfOrders, HangingOrdersTracker
 * from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_AssetPriceDelegate};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_hummingbot_strategy_asset_price, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":19
 * from hummingbot.strategy.asset_price_delegate cimport AssetPriceDelegate
 * from hummingbot.strategy.asset_price_delegate import AssetPriceDelegate
 * from hummingbot.strategy.hanging_orders_tracker import CreatedPairOfOrders, HangingOrdersTracker             # <<<<<<<<<<<<<<
 * from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
 * from hummingbot.strategy.order_book_asset_price_delegate cimport OrderBookAssetPriceDelegate
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_CreatedPairOfOrders,__pyx_mstate_global->__pyx_n_u_HangingOrdersTracker};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_hummingbot_strategy_hanging_orde, __pyx_imported_names, 2, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_CreatedPairOfOrders,__pyx_mstate_global->__pyx_n_u_HangingOrdersTracker};
    for (__pyx_t_3=0; __pyx_t_3 < 2; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 19, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 19, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":20
 * from hummingbot.strategy.asset_price_delegate import AssetPriceDelegate
 * from hummingbot.strategy.hanging_orders_tracker import CreatedPairOfOrders, HangingOrdersTracker
 * from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple             # <<<<<<<<<<<<<<
 * from hummingbot.strategy.order_book_asset_price_delegate cimport OrderBookAssetPriceDelegate
 * from hummingbot.strategy.strategy_base import StrategyBase
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_MarketTradingPairTuple};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_hummingbot_strategy_market_tradi, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_MarketTradingPairTuple};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 20, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 20, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":22
 * from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
 * from hummingbot.strategy.order_book_asset_price_delegate cimport OrderBookAssetPriceDelegate
 * from hummingbot.strategy.strategy_base import StrategyBase             # <<<<<<<<<<<<<<
 * from hummingbot.strategy.utils import order_age
 * from .data_types import PriceSize, Proposal
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_StrategyBase};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_hummingbot_strategy_strategy_bas, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":23
 * from hummingbot.strategy.order_book_asset_price_delegate cimport OrderBookAssetPriceDelegate
 * from hummingbot.strategy.strategy_base import StrategyBase
 * from hummingbot.strategy.utils import order_age             # <<<<<<<<<<<<<<
 * from .data_types import PriceSize, Proposal
 * from .inventory_cost_price_delegate import InventoryCostPriceDelegate
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_order_age};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_hummingbot_strategy_utils, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_order_age};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 23, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 23, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":24
 * from hummingbot.strategy.strategy_base import StrategyBase
 * from hummingbot.strategy.utils import order_age
 * from .data_types import PriceSize, Proposal             # <<<<<<<<<<<<<<
 * from .inventory_cost_price_delegate import InventoryCostPriceDelegate
 * from .inventory_skew_calculator cimport c_calculate_bid_ask_ratios_from_base_asset_ratio
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PriceSize,__pyx_mstate_global->__pyx_n_u_Proposal};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_data_types, __pyx_imported_names, 2, __pyx_mstate_global->__pyx_kp_u_hummingbot_strategy_pure_market_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PriceSize,__pyx_mstate_global->__pyx_n_u_Proposal};
    for (__pyx_t_3=0; __pyx_t_3 < 2; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 24, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 24, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":25
 * from hummingbot.strategy.utils import order_age
 * from .data_types import PriceSize, Proposal
 * from .inventory_cost_price_delegate import InventoryCostPriceDelegate             # <<<<<<<<<<<<<<
 * from .inventory_skew_calculator cimport c_calculate_bid_ask_ratios_from_base_asset_ratio
 * from .inventory_skew_calculator import calculate_total_order_size
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_InventoryCostPriceDelegate};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_inventory_cost_price_delegate, __pyx_imported_names, 1, __pyx_mstate_global->__pyx_kp_u_hummingbot_strategy_pure_market_3, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_InventoryCostPriceDelegate};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 25, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 25, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":27
 * from .inventory_cost_price_delegate import InventoryCostPriceDelegate
 * from .inventory_skew_calculator cimport c_calculate_bid_ask_ratios_from_base_asset_ratio
 * from .inventory_skew_calculator import calculate_total_order_size             # <<<<<<<<<<<<<<
 * from .pure_market_making_order_tracker import PureMarketMakingOrderTracker
 * from .moving_price_band import MovingPriceBand
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_calculate_total_order_size};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_inventory_skew_calculator, __pyx_imported_names, 1, __pyx_mstate_global->__pyx_kp_u_hummingbot_strategy_pure_market_4, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_calculate_total_order_size};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 27, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 27, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":28
 * from .inventory_skew_calculator cimport c_calculate_bid_ask_ratios_from_base_asset_ratio
 * from .inventory_skew_calculator import calculate_total_order_size
 * from .pure_market_making_order_tracker import PureMarketMakingOrderTracker             # <<<<<<<<<<<<<<
 * from .moving_price_band import MovingPriceBand
 * 
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PureMarketMakingOrderTracker};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pure_market_making_order_tracker, __pyx_imported_names, 1, __pyx_mstate_global->__pyx_kp_u_hummingbot_strategy_pure_market_5, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PureMarketMakingOrderTracker};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 28, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 28, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":29
 * from .inventory_skew_calculator import calculate_total_order_size
 * from .pure_market_making_order_tracker import PureMarketMakingOrderTracker
 * from .moving_price_band import MovingPriceBand             # <<<<<<<<<<<<<<
 * 
 * 
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_MovingPriceBand};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_moving_price_band, __pyx_imported_names, 1, __pyx_mstate_global->__pyx_kp_u_hummingbot_strategy_pure_market_6, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_MovingPriceBand};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 29, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 29, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":32
 * 
 * 
 * NaN = float("nan")             # <<<<<<<<<<<<<<
 * s_decimal_zero = Decimal(0)
 * s_decimal_neg_one = Decimal(-1)
*/
  __pyx_t_5 = __Pyx_PyUnicode_AsDouble(__pyx_mstate_global->__pyx_n_u_nan); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_5, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 32, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_NaN, __pyx_t_2) < (0)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":33
 * 
 * NaN = float("nan")
 * s_decimal_zero = Decimal(0)             # <<<<<<<<<<<<<<
 * s_decimal_neg_one = Decimal(-1)
 * pmm_logger = None
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_int_0};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_s_decimal_zero, __pyx_t_2) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":34
 * NaN = float("nan")
 * s_decimal_zero = Decimal(0)
 * s_decimal_neg_one = Decimal(-1)             # <<<<<<<<<<<<<<
 * pmm_logger = None
 * 
*/
  __pyx_t_6 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_mstate_global->__pyx_int_neg_1};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_s_decimal_neg_one, __pyx_t_2) < (0)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":35
 * s_decimal_zero = Decimal(0)
 * s_decimal_neg_one = Decimal(-1)
 * pmm_logger = None             # <<<<<<<<<<<<<<
 * 
 * 
*/
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pmm_logger, Py_None) < (0)) __PYX_ERR(0, 35, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":39
 * 
 * cdef class PureMarketMakingStrategy(StrategyBase):
 *     OPTION_LOG_CREATE_ORDER = 1 << 3             # <<<<<<<<<<<<<<
 *     OPTION_LOG_MAKER_ORDER_FILLED = 1 << 4
 *     OPTION_LOG_STATUS_REPORT = 1 << 5
*/
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_OPTION_LOG_CREATE_ORDER, __pyx_mstate_global->__pyx_int_8) < (0)) __PYX_ERR(0, 39, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":40
 * cdef class PureMarketMakingStrategy(StrategyBase):
 *     OPTION_LOG_CREATE_ORDER = 1 << 3
 *     OPTION_LOG_MAKER_ORDER_FILLED = 1 << 4             # <<<<<<<<<<<<<<
 *     OPTION_LOG_STATUS_REPORT = 1 << 5
 *     OPTION_LOG_ALL = 0x7fffffffffffffff
*/
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_OPTION_LOG_MAKER_ORDER_FILLED, __pyx_mstate_global->__pyx_int_16) < (0)) __PYX_ERR(0, 40, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":41
 *     OPTION_LOG_CREATE_ORDER = 1 << 3
 *     OPTION_LOG_MAKER_ORDER_FILLED = 1 << 4
 *     OPTION_LOG_STATUS_REPORT = 1 << 5             # <<<<<<<<<<<<<<
 *     OPTION_LOG_ALL = 0x7fffffffffffffff
 * 
*/
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_OPTION_LOG_STATUS_REPORT, __pyx_mstate_global->__pyx_int_32) < (0)) __PYX_ERR(0, 41, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":42
 *     OPTION_LOG_MAKER_ORDER_FILLED = 1 << 4
 *     OPTION_LOG_STATUS_REPORT = 1 << 5
 *     OPTION_LOG_ALL = 0x7fffffffffffffff             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
*/
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_OPTION_LOG_ALL, __pyx_mstate_global->__pyx_int_0x7fffffffffffffff) < (0)) __PYX_ERR(0, 42, __pyx_L1_error)

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":44
 *     OPTION_LOG_ALL = 0x7fffffffffffffff
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def logger(cls):
 *         global pmm_logger
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_1logger, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_logger, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_logger, __pyx_t_2) < (0)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject*)__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_logger, __pyx_t_4) < (0)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":57
 *                     order_amount: Decimal,
 *                     order_levels: int = 1,
 *                     order_level_spread: Decimal = s_decimal_zero,             # <<<<<<<<<<<<<<
 *                     order_level_amount: Decimal = s_decimal_zero,
 *                     order_refresh_time: float = 30.0,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_mstate_global->__pyx_k_ = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":58
 *                     order_levels: int = 1,
 *                     order_level_spread: Decimal = s_decimal_zero,
 *                     order_level_amount: Decimal = s_decimal_zero,             # <<<<<<<<<<<<<<
 *                     order_refresh_time: float = 30.0,
 *                     max_order_age: float = 1800.0,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_mstate_global->__pyx_k__2 = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":61
 *                     order_refresh_time: float = 30.0,
 *                     max_order_age: float = 1800.0,
 *                     order_refresh_tolerance_pct: Decimal = s_decimal_neg_one,             # <<<<<<<<<<<<<<
 *                     filled_order_delay: float = 60.0,
 *                     inventory_skew_enabled: bool = False,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_s_decimal_neg_one); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_mstate_global->__pyx_k__3 = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":64
 *                     filled_order_delay: float = 60.0,
 *                     inventory_skew_enabled: bool = False,
 *                     inventory_target_base_pct: Decimal = s_decimal_zero,             # <<<<<<<<<<<<<<
 *                     inventory_range_multiplier: Decimal = s_decimal_zero,
 *                     hanging_orders_enabled: bool = False,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_mstate_global->__pyx_k__4 = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":65
 *                     inventory_skew_enabled: bool = False,
 *                     inventory_target_base_pct: Decimal = s_decimal_zero,
 *                     inventory_range_multiplier: Decimal = s_decimal_zero,             # <<<<<<<<<<<<<<
 *                     hanging_orders_enabled: bool = False,
 *                     hanging_orders_cancel_pct: Decimal = Decimal("0.1"),
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_mstate_global->__pyx_k__5 = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":67
 *                     inventory_range_multiplier: Decimal = s_decimal_zero,
 *                     hanging_orders_enabled: bool = False,
 *                     hanging_orders_cancel_pct: Decimal = Decimal("0.1"),             # <<<<<<<<<<<<<<
 *                     order_optimization_enabled: bool = False,
 *                     ask_order_optimization_depth: Decimal = s_decimal_zero,
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_0_1};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_mstate_global->__pyx_k__6 = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":69
 *                     hanging_orders_cancel_pct: Decimal = Decimal("0.1"),
 *                     order_optimization_enabled: bool = False,
 *                     ask_order_optimization_depth: Decimal = s_decimal_zero,             # <<<<<<<<<<<<<<
 *                     bid_order_optimization_depth: Decimal = s_decimal_zero,
 *                     add_transaction_costs_to_orders: bool = False,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_mstate_global->__pyx_k__7 = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":70
 *                     order_optimization_enabled: bool = False,
 *                     ask_order_optimization_depth: Decimal = s_decimal_zero,
 *                     bid_order_optimization_depth: Decimal = s_decimal_zero,             # <<<<<<<<<<<<<<
 *                     add_transaction_costs_to_orders: bool = False,
 *                     asset_price_delegate: AssetPriceDelegate = None,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_mstate_global->__pyx_k__8 = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":76
 *                     price_type: str = "mid_price",
 *                     take_if_crossed: bool = False,
 *                     price_ceiling: Decimal = s_decimal_neg_one,             # <<<<<<<<<<<<<<
 *                     price_floor: Decimal = s_decimal_neg_one,
 *                     ping_pong_enabled: bool = False,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_s_decimal_neg_one); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_mstate_global->__pyx_k__9 = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":77
 *                     take_if_crossed: bool = False,
 *                     price_ceiling: Decimal = s_decimal_neg_one,
 *                     price_floor: Decimal = s_decimal_neg_one,             # <<<<<<<<<<<<<<
 *                     ping_pong_enabled: bool = False,
 *                     logging_options: int = OPTION_LOG_ALL,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_s_decimal_neg_one); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_mstate_global->__pyx_k__10 = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":79
 *                     price_floor: Decimal = s_decimal_neg_one,
 *                     ping_pong_enabled: bool = False,
 *                     logging_options: int = OPTION_LOG_ALL,             # <<<<<<<<<<<<<<
 *                     status_report_interval: float = 900,
 *                     minimum_spread: Decimal = Decimal(0),
*/
  __Pyx_GetNameInClass(__pyx_t_4, (PyObject*)__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_OPTION_LOG_ALL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_mstate_global->__pyx_k__11 = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":81
 *                     logging_options: int = OPTION_LOG_ALL,
 *                     status_report_interval: float = 900,
 *                     minimum_spread: Decimal = Decimal(0),             # <<<<<<<<<<<<<<
 *                     hb_app_notification: bool = False,
 *                     order_override: Dict[str, List[str]] = None,
*/
  __pyx_t_6 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_mstate_global->__pyx_int_0};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_mstate_global->__pyx_k__12 = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":57
 *                     order_amount: Decimal,
 *                     order_levels: int = 1,
 *                     order_level_spread: Decimal = s_decimal_zero,             # <<<<<<<<<<<<<<
 *                     order_level_amount: Decimal = s_decimal_zero,
 *                     order_refresh_time: float = 30.0,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":58
 *                     order_levels: int = 1,
 *                     order_level_spread: Decimal = s_decimal_zero,
 *                     order_level_amount: Decimal = s_decimal_zero,             # <<<<<<<<<<<<<<
 *                     order_refresh_time: float = 30.0,
 *                     max_order_age: float = 1800.0,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":61
 *                     order_refresh_time: float = 30.0,
 *                     max_order_age: float = 1800.0,
 *                     order_refresh_tolerance_pct: Decimal = s_decimal_neg_one,             # <<<<<<<<<<<<<<
 *                     filled_order_delay: float = 60.0,
 *                     inventory_skew_enabled: bool = False,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_s_decimal_neg_one); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":64
 *                     filled_order_delay: float = 60.0,
 *                     inventory_skew_enabled: bool = False,
 *                     inventory_target_base_pct: Decimal = s_decimal_zero,             # <<<<<<<<<<<<<<
 *                     inventory_range_multiplier: Decimal = s_decimal_zero,
 *                     hanging_orders_enabled: bool = False,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":65
 *                     inventory_skew_enabled: bool = False,
 *                     inventory_target_base_pct: Decimal = s_decimal_zero,
 *                     inventory_range_multiplier: Decimal = s_decimal_zero,             # <<<<<<<<<<<<<<
 *                     hanging_orders_enabled: bool = False,
 *                     hanging_orders_cancel_pct: Decimal = Decimal("0.1"),
*/
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":67
 *                     inventory_range_multiplier: Decimal = s_decimal_zero,
 *                     hanging_orders_enabled: bool = False,
 *                     hanging_orders_cancel_pct: Decimal = Decimal("0.1"),             # <<<<<<<<<<<<<<
 *                     order_optimization_enabled: bool = False,
 *                     ask_order_optimization_depth: Decimal = s_decimal_zero,
*/
  __pyx_t_11 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_mstate_global->__pyx_kp_u_0_1};
    __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":69
 *                     hanging_orders_cancel_pct: Decimal = Decimal("0.1"),
 *                     order_optimization_enabled: bool = False,
 *                     ask_order_optimization_depth: Decimal = s_decimal_zero,             # <<<<<<<<<<<<<<
 *                     bid_order_optimization_depth: Decimal = s_decimal_zero,
 *                     add_transaction_costs_to_orders: bool = False,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":70
 *                     order_optimization_enabled: bool = False,
 *                     ask_order_optimization_depth: Decimal = s_decimal_zero,
 *                     bid_order_optimization_depth: Decimal = s_decimal_zero,             # <<<<<<<<<<<<<<
 *                     add_transaction_costs_to_orders: bool = False,
 *                     asset_price_delegate: AssetPriceDelegate = None,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_s_decimal_zero); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":76
 *                     price_type: str = "mid_price",
 *                     take_if_crossed: bool = False,
 *                     price_ceiling: Decimal = s_decimal_neg_one,             # <<<<<<<<<<<<<<
 *                     price_floor: Decimal = s_decimal_neg_one,
 *                     ping_pong_enabled: bool = False,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_s_decimal_neg_one); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":77
 *                     take_if_crossed: bool = False,
 *                     price_ceiling: Decimal = s_decimal_neg_one,
 *                     price_floor: Decimal = s_decimal_neg_one,             # <<<<<<<<<<<<<<
 *                     ping_pong_enabled: bool = False,
 *                     logging_options: int = OPTION_LOG_ALL,
*/
  __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_s_decimal_neg_one); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":79
 *                     price_floor: Decimal = s_decimal_neg_one,
 *                     ping_pong_enabled: bool = False,
 *                     logging_options: int = OPTION_LOG_ALL,             # <<<<<<<<<<<<<<
 *                     status_report_interval: float = 900,
 *                     minimum_spread: Decimal = Decimal(0),
*/
  __Pyx_GetNameInClass(__pyx_t_15, (PyObject*)__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_OPTION_LOG_ALL); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":81
 *                     logging_options: int = OPTION_LOG_ALL,
 *                     status_report_interval: float = 900,
 *                     minimum_spread: Decimal = Decimal(0),             # <<<<<<<<<<<<<<
 *                     hb_app_notification: bool = False,
 *                     order_override: Dict[str, List[str]] = None,
*/
  __pyx_t_17 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_Decimal); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_17, __pyx_mstate_global->__pyx_int_0};
    __pyx_t_16 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_18, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
  }

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":51
 *         return pmm_logger
 * 
 *     def init_params(self,             # <<<<<<<<<<<<<<
 *                     market_info: MarketTradingPairTuple,
 *                     bid_spread: Decimal,
*/
  __pyx_t_18 = PyTuple_Pack(33, __pyx_mstate_global->__pyx_int_1, __pyx_t_4, __pyx_t_2, __pyx_mstate_global->__pyx_float_30_0, __pyx_mstate_global->__pyx_float_1800_0, __pyx_t_6, __pyx_mstate_global->__pyx_float_60_0, Py_False, __pyx_t_8, __pyx_t_9, Py_False, __pyx_t_10, Py_False, __pyx_t_12, __pyx_t_11, Py_False, Py_None, Py_None, __pyx_mstate_global->__pyx_n_u_mid_price, Py_False, __pyx_t_13, __pyx_t_14, Py_False, __pyx_t_15, __pyx_mstate_global->__pyx_int_900, __pyx_t_16, Py_False, Py_None, Py_False, Py_None, Py_None, Py_True, Py_None); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyDict_NewPresized(37); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_market_info, __pyx_mstate_global->__pyx_n_u_MarketTradingPairTuple) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_bid_spread, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_ask_spread, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_order_amount, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_order_levels, __pyx_mstate_global->__pyx_n_u_int) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_order_level_spread, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_order_level_amount, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_order_refresh_time, __pyx_mstate_global->__pyx_n_u_float) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_max_order_age, __pyx_mstate_global->__pyx_n_u_float) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_order_refresh_tolerance_pct, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_filled_order_delay, __pyx_mstate_global->__pyx_n_u_float) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_inventory_skew_enabled, __pyx_mstate_global->__pyx_n_u_bool) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_inventory_target_base_pct, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_inventory_range_multiplier, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_hanging_orders_enabled, __pyx_mstate_global->__pyx_n_u_bool) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_hanging_orders_cancel_pct, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_order_optimization_enabled, __pyx_mstate_global->__pyx_n_u_bool) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_ask_order_optimization_depth, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_bid_order_optimization_depth, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_add_transaction_costs_to_orders, __pyx_mstate_global->__pyx_n_u_bool) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_asset_price_delegate, __pyx_mstate_global->__pyx_n_u_AssetPriceDelegate) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_inventory_cost_price_delegate, __pyx_mstate_global->__pyx_n_u_InventoryCostPriceDelegate) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_price_type, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_take_if_crossed, __pyx_mstate_global->__pyx_n_u_bool) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_price_ceiling, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_price_floor, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_ping_pong_enabled, __pyx_mstate_global->__pyx_n_u_bool) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_logging_options, __pyx_mstate_global->__pyx_n_u_int) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_status_report_interval, __pyx_mstate_global->__pyx_n_u_float) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_minimum_spread, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_hb_app_notification, __pyx_mstate_global->__pyx_n_u_bool) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_order_override, __pyx_mstate_global->__pyx_kp_u_Dict_str_List_str) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_split_order_levels_enabled, __pyx_mstate_global->__pyx_n_u_bool) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_bid_order_level_spreads, __pyx_mstate_global->__pyx_kp_u_List_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_ask_order_level_spreads, __pyx_mstate_global->__pyx_kp_u_List_Decimal) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_should_wait_order_cancel_confirm, __pyx_mstate_global->__pyx_n_u_bool) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_moving_price_band, __pyx_mstate_global->__pyx_kp_u_Optional_MovingPriceBand) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_t_15 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_3init_params, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_init_pa, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_15);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_15, __pyx_t_18);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_15, __pyx_t_16);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_init_params, __pyx_t_15) < (0)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":149
 *         self.c_add_markets([market_info.market])
 * 
 *     def all_markets_ready(self):             # <<<<<<<<<<<<<<
 *         return all([market.ready for market in self._sb_markets])
 * 
*/
  __pyx_t_15 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_5all_markets_ready, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_all_mar, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9])); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_15);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_all_markets_ready, __pyx_t_15) < (0)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":413
 *         return self._moving_price_band
 * 
 *     def get_price(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         price_provider = self._asset_price_delegate or self._market_info
 *         if self._price_type is PriceType.LastOwnTrade:
*/
  __pyx_t_15 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyDict_SetItem(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 413, __pyx_L1_error)
  __pyx_t_16 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_7get_price, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_get_pri, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_16);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_16, __pyx_t_15);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_get_price, __pyx_t_16) < (0)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":427
 *         return price
 * 
 *     def get_mid_price(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_mid_price()
 * 
*/
  __pyx_t_16 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_Decimal) < (0)) __PYX_ERR(0, 427, __pyx_L1_error)
  __pyx_t_15 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_9get_mid_price, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_get_mid, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_15);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_15, __pyx_t_16);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_get_mid_price, __pyx_t_15) < (0)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":495
 *         self._inventory_cost_price_delegate = value
 * 
 *     def inventory_skew_stats_data_frame(self) -> Optional[pd.DataFrame]:             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
*/
  __pyx_t_15 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyDict_SetItem(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_kp_u_Optional_pd_DataFrame) < (0)) __PYX_ERR(0, 495, __pyx_L1_error)
  __pyx_t_16 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_11inventory_skew_stats_data_frame, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_invento, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_16);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_16, __pyx_t_15);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_inventory_skew_stats_data_frame, __pyx_t_16) < (0)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":552
 *         return inventory_skew_df
 * 
 *     def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self._market_info.get_mid_price()
*/
  __pyx_t_16 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_to_show_current_pct, __pyx_mstate_global->__pyx_n_u_bool) < (0)) __PYX_ERR(0, 552, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_kp_u_pd_DataFrame) < (0)) __PYX_ERR(0, 552, __pyx_L1_error)
  __pyx_t_15 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_13pure_mm_assets_df, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_pure_mm, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_15);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_15, __pyx_t_16);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_pure_mm_assets_df, __pyx_t_15) < (0)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":574
 *         return df
 * 
 *     def active_orders_df(self) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self.get_price()
*/
  __pyx_t_15 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyDict_SetItem(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_kp_u_pd_DataFrame) < (0)) __PYX_ERR(0, 574, __pyx_L1_error)
  __pyx_t_16 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_15active_orders_df, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_active, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_16);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_16, __pyx_t_15);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_active_orders_df, __pyx_t_16) < (0)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":614
 *         return pd.DataFrame(data=data, columns=columns)
 * 
 *     def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         markets_data = []
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"Ref Price ({self._price_type.name})"]
*/
  __pyx_t_16 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_market_trading_pair_tuples, __pyx_mstate_global->__pyx_kp_u_List_MarketTradingPairTuple) < (0)) __PYX_ERR(0, 614, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_kp_u_pd_DataFrame) < (0)) __PYX_ERR(0, 614, __pyx_L1_error)
  __pyx_t_15 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_17market_status_data_frame, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_market, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15])); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_15);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_15, __pyx_t_16);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_market_status_data_frame, __pyx_t_15) < (0)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":648
 *         return pd.DataFrame(data=markets_data, columns=markets_columns).replace(np.nan, '', regex=True)
 * 
 *     def format_status(self) -> str:             # <<<<<<<<<<<<<<
 *         if not self._all_markets_ready:
 *             return "Market connectors are not ready."
*/
  __pyx_t_15 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyDict_SetItem(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 648, __pyx_L1_error)
  __pyx_t_16 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_19format_status, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_format, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16])); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_16);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_16, __pyx_t_15);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_format_status, __pyx_t_16) < (0)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":688
 *     # The following exposed Python functions are meant for unit tests
 *     # ---------------------------------------------------------------
 *     def execute_orders_proposal(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_execute_orders_proposal(proposal)
 * 
*/
  __pyx_t_16 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_proposal, __pyx_mstate_global->__pyx_n_u_Proposal) < (0)) __PYX_ERR(0, 688, __pyx_L1_error)
  __pyx_t_15 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_21execute_orders_proposal, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_execute, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[17])); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_15);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_15, __pyx_t_16);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_execute_orders_proposal, __pyx_t_15) < (0)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":691
 *         return self.c_execute_orders_proposal(proposal)
 * 
 *     def cancel_order(self, order_id: str):             # <<<<<<<<<<<<<<
 *         return self.c_cancel_order(self._market_info, order_id)
 * 
*/
  __pyx_t_15 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyDict_SetItem(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_order_id, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 691, __pyx_L1_error)
  __pyx_t_16 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_23cancel_order, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_cancel, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[18])); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_16);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_16, __pyx_t_15);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_cancel_order, __pyx_t_16) < (0)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":920
 *             sell.size = size
 * 
 *     def adjusted_available_balance_for_orders_budget_constrain(self):             # <<<<<<<<<<<<<<
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
 *         non_hanging = []
*/
  __pyx_t_16 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_25adjusted_available_balance_for_orders_budget_constrain, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_adjuste, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19])); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_16);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_adjusted_available_balance_for_o, __pyx_t_16) < (0)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1310
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)
 * 
 *     def notify_hb_app(self, msg: str):             # <<<<<<<<<<<<<<
 *         if self._hb_app_notification:
 *             super().notify_hb_app(msg)
*/
  __pyx_t_16 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyDict_SetItem(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_msg, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __pyx_t_15 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_27notify_hb_app, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_notify, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20])); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_15);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_15, __pyx_t_16);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_notify_hb_app, __pyx_t_15) < (0)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1314
 *             super().notify_hb_app(msg)
 * 
 *     def get_price_type(self, price_type_str: str) -> PriceType:             # <<<<<<<<<<<<<<
 *         if price_type_str == "mid_price":
 *             return PriceType.MidPrice
*/
  __pyx_t_15 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyDict_SetItem(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_price_type_str, __pyx_mstate_global->__pyx_n_u_str) < (0)) __PYX_ERR(0, 1314, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_return, __pyx_mstate_global->__pyx_n_u_PriceType) < (0)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __pyx_t_16 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_29get_price_type, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy_get_pri_2, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[21])); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_16);
  #endif
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_16, __pyx_t_15);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_18pure_market_making_18pure_market_making_PureMarketMakingStrategy, __pyx_mstate_global->__pyx_n_u_get_price_type, __pyx_t_16) < (0)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_16 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_31__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy___reduc, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22])); if (unlikely(!__pyx_t_16)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_16);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_16) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
*/
  __pyx_t_16 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_18pure_market_making_18pure_market_making_24PureMarketMakingStrategy_33__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PureMarketMakingStrategy___setst, NULL, __pyx_mstate_global->__pyx_n_u_hummingbot_strategy_pure_market, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23])); if (unlikely(!__pyx_t_16)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_16);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_16) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "hummingbot/strategy/pure_market_making/pure_market_making.pyx":1
 * import logging             # <<<<<<<<<<<<<<
 * from decimal import Decimal
 * from math import ceil, floor
*/
  __pyx_t_16 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_16) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init hummingbot.strategy.pure_market_making.pure_market_making", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init hummingbot.strategy.pure_market_making.pure_market_making");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_builtin_all = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_all); if (!__pyx_builtin_all) __PYX_ERR(0, 150, __pyx_L1_error)
  __pyx_builtin_round = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_round); if (!__pyx_builtin_round) __PYX_ERR(0, 565, __pyx_L1_error)
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_max); if (!__pyx_builtin_max) __PYX_ERR(0, 667, __pyx_L1_error)
  __pyx_builtin_sorted = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_sorted); if (!__pyx_builtin_sorted) __PYX_ERR(0, 1010, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 1012, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 1171, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_super); if (!__pyx_builtin_super) __PYX_ERR(0, 1312, __pyx_L1_error)

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);
  __Pyx_RefNannyFinishContext();
  return 0;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 10; } index[] = {{3},{3},{1},{3},{3},{1},{3},{3},{12},{13},{9},{17},{8},{8},{21},{11},{9},{15},{20},{8},{24},{25},{26},{27},{7},{74},{15},{13},{28},{16},{17},{18},{21},{19},{22},{32},{10},{61},{25},{4},{25},{22},{14},{31},{9},{57},{20},{11},{20},{8},{14},{13},{31},{16},{136},{1},{3},{0},{1},{4},{3},{1},{1},{1},{2},{3},{2},{1},{1},{30},{30},{12},{7},{6},{8},{2},{29},{49},{68},{64},{71},{56},{61},{11},{9},{39},{34},{12},{64},{13},{14},{3},{18},{3},{7},{7},{3},{9},{19},{6},{9},{7},{4},{8},{12},{20},{4},{13},{26},{5},{12},{9},{5},{10},{4},{6},{22},{8},{15},{3},{13},{14},{23},{29},{24},{8},{9},{5},{9},{9},{8},{28},{24},{42},{44},{41},{79},{42},{69},{81},{67},{49},{37},{48},{38},{38},{34},{39},{36},{56},{31},{49},{38},{42},{20},{4},{4},{6},{12},{9},{9},{4},{14},{25},{13},{16},{34},{20},{57},{31},{54},{3},{3},{17},{22},{10},{6},{11},{5},{23},{28},{9},{9},{10},{20},{9},{18},{22},{23},{15},{10},{17},{16},{16},{16},{12},{13},{10},{10},{8},{8},{14},{23},{28},{9},{9},{10},{4},{3},{4},{26},{12},{24},{35},{4},{27},{28},{26},{17},{15},{18},{5},{3},{7},{6},{31},{17},{6},{4},{10},{7},{2},{8},{8},{12},{7},{9},{23},{18},{16},{5},{5},{6},{13},{10},{8},{7},{3},{9},{21},{11},{16},{20},{13},{9},{17},{20},{14},{12},{4},{17},{25},{26},{22},{22},{19},{6},{15},{21},{34},{32},{37},{32},{21},{40},{42},{45},{57},{33},{25},{3},{17},{5},{4},{11},{5},{3},{14},{29},{26},{25},{17},{22},{31},{25},{6},{13},{16},{6},{29},{26},{5},{4},{3},{8},{20},{10},{3},{5},{21},{12},{4},{5},{14},{6},{7},{15},{14},{20},{7},{8},{8},{13},{6},{12},{11},{28},{28},{24},{26},{15},{12},{10},{4},{3},{13},{9},{14},{10},{17},{25},{3},{8},{4},{3},{14},{15},{4},{8},{11},{9},{13},{28},{2},{5},{1},{5},{9},{12},{8},{18},{18},{12},{26},{14},{18},{27},{1},{6},{2},{7},{17},{10},{3},{5},{23},{13},{17},{11},{15},{14},{10},{14},{24},{12},{8},{32},{17},{11},{14},{12},{8},{11},{18},{17},{17},{13},{14},{11},{5},{10},{17},{13},{9},{5},{15},{7},{12},{6},{7},{5},{1},{17},{14},{4},{4},{10},{5},{4},{12},{10},{12},{19},{37},{4},{4},{6},{5},{26},{6},{22},{3},{31},{8},{5},{6},{15},{18},{27},{17},{19},{8},{5},{19},{9},{14},{16},{22},{11},{20},{10},{12},{6},{4},{17},{6},{45},{5},{6},{7},{13},{1},{3},{99},{17},{6},{106},{557},{23},{287},{6},{691},{133},{30},{23},{394},{2},{9},{20},{121},{2},{427},{2},{447},{15},{6}};
    #if (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (4964 bytes) */
const char* const cstring = "BZh91AY&SY\324\247\303#\000\002\317\177\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\376\300@@@@@@@@@@@@\000@\000`\024\031\363\327\213\331\356\307=w2\253JitI\315\r\354\351s\331\325\315V\022+-5F\205\206*a6\262\004\357T\001\364o\227\241\241\021\004\3314\020\302i\223OP\302a4\324\360\200D\364\00352\236\214\204\362\215\246I\342&\236\247\352h\321\246\210\366\222`\200J\010\t\210A\023D\3614\324\237\252~Tz\232\007\3525\036\243!\35246\220\32044\000\000\001\240\003&\200\320\365\001\240\024\311M \310=@\r\031\250\036\240\014\215\006\200\000\000\000\000\003M\000\3204\017I\264\324\365=A!\"h)\220SM\222\236(\362z\247\352O$\0312\031\032\000\000212\003 4\0322\031\014\2324\310\000\303\246\214\004\304\323\004\300\004\364\232a40\000\0022`\230\23112h\300L\000\230\023\000F\206\001\"@\223\010&\230\206\201M\251\203 \215&\232z\200\001\2404\000\000\000\000\000\000\000h k;\240_X\005\374!\225de\213YV\204A C\240\314P\201\r\211\215\2558w\016\342E'e\205\267l\210L\231\244P\324\005p\227\002\n\\\220\\\024\231\2030\014\323i\t1\266\332\033I\261\261\261&\232lm%v\242\376-\033dc\226z\244\262w\2020hlCh\033D\323i61\215\2421\013\251`B\260\024(u\230\315\372\351B\3415\317\035v+\264<\026\242\014\032\0062\3620\262\260\316\210\253A.\230\277\024d\030\206H\316\266\303|R,\303E\\\201\005j\244\"KUV\205\357t\264R\325\324#*s\272#\276)f\266\305J\205&-\025$X+S\006t\221):\024\242\211\030\343(I`\241\232JV\312\364\313\236\254\252\023\235\234\204\231\255v\214\031nQ\362\210n\224%\206\010\242\"i\352\023\300\3578\355\226\304\221r-r\352\251\252\224\255N\313\311\317I\244q\3518\370\215\267\212r6i\214kS3\202\037h4/\021\372\236\255|\200\314\301\364\020\366\032\271x<\r}\037\211\330\rk\\1\177\345\r\231bL\233\254\356\032K!f\303\222?g\257\272\246\353\236w\254\177h\252\037MO\351*\253\230\222+g(\233\214\301i\352A\336o\266\346\371\027\365\330\272*\314B\324k+A\225.\242\361}\027u\021\023\034\"~\341(i\2422N\354TR4\261C,\025&c>\243\267\337pw\0174;\335\240\360\235\340\273\277r\271\276\262\343\306\275\005\343\231J\304\332\2237""\271W\233\264\362\260N\023\2378\260R\222\226fb=\223\013\203;\320\3102\242\374\202\301d3<\333\\=\303\275\275\n$\322f\354\314\315\204\\\317w\311\252zK\351l\341\365<\206U\266\207\026\346\340\320\316\366\nH\256\225\364f\223E\2314\354p\266\377\235\342\272P*\300{y\366}\010w\327\333X\223z\226`\021^\307q\340\371\034\031\255B|f\r\341\272\031\220$\337\377\313\021\344\342\370\233j\203\200\240\223\347\272O\036\006\004\014\031s\237\006\266_\273\265\341\215\225&%\272\213\277\334\024\360S\204W\322{`\372\222[\022 -\240f\320V\021\341\303\253\227\036F\233\224\010\3569\224U\010\224\033\315\"PrdO\301\033u\243\014\231\202#W\315K\1775\370\225#\027\250 \024)\2318\207\021\364\326\251\007\373\267\324c\223\200A3\2777\203\260\364\216K\244\313\2337\266NR\360x\013cGh\261\027\016\010\212J\232p\316r\025\275N\023\352\364\353\231\236\245\2654\271\340;\232\033$\211\023_#\311Nj\361\277\010iJ\031gr\371\247\036]\233\031\315%n\3348\"c\210\346#X\235\353\251G\262\007\3278\271\317e\371\277?\352\361S~`\337\271\0062\343a\343\316\306\300dJ\024\014\026|\254)\365\305\016\330\212B\247n\00038x\017\3621/\334d3nn\366J@z\355I\236/\210\273\316\303\353Y\224\025f2K\210%\220\247\220B\225bY\031\3120\302m\304\207\024\210NQ\021\346Bh\355\td[\211\206N\205\360\371\353\r|\272\346g\371n\370\277k\263\244\003X\327P\311\310h\2104\376v%\237[\\\340\356O\021\372\014i\273\022`pA\022\337UYHo\233#_i!\343\037\302;9\3629\223K\363c\233\2557r\357<\251\037\304:\345\213\236\237\260\271\303\320\233\246g\237\247|\247\315#\255\010F\002\2738qdA\220\347%\333\370\304\3131+r{X\302\212Q\347\212\014\244\300\030\030\014\211\321\323\307\033n\256\233\362\255QT\220\313$\210:\370'\327B{\027\351\347\275w\316n\343\017'-t\344t\273\027\225\262\216\003\274\316\001\226\270T\232\034\365eK\217(a\305\303\357\314Vr\2461\030\n\224\335\303\002`\023\037\007`\020R\203\360\216\000\334K\360\346\312\2640\000fK\3223Q\375\021\3624\234\320\250\327\300\305\"Qcy\3413\216\221_\034X\243\263\2643\246'nZ\373JD9\326\305\213\233\345;\232<GF\232\030A\016a\341\302\204\320\200\032\026`""\217\216\263\3611\330n\027\363B\020\303{3K\034\201\014s\302\020\345V\311\314\227/\004\nl* \037\206\0048\223\247Q\233<\302\211\222\305\213\2137\001B\2057\013\342F\372\210f\205\256)\017\205\004\347'\217{~\327'A\341\245Cy\252o=\246\336\177[\275f\n\244\270\275\037\016\2534\221cr|^\200\262\013~\0338\035\261n`\246\364\201\240\211$7\004^\017\t\310\221\361!\365Fsx\265O\032\370\234\023Dn\226\215\211r%4+\201`K\352\200\222s\026\300\364\033L\240\024\205\264\002(Q\337\366\352\300?4N\314\346\375\365\311O\007\233\244\241hQ\216\306_h\247!BM\026\1773\374\246\213t\214D6\016U\t\2612\244\036\367\0165\246V\317_\324I\221\213\207p~\272\266\220\344+`\333\r\325!\2148M\022\354\024\034;\352Ep+N<\020\317<\354Ye\244\336\221\t\252xB%\305QHk\031\n\254\272%\002\030\214\234d\216\350HH\307f6\336\000\213\"H\204\305b\321\034\210\252\3315\347\025\332\351\225Aa\276\271\310h\347j\332\253\n\346ne\n\031\252vp\031\211;\227g\3031N\344\206\031\014\347\rC#\"\262\264\333\025Wd\331\203i\332;I\244\310\216\333\325\247\n\303\321aG\035\330Q\246M2\222\275\261\024\026\tQ1\225&\2334\336\\%\212\330f\271\215\344\326Y\303XSX\021\032f\216n$\215fqa\254\235ZZF9\333F\3702\327\177G\177\373\361\306\317G:\324\272\030\031\266A\010\2001\315\256{\340S\314k\215\333\343&\311\206\017\277w\033D\230\255i \266L\313\325\026\005\334\317\333\255\251qh\247\250\333c\030\322\032Hj\177\275\365\272\014\225h\325\231<3\244\203\262\212\337\224\344\351\360\317\207\325\213\221w\2444\261\271k\r\000\306\315\244\356\204B\221\321P\206,\302\366n5\032\213\266AA\334w^c\232:\320\3003\226)\253f\001\332|\231\034\342\365\246\264X\3366!4!\266\300\345\016Y\350\277\204\341@_\277\212\273\354w^\3568#\232\004S\323(t`\265\037\005f\021\311\304p+\2310\211^r\000\210[\002:d\222\220\236 \312\261\200\014d<\231\023\032&\330D'3\244\353\353:\317,\367G\223\270q_\013\022ry\320\250\262\232\250ch%c\006\261\017\010\332@39\260f@\257BF\3368\220\371\334\331\341[\0272+\347\3059x\314k\232\237\002=a\326\206\001;\263\240 \274\036\367\241\232\315\333|\276O.8x\273\277\275\25465\300\206K""\206\024\373\\\005\335\234\215)\262\230\017\247\0041\203a\006\332\033{\231<\022\020\317@\241P\307g\016\257\022\313n\343\227:.j\344\256\271\336Y\357\251\030\275:e(\223\201\301\r\266\332S,\0368w\266\221\004\006H\241\223Y\247@\310\230t\236\r\375\327\374\243\2033s\006\365\032\232\310Z\252N;\026R\252\230\332#\251\277\013\211\347\200\324e\356=K\"\306Vw\002&\2606:\r\0268\000\022J\016\240 \323\354D:\2254S\210\254\245\006&\2043\207\236@\224\244Z\r\250y\313\253\300a\003*B\021\326\242\021\204b\304\030\210\021\005\213t)E0w\362Dd\006\265Z\302\236\300\021Y\230\213\275\"1{\215\205\372=\274#\321\223\334[\260VI#L\272\232M\244\036\244\325\351i\213\271\255\235Z1\344\233\033\033tH\344\272&\265H\264\215\266\t\235Z0\256)\0312!\025M]t\261\213\3477\224d\350\025\244O27\264\032\333\024\2734\255\2727\257\2776Z-\264\364\016\025\335\207\026\375_\020\332z\265\034\347$\224\235\367\206G\014\370k\345U\314\326U\025\307\013wE\025\311 \275\225\211\241\343<AUT\025\026\273\005\006\231\026`\363\024\222\246\032\013R\273\341xi\271N\204\311R\222\036\201\241\266\004T\346\203&W\000\272\206\265ZH\213m\201~\030\272'0\306\206\236\255|\n\214\326D\2544\206\014#\021\301\255\253\\\3261aM\371\271\t6\232\300y\352\300d\250\010\312\2116w\252B\332\243>-\203\274\032t\215\026\217\217M\346\365\303 \n1\220Q\214cHd\220\203B\203\002\224R\363\211\033\312\223lE\035QQAd\246\\v\t\323\275p7\205\257\024.\373{c]\341\2261n\243\250\352\352\273\266`F\177{\344\224\331\312\353'\007xrq=0\256\236\177%:Z\334\373+Ut\303ev\361\3142L\306q\222\327\352\340\nM\010.j\346!i\234\205!\006\036\030\017\003m\311_\034\212o\023;\271\206\2221\315\r\215\2233\021\210\261/oP\321\203I7A\224iBH7\030\356\346\230\025h7\216\227uH,\027M\030M\360mTc\025N\207\343\223\354 %\205\274\366l\312\244\335\260\356\342\315s\030\252wb\210\t\334~\244\227G\261;\343\033\224\351U\031A\3137z\014\374\344`\024:\t\362\013T\240\200\322\212\233\020\317^\325\273\255@o7\031\217\006H\214\366\263\033'G\236s7\007,?`\320K\204\326HH\2464$1*)\361\325Wj\002\326\223\255\345\247S\223y\243=\265\024""\320m\211\233cXp\311\001;\253\300$\325\025j\017\224\362\316\366\272\360\206\323R\373Y(r\342\330t\205k\025\354i\022\017\002\316R\260o3\362\2353\222\356'\013>\347\004\014\201\2469\304\t\215\205$%B\326X\020X\230\223\233\302\357r7\0039Z\2158u\334teDu\023p\205\232\245\335\213\257\177\021Y#o\007\034\250\304\327L\"\217Y&*\216m\022/\010AaS\216-t\311\325\255L\255^\025\347\247\"\316;\336j\267\346T\251\224\004v\006V\006\016\007\266\3042\312\371[\223\206\210N\322\273\243\001N\232s\222\206\347\254\032\016\000\354\247\001\310\262\312\353W\230\206\345\253:\033\032FH\323\n\\\365\3254\207\255\273\355bf\232\242j\374\240\227W=\326X\014\013\232\233\212\250\017\006\333\320\030\227F\254.q\346\200\321\234\266\317N\360\331\006\335\032\2552\367\211\202\212\273Z\3258OGwj\306\031\235\247\n\032\033\033\030D\320JY\344\302\333[\000\216k\020m\030a{\327Yb\021\323\\\025\313\307\240o\265\2016\222\006\304\221x\r\241\r\023jC6\004\340\3651\264\255`\212\256L\214-\211\210O-\330?u\226\214wh\034\340\264\014llm3D:H\220\201\264\336\3357\352\263\030+\274\227\327\320c\244-MsKkZ\341\032E\243\233\207:T9\247N\214\340\266\350rOrm\323\211\035\311\250(h\231\262CW\215d\354\344\230\327l\322\271\231\234V\271\026\326h\325\246!\215\310E\013n\204U\356\210P\3274L\222E\326D\000\031\305|\356&\371\321\301\210X\2552\305\216\0003HO>r\336\210\302\361j@vf\025\334\250/j\241.\004m\303bFxQ\301\036\002KMZ\265k\332\315\213\273\242ERm\004\274z4,\374hX\361NH\321n\314\242\024\273\226*\260\333\364\261\256Q8\204\206\251\232\243\365%\371\025\021\266\023J\250G\321P)\355\254l\265D\267\262Z\007m\334\330<\004\375\320\276\360\372\214\323\215L\317\210\364{\032sU\211\027\311\352\263\264\275\331\337[\340\370YO\217\013Z\377#\030c\274m\263*\3178U\205l\313[*3f[!p0m6\215\320\213\261\312\240A\331\2530l\020\306\332)\214r\272\006\025\024Lh\372\235\340\330\372U8\nCI\004!F\347\346\274\302\336\306\30660\020\323\000-6\210\024\002\014Ps\027M\365\365\276\334\215\343\014;S\273$\035U\324\254W\r:\t\264g\2056?H\220\3201\214@_\252vU!\214\260\306\204\223\216\232\351:""\260rfl=X\\\344f\n\355\325=^\232 \226i\005\343Bm\003\244\024\2340\234@\007!\313\304o9V\376l\330kxy\2627^\354{\354\240\371\320\350\347\231\265\2232\216\355\276\245\356P\330\360\370\335\364\314\245P\213#;JV\346\252XCo\265\255N{\033be\262\003\032\302\253\254\034t\361\232\230\253I\222\n\200\252\202\234\272k+&\222,\352F03W\317\254P\034H\316[\\w\312\246\330t\327Xc[\210\316\261\231\266\271\300\021U1y\022Qm\026VGH\274\005\304\022\326\025EB\312T\251F=\231C\245\020\352[\320\337\257\224\351\262\332h\264U]yz\263\274X\211\317\001\341TW\203\016\221h\355\254P\t\223\\\200M\336G\320\340\3136pgV\033\023}\324\205]\220\321\200gL\031z\2248\0240x\210\325\216\320\"E\273\351\246\311L\256\267\364\370\263Z\200v1J\330\346\223ERm\206\344\341\255\0321\226b\2616\220\213\\Q\244@!:\332i\024\232\3117\333\240n\324{hVyn{\006\203\007n\300\345t\272\022\311\252&\233K4\263\322\022#f\255+\025\343\340\203\022\354\306\n \003\203\n\263\213k\344\235\236fB\255\241(\235\302#\020\3306\304\332cH\242\3545\032T\363\344\275\003%\364\231\225v\223F\224\221\213\363\007?<\016\205\340\377\206\224r\007\234nA=(\2454\360v\303\314\273\207\235\001\004.l\334\\{\244\221\234\026\356,\217\021b\257\261$s\241F\263\3024RIy\024\346\216\002\206\320\362\260\361ssw:\367\243\247\314\330\033\036\311\244e4\034\200\014!\304\013yS\305\213\235F\301\223\336\314\314%\274\224\224\274o\211\017\252\375]\343\rD\020zcGS\233K\035\027y\333vX\327\326\022\274\341:\006\222X\272P\351%Q\310\037R\246l\323\242\\\255\305\247\321k=8\355]\241\2666\005i\224\005!\235\325\314\221\275\235|\223\335\311\336\363\243\303\222\313\253\325\252\317\306\355\214\364\375C\316hM/M/\266z\307\315`\317\341\010\303\256\327\317\245[]6}\366\1774\035\003;\014\354\372\321\017\373\263\364\373\223\322\037\331\210)\226\267\224z\207\031\362q\004\006e\366,\300c\322`\010\374\342Yn\366\020\005}D\237\272\031\t#\n\367\205\024j \026\311u\317\007\033\324\275\226\324\271\006\352\366?%`_\337\2312u\267\342\006\007\223\222Y\034`c]\226\320X\264\013\261\247yfTy\010\223*\002\002\351#!T\030`\3150\276""\334\346eU\237\200U\325\227\264\353\205\342\344\200cg\370\271\203\020N\250\353\334\376\206\2660\343\301\205e\014\262\r\001\030\\5TAW\206;\342\326\336o\t\255\017\321\322\214i<Qgq\212D;\227\264U\244\244Q\357\317x\232\034\260\364t;\200i)\205\004\321@\210)\214\354\364\004\331\236\177\275n\225\237\177\236\004\320\254\336\005W\240V=\2106Mef\244\341\226\016\274!y\350t\266}\211Xkn\345\377\036\233(9\224wN\201\253b\227\337+*\365v\026}\360\263*\236\\\275U&\312\370\327\322v\001\250n\n\374\250Fa<\037Z\306G\260\251\264\255\325P\357\"\027\270\312\266\265<[\325\262\321C,\330\006tn\360T\326\254./\336:\r\217\217\320V\326B\032\306 \350\215\372UB\2137\274\320Z\354od\215J\225\235%\25137\353)\347\r^!\371\n\352\tm~\233\234\032`y\214Z\007\277\336U\263]\236n!\323\372\333_\301\237\305\342\026n\345\243\336M\201\\W\030\241\301\243f\322\215i\362\335;'j\300\206\306j1\305r\275\253\213L\254\276\027\022\255\307.~\3719~E\003\375\022\226\340\354\225Zz\0138_\355\003\031\232\276\204\246\023\360\343\325*n\352y\250\250\320|\365\016\217\237\312\252Y\002\216\257\315\336\354\377R&\306\252X\356\003\272\233\215\237\215\035\355N\352N\310t\374i\217\332Y\366\244\211p\250\002*v|\272\330~;\005\346}\240\025dps\236\243\270\0105\315\032\005\035M?\024\006R\003\343Xo\310 @\321/\220\"\023\301{Gh\t\272q)~\335\337\027\030\ru\373\206\252\317\214\316\245\370\245v\311h\014\021,\261\373\303\000m\260\215@\353XO+b[\356\357\257y\266\360\206\323{m\343\320\236\337\206\025\360\357\272\324]p\365\310\335\226\362%R}s\246\345\363\335rT\317v\245\235L|\3606\275\246\003\000j\027\203\2573\300e\312C\256\\\005\007\335\365\256<\205w \271x\005\221\217\320\370ctn\261+j\227\203O\305\260\001\336\327\233Q$4H\245A\272\263\322\021\003@1\023f\234f\322\026J)\007\232\021\274!.:\315\352\316Vqg\234\352J{\245\313\367\253\224\355h~\010\221\343o\037\255\245\232\335p\376\005\207yq\355\266GL\370\340Z\013\006\355\254\240\030\313*I\315\034\023\003d\327\250\207I\312\373Po8\234VI\202\256\324\233T\010\222F\006\315U\\\326\367\251\254\355T\356t\301\260\347\243\266W""\217Z\254\326VLn\325yV\3278\260\032\301\233\035Up\367\223\020\2775\211S\223t\251(\2251\025\236tG\374]\311\024\341BCR\237\014\214";
    PyObject *data = __Pyx_DecompressString(cstring, 4964, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (4962 bytes) */
const char* const cstring = "x\332\255XK[\324\350\266\336\270\331\275\351n\334\202b\253\355-\205 `ki\341\255\245=\332\305E\305\346\"\027\265\325\326tHR\020I%E\222\002\312s\372\331{\310\260\206\031f\230a\2065\254!\303oXC~\302\371\t\347]\337\227\024\251\002\334\366~\216\017\326w\277\254\265\336\365\256\365\345V6w+{7\227\273u+\233\033\034\315\216\016f\357\024\362E\273ly\322p^\3738\222\324\347\035cuD\222\362\256\253{\356X~S1Le\305\324\245q\305T,U\037\327]\017\203\353\274\0347\264\221\2544A\375\246a\255J\363\216\246;c\322\360\2104\341\350\212G]\023e\307\321\261\357`Ry\245\230e]\032\2364T\357\235\3539\327\245\031\303\345\265\367\357\007\237\215\r\316\216\r.=S\254UZ[T\326uG\032\177\371F\262ic\251\265\177ijf&\036\030i\033Z)W\216\030qu\323l\016MOJ7\244i\313\360\014\305\224\014k\023\327\263\235\212Tr\014U\227\014W\262l\017\013<ik\315\200\002\232\023d\325\206\354\005\010Xv\370<E\365\214M=;\335\334a\021g\352\\\256I]5\212\212\371\2367f\025g]\367\226\035E\303\205^(\206\263\\.\231\372\373\33161g\017\2127{@\254\366\036\273\320\354l\221\361@\027&\212{H\252mY\272\212\013C\002\010B\322\302jZ%+\3053\334\261\270l\037\237\263\241N\276\007\264J\312\365 \223.\246\225t\247hx\236\256a\033\314\023\272\211\265\317o\340f\347lK\237/y\206m)\346\273Y{\223\264A:\037W,\355}s\240\244e'\025Oy\342(E\375=G\226\004\234\226\241\372A\321\202\346Wt\323\336\222\212\206e\024\313E\311-\321\375\244aI\000\321\035{\241\320b\017s\271MeU78NU\305\"iVt\t\3131\352\255)V<\245`\332\266\203\253\277\300\274\033%\033\223\035\275ho\352\232\264\250\027$~Mi8U\23554^\033YV\234U(d0.c\230/\333\036\260\025\373\316K\313\321U{\3252>a;\2012\257R\322%\200\237nu\355\3325\351u~qnz\356\351\0225\342\372\230\264d\027\365X\341\373\246\210\255\207\235l\207wj\366\226%)\036\204\301d\254\260\274lb\351\330JE\245B2k\204N\307.\273\000\266n57\3066\226\356m\331\316\2728\244l\271\036y~v\004^\362\215\204\177\377=\366\360\017(wX\372Y\272.e\037K\212\273\036\233\224\316\035^\202\\\327\205^F\306$i\305\320>7\232 \327\315j\206K\347\350\026\347\231\202a\232\000\317\252:\"\255)da\\Q""\265\213p\024O7+\311\360Z\271\010\253\257\256\330^\026\332S<}\265\222-\301\035e!\215,$\316j\000\220L:v\277pA\253\223\313\002\023\232n\352\253X\362\247\367p\327\365-YUL\265l*\350\370\302\365\007\273d\256)\031kT\270\321\027nS\344\216\025\213\260\002\327\332_w3Yw\363\340\272\233\207\335\250\262M\336\026\353\336p\205\2514\253\\,U\262\320\224\243g\213e\0234\3528\200X\001!\003\260\364l\311(\226l\307KO+\027\025o\355\300\214\264\253\203\252\nYY'\025\272)G\005\3347u\307\023\253\024\351E\305[\263-\311^\371\010\037\220\n\300n\311P\327\271s\357s\235\233}(\\\313\265\313\216\252?\312\257\352<\254q\020N\306F\005\357R,\213C\032\"\332x\27121?775\261<5\311\303\230\256\021W\317\027\004\251L\200\201\354b\363\2721\301S8\233\332V\327\310\265\222r\\q\3658\372\210\245\313\302~\323sO\346\233\261b\0020ki\264\\nfzvzyFq\275\371-\213\342\206NuQ\201\206\314\031\003T\313\367\246\370\"\234\375\360 \223\320T\033\333\316)ss\302\347\227<\2043w\376\305\362\364\374\234<3\377T\316\317\314\244Z\023\213S\371\345)y~qrj1\325=\233\377ejQ\364\312O\246gf\246&S\203K\313\371\345\227K\362\342\324\213\371\305\345\204\326\371m\227\341\221\374\016\374\207\230\201WD\267]\262]\305|\001\024&\002\255'\032\214\025\330>\266\024\243\371\250\376\254,;\272V\246\000\300a#\313\237\231Ii\017t\361\005sEh\023\276\351\312Z\341\350\211<l\351\232\254$\331\024\374\221G\004\031\310M6X)k\010\034p\023\213\274\323\260\216\336\3164c\377te\036\220\217\234\251\312J\251dVb\372\260a\202\242\361I!Cd\037\2326\210\311}\204\212R\\\321\224G\237\331D\345\t\236\334*/TST\266eeU\227M\202\341\376\216\253\272\245o\227\234\317l\250o\353j\331knU\212m\376g\266\200\215\234?s\246\020\201\037x\344\244#\256u\344|\030\017tFW\201\353\0349\213\214Z44A\304\237\235\365\2053x@;r\032r!\314\243\304\307\375\314\234\226\010E\002\000\300\024+\013\304iG\2563\355\325\325\317\350/\016\032B\037_\262\037\270\335(T\344\265\025\302\351\221\263D@*\312\n\177\222\300\321\340\223\225m\374'\322\225\347\364m\017\371\030\245\313K`\376%\236\003&k\211\202\227\r\004\025O)\2268o\022\307""\320\377\030\314\261\037\305-\013\240^\023\214\035\203\240\005\363\355\016\337\336>\350T\212\006\247w[\367\244>U<\206`'\203\\\251\320\304Z\262Y\014EM^\201\215bah!\252\026\277\007nJ\031\212+{vr\273\377\210e\340\277\340\223\003\224B\035\207h\303L.\250\360\327\242\370E\227\261\312\211\006\ta\3145&\205'Y$\344\356~w\232\202\220V\225\220xc\214\243\232*\016\rPE,\344\006o\313\301\232 P\334\212\245\032v\026i\205]\306SSw\323r\273M\341\367{7\312\266\327\354NT\263\2458\026%E\264\202\357\275_\223\205x\251\016\207\302zK\033\367M\2657)\351O\037\316\353\302\330j\2057\366\227\210\311\200&\026\257\363\022\0313\376\344\246&\\j\035\242\317\375\356\203\372\\I\230\206*\3420T\304\302\025\3336\221w\343\317M\222R\220\t\275P\342\355\\D\3424S\242\256\031\032\315j\305\301Q\375pw\273(@MO-uMW\327e\314\225\313%>{?\027\025C-\2572`^\325uMo\031\343\317\261\346\210,\253&4-\003\301\206\247\027\251i\350\034\200t{CC\323\322e\303\342\2312\224\257\256\253\246\355\352\252\351\252\266Y.Z(,U\361\022\357SEv\267\357\205\261t\361\260\227\020\207\312S>\342\263\375\367\204&\362>\255\000:\242S]Y\326\210\216\360\353\226L\245\"[ \2768O\326\221\001\353\344\303G\004\030\221W\307R\000\350J\245\200\013\221\223\2320\274\265\352\255A\r\212\307u!bNK\344\021\r\217v\224\013eK\225\3458\032BM\370\233\341\244MN\177\200\030\2503U\345q<\276\334\252\310\376u'y{@\350\226H\326\014G\373q\t\\\225L\213\255\007\333m\222\346\365\326\340E7\214\263,\216\240\026\024\301\216n\033\254bL\226\240\337#G\332\006b\325\267\365&/(\036pd\036~\014\225{\317\032\374\003\003\306\352\032(\001\272\344t\330\326\024\376\224z~5\277\240 u\0209?\247\236\226\031x\3644Q\203f\261\210\263\216\034OY\240}R\3746\227\001|\247\3551)^V\236a\036\372\312=\214E\017\233w\204\252\016\231\031\307yO\2741\270\373\310\036\2752\376\343\007\356a\013\223J\273J\233\023\270\304\206\266\r\177/\230\260\224\027\243\3015,MGo\301N\345B\006,\205\247\240\327\372\300\377\354s\177\177\220\363\276\314\237\270%0\216s\344\013\277m@+\264u\304\250\3747\371\327\376\260\307?&\211x\006\214\033\024""\277+2\212f\3243\332\262\013\264-\305\302o\342KD\207\255\246\305`\t\221\320\242\317\236m\213A\252\356G\250m]\257$)\014\010\027\210\334\342\234\252\305\024\316\373DM\267xh\022\361\211\374=Q\006|*\315'\304\221\234'\221C\312[\206\267\006\356\266\361b\347\344D\277\270D\\\360\250\206\374\304\264\267R\316\327\332\022\256hn\232\244\020*\350\325/c\010\322\342\267DY\"2#`E\340,F\033\"\340\272\033\327\t \251*MoI\350\3421\201\356\303\307\016d\272Gz\206\233dWq J\232\264\264Y/\320\307\021<\252\350]%\314\207\354\254I\270\361W\3168\226CJ[+\203\312\345\003\337y\016t$\260+\272\2532\217K\342\0278IHE\010\222\264\342\234\0101\304\263l\256Y7\225\013Z\024nLd\n-\331{KC\030\270\031?\255\022\377\004ds\231\232\202\305\025\236c%`M';\351\221t\372\223\352q\017\311\202bY\343\221M\335q\014->\313\321\013\216\356\212\213\265\365\330&X\225\362AxY\251\004\235)nI+!\003\202\233\224\2706m\312P\304\336%<F\004pKv\211\2539\245\353\364!-\371Mk\262C\347\354g8\351d\2479\202\364`\023ww\366\343f*\202\002\331\030Wu7\361tJ\"\304g\263\244\3373\324\365$\305\370w\337\024\017yc\225\360\306\212c4U7\371\307`\031\3776\312\212)P\204\032X\304\253\210\314\232\257MUc\013\246{\270\327\246;x\022\334\222\230\213F\2223\213\026_\306\337%\315/9\362\201o:\315\016}\233\252\005\341\000\016R\217m\374\030.Q\207\370\254\350\350\310\320h\310\005\243'\323\274\262c9\030w\\\3726niP`\234\343\311\226\016mQ\212\227t|\302\024\372N\311]Co>\212\250\345\352\300\000}D\212\035\r5M/(t\320\376\247%\371\220\317L\356\232]65\370^B\231In\203$\003\351`\221\243\233\022t\200\336\243\377\272\206D\323\360\370O\372\225\320\314}\342\327\202\340(\310\2145`:\250\001J\007|\232\3615\311w[\243\004\206\013\334\203\313p\003\027\016\255\256y\310\007e\243\200\274\331\206\234Z:8\tS\037\354\241\350\303m\230\036\342\346\214;b0\210\311\262G\317 \331[s\354-b\3575p~\363\251\254z\202\320qE\361fin\335\366\202io\307\347\361^\216\267Tu\177\023\036\335\310\265\322\344\2156\252e\244\021e\253\rD\342I\023?l\232\272\214\223'\316\177\200\262\201C\232\017\024\227\237\310\177\334\230b""\343B\274 \266?\031\245\027\225y\361I\373\332\260f\227a\305\353\322\347\313\221\277\250\362\376c\256\365\371(\236c\355\357U\326y+\352ht\235\330\361\330\251\033\341B\250\324\276\251g\3529\326y\336\3175\272\316\371}\354\342\235h\"\362jh}\273sg\347\017\177\302\257\204\035\215\356\276\352\244\337\3218\336[\275\350\253\301\2710C=s\254\377N\224\217\336\324\224\275\375\326r\255w\017K\357V\377^\335h\233\204c7Xg.\352\335\353\356\017\006\202b\224\333\353\352\251vV\237\373J\243\353<;\177+\352dw^\262\227\257\330\253\267\354\355;\336w7Z\21046\366|w\200\315iL[ck\306\036\365\337\010\027\303\215F\327\005v\341f\350D\347jW\353\307\3308\026\276nt}\307\276\033\016\217\341\212]\022\223F\243\376\350im\264\266\260Gs\263\341\n\304\277\374C8\032.42\375X\357?\rr\301Tx\222\375\360\270\366?\273\243\273\257\331\362K\352\236\014\240\246\301`9\354mt]\362\315\260?\31472C\201\203\366\225\201F\327Uv\365\307\332\261\032\316\270\354;\301Iv\345^\264R\353\3407\273\036\216\207Z\224\241e\n\273\234\213.\324\026k[u\245\216\333~\317\276O\356\355\347\331\305[\32116:Y\317\343\242\301\tH\201\255o\362b\360j\243+\3032\271\350d\224k\014`\032\266\273:\324\350\272\302\357\212\033\221\372\354h1\332\252)\265r}\274\2564\006\305\344\241aX\221\235\233\336\355mt\367V{\253\231\317\025P\211?\356\177\014;\302\323\241\322\350>\316\216_\364\027|\215\365C\351\354\336\002[Xl\234\350a=8\025\203;2]\371F\3706\332`?\362\375\217\357|\360s\354\002\357\202.\372\331\303\331\335\005\276\215\024\034\0132\354\312\375h\273\266\301~&\2136\216\303\366\325c\325~\3267\022\236\215za\225qv\357\371n\206\317\317\004\275A6\372\246v\255\236\257\177`K\357\331\373\0171X0\"\200\270Aw>Y}\204\231\031\326y\227\360\324\000x\376\001\323dk\235\265|\203#i6\270D\332=S-\373y\377U\3600\334 \264\237\365\377\006\271\n\301L\324\001\345ve\202\323\201\022l\262\033\023\365\034)\277?\030\0146\302\257X\026\246\250/4\272NW\337\007\375\244g\202\3209\377|0\036(\324\373.8FX\236\257\217\326_\357b\335wU+\030'\034\234\250;\273g\330\013,=\276C\n\330Y\256^\360\205&.\371:l6\027""\rDz-W\233\337\315\361^~\"\033\376\271\336Y\237\332\355e\317U\246j\361t@\346fX\206\303\374V\357\253\277\334\315\354>\206'\220Cae\017\371\224\337\341_\206\305\276\217\306j\013\265u\330\001\207.\356|\362{\375A\177\203<{\203\253q\341_y\002\270\224\241\277\376+\215\376ax\361\360(\200$~\210\006\256\374\020\346\032\271\237\340\353\274y\3756\000t\363A\255\267qk\014\310N\232g\352\035ut\335\2176\032\243\017\241h\376s\227~\256=\256m4n\277 \311\245\253\320\321\340u8\311\225\033\344Z\264\363P\026\250\032\244\332u0@c \033\256\302\231\225\306\010\0356@\307\362Mo\374WmA\374\3744^\3174\206`\025\222\367\261\177\205t\010D\002\000\354\333\213\376\022<\262\233\034e\203\020\361\310?E\356\254E\331\3721\254\352>Y\275\006\223cn7\353\206\353\320vT\205\237\241\334\231'vK\n\326}\2314\205\362b\274@\224\315a\261\210u\023\360\017+/\021_\241$q\323\375C\201\330\366\032\001\007%\211\236\356\037bC\220\272\366\244\336:L\nK\227\234\241Q\016\006\242Mj\344\267\273\343\3773TB\261\227\024t\264\334%\271s\177 6'sS9\020\344[D\036!\372@I\366J\227\334\321Z\312\276\340}4\304\036\314\355*\215\030~\254g0x\023*-\347$\373'\227M\356\225\234\227\310>\030\274\016\363\261\200\367\242\364\036;s~\306\317\371\277\004\033\200,(\356\016\033z\304\036\255\260\025\225\\L\331\217C\036\273r\227\335}\002\262\353\346\201\352&\273\3713\031\277~\247\376\007[~\307\336\375\326\350\376\036{\335\303\325\351\232\347/\371Z0\300\256b\267\337\330o \025\261'I\266\020\250p\357\033\221S;U\313\023Tx@\2448\005\277\331\353\372\232\016\375\251\252\372\275t\324\231\340\2330C\266\202o\321\025O\354\270\325L5W}\342\017a\253\267\241\027!~N\260\t\204'\235\270\207\210\036\327x\030|\002\367\377\266{\021A\211\275\306\365V\331\352Z:\252\376#\322j\003\354\321rr\367\363\376\250\377\222Tr\246\352\370}\376\373p$\272\023\375A\014\305\t\260\332\311\372\256\207\023\260\177\367Y\377\357\210:}l\360\247\332\203\372${\366\201}\220I\223\317X\037\331\003\343_qZ\277\024m\034h\020\025B\203'v\266 \037\2352\014rB\204E\304U\230\262\262\327\203\330\307\244\307\265O\240\226""\236S\3251\356w}R\320\031<\207N{\317\372\177\005\016\377\010\237E\223\265N\366p\211-\361\240K\253HM{=}\254\357\207\360>(\246\247\257\372\324\277M\341\344\307Z\007\265&\375\343\000\301t\264\020\237q\217\335{\266\333\261\333GA\203$\332\300A=d\361\036\234|\272\272@\265tA!\301\251~\017\254L\201\346\267\331\375\247\365\355\335\r\366b\221-\376\312~\205\211\311\276\376\302\277:\032\235\337\354\374P]`\235\034\2240m\365q\220\t\356\004\377\214\024\326y#\3745\"JI\010\346T\365\001\\\347\370\005\230d\343\377\265\265'8\211]|\204 \275AQ\213\343\375@\314\342\331\317Y\277\223\034\370T\365vU\001\307\037\363\007\0208\036E\247\240\345\216\332Y\204\202\005\366\004(\372\265q\226\253\356G\366\343+\016+\010\376\206b\323K\000\356\016\273<\032\rSNr\022\370|\340?\013\246\303\267\354\3763\366L\314\002|\033]\275\325\363\374\216@\313\323\352]\004\223S\324\014x\2067K\301\265\001\327\205\247\375\304~z\303\336\000\0242\223\177\347~E\230\361\2529\002\005\3265\372.\340\000\212\214}p\365=\264\246\203E\370z\037\021@\017y\336\315\250/z\205\314k\t>:PW\366z\206\303o\243\273\265\343\010\261\3118\273\373\013R\256%nB\001\243\263\330\271\233\304\271\032\3745\270\037^A\212\252G\367k\375D\232\365\034\321|?)\214r\322\360\014\245\213\317w\007w=\266$\034\374C3Fr\014\235\363O\373E8.\207\217\016r?y\252z\017>\235\340\213\272zz\377\367\253\277|\335\323\004\326w\301v\270QG\346\301\t\214\034\217d\206Es{\240\002\360\001'\310\323\325\327\2106\313\251*ek\275Q\006\036\273\261\3073\210\014\021\331\317\024;Iz\212\322\257\253y\034\252\371YJN\341\"\375\321T\255\257\366k\375\355\356'\256\351\337\331\357\005V0\231YdEk\257K\330c\200\022\214\016\314[\256\363\274\232gBRp\034\212\321\330(\364\266D\316s\014TDrU\317\263\013\267\242\277\301\305\272\316\303\367V\202\257As\375\341\023\\,G\350z\340?\247\324'|\215\234\030\001\3726\256\273\211\364\324\255g\330\004\3224\301S\340\261\217\354\343:[7\323\267\276A\271o8\204,\363.\260\373\254\236'\275u\357\334\336Q\210:\257\004\243\360\265nz\036(\004\037\n\370h\235\001\300z\201\345\"%\334\010\377""\213Q\2716Y\357\332=\266{\ry6{\211\323@:D\214\224\327\247\026({\207*\366[~\336?!XGj{v\211\262\374\374\036wu\302tol\276\023\342j9\377\023%s\377\007\200\354,D";
    PyObject *data = __Pyx_DecompressString(cstring, 4962, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (11013 bytes) */
const char* const bytes = "0.10.51100.1%2.2%.4fAmount (Adj)Amount (Orig)  Assets:Available BalanceBest AskBest Bid). Canceling Order: () Creating Current %Current Value (Dict[str, List[str]]%H:%M:%SHanging maker BUY order Hanging maker SELL order ) Hanging maker buy order ) Hanging maker sell order ) ID - Initial inventory price is not set while inventory_cost feature is active.Inventory RangeList[Decimal]List[MarketTradingPairTuple]Maker BUY order Maker SELL order ) Maker buy order ) Maker buy order of ) Maker sell order ) Maker sell order of Market connectors are not ready.  Markets:Markets are not ready. No market making trades are permitted.  No active maker orders.NoneOptional[MovingPriceBand]Optional[pd.DataFrame]Order Adjust %Order is below minimum spread (  Orders:Parameter price_ceiling cannot be lower than price_floor.  Ping-pong removed Ref Price (Ref Price (MidPrice)Target %Target Value (Total BalanceUnrecognized price type string *** WARNINGS ***WARNING: Some markets are not connected or are down at the moment. Market making may be dangerous when markets or networks are unstable.) - \n    {:<}(  ( @ , .? ask orders at (Size, Price):  bid orders at (Size, Price):  buy orders.disableenable filled.gc) has been completely filled.hummingbot.strategy.pure_market_making.data_typeshummingbot.strategy.pure_market_making.inventory_cost_price_delegatehummingbot.strategy.pure_market_making.inventory_skew_calculatorhummingbot.strategy.pure_market_making.pure_market_making_order_trackerhummingbot.strategy.pure_market_making.moving_price_bandhummingbot/strategy/pure_market_making/pure_market_making.pyx is filled.isenablednumpy._core.multiarray failed to importnumpy._core.umath failed to importpd.DataFrameself._events cannot be converted to a Python object for pickling sell orders.<stringsource>AgeAssetPriceDelegateBUYBestAskBestBidBuyCONNECTEDCreatedPairOfOrdersCustomDataFrameDecimalDictExchangeExchangeBaseHangingOrdersTrackerINFOInventoryCostInventoryCostPriceDelegateLIMITLastOwnTradeLastTrade""LevelLimitOrderListMarketMarketTradingPairTupleMidPriceMovingPriceBandNaNNetworkStatusOPTION_LOG_ALLOPTION_LOG_CREATE_ORDEROPTION_LOG_MAKER_ORDER_FILLEDOPTION_LOG_STATUS_REPORTOptionalOrderTypePricePriceSizePriceTypeProposalPureMarketMakingOrderTrackerPureMarketMakingStrategyPureMarketMakingStrategy.__reduce_cython__PureMarketMakingStrategy.__setstate_cython__PureMarketMakingStrategy.active_orders_dfPureMarketMakingStrategy.adjusted_available_balance_for_orders_budget_constrainPureMarketMakingStrategy.all_markets_readyPureMarketMakingStrategy.c_apply_order_optimization.<locals>.<lambda>PureMarketMakingStrategy.c_cancel_active_orders_on_max_age_limit.<locals>.genexprPureMarketMakingStrategy.c_execute_orders_proposal.<locals>.genexprPureMarketMakingStrategy.c_start.<locals>.genexprPureMarketMakingStrategy.cancel_orderPureMarketMakingStrategy.execute_orders_proposalPureMarketMakingStrategy.format_statusPureMarketMakingStrategy.get_mid_pricePureMarketMakingStrategy.get_pricePureMarketMakingStrategy.get_price_typePureMarketMakingStrategy.init_paramsPureMarketMakingStrategy.inventory_skew_stats_data_framePureMarketMakingStrategy.loggerPureMarketMakingStrategy.market_status_data_framePureMarketMakingStrategy.notify_hb_appPureMarketMakingStrategy.pure_mm_assets_df__Pyx_PyDict_NextRefSELLSellSpreadStrategyBaseTimestampTradeTypeTypeactive_marketsactive_non_hanging_ordersactive_ordersactive_orders_dfactive_orders_df.<locals>.<lambda>add_as_hanging_orderadd_current_pairs_of_proposal_orders_executed_by_strategyadd_transaction_costs_to_ordersadjusted_available_balance_for_orders_budget_constrainageallall_markets_readyall_non_hanging_ordersall_ordersamountamount_origapplyask_order_level_spreadsask_order_optimization_depthask_priceask_ratioask_spreadasset_price_delegateassets_dfasyncio.coroutinesavailable_base_balanceavailable_quote_balancebalance_warningbase_assetbase_asset_amountbase_asset_rangebase_asset_ratiobase_asset_valuebase_balancebase_currencybase_ratiobase_valuebest_askb""est_bidbid_ask_ratiosbid_order_level_spreadsbid_order_optimization_depthbid_pricebid_ratiobid_spreadboolbuybuyscalculate_total_order_sizecancel_ordercandidate_hanging_orderscandidate_hanging_orders_from_pairsceilcheck_and_update_price_bandcheck_price_ceiling_exceededcheck_price_floor_exceeded__class_getitem__client_order_idcline_in_tracebackcloseclscolumnsconcatcurrent_created_pairs_of_orderscurrent_timestampcustomdatadata_typesdecimaldfdf_lines__dict__display_nameenabledenumerateexecute_orders_proposalfilled_order_delayfirst_col_lengthfloatfloorformatformat_statusformatters__func__genexprgetgetLoggerget_available_balanceget_balanceget_limit_ordersget_maker_order_typeget_mid_priceget_priceget_price_by_typeget_price_for_volumeget_price_type__getstate__hanghanging_order_idshanging_orders_cancel_pct_hanging_orders_cancel_pcthanging_orders_enabledhanging_orders_trackerhb_app_notificationheaderhigh_water_markhigh_water_mark_ratiohummingbot.connector.exchange_basehummingbot.core.data_type.commonhummingbot.core.data_type.limit_orderhummingbot.core.network_iteratorhummingbot.core.utilshummingbot.strategy.asset_price_delegatehummingbot.strategy.hanging_orders_trackerhummingbot.strategy.market_trading_pair_tuplehummingbot.strategy.pure_market_making.pure_market_makinghummingbot.strategy.strategy_basehummingbot.strategy.utilsidxin_flight_cancelsindexinfoinit_paramsinnerintinventory_costinventory_cost_price_delegateinventory_range_multiplierinventory_skew_calculatorinventory_skew_dfinventory_skew_enabledinventory_skew_stats_data_frameinventory_target_base_pctis_buy_is_coroutineis_hanging_orderis_nanis_order_id_in_hanging_ordersis_potential_hanging_orderitemsjoinkey<lambda>last_own_trade_pricelast_pricelenlevellevel_for_calculationlimit_orderslinelineslog_with_clockloggerlogginglogging_optionslow_water_marklow_water_mark_ratiolvl_buylvl_sell__main__map_df_to_strmarketmarket_booksmarket_infomarket_info_to_active_ordersmarket_pair_to_active_ordersmarket_status_data_framemarket_tra""ding_pair_tuplesmarkets_columnsmarkets_datamarkets_dfmathmaxmax_order_agemid_priceminimum_spread__module__moving_price_bandmoving_price_band_enabledmsg__name__namenannetwork_statusnetwork_warningnextno_sellsnon_hangingnormalizenotify_hb_appnotify_hb_app_with_timestampnpnumpyoorderorder_ageorder_amountorder_idorder_level_amountorder_level_spreadorder_levelsorder_optimization_enabledorder_overrideorder_refresh_timeorder_refresh_tolerance_pctppandaspdpercentping_pong_enabledpmm_loggerpoppriceprice_band_refresh_timeprice_ceilingprice_ceiling_pctprice_floorprice_floor_pctprice_providerprice_typeprice_type_strprocess_order_fill_eventprocess_tickproposalpure_market_making_order_trackerpure_mm_assets_df__pyx_state__pyx_vtable____qualname__quantityquote_assetquote_asset_amountquote_asset_ratioquote_asset_valuequote_balancequote_currencyquote_ratioready__reduce____reduce_cython____reduce_ex__ref_priceregexregister_eventsreplaceresult_pricereturnreverseroundss_decimal_neg_ones_decimal_zeroselfsellsell_ordersellssend__set_name__setdefault__setstate____setstate_cython__should_wait_order_cancel_confirmationsizesortsortedsplitsplit_order_levels_enabledspreadstatus_report_intervalstrstrategy_current_hanging_ordersstrftimesuperswitchtake_if_crossedtarget_base_amounttarget_base_amount_in_quotetarget_base_ratiotarget_quote_amount__test__throwto_show_current_pctto_stringtotal_in_quotetotal_order_sizetotal_order_size_ratiototal_valuetotal_value_in_quotetrade_typetrading_pairtypingunitunregister_eventsupdateupdate_strategy_orders_with_equivalent_ordersvaluevalueswarningwarning_linesxzipPyObject *(double, double, double, double, double)\000c_calculate_bid_ask_ratios_from_base_asset_ratio\320\0040\260\001\330\010\017\210t\320\023-\250Q\250a\270\n\300!\3001\320\004\033\2301\330\010\031\230\024\320\0354\260C\260t\2701\330\010\013\2104\210}\230C\230y\250\001\330\014\024\220D\230\001\330\r\021\220\035\230c\240\031\250!\330\014\024\220N\320\"4\260A\260Y\270a\340\014\024\220N\320\"4\260A\260T""\270\021\340\010\013\2105\220\007\220q\330\014\024\220N\320\"4\260A\260Y\270a\340\010\017\210q\320\0041\260\021\340\014\"\240$\240m\2601\340\010\020\220\004\220J\230a\330\010\033\320\0330\260\004\3204U\320UV\320VZ\320Z[\330\010\033\320\0335\260Q\260d\320:J\310$\320Nd\320dh\320hi\340\010\033\320\033-\250R\250q\330\010\034\320\034/\250r\260\031\270&\300\002\320BV\320VW\330\010\026\320\026(\250\002\250!\330\010 \320 2\260\"\260G\2702\270Q\340\010\034\320\034.\250b\260\001\330\037+\2502\250Q\330!\"\330\010\034\230G\2401\240E\250\022\320+?\270|\3102\310W\320TU\330\010\034\230D\240\001\330\010%\240T\250\021\330\010\036\230l\250\"\250A\330!'\240r\250\021\330#$\330\010&\320&8\270\002\270!\330\010\037\230r\240\022\320#6\260b\270\001\340\010\033\320\033,\250B\250d\260!\330\010\036\230a\320\0371\260\034\270R\270w\300a\300q\330\010\032\320\032-\250R\250q\330\010\034\230A\320\0350\260\002\3202D\300A\330\010 \240\017\250r\260\021\330#/\250r\260\021\330%&\330\010!\320!1\260\022\2601\330$0\260\002\260!\330&'\330\010#\2401\240E\250\021\330\010\"\240$\240o\260R\260w\270a\270u\300B\300a\330%1\260\022\2601\330'(\330\010\031\320\031I\310\021\330\014\021\220\021\220!\330\014\021\220\021\220!\330\014\021\220\021\220!\330\014\021\220\021\220!\330\014\021\220\021\220!\340\010\034\230B\230j\250\001\250\025\250a\330\014\r\320\r\035\230Q\230d\320\"5\260Q\3206Q\320QR\330\017\020\320\020#\2401\330\014\r\210_\230A\320\035-\250Z\260q\3208I\310\021\330\014\r\210^\2301\320\034-\250Z\260q\270\002\270\"\320<M\310Q\330\014\r\320\r \240\002\240!\320#7\260x\270q\320@U\320UV\330\r\017\210q\220\002\220\"\320\024)\250\030\260\021\260\"\260B\3206J\310!\330\014\r\320\r!\240\021\240.\260\n\270*\300A\300^\320S]\320]^\340\010\017\210q\320\004!\240\021\330\010\013\2104\210q\330\014\021\220\022\220>\240\021\240!\320\0045\260Y\270a\330\010\020\220\016\230l\250.\270\004\270A\330\010\020\220\004\220M\240\036\250q\330\010\027\220u\230A\230V\240<\250q\260\001\330\010\030\230\005\230Q\230f\240L\260\001\260\021\330\010!""\240\025\240a\240v\320-C\3001\300A\330\010\"\240%\240q\250\006\320.D\300A\300Q\330\010\025\220]\240\"\240E\250\021\250!\330\010\031\230\033\240B\240a\330\010\025\220[\240\002\320\"4\260O\3002\300W\310A\330\010\026\220n\240B\320&8\270\017\300r\310\027\320PQ\330\010\r\210Q\330\014\r\210T\220\034\230Q\330\014\r\320\r\036\230e\2401\240N\260$\260e\2701\270O\3101\330\014\r\320\r\"\240%\240q\320(@\300\004\300E\310\021\320Jc\320cd\330\014\r\320\r\036\230a\320\037/\250u\260A\260\\\300\024\300U\310!\310?\320Z[\340\010\013\2101\330\014\020\220\007\220q\230\001\230\037\250\001\250\032\260:\270Q\270k\310\021\330\010\r\210R\210z\230\021\230%\230q\330\010\017\210q\270*\300A\300Q\200A\330!\"\330 !\330 !\330\"#\330\"(\250\001\330(2\260!\330(2\260!\330(0\260\001\330#+\2501\3301;\2701\330(0\260\001\330,3\2601\330/9\270\021\3300:\270!\330,3\2601\330/9\270\027\300\001\300\021\33007\260q\3302<\270A\3302<\270A\3305<\270A\330*?\270q\3303P\320PQ\330 &\240a\330%,\250A\330#-\250Q\330!+\2501\330'.\250a\330%+\2501\330,4\260A\330$.\250g\260Q\260a\330)0\260\001\330$;\2701\33007\260q\330-=\270Q\330-=\270Q\330;B\300!\330'C\3001\340\010\013\210?\230#\230Q\330\014\035\230Q\330\010\013\320\013\035\230S\240\001\330\014 \240\017\250q\330\010\013\210>\230\023\320\034.\250d\260.\300\002\300!\330\014\022\220*\230A\230Q\330\010\014\320\014!\320!=\270Q\330\010\014\320\014\034\230A\330\010\014\210O\2301\330\010\014\210O\2301\330\010\014\320\014\037\230q\330\010\014\320\014\035\230Q\330\010\014\320\014\035\230Q\330\010\014\210O\2301\330\010\014\320\014\034\230A\330\010\014\320\014#\2401\330\010\014\320\014#\2401\330\010\014\320\014#\2401\330\010\014\320\014\036\230a\330\010\014\320\014,\250A\330\010\014\320\014#\2401\330\010\014\320\014'\240q\330\010\014\320\014*\250!\330\010\014\320\014+\2501\330\010\014\320\014'\240q\330\010\014\320\014'\320';\2701\270F\300!\330\010\014\320\014+\2501\330\010\014\320\014-\250Q\330\010\014\320\014-\250Q\330\010\014\320\0140\260\001\330\010\014\320\014%\240Q\330\010\014\320""\014.\250a\330\010\014\210O\2304\230\177\250a\250q\330\010\014\320\014 \240\001\330\010\014\320\014\036\230a\330\010\014\320\014\034\230A\330\010\014\320\014\"\240!\330\010\014\320\014(\250\001\330\010\014\320\014$\240A\330\010\014\320\014\037\230q\330\010\014\320\014)\250\021\330\010\014\320\014&\240a\330\010\014\320\014&\240a\330\010\014\320\014!\240\021\330\010\014\320\014!\240\021\330\010\014\320\014!\240\024\240]\260'\3209N\310a\330\010\013\2101\330\014\020\320\020%\240Y\250a\330\010\014\320\014\"\240!\330\010\014\320\014$\240A\330\010\014\320\014%\240Q\330\010\014\320\014 \240\001\330\010\014\320\014\037\230q\330\010\014\320\014'\240q\330\010\014\320\014%\240W\250A\250Q\330\010\014\320\0146\260a\330\010\014\320\014\"\240!\330\010\014\210N\230!\2301\230K\240q\200A\330\010#\2404\320'>\320>b\320bc\330\010\026\220a\330\010\013\2104\210}\230C\230t\320#5\3205F\300a\330\014\031\230\024\320\035/\320/@\300\002\300!\3004\300}\320T[\320[\\\330\014\032\230!\2306\240\024\240Y\250a\330\033\036\230d\240$\320&>\320>\\\320\\]\320]b\320bc\330\010!\240\024\240Q\240c\250\021\250-\260r\270\023\270A\270Q\330\010\017\210t\320\0234\260A\260Q\200A\340\010\t\330\010\013\210;\220c\230\021\330\014\031\230\027\240\n\250!\2501\330\010\017\210q\200A\330\010\017\210s\220!\2201\220F\230'\240\024\240Z\250t\2601\320\004C\320Cd\320de\330\010\027\220q\330\010\032\230!\230<\240z\260\034\270\\\310\035\320VW\320W[\320[g\320gh\330\010\013\2104\210}\230C\230y\250\016\260d\270$\320>T\320T[\320[\\\330\014\033\2302\230U\240!\330\010\027\220r\230\024\230]\250)\2604\260}\300A\330\010\013\2104\210q\220\004\320\024,\250C\250q\330\014\030\230\007\230r\240\024\320%;\2709\300D\320H^\320^_\330\010\014\210H\320\024$\240A\330\014\030\230\006\230j\250\001\250\036\260q\330\014\030\230\006\230j\250\001\250\036\260q\330\014\030\230\005\230Q\230a\330\014\017\210w\220c\230\024\230]\250(\260$\260d\320:Z\320Za\320ab\340\020\034\230D\320 ?\270z\310\021\330\020\023\220:\230S\240\001\330\024 \240\004\240J\250a\330\021\030""\230\003\2304\230}\250H\260D\270\004\320<S\320SV\320VW\330\020\034\230D\240\n\250!\340\020\024\320\024+\2507\260!\330\020\024\220G\2303\230d\320\"8\270\001\330\020\024\220D\230\r\240W\250I\260Q\340\020\034\230D\320 6\3206H\310\001\310\024\310Q\330\014\030\230\007\230q\240\001\330\020\026\220a\330\020\021\330\020\025\220Q\220a\330\020\025\220Q\220a\330\020\025\220Q\220a\340\010\017\210r\220\032\2301\230E\240\036\250x\3207G\300x\310q\320PR\320RX\320X\\\320\\b\320bc\230Q\200\001\330\004\n\210+\220Q\320\004%\240Q\330\010\017\210t\220?\240!\2404\240\177\260a\320\004-\250X\260Q\330\010\013\210?\230#\230Q\330\014\023\2209\230A\330\r\034\230C\230q\330\014\023\2209\230A\330\r\034\230C\230q\330\014\023\2209\230A\330\r\034\230C\230q\330\014\023\2209\230A\330\r\034\230C\230q\330\014\023\2209\230A\330\r\034\230C\230q\330\014\023\2209\230A\330\r\034\230C\230q\330\014\023\2209\230A\340\014\022\220*\230A\320\035>\270a\270q\240!\320\004\"\240!\330\010\020\220\016\230l\250.\270\004\270A\330\010\020\220\004\220J\230a\330\010\030\230\004\230A\330\010\023\2203\220a\220q\230\002\230$\230e\240>\260\023\260D\270\001\270\030\300\024\300Q\320FW\320WX\330\030\034\230D\320 8\3208V\320VW\320WX\320XY\330\010\025\220U\230!\2304\320\0372\260(\270!\330\010\022\220!\2209\230H\240I\250Z\3207H\320HX\320XY\330\010\017\210q\330\010\021\220\033\230C\230q\330\010\014\210G\2205\230\001\230\023\230C\230q\240\001\330\014\024\220M\240\021\240!\330\014\037\230t\320#;\320;Y\320YZ\320Z_\320_`\330\014\032\230!\330\014\017\210t\2201\330\020\023\2205\230\001\330\024\034\230H\240B\240a\330\024\037\230q\340\024\034\230I\240R\240q\330\024 \240\001\330\020\036\230d\240/\260\024\260V\2702\270S\300\002\300$\300a\340\020(\250\013\2605\270\r\300Q\330\020\036\230d\240/\260\024\3205K\3102\310S\320PR\320RV\320VW\330\020\030\230\001\330\014\025\220U\230&\240\003\2407\250#\250Q\250e\2607\270\"\270F\300!\3001\330\014\022\220\"\220J\230a\230y\250\001\250\027\260\004\3204J\310%\310t\320S\\\320\\]\320]^\330\014\020\220\007\220q""\230\001\330\020\021\330\020\031\230\025\230m\2501\330\020\025\220Q\220e\2301\330\022\023\2206\230\021\330\020\021\330\020\025\220Q\220e\2301\330\020\021\360\006\000\t\020\210r\220\032\2301\230E\240\026\240x\250q\300q\320\004\037\230q\330\010\013\2104\210t\2201\330\014\023\2201\340\014\031\230\021\330\014!\240\021\330\010\025\220W\230A\230T\240\021\330\010\025\220W\230A\230T\320!1\260\021\260!\2604\260q\340\010\025\220]\240!\2404\320'@\300\001\300\021\300$\300a\330\010\r\210W\220A\220Q\220d\230.\250\002\250!\2507\260\"\260E\270\024\270X\300Z\310z\320YZ\320Z`\320`f\320fl\320lm\320mn\340\010\024\220M\240\021\240$\320&8\270\001\270\024\270T\300\021\340\010\013\2104\210q\330\014 \240\r\250Q\250d\3202R\320RS\330\014\030\230\002\230'\240\021\330\020\021\220\033\320\0340\260\005\260Q\340\010\033\2303\230b\240\t\250\021\250\"\250F\260!\2601\330\010\023\2209\230J\240a\240v\250W\260G\2701\330'3\2604\260v\270R\270s\300!\320CU\320UW\320W[\320[d\320dj\320jk\320kl\330\010\r\210W\220A\220Q\220d\230-\240r\250\021\250'\260\022\2605\270\004\270H\300A\360\006\000\t\014\2103\210a\210t\320\023#\2402\240Q\330\014\021\220\035\230a\230t\320#4\260A\330\014\021\220\027\230\001\230\021\230$\230m\2502\250Q\250g\260R\260u\270D\300\010\310\002\310*\320TU\320U[\320[a\320ag\320gh\320hi\340\014\021\220\027\230\001\230\021\230$\230a\340\010\025\220W\230A\230T\320!1\260\021\260!\2604\260q\340\010\013\2103\210a\210\177\230b\240\001\330\014\021\220\027\230\001\230\021\230$\320\0362\260\"\260A\340\010\017\210t\2205\230\001\230\021\320\004\037\230q\330\010\017\210t\320\023#\2401\230z\250\021\250!";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 463; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 91) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 463; i < 486; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 486; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 463;
      for (Py_ssize_t i=0; i<23; ++i) {
        #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
        }
        #else
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab;
    double const c_constants[] = {30.0,60.0,1800.0};
    for (int i = 0; i < 3; i++) {
      numbertab[i] = PyFloat_FromDouble(c_constants[i]);
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab + 3;
    int8_t const cint_constants_1[] = {0,-1,1,4,8,16,32};
    int16_t const cint_constants_2[] = {900};
    int64_t const cint_constants_8[] = {9223372036854775807LL};
    for (int i = 0; i < 9; i++) {
      numbertab[i] = PyLong_FromLongLong((i < 7 ? cint_constants_1[i - 0] : (i < 8 ? cint_constants_2[i - 7] : cint_constants_8[i - 8])));
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_number_tab;
    for (Py_ssize_t i=0; i<12; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 6;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 1;
    unsigned int nlocals : 6;
    unsigned int flags : 10;
    unsigned int first_line : 11;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 580};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_x};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_lambda, __pyx_mstate->__pyx_kp_b_iso88591_z, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR), 707};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_o};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_genexpr, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1010};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_p};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_lambda, __pyx_mstate->__pyx_kp_b_iso88591_1_3, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1033};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_p};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_lambda, __pyx_mstate->__pyx_kp_b_iso88591_AQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR), 1184};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_o};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_genexpr, __pyx_mstate->__pyx_kp_b_iso88591_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR), 1274};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_o};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_genexpr, __pyx_mstate->__pyx_kp_b_iso88591__27, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {0, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR), 1297};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_o};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_genexpr, __pyx_mstate->__pyx_kp_b_iso88591__27, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 44};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_cls};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_logger, __pyx_mstate->__pyx_kp_b_iso88591_A_c_1_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {38, 0, 0, 38, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 51};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_market_info, __pyx_mstate->__pyx_n_u_bid_spread, __pyx_mstate->__pyx_n_u_ask_spread, __pyx_mstate->__pyx_n_u_order_amount, __pyx_mstate->__pyx_n_u_order_levels, __pyx_mstate->__pyx_n_u_order_level_spread, __pyx_mstate->__pyx_n_u_order_level_amount, __pyx_mstate->__pyx_n_u_order_refresh_time, __pyx_mstate->__pyx_n_u_max_order_age, __pyx_mstate->__pyx_n_u_order_refresh_tolerance_pct, __pyx_mstate->__pyx_n_u_filled_order_delay, __pyx_mstate->__pyx_n_u_inventory_skew_enabled, __pyx_mstate->__pyx_n_u_inventory_target_base_pct, __pyx_mstate->__pyx_n_u_inventory_range_multiplier, __pyx_mstate->__pyx_n_u_hanging_orders_enabled, __pyx_mstate->__pyx_n_u_hanging_orders_cancel_pct, __pyx_mstate->__pyx_n_u_order_optimization_enabled, __pyx_mstate->__pyx_n_u_ask_order_optimization_depth, __pyx_mstate->__pyx_n_u_bid_order_optimization_depth, __pyx_mstate->__pyx_n_u_add_transaction_costs_to_orders, __pyx_mstate->__pyx_n_u_asset_price_delegate, __pyx_mstate->__pyx_n_u_inventory_cost_price_delegate, __pyx_mstate->__pyx_n_u_price_type, __pyx_mstate->__pyx_n_u_take_if_crossed, __pyx_mstate->__pyx_n_u_price_ceiling, __pyx_mstate->__pyx_n_u_price_floor, __pyx_mstate->__pyx_n_u_ping_pong_enabled, __pyx_mstate->__pyx_n_u_logging_options, __pyx_mstate->__pyx_n_u_status_report_interval, __pyx_mstate->__pyx_n_u_minimum_spread, __pyx_mstate->__pyx_n_u_hb_app_notification, __pyx_mstate->__pyx_n_u_order_override, __pyx_mstate->__pyx_n_u_split_order_levels_enabled, __pyx_mstate->__pyx_n_u_bid_order_level_spreads, __pyx_mstate->__pyx_n_u_ask_order_level_spreads, __pyx_mstate->__pyx_n_u_should_wait_order_cancel_confirm, __pyx_mstate->__pyx_n_u_moving_price_band};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_init_params, __pyx_mstate->__pyx_kp_b_iso88591_A_2_2_0_11_1_0_31_9_0_31_9_07q2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 149};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_market};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_all_markets_ready, __pyx_mstate->__pyx_kp_b_iso88591_A_s_1F_Zt1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 413};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_price_provider, __pyx_mstate->__pyx_n_u_price};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_get_price, __pyx_mstate->__pyx_kp_b_iso88591_1_4Ct1_4_Cy_D_c_N_4AYa_N_4AT_5, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 427};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_get_mid_price, __pyx_mstate->__pyx_kp_b_iso88591_q_t_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 25, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 495};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_market, __pyx_mstate->__pyx_n_u_price, __pyx_mstate->__pyx_n_u_base_asset_amount, __pyx_mstate->__pyx_n_u_quote_asset_amount, __pyx_mstate->__pyx_n_u_total_order_size, __pyx_mstate->__pyx_n_u_base_asset_value, __pyx_mstate->__pyx_n_u_quote_asset_value, __pyx_mstate->__pyx_n_u_total_value, __pyx_mstate->__pyx_n_u_total_value_in_quote, __pyx_mstate->__pyx_n_u_base_asset_ratio, __pyx_mstate->__pyx_n_u_quote_asset_ratio, __pyx_mstate->__pyx_n_u_target_base_ratio, __pyx_mstate->__pyx_n_u_inventory_range_multiplier, __pyx_mstate->__pyx_n_u_target_base_amount, __pyx_mstate->__pyx_n_u_target_base_amount_in_quote, __pyx_mstate->__pyx_n_u_target_quote_amount, __pyx_mstate->__pyx_n_u_base_asset_range, __pyx_mstate->__pyx_n_u_high_water_mark, __pyx_mstate->__pyx_n_u_low_water_mark, __pyx_mstate->__pyx_n_u_low_water_mark_ratio, __pyx_mstate->__pyx_n_u_high_water_mark_ratio, __pyx_mstate->__pyx_n_u_total_order_size_ratio, __pyx_mstate->__pyx_n_u_bid_ask_ratios, __pyx_mstate->__pyx_n_u_inventory_skew_df};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_inventory_skew_stats_data_frame, __pyx_mstate->__pyx_kp_b_iso88591_1_m1_Ja_0_4UUVVZZ_5Qd_J_Nddhhi, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 17, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 552};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_to_show_current_pct, __pyx_mstate->__pyx_n_u_market, __pyx_mstate->__pyx_n_u_trading_pair, __pyx_mstate->__pyx_n_u_base_asset, __pyx_mstate->__pyx_n_u_quote_asset, __pyx_mstate->__pyx_n_u_price, __pyx_mstate->__pyx_n_u_base_balance, __pyx_mstate->__pyx_n_u_quote_balance, __pyx_mstate->__pyx_n_u_available_base_balance, __pyx_mstate->__pyx_n_u_available_quote_balance, __pyx_mstate->__pyx_n_u_base_value, __pyx_mstate->__pyx_n_u_total_in_quote, __pyx_mstate->__pyx_n_u_base_ratio, __pyx_mstate->__pyx_n_u_quote_ratio, __pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_df};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_pure_mm_assets_df, __pyx_mstate->__pyx_kp_b_iso88591_5Ya_l_A_M_q_uAV_q_QfL_av_C1A_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 21, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 574};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_market, __pyx_mstate->__pyx_n_u_trading_pair, __pyx_mstate->__pyx_n_u_base_asset, __pyx_mstate->__pyx_n_u_quote_asset, __pyx_mstate->__pyx_n_u_price, __pyx_mstate->__pyx_n_u_active_orders, __pyx_mstate->__pyx_n_u_no_sells, __pyx_mstate->__pyx_n_u_columns, __pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_lvl_buy, __pyx_mstate->__pyx_n_u_lvl_sell, __pyx_mstate->__pyx_n_u_idx, __pyx_mstate->__pyx_n_u_order, __pyx_mstate->__pyx_n_u_is_hanging_order, __pyx_mstate->__pyx_n_u_amount_orig, __pyx_mstate->__pyx_n_u_level, __pyx_mstate->__pyx_n_u_level_for_calculation, __pyx_mstate->__pyx_n_u_spread, __pyx_mstate->__pyx_n_u_age, __pyx_mstate->__pyx_n_u_o};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_active_orders_df, __pyx_mstate->__pyx_kp_b_iso88591_l_A_Ja_A_3aq_e_D_QFWWX_D_88VVWW, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 614};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_market_trading_pair_tuples, __pyx_mstate->__pyx_n_u_markets_data, __pyx_mstate->__pyx_n_u_markets_columns, __pyx_mstate->__pyx_n_u_market_books, __pyx_mstate->__pyx_n_u_market, __pyx_mstate->__pyx_n_u_trading_pair, __pyx_mstate->__pyx_n_u_bid_price, __pyx_mstate->__pyx_n_u_ask_price, __pyx_mstate->__pyx_n_u_ref_price};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_market_status_data_frame, __pyx_mstate->__pyx_kp_b_iso88591_CCdde_q_z_VWW_ggh_4_Cy_d_TT_2U, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 12, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 648};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_lines, __pyx_mstate->__pyx_n_u_warning_lines, __pyx_mstate->__pyx_n_u_markets_df, __pyx_mstate->__pyx_n_u_assets_df, __pyx_mstate->__pyx_n_u_inventory_skew_df, __pyx_mstate->__pyx_n_u_first_col_length, __pyx_mstate->__pyx_n_u_df_lines, __pyx_mstate->__pyx_n_u_df, __pyx_mstate->__pyx_n_u_line, __pyx_mstate->__pyx_n_u_line, __pyx_mstate->__pyx_n_u_line};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_format_status, __pyx_mstate->__pyx_kp_b_iso88591_q_4t1_1_WAT_WAT_1_4q_4_a_WAQd_7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 688};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_proposal};
    __pyx_mstate_global->__pyx_codeobj_tab[17] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_execute_orders_proposal, __pyx_mstate->__pyx_kp_b_iso88591_0_t_Qa, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[17])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 691};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_order_id};
    __pyx_mstate_global->__pyx_codeobj_tab[18] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_cancel_order, __pyx_mstate->__pyx_kp_b_iso88591_Q_t_4_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[18])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 920};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_candidate_hanging_orders, __pyx_mstate->__pyx_n_u_non_hanging, __pyx_mstate->__pyx_n_u_all_orders, __pyx_mstate->__pyx_n_u_all_non_hanging_orders, __pyx_mstate->__pyx_n_u_order};
    __pyx_mstate_global->__pyx_codeobj_tab[19] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_adjusted_available_balance_for_o, __pyx_mstate->__pyx_kp_b_iso88591_A_4_bbc_a_4_Ct_55Fa_4_T_6_Ya_d_b, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[19])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1310};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_msg};
    __pyx_mstate_global->__pyx_codeobj_tab[20] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_notify_hb_app, __pyx_mstate->__pyx_kp_b_iso88591_4q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[20])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1314};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_price_type_str};
    __pyx_mstate_global->__pyx_codeobj_tab[21] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_hummingbot_strategy_pure_market_7, __pyx_mstate->__pyx_n_u_get_price_type, __pyx_mstate->__pyx_kp_b_iso88591_XQ_Q_9A_Cq_9A_Cq_9A_Cq_9A_Cq_9A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[21])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[22] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[22])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[23] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[23])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* Generator.init */
  if (likely(__pyx_Generator_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* GetTopmostException (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C0000
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#elif __PYX_LIMITED_VERSION_HEX > 0x030C0000
    local_value = PyErr_GetRaisedException();
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
#if __PYX_LIMITED_VERSION_HEX > 0x030C0000
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
#else
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
#endif // __PYX_LIMITED_VERSION_HEX > 0x030C0000
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#elif __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    PyErr_SetHandledException(local_value);
    Py_XDECREF(local_value);
    Py_XDECREF(local_type);
    Py_XDECREF(local_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
#if __PYX_LIMITED_VERSION_HEX <= 0x030C0000
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
#endif
}

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* TupleAndListFromArray (used by fastcall) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyObjectCallOneArg (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by RejectKeywords) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *pos = NULL;
#else
        Py_ssize_t pos = 0;
#endif
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL);
#if CYTHON_AVOID_BORROWED_REFS
        Py_XDECREF(pos);
#endif
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* IterFinish (used by set_iter) */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* set_iter */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    is_set = is_set || likely(PySet_CheckExact(iterable) || PyFrozenSet_CheckExact(iterable));
    *p_source_is_set = is_set;
    if (likely(is_set)) {
        *p_orig_length = PySet_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#else
    CYTHON_UNUSED_VAR(is_set);
    *p_source_is_set = 0;
#endif
    *p_orig_length = 0;
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set) {
    if (!CYTHON_COMPILING_IN_CPYTHON || PY_VERSION_HEX >= 0x030d0000 || unlikely(!source_is_set)) {
        *value = PyIter_Next(iter_obj);
        if (unlikely(!*value)) {
            return __Pyx_IterFinish();
        }
        CYTHON_UNUSED_VAR(orig_length);
        CYTHON_UNUSED_VAR(ppos);
        return 1;
    }
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    if (unlikely(PySet_GET_SIZE(iter_obj) != orig_length)) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "set changed size during iteration");
        return -1;
    }
    {
        Py_hash_t hash;
        int ret = _PySet_NextEntry(iter_obj, ppos, value, &hash);
        assert (ret != -1);
        if (likely(ret)) {
            Py_INCREF(*value);
            return 1;
        }
    }
#endif
    return 0;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS)) {
        return __Pyx_PyList_GetItemRefFast(o, wrapped_i, unsafe_shared);
    } else
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyList_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyTuple_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)) {
            return __Pyx_PyList_GetItemRefFast(o, n, unsafe_shared);
        } else if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_NewRef(PyList_GET_ITEM(o, n));
        }
    } else
    #if !CYTHON_AVOID_BORROWED_REFS
    if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            return __Pyx_NewRef(PyTuple_GET_ITEM(o, n));
        }
    } else
    #endif
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_mstate_global->__pyx_n_u_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_SubtractCObj(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_SubtractCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long a = intval;
    long b;
    const PY_LONG_LONG lla = intval;
    PY_LONG_LONG llb;
    if (unlikely(__Pyx_PyLong_IsZero(op2))) {
        return __Pyx_NewRef(op1);
    }
    const int is_positive = __Pyx_PyLong_IsPos(op2);
    const digit* digits = __Pyx_PyLong_Digits(op2);
    const Py_ssize_t size = __Pyx_PyLong_DigitCount(op2);
    if (likely(size == 1)) {
        b = (long) digits[0];
        if (!is_positive) b *= -1;
    } else {
        switch (size) {
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    b = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) b *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    llb = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) llb *= -1;
                    goto calculate_long_long;
                }
                break;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    b = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) b *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    llb = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) llb *= -1;
                    goto calculate_long_long;
                }
                break;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    b = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) b *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    llb = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) llb *= -1;
                    goto calculate_long_long;
                }
                break;
        }
        return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    calculate_long:
        {
            long x;
            x = a - b;
            return PyLong_FromLong(x);
        }
    calculate_long_long:
        {
            PY_LONG_LONG llx;
            llx = lla - llb;
            return PyLong_FromLongLong(llx);
        }
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_SubtractCObj(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long a = intval;
    double b = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) - (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_SubtractCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op2))) {
        return __Pyx_Unpacked___Pyx_PyLong_SubtractCObj(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op2)) {
        return __Pyx_Float___Pyx_PyLong_SubtractCObj(op2, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_SubtractCObj(op1, op2, inplace);
}
#endif

/* pybytes_as_double (used by pyobject_as_double) */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj) {
    PyObject *float_value = PyFloat_FromString(obj);
    if (likely(float_value)) {
        double value = __Pyx_PyFloat_AS_DOUBLE(float_value);
        Py_DECREF(float_value);
        return value;
    }
    return (double)-1;
}
static const char* __Pyx__PyBytes_AsDouble_Copy(const char* start, char* buffer, Py_ssize_t length) {
    int last_was_punctuation = 1;
    int parse_error_found = 0;
    Py_ssize_t i;
    for (i=0; i < length; i++) {
        char chr = start[i];
        int is_punctuation = (chr == '_') | (chr == '.') | (chr == 'e') | (chr == 'E');
        *buffer = chr;
        buffer += (chr != '_');
        parse_error_found |= last_was_punctuation & is_punctuation;
        last_was_punctuation = is_punctuation;
    }
    parse_error_found |= last_was_punctuation;
    *buffer = '\0';
    return unlikely(parse_error_found) ? NULL : buffer;
}
static double __Pyx__PyBytes_AsDouble_inf_nan(const char* start, Py_ssize_t length) {
    int matches = 1;
    char sign = start[0];
    int is_signed = (sign == '+') | (sign == '-');
    start += is_signed;
    length -= is_signed;
    switch (start[0]) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            matches &= (start[1] == 'a' || start[1] == 'A');
            matches &= (start[2] == 'n' || start[2] == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            matches &= (start[1] == 'n' || start[1] == 'N');
            matches &= (start[2] == 'f' || start[2] == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            matches &= (start[3] == 'i' || start[3] == 'I');
            matches &= (start[4] == 'n' || start[4] == 'N');
            matches &= (start[5] == 'i' || start[5] == 'I');
            matches &= (start[6] == 't' || start[6] == 'T');
            matches &= (start[7] == 'y' || start[7] == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static CYTHON_INLINE int __Pyx__PyBytes_AsDouble_IsSpace(char ch) {
    return (ch == 0x20) | !((ch < 0x9) | (ch > 0xd));
}
CYTHON_UNUSED static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length) {
    double value;
    Py_ssize_t i, digits;
    const char *last = start + length;
    char *end;
    while (__Pyx__PyBytes_AsDouble_IsSpace(*start))
        start++;
    while (start < last - 1 && __Pyx__PyBytes_AsDouble_IsSpace(last[-1]))
        last--;
    length = last - start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyBytes_AsDouble_inf_nan(start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    digits = 0;
    for (i=0; i < length; digits += start[i++] != '_');
    if (likely(digits == length)) {
        value = PyOS_string_to_double(start, &end, NULL);
    } else if (digits < 40) {
        char number[40];
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((digits + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}

/* pyobject_as_double */
static double __Pyx__PyObject_AsDouble(PyObject* obj) {
    if (PyUnicode_CheckExact(obj)) {
        return __Pyx_PyUnicode_AsDouble(obj);
    } else if (PyBytes_CheckExact(obj)) {
        return __Pyx_PyBytes_AsDouble(obj);
    } else if (PyByteArray_CheckExact(obj)) {
        return __Pyx_PyByteArray_AsDouble(obj);
    } else {
        PyObject* float_value;
#if !CYTHON_USE_TYPE_SLOTS
        float_value = PyNumber_Float(obj);  if ((0)) goto bad;
        (void)__Pyx_PyObject_CallOneArg;
#else
        PyNumberMethods *nb = Py_TYPE(obj)->tp_as_number;
        if (likely(nb) && likely(nb->nb_float)) {
            float_value = nb->nb_float(obj);
            if (likely(float_value) && unlikely(!PyFloat_Check(float_value))) {
                __Pyx_TypeName float_value_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(float_value));
                PyErr_Format(PyExc_TypeError,
                    "__float__ returned non-float (type " __Pyx_FMT_TYPENAME ")",
                    float_value_type_name);
                __Pyx_DECREF_TypeName(float_value_type_name);
                Py_DECREF(float_value);
                goto bad;
            }
        } else {
            float_value = __Pyx_PyObject_CallOneArg((PyObject*)&PyFloat_Type, obj);
        }
#endif
        if (likely(float_value)) {
            double value = __Pyx_PyFloat_AS_DOUBLE(float_value);
            Py_DECREF(float_value);
            return value;
        }
    }
bad:
    return (double)-1;
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject** values, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    if (max_char > 1114111) max_char = 1114111;
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - result_ulength < 0))
        goto overflow;
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = values[i];
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (__Pyx_PyUnicode_READY(uval) == (-1))
            goto bad;
        #endif
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(ulength < 0)) goto bad;
        #endif
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    Py_ssize_t i;
    PyObject *result = NULL;
    PyObject *value_tuple = PyTuple_New(value_count);
    if (unlikely(!value_tuple)) return NULL;
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    for (i=0; i<value_count; i++) {
        if (__Pyx_PyTuple_SET_ITEM(value_tuple, i, values[i]) != (0)) goto bad;
        Py_INCREF(values[i]);
    }
    result = PyUnicode_Join(__pyx_mstate_global->__pyx_empty_unicode, value_tuple);
bad:
    Py_DECREF(value_tuple);
    return result;
#endif
}

/* PyObjectFormat */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* obj, PyObject* format_spec) {
    int ret;
    _PyUnicodeWriter writer;
    if (likely(PyFloat_CheckExact(obj))) {
        _PyUnicodeWriter_Init(&writer);
        ret = _PyFloat_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else if (likely(PyLong_CheckExact(obj))) {
        _PyUnicodeWriter_Init(&writer);
        ret = _PyLong_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else {
        return PyObject_Format(obj, format_spec);
    }
    if (unlikely(ret == -1)) {
        _PyUnicodeWriter_Dealloc(&writer);
        return NULL;
    }
    return _PyUnicodeWriter_Finish(&writer);
}
#endif

/* PyObjectVectorCallKwBuilder */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* dict_setdefault (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType (used by FetchCommonType) */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_vectorcall_offset = memb->offset;
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#else
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* AddModuleRef (used by FetchSharedCythonModule) */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* FetchSharedCythonModule (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* FetchCommonType (used by CommonTypesMetaclass) */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    cached_type = __Pyx_PyType_FromMetaclass(
        metaclass,
        CYTHON_USE_MODULE_STATE ? module : abi_module,
        spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass (used by CythonFunctionShared) */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
#if __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
    PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
    return NULL;
}
static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
    PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
    return -1;
}
#endif
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
    #endif
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    Py_DECREF(bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* CallTypeTraverse (used by CythonFunctionShared) */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* PyMethodNew (used by CythonFunctionShared) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i;
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos;
    #else
    Py_ssize_t pos;
    #endif
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = 0;
    i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(pos);
    #endif
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared (used by CythonFunction) */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
#endif
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
#else
    {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
#endif
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#endif
#if CYTHON_METH_FASTCALL
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_dict = NULL;
#endif
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func_dict);
#elif PY_VERSION_HEX < 0x030d0000
    _PyObject_ClearManagedDict((PyObject*)m);
#else
    PyObject_ClearManagedDict((PyObject*)m);
#endif
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func_dict);
#else
    {
        int e =
#if PY_VERSION_HEX < 0x030d0000
            _PyObject_VisitManagedDict
#else
            PyObject_VisitManagedDict
#endif
                ((PyObject*)m, visit, arg);
        if (e != 0) return e;
    }
#endif
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    #if PY_VERSION_HEX < 0x030e00A6
    size_t nargs_value = (size_t) nargs;
    #else
    Py_ssize_t nargs_value = nargs;
    #endif
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_DICT |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* PyObjectVectorCallMethodKwBuilder */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames) {
    PyObject *result;
    PyObject *obj = PyObject_GetAttr(args[0], name);
    if (unlikely(!obj))
        return NULL;
    result = __Pyx_Object_Vectorcall_CallFromBuilder(obj, args+1, nargsf-1, kwnames);
    Py_DECREF(obj);
    return result;
}
#endif

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    long a;
    const PY_LONG_LONG llb = intval;
    PY_LONG_LONG lla;
    if (unlikely(__Pyx_PyLong_IsZero(op1))) {
        return __Pyx_NewRef(op2);
    }
    const int is_positive = __Pyx_PyLong_IsPos(op1);
    const digit* digits = __Pyx_PyLong_Digits(op1);
    const Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
    if (likely(size == 1)) {
        a = (long) digits[0];
        if (!is_positive) a *= -1;
    } else {
        switch (size) {
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
        }
        return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    calculate_long:
        {
            long x;
            x = a + b;
            return PyLong_FromLong(x);
        }
    calculate_long_long:
        {
            PY_LONG_LONG llx;
            llx = lla + llb;
            return PyLong_FromLongLong(llx);
        }
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_AddObjC(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    double a = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) + (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        return __Pyx_Unpacked___Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        return __Pyx_Float___Pyx_PyLong_AddObjC(op1, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_AddObjC(op1, op2, inplace);
}
#endif

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_SubtractObjC(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    long a;
    const PY_LONG_LONG llb = intval;
    PY_LONG_LONG lla;
    if (unlikely(__Pyx_PyLong_IsZero(op1))) {
        return PyLong_FromLong(-intval);
    }
    const int is_positive = __Pyx_PyLong_IsPos(op1);
    const digit* digits = __Pyx_PyLong_Digits(op1);
    const Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
    if (likely(size == 1)) {
        a = (long) digits[0];
        if (!is_positive) a *= -1;
    } else {
        switch (size) {
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
        }
        return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    calculate_long:
        {
            long x;
            x = a - b;
            return PyLong_FromLong(x);
        }
    calculate_long_long:
        {
            PY_LONG_LONG llx;
            llx = lla - llb;
            return PyLong_FromLongLong(llx);
        }
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_SubtractObjC(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    double a = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) - (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        return __Pyx_Unpacked___Pyx_PyLong_SubtractObjC(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        return __Pyx_Float___Pyx_PyLong_SubtractObjC(op1, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_SubtractObjC(op1, op2, inplace);
}
#endif

/* PyLongCompare */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 1;
    }
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) == 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 0;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 0;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal == 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = __Pyx_PyFloat_AS_DOUBLE(op1);
        return ((double)a == (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* py_abs */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *n) {
#if PY_VERSION_HEX >= 0x030C00A7
    if (likely(__Pyx_PyLong_IsCompact(n))) {
        return PyLong_FromSize_t(__Pyx_PyLong_CompactValueUnsigned(n));
    }
#else
    if (likely(Py_SIZE(n) == -1)) {
        return PyLong_FromUnsignedLong(__Pyx_PyLong_Digits(n)[0]);
    }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    {
        PyObject *copy = _PyLong_Copy((PyLongObject*)n);
        if (likely(copy)) {
            #if PY_VERSION_HEX >= 0x030C00A7
            ((PyLongObject*)copy)->long_value.lv_tag ^= ((PyLongObject*)copy)->long_value.lv_tag & _PyLong_SIGN_MASK;
            #else
            __Pyx_SET_SIZE(copy, -Py_SIZE(copy));
            #endif
        }
        return copy;
    }
#else
    return PyNumber_Negative(n);
#endif
}
#endif

/* pynumber_float */
static CYTHON_INLINE PyObject* __Pyx__PyNumber_Float(PyObject* obj) {
    double val;
    if (PyLong_CheckExact(obj)) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (likely(__Pyx_PyLong_IsCompact(obj))) {
            val = (double) __Pyx_PyLong_CompactValue(obj);
            goto no_error;
        }
#endif
        val = PyLong_AsDouble(obj);
    } else if (PyUnicode_CheckExact(obj)) {
        val = __Pyx_PyUnicode_AsDouble(obj);
    } else if (PyBytes_CheckExact(obj)) {
        val = __Pyx_PyBytes_AsDouble(obj);
    } else if (PyByteArray_CheckExact(obj)) {
        val = __Pyx_PyByteArray_AsDouble(obj);
    } else {
        return PyNumber_Float(obj);
    }
    if (unlikely(val == -1 && PyErr_Occurred())) {
        return NULL;
    }
#if CYTHON_USE_PYLONG_INTERNALS
no_error:
#endif
    return PyFloat_FromDouble(val);
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (unlikely(!j)) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE && !CYTHON_AVOID_BORROWED_REFS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared))) {
            Py_INCREF(v);
            return PyList_SetItem(o, n, v);
        } else if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old;
            Py_INCREF(v);
            old = PyList_GET_ITEM(o, n);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 0;
        }
    } else
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_ass_subscript) {
            int r;
            PyObject *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return -1;
            r = mm->mp_ass_subscript(o, key, v);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return sm->sq_ass_item(o, i, v);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_SetItem(o, i, v);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_SetItemInt_Generic(o, PyLong_FromSsize_t(i), v);
}

/* UnicodeConcatInPlace */
# if CYTHON_COMPILING_IN_CPYTHON
static int
__Pyx_unicode_modifiable(PyObject *unicode, int unsafe_shared)
{
    if (!__Pyx_IS_UNIQUELY_REFERENCED(unicode, unsafe_shared))
        return 0;
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030F0000
    if (PyUnstable_Unicode_GET_CACHED_HASH(unicode) != -1)
        return 0;
#endif
    if (!PyUnicode_CheckExact(unicode))
        return 0;
    if (PyUnicode_CHECK_INTERNED(unicode))
        return 0;
    return 1;
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right, int unsafe_shared
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    ) {
    PyObject *left = *p_left;
    Py_ssize_t left_len, right_len, new_len;
    if (unlikely(__Pyx_PyUnicode_READY(left) == -1))
        return NULL;
    if (unlikely(__Pyx_PyUnicode_READY(right) == -1))
        return NULL;
    left_len = PyUnicode_GET_LENGTH(left);
    if (left_len == 0) {
        Py_INCREF(right);
        return right;
    }
    right_len = PyUnicode_GET_LENGTH(right);
    if (right_len == 0) {
        Py_INCREF(left);
        return left;
    }
    if (unlikely(left_len > PY_SSIZE_T_MAX - right_len)) {
        PyErr_SetString(PyExc_OverflowError,
                        "strings are too large to concat");
        return NULL;
    }
    new_len = left_len + right_len;
    if (__Pyx_unicode_modifiable(left, unsafe_shared)
            && PyUnicode_CheckExact(right)
            && PyUnicode_KIND(right) <= PyUnicode_KIND(left)
            && !(PyUnicode_IS_ASCII(left) && !PyUnicode_IS_ASCII(right))) {
        int ret;
        __Pyx_GIVEREF(*p_left);
        ret = PyUnicode_Resize(p_left, new_len);
        __Pyx_GOTREF(*p_left);
        if (unlikely(ret != 0))
            return NULL;
        #if PY_VERSION_HEX >= 0x030d0000
        if (unlikely(PyUnicode_CopyCharacters(*p_left, left_len, right, 0, right_len) < 0)) return NULL;
        #else
        _PyUnicode_FastCopyCharacters(*p_left, left_len, right, 0, right_len);
        #endif
        __Pyx_INCREF(*p_left);
        __Pyx_GIVEREF(*p_left);
        return *p_left;
    } else {
        return __Pyx_PyUnicode_Concat(left, right);
    }
  }
#endif

/* RaiseUnboundLocalError */
static void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* RaiseClosureNameError */
static void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc, *new_exc;
    __Pyx_PyThreadState_declare
    int is_async_stopiteration = 0;
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    __Pyx_PyThreadState_assign
    cur_exc = __Pyx_PyErr_CurrentExceptionType();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else {
            return;
        }
    }
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(tb);
    new_exc = PyObject_CallFunction(PyExc_RuntimeError, "s",
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        "generator raised StopIteration");
    if (!new_exc) {
        Py_XDECREF(val);
        return;
    }
    PyException_SetCause(new_exc, val); // steals ref to val
    PyErr_SetObject(PyExc_RuntimeError, new_exc);
}

/* IterNextPlain (used by IterNext) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *__Pyx_GetBuiltinNext_LimitedAPI(void) {
    if (unlikely(!__pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache))
        __pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache = __Pyx_GetBuiltinName(__pyx_mstate_global->__pyx_n_u_next);
    return __pyx_mstate_global->__Pyx_GetBuiltinNext_LimitedAPI_cache;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next_Plain(PyObject *iterator) {
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    PyObject *result;
    PyObject *next = __Pyx_GetBuiltinNext_LimitedAPI();
    if (unlikely(!next)) return NULL;
    result = PyObject_CallFunctionObjArgs(next, iterator, NULL);
    return result;
#else
    (void)__Pyx_GetBuiltinName; // only for early limited API
    iternextfunc iternext = __Pyx_PyObject_GetIterNextFunc(iterator);
    assert(iternext);
    return iternext(iterator);
#endif
}

/* IterNext */
static PyObject *__Pyx_PyIter_Next2Default(PyObject* defval) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (!defval || unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return NULL;
        __Pyx_PyErr_Clear();
        Py_INCREF(defval);
        return defval;
    }
    if (defval) {
        Py_INCREF(defval);
        return defval;
    }
    __Pyx_PyErr_SetNone(PyExc_StopIteration);
    return NULL;
}
static void __Pyx_PyIter_Next_ErrorNoIterator(PyObject *iterator) {
    __Pyx_TypeName iterator_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(iterator));
    PyErr_Format(PyExc_TypeError,
        __Pyx_FMT_TYPENAME " object is not an iterator", iterator_type_name);
    __Pyx_DECREF_TypeName(iterator_type_name);
}
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject* iterator, PyObject* defval) {
    PyObject* next;
#if !CYTHON_COMPILING_IN_LIMITED_API
    iternextfunc iternext = __Pyx_PyObject_TryGetSlot(iterator, tp_iternext, iternextfunc);
    if (likely(iternext)) {
        next = iternext(iterator);
        if (likely(next))
            return next;
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
        if (unlikely(iternext == &_PyObject_NextNotImplemented))
            return NULL;
    #endif
    } else if (CYTHON_USE_TYPE_SLOTS) {
        __Pyx_PyIter_Next_ErrorNoIterator(iterator);
        return NULL;
    } else
#endif
    if (unlikely(!PyIter_Check(iterator))) {
        __Pyx_PyIter_Next_ErrorNoIterator(iterator);
        return NULL;
    } else {
        next = defval ? PyIter_Next(iterator) : __Pyx_PyIter_Next_Plain(iterator);
        if (likely(next))
            return next;
    }
    return __Pyx_PyIter_Next2Default(defval);
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCallNoArg (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}
#endif

/* PyObjectCallMethod0 (used by dict_iter) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[1] = {obj};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_CallNoArg;
    return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
#endif
}

/* UnpackTupleError (used by UnpackTuple2) */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else {
      Py_ssize_t size = __Pyx_PyTuple_GET_SIZE(t);
 #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely(size < 0)) return;
 #endif
      if (size < index) {
        __Pyx_RaiseNeedMoreValuesError(size);
      } else {
        __Pyx_RaiseTooManyValuesError(index);
      }
    }
}

/* UnpackTuple2 (used by dict_iter) */
static CYTHON_INLINE int __Pyx_unpack_tuple2(
        PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple) {
    if (likely(is_tuple || PyTuple_Check(tuple))) {
        Py_ssize_t size;
        if (has_known_size) {
            return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
        }
        size = __Pyx_PyTuple_GET_SIZE(tuple);
        if (likely(size == 2)) {
            return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
        }
        if (size >= 0) {
            __Pyx_UnpackTupleError(tuple, 2);
        }
        return -1;
    } else {
        return __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple);
    }
}
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
    value1 = __Pyx_PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = __Pyx_PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#else
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
#if !CYTHON_AVOID_BORROWED_REFS
static CYTHON_INLINE int __Pyx_dict_iter_next_source_is_dict(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem) {
    PyObject *key, *value;
    if (unlikely(orig_length != PyDict_Size(iter_obj))) {
        PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
        return -1;
    }
    if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
        return 0;
    }
    if (pitem) {
        PyObject* tuple = PyTuple_New(2);
        if (unlikely(!tuple)) {
            return -1;
        }
        Py_INCREF(key);
        Py_INCREF(value);
        #if CYTHON_ASSUME_SAFE_MACROS
        PyTuple_SET_ITEM(tuple, 0, key);
        PyTuple_SET_ITEM(tuple, 1, value);
        #else
        if (unlikely(PyTuple_SetItem(tuple, 0, key) < 0)) {
            Py_DECREF(value);
            Py_DECREF(tuple);
            return -1;
        }
        if (unlikely(PyTuple_SetItem(tuple, 1, value) < 0)) {
            Py_DECREF(tuple);
            return -1;
        }
        #endif
        *pitem = tuple;
    } else {
        if (pkey) {
            Py_INCREF(key);
            *pkey = key;
        }
        if (pvalue) {
            Py_INCREF(value);
            *pvalue = value;
        }
    }
    return 1;
}
#endif
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_AVOID_BORROWED_REFS
    if (source_is_dict) {
        int result;
#if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_BEGIN_CRITICAL_SECTION(iter_obj);
#endif
        result = __Pyx_dict_iter_next_source_is_dict(iter_obj, orig_length, ppos, pkey, pvalue, pitem);
#if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_END_CRITICAL_SECTION();
#endif
        return result;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        Py_ssize_t tuple_size = __Pyx_PyTuple_GET_SIZE(iter_obj);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(tuple_size < 0)) return -1;
        #endif
        if (unlikely(pos >= tuple_size)) return 0;
        *ppos = pos + 1;
        #if CYTHON_ASSUME_SAFE_MACROS
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        #else
        next_item = PyTuple_GetItem(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
        #endif
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        Py_ssize_t list_size = __Pyx_PyList_GET_SIZE(iter_obj);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(list_size < 0)) return -1;
        #endif
        if (unlikely(pos >= list_size)) return 0;
        *ppos = pos + 1;
        next_item = __Pyx_PyList_GetItemRef(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyLong_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyLong_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is unsliceable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return NULL;
}

/* CIntToDigits (used by CIntToPyUnicode) */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode (used by COrdinalToPyUnicode) */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, const char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp = PySequence_Repeat(padding, uoffset - prepend_sign);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyUnicode_Concat(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyUnicode_Concat(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* COrdinalToPyUnicode (used by CIntToPyUnicode) */
static CYTHON_INLINE int __Pyx_CheckUnicodeValue(int value) {
    return value <= 1114111;
}
static PyObject* __Pyx_PyUnicode_FromOrdinal_Padded(int value, Py_ssize_t ulength, char padding_char) {
    Py_ssize_t padding_length = ulength - 1;
    if (likely((padding_length <= 250) && (value < 0xD800 || value > 0xDFFF))) {
        char chars[256];
        if (value <= 255) {
            memset(chars, padding_char, (size_t) padding_length);
            chars[ulength-1] = (char) value;
            return PyUnicode_DecodeLatin1(chars, ulength, NULL);
        }
        char *cpos = chars + sizeof(chars);
        if (value < 0x800) {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xc0 | (value & 0x1f));
        } else if (value < 0x10000) {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xe0 | (value & 0x0f));
        } else {
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0x80 | (value & 0x3f));
            value >>= 6;
            *--cpos = (char) (0xf0 | (value & 0x07));
        }
        cpos -= padding_length;
        memset(cpos, padding_char, (size_t) padding_length);
        return PyUnicode_DecodeUTF8(cpos, chars + sizeof(chars) - cpos, NULL);
    }
    if (value <= 127 && CYTHON_USE_UNICODE_INTERNALS) {
        const char chars[1] = {(char) value};
        return __Pyx_PyUnicode_BuildFromAscii(ulength, chars, 1, 0, padding_char);
    }
    {
        PyObject *uchar, *padding_uchar, *padding, *result;
        padding_uchar = PyUnicode_FromOrdinal(padding_char);
        if (unlikely(!padding_uchar)) return NULL;
        padding = PySequence_Repeat(padding_uchar, padding_length);
        Py_DECREF(padding_uchar);
        if (unlikely(!padding)) return NULL;
        uchar = PyUnicode_FromOrdinal(value);
        if (unlikely(!uchar)) {
            Py_DECREF(padding);
            return NULL;
        }
        result = PyUnicode_Concat(padding, uchar);
        Py_DECREF(padding);
        Py_DECREF(uchar);
        return result;
    }
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!(is_unsigned || value == 0 || value > 0) ||
                    !(sizeof(value) <= 2 || value & ~ (int) 0x01fffff || __Pyx_CheckUnicodeValue((int) value)))) {
        PyErr_SetString(PyExc_OverflowError, "%c arg not in range(0x110000)");
        return NULL;
    }
    if (width <= 1) {
        return PyUnicode_FromOrdinal((int) value);
    }
    return __Pyx_PyUnicode_FromOrdinal_Padded((int) value, width, padding_char);
}
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_AddCObj(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_AddCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long a = intval;
    long b;
    const PY_LONG_LONG lla = intval;
    PY_LONG_LONG llb;
    if (unlikely(__Pyx_PyLong_IsZero(op2))) {
        return __Pyx_NewRef(op1);
    }
    const int is_positive = __Pyx_PyLong_IsPos(op2);
    const digit* digits = __Pyx_PyLong_Digits(op2);
    const Py_ssize_t size = __Pyx_PyLong_DigitCount(op2);
    if (likely(size == 1)) {
        b = (long) digits[0];
        if (!is_positive) b *= -1;
    } else {
        switch (size) {
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    b = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) b *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    llb = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) llb *= -1;
                    goto calculate_long_long;
                }
                break;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    b = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) b *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    llb = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) llb *= -1;
                    goto calculate_long_long;
                }
                break;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    b = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) b *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    llb = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) llb *= -1;
                    goto calculate_long_long;
                }
                break;
        }
        return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    calculate_long:
        {
            long x;
            x = a + b;
            return PyLong_FromLong(x);
        }
    calculate_long_long:
        {
            PY_LONG_LONG llx;
            llx = lla + llb;
            return PyLong_FromLongLong(llx);
        }
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_AddCObj(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long a = intval;
    double b = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) + (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op2))) {
        return __Pyx_Unpacked___Pyx_PyLong_AddCObj(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op2)) {
        return __Pyx_Float___Pyx_PyLong_AddCObj(op2, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_AddCObj(op1, op2, inplace);
}
#endif

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_mstate_global->__pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_uchar___Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!(is_unsigned || value == 0 || value > 0) ||
                    !(sizeof(value) <= 2 || value & ~ (Py_ssize_t) 0x01fffff || __Pyx_CheckUnicodeValue((int) value)))) {
        PyErr_SetString(PyExc_OverflowError, "%c arg not in range(0x110000)");
        return NULL;
    }
    if (width <= 1) {
        return PyUnicode_FromOrdinal((int) value);
    }
    return __Pyx_PyUnicode_FromOrdinal_Padded((int) value, width, padding_char);
}
static CYTHON_INLINE PyObject* __Pyx____Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* CallNextTpDealloc */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    destructor tp_dealloc = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_dealloc, destructor) != current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_dealloc = __Pyx_PyType_GetSlot(type, tp_dealloc, destructor)) == current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type)
        tp_dealloc(obj);
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    traverseproc tp_traverse = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc) != current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_traverse = __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc)) == current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_traverse)
        return tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    inquiry tp_clear = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_clear, inquiry) != current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_clear = __Pyx_PyType_GetSlot(type, tp_clear, inquiry)) == current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_clear)
        tp_clear(obj);
}

/* AllocateExtensionType */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final) {
    if (is_final || likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
        allocfunc alloc_func = __Pyx_PyType_GetSlot(t, tp_alloc, allocfunc);
        return alloc_func(t, 0);
    } else {
        newfunc tp_new = __Pyx_PyType_TryGetSlot(&PyBaseObject_Type, tp_new, newfunc);
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (!tp_new) {
            PyObject *new_str = PyUnicode_FromString("__new__");
            if (likely(new_str)) {
                PyObject *o = PyObject_CallMethodObjArgs((PyObject *)&PyBaseObject_Type, new_str, t, NULL);
                Py_DECREF(new_str);
                return o;
            } else
                return NULL;
        } else
    #endif
        return tp_new(t, __pyx_mstate_global->__pyx_empty_tuple, 0);
    }
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_2_4
#define __PYX_HAVE_RT_ImportType_3_2_4
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API)
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !( defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API) )
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_2_4 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_2_4 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* ValidateBasesTuple (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_SIZE
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (unlikely(n < 0)) return -1;
#endif
    for (i = 1; i < n; i++)
    {
        PyTypeObject *b;
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !CYTHON_USE_TYPE_SLOTS
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
    while (t) {
        PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
        if (bases) {
            return 1;
        }
        t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
    }
    return 0;
}
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    if (!__Pyx_PyType_HasMultipleInheritance(t)) {
        return PyType_Ready(t);
    }
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                !CYTHON_COMPILING_IN_GRAAL
        gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* MergeVTables */
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i=0;
    Py_ssize_t size;
    void** base_vtables;
    __Pyx_TypeName tp_base_name = NULL;
    __Pyx_TypeName base_name = NULL;
    void* unknown = (void*)-1;
    PyObject* bases = __Pyx_PyType_GetSlot(type, tp_bases, PyObject*);
    int base_depth = 0;
    {
        PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        while (base) {
            base_depth += 1;
            base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
        }
    }
    base_vtables = (void**) PyMem_Malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
#if CYTHON_COMPILING_IN_LIMITED_API
    size = PyTuple_Size(bases);
    if (size < 0) goto other_failure;
#else
    size = PyTuple_GET_SIZE(bases);
#endif
    for (i = 1; i < size; i++) {
        PyObject *basei;
        void* base_vtable;
#if CYTHON_AVOID_BORROWED_REFS
        basei = PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#else
        basei = PyTuple_GET_ITEM(bases, i);
#endif
        base_vtable = __Pyx_GetVtable((PyTypeObject*)basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
            }
        }
    }
    PyErr_Clear();
    PyMem_Free(base_vtables);
    return 0;
bad:
    {
        PyTypeObject* basei = NULL;
        PyTypeObject* tp_base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        tp_base_name = __Pyx_PyType_GetFullyQualifiedName(tp_base);
#if CYTHON_AVOID_BORROWED_REFS
        basei = (PyTypeObject*)PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = (PyTypeObject*)PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#else
        basei = (PyTypeObject*)PyTuple_GET_ITEM(bases, i);
#endif
        base_name = __Pyx_PyType_GetFullyQualifiedName(basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
    }
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
really_bad: // bad has failed!
#endif
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
other_failure:
#endif
    PyMem_Free(base_vtables);
    return -1;
}

/* DelItemOnTypeDict (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_DelItem(tp_dict, k);
    if (likely(!result)) PyType_Modified(tp);
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
            if (likely(reduce_cython)) {
                ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* PxdImportShared (used by FunctionImport) */
#ifndef __PYX_HAVE_RT_ImportFromPxd_3_2_4
#define __PYX_HAVE_RT_ImportFromPxd_3_2_4
static int __Pyx_ImportFromPxd_3_2_4(PyObject *module, const char *name, void **p, const char *sig, const char *what) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, "__pyx_capi__");
    if (!d)
        goto bad;
#if (defined(Py_LIMITED_API) && Py_LIMITED_API >= 0x030d0000) || (!defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030d0000)
    PyDict_GetItemStringRef(d, name, &cobj);
#else
    cobj = PyDict_GetItemString(d, name);
    Py_XINCREF(cobj);
#endif
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C %.8s %.200s",
                PyModule_GetName(module), what, name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C %.8s %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             what, PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    Py_DECREF(cobj);
    return 0;
bad:
    Py_XDECREF(d);
    Py_XDECREF(cobj);
    return -1;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction_3_2_4
#define __PYX_HAVE_RT_ImportFunction_3_2_4
static int __Pyx_ImportFunction_3_2_4(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    int result = __Pyx_ImportFromPxd_3_2_4(module, funcname, &tmp.p, sig, "function");
    if (result == 0) {
        *f = tmp.fp;
    }
    return result;
}
#endif

/* HasAttr (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u__25);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* ClassMethod */
static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM <= 0x05080000
    if (PyObject_TypeCheck(method, &PyWrapperDescr_Type)) {
        return PyClassMethod_New(method);
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (PyMethodDescr_Check(method))
#else
    if (__Pyx_TypeCheck(method, &PyMethodDescr_Type))
#endif
    {
#if CYTHON_COMPILING_IN_LIMITED_API
        return PyErr_Format(
            PyExc_SystemError,
            "Cython cannot yet handle classmethod on a MethodDescriptorType (%S) in limited API mode. "
            "This is most likely a classmethod in a cdef class method with binding=False. "
            "Try setting 'binding' to True.",
            method);
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
        PyTypeObject *d_type = GraalPyDescrObject_GetType(method);
        return PyDescr_NewClassMethod(d_type, GraalPyMethodDescrObject_GetMethod(method));
#elif CYTHON_COMPILING_IN_GRAAL
        PyTypeObject *d_type = PyDescrObject_GetType(method);
        return PyDescr_NewClassMethod(d_type, PyMethodDescrObject_GetMethod(method));
#else
        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
        PyTypeObject *d_type = descr->d_common.d_type;
        return PyDescr_NewClassMethod(d_type, descr->d_method);
#endif
    }
#endif
#if !CYTHON_COMPILING_IN_LIMITED_API
    else if (PyMethod_Check(method)) {
        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
    }
    else {
        return PyClassMethod_New(method);
    }
#else
    {
        PyObject *func=NULL;
        PyObject *builtins, *classmethod, *classmethod_str, *result=NULL;
        if (__Pyx_TypeCheck(method, __pyx_mstate_global->__Pyx_CachedMethodType)) {
            func = PyObject_GetAttrString(method, "__func__");
            if (!func) goto bad;
        } else {
            func = method;
            Py_INCREF(func);
        }
        builtins = PyEval_GetBuiltins(); // borrowed
        if (unlikely(!builtins)) goto bad;
        classmethod_str = PyUnicode_FromString("classmethod");
        if (unlikely(!classmethod_str)) goto bad;
        classmethod = PyObject_GetItem(builtins, classmethod_str);
        Py_DECREF(classmethod_str);
        if (unlikely(!classmethod)) goto bad;
        result = PyObject_CallFunctionObjArgs(classmethod, func, NULL);
        Py_DECREF(classmethod);
        bad:
        Py_XDECREF(func);
        return result;
    }
#endif
}

/* GetNameInClass */
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    PyObject *dict;
    assert(PyType_Check(nmspace));
#if CYTHON_USE_TYPE_SLOTS
    dict = ((PyTypeObject*)nmspace)->tp_dict;
    Py_XINCREF(dict);
#else
    dict = PyObject_GetAttr(nmspace, __pyx_mstate_global->__pyx_n_u_dict);
#endif
    if (likely(dict)) {
        result = PyObject_GetItem(dict, name);
        Py_DECREF(dict);
        if (result) {
            return result;
        }
    }
    PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      return ::std::complex< long double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      return x + y*(__pyx_t_long_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      __pyx_t_long_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_sum_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_diff_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_prod_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsl(b.real) >= fabsl(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                long double r = b.imag / b.real;
                long double s = (long double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_long_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            long double r = b.real / b.imag;
            long double s = (long double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_long_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            long double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_long_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_neg_long__double(__pyx_t_long_double_complex a) {
        __pyx_t_long_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_long__double(__pyx_t_long_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_conj_long__double(__pyx_t_long_double_complex a) {
        __pyx_t_long_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE long double __Pyx_c_abs_long__double(__pyx_t_long_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtl(z.real*z.real + z.imag*z.imag);
          #else
            return hypotl(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_pow_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
            __pyx_t_long_double_complex z;
            long double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    long double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_long__double(a, a);
                    case 3:
                        z = __Pyx_c_prod_long__double(a, a);
                        return __Pyx_c_prod_long__double(z, a);
                    case 4:
                        z = __Pyx_c_prod_long__double(a, a);
                        return __Pyx_c_prod_long__double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powl(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2l(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_long__double(a);
                theta = atan2l(a.imag, a.real);
            }
            lnr = logl(r);
            z_r = expl(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosl(z_theta);
            z.imag = z_r * sinl(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPy */
static CYTHON_INLINE int64_t __Pyx_PyLong_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyLong_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 2 * PyLong_SHIFT)) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 3 * PyLong_SHIFT)) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 4 * PyLong_SHIFT)) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
        } else if ((sizeof(int64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int64_t) 1) << (sizeof(int64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__26);
    }
    goto done;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* PyObjectCall2Args (used by PyObjectCallMethod1) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod1 (used by CoroutineBase) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* ReturnWithStopIteration (used by CoroutineBase) */
static void __Pyx__ReturnWithStopIteration(PyObject* value, int async);
static CYTHON_INLINE void __Pyx_ReturnWithStopIteration(PyObject* value, int async, int iternext) {
    if (value == Py_None) {
        if (async || !iternext)
            PyErr_SetNone(async ? PyExc_StopAsyncIteration : PyExc_StopIteration);
        return;
    }
    __Pyx__ReturnWithStopIteration(value, async);
}
static void __Pyx__ReturnWithStopIteration(PyObject* value, int async) {
#if CYTHON_COMPILING_IN_CPYTHON
    __Pyx_PyThreadState_declare
#endif
    PyObject *exc;
    PyObject *exc_type = async ? PyExc_StopAsyncIteration : PyExc_StopIteration;
#if CYTHON_COMPILING_IN_CPYTHON
    if ((PY_VERSION_HEX >= (0x030C00A6)) || unlikely(PyTuple_Check(value) || PyExceptionInstance_Check(value))) {
        if (PY_VERSION_HEX >= (0x030e00A1)) {
            exc = __Pyx_PyObject_CallOneArg(exc_type, value);
        } else {
            PyObject *args_tuple = PyTuple_New(1);
            if (unlikely(!args_tuple)) return;
            Py_INCREF(value);
            PyTuple_SET_ITEM(args_tuple, 0, value);
            exc = PyObject_Call(exc_type, args_tuple, NULL);
            Py_DECREF(args_tuple);
        }
        if (unlikely(!exc)) return;
    } else {
        Py_INCREF(value);
        exc = value;
    }
    #if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    #if CYTHON_USE_EXC_INFO_STACK
    if (!__pyx_tstate->exc_info->exc_value)
    #else
    if (!__pyx_tstate->exc_type)
    #endif
    {
        Py_INCREF(exc_type);
        __Pyx_ErrRestore(exc_type, exc, NULL);
        return;
    }
    #endif
#else
    exc = __Pyx_PyObject_CallOneArg(exc_type, value);
    if (unlikely(!exc)) return;
#endif
    PyErr_SetObject(exc_type, exc);
    Py_DECREF(exc);
}

/* CoroutineBase (used by Generator) */
#if !CYTHON_COMPILING_IN_LIMITED_API
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6 && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#endif // CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void
__Pyx_Coroutine_Undelegate(__pyx_CoroutineObject *gen) {
#if CYTHON_USE_AM_SEND
    gen->yieldfrom_am_send = NULL;
#endif
    Py_CLEAR(gen->yieldfrom);
}
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            #if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
            value = PyObject_GetAttr(ev, __pyx_mstate_global->__pyx_n_u_value);
            if (unlikely(!value)) goto limited_api_failure;
            #else
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            #endif
            Py_DECREF(ev);
        }
        else if (unlikely(PyTuple_Check(ev))) {
            Py_ssize_t tuple_size = __Pyx_PyTuple_GET_SIZE(ev);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely(tuple_size < 0)) {
                Py_XDECREF(tb);
                Py_DECREF(ev);
                Py_DECREF(et);
                return -1;
            }
            #endif
            if (tuple_size >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#elif CYTHON_ASSUME_SAFE_MACROS
                value = PySequence_ITEM(ev, 0);
#else
                value = PySequence_GetItem(ev, 0);
                if (!value) goto limited_api_failure;
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_GetAttr(ev, __pyx_mstate_global->__pyx_n_u_value);
#else
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
#endif
    Py_DECREF(ev);
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!value)) return -1;
#endif
    *pvalue = value;
    return 0;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL || !CYTHON_ASSUME_SAFE_MACROS
  limited_api_failure:
    Py_XDECREF(et);
    Py_XDECREF(tb);
    Py_XDECREF(ev);
    return -1;
#endif
}
static CYTHON_INLINE
__Pyx_PySendResult __Pyx_Coroutine_status_from_result(PyObject **retval) {
    if (*retval) {
        return PYGEN_NEXT;
    } else if (likely(__Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, retval) == 0)) {
        return PYGEN_RETURN;
    } else {
        return PYGEN_ERROR;
    }
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
static void __Pyx_Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
__Pyx_PySendResult __Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, PyObject **result, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(__Pyx_Coroutine_get_is_running(self));  // Callers should ensure is_running
    if (unlikely(self->resume_label == -1)) {
        __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
        return PYGEN_ERROR;
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    retval = self->body(self, tstate, value);
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    *result = retval;
    if (self->resume_label == -1) {
        return likely(retval) ? PYGEN_RETURN : PYGEN_ERROR;
    }
    return PYGEN_NEXT;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
#define __Pyx_Coroutine_MethodReturnFromResult(gen, result, retval, iternext)\
    ((result) == PYGEN_NEXT ? (retval) : __Pyx__Coroutine_MethodReturnFromResult(gen, result, retval, iternext))
static PyObject *
__Pyx__Coroutine_MethodReturnFromResult(PyObject* gen, __Pyx_PySendResult result, PyObject *retval, int iternext) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (likely(result == PYGEN_RETURN)) {
        int is_async = 0;
        #ifdef __Pyx_AsyncGen_USED
        is_async = __Pyx_AsyncGen_CheckExact(gen);
        #endif
        __Pyx_ReturnWithStopIteration(retval, is_async, iternext);
        Py_XDECREF(retval);
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE __Pyx_PySendResult
__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen, PyObject** retval) {
    __Pyx_PySendResult result;
    PyObject *val = NULL;
    assert(__Pyx_Coroutine_get_is_running(gen));
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    result = __Pyx_Coroutine_SendEx(gen, val, retval, 0);
    Py_XDECREF(val);
    return result;
}
#if CYTHON_USE_AM_SEND
static __Pyx_PySendResult
__Pyx_Coroutine_SendToDelegate(__pyx_CoroutineObject *gen, __Pyx_pyiter_sendfunc gen_am_send, PyObject *value, PyObject **retval) {
    PyObject *ret = NULL;
    __Pyx_PySendResult delegate_result, result;
    assert(__Pyx_Coroutine_get_is_running(gen));
    delegate_result = gen_am_send(gen->yieldfrom, value, &ret);
    if (delegate_result == PYGEN_NEXT) {
        assert (ret != NULL);
        *retval = ret;
        return PYGEN_NEXT;
    }
    assert (delegate_result != PYGEN_ERROR || ret == NULL);
    __Pyx_Coroutine_Undelegate(gen);
    result = __Pyx_Coroutine_SendEx(gen, ret, retval, 0);
    Py_XDECREF(ret);
    return result;
}
#endif
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval = NULL;
    __Pyx_PySendResult result = __Pyx_Coroutine_AmSend(self, value, &retval);
    return __Pyx_Coroutine_MethodReturnFromResult(self, result, retval, 0);
}
static __Pyx_PySendResult
__Pyx_Coroutine_AmSend(PyObject *self, PyObject *value, PyObject **retval) {
    __Pyx_PySendResult result;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        *retval = __Pyx_Coroutine_AlreadyRunningError(gen);
        return PYGEN_ERROR;
    }
    #if CYTHON_USE_AM_SEND
    if (gen->yieldfrom_am_send) {
        result = __Pyx_Coroutine_SendToDelegate(gen, gen->yieldfrom_am_send, value, retval);
    } else
    #endif
    if (gen->yieldfrom) {
        PyObject *yf = gen->yieldfrom;
        PyObject *ret;
      #if !CYTHON_USE_AM_SEND
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
      #endif
        {
            #if !CYTHON_COMPILING_IN_LIMITED_API || __PYX_LIMITED_VERSION_HEX >= 0x03080000
            if (value == Py_None && PyIter_Check(yf))
                ret = __Pyx_PyIter_Next_Plain(yf);
            else
            #endif
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_mstate_global->__pyx_n_u_send, value);
        }
        if (likely(ret)) {
            __Pyx_Coroutine_unset_is_running(gen);
            *retval = ret;
            return PYGEN_NEXT;
        }
        result = __Pyx_Coroutine_FinishDelegation(gen, retval);
    } else {
        result = __Pyx_Coroutine_SendEx(gen, value, retval, 0);
    }
    __Pyx_Coroutine_unset_is_running(gen);
    return result;
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    __Pyx_PySendResult result;
    PyObject *retval = NULL;
    CYTHON_UNUSED_VAR(gen);
    assert(__Pyx_Coroutine_get_is_running(gen));
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        result = __Pyx_Coroutine_Close(yf, &retval);
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        result = __Pyx_Coroutine_Close(yf, &retval);
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        result = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf);
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
        result = PYGEN_RETURN;
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
        result = PYGEN_RETURN;
    } else
    #endif
    {
        PyObject *meth;
        result = PYGEN_RETURN;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_mstate_global->__pyx_n_u_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval)) {
                result = PYGEN_ERROR;
            }
        }
    }
    Py_XDECREF(retval);
    return result == PYGEN_ERROR ? -1 : 0;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __Pyx_PySendResult result;
    PyObject *retval = NULL;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    }
    #if CYTHON_USE_AM_SEND
    if (gen->yieldfrom_am_send) {
        result = __Pyx_Coroutine_SendToDelegate(gen, gen->yieldfrom_am_send, Py_None, &retval);
    } else
    #endif
    if (gen->yieldfrom) {
        PyObject *yf = gen->yieldfrom;
        PyObject *ret;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && (PY_VERSION_HEX < 0x030A00A3 || !CYTHON_USE_AM_SEND)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
            ret = __Pyx_PyIter_Next_Plain(yf);
        if (likely(ret)) {
            __Pyx_Coroutine_unset_is_running(gen);
            return ret;
        }
        result = __Pyx_Coroutine_FinishDelegation(gen, &retval);
    } else {
        result = __Pyx_Coroutine_SendEx(gen, Py_None, &retval, 0);
    }
    __Pyx_Coroutine_unset_is_running(gen);
    return __Pyx_Coroutine_MethodReturnFromResult(self, result, retval, 1);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    PyObject *retval = NULL;
    __Pyx_PySendResult result;
    CYTHON_UNUSED_VAR(arg);
    result = __Pyx_Coroutine_Close(self, &retval);
    if (unlikely(result == PYGEN_ERROR))
        return NULL;
    Py_XDECREF(retval);
    Py_RETURN_NONE;
}
static __Pyx_PySendResult
__Pyx_Coroutine_Close(PyObject *self, PyObject **retval) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PySendResult result;
    PyObject *yf;
    int err = 0;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        *retval = __Pyx_Coroutine_AlreadyRunningError(gen);
        return PYGEN_ERROR;
    }
    yf = gen->yieldfrom;
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    result = __Pyx_Coroutine_SendEx(gen, NULL, retval, 1);
    if (result == PYGEN_ERROR) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_Coroutine_unset_is_running(gen);
        if (!__Pyx_PyErr_Occurred()) {
            return PYGEN_RETURN;
        } else if (likely(__Pyx_PyErr_ExceptionMatches2(PyExc_GeneratorExit, PyExc_StopIteration))) {
            __Pyx_PyErr_Clear();
            return PYGEN_RETURN;
        }
        return PYGEN_ERROR;
    } else if (likely(result == PYGEN_RETURN && *retval == Py_None)) {
        __Pyx_Coroutine_unset_is_running(gen);
        return PYGEN_RETURN;
    } else {
        const char *msg;
        Py_DECREF(*retval);
        *retval = NULL;
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
            msg = "async generator ignored GeneratorExit";
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        __Pyx_Coroutine_unset_is_running(gen);
        return PYGEN_ERROR;
    }
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen)))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    yf = gen->yieldfrom;
    if (yf) {
        __Pyx_PySendResult result;
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                goto propagate_exception;
            goto throw_here;
        }
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_mstate_global->__pyx_n_u_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    __Pyx_Coroutine_unset_is_running(gen);
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        Py_DECREF(yf);
        if (ret) {
            __Pyx_Coroutine_unset_is_running(gen);
            return ret;
        }
        result = __Pyx_Coroutine_FinishDelegation(gen, &ret);
        __Pyx_Coroutine_unset_is_running(gen);
        return __Pyx_Coroutine_MethodReturnFromResult(self, result, ret, 0);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
propagate_exception:
    {
        PyObject *retval = NULL;
        __Pyx_PySendResult result = __Pyx_Coroutine_SendEx(gen, NULL, &retval, 0);
        __Pyx_Coroutine_unset_is_running(gen);
        return __Pyx_Coroutine_MethodReturnFromResult(self, result, retval, 0);
    }
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, "throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    {
        int e = __Pyx_call_type_traverse((PyObject*)gen, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    #if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    if (gen->gi_weakreflist != NULL)
    #endif
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        {
            destructor del = __Pyx_PyObject_GetSlot(gen, tp_del, destructor);
            if (del) del(self);
        }
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
#if CYTHON_USE_TP_FINALIZE
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *retval = NULL;
        __Pyx_PySendResult result = __Pyx_Coroutine_Close(self, &retval);
        if (result == PYGEN_ERROR) {
            PyErr_WriteUnraisable(self);
        } else {
            Py_XDECREF(retval);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
}
#endif
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx__Coroutine_get_frame(__pyx_CoroutineObject *self)
{
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyObject *frame;
    #if PY_VERSION_HEX >= 0x030d0000
    Py_BEGIN_CRITICAL_SECTION(self);
    #endif
    frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        PyObject *globals = PyDict_New();
        if (unlikely(!globals)) return NULL;
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            globals,                        /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        Py_DECREF(globals);
        if (unlikely(!frame))
            return NULL;
        if (unlikely(self->gi_frame)) {
            Py_DECREF(frame);
            frame = self->gi_frame;
        } else {
            self->gi_frame = frame;
        }
    }
    Py_INCREF(frame);
    #if PY_VERSION_HEX >= 0x030d0000
    Py_END_CRITICAL_SECTION();
    #endif
    return frame;
#else
    CYTHON_UNUSED_VAR(self);
    Py_RETURN_NONE;
#endif
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context) {
    CYTHON_UNUSED_VAR(context);
    PyObject *frame = self->gi_frame;
    if (frame)
        return __Pyx_NewRef(frame);
    return __Pyx__Coroutine_get_frame(self);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->yieldfrom_am_send = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_LIMITED_API
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    gen->gi_weakreflist = NULL;
#endif
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}
static char __Pyx_Coroutine_test_and_set_is_running(__pyx_CoroutineObject *gen) {
    char result;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_BEGIN_CRITICAL_SECTION(gen);
    #endif
    result = gen->is_running;
    gen->is_running = 1;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_END_CRITICAL_SECTION();
    #endif
    return result;
}
static void __Pyx_Coroutine_unset_is_running(__pyx_CoroutineObject *gen) {
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_BEGIN_CRITICAL_SECTION(gen);
    #endif
    assert(gen->is_running);
    gen->is_running = 0;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_END_CRITICAL_SECTION();
    #endif
}
static char __Pyx_Coroutine_get_is_running(__pyx_CoroutineObject *gen) {
    char result;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_BEGIN_CRITICAL_SECTION(gen);
    #endif
    result = gen->is_running;
    #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_END_CRITICAL_SECTION();
    #endif
    return result;
}
static PyObject *__Pyx_Coroutine_get_is_running_getter(PyObject *gen, void *closure) {
    CYTHON_UNUSED_VAR(closure);
    char result = __Pyx_Coroutine_get_is_running((__pyx_CoroutineObject*)gen);
    if (result) Py_RETURN_TRUE;
    else Py_RETURN_FALSE;
}
#if __PYX_HAS_PY_AM_SEND == 2
static void __Pyx_SetBackportTypeAmSend(PyTypeObject *type, __Pyx_PyAsyncMethodsStruct *static_amsend_methods, __Pyx_pyiter_sendfunc am_send) {
    Py_ssize_t ptr_offset = (char*)(type->tp_as_async) - (char*)type;
    if (ptr_offset < 0 || ptr_offset > type->tp_basicsize) {
        return;
    }
    memcpy((void*)static_amsend_methods, (void*)(type->tp_as_async), sizeof(*type->tp_as_async));
    static_amsend_methods->am_send = am_send;
    type->tp_as_async = __Pyx_SlotTpAsAsync(static_amsend_methods);
}
#endif
static PyObject *__Pyx_Coroutine_fail_reduce_ex(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    __Pyx_TypeName self_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE((PyObject*)self));
    PyErr_Format(PyExc_TypeError, "cannot pickle '" __Pyx_FMT_TYPENAME "' object",
                         self_type_name);
    __Pyx_DECREF_TypeName(self_type_name);
    return NULL;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {"__reduce_ex__", (PyCFunction) __Pyx_Coroutine_fail_reduce_ex, METH_O, 0},
    {"__reduce__", (PyCFunction) __Pyx_Coroutine_fail_reduce_ex, METH_NOARGS, 0},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {"gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     PyDoc_STR("object being iterated by 'yield from', or None")},
    {"gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {"__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {"__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     PyDoc_STR("name of the generator"), 0},
    {"__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     PyDoc_STR("qualified name of the generator"), 0},
    {"gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     PyDoc_STR("Frame of the generator"), 0},
    {"gi_running", __Pyx_Coroutine_get_is_running_getter, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) PyObject_GenericGetAttr},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    {Py_am_send, (void *)__Pyx_Coroutine_AmSend},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_WEAKREF |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | __Pyx_TPFLAGS_HAVE_AM_SEND,
    __pyx_GeneratorType_slots
};
#if __PYX_HAS_PY_AM_SEND == 2
static __Pyx_PyAsyncMethodsStruct __pyx_Generator_as_async;
#endif
static int __pyx_Generator_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_GeneratorType_spec, NULL);
    if (unlikely(!mstate->__pyx_GeneratorType)) {
        return -1;
    }
#if __PYX_HAS_PY_AM_SEND == 2
    __Pyx_SetBackportTypeAmSend(mstate->__pyx_GeneratorType, &__pyx_Generator_as_async, &__Pyx_Coroutine_AmSend);
#endif
    return 0;
}
static PyObject *__Pyx_Generator_GetInlinedResult(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *retval = NULL;
    if (unlikely(__Pyx_Coroutine_test_and_set_is_running(gen))) {
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    }
    __Pyx_PySendResult result = __Pyx_Coroutine_SendEx(gen, Py_None, &retval, 0);
    __Pyx_Coroutine_unset_is_running(gen);
    (void) result;
    assert (result == PYGEN_RETURN || result == PYGEN_ERROR);
    assert ((result == PYGEN_RETURN && retval != NULL) || (result == PYGEN_ERROR && retval == NULL));
    return retval;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
    if (likely(result))
        result->_co_firsttraceable = 0;
    #endif
    return result;
  }
#elif !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
        Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(line_table_length == -1)) goto done;
        #endif
        Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
